/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/DrawAxis.c,v 1.2 1994/05/26 19:48:53 vs Exp vs $ */

#include <stdio.h>
#include <ctype.h>

#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCal.h>
#include <tv/vsTools.h>

#include "DrawAxis.h"

#define exp_10(x) (pow ((double)10, ((double)x)))


/* ---------------------------------------------------------------------- */
typedef struct {
  double pos;
  double inc;
  int canonic;
  int n_inc;
  int l10;
  int siz;
  int p_pos;
} ScaleLblRec, *ScaleLbl;

typedef struct {
  VecCorePart core;
  ScaleLblRec *p;
} ScaleLblVecRec, *ScaleLblVec_t;


typedef struct {
  Cal_t cal;
  double (*cal_l)();
  double (*cal_s)();
  int long_tic, medium_tic, short_tic, space;
  int fontsize;
  char justify;
  int lbl_dist, tic_dist;
  int (*sprintlabel)();
  void (*drawtic)();
  void (*drawlbl)();
  GowScale scale;
  GowPixelFunc px;
  int frame;
  int x, y;
  int minLimit;
  int maxLimit;
} AxisArgs;

typedef struct {
  int step, old_step;
  int first, last;
  int canonic;
  int min_distance;
} CheckArgsRec, *CheckArgs;

#define Lbls(i)			VecP(lbls)[i]
#define FMTMAX(l10)		((l10 > 0) ? 0 : ((l10 < -12) ? 12 : -l10))
#define FMT(l10)		((l10 > 0) ? 0 : -l10)
/* ---------------------------------------------------------------------- */
/* ARGSUSED */
static int SprintLabelX(text, lbl, lmt)
     char *text;
     ScaleLbl lbl;
     int lmt;
{
  char fmt[64];
  int fl= FMT(lbl->l10);
  sprintf(fmt, "%%.%df", fl);
  sprintf(text, fmt, lbl->pos);
  return strlen(text);
}  
/* ---------------------------------------------------------------------- */
static int SprintLabelY(text, lbl, lmt)
     char *text;
     ScaleLbl lbl;
     int lmt;
{
  char fmt[64];
  int fl= FMT(lbl->l10);
  sprintf(fmt, "%%.%df", fl);
  sprintf(text, fmt, lbl->pos);
  if (strlen(text) > lmt) {
    if (lbl->inc >= 1000.0) {
      sprintf(text, "%.0fe%d", lbl->pos / lbl->inc, (int)log10(lbl->inc)); }
    else {
      double pos= ABS(lbl->pos);
      if (pos < .001) {
	double l10= floor(log10(pos));
	fl= FMT(lbl->l10) + NINT(l10);
	if (fl > 12) { fl= 12; }
	sprintf(fmt, "%%.%dfe%%.0f", fl);
	sprintf(text, fmt, lbl->pos / exp_10(l10), l10); } }
    if (strlen(text) > lmt) { *text= '\0'; } }
  return 1;
}  
/* ---------------------------------------------------------------------- */
void DrawTicAxisX(seg, x, y, pos, tic)
     XSegment *seg;
     int x, y;
     int pos;
     int tic;
{
  seg->x1= seg->x2= x + pos; seg->y1= y; seg->y2= y + tic;
}
/* ---------------------------------------------------------------------- */
void DrawTicAxisY(seg, x, y, pos, tic)
     XSegment *seg;
     int x, y;
     int pos;
     int tic;
{
  seg->x1= x; seg->x2= x + tic; seg->y1= seg->y2= y + pos;
}
/* ---------------------------------------------------------------------- */
void DrawLblAxisX(d, x, y, pos, space, justify, justify_o, lbl)
     GowDestination d;
     int x, y;
     int pos;
     int space;
     int justify, justify_o;
     char *lbl;
{
  if (*lbl) {
    GowDMove(d, x + pos, y + space);
    GowDAlabel(d, justify_o, justify, lbl); }
}
/* ---------------------------------------------------------------------- */
void DrawLblAxisY(d, x, y, pos, space, justify, justify_o, lbl)
     GowDestination d;
     int x, y;
     int pos;
     int space;
     int justify, justify_o;
     char *lbl;
{
  if (*lbl) {
    GowDMove(d, x + space, y + pos);
    GowDAlabel(d, justify, justify_o, lbl); }
}
/* ---------------------------------------------------------------------- */
void LabelInit(lbls, args)
     ScaleLblVec_t lbls;
     AxisArgs *args;
{
  Cal_t cal= args->cal;
  GowScale scale= args->scale;
  double l_min, l_max;

  { double u_min, u_max;
    GowGetScaleUserRange(scale, &u_min, &u_max);
    l_min= args->cal_l(cal, u_min);
    l_max= args->cal_l(cal, u_max); }

  if (l_min > l_max) {
    double tmp= l_min; l_min= l_max; l_max= tmp;
    args->minLimit= GowToLarge;
    args->maxLimit= GowToSmall; }
  else {
    args->minLimit= GowToSmall;
    args->maxLimit= GowToLarge; }

  VecInsert(lbls, 0, 2);
  Lbls(0).siz= Lbls(1).siz= 0;
  Lbls(0).canonic= Lbls(1).canonic= TRUE;

  if (GowGetScaleType(scale) == GowLogarithmicScale) {
    if (l_min <= 0.0) { Lbls(0).n_inc= 0; return; }
    else {
      double e_min= floor(log10(l_min));
      double e_max= ceil(log10(l_max));
      if (e_max - e_min >= 1.0) {
	Lbls(0).inc= Lbls(0).pos= exp_10(e_min);
	Lbls(0).l10= e_min;
	Lbls(0).n_inc= 1;
	Lbls(1).inc= Lbls(1).pos= exp_10(e_max);
	Lbls(1).l10= e_max;
	return; } } }

  { double inc, n_inc= 0;
    double e, r= l_max - l_min;
    int l10;
    if (r == 0.0) { Lbls(0).n_inc= 0; return; }
    e= ceil(log10(r)); inc= exp_10(e); l10= NINT(e);


    Lbls(0).pos= l_min - fmod(l_min, inc);
    while (Lbls(0).pos >= l_min) { Lbls(0).pos -= inc; }

    Lbls(1).pos= Lbls(0).pos;
    while (Lbls(1).pos <= l_max) { Lbls(1).pos += inc; n_inc++; }

    Lbls(0).p_pos= GowGetScaleLimitPixel(scale, args->minLimit);
    Lbls(1).p_pos= GowGetScaleLimitPixel(scale, args->maxLimit);

    Lbls(0).inc= inc;
    Lbls(0).n_inc= n_inc;
    Lbls(0).l10= l10; }
}
/* ---------------------------------------------------------------------- */
static Boolean CheckNeighbours(tmpLbls, lbls, i, i_step, i0, i1, cargs)
     ScaleLbl tmpLbls;
     ScaleLblVec_t lbls;
     int i, i_step, i0, i1;
     CheckArgs cargs;
{
  if (cargs->canonic) {
    int di= i_step - i;
    tmpLbls[i].n_inc= ABS(di); }

  if (tmpLbls[i].siz && tmpLbls[i_step].siz) {
    int dp= tmpLbls[i].p_pos - tmpLbls[i_step].p_pos;
    if (cargs->min_distance > ABS(dp)) {
      if (!cargs->canonic) { return FALSE; }
      switch (cargs->step) { /* divide canonically */
      case 1:
	cargs->step= 2;
	if (cargs->first - 1 + cargs->step <= cargs->last) { break; }
      case 2:
	cargs->step= 5;
	if (cargs->first - 1 + cargs->step <= cargs->last) { break; }
      default:
	cargs->canonic= FALSE;
	if (tmpLbls[cargs->first-1].siz && tmpLbls[cargs->last+1].siz) {
	  int c= (tmpLbls[cargs->first-1].p_pos +
		  tmpLbls[cargs->last+1].p_pos) / 2;
	  int dc= (tmpLbls[cargs->first-1].p_pos -
		   tmpLbls[cargs->last+1].p_pos);
	  dc= ABS(dc);
	  { int l; for (l= cargs->first; l<= cargs->last; l++) {
	    int d= tmpLbls[l].p_pos - c;
	    d= ABS(d);
	    if (d >= dc) { l--; break; }
	    else { dc= d; } }
	    tmpLbls[l-1].n_inc= l - cargs->first + 1;
	    tmpLbls[l].n_inc= cargs->last - l + 1;
	    cargs->last= l; } }
	else if (tmpLbls[cargs->last+1].siz) {
	  tmpLbls[cargs->first-1].n_inc= 1;
	  tmpLbls[cargs->first].n_inc= cargs->last - cargs->first + 1;
	  cargs->last= cargs->first; }
	else {
	  tmpLbls[cargs->last-1].n_inc= cargs->last - cargs->first + 1;
	  tmpLbls[cargs->last].n_inc= 1; }
	cargs->first= cargs->last; cargs->step= 1;
	cargs->old_step= cargs->step+1;
	tmpLbls[cargs->first-1].pos= Lbls(i0).pos;
	tmpLbls[cargs->first-1].siz= Lbls(i0).siz;
	tmpLbls[cargs->first-1].p_pos= Lbls(i0).p_pos;
	tmpLbls[cargs->last+1].pos= Lbls(i1).pos;
	tmpLbls[cargs->last+1].siz= Lbls(i1).siz;
	tmpLbls[cargs->last+1].p_pos= Lbls(i1).p_pos;
	break; } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int LabelInsert(dest, lbls, args, i0, i1)
     GowDestination dest;
     ScaleLblVec_t lbls;
     AxisArgs *args;
     int i0, i1;
{
  double inc0, inc, pos;
  int l10;
  int siz;
  char text[11][256];
  ScaleLblRec tmpLbls[11];

  CheckArgsRec cargs;

  if (Lbls(i0).inc == 0) { return FALSE; }

  switch (GowGetScaleType(args->scale)) {
  case GowLogarithmicScale:
    if (Lbls(i0).pos == Lbls(i0).inc &&
	Lbls(i1).pos == Lbls(i1).inc) {
      if (Lbls(i1).l10 - Lbls(i0).l10 > 1) {
	cargs.first= cargs.last= 1;
	l10= (Lbls(i1).l10 + Lbls(i0).l10) / 2;
	inc= exp_10(l10);
	inc0= Lbls(i0).inc;
	pos= 0.0;
	break; }
      else {
	cargs.first= 2;
	cargs.last= 9;
	l10= Lbls(i0).l10;
	inc0= inc= Lbls(i0).inc;
	pos= 0.0;
	break; } }
  default:
    cargs.first= 1;
    if (Lbls(i0).n_inc == 1) {
      cargs.last= 9;
      l10= Lbls(i0).l10 - 1;
      inc0= inc= Lbls(i0).inc * 0.1; }
    else {
      cargs.last= Lbls(i0).n_inc - 1;
      if (!Lbls(i1).siz || Lbls(i0).l10 <= Lbls(i1).l10) {
	l10= Lbls(i0).l10;
	inc0= inc = Lbls(i0).inc; }
      else {
	l10= Lbls(i1).l10;
	inc0= inc = Lbls(i1).inc; } }
    pos= Lbls(i0).pos;
    break; }

  { int i;
    int minLimit= args->minLimit;
    int maxLimit= args->maxLimit;
    GowPixelFunc px= args->px;
    Cal_t cal= args->cal;
    GowScale scale= args->scale;
    double (*cal_s)()= args->cal_s;
    int (*sprintlabel)()= args->sprintlabel;
    for (i= cargs.first; i< cargs.last+1; i++) {
      GowScaleConversion conv;
      tmpLbls[i].pos= pos + inc * i;
      tmpLbls[i].inc= inc;
      tmpLbls[i].l10= l10;
      tmpLbls[i].n_inc= 1;
      conv= px(scale, &tmpLbls[i].p_pos, cal_s(cal, tmpLbls[i].pos));
      if (conv == minLimit) {
	cargs.first= i+1;
	tmpLbls[i].siz= 0;
	Lbls(i0).pos= tmpLbls[i].pos;
	Lbls(i0).inc= inc;
	Lbls(i0).l10= l10;
	Lbls(i0).n_inc= cargs.last - i + 1;
	Lbls(i0).canonic= (Lbls(i0).n_inc == 1);
	if (cargs.first == cargs.last + 1 && Lbls(i1).siz) {
	  int dp= (Lbls(i1).p_pos -
		   GowGetScaleLimitPixel(scale, minLimit));
	  return (ABS(dp) > args->fontsize * args->lbl_dist); } }
      else if (conv == maxLimit) {
	cargs.last= i-1;
	tmpLbls[i].siz= 0;
	Lbls(i1).pos= tmpLbls[i].pos;
	Lbls(i0).inc= inc;
	Lbls(i0).n_inc= i;
	if (Lbls(i0).pos == 0.0) { Lbls(i0).l10--; }
	if (cargs.first > cargs.last) {
	  int dp= (Lbls(i0).p_pos -
		   GowGetScaleLimitPixel(scale, maxLimit));
	  return (ABS(dp) > args->fontsize * args->lbl_dist); } }
      else {
	tmpLbls[i].siz= sprintlabel(text[i], tmpLbls + i, args->frame); } } }

  tmpLbls[cargs.first-1].n_inc	= 1;
  tmpLbls[cargs.first-1].pos	= Lbls(i0).pos;
  tmpLbls[cargs.first-1].siz	= Lbls(i0).siz;
  tmpLbls[cargs.first-1].p_pos	= Lbls(i0).p_pos;

  tmpLbls[cargs.last+1].pos	= Lbls(i1).pos;
  tmpLbls[cargs.last+1].siz	= Lbls(i1).siz;
  tmpLbls[cargs.last+1].p_pos	= Lbls(i1).p_pos;

  cargs.canonic= TRUE;
  cargs.step= 1;

  { int lbl_dist;
    int new, i_start;
    lbl_dist= MAX(Lbls(i0).siz, args->lbl_dist);
    lbl_dist= MAX(Lbls(i1).siz, lbl_dist);

    if (Lbls(i0).canonic || !Lbls(i1).canonic) {
      do {
	cargs.old_step= cargs.step;

	siz= lbl_dist;
    	{ int i;
	  for (i= cargs.first - 1 + cargs.step;
	       i <= cargs.last; i += cargs.step) {
	    siz= MAX(tmpLbls[i].siz, siz); } }
	
	{ int i, i_step;

	  cargs.min_distance= siz * args->fontsize;

	  for (i= cargs.first - 1, i_step= i + cargs.step;
	       cargs.step == cargs.old_step && i<= cargs.last;
	       i= i_step, i_step += cargs.step) {
	    if (i_step > cargs.last+1) { i_step= cargs.last+1; }

	    if (!CheckNeighbours(tmpLbls, lbls, i, i_step, i0, i1, &cargs)) {
	      return FALSE; } } } /* i < i_step !!! */
      } while (cargs.step != cargs.old_step);

      new= (cargs.last - cargs.first + 1) / cargs.step;
      i_start= cargs.first - 1 + cargs.step; }
    else {
      do {
	cargs.old_step= cargs.step;

	siz= lbl_dist;
    	{ int i;
	  for (i= cargs.last + 1 - cargs.step;
	       i >= cargs.first; i -= cargs.step) {
	    siz= MAX(tmpLbls[i].siz, siz); } }

	{ int i, i_step;

	  cargs.min_distance= siz * args->fontsize;

	  for (i= cargs.last + 1, i_step= i - cargs.step;
	       cargs.step == cargs.old_step && i>= cargs.first;
	       i= i_step, i_step -= cargs.step) {
	    if (i_step < cargs.first - 1) { i_step= cargs.first - 1; }
	  
	    if (!CheckNeighbours(tmpLbls, lbls, i_step, i, i0, i1, &cargs)) {
	      return FALSE; } } } /* i > i_step !!! */
      } while (cargs.step != cargs.old_step);

      new= (cargs.last - cargs.first + 1) / cargs.step;
      i_start= cargs.last+1 - new * cargs.step; }

    if (!new || !VecInsert(lbls, i0 + 1, new)) { return FALSE; }

    { int i, li;

      i1 += new;

      for (li= (i0 + 1), i= i_start; li < i1; i += cargs.step, li++) {
	Lbls(li).pos= tmpLbls[i].pos;
	Lbls(li).n_inc= tmpLbls[i].n_inc;
	Lbls(li).canonic= (tmpLbls[i].n_inc == 1);
	Lbls(li).l10= l10;
	Lbls(li).inc= inc;
	Lbls(li).p_pos= tmpLbls[i].p_pos;
	Lbls(li).siz= siz;
	args->drawlbl(dest, args->x, args->y, tmpLbls[i].p_pos,
		      args->space, args->justify, 'c', text[i]); } }
    Lbls(i0).n_inc= tmpLbls[cargs.first-1].n_inc;
    Lbls(i0).canonic= Lbls(i0).canonic || (Lbls(i0).n_inc == 1);
    Lbls(i0).inc= inc0; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static XSegment *DrawTics(tic, first, last, lbl, args, segs, canonic)
     int tic[];
     int first, last, lbl;
     AxisArgs *args;
     XSegment *segs;
     int canonic;
{
  XSegment *seg= segs;
  int step= 1, old_step= step, min_distance= args->tic_dist;

  do {
    int i, i_step;

    old_step= step;
    for (i= 0, i_step= i + step; step && i< last;
	 i = i_step, i_step += step) {
      if (i_step > last) { i_step= last; }
      if (i >= first) {
	int dp= tic[i] - tic[i_step];
	if (ABS(dp) < min_distance) {
	  switch (step) {
	  case 1: if (canonic || lbl != last || last == 10) { step= 2; break; }
	  case 2: if (canonic || lbl != last || last == 10) { step= 5; break; }
	  case 5: old_step= step= 0; break; } } } }
  } while (step != old_step);

  if (step && step != 1) {
    if (lbl == last && lbl != 10 && !((last - first) % step)) { step= 0; } }

  if (!first) { first++; }

  { int i;
    int x= args->x, y= args->y;
    void (*drawtic)()= args->drawtic;

    if (step) {
      for (i= 0; i<= last; i += step) {
	if (i >= first) {
	  if (i == lbl) { continue; }
	  else if (i == 5) { drawtic(seg, x, y, tic[i], args->medium_tic); }
	  else { drawtic(seg, x, y, tic[i], args->short_tic); }
	  seg++; } } }
    if (last == lbl) {
      drawtic(seg, x, y, tic[lbl], args->long_tic); seg++; } }

  return seg;
}
/* ---------------------------------------------------------------------- */
static XSegment *TicInsert(lbls, args, segs, i0, i1)
     ScaleLblVec_t lbls;
     AxisArgs *args;
     XSegment *segs;
     int i0, i1;
{
  GowScale scale= args->scale;
  GowPixelFunc px= args->px;
  double (*cal_s)()= args->cal_s;
  Cal_t cal= args->cal;
  XSegment *seg= segs;
  int tic[11];
  int first, last;

  switch (GowGetScaleType(args->scale)) {
  case GowLogarithmicScale: {
    if (Lbls(i0).pos == Lbls(i0).inc &&
	(Lbls(i1).pos == Lbls(i1).inc ||
	 Lbls(i1).pos == Lbls(i0).pos * 10.0)) {
      int l10;
      double pos= Lbls(i0).pos;
      for (l10= Lbls(i0).l10; l10< Lbls(i1).l10; l10++, pos *= 10.0) {
	first= 1; last= 10;
	{ int i;
	  int minLimit= args->minLimit;
	  int maxLimit= args->maxLimit;
	  for (i= first; i<=last; i++) {
	    GowScaleConversion conv
	      = px(scale, tic + i, cal_s(cal, pos * (double)i));
	    if (conv == minLimit) { first= i + 1; }
	    else if (conv == maxLimit) { last= i - 1; } } }
	seg= DrawTics(tic, first, last, 10, args, seg, TRUE); }
      return seg; } }
  case GowLinearScale:
  default: {
    double inc;
    { int n_inc;

      if (Lbls(i0).n_inc == 1) { inc= Lbls(i0).inc * 0.1; n_inc= 10; }
      else { inc= Lbls(i0).inc; n_inc= Lbls(i0).n_inc; }

      if (!Lbls(i1).canonic || Lbls(i0).canonic || Lbls(i0).n_inc == 1) {
	first= 0; last= n_inc; }
      else { last= 10; first= 10 - n_inc; } }

    { int i;
      double pos;

      for (i= first, pos= Lbls(i0).pos; i<= last; i++, pos += inc) {
	int minLimit= args->minLimit;
	int maxLimit= args->maxLimit;
	GowScaleConversion conv= px(scale, tic + i, cal_s(cal, pos));
	if (conv == minLimit) { first= i + 1; }
	else if (conv == maxLimit) { last= i - 1; }
	else {
	  if (Lbls(i0).siz && tic[i] == Lbls(i0).p_pos) { first= i; }
	  if (Lbls(i1).siz && tic[i] == Lbls(i1).p_pos) { last= i; } } } }

    return DrawTics(tic, first, last, (Lbls(i1).siz ? last : last+1),
		    args, seg, (Lbls(i0).n_inc == 1)); } }
}
/* ---------------------------------------------------------------------- */
Boolean DrawAxis(dest, x, y, axis, type, scale, tic_dim, tic_fac, frame)
     GowDestination dest;
     int x, y;
     DrawAxis_t axis;
     int type;
     GowScale scale;
     float *tic_dim;
     int tic_fac;
     int frame;
{
  AxisArgs args_rec, *args= &args_rec;
  ScaleLblVecRec lbls_rec, *lbls= &lbls_rec;
  XSegment *segs, *seg;
  int p_min, p_max, fs;
  char *axis_name;
  char *unit;
  int ulen, upos;
  double uspc;
  char ujust, ujust_o;

  if (!dest || !axis || !scale) { return FALSE; }

  VecInit(lbls, sizeof(ScaleLblRec));

  GowDSet_gc(dest, DrawAxisScaleGC);
  GowGetScalePixelRange(scale, &p_min, &p_max);

  args->scale= scale;
  args->px= GowGetScalePixelFunction(scale);
  args->cal= DrawAxisCal(axis);
  args->cal_l= DrawAxisCalL(axis);
  args->cal_s= DrawAxisCalS(axis);
  args->fontsize= fs= MAX(GowDFontwidth(dest), GowDFontheight(dest));
  args->lbl_dist= NINT(tic_dim[TIC_LABEL]);
  args->tic_dist= NINT(tic_dim[TIC_TIC] * fs);

  unit= DrawAxisUnit(axis) ? DrawAxisUnit(axis)(args->cal) : NULL;

  { int range= p_max - p_min;
    int tic_num= range / (fs * tic_dim[TIC_TIC]) + 8;
    int lbl_num= range / (fs * tic_dim[TIC_LABEL]) + 8;

    if (!(seg= segs= (XSegment *)Malloc(tic_num * sizeof(XSegment)))) {
      return FALSE; }
    else if (!VecAlloc(lbls, lbl_num)) { Free(segs); return FALSE; } }

  { double fac;
    int offset, range, linewidth, space;

    linewidth= MAX(GowDLinewidth(dest), 1);
    space= NINT(tic_dim[TIC_SPACE] * (double)fs);
    if (GowGetScaleDirection(scale) == GowNormalScaleDirection) {
      offset= GowGetScaleLimitPixel(scale, GowToSmall);
      range= GowGetScaleLimitPixel(scale, GowToLarge);
      upos= range; }
    else {
      offset= GowGetScaleLimitPixel(scale, GowToLarge);
      range= GowGetScaleLimitPixel(scale, GowToSmall);
      upos= offset; }
    range += linewidth - 1;

    switch (type) {
    case DrawAxisLeft:
      args->sprintlabel= SprintLabelY;
      args->drawtic= DrawTicAxisY;
      args->drawlbl= DrawLblAxisY;
      fac= -1;
      args->justify= 'r';
      seg->y1= y - linewidth;
      args->y= y -= offset;
      seg->x1= seg->x2= args->x= x;
      seg->y2= y + range + linewidth;
      args->frame= (frame - space) / GowDFontwidth(dest);
      axis_name= "left-axis";
      ulen= 1;
      ujust= 'r';
      ujust_o= 'b';
      uspc= 2.0 * tic_dim[TIC_SPACE];
      break;
    case DrawAxisRight:
      args->sprintlabel= SprintLabelY;
      args->drawtic= DrawTicAxisY;
      args->drawlbl= DrawLblAxisY;
      fac= 1;
      args->justify= 'l';
      seg->y1= y - linewidth;
      args->y= y -= offset;
      seg->x1= seg->x2= args->x= x;
      seg->y2= y + range + linewidth;
      args->frame= (frame - space) / GowDFontwidth(dest);
      axis_name= "right-axis";
      ulen= 1;
      ujust= 'l';
      ujust_o= 'b';
      uspc= 2.0 * tic_dim[TIC_SPACE];
      break;
    case DrawAxisBottom:
      args->sprintlabel= SprintLabelX;
      args->drawtic= DrawTicAxisX;
      args->drawlbl= DrawLblAxisX;
      fac= -1;
      args->justify= 't';
      seg->x1= x - linewidth;
      args->x= x -= offset;
      seg->x2= x + range + linewidth;
      seg->y1= seg->y2= args->y= y;
      args->frame= 0;
      axis_name= "bottom-axis";
      ulen= unit ? strlen(unit) : 0;
      ujust= 'r';
      ujust_o= 't';
      uspc= (2.0 * tic_dim[TIC_SPACE] + 1.0);
      break;
    case DrawAxisTop:
      args->sprintlabel= SprintLabelX;
      args->drawtic= DrawTicAxisX;
      args->drawlbl= DrawLblAxisX;
      fac= 1;
      args->justify= 'b';
      seg->x1= x - linewidth;
      args->x= x -= offset;
      seg->x2= x + range + linewidth;
      seg->y1= seg->y2= args->y= y;
      args->frame= 0;
      axis_name= "top_axis";
      ulen= unit ? strlen(unit) : 0;
      ujust= 'r';
      ujust_o= 'b';
      uspc= (2.0 * tic_dim[TIC_SPACE] + 1.0);
      break;
    default: Free(segs); VecFree(lbls);  return FALSE; }
    seg++;

    fac *= fs * tic_fac;
    uspc *= fac;
    { double t= tic_dim[TIC_LONG] * fac; args->long_tic= NINT(t); }
    { double t= tic_dim[TIC_MEDIUM] * fac; args->medium_tic= NINT(t); }
    { double t= tic_dim[TIC_SHORT] * fac; args->short_tic= NINT(t); }
    { double t= tic_dim[TIC_SPACE] * fac; args->space= NINT(t); } }

  GowDStart_object(dest, axis_name);

  if (frame) {
    LabelInit(lbls, args);

    { int i0, i1;
      for (i0= 0, i1= 1; i1 < VecNum(lbls); i1= i0 + 1) {
	if (!LabelInsert(dest, lbls, args, i0, i1)) { i0= i1; } } }
    { int l;
      for (l= 0; l+1 < VecNum(lbls); l++) {
	seg= TicInsert(lbls, args, seg, l, l+1); } }
    if (unit && (ulen + 2 * tic_dim[TIC_LONG]) < frame) {
      args->drawlbl(dest, args->x, args->y, upos,
		    NINT(uspc), ujust_o, ujust, unit); } }
  GowDSegments(dest, segs, seg - segs);

  GowDEnd_object(dest, axis_name);

  Free(segs); VecFree(lbls);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
