/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/DrawFit.c,v 1.3 1994/12/02 21:22:00 vs Exp vs $ */

#include <stdio.h>
#include <ctype.h>

#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCal.h>
#include <tv/vsFit.h>

#include "DrawFit.h"

static Boolean GetRange();

/*	draws from PIXELmin -> PIXELmax ,CHNmin -> CHNmax 
 *	in each pixel column is drawn only one channel;
 *	all other channels which fall into the same column are omitted
 */


Boolean DrawFit(dest, dfit, x_scale, y_scale,
		step_width, min_steps,
		calibrated, modification, normalized)
     GowDestination dest;
     DrawFit_t dfit;
     GowScale x_scale;
     GowScale y_scale;
     int step_width;
     int min_steps;
     int calibrated;
     GowScaleModificationType modification;
     int normalized;
{
  int (*setup)()= DrawFitFctSetup(dfit);
  int (*finish)()= DrawFitFctFinish(dfit);
  double (*fct)()= DrawFitFct(dfit);
  Cal_t cal= DrawFitCal(dfit);
  Fit_t fit= DrawFitFit(dfit);
  Boolean visible= FALSE;

  GowUserFunc ux;
  GowPixelFunc px, py;
  int xp_min, xp_max, yp_min, yp_max;
  int x_start, x_end;
  double xu_min, xu_max, yu_min, yu_max;
  XPoint *pnts;
  double chn_min, chn_max;
  double (*cal_s)();
  double (*cal_c)();
  double (*cal_efc)();
  double nrm= normalized ? *DrawFitNorm(dfit) : 1.0;

  if (calibrated) { cal_s= CalP; cal_c= CalC; }
  else { cal_s= CalId; cal_c= CalId; }
  switch (modification) {
  case GowMmultiplied: cal_efc= cal_c; break;
  case GowMefficiency: cal_efc= CalEfc; break;
  case GowMnone:
  default: cal_efc= CalUnity; break; }

  if (!GetRange(DrawFitTyp(dfit), fit, &chn_min, &chn_max)) {
    return FALSE; }

  GowDSet_gc(dest, DrawFitGCIdx(dfit));

  ux= GowGetScaleUserFunction(x_scale);
  px= GowGetScalePixelFunction(x_scale);
  py= GowGetScalePixelFunction(y_scale);
  GowGetScaleUserRange(x_scale, &xu_min, &xu_max);
  GowGetScaleUserRange(y_scale, &yu_min, &yu_max);
  xp_min= GowGetScaleLimitPixel(x_scale, GowClipSmall);
  xp_max= GowGetScaleLimitPixel(x_scale, GowClipLarge);
  yp_min= GowGetScaleLimitPixel(y_scale, GowClipSmall);
  yp_max= GowGetScaleLimitPixel(y_scale, GowClipLarge);

  chn_min= (double)NINT(chn_min);
  chn_max= (double)NINT(chn_max);

  { double tmp= cal_s(cal, chn_min - 0.5);

    switch (px(x_scale, &x_start, tmp)) {
    case GowClipSmall:
    case GowToSmall:	x_start= xp_min; break;
    case GowValid:	break;
    default:		return FALSE; } }

  { double tmp= cal_s(cal, (double)chn_max + 0.5);

    switch (px(x_scale, &x_end, tmp)) {
    case GowClipSmall:
    case GowToSmall:	return FALSE;
    case GowValid:	break;
    default:		x_end= xp_max; break; } }

  if ((x_end - x_start + 1) / step_width < min_steps) { return FALSE; }

  if (!setup(fit, FitPar(fit))) { return FALSE; }

  if (!(pnts= (XPoint *)Malloc(sizeof(XPoint) * 2 * (xp_max - xp_min + 4)))) {
    finish(fit); return FALSE; }

  GowDStart_object(dest, "fit");

  { XPoint *pnt= pnts;

    { int x;
      int x_prev= -1;
      int y_prev= -1;		/* undefined */
      GowScaleConversion y_prevc= GowToSmall;

      for (x= x_start; x <= x_end; x += step_width) {
	double x_user;
	GowScaleConversion yconv;
	int y;

	ux(x_scale, x, &x_user);
	x_user= cal_c(cal, x_user);

	yconv= py(y_scale, &y, nrm * cal_efc(cal, x_user) * fct(fit, x_user));

	switch (yconv) {	/* handle clipped as out of bounds */
	case GowValid: break;
	case GowClipSmall: yconv= GowToSmall;
	case GowToSmall: y= yp_min; break;
	case GowClipLarge: yconv= GowToLarge;
	case GowToLarge: y= yp_max; break; }

	if (y_prev >= 0 &&	/* defined */
	    (y_prevc == GowValid || yconv != y_prevc)) { /* visible */
	  pnt->x= x_prev; pnt->y= y_prev; pnt++; } /* 1st point */

	if (yconv == GowValid) {
	  pnt->x= x; pnt->y= y; pnt++;
	  y_prev= -1; }
	else {
	  if (y_prevc != yconv) { /* last point */
	    pnt->x= x; pnt->y= y; pnt++; }

	  if (pnt - pnts > 1 && GowDLines(dest, pnts, pnt - pnts)) {
	    visible= TRUE; }
	  pnt= pnts;

	  x_prev= x;		/* not yet drawn */
	  y_prev= y;
	  y_prevc= yconv; } }

      if (pnt - pnts > 1 && GowDLines(dest, pnts, pnt - pnts)) {
	visible= TRUE; } } }

  finish(fit);

  GowDEnd_object(dest, "decomposition");

  Free(pnts);
  return visible;
}

void DrawFitExpand(dfit,
		   x_scale, x_type, x_full,
		   y_scale, y_type, y_full,
		   calibrated, modification, normalized)
     DrawFit_t dfit;
     GowScale x_scale;
     GowScaleType x_type;
     int x_full;
     GowScale y_scale;
     GowScaleType y_type;
     int y_full;
     int calibrated;
     GowScaleModificationType modification;
     int normalized;
{
  Cal_t cal= DrawFitCal(dfit);
  Fit_t fit= DrawFitFit(dfit);
  double x_offset, x_range;
  double chn_min, chn_max;
  double (*cal_s)();
  double (*cal_c)();
  double (*cal_efc)();
  double nrm= normalized ? *DrawFitNorm(dfit) : 1.0;

  if (calibrated) { cal_s= CalP; cal_c= CalC; }
  else { cal_s= CalId; cal_c= CalId; }
  switch (modification) {
  case GowMmultiplied: cal_efc= cal_c; break;
  case GowMefficiency: cal_efc= CalEfc; break;
  case GowMnone:
  default: cal_efc= CalUnity; break; }

  if (!GetRange(DrawFitTyp(dfit), fit, &chn_min, &chn_max)) {
    return; }

  if (x_full) {
    x_offset	= cal_s(cal, chn_min - 0.5);
    x_range	= cal_s(cal, chn_max + 0.5) - x_offset;
    GowExtendNewScaleOffsetRange(x_scale, x_offset, x_range); }
  else if (!GowGetScaleOffset(x_scale, &x_offset) ||
	   !GowGetScaleRange(x_scale, &x_range)) { return; }
  if (y_full) {
    { double tmp;
      tmp= cal_c(cal, x_offset);
      chn_min= MAX(chn_min, tmp);
      tmp= cal_c(cal, x_offset + x_range); 
      chn_max= MIN(chn_max, tmp); }

    if (chn_min >= chn_max) { return; }
    else if (y_type == GowLogarithmicScale) {
      double y_min= 0.0, y_max= 0.0;
      Boolean found= FALSE;
      double chn;
      double (*fct)()= DrawFitFct(dfit);
      int (*setup)()= DrawFitFctSetup(dfit);
      int (*finish)()= DrawFitFctFinish(dfit);

      if (!setup(fit, FitPar(fit))) { return; }

      for (chn= chn_min, chn_max += 0.25; chn< chn_max; chn += 0.5) {
	double tmp= fct(fit, chn) * nrm * cal_efc(cal, chn);

	if (tmp > 0) {
	  if (found) {
	    if (tmp < y_min) { y_min= tmp; }
	    else if (tmp > y_max) { y_max= tmp; } }
	  else { y_min= y_max= tmp; found= TRUE; } } }
      if (found) {
	GowExtendNewScaleOffsetRange(y_scale, y_min, y_max - y_min); }
      finish(fit);
      return; }
    else {
      double y_min, y_max;
      double chn;
      double (*fct)()= DrawFitFct(dfit);
      int (*setup)()= DrawFitFctSetup(dfit);
      int (*finish)()= DrawFitFctFinish(dfit);

      if (!setup(fit, FitPar(fit))) { return; }

      y_min= y_max= fct(fit, chn_min) * nrm * cal_efc(cal, chn_min);
      for (chn= chn_min, chn_max += 0.25; chn< chn_max; chn += 0.5) {
	double tmp= fct(fit, chn) * nrm * cal_efc(cal, chn);

	if (tmp < y_min) { y_min= tmp; }
	else if (tmp > y_max) { y_max= tmp; } }
      GowExtendNewScaleOffsetRange(y_scale, y_min, y_max - y_min);
      finish(fit);
      return; } }
  return;
}

static Boolean GetRange(type, fit, chn_min, chn_max)
     GowFitType type;
     Fit_t fit;
     double *chn_min;
     double *chn_max;
{
  PVec_t mr= FitRegMrk(fit);
  PVec_t mb= FitBgMrk(fit);

  switch (type) {
  case GowRegionFit:
    if (!FitTime(fit) || FitFitTyp(fit) != FIT_FTFIT || !VecNum(mr)) {
      return FALSE; }
    else {
      *chn_min= PVecX1(mr,0);
      *chn_max= PVecX2(mr,VecNum(mr)-1);
      return TRUE; }
  case GowBgFit:
    if ((!FitTime(fit) || FitFitTyp(fit) != FIT_FTFIT) && !FitBgTime(fit)) {
      return FALSE; }
    else if (VecNum(mr) || VecNum(mb)) {
      if (VecNum(mr) && VecNum(mb)) {
	*chn_min= MIN(PVecX1(mr,0), PVecX1(mb,0));
	*chn_max= MAX(PVecX2(mr,VecNum(mr)-1), PVecX2(mb,VecNum(mb)-1)); }
      else if (VecNum(mr)) {
	*chn_min= PVecX1(mr,0);
	*chn_max= PVecX2(mr,VecNum(mr)-1); }
      else {			/* VecNum(mb) */
	*chn_min= PVecX1(mb,0);
	*chn_max= PVecX2(mb,VecNum(mb)-1); }
      return TRUE; }
    else { return FALSE; }
  default: return FALSE; }
}
