/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/DrawLabel.c,v 1.2 1994/05/26 19:48:56 vs Exp vs $ */

#include <stdio.h>
#include <ctype.h>

#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCal.h>
#include <tv/vsSpectra.h>
#include <tv/vsTools.h>

#include "DrawLabel.h"

#define L_DRAWN		(1 << 0)
#define L_POSITIONED	(1 << 1)
#define L_USED		(1 << 2)

static Boolean DrawLabels();
static void CreateLabels();
static void PositionLabels();
static void CreateLabel();
static void InsertLabel();

typedef struct _CreateLabelArgs {
  double cl, pl;
  int x;
  int yb, yt;
  int direction;
  int fh;
  int gc;
  int lnum;
} CreateLabelArgs;
/* ---------------------------------------------------------------------- */
Boolean DrawLabel(dest, dlbl, x_scale, y_scale,
		  bot_frac, top_frac, cross,
		  distance, line , top_spectrum, chn_radius, lib_radius,
		  calibrated, modification, normalized)
     GowDestination dest;
     DrawLbl_t dlbl;
     GowScale x_scale;
     GowScale y_scale;
     double bot_frac;
     double top_frac;
     double cross;
     double distance;
     double line;
     int top_spectrum;
     int chn_radius;
     double lib_radius;
     int calibrated;
     GowScaleModificationType modification;
     int normalized;
{
  int max_gc= 0;
  IVecRec gc_rec;
  IVec_t gcs= &gc_rec;

  double (*cal_s)();
  double (*cal_c)();
  double (*cal_p)();
  double (*cal_efc)();
  double nrm= normalized ? *DrawLblNorm(dlbl) : 1.0;

  VecInit(gcs, sizeof(int));

  if (DrawLblCalibrated(dlbl)(DrawLblLbl(dlbl))) { cal_c= CalC; cal_p= CalId; }
  else { cal_c= CalId; cal_p= CalP; }

  cal_s= calibrated ? CalP : CalId;

  switch (modification) {
  case GowMmultiplied: cal_efc= CalId; break;
  case GowMefficiency: cal_efc= CalEfc; break;
  case GowMnone:
  default: cal_efc= CalUnity; break; }

  { Lbl_t v; ForAllVisibleLabels(dlbl, v) {
    v->status= L_DRAWN | L_POSITIONED;
    v->idx= -1;
    max_gc= MAX(max_gc, v->gc); } }

  CreateLabels(dest, dlbl, x_scale, y_scale, &max_gc,
	       top_frac, bot_frac, top_spectrum,
	       chn_radius, lib_radius,
	       cal_c, cal_p, cal_s, cal_efc, nrm);

  max_gc++;

  if (!DrawLblVis(dlbl) ||
      !VecAlloc(gcs, max_gc)) {
    return FALSE; }
  else { memset(VecP(gcs), 0, max_gc * sizeof(int)); VecNum(gcs)= max_gc; }

  PositionLabels(dlbl, y_scale, gcs,
		 bot_frac, top_frac, cross, distance, line);

  { int ret_val;

    GowDStart_object(dest, "label");
    ret_val= DrawLabels(dest, dlbl, x_scale, y_scale, gcs, cross);
    VecFree(gcs);
    GowDEnd_object(dest, "label");
    return ret_val; }
}
/* ---------------------------------------------------------------------- */
static Boolean DrawLabels(dest, dlbl, x_scale, y_scale, gcs, cross)
     GowDestination dest;
     DrawLbl_t dlbl;
     GowScale x_scale;
     GowScale y_scale;
     IVec_t gcs;
     double cross;
{
  int gc;
  int cw;
  Boolean visible= FALSE;

  { int cx, cy;
    int xp_min, xp_max;
    int yp_min, yp_max;

    xp_min= GowGetScaleLimitPixel(x_scale, GowClipSmall);
    xp_max= GowGetScaleLimitPixel(x_scale, GowClipLarge);
    yp_min= GowGetScaleLimitPixel(y_scale, GowClipSmall);
    yp_max= GowGetScaleLimitPixel(y_scale, GowToLarge);
    cx= (xp_max - xp_min) * cross; cx= ABS(cx);
    cy= (yp_max - yp_min) * cross; cy= ABS(cy);
    cw= MIN(cx, cy); }

  for (gc= 0; gc< VecNum(gcs); gc++) {
    XSegment *segs;
    if (VecP(gcs)[gc] &&
	(segs= (XSegment *)Malloc(VecP(gcs)[gc] * sizeof(XSegment)))) {
      Lbl_t v;
      XSegment *seg= segs;

      GowDSet_gc(dest, gc);

      ForAllVisibleLabels(dlbl, v) {
	if (v->gc == gc) {
	  int x= v->x;

	  GowDMove(dest, x, v->yl);
	  GowDLabel(dest, v->txt);
	  seg->x1= seg->x2= v->x;
	  seg->y1= v->yt; seg->y2= v->yb;
	  seg++;
	  if (cross != 0.0 && !(v->status & L_USED)) {
	    int yd= (v->yt + v->yb) / 2;
	    seg->x1= x-cw; seg->y1= yd-cw;
	    seg->x2= x+cw; seg->y2= yd+cw; seg++;
	    seg->x1= x+cw; seg->y1= yd-cw;
	    seg->x2= x-cw; seg->y2= yd+cw; seg++; } } }

      GowDSegments(dest, segs, seg - segs);
      visible= TRUE;
      Free(segs); } }
  return visible;
}
/* ---------------------------------------------------------------------- */
static void CreateLabels(dest, dlbl, x_scale, y_scale, max_gc,
			 top_frac, bot_frac, top_spectrum,
			 chn_radius, lib_radius,
			 cal_c, cal_p, cal_s, cal_efc, nrm)
     GowDestination dest;
     DrawLbl_t dlbl;
     GowScale x_scale;
     GowScale y_scale;
     int *max_gc;
     double bot_frac;
     double top_frac;
     Boolean top_spectrum;
     int chn_radius;
     double lib_radius;
     double (*cal_c)();
     double (*cal_p)();
     double (*cal_s)();
     double (*cal_efc)();
     double nrm;
{
  CreateLabelArgs args;
  Spc_t spc= DrawLblSpc(dlbl);
  Cal_t cal= DrawLblCal(dlbl);
  void *dlib= DrawLblLibFct(dlbl) ? DrawLblLib(dlbl) : NULL;
  SpcReal_t *s= SpcBufReal(SpcSpc(spc));
  GowPixelFunc px, py;
  int yp_clip, yp_min;
  double yu_min, yu_max;
  void *l= NULL;
  void *l_lib= NULL;
  double (*cal_l)();

  args.lnum= 0;

  if (dlib) { cal_l= DrawLblLibCalibrated(dlbl)(dlib) ? CalC : CalId; }

  switch (GowGetScaleDirection(y_scale)) {
  case GowNormalScaleDirection: args.direction= 1; break;
  case GowInvertedScaleDirection: args.direction= -1; break;
  default: return; }

  px= GowGetScalePixelFunction(x_scale);
  py= SpcNum(spc) ? GowGetScalePixelFunction(y_scale) : NULL;

  yp_clip= GowGetScaleLimitPixel(y_scale, GowClipLarge);
  yp_min= GowGetScaleLimitPixel(y_scale, GowClipSmall);
  GowGetScaleUserRange(y_scale, &yu_min, &yu_max);

  while ((l= DrawLblNext(dlbl)(DrawLblLbl(dlbl), l))) {
    double y= yu_min;
    double pos= DrawLblPos(dlbl)(DrawLblLbl(dlbl), l);
    double sl;
    int yp;
    int drawn= FALSE;

    args.cl= cal_c(cal, pos);
    sl= cal_s(cal, args.cl);

    if (px(x_scale, &args.x, sl) != GowValid) { args.lnum++; continue; }

    args.pl= cal_p(cal, pos);

    { int ci, c_min, c_max;
      c_min= args.cl - chn_radius; c_min= MAX(c_min, 0);
      c_max= args.cl + chn_radius; c_max= MIN(c_max, SpcNum(spc));
      for (ci= c_min; ci<= c_max; ci++) {
	double tmp= nrm * cal_efc(cal, (double)ci) * (double)s[ci];
	y= MAX(tmp, y); } }

    if (py) {
      switch (py(y_scale, &yp, y)) {
      case GowClipSmall:
      case GowToSmall: yp= yp_min; break;
      case GowClipLarge:
      case GowToLarge: yp= yp_clip; break;
      case GowValid: break; } }
    else { yp= yp_min; }

    args.yb= top_spectrum ? yp : yp_min;
    args.yt= yp;

    if (dlib) {
      double cl_min= args.cl - lib_radius;
      double cl_max= args.cl + lib_radius;

				/* libary labels must be sorted on positions */
      if (l_lib) {		/* check previous lib-labels */
	void *p= l_lib;
	while ((p= DrawLblLibPrev(dlbl)(dlib, p))) {
	  double clib= cal_l(cal, DrawLblLibPos(dlbl)(dlib, p));
	  if (clib < cl_min) { break; }
	  else if (clib <= cl_max) {
	    CreateLabel(dest, dlbl, l, dlib, p, &args, max_gc);
	    drawn= TRUE; } } }
      else { l_lib= DrawLblLibNext(dlbl)(dlib, l_lib); }

      { void *n= l_lib;
	while (n) {
	  double clib= cal_l(cal, DrawLblLibPos(dlbl)(dlib, n));

	  l_lib= n;
	  if (clib < cl_min) { break; }
	  else if (clib <= cl_max) {
	    CreateLabel(dest, dlbl, l, dlib, n, &args, max_gc);
	    drawn= TRUE; }
	  n= DrawLblLibNext(dlbl)(dlib, n); } } }
    if (!drawn) {
      CreateLabel(dest, dlbl, l, NULL, NULL, &args, max_gc); }
    args.lnum++; }
}
/* ---------------------------------------------------------------------- */
static void PositionLabels(dlbl, y_scale, gcs,
			   bot_frac, top_frac, cross, distance, line)
     DrawLbl_t dlbl;
     GowScale y_scale;
     IVec_t gcs;
     double bot_frac, top_frac, cross, distance, line;
{
  Lbl_t v;
  int yp_max;
  int direction;

  distance += 1.0;
  yp_max= GowGetScaleLimitPixel(y_scale, GowToLarge);

  switch (GowGetScaleDirection(y_scale)) {
  case GowNormalScaleDirection: direction= 1; break;
  case GowInvertedScaleDirection: direction= -1; break;
  default: return; }

  ForAllVisibleLabels(dlbl, v) {
    Lbl_t n= v->next;

    if (cross == 0.0 || (v->status & L_USED)) { VecP(gcs)[v->gc]++; }
    else { VecP(gcs)[v->gc] += 3; }
    
    if ((v->status & L_DRAWN) && n &&
	 ((n->x != v->x) || (n->status & L_DRAWN))) {
      continue; }
    else {
      int y= v->yt;
      int dy= (v->status & L_DRAWN) ? v->fh : (v->fh * distance);
      double fac;
      if (y * direction >= yp_max * direction) { continue; }
      while (n && v->x == n->x) { dy += n->fh * line; n= n->next; }

      fac= ((yp_max * direction < (y + dy) * direction) ?
	    (double)(yp_max - (double)y) / dy : 1.0);

      if (!(v->status & L_POSITIONED)) {
	v->yl= v->yt + NINT(fac * distance * v->fh);
	v->yt= v->yl - NINT(fac * top_frac * v->fh);
	v->yb= v->yb + NINT(fac * bot_frac * v->fh);
	v->status |= L_POSITIONED; }
      n= v->next;

      while (n && !(n->status & L_POSITIONED) && v->x == n->x) {
	Lbl_t p= n->prev;
	n->yl= p->yl + (int)(fac * p->fh * line);
	n->yt= n->yl - NINT(fac * top_frac * n->fh);
	n->yb= n->yb + NINT(fac * bot_frac * n->fh);
	n->status |= L_POSITIONED;
	n= n->next; } } }
}
/* ---------------------------------------------------------------------- */
static void CreateLabel(dest, dlbl, l, dlib, l_lib, args, max_gc)
     GowDestination dest;
     DrawLbl_t dlbl;
     void *l;
     DrawLbl_t dlib;
     void *l_lib;
     CreateLabelArgs *args;
     int *max_gc;
{
  char *txt= NULL;
  char pos_txt[64];

  args->gc= ((dlib && DrawLblLibGC(dlbl)) ?
	     DrawLblLibGC(dlbl)(dlib, l_lib) :
	     (DrawLblGC(dlbl) ?
	      DrawLblGC(dlbl)(DrawLblLbl(dlbl), l) : DrawLblGCIdx(dlbl)));
  *max_gc= MAX(args->gc, *max_gc);
  GowDSet_gc(dest, args->gc);
  args->fh= args->direction * GowDFontheight(dest);

  if (DrawLblString(dlbl)) {
    txt= (dlib ?
	  (DrawLblLibTxt(dlbl) ? DrawLblLibTxt(dlbl)(dlib, l_lib) : NULL) :
	  (DrawLblTxt(dlbl) ? DrawLblTxt(dlbl)(DrawLblLbl(dlbl), l) : NULL));
    if (txt) { InsertLabel(dlbl, args, txt); } }

  if (DrawLblPosCalib(dlbl) || (!txt && !DrawLblPosUncalib(dlbl))) {
    sprintf(pos_txt, "%g", args->pl);
    InsertLabel(dlbl, args, pos_txt); }
  if (DrawLblPosUncalib(dlbl) || (!txt && !DrawLblPosCalib(dlbl))) {
    sprintf(pos_txt, "%g", args->cl);
    InsertLabel(dlbl, args, pos_txt); }
}
/* ---------------------------------------------------------------------- */
static void InsertLabel(dlbl, args, txt)
     DrawLbl_t dlbl;
     CreateLabelArgs *args;
     char *txt;
{
  Lbl_t v;

  if ((v= DrawLblCur(dlbl))) {
    while (v->x >= args->x && v->prev) { v= v->prev; }
    while (v->x < args->x && v->next) { v= v->next; }

    while (v->x == args->x) {
      if  (!(v->status & L_USED) &&
	   v->gc == args->gc && v->fh == args->fh &&
	   !strcmp(v->txt, txt)) {
	if  (!(v->status & L_POSITIONED)) {
	  v->yt= args->yt;
	  v->yb= args->yb; }
	v->idx= args->lnum;
	v->status |= L_USED; return; }
      else if (v->next) { v= v->next; }
      else { break; } } }

  { Lbl_t new;
    if ((new= (Lbl_t)Malloc(sizeof(LblRec)))) {
      if ((new->txt= Strdup(txt))) {
	new->x= args->x;
	new->yt= args->yt;
	new->yb= args->yb;
	new->gc= args->gc;
	new->fh= args->fh;
	new->idx= args->lnum;
	new->status= L_USED;
	if (v) {
	  if (v->x <= args->x) {
	    new->prev= v; new->next= v->next; v->next= new; 
	    if (new->next) { new->next->prev= new;} }
	  else {
	    new->prev= v->prev; new->next= v; v->prev= new;
	    if (new->prev) { new->prev->next= new; }
	    if (v == DrawLblVis(dlbl)) { DrawLblVis(dlbl)= new; } } }
	else { DrawLblVis(dlbl)= new; new->prev= new->next= NULL; }
	DrawLblCur(dlbl)= new; }
      else { Free(new); return; } } }
}
/* ---------------------------------------------------------------------- */
void DrawLabelFree(dlbl)
     DrawLbl_t dlbl;
{
  Lbl_t v, last= NULL;

  ForAllVisibleLabels(dlbl, v) { Free(v->txt); last= v; Free(v->prev); }
  Free(last);
  DrawLblVis(dlbl)= DrawLblCur(dlbl)= NULL;
}
/* ---------------------------------------------------------------------- */
void DrawLabelInit(dlbl)
     DrawLbl_t dlbl;
{
  DrawLblGCIdx(dlbl)= 0;
  DrawLblFct(dlbl)= NULL;
  DrawLblLbl(dlbl)= NULL;
  DrawLblVis(dlbl)= NULL;
  DrawLblCur(dlbl)= NULL;
  DrawLblCal(dlbl)= NULL;
  DrawLblSpc(dlbl)= NULL;
  DrawLblNorm(dlbl)= NULL;
  DrawLblLibFct(dlbl)= NULL;
  DrawLblLib(dlbl)= NULL;
  DrawLblPosCalib(dlbl)= FALSE;
  DrawLblPosUncalib(dlbl)= FALSE;
  DrawLblString(dlbl)= FALSE;
}
/* ---------------------------------------------------------------------- */
int DrawLabelGetIndex(dlbl, x, y)
     DrawLbl_t dlbl;
     int x, y;
{
  int idx= -1;
  Lbl_t v;
  int min_dx= -1, min_dy;
  ForAllVisibleLabels(dlbl, v) {
    if (v->idx >= 0) {
      int dx= v->x - x;
      int dy= v->yl - y;
      dx= ABS(dx); dy= ABS(dy);
      if (min_dx < 0 || min_dx > dx) { idx= v->idx; min_dx= dx; min_dy= dy; }
      else if (min_dx == dx && min_dy > dy) { idx= v->idx; min_dy= dy; } } }
  return idx;
}
/* ---------------------------------------------------------------------- */
