/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/DrawPeak.c,v 1.2 1994/05/26 19:48:57 vs Exp vs $ */

#include <stdio.h>
#include <ctype.h>

#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCal.h>

#include "DrawPeak.h"

/*	draws from PIXELmin -> PIXELmax ,CHNmin -> CHNmax 
 *	in each pixel column is drawn only one channel;
 *	all other channels which fall into the same column are omitted
 */


Boolean DrawPeak(dest, dpeak, x_scale, y_scale,
		 step_width, min_steps,
		 calibrated, modification, normalized)
     GowDestination dest;
     DrawPeak_t dpeak;
     GowScale x_scale;
     GowScale y_scale;
     int step_width;
     int min_steps;
     int calibrated;
     GowScaleModificationType modification;
     int normalized;
{
  Cal_t cal= DrawPeakCal(dpeak);
  void *par= DrawPeakPar(dpeak);
  double (*fct)()= DrawPeakFct(dpeak);
  double pos= DrawPeakChn(dpeak);
  Boolean visible= FALSE;

  GowUserFunc ux;
  GowPixelFunc px, py;
  int xp_min, xp_max, yp_min, yp_max;
  double xu_min, xu_max, yu_min, yu_max;
  double y_offset;
  XPoint *pnts;
  double (*cal_s)();
  double (*cal_c)();
  double (*cal_efc)();
  double nrm= normalized ? *DrawPeakNorm(dpeak) : 1.0;

  if (calibrated) { cal_s= CalP; cal_c= CalC; }
  else { cal_s= CalId; cal_c= CalId; }
  switch (modification) {
  case GowMmultiplied: cal_efc= cal_c; break;
  case GowMefficiency: cal_efc= CalEfc; break;
  case GowMnone:
  default: cal_efc= CalUnity; break; }

  ux= GowGetScaleUserFunction(x_scale);
  px= GowGetScalePixelFunction(x_scale);
  py= GowGetScalePixelFunction(y_scale);
  GowGetScaleUserRange(x_scale, &xu_min, &xu_max);
  GowGetScaleUserRange(y_scale, &yu_min, &yu_max);
  xp_min= GowGetScaleLimitPixel(x_scale, GowClipSmall);
  xp_max= GowGetScaleLimitPixel(x_scale, GowClipLarge);
  yp_min= GowGetScaleLimitPixel(y_scale, GowClipSmall);
  yp_max= GowGetScaleLimitPixel(y_scale, GowClipLarge);

  if (!GowGetScaleOffset(y_scale, &y_offset)) { return FALSE; }

  if (!(pnts= (XPoint *)Malloc(sizeof(XPoint) * 2 * (xp_max - xp_min + 4)))) {
    return FALSE; }

  { double tmp= fct(par, pos) * nrm * cal_efc(cal, pos);
    if (tmp >= 0.0) { GowDSet_gc(dest, DrawPeakGCNeg(dpeak)); }
    else { GowDSet_gc(dest, DrawPeakGCPos(dpeak)); } }

  GowDStart_object(dest, "peak");

  { double offset;
    int x_end;
    int fac;
    int xpos;

    switch (px(x_scale, &xpos, cal_s(cal, pos))) {
    case GowValid: break;
    case GowClipSmall:
    case GowToSmall: xpos= xp_min; break;
    case GowClipLarge:
    case GowToLarge: xpos= xp_max; break; }
	  
    for (offset= -step_width, x_end= -xp_min, fac= -1;
	 offset <= step_width;
	 offset += 2 * step_width, x_end= xp_max, fac= 1) {
      XPoint *pnt= pnts;
      int x;
      int x_prev;
      int y_prev= -1;		/* undefined */
      GowScaleConversion y_prevc;
      for (x= xpos; x * fac <= x_end; x += offset) {
	double x_user;
	GowScaleConversion yconv;
	int y;

	ux(x_scale, x, &x_user);
	x_user= cal_c(cal, x_user);

	{ double tmp= nrm * cal_efc(cal, x_user) * fct(par, x_user);
	  tmp= ABS(tmp);
	  yconv= py(y_scale, &y, y_offset + tmp); }

	switch (yconv) {	/* handle clipped as out of bounds */
	case GowValid: break;
	case GowClipSmall: yconv= GowToSmall;
	case GowToSmall: y= yp_min; break;
	case GowClipLarge: yconv= GowToLarge;
	case GowToLarge: y= yp_max; break; }

	if (y_prev >= 0 &&	/* defined */
	    (y_prevc == GowValid || yconv != y_prevc)) { /* visible */
	  pnt->x= x_prev; pnt->y= y_prev; pnt++; } /* 1st point */

	if (yconv == GowValid) {
	  pnt->x= x; pnt->y= y; pnt++;
	  y_prev= -1; }
	else {
	  if (y_prevc != yconv) { /* last point */
	    pnt->x= x; pnt->y= y; pnt++; }

	  if (pnt - pnts > 1 && GowDLines(dest, pnts, pnt - pnts)) {
	    visible= TRUE; }
	  pnt= pnts;

	  x_prev= x;		/* not yet drawn */
	  y_prev= y;
	  y_prevc= yconv; }
	if (yconv == GowValid && y == yp_min) { break; } }

      if (pnt - pnts > 1 && GowDLines(dest, pnts, pnt - pnts)) {
	visible= TRUE; } } }

  GowDEnd_object(dest, "peak");

  Free(pnts);
  return visible;
}

void DrawPeakExpand(dpeak,
		    x_scale, x_type, x_full,
		    y_scale, y_type, y_full,
		    calibrated, modification, normalized)
     DrawPeak_t dpeak;
     GowScale x_scale;
     GowScaleType x_type;
     int x_full;
     GowScale y_scale;
     GowScaleType y_type;
     int y_full;
     int calibrated;
     GowScaleModificationType modification;
     int normalized;
{
  double x_offset, x_range;

  if (y_full &&
      GowGetScaleOffset(x_scale, &x_offset) &&
      GowGetScaleRange(x_scale, &x_range)) {
    Cal_t cal= DrawPeakCal(dpeak);
    void *par= DrawPeakPar(dpeak);
    double (*fct)()= DrawPeakFct(dpeak);
    double pos= DrawPeakChn(dpeak);
    double chn_min;
    double chn_max;
    double y_range= 0.0;
    double (*cal_s)();
    double (*cal_c)();
    double (*cal_efc)();
    double nrm= normalized ? *DrawPeakNorm(dpeak) : 1.0;

    if (calibrated) { cal_s= CalP; cal_c= CalC; }
    else { cal_s= CalId; cal_c= CalId; }
    switch (modification) {
    case GowMmultiplied: cal_efc= cal_c; break;
    case GowMefficiency: cal_efc= CalEfc; break;
    case GowMnone:
    default: cal_efc= CalUnity; break; }

    { double tmp;
      tmp= cal_c(cal, x_offset);
      chn_min= MAX(chn_min, tmp);
      tmp= cal_c(cal, x_offset + x_range); 
      chn_max= MIN(chn_max, tmp); }

    if (chn_min >= chn_max) { return; }

    if (pos < chn_min) {
      double y= fct(par, chn_min) * nrm * cal_efc(cal, chn_min);
      y= ABS(y);
      y_range= MAX(y_range, y); }
    else if (pos > chn_max) {
      double y= fct(par, chn_min) * nrm * cal_efc(cal, chn_max);
      y= ABS(y);
      y_range= MAX(y_range, y); }
    else {
      double offset;
      double chn_end;
      double fac;
      for (offset= -0.5, chn_end= -chn_min, fac= -1.0;
	   offset < 1.0;
	   offset += 1.0, chn_end= chn_max, fac= 1.0) {
	double chn;
	double y, yp;
	for (chn= pos, yp= 0.0; chn * fac < chn_end; chn += offset) {
	  y= fct(par, chn) * nrm * cal_efc(cal, chn); y= ABS(y);
	  if (y < yp) { break; }
	  else { y_range= MAX(y_range, y); yp= y; } } } } }

  return;
}
