/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowCutWidget.c,v 1.2 1994/05/26 19:49:03 vs Exp vs $ */

#ifdef SOLARISBUG_X_INCLUDES
#undef NOSTDHDRS
#endif

#include <X11/IntrinsicP.h>
#include <X11/cursorfont.h>
#include <X11/StringDefs.h>

#include <X11/Xmu/Misc.h>
#include <X11/Xmu/Converters.h>

#ifdef HAS_XAW3D
#include <X11/Xaw3d/XawInit.h>
#include <X11/Xaw3d/Grip.h>
#else
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Grip.h>
#endif

#include <Gow/GowPaneGeo.h>
#include <Gow/GowPlot.h>
#include <Gow/GowFramWidgetP.h>

#include "GowCutWidgetP.h"

#include <ctype.h>

#define ForAllChildren(pw, childP) \
  for ( (childP) = (pw)->composite.children ; \
        (childP) < (pw)->composite.children + (pw)->composite.num_children ; \
        (childP)++ )

/*****************************************************************************
 *
 * Full instance record declaration
 *
 ****************************************************************************/

static char defGripTranslations[] =
    "<BtnDown>:		GripAction(Start) \n\
     <BtnMotion>:	GripAction(Move) \n\
     Any<BtnUp>:	GripAction(Commit)";

#define offset(field) XtOffsetOf(GowCutWidgetRec, cut_widget.field)

static float def_hr= 0.3;
static float def_w2h= 1.2;

static XtResource resources[] = {
    {XtNinternalBorderWidth, XtCBorderWidth, XtRDimension, sizeof(Dimension),
	 offset(internal_bw), XtRImmediate, (XtPointer) 1},
    {XtNgripTranslations, XtCTranslations, XtRTranslationTable,
         sizeof(XtTranslations),
         offset(grip_translations), XtRString, (XtPointer)defGripTranslations},
    {XtNheightRatio, XtCHeightRatio, XtRFloat,
         sizeof(float),
         offset(height_ratio), XtRFloat, (XtPointer)&def_hr},
    {XtNwidthToHeight, XtCWidthToHeight, XtRFloat,
         sizeof(float),
         offset(width_to_height), XtRFloat, (XtPointer)&def_w2h},

    /* Cursors - both horiz and vertical have to work. */

    {XtNcursor, XtCCursor, XtRCursor, sizeof(Cursor),
         offset(cursor), XtRImmediate, None},
    {XtNgripCursor, XtCCursor, XtRCursor, sizeof(Cursor),
         offset(grip_cursor), XtRImmediate, None},
    {XtNadjustCursor, XtCCursor, XtRCursor, sizeof(Cursor),
         offset(adjust_cursor), XtRString, "bottom_left_corner"},
};

#undef offset

static void CreateGrip();
static void Layout();
static void ChangeGripCursor();
static void Initialize();
static void Realize(), Resize();
static void GetGC(), ReleaseGC();
static Boolean SetValues();
static XtGeometryResult GeometryManager();
static void ChangeManaged();
static void InsertChild();
static void DeleteChild();

#define SuperClass ((CompositeWidgetClass)&compositeClassRec)

GowCutWidgetClassRec gowCutWidgetClassRec = {
   {
/* core class fields */
    /* superclass         */   (WidgetClass) SuperClass,
    /* class name         */   "GowCutWidget",
    /* size               */   sizeof(GowCutWidgetRec),
    /* class_initialize   */   NULL,
    /* class_part init    */   NULL,
    /* class_inited       */   FALSE,
    /* initialize         */   Initialize,
    /* initialize_hook    */   NULL,
    /* realize            */   Realize,
    /* actions            */   NULL,
    /* num_actions        */   0,
    /* resources          */   resources,
    /* resource_count     */   XtNumber(resources),
    /* xrm_class          */   NULLQUARK,
    /* compress_motion    */   TRUE,
    /* compress_exposure  */   TRUE,
    /* compress_enterleave*/   TRUE,
    /* visible_interest   */   FALSE,
    /* destroy            */   ReleaseGC,
    /* resize             */   Resize,
    /* expose             */   NULL,
    /* set_values         */   SetValues,
    /* set_values_hook    */   NULL,
    /* set_values_almost  */   XtInheritSetValuesAlmost,
    /* get_values_hook    */   NULL,
    /* accept_focus       */   NULL,
    /* version            */   XtVersion,
    /* callback_private   */   NULL,
    /* tm_table           */   NULL,
    /* query_geometry	  */   GowQueryPaneGeometry,
    /* display_accelerator*/   XtInheritDisplayAccelerator,
    /* extension          */   NULL
   }, {
/* composite class fields */
    /* geometry_manager   */   GeometryManager,
    /* change_managed     */   ChangeManaged,
    /* insert_child       */   InsertChild,
    /* delete_child       */   DeleteChild,
    /* extension          */   NULL
   }
};

WidgetClass gowCutWidgetClass = (WidgetClass) &gowCutWidgetClassRec;

/***********************************************************
 *
 * Private Functions.
 *
 ************************************************************/

static void Layout(pw)
     GowCutWidget pw;
{
  Widget child0;

  if (!pw->composite.num_children || !XtIsRealized((Widget)pw)) { return; }
  
  child0= pw->composite.children[0];

  if (XtIsManaged(child0) && XtIsRealized(child0)) {
    XtConfigureWidget(child0,
		      /*POSITIONCAST*/ pw->core.x,
		      /*POSITIONCAST*/ pw->core.y,
		      /*DIMENSIONCAST*/ pw->core.width,
		      /*DIMENSIONCAST*/ pw->core.height,
		      /*DIMENSIONCAST*/ 0); }
  

  if (pw->composite.num_children == 3) {
    Widget child1= pw->composite.children[1];
    Widget grip= pw->cut_widget.grip;
    Dimension width;
    Dimension height;
    Dimension border_width;

    if (XtIsManaged(child1) && XtIsRealized(child1)) {
      if (XtIsManaged(child0) && XtIsRealized(child0)) {
	border_width= pw->cut_widget.internal_bw;
	height	= pw->core.height * pw->cut_widget.height_ratio;
	width	= height * pw->cut_widget.width_to_height;
	if (width > pw->core.width) { width= pw->core.width; }
	if (width < 1) { width= 1; }
	if (height < 1) { height= 1; } }
      else {
	border_width= 0;
	height	= pw->core.height;
	width	= pw->core.width; }
      { Position x	= (pw->core.width - width - border_width);
	Position y	= -border_width;
	XtConfigureWidget(child1,
			  /*POSITIONCAST*/ x,
			  /*POSITIONCAST*/ y,
			  /*DIMENSIONCAST*/ width,
			  /*DIMENSIONCAST*/ height,
			  /*DIMENSIONCAST*/ border_width);
	if (XtIsManaged(grip)) {
	  XWindowChanges changes;
	  Dimension bw= grip->core.border_width;

	  changes.stack_mode = Above;
	  changes.x	= x - grip->core.width / 2 - bw;
	  changes.y	= height - grip->core.height / 2 - bw;
	  if (changes.x < 0) { changes.x= 0; }
	  if (changes.y < 0) { changes.y= 0; }
	  if (XtIsRealized(grip)) {
	    XConfigureWindow(XtDisplay(grip), XtWindow(grip),
			     (unsigned int)(CWX | CWY | CWStackMode),
			     &changes ); } } } } }
}

/*	Function Name: GetGC
 *	Description: Gets new GC's.
 *	Arguments: w - the paned widget.
 *	Returns: none.
 */

static void
GetGC(w)
Widget w;
{
  GowCutWidget pw = (GowCutWidget) w;
  XtGCMask valuemask;
  XGCValues values;

/*
 * Draw Track lines (animate pane borders) in internal border color ^ bg color.
 */

  if (!pw->cut_widget.grip) { return; }

  values.function = GXinvert;
  values.plane_mask = (pw->core.border_pixel ^ pw->core.background_pixel);
  values.subwindow_mode = IncludeInferiors; 
  valuemask = GCPlaneMask | GCFunction | GCSubwindowMode;
  pw->cut_widget.flipgc = XtGetGC(w, valuemask, &values);
}

/*	Function Name: _DrawTrackLines
 *	Description: Draws the lines that animate the pane borders when the
 *                   grips are moved.
 *	Arguments: pw - the Paned widget.
 *                 erase - if True then just erase track lines, else
 *                         draw them in.
 *	Returns: none.
 */

static void
_DrawTrackLines(pw, x, y, erase)
GowCutWidget pw;
int x, y;
Boolean erase;
{
  XSegment *track= pw->cut_widget.track_lines;

  x += pw->cut_widget.x_offset;
  y += pw->cut_widget.y_offset;

  if (x < 0) { x= 0; }
  if (x >= pw->core.width) { x= pw->core.width; }
  if (y < 1) { y= 1; }
  if (y >= pw->core.height) { y= pw->core.height; }

  if (erase || track[0].x1 != x || track[0].y1 != y) {
    if (!erase) {
      XDrawSegments(XtDisplay(pw), XtWindow(pw),
		    pw->cut_widget.flipgc, track, 2);
      track[0].x1= track[0].x2= x;
      track[0].y1= y;
      track[0].y2= 0;

      track[1].x1= x;
      track[1].x2= pw->core.width;
      track[1].y1= track[1].y2= y; }
    XDrawSegments(XtDisplay(pw), XtWindow(pw),
		  pw->cut_widget.flipgc, track, 2); }
}

/* 
 * This allows good reuse of code, as well as descriptive function names.
 */

#define DrawTrackLines(pw,x,y) _DrawTrackLines((pw), x, y, FALSE);
#define EraseTrackLines(pw) _DrawTrackLines((pw), 0, 0, TRUE);

/*	Function Name: GetEventLocation
 *	Description: Converts and event to an x and y location.
 *	Arguments: pw - the paned widget.
 *                 event - a pointer to an event.
 *	Returns: if this is a vertical pane then (y) else (x).
 */

static Boolean
GetEventLocation(pw, event, x, y)
GowCutWidget pw;
XEvent *event;
int *x, *y;
{
  switch (event->type) {
  case ButtonPress:
  case ButtonRelease:
    *x = event->xbutton.x_root;
    *y = event->xbutton.y_root;
    break;
  case KeyPress:
  case KeyRelease:
    *x = event->xkey.x_root;
    *y = event->xkey.y_root;
    break;
  case MotionNotify:
    *x = event->xmotion.x_root;
    *y = event->xmotion.y_root;
    break;
  default:
    return FALSE; }
  return TRUE;
}

/*	Function Name: StartGripAdjustment
 *	Description: Starts the grip adjustment procedure.
 *	Arguments: pw - the paned widget.
 *                 grip - the grip widget selected.
 *                 dir - the direction that we are to be moving.
 *	Returns: none.
 */

static void
StartGripAdjustment(pw, x, y)
GowCutWidget pw;
int x, y;
{
  Widget grip= pw->cut_widget.grip;

  if (grip && XtIsRealized(grip) && XtIsManaged(grip)) {
    XSegment *track= pw->cut_widget.track_lines;
    Widget child= pw->composite.children[1];
    int gx= child->core.x;
    int gy= child->core.y + child->core.height;
    pw->cut_widget.x_offset= gx - x;
    pw->cut_widget.y_offset= gy - y;
    track[0].x1= track[0].x2= gx;
    track[0].y1= gy;
    track[0].y2= 0;

    track[1].x1= gx;
    track[1].x2= pw->core.width;
    track[1].y1= track[1].y2= gy;
    pw->cut_widget.grip_adjust= TRUE;
    XDefineCursor(XtDisplay(grip), XtWindow(grip),
		  pw->cut_widget.adjust_cursor); }
}

/*	Function Name: MoveGripAdjustment
 *	Description: This routine moves all panes around when a grip is moved. 
 *	Arguments: pw - the paned widget.
 *                 grip - the grip that we are moving.
 *                 dir - the direction the pane we are interested is w.r.t the
 *                       grip.
 *                 loc - location of pointer in proper direction.
 *	Returns: none.
 */

static void
MoveGripAdjustment(pw, x, y)
GowCutWidget pw;
int x, y;
{
  if (!pw->cut_widget.grip_adjust) { return; }
  DrawTrackLines(pw, x, y);
}

/*	Function Name: CommitGripAdjustment
 *	Description: Commits the grip adjustment.
 *	Arguments: pw - the paned widget.
 *	Returns: none
 */

static void
CommitGripAdjustment(pw)
GowCutWidget pw;
{
  if (!pw->cut_widget.grip_adjust) { return; }
  else {
    double w2h, hr;
    XSegment *track= pw->cut_widget.track_lines;
    Arg arglist[1];
    Widget grip= pw->cut_widget.grip;
    Cursor cursor;

    XtSetArg(arglist[0], XtNcursor, &cursor);
    XtGetValues(grip, arglist, (Cardinal) 1);
    XDefineCursor(XtDisplay(grip), XtWindow(grip), cursor);

    EraseTrackLines(pw);
    pw->cut_widget.grip_adjust= FALSE;
    hr= (double)track[0].y1 / (double)pw->core.height;
    w2h= (double)(pw->core.width - track[0].x1) / (double)track[0].y1;

    if (hr > 0.0 && hr <= 1.0) { pw->cut_widget.height_ratio= hr; }
    if (w2h > 0.0) { pw->cut_widget.width_to_height= w2h; }

    Layout(pw); }
}

/*	Function Name: HandleGrip
 *	Description: Handles the grip manipulations.
 *	Arguments: grip - the grip widget that has been moved.
 *                 junk - ** NOT USED **
 *                 call_data - data passed to us from the grip widget.
 *	Returns: none.
 */

/* ARGSUSED */
static void
HandleGrip(grip, junk, callData)
Widget grip;
XtPointer junk, callData;
{
  XawGripCallData call_data = (XawGripCallData)callData;
  GowCutWidget pw = (GowCutWidget) XtParent(grip);
  char action_type;
  int x, y;

  action_type = *call_data->params[0];

  if (call_data->num_params == 0 || call_data->num_params != 1) {
    XtError( "Paned GripAction has been passed incorrect parameters."); }

  if (islower(action_type)) action_type = toupper(action_type);

  if (!GetEventLocation(pw, (XEvent *) (call_data->event), &x, &y)) { return; }

  switch (action_type) {
  case 'S':		/* Start adjustment */
    StartGripAdjustment(pw, x, y);
    break;
  case 'M': 
    MoveGripAdjustment(pw, x, y);
    break;
  case 'C':
    CommitGripAdjustment(pw);
    break;
  default:
    XtError( "GowCutWidget GripAction(); 1st parameter invalid" ); }
}

/*	Function Name: ManageAndUnmanageGrip
 *	Description: This function manages and unmanages the grips so that
 *                   the managed state of each grip matches that of its pane.
 *	Arguments: pw - the paned widget.
 *	Returns: none.
 */

static void   
ManageAndUnmanageGrip(pw)
GowCutWidget pw;
{
  if (pw->cut_widget.grip) {
    if (XtIsManaged(pw->composite.children[0]) &&
	XtIsManaged(pw->composite.children[1])) {
      if (!XtIsManaged(pw->cut_widget.grip)) {
	XtManageChild(pw->cut_widget.grip); } }
    else {
      if (XtIsManaged(pw->cut_widget.grip)) {
	XtUnmanageChild(pw->cut_widget.grip); } } }
}

/*	Function Name: CreateGrip
 *	Description: Creates a grip widget.
 *	Arguments: child - the child that wants a grip to be created for it.
 *	Returns: none.
 */

static void
CreateGrip(pw)
GowCutWidget pw;
{
  Arg arglist[2];
  Cardinal num_args = 0;
  Cursor cursor;
     
  XtSetArg(arglist[num_args], XtNtranslations,
	   pw->cut_widget.grip_translations);
  num_args++;
  if ((cursor = pw->cut_widget.grip_cursor) == None) {
    cursor = pw->cut_widget.grip_cursor; }

  XtSetArg(arglist[num_args], XtNcursor, cursor);
  num_args++;
  pw->cut_widget.grip = XtCreateWidget("grip", gripWidgetClass, (Widget)pw,
				       arglist, num_args);
  GetGC(pw);

  XtAddCallback(pw->cut_widget.grip, XtNcallback, 
		HandleGrip, (XtPointer)pw);
}

/*	Function Name: ChangeGripCursor
 *	Description: Changes the grip cursor.
 *	Arguments: pw - the paned widget.
 *	Returns: none
 */

static void
ChangeGripCursor(pw)
GowCutWidget pw;
{
  Cursor cursor= pw->cut_widget.cursor;
  Arg arglist[1];

  XtSetArg(arglist[0], XtNcursor, cursor);
  XtSetValues(pw->cut_widget.grip, arglist, (Cardinal) 1);
}
      
/************************************************************
 *
 * Semi-public routines. 
 *
 ************************************************************/

/* The Geometry Manager only allows changes after Realize if
 * allow_resize is True in the constraints record.  
 * 
 * For vertically paned widgets:
 *
 * It only allows height changes, but offers the requested height
 * as a compromise if both width and height changes were requested.
 *
 * For horizontal widgets the converse is true.
 * As all good Geometry Managers should, we will return No if the
 * request will have no effect; i.e. when the requestor is already
 * of the desired geometry.
 */

static XtGeometryResult GeometryManager(w, request, reply)
Widget w;
XtWidgetGeometry *request, *reply;
{
  GowCutWidget pw = (GowCutWidget) XtParent(w);
  XtGeometryMask mask = request->request_mode;
  Widget grip= pw->cut_widget.grip;
  Boolean both_managed= grip && XtIsManaged(grip);

/*
 * If any of the following is true, disallow the geometry change.
 *
 * o The request is not a width or height request.
 * o The requested size is the same as the current size.
 */

  if (((mask & ~(XtGeometryMask)CWWidth ||
	w->core.width == request->width) &&
       (mask & ~(XtGeometryMask)CWHeight ||
	w->core.height == request->height)) ||
      w == pw->cut_widget.grip) {
    return XtGeometryNo; }

  if (w == pw->composite.children[0] ||
      (pw->composite.num_children == 3 && !both_managed)) {
    XtGeometryResult result= XtMakeGeometryRequest(w, request, reply);

    switch (result) {
    case XtGeometryYes: Resize(w); return XtGeometryDone;
    default: return result; } }
  else {
    if (both_managed) {
      { Dimension width= grip->core.width + 2 * grip->core.border_width + 1;
	Dimension height= grip->core.height + 2 * grip->core.border_width + 1;
	int almost= FALSE;

	if (mask & CWWidth) {
	  if (request->width > width) {
	    almost= TRUE;
	    reply->width= width; }
	  else { reply->width= request->width; }
	  reply->request_mode |= CWWidth; }
	if (mask & CWHeight) {
	  if (request->height > height) {
	    almost= TRUE;
	    reply->height= height; }
	  else { reply->height= request->height; }
	  reply->request_mode |= CWHeight; }
	if (almost) { return XtGeometryAlmost; } }

      if (mask & ~(XtGeometryMask)XtCWQueryOnly) {
	Dimension width= (mask & CWWidth) ? request->width : w->core.width;
	Dimension height= (mask & CWWidth) ? request->height : w->core.height;
	Position x= pw->core.width - width;
	Position y= 0;
	Dimension border_width= w->core.border_width;
	XtConfigureWidget(w,
			  /*POSITIONCAST*/ x,
			  /*POSITIONCAST*/ y,
			  /*DIMENSIONCAST*/ width,
			  /*DIMENSIONCAST*/ height,
			  /*DIMENSIONCAST*/ border_width);

	{ XWindowChanges changes;
	  changes.stack_mode = Above;
	  if (XtIsRealized(grip)) {
	    XConfigureWindow(XtDisplay(grip), XtWindow(grip),
			     (unsigned int)CWStackMode, &changes ); } } }
      return XtGeometryDone; } }
  return XtGeometryNo;
}

/* ARGSUSED */
static void Initialize(request, new)
Widget request, new;
{
  GowCutWidget pw = (GowCutWidget)new;

  GetGC(pw);

  if (!new->core.width) { new->core.width= 500; }
  if (!new->core.height) { new->core.height= 500; }

  pw->cut_widget.grip= NULL;
  pw->cut_widget.grip_adjust= FALSE;
  if (pw->cut_widget.height_ratio > 1.0 ||
      pw->cut_widget.height_ratio <= 0.0) {
    pw->cut_widget.height_ratio= def_hr; }
  if (pw->cut_widget.width_to_height > 1.0 ||
      pw->cut_widget.width_to_height <= 0.0) {
    pw->cut_widget.width_to_height= def_w2h; }
}

static void 
Realize(w, valueMask, attributes)
Widget w;
Mask *valueMask;
XSetWindowAttributes *attributes;
{
    GowCutWidget pw = (GowCutWidget) w;
    Widget *childP;

    if ((attributes->cursor = (pw)->cut_widget.cursor) != None)
	*valueMask |= CWCursor;

    (*SuperClass->core_class.realize) (w, valueMask, attributes);

/*
 * Before we commit the new locations we need to realize all the panes and
 * their grips.
 */

    ForAllChildren(pw, childP) { XtRealizeWidget( *childP ); }

    Layout((GowCutWidget)w);
} /* Realize */

static void 
ReleaseGC(w)
Widget w;
{
  register GowCutWidget pw = (GowCutWidget)w;

  XtReleaseGC( w, pw->cut_widget.flipgc );
} 

static void InsertChild(w)
register Widget w;
{
  GowCutWidget pw= (GowCutWidget)XtParent(w);

  if (pw->composite.num_children == 3) { XtDestroyWidget(w); return; }

  /* insert the child widget in the composite children list with the */
  /* superclass insert_child routine.                                */

  (*SuperClass->composite_class.insert_child)(w);

  if (pw->composite.num_children == 2) { CreateGrip(pw); }

} /* InsertChild */

static void DeleteChild(w)
Widget w;
{
  GowCutWidget pw= (GowCutWidget)XtParent(w);

  if (pw->composite.num_children == 3) {
    pw->cut_widget.grip= NULL;

    /* delete the child widget in the composite children list with the */
    /* superclass delete_child routine.                                */
    (*SuperClass->composite_class.delete_child)(w);
    /* delete grip (now on second place) or second child */
    XtDestroyWidget(pw->composite.children[1]); }
  else { (*SuperClass->composite_class.delete_child)(w); }
} /* DeleteChild */

static void ChangeManaged(w)
   Widget w;
{
   GowCutWidget pw = (GowCutWidget)w;

   ManageAndUnmanageGrip(pw);

   Layout(pw);

} /* ChangeManaged */

/*	Function Name: Resize
 *	Description: The paned widget's resize proc.
 *	Arguments: w - the paned widget.
 *	Returns: none.
 */

static void
Resize(w)
Widget w;
{
  /*  printf("Hallo\n");fflush(stdout);*/
  Layout((GowCutWidget)w);
}

/* ARGSUSED */
static Boolean 
SetValues(old, request, new)
Widget old, request, new;
{
  GowCutWidget old_pw = (GowCutWidget) old;
  GowCutWidget new_pw = (GowCutWidget) new;
  Boolean redisplay = FALSE;

  if ((old_pw->cut_widget.cursor != new_pw->cut_widget.cursor) &&
      XtIsRealized(new)) {
    XDefineCursor(XtDisplay(new), XtWindow(new),
		  new_pw->cut_widget.cursor); }

  if (old_pw->core.background_pixel != new_pw->core.background_pixel) {
    ReleaseGC(old);
    GetGC(new);
    redisplay = TRUE; }

  if (old_pw->cut_widget.grip_cursor != new_pw->cut_widget.grip_cursor) {
    ChangeGripCursor(new_pw); }

  if (new_pw->cut_widget.width_to_height <= 0.0 ||
      new_pw->cut_widget.width_to_height > 1.0) {
    new_pw->cut_widget.width_to_height= old_pw->cut_widget.width_to_height; }
  if (new_pw->cut_widget.height_ratio <= 0.0 ||
      new_pw->cut_widget.height_ratio > 1.0) {
    new_pw->cut_widget.height_ratio= old_pw->cut_widget.height_ratio; }

  if (old_pw->cut_widget.width_to_height !=
      new_pw->cut_widget.width_to_height ||
      old_pw->cut_widget.height_ratio != new_pw->cut_widget.height_ratio ||
      old_pw->cut_widget.internal_bw != new_pw->cut_widget.internal_bw) {
    Layout(new_pw);
    return FALSE; }		/* We have done a full configuration, return.*/

  return redisplay;
} /* SetValues */

/************************************************************
 *
 * Public routines. 
 *
 ************************************************************/

void GowCutWidgetPlot(w, fp, type, geo)
     Widget w;
     GowDestFile fp;
     int type;
     GowDestGeometry geo;
{
  GowCutWidget pw= (GowCutWidget)w;
  Widget child0;

  if (!XtIsRealized(w) ||
      !pw->composite.num_children) { return; }
  
  child0= pw->composite.children[0];

  if (XtIsManaged(child0)) {
    GowPlotWidget(child0, fp, type, geo); }

  if (pw->composite.num_children == 3) {
    Widget child1= pw->composite.children[1];
    Dimension width, height;
    Position x, y;

    if (XtIsManaged(child1)) {
      GowDestGeometryRec geometry;

      if (XtIsManaged(child0) && XtIsRealized(child0)) {
	if (XtIsSubclass(child0, gowFrameWidgetClass)) {
	  GowFrameWidget fw= (GowFrameWidget)child0;
	  if (fw->frame_widget.plot_width && fw->frame_widget.plot_height) {
	    height= fw->frame_widget.plot_height * pw->cut_widget.height_ratio;
	    width= height * pw->cut_widget.width_to_height;
	    if (width > fw->frame_widget.plot_width) {
	      width= fw->frame_widget.plot_width; }
	    x= (fw->frame_widget.plot_x +
		fw->frame_widget.plot_width - width);
	    y= (fw->frame_widget.plot_y +
		fw->frame_widget.plot_height - height); } }
	else {
	  height= geo->y_range * pw->cut_widget.height_ratio;
	  width	= height * pw->cut_widget.width_to_height;
	  if (width > geo->x_range) { width= geo->x_range; }
	  x= geo->x_range - width;
	  y= geo->y_range - height; }
	if (width < 1 || height < 1) { return; } }
      else {
	height	= geo->y_range;
	width	= geo->x_range; }

      if (GowConfigureGeometry(&geometry, geo, type, x, y, width, height)) {
	GowPlotWidget(child1, fp, type, &geometry); } } }
}
