/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowDestX11.c,v 1.2 1994/05/26 19:49:09 vs Exp vs $ */

#include <Gow/GowConfig.h>

#ifdef HAS_XT_GRAPHIC

#include <math.h>
#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "GowDestP.h"
#include "GowGCsX11P.h"
#include "GowDestX11.h"

				/* basic draw functions */
#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

/* GowDestX11.c */
static Boolean Alabel P_((GowDestination d,
			  int x_justify, int y_justify, char *label));
static Boolean Arc P_((GowDestination d, int center_x, int center_y,
		       int start_x, int start_y, int end_x, int end_y));
static Boolean Circle P_((GowDestination d, int x, int y, int r));
static Boolean Comment P_((GowDestination d, char *comment));
static Boolean Cont P_((GowDestination d, int x, int y));
static Boolean Label P_((GowDestination d, char *label));
static Boolean Line P_((GowDestination d,
			int x_start, int y_start, int x_end, int y_end));
static Boolean Lines P_((GowDestination d, XPoint *points, int npoints));
static Boolean Move P_((GowDestination d, int x, int y));
static Boolean Point P_((GowDestination d, int x, int y));
static Boolean Points P_((GowDestination d, XPoint *points, int npoints));
static Boolean Segments P_((GowDestination d,
			    XSegment *segments, int nsegments));
static Boolean Set_angle360 P_((GowDestination d, int angle));
static Boolean Set_gc P_((GowDestination d, int idx));
static void Text_extents P_((GowDestination d, char *label,
			     int *height, int *width));
static int Linewidth P_((GowDestination d));
static int Fontheight P_((GowDestination d));
static int Fontwidth P_((GowDestination d));

#undef P_

static GowDestFuncsRec fct= {
  Alabel,
  Arc,
  Circle,
  Comment,
  Cont,
  Label,
  Line,
  Lines,
  Move,
  Point,
  Points,
  Segments,

  Comment,			/* start_object */
  Comment,			/* end_object */

  Set_angle360,
  Set_gc,
  Text_extents,

  Linewidth,
  Fontheight,
  Fontwidth };

/* ---------------------------------------------------------------------- */
Boolean GowCreateX11Destination(d, w)
     GowDestination d;
     Widget w;
{
  if (!XtIsRealized(w)) { return FALSE; }

  d->widget		= w;
  d->drawable		= (Drawable)XtWindow(w);
  d->display		= XtDisplay(w);

  d->x11.x_offset	= 0;	/* always constant */
  d->x11.y_offset	= 0;
  d->x11.x_direction	= 1;
  d->x11.y_direction	= -1;
  d->x11.fontscale	= 1.0;

  GowSetupX11Destination(d);

  return TRUE;
}
/* ---------------------------------------------------------------------- */
void GowDestroyX11Destination(d)
     GowDestination d;
{
  return;
}
/* ---------------------------------------------------------------------- */
void GowChangeX11Destination(d)
     GowDestination d;
{
  Widget w= d->widget;

  d->x11.x_range	= w->core.width;
  d->x11.y_range	= w->core.height;
}
/* ---------------------------------------------------------------------- */
void GowSetupX11Destination(d)
     GowDestination d;
{
  d->fct= &fct;
  GowChangeX11Destination(d);
}
/* ---------------------------------------------------------------------- */
/*
 * Functions of standard interface for drawing (GowDestination)
 */
/* ---------------------------------------------------------------------- */
static Boolean Alabel(d, x_justify, y_justify, label)
     GowDestination d;
     int x_justify;
     int y_justify;
     char *label;
{
  int width, height;
  Position x= d->actual_x;
  Position y= d->actual_y;
  int len= strlen(label);

  Text_extents(d, label, &height, &width);

  switch (x_justify) {
  case 'l': break;
  case 'c': x -= width / 2; break;
  case 'r': x -= width; break;
  default:  return FALSE; }

  switch (y_justify) {
  case 'b': break;
  case 'c': y += height / 2; break;
  case 't': y += height; break;
  default:  return FALSE; }

  XDrawString(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	      x, y, label, (int)len);

  d->actual_x= x + width;
  /* d->actual_y= unchanged */

  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Arc(d, center_x, center_y, start_x, start_y, end_x, end_y)
     GowDestination d;
     int center_x, center_y;
     int start_x, start_y;
     int end_x, end_y;
{
  int r = (int)sqrt((double)((center_x - start_x) * (center_x - start_x)
			     +(center_y - start_y) * (center_y - start_y)));

				/* 3666.93 = (180 * 64) / pi */
  XDrawArc(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	   center_x - r, center_y - r,
	   (unsigned int)(2 * r), (unsigned int)(2 * r),
	   (int)(atan2((double)(start_y - center_y),
		       (double)(start_x - center_x)) * 3666.93),
	   (int)(atan2((double)(end_y - center_y),
		       (double)(end_x - center_x)) * 3666.93));
  d->actual_x= center_x;
  d->actual_x= center_y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Circle(d, x, y, r)
     GowDestination d;
     int x, y, r;
{
				/* 23040 = 360 * 64 */
  XDrawArc(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	   x - r, y - r,
	   (unsigned int)(2 * r), (unsigned int)(2 * r),
	   (int)0, (int)23040);
  d->actual_x= x;
  d->actual_x= y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Comment(d, comment)
     GowDestination d;
     char *comment;
{
  return FALSE;			/* do nothing */
}
/* ---------------------------------------------------------------------- */
static Boolean Cont(d, x, y)
     GowDestination d;
     int x, y;
{
  XDrawLine(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	    d->actual_x, d->actual_y, x, y);
  d->actual_x = x;
  d->actual_y = y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Label(d, label)
     GowDestination d;
     char *label;
{
  return Alabel(d, 'l', 'c', label);
}
/* ---------------------------------------------------------------------- */
static Boolean Line(d, x_start, y_start, x_end, y_end)
     GowDestination d;
     int x_start, y_start, x_end, y_end;
{
  XDrawLine(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	    x_start, y_start, x_end, y_end);
  d->actual_x = x_end;
  d->actual_y = y_end;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Lines(d, points, npoints)
     GowDestination d;
     XPoint *points;
     int npoints;
{
  if (!npoints) { return FALSE; }
  XDrawLines(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	     points, npoints, CoordModeOrigin);
  d->actual_x = points[npoints-1].x;
  d->actual_y = points[npoints-1].y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Move(d, x, y)
     GowDestination d;
     int x, y;
{
  d->actual_x = x;
  d->actual_y = y;
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static Boolean Point(d, x, y)
     GowDestination d;
     int x, y;
{
  XDrawPoint(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc, x, y);
  d->actual_x = x;
  d->actual_y = y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Points(d, points, npoints)
     GowDestination d;
     XPoint *points;
     int npoints;
{
  if (!npoints) { return FALSE; }
  XDrawPoints(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
	     points, npoints, CoordModeOrigin);
  d->actual_x = points[npoints-1].x;
  d->actual_y = points[npoints-1].y;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Segments(d, segments, nsegments)
     GowDestination d;
     XSegment *segments;
     int nsegments;
{
  if (!nsegments) { return FALSE; }
  XDrawSegments(d->display, d->drawable, ((GowGCsX11)d->actual_gc)->gc,
		segments, nsegments);
  d->actual_x = segments[nsegments-1].x2;
  d->actual_y = segments[nsegments-1].y2;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean Set_angle360(d, angle)
     GowDestination d;
     int angle;
{
  return (angle == 0);
}
/* ---------------------------------------------------------------------- */
static Boolean Set_gc(d, idx)
     GowDestination d;
     int idx;
{
  if (d->gcs) {
    d->actual_idx	= idx % d->num_gcs;
    d->actual_gc	= (GowGC)(((GowGCsX11)d->gcs) + d->actual_idx);
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static void Text_extents(d, label, height, width)
     GowDestination d;
     char *label;
     int *height, *width;
{
  int direction, ascent, descent;
  XCharStruct overall;
  int sl= strlen(label);

  if (sl) {
    XTextExtents(((GowGCsX11)d->actual_gc)->font,
		 label, (int)strlen(label),
		 &direction, &ascent, &descent,
		 &overall);
    *height	= overall.ascent;
    *width	= overall.width; }
  else { *height= *width= 0; }
  return;
}
/* ---------------------------------------------------------------------- */
static int Linewidth(d)
     GowDestination d;
{
  XGCValues values;

  if (XGetGCValues(d->display, ((GowGCsX11)d->actual_gc)->gc,
		   (unsigned int)GCLineWidth, &values)) {
    return values.line_width; }
  else { return 0; }
}
/* ---------------------------------------------------------------------- */
static int Fontheight(d)
     GowDestination d;
{
  return (((GowGCsX11)d->actual_gc)->font->max_bounds.ascent +
	  ((GowGCsX11)d->actual_gc)->font->max_bounds.descent);
}
/* ---------------------------------------------------------------------- */
static int Fontwidth(d)
     GowDestination d;
{
  return ((GowGCsX11)d->actual_gc)->font->max_bounds.width;
}
/* ---------------------------------------------------------------------- */

#endif				/* HAS_XT_GRAPHIC */
