/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowFramWidget.c,v 1.3 1994/12/02 21:22:11 vs Exp vs $ */

#include <X11/IntrinsicP.h>
#include <X11/cursorfont.h>
#include <X11/StringDefs.h>

#include <X11/Xmu/Misc.h>
#include <X11/Xmu/Converters.h>

#ifdef HAS_XAW3D
#include <X11/Xaw3d/XawInit.h>
#include <X11/Xaw3d/Grip.h>
#else
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Grip.h>
#endif

#include <Gow/GowPaneGeo.h>
#include <Gow/GowPlot.h>

#include "GowWidgetP.h"

#include "GowFramWidgetP.h"

#include <ctype.h>

#define ForAllChildren(pw, childP) \
  for ( (childP) = (pw)->composite.children ; \
        (childP) < (pw)->composite.children + (pw)->composite.num_children ; \
        (childP)++ )

/*****************************************************************************
 *
 * Full instance record declaration
 *
 ****************************************************************************/

#define offset(field) XtOffsetOf(GowFrameWidgetRec, frame_widget.field)

static float def_short= 0.2;
static float def_medium= 0.4;
static float def_long= 0.6;
static float def_space= 0.75;
static float def_tic= 1.0;
static float def_lbl= 5.0;

static XtResource resources[] = {
  {XtNticLong, XtCTic, XtRFloat, sizeof(float),
     offset(tic[TIC_LONG]), XtRFloat, (XtPointer)&def_long},
  {XtNticMedium, XtCTic, XtRFloat, sizeof(float),
     offset(tic[TIC_MEDIUM]), XtRFloat, (XtPointer)&def_medium},
  {XtNticShort, XtCTic, XtRFloat, sizeof(float),
     offset(tic[TIC_SHORT]), XtRFloat, (XtPointer)&def_short},

  {XtNticSpace, XtCTicSpace, XtRFloat, sizeof(float),
     offset(tic[TIC_SPACE]), XtRFloat, (XtPointer)&def_space},

  {XtNticLabel, XtCTicLabel, XtRFloat, sizeof(float),
     offset(tic[TIC_LABEL]), XtRFloat, (XtPointer)&def_lbl},
  {XtNticTic, XtCTicTic, XtRFloat, sizeof(float),
     offset(tic[TIC_TIC]), XtRFloat, (XtPointer)&def_tic},

  {XtNleftFrame, XtCYFrame, XtRInt, sizeof(int),
     offset(frame[DrawAxisLeft]), XtRImmediate, (XtPointer)0},
  {XtNleftAxis, XtCAxis, XtRPointer, sizeof(DrawAxis_t),
     offset(axis[DrawAxisLeft]), XtRImmediate, (XtPointer)NULL},
  {XtNrightFrame, XtCYFrame, XtRInt, sizeof(int),
     offset(frame[DrawAxisRight]), XtRImmediate, (XtPointer)0},
  {XtNrightAxis, XtCAxis, XtRPointer, sizeof(DrawAxis_t),
     offset(axis[DrawAxisRight]), XtRImmediate, (XtPointer)NULL},
  {XtNbottomFrame, XtCXFrame, XtRInt, sizeof(int),
     offset(frame[DrawAxisBottom]), XtRImmediate, (XtPointer)0},
  {XtNbottomAxis, XtCAxis, XtRPointer, sizeof(DrawAxis_t),
     offset(axis[DrawAxisBottom]), XtRImmediate, (XtPointer)NULL},
  {XtNtopFrame, XtCXFrame, XtRInt, sizeof(int),
     offset(frame[DrawAxisTop]), XtRImmediate, (XtPointer)0},
  {XtNtopAxis, XtCAxis, XtRPointer, sizeof(DrawAxis_t),
     offset(axis[DrawAxisTop]), XtRImmediate, (XtPointer)NULL},
};

#undef offset

static void Layout(), DrawFrame();
static void CreateGCs(), DestroyGCs();
static void Initialize();
static void Realize(), Resize(), Redisplay();
static Boolean SetValues();
static XtGeometryResult GeometryManager();
static void ChangeManaged();
static void InsertChild();


/*
 * Action Routine Definitions
 */


static void CursorFull();
static void CursorLocalFull();
static void CursorShift();

/*
 * Actions list
 */

static XtActionsRec actions[] = {
  { "full",		CursorFull },
  { "local-full",	CursorLocalFull },
  { "shift",		CursorShift },
};

#define SuperClass ((GowDestinationWidgetClass)&gowDestinationWidgetClassRec)

GowFrameWidgetClassRec gowFrameWidgetClassRec = {
   {
/* core class fields */
    /* superclass         */   (WidgetClass) SuperClass,
    /* class name         */   "GowFrameWidget",
    /* size               */   sizeof(GowFrameWidgetRec),
    /* class_initialize   */   NULL,
    /* class_part init    */   NULL,
    /* class_inited       */   FALSE,
    /* initialize         */   Initialize,
    /* initialize_hook    */   NULL,
    /* realize            */   Realize,
    /* actions            */   actions,
    /* num_actions        */   XtNumber(actions),
    /* resources          */   resources,
    /* resource_count     */   XtNumber(resources),
    /* xrm_class          */   NULLQUARK,
    /* compress_motion    */   TRUE,
    /* compress_exposure  */   TRUE,
    /* compress_enterleave*/   TRUE,
    /* visible_interest   */   FALSE,
    /* destroy            */   DestroyGCs,
    /* resize             */   Resize,
    /* expose             */   Redisplay,
    /* set_values         */   SetValues,
    /* set_values_hook    */   NULL,
    /* set_values_almost  */   XtInheritSetValuesAlmost,
    /* get_values_hook    */   NULL,
    /* accept_focus       */   NULL,
    /* version            */   XtVersion,
    /* callback_private   */   NULL,
    /* tm_table           */   NULL,
    /* query_geometry	  */   GowQueryPaneGeometry,
    /* display_accelerator*/   XtInheritDisplayAccelerator,
    /* extension          */   NULL
   }, {
/* composite class fields */
    /* geometry_manager   */   GeometryManager,
    /* change_managed     */   ChangeManaged,
    /* insert_child       */   InsertChild,
    /* delete_child       */   XtInheritDeleteChild,
    /* extension          */   NULL
   }
};

WidgetClass gowFrameWidgetClass = (WidgetClass) &gowFrameWidgetClassRec;

/***********************************************************
 *
 * Private Functions.
 *
 ************************************************************/

static void Layout(fw)
     GowFrameWidget fw;
{
  Widget c0;

  if (XtIsRealized((Widget)fw) &&
      fw->composite.num_children &&
      XtIsManaged(c0= fw->composite.children[0])) {
    int left= 0, right= 0, bottom= 0, top= 0, fontsize= 0;
    int *frame_px= fw->frame_widget.frame_px;
    int *frame= fw->frame_widget.frame;

    if (GowUseGCs(fw->destination_widget.destination, fw->frame_widget.gcs)) {
      GowDestination dest= fw->destination_widget.destination;
      GowDSet_gc(dest, DrawAxisScaleGC);
      fontsize = MAX(GowDFontwidth(dest), GowDFontheight(dest)); }

    left= frame_px[DrawAxisLeft]= fontsize * frame[DrawAxisLeft];
    right= frame_px[DrawAxisRight]= fontsize * frame[DrawAxisRight];
    bottom= frame_px[DrawAxisBottom]= fontsize * frame[DrawAxisBottom];
    top= frame_px[DrawAxisTop]= fontsize * frame[DrawAxisTop];

    if (bottom >= fw->core.height) { bottom= top= 0; }
    else if (bottom + top >= fw->core.height) { top= 0; }

    if (left >= fw->core.width) { left= right= 0; }
    else if (left + right >= fw->core.width) { right= 0; }

    { Dimension height	= fw->core.height - bottom - top;
      Dimension width	= fw->core.width - left - right;

      if (width < 1) { width= 1; }
      if (height < 1) { height= 1; }

      fw->frame_widget.needs_redisplay
	= (left != fw->core.x || top != fw->core.y ||
	   width != fw->core.width || height != fw->core.height);

      XtConfigureWidget(c0, left, top, (unsigned)width, (unsigned)height,
			(unsigned)0);
      if (fw->frame_widget.needs_redisplay) {
	Redisplay((Widget)fw, (XEvent *)NULL, (Region)NULL); } } }
}

/************************************************************
 *
 * Semi-public routines. 
 *
 ************************************************************/

/* The Geometry Manager only allows no resize
 */

static XtGeometryResult GeometryManager(w, request, reply)
Widget w;
XtWidgetGeometry *request, *reply;
{
  return XtGeometryNo;		/* disallow any geometry change */
}

/* ARGSUSED */
static void Initialize(request, new)
Widget request, new;
{
  GowFrameWidget fw = (GowFrameWidget)new;

  { int i; for (i= 0; i< AXIS_NUM; i++) { fw->frame_widget.frame_px[i]= 0; } }

  if (!new->core.width) { new->core.width= DEFAULT_WIDTH; }
  if (!new->core.height) { new->core.height= DEFAULT_HEIGHT; }

  if (fw->frame_widget.frame[DrawAxisLeft] < 0) {
    fw->frame_widget.frame[DrawAxisLeft]= 0; }
  if (fw->frame_widget.frame[DrawAxisRight] < 0) {
    fw->frame_widget.frame[DrawAxisRight]= 0; }
  if (fw->frame_widget.frame[DrawAxisBottom] < 0) {
    fw->frame_widget.frame[DrawAxisBottom]= 0; }
  if (fw->frame_widget.frame[DrawAxisTop] < 0) {
    fw->frame_widget.frame[DrawAxisTop]= 0; }

  if (fw->frame_widget.tic[TIC_LONG] < 0.0) {
    fw->frame_widget.tic[TIC_LONG]= def_long; }
  if (fw->frame_widget.tic[TIC_MEDIUM] < 0.0) {
    fw->frame_widget.tic[TIC_MEDIUM]= def_medium; }
  if (fw->frame_widget.tic[TIC_SHORT] < 0.0) {
    fw->frame_widget.tic[TIC_SHORT]= def_short; }

  if (fw->frame_widget.tic[TIC_SPACE] < 0.0) {
    fw->frame_widget.tic[TIC_SPACE]= def_space; }

  if (fw->frame_widget.tic[TIC_LABEL] < 1.0) {
    fw->frame_widget.tic[TIC_LABEL]= def_lbl; }
  if (fw->frame_widget.tic[TIC_TIC] <= 0.0) {
    fw->frame_widget.tic[TIC_TIC]= def_tic; }

  fw->frame_widget.needs_redisplay= FALSE;
}

static void 
Realize(w, valueMask, attributes)
Widget w;
Mask *valueMask;
XSetWindowAttributes *attributes;
{
    GowFrameWidget fw = (GowFrameWidget) w;
    Widget *childP;

    (*SuperClass->core_class.realize) (w, valueMask, attributes);

    CreateGCs(w);
/*
 * Before we commit the new locations we need to realize all the panes and
 * their grips.
 */

    ForAllChildren(fw, childP) { XtRealizeWidget( *childP ); }

    Layout((GowFrameWidget)w);
} /* Realize */

/*	Function Name: DelayCallback
 *	Description: notifies that delaycount reached 0
 *	Arguments: gow - the no more delayed gow widget
 *                 client_data - registering frame_widget
 *                 call_data - ** NOT USED **
 *	Returns: none.
 */

/* ARGSUSED */
static void
DelayCallback(gow, clientData, callData)
Widget gow;
XtPointer clientData, callData;
{
  GowFrameWidget fw = (GowFrameWidget)clientData;
  if (fw->frame_widget.needs_redisplay) { Redisplay((Widget)fw); }
}

/*	Function Name: ScaleCallback
 *	Description: notifies a scale change
 *	Arguments: gow - the gow widget with a new scale
 *                 client_data - registering frame_widget
 *                 call_data - ** NOT USED **
 *	Returns: none.
 */

/* ARGSUSED */
static void
ScaleCallback(gow, clientData, callData)
Widget gow;
XtPointer clientData, callData;
{
  Redisplay((Widget)clientData);
}

static void InsertChild(w)
register Widget w;
{
  GowFrameWidget fw= (GowFrameWidget)XtParent(w);

  if (fw->composite.num_children == 1 || !XtIsSubclass(w, gowWidgetClass)) {
    XtDestroyWidget(w); return; }

  /* insert the child widget in the composite children list with the */
  /* superclass insert_child routine.                                */

  (*SuperClass->composite_class.insert_child)(w);

  XtAddCallback(w, XtNdelayCallback, DelayCallback, fw);
  XtAddCallback(w, XtNscaleCallback, ScaleCallback, fw);
} /* InsertChild */

static void ChangeManaged(w)
   Widget w;
{
   Layout((GowFrameWidget)w);

} /* ChangeManaged */

/*	Function Name: Resize
 *	Description: The paned widget's resize proc.
 *	Arguments: w - the paned widget.
 *	Returns: none.
 */

static void
Resize(w)
Widget w;
{
  Layout((GowFrameWidget)w);
}

/* ARGSUSED */
static Boolean 
SetValues(old, request, new)
Widget old, request, new;
{
  GowFrameWidget ow = (GowFrameWidget) old;
  GowFrameWidget nw = (GowFrameWidget) new;

  if (nw->frame_widget.frame[DrawAxisLeft] < 0) {
    nw->frame_widget.frame[DrawAxisLeft]
      = ow->frame_widget.frame[DrawAxisLeft]; }
  if (nw->frame_widget.frame[DrawAxisRight] < 0) {
    nw->frame_widget.frame[DrawAxisRight]
      = ow->frame_widget.frame[DrawAxisRight]; }
  if (nw->frame_widget.frame[DrawAxisBottom] < 0) {
    nw->frame_widget.frame[DrawAxisBottom]
      = ow->frame_widget.frame[DrawAxisBottom]; }
  if (nw->frame_widget.frame[DrawAxisTop] < 0) {
    nw->frame_widget.frame[DrawAxisTop]
      = ow->frame_widget.frame[DrawAxisTop]; }

  if (nw->frame_widget.tic[TIC_LONG] < 0) {
    nw->frame_widget.tic[TIC_LONG]= ow->frame_widget.tic[TIC_LONG]; }
  if (nw->frame_widget.tic[TIC_MEDIUM] < 0) {
    nw->frame_widget.tic[TIC_MEDIUM]= ow->frame_widget.tic[TIC_MEDIUM]; }
  if (nw->frame_widget.tic[TIC_SHORT] < 0) {
    nw->frame_widget.tic[TIC_SHORT]= ow->frame_widget.tic[TIC_SHORT]; }

  if (nw->frame_widget.tic[TIC_SPACE] < 0) {
    nw->frame_widget.tic[TIC_SPACE]= ow->frame_widget.tic[TIC_SPACE]; }

  if (nw->frame_widget.tic[TIC_LABEL] < 1.0) {
    nw->frame_widget.tic[TIC_LABEL]= ow->frame_widget.tic[TIC_LABEL]; }
  if (nw->frame_widget.tic[TIC_TIC] <= 0) {
    nw->frame_widget.tic[TIC_TIC]= ow->frame_widget.tic[TIC_TIC]; }

  if (memcmp(nw->frame_widget.frame, ow->frame_widget.frame,
	     AXIS_NUM * sizeof(int))) {
    Layout(nw); } /* full reconfiguration */
  else if (memcmp(nw->frame_widget.tic, ow->frame_widget.tic,
		  TIC_NUM * sizeof(float)) ||
	   memcmp(nw->frame_widget.axis, ow->frame_widget.axis,
		  AXIS_NUM * sizeof(DrawAxis_t))) {
    Redisplay(new); } /* redraw scales */
  return FALSE;
} /* SetValues */

/*      Function Name: Redisplay
 *      Description: Redisplays the contents of the widget.
 *      Arguments:	w - the gow widget.
 *              	event - the X event that caused this redisplay.
 *              	region - the region the needs to be repainted. 
 *				 NULL if full redisplay
 *      Returns:	none.
 */

/* ARGSUSED */
static void
Redisplay(w, event, region)
Widget w;
XEvent * event;
Region region;
{
  GowFrameWidget fw = (GowFrameWidget) w;
  GowWidget gow;

  fw->frame_widget.needs_redisplay= TRUE;
  if (!XtIsRealized(w)) { return; }
  XClearWindow(XtDisplay(w), XtWindow(w));

  if (!fw->composite.num_children) { return; }

  gow= (GowWidget)fw->composite.children[0];
  if (gow->gow_widget.delaycount) { return; }

  DrawFrame(fw, gow->core.x, gow->core.y);

  fw->frame_widget.needs_redisplay= FALSE;
}

/*	Function Name: CreateGCs
 *	Description: Creates the GC's for the frame.
 *	Arguments: w - the frame widget.
 *	Returns: none
 */

static void
CreateGCs(w)
Widget w;
{
  ((GowFrameWidget)w)->frame_widget.gcs= GowCreateGCs(w, 2);
}

static void
DestroyGCs(w)
Widget w;
{
  GowFrameWidget fw= (GowFrameWidget)w;
  GowDestroyGCs(fw->frame_widget.gcs);
}

static void DrawFrame(fw, x, y)
     GowFrameWidget fw;
     int x, y;
{
  GowWidget gow= (GowWidget)fw->composite.children[0];

  if (!GowInitializeWidgetScales((Widget)gow, TRUE, TRUE)) {
    return; }

  { GowDestination dest= fw->destination_widget.destination;
    GowDestination gdest= gow->destination_widget.destination;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;
    int wd= GowDXRange(gdest);
    int hg= GowDYRange(gdest);
    float *tic= fw->frame_widget.tic;
    DrawAxis_t *axis= fw->frame_widget.axis;
    int *frame_px= fw->frame_widget.frame_px;
    int x_fac, y_fac;
    int bw_off;

    if (!GowUseGCs(dest, fw->frame_widget.gcs)) { return; }
    GowDSet_gc(dest, DrawAxisScaleGC);

    x_fac=(GowGetScaleDirection(x_scale)==GowNormalScaleDirection) ? 1 : -1;
    y_fac=(GowGetScaleDirection(y_scale)==GowNormalScaleDirection) ? 1 : -1;
    bw_off= GowDLinewidth(dest);
    bw_off= bw_off ? ((bw_off + 1) / 2) : 1;
    DrawAxis(dest,
	     ((x_fac > 0) ? (x - bw_off) : (x + wd + bw_off)), y,
	     axis[DrawAxisLeft], DrawAxisLeft,
	     y_scale, tic, x_fac, frame_px[DrawAxisLeft]);
    DrawAxis(dest,
	     ((x_fac > 0) ? (x + wd + bw_off) : (x - bw_off)), y,
	     axis[DrawAxisRight], DrawAxisRight,
	     y_scale, tic, x_fac, frame_px[DrawAxisRight]);

    DrawAxis(dest,
	     x, ((y_fac > 0) ? (y - bw_off) : (y + hg + bw_off)),
	     axis[DrawAxisBottom], DrawAxisBottom,
	     x_scale, tic, y_fac, frame_px[DrawAxisBottom]);
    DrawAxis(dest,
	     x, ((y_fac > 0) ? (y + hg + bw_off) : (y - bw_off)),
	     axis[DrawAxisTop], DrawAxisTop,
	     x_scale, tic, y_fac, frame_px[DrawAxisTop]); }
}


/************************************************************
 *
 * Global Action Routines.
 * 
 * These actions routines will be added to the application's
 * global action list. 
 * 
 ************************************************************/

/************************************************************
 *
 * Widget Action Routines.
 * 
 ************************************************************/

/* XtActionProc */
static void CursorFull(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  GowFrameWidget fw= (GowFrameWidget)w;

  if (*num_params != 1) {
    XtError("GowWidget CursorFull has been passed incorrect parameter.");
    return; }
  if (!XtIsSubclass(w, gowFrameWidgetClass) ||
      fw->composite.num_children != 1){
    return; }
  w = fw->composite.children[0];
  if (XtIsSubclass(w, gowWidgetClass)) {
    int xf, yf;
    xf= (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL);
    yf= (strchr(params[0], 'y')  != NULL || strchr(params[0], 'Y') != NULL);

    GowWidgetFull(w, xf, yf); }
}

/* XtActionProc */
static void CursorLocalFull(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  GowFrameWidget fw= (GowFrameWidget)w;

  if (*num_params != 1) {
    XtError("GowWidget CursorLocalFull has been passed incorrect parameter.");
    return; }
  if (!XtIsSubclass(w, gowFrameWidgetClass) ||
      fw->composite.num_children != 1){
    return; }
  w = fw->composite.children[0];
  if (XtIsSubclass(w, gowWidgetClass)) {
    int xf, yf;
    xf= (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL);
    yf= (strchr(params[0], 'y')  != NULL || strchr(params[0], 'Y') != NULL);

    GowWidgetLocalFull(w, xf, yf); }
}

				/* XtActionProc */
static void CursorShift(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  double x_perc= 0.0, y_perc= 0.0;
  GowFrameWidget fw= (GowFrameWidget)w;

  if (*num_params != 2 ||
      sscanf(params[0], "%lf", &x_perc) != 1 ||
      sscanf(params[1], "%lf", &y_perc) != 1) {
    XtError("GowWidget CursorShift has been passed incorrect parameters.");
    return; }
  if (!XtIsSubclass(w, gowFrameWidgetClass) ||
      fw->composite.num_children != 1){
    return; }
  w = fw->composite.children[0];
  if (XtIsSubclass(w, gowWidgetClass)) {
    if (x_perc != 0.0 || y_perc != 0.0) {
      GowWidgetShift(w, x_perc, y_perc); }}
}

/************************************************************
 *
 * Public routines. 
 *
 ************************************************************/

void GowFrameRedisplay(w, type)
     Widget w;
     int type;
{
  GowFrameWidget fw= (GowFrameWidget)w;

  if (0 <= type && type < AXIS_NUM && !fw->frame_widget.frame[type]) {
    return; }
  else { Redisplay(w, (XEvent *)NULL, (Region)NULL); }
}

void GowFrameWidgetPlot(w, fp, type, geo)
     Widget w;
     GowDestFile fp;
     int type;
     GowDestGeometry geo;
{
  GowFrameWidget fw= (GowFrameWidget)w;
  Widget c0;

  fw->frame_widget.plot_x	= 0;
  fw->frame_widget.plot_y	= 0;
  fw->frame_widget.plot_width	= 0;
  fw->frame_widget.plot_height	= 0;

  if (XtIsRealized(w) &&
      fw->composite.num_children &&
      XtIsManaged(c0= fw->composite.children[0])) {
    int left= 0, right= 0, bottom= 0, top= 0, fontsize= 0;
    int height, width;
    int *frame_px= fw->frame_widget.frame_px;
    int *frame= fw->frame_widget.frame;
    GowDestination dest= fw->destination_widget.destination;
    GowDestGeometryRec geometry;

    if (GowSetupPlotDestination(dest, type, fp, geo)) {
      if (GowUseGCs(dest, fw->frame_widget.gcs)) {
	GowDSet_gc(dest, DrawAxisScaleGC);
	fontsize = MAX(GowDFontwidth(dest), GowDFontheight(dest));

	left= frame_px[DrawAxisLeft]= fontsize * frame[DrawAxisLeft];
	right= frame_px[DrawAxisRight]= fontsize * frame[DrawAxisRight];
	bottom= frame_px[DrawAxisBottom]= fontsize * frame[DrawAxisBottom];
	top= frame_px[DrawAxisTop]= fontsize * frame[DrawAxisTop];

	if (bottom >= geo->y_range) {
	  frame_px[DrawAxisBottom]= bottom= 0;
	  frame_px[DrawAxisTop]= top= 0; }
	else if (bottom + top >= geo->y_range) {
	  frame_px[DrawAxisTop]= top= 0; }

	if (left >= geo->x_range) {
	  frame_px[DrawAxisLeft]= left= 0;
	  frame_px[DrawAxisRight]= right= 0; }
	else if (left + right >= geo->x_range) {
	  frame_px[DrawAxisRight]= right= 0; }

	if (left || right || bottom || top) {
	  height= geo->y_range - bottom - top;
	  width	= geo->x_range - left - right;

	  if (GowConfigureGeometry(&geometry, geo, type,
				   left, bottom, width, height)) {
	    GowDestination cdest
	      = ((GowWidget)c0)->destination_widget.destination;
	    if (GowSetupPlotDestination(cdest, type, fp, &geometry)) {
	      DrawFrame(fw, GowDXOffset(cdest), GowDYOffset(cdest));
	      GowFinishPlotDestination(cdest); } } } }
      GowFinishPlotDestination(dest); }

    height	= geo->y_range - bottom - top;
    width	= geo->x_range - left - right;

    if (GowConfigureGeometry(&geometry, geo, type,
			     left, bottom, width, height)) {
      fw->frame_widget.plot_x		= left;
      fw->frame_widget.plot_y		= bottom;
      fw->frame_widget.plot_width	= width;
      fw->frame_widget.plot_height	= height;
      GowPlotWidget(c0, fp, type, &geometry); } }
}
