/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowGCsX11.c,v 1.2 1994/05/26 19:49:19 vs Exp vs $ */

/*
 * GowGCs.c - gow GCs.
 *
 */

#include <Gow/GowConfig.h>

#ifdef HAS_XT_GRAPHIC

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/Xmu/Initer.h>
#include <X11/Xmu/CharSet.h>

#include <tv/vsAlloc.h>

#include "GowGCsX11P.h"

static Boolean InitializeResources(), GetDashes();

static XrmQuark	XtQSolid, XtQLongdashed, XtQDisconnected, XtQDotdashed,
  XtQDotted, XtQShortdashed, XtQOdddashed;

#define MAXNAMELEN		256
static int max_num_gcs= 0;
static XtResourceList gc_resource_list= NULL;

#define SetGCResources(i) {\
  sprintf(Nfont, "%s%d", XtNfont, i); \
  sprintf(Nforeground, "%s%d", XtNforeground, i); \
  sprintf(Nbackground, "%s%d", XtNbackground, i); \
  sprintf(Nlinewidth, "%s%d", XtNlinewidth, i); \
  sprintf(Nlinestyle, "%s%d", XtNlinestyle, i); }

typedef struct _GowGCResources {
  XFontStruct	*font;
  Pixel		foreground;
  Pixel		background;
  Dimension	linewidth;
  GowLineStyle	linestyle;
} GowGCResources;

#define Offset(field) XtOffsetOf(GowGCResources,field) 

static XtResource gc_resources[] = {
  {NULL, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
     Offset(font), XtRString, XtDefaultFont},
  {NULL, XtCForeground, XtRPixel, sizeof (Pixel),
     Offset(foreground), XtRString, XtDefaultForeground},
  {NULL, XtCBackground, XtRPixel, sizeof (Pixel),
     Offset(background), XtRString, XtDefaultBackground}, 
  {NULL, XtCLinewidth, XtRDimension, sizeof(Dimension),
     Offset(linewidth), XtRImmediate, (XtPointer) 0},
  {NULL, XtCLinestyle, XtRLineStyle, sizeof(GowLineStyle),
     Offset(linestyle), XtRImmediate, (XtPointer) GowLineSolid },
};

#undef Offset

/* ---------------------------------------------------------------------- */
GowGCsX11 GowCreateGCsX11(w, num)
     Widget w;
     int num;
{
  GowGCsX11 x11;

  if ((!gc_resource_list && !InitializeResources(num)) ||
      num > max_num_gcs) {
    return NULL; }
  else if ((x11= (GowGCsX11)Calloc(num, sizeof(GowGCsX11Rec)))) {
    GowGCResources *gc_base;
    if ((gc_base= (GowGCResources *)Malloc(num * sizeof(GowGCResources)))) {

      XtGetSubresources(XtParent(w), gc_base, XtName(w),
			w->core.widget_class->core_class.class_name,
			gc_resource_list, num * XtNumber(gc_resources),
			NULL, 0);

      { Display *dpy;
	Screen *scr;
	Window wdw;
	Widget ancestor;
	XGCValues values;
	XtGCMask  mask= (GCFont | GCLineWidth  | GCLineStyle |
			 GCForeground | GCBackground);

	dpy= XtDisplayOfObject(w);
	scr= XtScreenOfObject(w);
	wdw= RootWindowOfScreen(scr);
	ancestor= _XtWindowedAncestor(w);

	if (XDefaultDepthOfScreen(scr) == 1) {
	  values.foreground	= ~w->core.background_pixel;
	  values.background	= w->core.background_pixel; }

	{ int i;
	  GowGCsX11 gx;
	  for (i= 0, gx= x11; i< num; i++, gx++) {
	    GC gc;
	    char *dash_list;
	    int nlist;

	    if (XDefaultDepthOfScreen(scr) != 1) {
	      values.foreground	= gc_base[i].foreground;
	      values.background	= gc_base[i].background; }
	    values.font		= gc_base[i].font->fid;
	    values.line_width	= gc_base[i].linewidth;

	    if (!GetDashes(gc_base[i].linestyle, &dash_list, &nlist)) {
	      values.line_style= LineSolid; }
	    else if (values.background == ancestor->core.background_pixel) {
	      values.line_style= LineOnOffDash; }
	    else { values.line_style= LineDoubleDash; }

	    if (gc_base[i].linestyle != GowLineSolid) {
	      if ((gc= XCreateGC(dpy, wdw, mask, &values))) {
		XSetDashes(dpy, gc, dash_list[0], dash_list, nlist); }
	      else { break; } }	/* no new gc created*/
	    else if (!(gc= XtGetGC(w, mask, &values))) {
	      break; }		/* no new gc got */

	    gx->gc_created	= (gc_base[i].linestyle != GowLineSolid);
	    gx->font		= gc_base[i].font;
	    gx->gc		= gc; }
	  Free(gc_base);
	  if (i == num) { return x11; }
	  GowDestroyGCsX11(w, x11, i); } } } }
  return NULL;
}
/* ---------------------------------------------------------------------- */
void GowDestroyGCsX11(w, x11, num)
     Widget w;
     GowGCsX11 x11;
     int num;
{
  int i;
  GowGCsX11 x;
  Display *dpy= XtDisplayOfObject(w);

  if (x11) {
    for (i= 0, x= x11; i< num; i++, x++) {
      if (x->gc_created) { XFreeGC(dpy, x->gc); }
      else { XtReleaseGC(w, x->gc); } }
    Free(x11); }
}
/* ---------------------------------------------------------------------- */
Boolean GowInitializeGCsX11(num)
     int num;
{
  if (gc_resource_list) {
    if (num > max_num_gcs) { return FALSE; } }
  else if (num > max_num_gcs) { max_num_gcs= num; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static Boolean InitializeResources(num)
     int num;
{
  if (num > max_num_gcs) { max_num_gcs= num; }

  if (!(gc_resource_list=
	(XtResourceList)Calloc(max_num_gcs, sizeof(gc_resources)))) {
    return FALSE; }
  else {
    int i;
    XtResourceList gc_i;
    extern void _XtCompileResourceList();
    char Nfont[MAXNAMELEN];
    char Nforeground[MAXNAMELEN];
    char Nbackground[MAXNAMELEN];
    char Nlinewidth[MAXNAMELEN];
    char Nlinestyle[MAXNAMELEN];

    XtQSolid		= XrmPermStringToQuark(GowEsolid);
    XtQLongdashed	= XrmPermStringToQuark(GowElongdashed);
    XtQDisconnected	= XrmPermStringToQuark(GowEdisconnected);
    XtQDotdashed	= XrmPermStringToQuark(GowEdotdashed);
    XtQDotted		= XrmPermStringToQuark(GowEdotted);
    XtQShortdashed	= XrmPermStringToQuark(GowEshortdashed);
    XtQOdddashed	= XrmPermStringToQuark(GowEodddashed);

    XtAddConverter( XtRString, XtRLineStyle,
		   GowCvtStringToLineStyle, NULL, 0 );

    gc_resources[0].resource_name= Nfont;
    gc_resources[1].resource_name= Nforeground;
    gc_resources[2].resource_name= Nbackground;
    gc_resources[3].resource_name= Nlinewidth;
    gc_resources[4].resource_name= Nlinestyle;

    for (i = 0, gc_i= gc_resource_list;
	 i < max_num_gcs;
	 i++, gc_i += XtNumber(gc_resources)) {

      SetGCResources(i);

      memcpy(gc_i, gc_resources, sizeof(gc_resources));

      { int g;			/* compile resources of each gc in place */

	for (g= 0; g< XtNumber(gc_resources); g++) {
	  int value_offset= i * sizeof(GowGCResources);

	  if (!(gc_i[g].resource_name=
		Strdup(gc_resources[g].resource_name))) {
	    for (g--; g>= 0; g--) { Free(gc_i[g].resource_name); }
	    for (i--, gc_i -= XtNumber(gc_resources); i>= 0;
		 i--, gc_i -= XtNumber(gc_resources)) {
	      for (g= 0; g< XtNumber(gc_resources); g++) {
		Free(gc_i[g].resource_name); } }
	    Free(gc_resource_list);
	    return FALSE; }
	      
	  gc_i[g].resource_offset
	    = gc_resources[g].resource_offset + value_offset; } }

      _XtCompileResourceList(gc_i, XtNumber(gc_resources)); }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
#define	done(address, type) \
	{ toVal->size = sizeof(type); \
	  toVal->addr = (caddr_t) address; \
	  return; \
	}

/* ARGSUSED */
void GowCvtStringToLineStyle(args, num_args, fromVal, toVal)
    XrmValuePtr args;		/* unused */
    Cardinal    *num_args;      /* unused */
    XrmValuePtr fromVal;
    XrmValuePtr toVal;
{
    static GowLineStyle lineStyle;
    XrmQuark q;
    char lowerName[1000];

    XmuCopyISOLatin1Lowered (lowerName, (char*)fromVal->addr);
    q = XrmStringToQuark(lowerName);
    if (q == XtQSolid) {
      lineStyle = GowLineSolid;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQLongdashed) {
      lineStyle = GowLineLongdashed;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQDisconnected) {
      lineStyle = GowLineDisconnected;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQDotdashed) {
      lineStyle = GowLineDotdashed;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQDotted) {
      lineStyle = GowLineDotted;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQShortdashed) {
      lineStyle = GowLineShortdashed;
      done(&lineStyle, GowLineStyle); }
    if (q == XtQOdddashed) {
      lineStyle = GowLineOdddashed;
      done(&lineStyle, GowLineStyle); }
    XtStringConversionWarning(fromVal->addr, "LineStyle");
    toVal->addr = NULL;
    toVal->size = 0;
}
/* ---------------------------------------------------------------------- */
static Boolean GetDashes(linestyle, dash_list, nlist)
     GowLineStyle linestyle;
     char **dash_list;
     int *nlist;
{
  typedef struct _GowLinestyleRec {
    char	*dash_list;
    int		nlist;
  } GowLinestyleRec;

  static char longdashed_line[]		= {8, 14};
  static char disconnected_line[]	= {12, 2};
  static char dotdashed_line[]		= {4, 10, 4, 2};
  static char dotted_line[]		= {6, 2};
  static char shortdashed_line[]	= {8, 8};
  static char odddashed_line[]		= {2, 6, 4};

  static GowLinestyleRec gowLinestyles[] = {
    { NULL, 0 },
    { longdashed_line, sizeof(longdashed_line) },
    { disconnected_line, sizeof(disconnected_line) },
    { dotdashed_line, sizeof(dotdashed_line) },
    { dotted_line, sizeof(dotted_line) },
    { shortdashed_line, sizeof(shortdashed_line) },
    { odddashed_line, sizeof(odddashed_line) },
  };

  switch (linestyle) {
  case GowLineLongdashed:
  case GowLineDisconnected:
  case GowLineDotdashed:
  case GowLineDotted:
  case GowLineShortdashed:
  case GowLineOdddashed:
    *dash_list= gowLinestyles[linestyle].dash_list;
    *nlist= gowLinestyles[linestyle].nlist;
    return TRUE;
  case GowLineSolid:
  default:
    *dash_list= NULL;
    *nlist= 0;
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
#endif				/* HAS_XT_GRAPHIC */
