/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowObject.c,v 1.2 1994/05/26 19:49:25 vs Exp vs $ */

/*
 * GowObject.c - gow object. (For use with the gow widget).
 *
 */

#include <Gow/GowConfig.h>

#ifdef HAS_XT_GRAPHIC

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <tv/vsAlloc.h>

#include "GowWidgetP.h"
#include "GowObjectP.h"

/****************************************************************
 *
 * Full class record constant
 *
 ****************************************************************/

static void ClassPartInitialize(), InitializeObject(), DestroyObject();
static Boolean SetValues();

static void RedisplayObject(), EraseObject();
static Boolean InsertSubobject();
static GowAppDrawdata CreateDrawdata();
static void DestroyDrawdata();
static Boolean InsertHere();

static void DestroyMarkedSubobjects();

#define Offset(field) XtOffsetOf(GowObjectRec,gow_object.field)

static XtResource resources[] = {
  { XtNnumGCs, XtCNumGCs, XtRInt, sizeof (int),
    Offset(num_gcs), XtRImmediate, (XtPointer) 1},
  { XtNxStepWidth, XtCStepWidth, XtRDimension, sizeof(Dimension),
    Offset(x_step_width), 
    XtRImmediate, (XtPointer) 1},
  { XtNxMinSteps, XtCMinSteps, XtRDimension, sizeof(Dimension),
    Offset(x_min_steps), 
    XtRImmediate, (XtPointer) 1},
  { XtNyStepWidth, XtCStepWidth, XtRDimension, sizeof(Dimension),
    Offset(y_step_width), 
    XtRImmediate, (XtPointer) 1},
  { XtNyMinSteps, XtCMinSteps, XtRDimension, sizeof(Dimension),
    Offset(y_min_steps), 
    XtRImmediate, (XtPointer) 1},
};

#undef Offset

#define SUPERCLASS		(&rectObjClassRec)

GowObjectClassRec gowObjectClassRec = {
  {
/* core_class fields */	
    /* superclass	  	*/	(WidgetClass) SUPERCLASS,
    /* class_name	  	*/	"GowObject",
    /* widget_size	  	*/	sizeof(GowObjectRec),
    /* class_initialize   	*/	NULL,
    /* class_part_initialize	*/	ClassPartInitialize,
    /* class_inited       	*/	FALSE,
    /* initialize	  	*/	InitializeObject,
    /* initialize_hook		*/	NULL,
    /* realize		  	*/	NULL,
    /* actions		  	*/	NULL,
    /* num_actions  		*/	0,
    /* resources	  	*/	resources,
    /* num_resources	  	*/	XtNumber(resources),
    /* xrm_class	  	*/	NULLQUARK,
    /* compress_motion	  	*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest	  	*/	FALSE,
    /* destroy		  	*/	DestroyObject,
    /* resize		  	*/	NULL,
    /* expose		  	*/	RedisplayObject,
    /* set_values	  	*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus	 	*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private   	*/	NULL,
    /* tm_table		   	*/	NULL,
    /* query_geometry		*/	NULL,
    /* display_accelerator	*/	NULL,
    /* extension		*/	NULL
  },
/* gow_object_class fields */
  {
    /* EraseObject		*/	EraseObject,
    /* InsertSubobject		*/	InsertSubobject,
    /* CreateDrawdata		*/	CreateDrawdata,
    /* DestroyDrawdata		*/	DestroyDrawdata,
    /* InsertHere		*/	InsertHere,
    /* ExposeSubobject		*/	NULL,
    /* ExpandFullScale		*/	NULL,
    /* uses_rectangles		*/	FALSE
  }
};

WidgetClass gowObjectClass = (WidgetClass)&gowObjectClassRec;

static void
ClassPartInitialize(wc)
WidgetClass wc;
{
  register GowObjectClass t_src, superC;

  t_src = (GowObjectClass) wc;
  superC = (GowObjectClass)t_src->rect_class.superclass;
/* 
 * We don't need to check for null super since we'll get to GowObject
 * eventually.
 */

  if (t_src->gow_object_class.EraseObject == XtInheritEraseObject) {
    t_src->gow_object_class.EraseObject
      = superC->gow_object_class.EraseObject; }

  if (t_src->gow_object_class.InsertSubobject == XtInheritInsertSubobject) {
    t_src->gow_object_class.InsertSubobject
      = superC->gow_object_class.InsertSubobject; }

  if (t_src->gow_object_class.CreateDrawdata == XtInheritCreateDrawdata) {
    t_src->gow_object_class.CreateDrawdata
      = superC->gow_object_class.CreateDrawdata; }

  if (t_src->gow_object_class.DestroyDrawdata == XtInheritDestroyDrawdata) {
    t_src->gow_object_class.DestroyDrawdata
      = superC->gow_object_class.DestroyDrawdata; }

  if (t_src->gow_object_class.InsertHere == XtInheritInsertHere) {
    t_src->gow_object_class.InsertHere
      = superC->gow_object_class.InsertHere; }
}

/*	Function Name: InitializeObject
 *	Description: Initializes the GowObject.
 *	Arguments: request, new - the requested and new values for the object
 *                                instance.
 *	Returns: none.
 *
 */

/* ARGSUSED */
static void
InitializeObject(request, new)
Widget request, new;
{
  GowObject goo= (GowObject)new;

  if (goo->gow_object.x_step_width < 1) { goo->gow_object.x_step_width= 1; }
  if (goo->gow_object.x_min_steps < 1) { goo->gow_object.x_min_steps= 1; }
  if (goo->gow_object.y_step_width < 1) { goo->gow_object.y_step_width= 1; }
  if (goo->gow_object.y_min_steps < 1) { goo->gow_object.y_min_steps= 1; }

  goo->gow_object.gcs			= NULL;

  goo->gow_object.subobjects		= NULL;
  goo->gow_object.visible		= FALSE;
  goo->gow_object.needs_update		= FALSE;
  goo->gow_object.managed		= FALSE;

  GowInitializeGCs(goo->gow_object.num_gcs);
}

/*	Function Name: DestroyObject
 *	Description: This function cleans up when the object is 
 *                   destroyed.
 *	Arguments: w - the gow object.
 *	Returns: none.
 */

static void
DestroyObject(w)
     Widget w;
{
  GowObject goo = (GowObject)w;

  GowDestroyGCs(goo->gow_object.gcs);

  { GowSubobject sub;

    ForAllSubobjects(goo, sub) {
      sub->status= GowDestroySubobj; } }

  DestroyMarkedSubobjects(goo);
}

/*	Function Name: RedisplayObject
 *	Description: Paints the GowSubobject's
 *	Arguments: w - the gow object.
 *                 event, region - NOT USED.
 *	Returns: none
 */

/*ARGSUSED*/
static void
RedisplayObject(w, event, region)
Widget w;
XEvent * event;
Region region;
{
  GowObject goo = (GowObject)w;
  GowWidget gow = (GowWidget)goo->object.parent;
  GowObjectClass class= (GowObjectClass)goo->object.widget_class;
  GowSubobjDrawFunc redisplay= class->gow_object_class.ExposeSubobject;

  if (!GowIsManaged(w) ||	/* only managed objects become drawn */
      (!goo->gow_object.needs_update &&
       !region)) { /* nothing to redisplay */
    return; }

  if (redisplay &&	/* draw if redisplay-function & GCs available */
      (goo->gow_object.gcs ||
       (goo->gow_object.gcs= GowCreateGCs(w, goo->gow_object.num_gcs))) &&
      GowUseGCs(gow->destination_widget.destination, goo->gow_object.gcs)) {
    GowSubobject sub;
    Boolean visible= FALSE;
    GowDestination dest= gow->destination_widget.destination;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;

    ForAllSubobjects(goo, sub) {
      if ((region || !sub->drawn) && sub->status == GowDrawSubobj) {
	visible |= (sub->visible= redisplay((GowDestination)dest,
					    x_scale, y_scale,
					    sub->drawdata, region));
	sub->drawn= TRUE; } }
    goo->gow_object.visible |= visible; }
  goo->gow_object.needs_update= FALSE;
}

/*      Function Name: SetValues
 *      Description: checks changes of num_gcs on validity.
 *      Arguments: current - current state of the widget.
 *                 request - what was requested.
 *                 new - what the widget will become.
 *      Returns: TRUE - redisplay, FALSE - do nothing
 */

/* ARGSUSED */
static Boolean
SetValues(current, request, new)
Widget current, request, new;
{
  GowObject goo = (GowObject)new;
  GowObject old_goo = (GowObject)current;

  goo->gow_object.num_gcs = old_goo->gow_object.num_gcs;

  if (goo->gow_object.x_step_width < 1) { goo->gow_object.x_step_width= 1; }
  if (goo->gow_object.x_min_steps < 1) { goo->gow_object.x_min_steps= 1; }
  if (goo->gow_object.y_step_width < 1) { goo->gow_object.y_step_width= 1; }
  if (goo->gow_object.y_min_steps < 1) { goo->gow_object.y_min_steps= 1; }

  return GowSetValuesReturn(new, 
			    (goo->gow_object.x_step_width !=
			     old_goo->gow_object.x_step_width ||
			     goo->gow_object.x_min_steps !=
			     old_goo->gow_object.x_min_steps ||
			     goo->gow_object.y_step_width !=
			     old_goo->gow_object.y_step_width ||
			     goo->gow_object.y_min_steps !=
			     old_goo->gow_object.y_min_steps));
}

/************************************************************
 *
 * Class specific methods.
 *
 ************************************************************/

/*	Function Name: EraseObject
 *	Description: notifies GowSubobject's on erase
 *	Arguments: w - the gow object.
 *                 event, region - NOT USED.
 *	Returns: none
 */

static void
EraseObject(w)
Widget w;
{
  GowObject goo = (GowObject)w;
  GowWidget gow= (GowWidget)goo->object.parent;
  GowSubobject sub;
  Boolean gow_update= FALSE;

  ForAllSubobjects(goo, sub) {
    if (sub->status == GowDrawSubobj) {
      gow_update= goo->gow_object.needs_update= TRUE; }
    sub->visible	= FALSE;
    sub->drawn		= FALSE; }
  if (gow_update) { gow->gow_widget.needs_update= TRUE; }
}

/*
 *	methods operating on subobjects and subobject data
 */


static GowAppDrawdata CreateDrawdata(new)
     GowAppData new;
{
  return (GowAppDrawdata)new;
}

static void DestroyDrawdata(old)
     GowAppDrawdata old;
{
  return;
}

static Boolean InsertHere(new, previous, next)
     GowAppData new;
     GowAppData previous;
     GowAppData next;
{
  return TRUE;
}

/*
 *	Function Name: DestroyMarkedSubobjects
 *	Description: removes destroyed GowSubobject's from subobject list
 *	Arguments: goo - the gow object.
 *	Returns: none
 */

static void DestroyMarkedSubobjects(goo)
     GowObject goo;
{
  GowObjectClass class= (GowObjectClass)goo->object.widget_class;
  GowAppDataDestroyProc destroy_drawdata
    = class->gow_object_class.DestroyDrawdata;
  GowSubobject sub;
  GowSubobject previous= NULL;

  for (sub= goo->gow_object.subobjects; sub; ) {
    GowSubobject next= sub->next;

    if (sub->status == GowDestroySubobj) {
      destroy_drawdata(sub->drawdata);
      Free(sub);
      sub= next;
      if (previous) { previous->next= next; } }
    else { previous= sub; } }
}

/*
 *	Function Name: InsertSubobject
 *	Description: inserts new application data in subobject list
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - initialization for subobject status
 *	Returns: TRUE - subobject inserted, FALSE - alloc failed
 */

static Boolean InsertSubobject(w, data, status)
     Widget w;
     GowAppData data;
     GowSubobjStat status;
{
  GowObject goo = (GowObject)w;
  GowObjectClass class= (GowObjectClass)goo->object.widget_class;
  GowSubobjHereFunc insert_here= class->gow_object_class.InsertHere;
  GowAppDataCreateFunc create_drawdata= class->gow_object_class.CreateDrawdata;
  GowSubobject new;

  if (status == GowDestroySubobj) { return FALSE; }
  else if ((new= (GowSubobject)Malloc(sizeof(GowSubobjectRec)))) {
    if ((new->drawdata= create_drawdata(data))) {
      new->appdata	= data;
      new->drawn	= FALSE;
      new->visible	= FALSE;
      new->status	= status; }
    else { Free(new); return FALSE; } } /* drawdata creation failed */
  else { return FALSE; }

  if (goo->gow_object.subobjects) {
    GowSubobject sub;

    if (insert_here(new->drawdata, NULL, goo->gow_object.subobjects)) {
      new->next= goo->gow_object.subobjects;
      goo->gow_object.subobjects= new; }
    else {
      ForAllSubobjects(goo, sub) {
	if (!sub->next || insert_here(new->drawdata,
				      sub->drawdata,
				      sub->next->drawdata)) {
	  new->next= sub->next;
	  sub->next= new;
	  break; } } } }
  else {
    goo->gow_object.subobjects= new; new->next= NULL; }
    
  return TRUE;
}

/*************************************************************
 *
 * public functions
 *
 *************************************************************/

/*
 *	Function Name: GowInsertSubobject
 *	Description: inserts new application data in subobject list
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - initialization for subobject status
 *	Returns: TRUE - subobject inserted, FALSE - alloc failed
 */

Boolean GowInsertSubobject(w, data, status)
     Widget w;
     GowAppData data;
     GowSubobjStat status;
{
  GowObject goo = (GowObject)w;
  GowObjectClass class= (GowObjectClass)goo->object.widget_class;
  GowSubobjFunc insert= class->gow_object_class.InsertSubobject;

  if (insert(w, data, status)) {
    if (status == GowDrawSubobj) { GowUpdateObjectWidget(w); }
    return TRUE; }
  else { return FALSE; }
}

/*
 *	Function Name: GowChangeSubobjectStat
 *	Description: changes status of subobject belonging to application data
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - new subobject status
 *	Returns: 	TRUE - subobject found, FALSE - subobject not found
 */

Boolean GowChangeSubobjectStat(w, data, status)
     Widget w;
     GowAppData data;
     GowSubobjStat status;
{
  GowObject goo= (GowObject)w;
  GowSubobject sub;
  Boolean redisplay= FALSE;
  Boolean update= FALSE;
  Boolean found= FALSE;

  ForAllSubobjects(goo, sub) {
    if (!data || sub->appdata == data) {
      found= TRUE;
      if (status == GowDrawSubobj) { update= TRUE; sub->drawn= FALSE; }
      if (sub->status != status) {
	if (sub->visible && status != GowDrawSubobj) { redisplay= TRUE; }
	sub->status= status; }
      if (data) { break; } } }

  if (status == GowDestroySubobj) { DestroyMarkedSubobjects(goo); }

  if (redisplay) { GowRedisplayObjectWidget(w); }
  else if (update) { GowUpdateObjectWidget(w); }

  return found;
}

/*
 *	Function Name: GowChangeSubobjectStat
 *	Description: changes status of subobject belonging to application data
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - new subobject status
 *	Returns: 	TRUE - subobject found, FALSE - subobject not found
 */

Boolean GowGetSubobjectStat(w, data, status, visible)
     Widget w;
     GowAppData data;
     GowSubobjStat *status;
     Boolean *visible;
{
  GowObject goo= (GowObject)w;
  GowSubobject sub;
  Boolean found= FALSE;

  *status= GowHideSubobj;
  *visible= FALSE;

  ForAllSubobjects(goo, sub) {
    if (sub->appdata == data) {
      found	= TRUE;
      *status	= sub->status;
      *visible	= sub->visible;
      break; } }

  return found;
}

/*
 *	Function Name:	GowRedisplayObjectWidget
 *	Description:	redisplays parent widget if object is managed
 *	Arguments:	w - the gow object.
 *	Returns:	none
 */

void GowRedisplayObjectWidget(w)
     Widget w;
{
  if (GowIsManaged(w)) {
    GowObject goo = (GowObject)w;

    GowRedisplayWidget(goo->object.parent); }
}
  
/*
 *	Function Name:	GowUpdateObjectWidget
 *	Description:	draws the gow object if it is managed
 *	Arguments:	w - the gow object.
 *	Returns:	none
 */

void GowUpdateObjectWidget(w)
     Widget w;
{
  if (GowIsManaged(w)) {
    GowObject goo= (GowObject)w;

    goo->gow_object.needs_update= TRUE;
    GowUpdateWidget(goo->object.parent); }
}

/*      Function Name: ExpandFullScaleOfObject
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow object.
 *			x/y_scale - the gow x/y-scale
 *			x/y_type -  type of x/y-scale
 *			x/y_full - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

void
GowExpandFullScaleOfObject(w, x_scale, x_full, y_scale, y_full)
Widget w;
GowScale x_scale;
int x_full;
GowScale y_scale;
int y_full;
{
  GowObject goo = (GowObject)w;
  GowObjectClass class= (GowObjectClass)goo->object.widget_class;
  GowSubobjScaleProc expand= class->gow_object_class.ExpandFullScale;
  GowScaleType x_type= GowGetScaleType(x_scale);
  GowScaleType y_type= GowGetScaleType(y_scale);

  if (!GowIsManaged(w) || !expand) { return; }

  { GowSubobject sub;

    ForAllSubobjects(goo, sub) {
      if (sub->status == GowDrawSubobj) {
	expand(sub->drawdata,
	       x_scale, x_type, x_full,
	       y_scale, y_type, y_full); } } }
}

Boolean GowIsManaged(w)
     Widget w;
{
  return (XtIsManaged(w) && ((GowObject)w)->gow_object.managed);
}

void GowManageChildren(children, num_children)
     WidgetList children;
     Cardinal num_children;
{
  WidgetList child= children;
  int i;
  Boolean update= FALSE;

  for (i= 0; i< num_children; i++, child++) {
    if (XtIsSubclass(*child, gowObjectClass)) {
      GowObject goo= (GowObject)*child;

      if (!goo->gow_object.managed) {
	update= TRUE; goo->gow_object.managed= TRUE; } } }
      
  XtManageChildren(children, num_children);
  if (update) { GowUpdateObjectWidget(*children); }
}

void GowManageChild(child)
     Widget child;
{
  GowManageChildren(&child, 1);
}

void GowUnmanageChildren(children, num_children)
     WidgetList children;
     Cardinal num_children;
{
  WidgetList child= children;
  int i;
  Boolean redisplay= FALSE;

  for (i= 0; i< num_children; i++, child++) {
    if (XtIsManaged(*child) && XtIsSubclass(*child, gowObjectClass)) {
      GowObject goo= (GowObject)*child;

      if (goo->gow_object.managed) {
	goo->gow_object.managed= FALSE;
	if (goo->gow_object.visible) { redisplay= TRUE; } } } }
      
  if (redisplay) { GowRedisplayWidget(((GowObject)*children)->object.parent); }
}

void GowUnmanageChild(child)
     Widget child;
{
  GowUnmanageChildren(&child, 1);
}

/*      Function Name: GowSetValuesReturn
 *      Description: Avoids multiple expose requests by different GowObject's
 *      Arguments:	w - the gow object.
 *			redisplay - TRUE if expose requested
 *      Returns: none.
 */

Boolean GowSetValuesReturn(w, redisplay)
     Widget w;
     int redisplay;
{
  GowWidget gow= (GowWidget)((GowObject)w)->object.parent;
  if (redisplay &&
      GowIsManaged(w) &&
      !gow->gow_widget.needs_redisplay) {
    gow->gow_widget.needs_redisplay= TRUE; /* forces redisplay of parent */
    if (!gow->gow_widget.set_values_expose) {
      return (gow->gow_widget.set_values_expose= TRUE); } }
  return FALSE;
}

#endif				/* HAS_XT_GRAPHIC */
