/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowObject.h,v 1.2 1994/05/26 19:49:25 vs Exp vs $ */

#ifndef _GowObject_h
#define _GowObject_h

/***********************************************************************
 *
 * GraphicObject Object
 *
 ***********************************************************************/

#include <X11/Object.h>
#include <X11/Xfuncproto.h>

#include <Gow/GowConfig.h>
#include <Gow/GowScale.h>

/* Resources:

 Name			Class		RepType		Default Value
 ----			-----		-------		-------------
 numGCs			NumGCs		int		1 ( <= GOW_MAXGCS)
 xMinSteps		MinSteps	int		1
 xStepWidth		StepWidth	int		1
 yMinSteps		MinSteps	int		1
 yStepWidth		StepWidth	int		1

*/
 
				/* resource names */
#define XtNnumGCs	"numGCs"

#define XtNxMinSteps	"xMinSteps"
#define XtNxStepWidth	"xStepWidth"
#define XtNyMinSteps	"yMinSteps"
#define XtNyStepWidth	"yStepWidth"
				/* resource class names */
#define XtCNumGCs	"NumGCs"

#define XtCMinSteps	"MinSteps"
#define XtCStepWidth	"StepWidth"

/* Class record constants */

extern WidgetClass gowObjectClass;

typedef struct _GowObjectClassRec	*GowObjectClass;
typedef struct _GowObjectRec		*GowObject;
typedef struct _GowSubobjectRec		*GowSubobject;

typedef void				*GowAppDrawdata;
typedef void				*GowAppData;

typedef enum {
  GowDestroySubobj,
  GowHideSubobj,
  GowDrawSubobj
} GowSubobjStat;

/************************************************************
 *
 * Public Functions.
 *
 ************************************************************/

_XFUNCPROTOBEGIN

/*	Function Name: GowSetMaxGCsOfObjectClass
 *	Description:	initializes num_gcs of GowObjectClass to a value > 1 
 *	Arguments:	object_class - the gow object class.
 *			num - max. number of GC's requested with this class.
 *	Returns:	none
 */

extern void GowSetNumGCsOfObjectClass(
#if NeedFunctionPrototypes
     WidgetClass		/* object_class */,
     int			/* num */
#endif
);

/*
 *	Function Name: GowInsertSubobject
 *	Description: inserts new application data in subobject list
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - initialization for subobject status
 *	Returns: TRUE - subobject inserted, FALSE - alloc failed
 */

extern Boolean GowInsertSubobject(
#if NeedFunctionPrototypes
     Widget			/* w */,
     GowAppData			/* data */,
     GowSubobjStat		/* status */
#endif
);

/*
 *	Function Name: GowChangeSubobjectStat
 *	Description: changes status of subobject belonging to application data
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - new subobject status
 *	Returns: TRUE - subobject found, FALSE - subobject not found
 */

extern Boolean GowChangeSubobjectStat(
#if NeedFunctionPrototypes
     Widget			/* w */,
     GowAppData			/* data */,
     GowSubobjStat		/* status */
#endif
);

/*
 *	Function Name: GowGetSubobjectStat
 *	Description: changes status of subobject belonging to application data
 *	Arguments:	w - the gow object.
 *			data - application data for subobject
 *			status - subobject status
 *			visible - subobject is visible
 *	Returns: TRUE - subobject found, FALSE - subobject not found
 */

extern Boolean GowGetSubobjectStat(
#if NeedFunctionPrototypes
     Widget			/* w */,
     GowAppData			/* data */,
     GowSubobjStat*		/* status */,
     Boolean*			/* visible */
#endif
);

/*
 *	Function Name:	GowRedisplayObjectWidget
 *	Description:	redisplays parent widget if object is managed
 *	Arguments:	w - the gow object.
 *	Returns:	none
 */

extern void GowRedisplayObjectWidget(
#if NeedFunctionPrototypes
     Widget			/* w */
#endif
);

/*
 *	Function Name:	GowUpdateObjectWidget
 *	Description:	draws the gow object if it is managed
 *	Arguments:	w - the gow object.
 *	Returns:	none
 */

extern void GowUpdateObjectWidget(
#if NeedFunctionPrototypes
    Widget			/* w */
#endif
);

/*      Function Name: ExpandFullScaleOfObject
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow object.
 *			x/y_scale - the gow x/y-scale
 *			x/y_full - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

extern void GowExpandFullScaleOfObject(
#if NeedFunctionPrototypes
    Widget			/* w */,
    GowScale			/* x_scale */,
    int				/* x_full */,
    GowScale			/* y_scale */,
    int				/* y_full */
#endif
);

extern Boolean GowIsManaged(
#if NeedFunctionPrototypes
    Widget			/* w */
#endif
);

void GowManageChildren(
#if NeedFunctionPrototypes
    WidgetList			/* children */,
    Cardinal			/* num_children */
#endif
);

void GowManageChild(
#if NeedFunctionPrototypes
    Widget			/* child */
#endif
);

void GowUnmanageChildren(
#if NeedFunctionPrototypes
    WidgetList			/* children */,
    Cardinal			/* num_children */
#endif
);

void GowUnmanageChild(
#if NeedFunctionPrototypes
    Widget			/* child */
#endif
);

/*      Function Name: GowSetValuesReturn
 *      Description: Avoids multiple expose requests by different GowObject's
 *      Arguments:	w - the gow object.
 *			redisplay - TRUE if expose requested
 *      Returns: none.
 */

Boolean GowSetValuesReturn(
#if NeedFunctionPrototypes
     Widget			/* w */,
			   int			/* redisplay */
#endif
);

_XFUNCPROTOEND

#endif				/* _GowObject_h */
