/* $Header: /usr/files/home/Gamma/vs/tv/edit/Gow/RCS/GowObjectP.h,v 1.1 1993/09/17 14:56:03 vs Exp vs $ */

/* 
 * GowObjectP.h - Private definitions for GowObject object
 * 
 */

#ifndef _GowObjectP_h
#define _GowObjectP_h

/***********************************************************************
 *
 * GowObject object Private Data
 *
 ***********************************************************************/

#include <X11/RectObjP.h>

#include <Gow/GowObject.h>
#include <Gow/GowGCs.h>

/************************************************************
 *
 * New fields for the GowObject object class record.
 *
 ************************************************************/

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

typedef GowAppDrawdata	(*GowAppDataCreateFunc) P_((GowAppData new));
typedef void	(*GowAppDataDestroyProc) P_((GowAppDrawdata old));
typedef Boolean	(*GowSubobjFunc) P_((Widget w, GowAppData data,
				     GowSubobjStat status));
typedef GowSubobjStat (*GowSubobjStatFunc) P_((GowObject obj,
					       GowSubobject sub));
typedef Boolean (*GowSubobjDrawFunc) P_((GowDestination dest,
					 GowScale x_scale, GowScale y_scale,
					 GowAppDrawdata data, Region region));
typedef Boolean	(*GowSubobjHereFunc) P_((GowAppDrawdata new, 
					 GowAppDrawdata previous,
					 GowAppDrawdata next));
typedef void	(*GowSubobjScaleProc) P_((GowAppDrawdata drawdata,
					  GowScale x_scale,
					  GowScaleType x_type, int x_full,
					  GowScale y_scale,
					  GowScaleType y_type, int y_full));
#undef P_

typedef struct _GowObjectClassPart {
				/* methods operating on gow objects */
  XtWidgetProc		EraseObject; /* notify gow object on erase */
  GowSubobjFunc		InsertSubobject; /* insert new subobject */
				/* methods operating on gow subobjects */
  GowAppDataCreateFunc	CreateDrawdata; /* create subobject draw struct */
  GowAppDataDestroyProc	DestroyDrawdata; /* destroy subobject draw struct */
  GowSubobjHereFunc	InsertHere; /* insert new subobject in between */
  GowSubobjDrawFunc	ExposeSubobject; /* redraw subobject */
  GowSubobjScaleProc	ExpandFullScale; /* expand scale for subobject */
  Boolean		uses_rectangles;
} GowObjectClassPart;

/* Full class record declaration */
typedef struct _GowObjectClassRec {
  RectObjClassPart	rect_class;
  GowObjectClassPart	gow_object_class;
} GowObjectClassRec;

extern GowObjectClassRec gowObjectClassRec;

/* instance record for subobjects of a GowObject */

typedef struct _GowSubobjectRec {
  Boolean		drawn;
  Boolean		visible;
  GowSubobjStat		status;
  GowAppData		appdata;
  GowAppDrawdata	drawdata;
  struct _GowSubobjectRec *next;
} GowSubobjectRec;

/* New fields for the GowObject object record */

typedef struct {
  /* resources */
  int		num_gcs;	/* number of GC's */
  Dimension	x_min_steps;
  Dimension	x_step_width;
  Dimension	y_min_steps;
  Dimension	y_step_width;
  /* private state. */
  GowDestGCs	gcs;		/* array of GC's */

  GowSubobject	subobjects;	/* subobjects */
  Boolean	visible;	/* gow object is visible */
  Boolean	needs_update;	/* object must become drawn */
  Boolean	managed;
} GowObjectPart;

/****************************************************************
 *
 * Full instance record declaration
 *
 ****************************************************************/

typedef struct _GowObjectRec {
  ObjectPart	object;
  RectObjPart	rectangle;
  GowObjectPart	gow_object;
} GowObjectRec;

/************************************************************
 *
 * Private declarations.
 *
 ************************************************************/

#define ForAllSubobjects(gow, sub) \
  for ((sub)= (gow)->gow_object.subobjects; (sub); (sub)= (sub)->next)

#define XtInheritEraseObject		((XtWidgetProc) _XtInherit)
#define XtInheritInsertSubobject	((GowSubobjFunc) _XtInherit)
#define XtInheritCreateDrawdata		((GowAppDataCreateFunc) _XtInherit)
#define XtInheritDestroyDrawdata	((GowAppDataDestroyProc) _XtInherit)
#define XtInheritInsertHere		((GowSubobjHereFunc) _XtInherit)
#define XtInheritExposeSubobject	((GowSubobjDrawFunc) _XtInherit)
#define XtInheritExpandFullScale	((GowSubobjScaleProc) _XtInherit)

#endif /* _GowObjectP_h */
