/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowScale.c,v 1.2 1994/05/26 19:49:43 vs Exp vs $ */

#define _GowScale_c_

#include <math.h>
#include <string.h>
#include <stdio.h>
				/* library include files */
#include <tv/vsAlloc.h>

#include "GowScaleP.h"

#define GowNull(fct) fprintf(stderr, \
			    "%s: illegal NULL-pointer in argument list\n", fct)
#define GowError(err)	fprintf(stderr, "%s\n", err)

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

static GowScaleConversion ErrorPixel P_((GowScale scale,
					 int *pixel, double user));
static GowScaleConversion ErrorUser P_((GowScale scale,
					int pixel, double *user));
static void InitializeError P_((GowScale scale, char *message));
static GowScaleConversion LinearPixel P_((GowScale scale,
					  int *pixel, double user));
static GowScaleConversion LinearPixelInv P_((GowScale scale,
					     int *pixel, double user));
static GowScaleConversion LinearUser P_((GowScale scale,
					 int pixel, double *user));
static GowScaleConversion LinearUserInv P_((GowScale scale,
					    int pixel, double *user));
static GowScaleConversion LogarithmicPixel P_((GowScale scale,
					       int *pixel, double user));
static GowScaleConversion LogarithmicPixelInv P_((GowScale scale,
						  int *pixel, double user));
static GowScaleConversion LogarithmicUser P_((GowScale scale,
					      int pixel, double *user));
static GowScaleConversion LogarithmicUserInv P_((GowScale scale,
						 int pixel, double *user));
static GowScaleConversion SquaredPixel P_((GowScale scale,
					       int *pixel, double user));
static GowScaleConversion SquaredPixelInv P_((GowScale scale,
						  int *pixel, double user));
static GowScaleConversion SquaredUser P_((GowScale scale,
					      int pixel, double *user));
static GowScaleConversion SquaredUserInv P_((GowScale scale,
						 int pixel, double *user));
static GowScaleConversion Range0Pixel P_((GowScale scale,
					  int *pixel, double user));
static GowScaleConversion Range0User P_((GowScale scale,
					 int pixel, double *user));
#undef P_

#define SQRED(x)	((x) * ABS(x))

/*
 * global functions operating on scale (GowScale)
 */

#ifdef HAS_XT_GRAPHIC
#include <X11/Intrinsic.h>

void GowDeleteScaleWidget(scale, w)
     GowScale scale;
     Widget w;
{
  if (!scale) { GowNull("GowDeleteScaleWidget"); }
  else {
    int i;

    for (i= 0; i< scale->num_widgets; i++) {
      if (w == scale->widgets[i].widget) {
	while (scale->widgets[i].delaycount > 0) {
	  scale->widgets[i].delaycount--;
	  scale->widgets[i].delay(w, FALSE); }
	if (--(scale->widgets[i].linkcount) == 0) {
	  scale->num_widgets--;
	  if (i < scale->num_widgets) {
	    memmove(scale->widgets + i, scale->widgets + i + 1, (size_t)
		    (sizeof(GowScaleNotifyRec) * (scale->num_widgets - i))); }
	  if (!scale->num_widgets && scale->being_destroyed) {
	    GowDestroyScale(scale); } }
	return; } } }
  return;
}

/* ARGSUSED */
static void Notify(w, scale)	/* if no notify required */
     Widget w;
     GowScale scale;
{
  return;
}

/* ARGSUSED */
static void Delay(w, delay)	/* if no delay required */
     Widget w;
     Boolean delay;
{
  return;
}

/* ARGSUSED */
static void Inherit(w, const_scale, full_scale)	/* if no inherit required */
     Widget w;
     GowScale const_scale, full_scale;
{
  return;
}

Boolean GowInsertScaleWidget(scale, w, notify, delay, inherit)
     GowScale scale;
     Widget w;
     GowScaleNotifyProc notify;
     GowScaleDelayProc delay;
     GowScaleInheritProc inherit;
{
  if (!scale) { GowNull("GowInsertScaleWidget"); return FALSE; }
  else if (!w || !notify) { /* notify not possible */
    return FALSE; }

  if (!delay) { delay= (GowScaleDelayProc)Delay; }
  if (!notify) { notify= (GowScaleNotifyProc)Notify; }
  if (!inherit) { inherit= (GowScaleInheritProc)Inherit; }

  { GowScaleNotifyRec *entry;

    ForAllWidgets(scale, entry) {
      if (entry->widget == w) { entry->linkcount++; return TRUE; } } }

  if (scale->num_widgets == scale->num_slots) {
    GowScaleNotifyRec *widgets;
    if ((widgets= (GowScaleNotifyRec *)Realloc(scale->widgets,
					      (size_t)
					      (sizeof(GowScaleNotifyRec)
					       * (scale->num_slots + 1))))) {
      scale->widgets= widgets;
      scale->num_slots++; }
    else { return FALSE; } }
  scale->widgets[scale->num_widgets].widget		= w;
  scale->widgets[scale->num_widgets].notify		= notify;
  scale->widgets[scale->num_widgets].delay		= delay;
  scale->widgets[scale->num_widgets].inherit		= inherit;
  scale->widgets[scale->num_widgets].delaycount		= 0;
  scale->widgets[scale->num_widgets].linkcount		= 1;
  scale->num_widgets++;
  return TRUE;
}
#endif				/* HAS_XT_GRAPHIC */

GowScale GowCreateScale()
{
  GowScale scale;

  if ((scale= (GowScale)Malloc(sizeof(GowScaleRec)))) {
    scale->widgets		= NULL;
    scale->num_widgets		= 0;
    scale->num_slots		= 0;
    scale->direction		= GowNormalScaleDirection;
    scale->offset_defined	= FALSE;
    scale->range_defined	= FALSE;
    scale->type			= GowLinearScale;

    scale->being_destroyed	= FALSE;

    scale->error_printed	= FALSE;
    scale->error		= NULL;

    GowResetNewScale(scale);

    return scale; }
  else { return NULL; }
}

void GowDelayScaleUpdate(scale, delay_output)
     GowScale scale;
     int delay_output;
{
  GowScaleNotifyRec *entry;
  ForAllWidgets(scale, entry) { entry->delay(entry->widget, delay_output); }
}

void GowInheritFullScale(const_scale, full_scale)
     GowScale const_scale;
     GowScale full_scale;
{
  GowScaleNotifyRec *entry;
  ForAllWidgets(const_scale, entry) {
    entry->inherit(entry->widget, const_scale, full_scale); }
}

void GowDestroyScale(scale)
     GowScale scale;
{
  if (!scale) { ; }
  else if (scale->num_widgets == 0) {
    Free(scale->widgets);
    Free(scale); }
  else { scale->being_destroyed= TRUE; }
  return;
}

void GowExtendNewScaleOffsetRange(scale, offset, range)
     GowScale scale;
     double offset;
     double range;
{
  if (!scale) { GowNull("GowExtendNewScaleOffsetRange"); }
  else if (scale->new_offset_defined) {
    double new_max= scale->new_offset + scale->new_range;

    if (offset + range > new_max) { new_max= offset + range; }
    if (offset < scale->new_offset) { scale->new_offset= offset; }
    scale->new_range= new_max - scale->new_offset; }
  else {
    scale->new_value_type	= GowExplicitScale;
    scale->new_offset		= offset;
    scale->new_range		= range;
    scale->new_range_defined	= TRUE;
    scale->new_offset_defined	= TRUE; }
  return;
}

void GowExtendNewScaleRange(scale, range)
     GowScale scale;
     double range;
{
  if (!scale) { GowNull("GowExtendScaleRange"); return; }
  else if (scale->new_range_defined) {
    if (range < scale->new_range) { scale->new_range= range; } }
  else {
    scale->new_value_type	= GowExplicitScale;
    scale->new_range		= range;
    scale->new_range_defined	= TRUE; }
  return;
}

GowScaleDirection GowGetScaleDirection(scale)
     GowScale scale;
{
  return scale->direction;
}
int GowGetScaleLimitPixel(scale, conversion)
     GowScale scale;
     int conversion;
{
  switch (conversion) {
  case GowToSmall:
    switch (scale->direction) {
    case GowNormalScaleDirection: return scale->pixel_min;
    case GowInvertedScaleDirection: return scale->pixel_max;
    default: return -1; }
  case GowToLarge:
    switch (scale->direction) {
    case GowNormalScaleDirection: return scale->pixel_max;
    case GowInvertedScaleDirection: return scale->pixel_min;
    default: return -1; }
  case GowClipSmall:
    switch (scale->direction) {
    case GowNormalScaleDirection: return scale->clip_min;
    case GowInvertedScaleDirection: return scale->clip_max;
    default: return -1; }
  case GowClipLarge:
    switch (scale->direction) {
    case GowNormalScaleDirection: return scale->clip_max;
    case GowInvertedScaleDirection: return scale->clip_min;
    default: return -1; }
  default: return -1; }
}

Boolean GowGetScaleOffset(scale, offset)
     GowScale scale;
     double *offset;
{
  if (!scale) { GowNull("GowGetScaleOffset"); return FALSE; }
  else if (scale->offset_defined) { *offset= scale->offset; }
  return scale->offset_defined;
}

Boolean GowGetScaleRange(scale, range)
     GowScale scale;
     double *range;
{
  if (!scale) { GowNull("GowGetScaleRange"); return FALSE; }
  else if (scale->range_defined) { *range= scale->range; }
  return scale->range_defined;
}

GowScaleType GowGetScaleType(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowGetScaleRange"); return GowLinearScale; }
  else { return scale->type; }
}

GowScaleType GowGetNewScaleType(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowGetScaleRange"); return GowLinearScale; }
  else { return scale->new_type; }
}

Boolean GowNewScaleFull(scale)
     GowScale scale;
{
  return (scale->new_value_type == GowFullScale);
}

GowPixelFunc GowGetScalePixelFunction(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowGetScalePixelFunction"); return NULL; }
  else { return scale->pixel; }
}

GowUserFunc GowGetScaleUserFunction(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowGetScaleUserFunction"); return NULL; }
  else { return scale->user; }
}

void GowGetScalePixelRange(scale, pixel_min, pixel_max)
     GowScale scale;
     int *pixel_min;
     int *pixel_max;
{
  if (!scale || !pixel_min || !pixel_max) {
    GowNull("GowGetScalePixelRange"); return; }
  else {
    *pixel_min= scale->pixel_min;
    *pixel_max= scale->pixel_max;
    return; }
}

void GowGetScaleUserRange(scale, user_min, user_max)
     GowScale scale;
     double *user_min;
     double *user_max;
{
  if (!scale || !user_min || !user_max) {
    GowNull("GowGetScaleUserRange"); return; }
  else {
    *user_min= scale->user_min;
    *user_max= scale->user_max;
    return; }
}



Boolean GowInitializeScale(scale, direction, offset, range, clip_min, clip_max)
     GowScale scale;
     GowScaleDirection direction;
     int offset;
     int range;
     double clip_min;
     double clip_max;
{
  if (!scale) { GowNull("GowInitializeScale"); return FALSE; }
  else if (--range == 0 || !scale->range_defined ||
      !scale->offset_defined) {
    InitializeError(scale, "GowInitializeScale: undefined offset or range");
    return FALSE; }
  else if (scale->offset <= 0 && scale->type == GowLogarithmicScale) {
    InitializeError(scale, "GowInitializeScale: illegal log(x < 0)");
    return FALSE; }
  else if (clip_min + clip_max >= 1.0 ||
      clip_min < 0.0 || 1.0 <= clip_min ||
      clip_max < 0.0 || 1.0 <= clip_max) {
    InitializeError(scale,
		    "GowInitializeScale: illegal clipping range specified");
    return FALSE; }
  else {			/* range limit setup */
    int clip_range;

    if (direction == GowInvertedScaleDirection) {
      double tmp= clip_min; clip_min= clip_max; clip_max= tmp; }

    scale->direction		= direction;
    scale->user_min		= scale->offset;
    scale->user_max		= scale->offset + scale->range;
    scale->pixel_min		= offset;
    scale->pixel_max		= offset + range;
    scale->clip_min		= scale->pixel_min + NINT(clip_min * range);
    scale->clip_max		= scale->pixel_max - NINT(clip_max * range);
    clip_range= scale->clip_max - scale->clip_min; 

    if (scale->range == 0.0) {
      scale->pixel_offset	= (scale->clip_min + scale->clip_max) / 2;
      scale->user_offset	= scale->offset;
      scale->pixel		= Range0Pixel;
      scale->user		= Range0User;
      return TRUE; }

    switch (scale->type) {
    case GowLinearScale:	/* function parameter setup */
      scale->user_offset	= scale->offset;
      scale->user_factor	= clip_range / scale->range;
      scale->user_factor_inv	= scale->range / clip_range;
				/* function setup */
      switch (direction) {
      case GowNormalScaleDirection:
	scale->pixel		= LinearPixel;
	scale->user		= LinearUser;
	scale->pixel_offset	= scale->clip_min;
	scale->user_min		= (((scale->pixel_min - scale->clip_min)
				    * scale->user_factor_inv)
				   + scale->user_offset);
	scale->user_max		= (((scale->pixel_max - scale->clip_min)
				    * scale->user_factor_inv)
				   + scale->user_offset);
	break;
      case GowInvertedScaleDirection:
	scale->pixel		= LinearPixelInv;
	scale->user		= LinearUserInv;
	scale->pixel_offset	= scale->clip_max;
	scale->user_min		= (((scale->clip_max - scale->pixel_max)
				    * scale->user_factor_inv)
				   + scale->user_offset);
	scale->user_max		= (((scale->clip_max - scale->pixel_min)
				    * scale->user_factor_inv)
				   + scale->user_offset);
	break;
      default:
	InitializeError(scale,
			"GowInitializeScale: illegal direction specified"); }
      break;
    case GowLogarithmicScale:	/* function parameter setup */
      scale->user_offset	= log(scale->offset);
      scale->user_factor	= (clip_range /
				   (log(scale->user_max) -
				    log(scale->user_min)));
      scale->user_factor_inv	= 1.0 / scale->user_factor;
				/* function setup */
      switch (direction) {
      case GowNormalScaleDirection:
	scale->pixel		= LogarithmicPixel;
	scale->user		= LogarithmicUser;
	scale->pixel_offset	= scale->clip_min;
	scale->user_min		= exp(((scale->pixel_min - scale->clip_min)
				       * scale->user_factor_inv)
				      + scale->user_offset);
	scale->user_max		= exp(((scale->pixel_max - scale->clip_min)
				       * scale->user_factor_inv)
				      + scale->user_offset);
	break;
      case GowInvertedScaleDirection:
	scale->pixel		= LogarithmicPixelInv;
	scale->user		= LogarithmicUserInv;
	scale->pixel_offset	= scale->clip_max;
	scale->user_min		= exp(((scale->clip_max - scale->pixel_max)
				       * scale->user_factor_inv)
				      + scale->user_offset);
	scale->user_max		= exp(((scale->clip_max - scale->pixel_min)
				       * scale->user_factor_inv)
				      + scale->user_offset);
	break;
      default:
	InitializeError(scale,
			"GowInitializeScale: illegal direction specified"); }
      break;
    case GowSquaredScale:	/* function parameter setup */
      scale->user_offset	= SQRED(scale->offset);
      scale->user_factor	= (clip_range /
				   (SQRED(scale->user_max) -
				    SQRED(scale->user_min)));
      scale->user_factor_inv	= 1.0 / scale->user_factor;
				/* function setup */
      switch (direction) {
      case GowNormalScaleDirection:
	scale->pixel		= SquaredPixel;
	scale->user		= SquaredUser;
	scale->pixel_offset	= scale->clip_min;
	{ double tmp= (((scale->pixel_min - scale->clip_min)
			* scale->user_factor_inv) + scale->user_offset);
	  scale->user_min	= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp); }
	{ double tmp= (((scale->pixel_max - scale->clip_min)
			* scale->user_factor_inv) + scale->user_offset);
	  scale->user_max	= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp); }
	break;
      case GowInvertedScaleDirection:
	scale->pixel		= SquaredPixelInv;
	scale->user		= SquaredUserInv;
	scale->pixel_offset	= scale->clip_max;
	{ double tmp= (((scale->clip_max - scale->pixel_max)
			* scale->user_factor_inv) + scale->user_offset);
	  scale->user_min	= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp); }
	{ double tmp= (((scale->clip_max - scale->pixel_min)
			* scale->user_factor_inv) + scale->user_offset);
	  scale->user_max	= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp); }
	break;
      default:
	InitializeError(scale,
			"GowInitializeScale: illegal direction specified"); }
      break;
    default:
      InitializeError(scale, "GowInitializeScale: illegal scale type");
      return FALSE; }
    return TRUE; }
}

Boolean GowNewScale(scale)
     GowScale scale;
{
  double new_max;
  double new_offset;
  double new_range;
  Boolean new_defined= FALSE;
  Boolean old_defined= scale->offset_defined && scale->range_defined;

  if (old_defined) { new_max= scale->offset + scale->range; }

  switch (scale->new_value_type) {
  case GowMinValue:
    if (old_defined &&		/* reference defined */
	new_max >= scale->new_value[0]) { /* legal */
      new_offset	= scale->new_value[0];
      new_range		= new_max - scale->new_value[0];
      new_defined	= TRUE; }
    break;
  case GowMaxValue:
    if (old_defined &&		/* reference defined */
	scale->offset < scale->new_value[0]) { /* legal */
      new_offset	= scale->offset;
      new_range		= scale->new_value[0] - scale->offset;
      new_defined	= TRUE; }
    break;
  case GowFifoValue:
    if (scale->num_new_values > 1 && /* enough values */
	scale->new_value[0] != scale->new_value[1]) { /* values differ */
      new_offset	= MIN(scale->new_value[0], scale->new_value[1]);
      new_range		= ABS(scale->new_value[0] - scale->new_value[1]);
      new_defined	= TRUE; }
    break;
  case GowLeaveScale:		/* will be checked after end of switch */
    if (old_defined) {		/* reference defined */
      new_offset	= scale->offset;
      new_range		= scale->range;
      new_defined	= TRUE; }
    break;
  case GowExplicitScale:
    if (scale->new_offset_defined && scale->new_range_defined) {
      new_offset	= scale->new_offset;
      new_range		= scale->new_range;
      new_defined	= TRUE; }
    break;			/* will be checked after end of switch too */
  case GowFullScale:
  default:
    old_defined= FALSE;
    break; }			/* end of switch */

  if (new_defined) {
    switch (scale->new_type) {
    case GowLogarithmicScale:
      if (new_offset <= 0) { /* illegal value for log() */
	scale->type= scale->new_type;
	scale->offset_defined= scale->range_defined= FALSE;
	return FALSE; }
      break;
    case GowSquaredScale:
    case GowLinearScale:
    default:
      break; }

    if (old_defined &&
	scale->offset == new_offset && scale->range == new_range &&
	scale->new_type == scale->type) {
      return TRUE; }
    else {
      scale->offset		= new_offset;
      scale->offset_defined	= TRUE;
      if (new_range > 0.0 &&
	  new_range / (ABS(new_offset) + new_range) > SIGNIF_RANGE) {
	scale->range		= new_range; }
      else {
	scale->range		= 0.0; }
      scale->range_defined	= TRUE;
      scale->type		= scale->new_type;
      GowResetNewScale(scale); } }
  else if (old_defined) {
    switch (scale->new_type) {
    case GowLogarithmicScale:
      if (scale->offset <= 0) { /* illegal value for log() */
	scale->type= scale->new_type;
	scale->offset_defined= scale->range_defined= FALSE;
	return FALSE; }
      break;
    case GowSquaredScale:
    case GowLinearScale:
    default:
      break; }
    scale->type= scale->new_type; }
  else {
    scale->type= scale->new_type;
    return FALSE; }
    
  { GowScaleNotifyRec *entry;	/* notify all widgets on change of scale */

    ForAllWidgets(scale, entry) {
      entry->delay(entry->widget, TRUE);
      entry->notify(entry->widget, scale); }
    ForAllWidgets(scale, entry) {
      entry->delay(entry->widget, FALSE); } }
  return TRUE;
}

void GowResetNewScale(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowResetNewScale"); }
  else {
    GowResetNewScaleOffsetRange(scale);
    scale->num_new_values	= 0;
    scale->new_type		= scale->type;
    if (scale->offset_defined && scale->range_defined) {
      scale->new_value_type	= GowLeaveScale; }
    else {			/* nothing to leave */
      scale->new_value_type	= GowFullScale; } }
  return;
}

void GowResetNewScaleOffsetRange(scale)
     GowScale scale;
{
  if (!scale) { GowNull("GowResetNewScaleOffsetRange"); }
  else {
    scale->new_offset_defined	= FALSE;
    scale->new_range_defined	= FALSE; }
}
void GowNewScaleType(scale, type)
     GowScale scale;
     int type;
{
  if (!scale) { GowNull("GowNewScaleType"); return; }
  else {
    switch (type) {
    case GowLinearScale:
    case GowLogarithmicScale:
    case GowSquaredScale:
      scale->new_type= type;
      break;
    default:
      GowError("GowNewScaleType: illegal scale type");
      break; }
    return; }
}

void GowNewScaleValue(scale, value, value_type)
     GowScale scale;
     double value;
     int value_type;
{
  if (!scale) { GowNull("GowNewScaleValue"); return; }
  else {
    switch (value_type) {
    case GowFullScale:
      scale->new_value_type	= value_type;
      scale->num_new_values	= 0;
      break;
    case GowMinValue:
    case GowMaxValue:
      scale->new_value_type	= value_type;
      scale->new_value[0]	= value;
      scale->num_new_values	= 0;
      break;
    case GowFifoValue:
      scale->new_value_type				= value_type;
      scale->new_value[scale->num_new_values % 2]	= value;
      scale->num_new_values++;
      break;
    case GowOffsetValue:
      scale->new_value_type	= GowExplicitScale;
      scale->new_offset		= value;
      scale->new_offset_defined	= TRUE;
      scale->num_new_values	= 0;
      break;
    case GowRangeValue:
      scale->new_value_type	= GowExplicitScale;
      scale->new_range		= value;
      scale->new_range_defined	= TRUE;
      scale->num_new_values	= 0;
      break;
    default:
      GowError("GowNewScaleValue: illegal value type");
      break; }

    return; }
}

Boolean GowStretchScale(scale, center, fac)
     GowScale scale;
     double center;
     double fac;
{
  if (scale->range_defined) {
    if (fac > 0.0) {
      double range= scale->range * fac;
      double offset= center - range * 0.5;

      if (scale->new_type != GowLogarithmicScale || offset > 0.0) {
	GowNewScaleValue(scale, offset, GowOffsetValue); }
      GowNewScaleValue(scale, range, GowRangeValue); }
    return TRUE; }
  return FALSE;
}

Boolean GowValidScale(scale)
     GowScale scale;
{
  return (scale->offset_defined && scale->range_defined);
}

/*
 * static functions operating on scale (GowScale)
 */

/* ARGSUSED */
static GowScaleConversion ErrorPixel(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (!scale->error_printed) {
    GowError(scale->error); scale->error_printed= FALSE; }
  return GowToSmall;
}

/* ARGSUSED */
static GowScaleConversion ErrorUser(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (!scale->error_printed) {
    GowError(scale->error); scale->error_printed= TRUE; }
  return GowToSmall;
}

static void InitializeError(scale, message)
     GowScale scale;
     char *message;
{
  scale->error_printed	= FALSE;
  scale->pixel		= ErrorPixel;
  scale->user		= ErrorUser;
  scale->error		= message;
  return;
}

static GowScaleConversion LinearPixel(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register double px= (((user - scale->user_offset) * scale->user_factor)
			 + scale->pixel_offset);
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipSmall; }
    else if (*pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion LinearPixelInv(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register double px= (scale->pixel_offset -
			 ((user - scale->user_offset) * scale->user_factor));
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipLarge; }
    else if (*pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion LinearUser(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToSmall; }
  else if (pixel > scale->pixel_max) { return GowToLarge; }
  else {
    *user= (((pixel - scale->clip_min) * scale->user_factor_inv)
	    + scale->user_offset);
    if (pixel < scale->clip_min) { return GowClipSmall; }
    else if (pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion LinearUserInv(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToLarge; }
  else if (pixel > scale->pixel_max) { return GowToSmall; }
  else {
    *user= (((scale->clip_max - pixel) * scale->user_factor_inv)
	    + scale->user_offset);
    if (pixel < scale->clip_min) { return GowClipLarge; }
    else if (pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion LogarithmicPixel(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register px= (((log(user) - scale->user_offset) * scale->user_factor)
		  + scale->pixel_offset);
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipSmall; }
    else if (*pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion LogarithmicPixelInv(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register px= (scale->pixel_offset -
		  ((log(user) - scale->user_offset) * scale->user_factor));
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipLarge; }
    else if (*pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion LogarithmicUser(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToSmall; }
  else if (pixel > scale->pixel_max) { return GowToLarge; }
  else {
    *user= exp(((pixel - scale->clip_min) * scale->user_factor_inv)
	       + scale->user_offset);
    if (pixel < scale->clip_min) { return GowClipSmall; }
    else if (pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion LogarithmicUserInv(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToLarge; }
  else if (pixel > scale->pixel_max) { return GowToSmall; }
  else {
    *user= exp(((scale->clip_max - pixel) * scale->user_factor_inv)
	       + scale->user_offset);
    if (pixel < scale->clip_min) { return GowClipLarge; }
    else if (pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion SquaredPixel(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register px= (((SQRED(user) - scale->user_offset) * scale->user_factor)
		  + scale->pixel_offset);
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipSmall; }
    else if (*pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion SquaredPixelInv(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else {
    register px= (scale->pixel_offset -
		  ((SQRED(user) - scale->user_offset) * scale->user_factor));
    *pixel= NINT(px);
    if (*pixel < scale->clip_min) { return GowClipLarge; }
    else if (*pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion SquaredUser(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToSmall; }
  else if (pixel > scale->pixel_max) { return GowToLarge; }
  else {
    double tmp= (((pixel - scale->clip_min) * scale->user_factor_inv)
		 + scale->user_offset);
    *user= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp);
    if (pixel < scale->clip_min) { return GowClipSmall; }
    else if (pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}

static GowScaleConversion SquaredUserInv(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToLarge; }
  else if (pixel > scale->pixel_max) { return GowToSmall; }
  else {
    double tmp= (((scale->clip_max - pixel) * scale->user_factor_inv)
		 + scale->user_offset);
    *user= (tmp < 0.0) ? -sqrt(-tmp) : sqrt(tmp);
    if (pixel < scale->clip_min) { return GowClipLarge; }
    else if (pixel > scale->clip_max) { return GowClipSmall; }
    else { return GowValid; } }
}

static GowScaleConversion Range0Pixel(scale, pixel, user)
     GowScale scale;
     int *pixel;
     double user;
{
  if (user < scale->user_min) { return GowToSmall; }
  else if (user > scale->user_max) { return GowToLarge; }
  else { *pixel= scale->pixel_offset; return GowValid; }
}

static GowScaleConversion Range0User(scale, pixel, user)
     GowScale scale;
     int pixel;
     double *user;
{
  if (pixel < scale->pixel_min) { return GowToSmall; }
  else if (pixel > scale->pixel_max) { return GowToLarge; }
  else {
    *user= scale->user_offset;
    if (pixel < scale->clip_min) { return GowClipSmall; }
    else if (pixel > scale->clip_max) { return GowClipLarge; }
    else { return GowValid; } }
}
