/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowWidget.c,v 1.4 1994/12/02 21:22:24 vs Exp vs $ */

#include <Gow/GowConfig.h>

#ifdef HAS_XT_GRAPHIC

#include <math.h>
#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#ifdef HAS_XAW3D
#include <X11/Xaw3d/XawInit.h>
#include <X11/Xaw3d/Cardinals.h>
#include <X11/Xaw3d/Label.h>
#else
#include <X11/Xaw/XawInit.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Label.h>
#endif

#include <X11/Xmu/Initer.h>
#include <X11/Xmu/CharSet.h>

#include <tv/vsAlloc.h>
#include <Gow/GowPaneGeo.h>
#include <Gow/GowPlot.h>

#include "GowWidgetP.h"
#include "GowObjectP.h"

#define Offset(field) XtOffsetOf(GowWidgetRec, gow_widget.field)

static float zero= 0.0;

static XtResource resources[] = { 
  { XtNdelayOutput, XtCDelayOutput, XtREnum, sizeof(GowDelayOutput),
    Offset(delay_output),
    XtRImmediate, (XtPointer)GowDelayUnchanged },
  { XtNdelayCallback, XtCCallback, XtRCallback, sizeof(XtCallbackList),
    Offset(delay_callbacks),
    XtRImmediate, (XtPointer)NULL },
  { XtNscaleCallback, XtCCallback, XtRCallback, sizeof(XtCallbackList),
    Offset(scale_callbacks),
    XtRImmediate, (XtPointer)NULL },
  { XtNcursorShow, XtCCursorShow, XtRBoolean, sizeof(Boolean),
    Offset(cursor_show),
    XtRImmediate, (XtPointer)TRUE },
  { XtNcursorForeground, XtCCursorForeground, XtRPixel, sizeof (Pixel),
    Offset(cursor_foreground), XtRString, XtDefaultForeground },
  { XtNcursorBackground, XtCCursorBackground, XtRPixel, sizeof (Pixel),
    Offset(cursor_background), XtRString, XtDefaultBackground }, 
  { XtNcursorLinewidth, XtCCursorLinewidth, XtRDimension, sizeof(Dimension),
    Offset(cursor_linewidth), XtRImmediate, (XtPointer) 0 },
  { XtNxScale, XtCScale, XtRPointer, sizeof(GowScale),
    Offset(x_scale), XtRPointer, (XtPointer) NULL },
  { XtNyScale, XtCScale, XtRPointer, sizeof(GowScale),
    Offset(y_scale), XtRPointer, (XtPointer) NULL },
  { XtNinheritableScale, XtCInheritableScale, XtRInt, sizeof(int),
    Offset(inheritable_scale), XtRImmediate, (XtPointer) '\0' },
  { XtNxClipMin, XtCClip, XtRFloat, sizeof(float),
    Offset(x_clip_min), XtRFloat, (XtPointer) &zero },
  { XtNxClipMax, XtCClip, XtRFloat, sizeof(float),
    Offset(x_clip_max), XtRFloat, (XtPointer) &zero },
  { XtNyClipMin, XtCClip, XtRFloat, sizeof(float),
    Offset(y_clip_min), XtRFloat, (XtPointer) &zero },
  { XtNyClipMax, XtCClip, XtRFloat, sizeof(float),
    Offset(y_clip_max), XtRFloat, (XtPointer) &zero },
};

#undef Offset

/*
 * Default Translations
 */

/* none */

/*
 * Semi Public function definitions. 
 */

static void Redisplay(), Realize(), Resize(), ChangeManaged();
static void ClassPartInitialize();
static void Initialize(), Destroy();
static Boolean SetValues();
static XtGeometryResult GeometryManager();
static Boolean CheckClip();
static void CreateCursorGC();
static void CursorDisplay();
static void CursorUndisplay();
static Boolean CursorGetXY();
static void CursorMotion();
static void ExpandFullScaleOfWidget();
static void NotifyWidgetScaleChange();
static void InheritFull();
static void Update();
static void Erase();

/* these Core methods not needed:
 *
 */

/*
 * Action Routine Definitions
 */

static void CursorStretch(), CursorExpand(), CursorExpandToLimit();
static void CursorResetScales(), CursorCancelExpand();
static void CursorExpandX(), CursorExpandY(), CursorExpandXY();
static void CursorCenter();
static void CursorFull();
static void CursorLocalFull();
static void CursorShift();

/* 
 * Private Function Definitions.
 */

/*
 * Actions list
 */

static XtActionsRec actions[] = {
  { "stretch",		CursorStretch },
  { "expand",		CursorExpand },
  { "expand-to-limit",	CursorExpandToLimit },
  { "cancel-expand",	CursorCancelExpand },
  { "scalereset",	CursorResetScales },
  { "expand-x",		CursorExpandX },
  { "expand-xy",	CursorExpandXY },
  { "expand-y",		CursorExpandY },
  { "full",		CursorFull },
  { "local-full",	CursorLocalFull },
  { "center",		CursorCenter },
  { "shift",		CursorShift },
};
 
static CompositeClassExtensionRec extension_rec = {
    /* next_extension */  NULL,
    /* record_type */     NULLQUARK,
    /* version */         XtCompositeExtensionVersion,
    /* record_size */     sizeof(CompositeClassExtensionRec),
    /* accepts_objects */ TRUE,
};

#define SUPERCLASS (&gowDestinationWidgetClassRec)
    
GowWidgetClassRec gowWidgetClassRec = {
  {				/* core_class */
    /* superclass         */    (WidgetClass) SUPERCLASS,
    /* class_name         */    "GowWidget",
    /* size               */    sizeof(GowWidgetRec),
    /* class_initialize   */	NULL,
    /* class_part_initialize*/	ClassPartInitialize,
    /* Class init'ed      */	FALSE,
    /* initialize         */    Initialize,
    /* initialize_hook    */	NULL,
    /* realize            */    Realize,
    /* actions            */    actions,
    /* num_actions        */    XtNumber(actions),
    /* resources          */    resources,
    /* resource_count     */	XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion    */    TRUE, 
    /* compress_exposure  */    TRUE,
    /* compress_enterleave*/ 	TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    Destroy,
    /* resize             */    Resize,
    /* expose             */    Redisplay,
    /* set_values         */    SetValues,
    /* set_values_hook    */	NULL,
    /* set_values_almost  */	XtInheritSetValuesAlmost,  
    /* get_values_hook    */	NULL,			
    /* accept_focus       */    NULL,
    /* intrinsics version */	XtVersion,
    /* callback offsets   */    NULL,
    /* tm_table		  */    NULL,
    /* query_geometry	  */    GowQueryPaneGeometry,
    /* display_accelerator*/    NULL,
    /* extension	  */    NULL
  },
  {				/* composite_class */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */    NULL
  },
  {				/* gow_widget_class */
    /* UpdateWidget	  */	Update,
    /* EraseWidget	  */	Erase,
    /* Shell extension	  */    NULL
  }
};

WidgetClass gowWidgetClass = (WidgetClass)&gowWidgetClassRec;

/************************************************************
 *
 * Semi-Public Functions.
 *
 ************************************************************/


/*      Function Name: ClassPartInitialize
 *      Description: Class Part Initialize routine, called for every
 *                   subclass.  Makes sure that the subclasses pick up 
 *                   the extension record.
 *      Arguments: wc - the widget class of the subclass.
 *      Returns: none.
 */

static void
ClassPartInitialize(wc)
WidgetClass wc;
{
  GowWidgetClass gowc = (GowWidgetClass)wc;
  GowWidgetClass swc = (GowWidgetClass)gowc->core_class.superclass;

/*
 * Make sure that our subclass gets the extension rec too.
 */

  extension_rec.next_extension = gowc->composite_class.extension;
  gowc->composite_class.extension = (XtPointer) &extension_rec;

/* 
 * We don't need to check for null super since we'll get to GowWidget
 * eventually.
 */

  if (gowc->gow_widget_class.UpdateWidget == XtInheritUpdateWidget) {
    gowc->gow_widget_class.UpdateWidget= swc->gow_widget_class.UpdateWidget; }

  if (gowc->gow_widget_class.EraseWidget == XtInheritEraseWidget) {
    gowc->gow_widget_class.EraseWidget= swc->gow_widget_class.EraseWidget; }
}

/*      Function Name: Initialize
 *      Description: Initializes the gow widget
 *      Arguments: request - the widget requested by the argument list.
 *                 new     - the new widget with both resource and non
 *                           resource values.
 *      Returns: none.
 */

/* ARGSUSED */
static void
Initialize(request, new, args, num_args)
Widget request, new;
ArgList args;
Cardinal *num_args;
{
  GowWidget gow = (GowWidget) new;

  if (!CheckClip(gow->gow_widget.x_clip_min, gow->gow_widget.x_clip_max)) {
    gow->gow_widget.x_clip_min= 0.0; gow->gow_widget.x_clip_max= 0.0; }
  if (!CheckClip(gow->gow_widget.y_clip_min, gow->gow_widget.y_clip_max)) {
    gow->gow_widget.y_clip_min= 0.0; gow->gow_widget.y_clip_max= 0.0; }

  if (gow->gow_widget.x_scale) {
    GowInsertScaleWidget(gow->gow_widget.x_scale, new,
			 NotifyWidgetScaleChange,
			 GowDelayWidgetOutput,
			 InheritFull); }

  if (gow->gow_widget.y_scale) {
    GowInsertScaleWidget(gow->gow_widget.y_scale, new,
			 NotifyWidgetScaleChange,
			 GowDelayWidgetOutput,
			 InheritFull); }

  if (gow->gow_widget.inheritable_scale == 'X' ||
      gow->gow_widget.inheritable_scale == 'Y') {
    gow->gow_widget.inheritable_scale += 'a' - 'A'; }

  gow->gow_widget.set_values_expose= FALSE;

  gow->gow_widget.delaycount= 0;
  gow->gow_widget.delay_output= GowDelayUnchanged;

  gow->gow_widget.needs_update		= FALSE;
  gow->gow_widget.needs_erase		= FALSE;
  gow->gow_widget.needs_redisplay	= FALSE;
  gow->gow_widget.needs_expand_x	= FALSE;
  gow->gow_widget.needs_expand_y	= FALSE;
  gow->gow_widget.local_full_x		= TRUE;
  gow->gow_widget.local_full_y		= TRUE;

  gow->gow_widget.recursive_redisplay	= FALSE;

  gow->gow_widget.cursor_drawn	= FALSE;
  gow->gow_widget.cursor_type	= CrossHairCursor;
  gow->gow_widget.cursor_crosshair_type	= CrossHairCursor;

  CreateCursorGC(new);
  XtAddEventHandler(new,
		    (unsigned int)(PointerMotionMask |
				   EnterWindowMask |
				   LeaveWindowMask),
		    FALSE, CursorMotion, NULL);

  gow->gow_widget.cursor_grid		= NULL;
  gow->gow_widget.cursor_grid_num	= 0;
}

/*      Function Name: Destroy
 *      Description: Destroys the gow widget
 *      Arguments: w - the widget to destroy.
 *      Returns: none.
 */

/* ARGSUSED */
static void
Destroy(w)
Widget w;
{
  GowWidget gow = (GowWidget)w;

  XtReleaseGC(w, gow->gow_widget.cursor_gc);
  XtRemoveEventHandler(w,
		       (unsigned int)(PointerMotionMask |
				      EnterWindowMask |
				      LeaveWindowMask),
		       FALSE, CursorMotion, NULL);
}

/*      Function Name: Redisplay
 *      Description: Redisplays the contents of the widget.
 *      Arguments:	w - the gow widget.
 *              	event - the X event that caused this redisplay.
 *              	region - the region the needs to be repainted. 
 *				 NULL if full redisplay
 *      Returns:	none.
 */

/* ARGSUSED */
static void
Redisplay(w, event, region)
Widget w;
XEvent * event;
Region region;
{
  GowWidget gow = (GowWidget) w;

  gow->gow_widget.needs_redisplay= TRUE;
  if (!region) { gow->gow_widget.needs_erase= TRUE; }

  Update(w, event, region);
}

/*      Function Name: Realize
 *      Description: Realizes the widget.
 *      Arguments: w - the gow widget.
 *                 mask - value mask for the window to create.
 *                 attrs - attributes for the window to create.
 *      Returns: none
 */

static void
Realize(w, mask, attrs)
Widget w;
XtValueMask *mask;
XSetWindowAttributes *attrs;
{
  GowWidget gow = (GowWidget) w;

  (*SUPERCLASS->core_class.realize)(w, mask, attrs);

  gow->gow_widget.cursor_display= XtDisplay(w);
  gow->gow_widget.cursor_window= XtWindow(w);
}

/*      Function Name: Resize
 *      Description: Handle the menu being resized bigger.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

static void
Resize(w)
Widget w;
{
  GowWidget gow = (GowWidget)w;
  GowObject *entry;

  if (!XtIsRealized(w)) { return; }

  ForAllObjects(gow, entry) {	/* reset width of all entries. */
    if (GowIsManaged((Widget)*entry)) {
      (*entry)->rectangle.width = gow->core.width;
      (*entry)->rectangle.height = gow->core.height; } }

  gow->gow_widget.needs_redisplay= TRUE;
  gow->gow_widget.needs_erase= TRUE;

  (*SUPERCLASS->core_class.resize)(w);
}

/*      Function Name: SetValues
 *      Description: Relayout the menu when one of the resources is changed.
 *      Arguments: current - current state of the widget.
 *                 request - what was requested.
 *                 new - what the widget will become.
 *      Returns: none
 */

/* ARGSUSED */
static Boolean
SetValues(current, request, new, args, num_args)
Widget current, request, new;
ArgList args;
Cardinal *num_args;
{
  GowWidget gow_old = (GowWidget) current;
  GowWidget gow_new = (GowWidget) new;
  Boolean ret_val = FALSE;
  Boolean geo_change = FALSE;
  Boolean update = FALSE;

  if (!XtIsRealized(new)) { return FALSE; }

  geo_change= ((gow_new->core.width != gow_old->core.width) ||
	       (gow_new->core.height != gow_old->core.height));

  if (gow_old->gow_widget.cursor_show != gow_new->gow_widget.cursor_show) {
    if (!gow_new->gow_widget.cursor_show) { CursorUndisplay(new); } }

  { Boolean scale_change= FALSE;

    if (gow_old->gow_widget.x_clip_min != gow_new->gow_widget.x_clip_min ||
	gow_old->gow_widget.x_clip_max != gow_new->gow_widget.x_clip_max) {
    if (!CheckClip(gow_new->gow_widget.x_clip_min,
		   gow_new->gow_widget.x_clip_max)) {
      gow_new->gow_widget.x_clip_min= gow_old->gow_widget.x_clip_min;
      gow_new->gow_widget.x_clip_max= gow_old->gow_widget.x_clip_max; }
    else { scale_change= ret_val= TRUE; } }

    if (gow_old->gow_widget.y_clip_min != gow_new->gow_widget.y_clip_min ||
	gow_old->gow_widget.y_clip_max != gow_new->gow_widget.y_clip_max) {
      if (!CheckClip(gow_new->gow_widget.y_clip_min,
		     gow_new->gow_widget.y_clip_max)) {
	gow_new->gow_widget.y_clip_min= gow_old->gow_widget.y_clip_min;
	gow_new->gow_widget.y_clip_max= gow_old->gow_widget.y_clip_max; }
      else { scale_change= ret_val= TRUE; } }

    if (gow_new->gow_widget.x_scale != gow_old->gow_widget.x_scale) {
      GowInsertScaleWidget(gow_new->gow_widget.x_scale, new,
			   NotifyWidgetScaleChange,
			   GowDelayWidgetOutput,
			   InheritFull);
      GowDeleteScaleWidget(gow_new->gow_widget.x_scale, new);
      scale_change= TRUE; }

    if (gow_new->gow_widget.y_scale != gow_old->gow_widget.y_scale) {
      GowInsertScaleWidget(gow_new->gow_widget.y_scale, new,
			   NotifyWidgetScaleChange,
			   GowDelayWidgetOutput,
			   InheritFull);
      GowDeleteScaleWidget(gow_new->gow_widget.y_scale, new);
      scale_change= TRUE; }

    if (scale_change) { XtCallCallbacks(new, XtNscaleCallback, NULL); } }

  if (gow_new->gow_widget.inheritable_scale == 'X' ||
      gow_new->gow_widget.inheritable_scale == 'Y') {
    gow_new->gow_widget.inheritable_scale += 'a' - 'A'; }

  if (gow_new->gow_widget.delay_output != gow_old->gow_widget.delay_output) {
    Boolean callback = FALSE;

    switch (gow_new->gow_widget.delay_output) {
    case GowDelayBlock:
      (gow_new->gow_widget.delaycount)++;
      break;
    case GowDelayUnblock:
      if (gow_new->gow_widget.delaycount &&
	  !(--(gow_new->gow_widget.delaycount))) {
	callback= TRUE; }
      break;
    case GowDelayReset:
      if (gow_new->gow_widget.delaycount) {
	gow_new->gow_widget.delaycount= 0; callback= TRUE; }
      break;
    default: break; }

    gow_new->gow_widget.delay_output= GowDelayUnchanged;

    if (callback) { XtCallCallbacks(new, XtNdelayCallback, NULL); }

    if (!gow_new->gow_widget.delaycount &&
	!gow_new->gow_widget.recursive_redisplay) {
      update= (gow_new->gow_widget.needs_update &&
	       !gow_new->gow_widget.needs_redisplay);
      ret_val= gow_new->gow_widget.needs_redisplay; } }
  else if (ret_val) {
    ret_val= !gow_new->gow_widget.needs_redisplay; }

  if (ret_val) {
    gow_new->gow_widget.needs_redisplay= TRUE;
    if (!gow_new->gow_widget.set_values_expose) {
      return (gow_new->gow_widget.set_values_expose= TRUE); } }
  if (update && !gow_new->gow_widget.delaycount) {
    Update(new, (XEvent *)NULL, (Region)NULL); }
  return FALSE;
}

/************************************************************
 *
 * Geometry Management routines.
 *
 ************************************************************/

/*	Function Name: GeometryManager
 *	Description: This is the Gow Widget's Geometry Manager.
 *	Arguments: w - the Menu Entry making the request.
 *                 request - requested new geometry.
 *                 reply - the allowed geometry.
 *	Returns: XtGeometry{Yes, No, Almost}.
 */

static XtGeometryResult
GeometryManager(w, request, reply)
Widget w;
XtWidgetGeometry * request, * reply;
{
  XtGeometryMask mode = request->request_mode;

  if (!(mode & CWWidth) && !(mode & CWHeight)) { return XtGeometryNo; }
  else {
    if (mode & CWWidth) { w->core.width = request->width; }
    if (mode & CWHeight) { w->core.height = request->height; }
    return XtGeometryYes; }
}

/*	Function Name: ChangeManaged
 *	Description: called whenever a new child is managed.
 *	Arguments: w - the gow widget.
 *	Returns: none.
 */

static void
ChangeManaged(w)
Widget w;
{
  { GowObject *entry;
    GowWidget gow= (GowWidget)w;
    ForAllObjects(gow, entry) {
      (*entry)->gow_object.managed= XtIsManaged((Widget)*entry); } }

  if (XtIsRealized(w)) {
    GowObject *entry;
    GowWidget gow= (GowWidget)w;
    Boolean update= FALSE;

    ForAllObjects(gow, entry) {
      update |= (XtIsManaged((Widget)*entry) &&
		 !(*entry)->gow_object.visible);
      if (!XtIsManaged((Widget)*entry) && (*entry)->gow_object.visible) {
	Redisplay(w, (XEvent *)NULL, (Region)NULL); return; } }
    if (update) { Update(w, (XEvent *)NULL, (Region)NULL); } }
  return;
}

/************************************************************
 *
 * Global Action Routines.
 * 
 * These actions routines will be added to the application's
 * global action list. 
 * 
 ************************************************************/

/************************************************************
 *
 * Widget Action Routines.
 * 
 ************************************************************/

/* XtActionProc */
static void CursorStretch(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  double x_fac= 1.0;
  double y_fac= 1.0;

  if (*num_params != 2 ||
      sscanf(params[0], "%lf", &x_fac) != 1 ||
      sscanf(params[1], "%lf", &y_fac) != 1) {
    XtError("GowWidget CursorStretch has been passed incorrect parameter.");
    return; }

  if (XtIsSubclass(w, gowWidgetClass)) {
    int x, y;
    GowWidget gow= (GowWidget)w;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;

    if (!CursorGetXY(event, &x, &y)) { return; }

    if (x_scale || y_scale) {
      double u_x, u_y;
      GowUserFunc user;
      GowScaleConversion conv;


      if (x_fac != 0.0 &&
	  !(x_scale &&
	    GowInitializeWidgetScales(w, TRUE, FALSE) &&
	    (user= GowGetScaleUserFunction(x_scale)) &&
	    (conv= user(x_scale, x, &u_x)) != GowToSmall &&
	    conv != GowToLarge)) {
	x_fac= 0.0; }

      if (y_fac != 0.0 &&
	  !(y_scale &&
	    GowInitializeWidgetScales(w, FALSE, TRUE) &&
	    (user= GowGetScaleUserFunction(y_scale)) &&
	    (conv= user(y_scale, y, &u_y)) != GowToSmall &&
	    conv != GowToLarge)) {
	y_fac= 0.0; }

      if (x_fac != 0.0 || y_fac != 0.0) {
	CursorUndisplay(w);
	GowWidgetStretch(w, x_fac, y_fac, u_x, u_y);
	CursorDisplay(w, x, y); } } }
}

void GowWidgetStretch(w, x_fac, u_x, y_fac, u_y)
     Widget w;
     double x_fac, y_fac;
     double u_x, u_y;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    GowWidget gow = (GowWidget)w;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;

    if (x_fac > 0.0) {
      x_fac= ((x_fac > 32.0) ? 2.e-10 : pow(0.5, x_fac)); }
    else if (x_fac < 0.0) {
      x_fac= ((x_fac < -32.0) ? 2.e10 : pow(2.0, -x_fac)); }

    if (y_fac > 0.0) {
      y_fac= ((y_fac > 32.0) ? 2.e-10 : pow(0.5, y_fac)); }
    else if (y_fac < 0.0) {
      y_fac= ((y_fac < -32.0) ? 2.e10 : pow(2.0, -y_fac)); }

    if (x_fac == 0.0 && y_fac == 0.0) { return; }

    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    if (x_fac != 0.0) { GowStretchScale(x_scale, u_x, x_fac); }
    if (y_fac != 0.0) { GowStretchScale(y_scale, u_y, y_fac); }

    GowExpandScalesOfWidget(w, TRUE, TRUE);

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE);

    gow->gow_widget.cursor_type	= CrossHairCursor;
    XDefineCursor(XtDisplay(w), XtWindow(w), gow->destination_widget.cursor); }
}

/* ARGSUSED */ /* XtActionProc */
static void CursorExpand(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  int xmin, xmax, ymin, ymax;
  GowWidget gow= (GowWidget)w;
  GowScale x_scale= gow->gow_widget.x_scale;
  GowScale y_scale= gow->gow_widget.y_scale;
  Boolean x_expand= FALSE;
  Boolean y_expand= FALSE;

  CursorUndisplay(w);

  if (!CursorGetXY(event, &x, &y)) { return; }

  switch (gow->gow_widget.cursor_type) {
  case ExpandHorizontalCursor:
    xmin= MIN(x, gow->gow_widget.cursor_segments[0].x1);
    xmax= MAX(x, gow->gow_widget.cursor_segments[0].x1);
    x_expand= TRUE;
    break;
  case ExpandRectangleCursor:
    xmin= MIN(x, gow->gow_widget.cursor_rectangle_x);
    xmax= MAX(x, gow->gow_widget.cursor_rectangle_x);
    ymin= MIN(y, gow->gow_widget.cursor_rectangle_y);
    ymax= MAX(y, gow->gow_widget.cursor_rectangle_y);
    x_expand= y_expand= TRUE;
    break;
  case ExpandVerticalCursor:
    ymin= MIN(y, gow->gow_widget.cursor_segments[0].y1);
    ymax= MAX(y, gow->gow_widget.cursor_segments[0].y1);
    y_expand= TRUE;
    break;
  default:
    break; }

  if (x_scale || y_scale) {
    double u_min, u_max;
    GowUserFunc user;
    GowScaleConversion conv;

    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    if (x_scale && x_expand &&
	GowInitializeWidgetScales(w, TRUE, FALSE) &&
	(user= GowGetScaleUserFunction(x_scale)) &&
	(conv= user(x_scale, xmin, &u_min)) != GowToSmall &&
	conv != GowToLarge &&
	(conv= user(x_scale, xmax, &u_max)) != GowToSmall &&
	conv != GowToLarge &&
	u_min != u_max) {
      GowNewScaleValue (x_scale, u_min, GowFifoValue);
      GowNewScaleValue (x_scale, u_max, GowFifoValue); }

    if (y_scale && y_expand &&
	GowInitializeWidgetScales(w, FALSE, TRUE) &&
	(user= GowGetScaleUserFunction(y_scale)) &&
	(conv= user(y_scale, ymin, &u_min)) != GowToSmall &&
	conv != GowToLarge &&
	(conv= user(y_scale, ymax, &u_max)) != GowToSmall &&
	conv != GowToLarge &&
	u_min != u_max) {
      GowNewScaleValue (y_scale, u_min, GowFifoValue);
      GowNewScaleValue (y_scale, u_max, GowFifoValue); }

    GowExpandScalesOfWidget(w, TRUE, TRUE);

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE); }

  gow->gow_widget.cursor_type	= CrossHairCursor;
  XDefineCursor(XtDisplay(w), XtWindow(w), gow->destination_widget.cursor);
  CursorDisplay(w, x, y);
}

/* ARGSUSED */ /* XtActionProc */
static void CursorExpandToLimit(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  GowWidget gow= (GowWidget)w;
  GowScale x_scale= gow->gow_widget.x_scale;
  GowScale y_scale= gow->gow_widget.y_scale;
  Boolean x_expand= FALSE;
  Boolean y_expand= FALSE;

  CursorUndisplay(w);

  if (!CursorGetXY(event, &x, &y)) { return; }

  GowDelayScaleUpdate(x_scale, TRUE);
  GowDelayScaleUpdate(y_scale, TRUE);

  switch (gow->gow_widget.cursor_type) {
  case ExpandHorizontalCursor: x_expand= TRUE; break;
  case ExpandRectangleCursor: y_expand= x_expand= TRUE; break;
  case ExpandVerticalCursor: y_expand= TRUE; break;
  default: break; }

  if (x_expand) {
    GowUserFunc user;
    GowScaleConversion conv;
    double u_x;

    if (GowInitializeWidgetScales(w, TRUE, FALSE) &&
	(user= GowGetScaleUserFunction(x_scale)) &&
	(conv= user(x_scale, gow->gow_widget.cursor_segments[0].x1, &u_x))
	!= GowToSmall && conv != GowToLarge) {
      if (x < gow->gow_widget.cursor_segments[0].x1) {
	GowNewScaleValue(x_scale, u_x, GowMaxValue); }
      else {
	GowNewScaleValue(x_scale, u_x, GowMinValue); } } }

  if (y_expand) {
    GowUserFunc user;
    GowScaleConversion conv;
    double uy;
    double uy1;
				/* y scale has inverse direction */
    if (GowInitializeWidgetScales(w, FALSE, TRUE) &&
	(user= GowGetScaleUserFunction(y_scale)) &&
	(conv= user(y_scale, y, &uy))
	!= GowToSmall && conv != GowToLarge &&
	(conv= user(y_scale, gow->gow_widget.cursor_segments[0].y1, &uy1))
	!= GowToSmall && conv != GowToLarge) {
      GowNewScaleValue(y_scale, uy1,
		       ((uy < uy1) ? GowMaxValue : GowMinValue)); } }

  GowExpandScalesOfWidget(w, TRUE, TRUE);

  GowDelayScaleUpdate(x_scale, FALSE);
  GowDelayScaleUpdate(y_scale, FALSE);

  gow->gow_widget.cursor_type	= CrossHairCursor;
  XDefineCursor(XtDisplay(w), XtWindow(w), gow->destination_widget.cursor);
  CursorDisplay(w, x, y);
}

/* ARGSUSED */ /* XtActionProc */
static void CursorCancelExpand(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  GowWidget gow = (GowWidget)w;

  CursorUndisplay(w);
  gow->gow_widget.cursor_type= CrossHairCursor;
  if (CursorGetXY(event, &x, &y)) { CursorDisplay(w, x, y); }
}

/* ARGSUSED */ /* XtActionProc */
static void CursorResetScales(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int xr, yr;
  if (*num_params != 1) {
    XtError
      ("GowWidget CursorResetScales has been passed incorrect parameter.");
    return; }

    xr= (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL);
    yr= (strchr(params[0], 'y') != NULL || strchr(params[0], 'Y') != NULL);
  GowWidgetResetScales(w, xr, yr);
}

void GowWidgetResetScales(w, xr, yr)
     Widget w;
     int xr, yr;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    GowWidget gow = (GowWidget)w;

    if (xr) { GowResetNewScale(gow->gow_widget.x_scale); }
    if (yr) { GowResetNewScale(gow->gow_widget.y_scale); } }
}

/* ARGSUSED */ /* XtActionProc */
static void CursorExpandX(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  GowWidget gow = (GowWidget)w;

  CursorUndisplay(w);

  if (!CursorGetXY(event, &x, &y)) { return; }

  gow->gow_widget.cursor_segments[0].x1 = x;
  gow->gow_widget.cursor_segments[0].x2 = x;
  gow->gow_widget.cursor_segments[0].y1 = 0;
  gow->gow_widget.cursor_segments[0].y2 = w->core.height;
  gow->gow_widget.cursor_type= ExpandHorizontalCursor;
  CursorDisplay(w, x, y);
}

/* ARGSUSED */ /* XtActionProc */
static void CursorExpandXY(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  GowWidget gow = (GowWidget)w;

  CursorUndisplay(w);

  if (!CursorGetXY(event, &x, &y)) { return; }

  gow->gow_widget.cursor_rectangle_x= x;
  gow->gow_widget.cursor_rectangle_y= y;
  gow->gow_widget.cursor_type	= ExpandRectangleCursor;
  CursorDisplay(w, x, y);
}

/* ARGSUSED */ /* XtActionProc */
static void CursorExpandY(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  int x, y;
  GowWidget gow = (GowWidget)w;

  CursorUndisplay(w);

  if (!CursorGetXY(event, &x, &y)) { return; }

  gow->gow_widget.cursor_segments[0].x1 = 0;
  gow->gow_widget.cursor_segments[0].x2 = w->core.width;
  gow->gow_widget.cursor_segments[0].y1 = y;
  gow->gow_widget.cursor_segments[0].y2 = y;
  gow->gow_widget.cursor_type= ExpandVerticalCursor;
  CursorDisplay(w, x, y);
}
				/* XtActionProc */
static void CursorCenter(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  if (*num_params != 1) {
    XtError("GowWidget CursorCenter has been passed incorrect parameter.");
    return; }
  if (XtIsSubclass(w, gowWidgetClass)) {
    GowWidget gow = (GowWidget)w;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;
    int x, y;
    int xc, yc;

    xc= (GowGetScaleType(x_scale) == GowLinearScale &&
	 (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL));
    yc= (GowGetScaleType(y_scale) == GowLinearScale &&
	 (strchr(params[0], 'y') != NULL || strchr(params[0], 'Y') != NULL));

    if ((xc || yc) && CursorGetXY(event, &x, &y)) {
      GowUserFunc user;
      GowScaleConversion conv;
      double ux, uy;
				/* y scale has inverse direction */
      if (GowInitializeWidgetScales(w, xc, yc) &&
	  (!xc || ((user= GowGetScaleUserFunction(x_scale)) &&
		   (conv= user(x_scale, x, &ux))
		   != GowToSmall && conv != GowToLarge)) &&
	  (!yc || ((user= GowGetScaleUserFunction(y_scale)) &&
		   (conv= user(y_scale, y, &uy))
		   != GowToSmall && conv != GowToLarge))) {
	  
	CursorUndisplay(w);
	GowWidgetCenter(w, ux, uy, xc, yc);
	CursorDisplay(w, x, y); } } }
}

void GowWidgetCenter(w, ux, uy, xc, yc)
     Widget w;
     double ux, uy;
     int xc, yc;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    double x_offset, x_range;
    double y_offset, y_range;
    GowWidget gow = (GowWidget)w;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;

    xc= (xc &&
	 GowGetScaleType(x_scale) == GowLinearScale &&
	 GowGetScaleOffset(x_scale, &x_offset) &&
	 GowGetScaleRange(x_scale, &x_range));
    yc= (yc &&
	 GowGetScaleType(y_scale) == GowLinearScale &&
	 GowGetScaleOffset(y_scale, &y_offset) &&
	 GowGetScaleRange(y_scale, &y_range));
				/* y scale has inverse direction */

    if (!xc && !yc) { return; }

    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    if (xc) {
      GowNewScaleValue(x_scale, ux - x_range * 0.5, GowOffsetValue);
      GowNewScaleValue(x_scale, x_range, GowRangeValue); }

    if (yc) {
      GowNewScaleValue(y_scale, uy - y_range * 0.5, GowOffsetValue);
      GowNewScaleValue(y_scale, y_range, GowRangeValue); }

    GowExpandScalesOfWidget(w, TRUE, TRUE);

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE);

    gow->gow_widget.cursor_type	= CrossHairCursor;
    XDefineCursor(XtDisplay(w), XtWindow(w), gow->destination_widget.cursor); }
}

/* XtActionProc */
static void CursorFull(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  if (*num_params != 1) {
    XtError("GowWidget CursorFull has been passed incorrect parameter.");
    return; }
  if (XtIsSubclass(w, gowWidgetClass)) {
    int x, y;
    int xf, yf;
    xf= (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL);
    yf= (strchr(params[0], 'y')  != NULL || strchr(params[0], 'Y') != NULL);

    CursorUndisplay(w);
    GowWidgetFull(w, xf, yf);
    if (CursorGetXY(event, &x, &y)) { CursorDisplay(w, x, y); } }
}

/* XtActionProc */
static void CursorLocalFull(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  if (*num_params != 1) {
    XtError("GowWidget CursorLocalFull has been passed incorrect parameter.");
    return; }
  if (XtIsSubclass(w, gowWidgetClass)) {
    int x, y;
    int xf, yf;
    xf= (strchr(params[0], 'x') != NULL || strchr(params[0], 'X') != NULL);
    yf= (strchr(params[0], 'y')  != NULL || strchr(params[0], 'Y') != NULL);

    CursorUndisplay(w);
    GowWidgetLocalFull(w, xf, yf);
    if (CursorGetXY(event, &x, &y)) { CursorDisplay(w, x, y); } }
}

void GowWidgetFull(w, x_full, y_full)
     Widget w;
     int x_full, y_full;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    GowWidget gow = (GowWidget)w;
    if (x_full || y_full) {
      GowScale x_scale= gow->gow_widget.x_scale;
      GowScale y_scale= gow->gow_widget.y_scale;
      if (x_full) {
	gow->gow_widget.local_full_x
	  = (gow->gow_widget.inheritable_scale != 'x');
	GowNewScaleValue(x_scale, 0.0, GowFullScale); }
      if (y_full) {
	gow->gow_widget.local_full_y
	  = (gow->gow_widget.inheritable_scale != 'y');
	GowNewScaleValue(y_scale, 0.0, GowFullScale); } }

    gow->gow_widget.cursor_type	= CrossHairCursor; }
}

void GowWidgetLocalFull(w, x_full, y_full)
     Widget w;
     int x_full, y_full;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    GowWidget gow = (GowWidget)w;
    if (x_full || y_full) {
      GowScale x_scale= gow->gow_widget.x_scale;
      GowScale y_scale= gow->gow_widget.y_scale;
      if (x_full) {
	gow->gow_widget.local_full_x= TRUE;
	GowNewScaleValue(x_scale, 0.0, GowFullScale); }
      if (y_full) {
	gow->gow_widget.local_full_y= TRUE;
	GowNewScaleValue(y_scale, 0.0, GowFullScale); } }

    gow->gow_widget.cursor_type	= CrossHairCursor; }
}

				/* XtActionProc */
static void CursorShift(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  double x_perc= 0.0, y_perc= 0.0;

  if (*num_params != 2 ||
      sscanf(params[0], "%lf", &x_perc) != 1 ||
      sscanf(params[1], "%lf", &y_perc) != 1) {
    XtError("GowWidget CursorShift has been passed incorrect parameters.");
    return; }
  if (XtIsSubclass(w, gowWidgetClass)) {
    if (x_perc != 0.0 || y_perc != 0.0) {
      CursorUndisplay(w);
      GowWidgetShift(w, x_perc, y_perc); }}
}

void GowWidgetShift(w, x_perc, y_perc)
     Widget w;
     double x_perc, y_perc;
{
  if (XtIsSubclass(w, gowWidgetClass)) {
    double x_offset, x_range;
    double y_offset, y_range;
    GowWidget gow = (GowWidget)w;
    GowScale x_scale= gow->gow_widget.x_scale;
    GowScale y_scale= gow->gow_widget.y_scale;

    if (x_perc != 0.0 &&
	GowGetScaleType(x_scale) == GowLinearScale &&
	GowGetScaleOffset(x_scale, &x_offset) &&
	GowGetScaleRange(x_scale, &x_range)) {
      x_offset += x_perc * 0.01 * x_range; }
    else { x_perc= 0.0; }
  
    if (y_perc != 0.0 &&
	GowGetScaleType(y_scale) == GowLinearScale &&
	GowGetScaleOffset(y_scale, &y_offset) &&
	GowGetScaleRange(y_scale, &y_range)) {
      y_offset += y_perc * 0.01 * y_range; }
    else { y_perc= 0.0; }
				/* y scale has inverse direction */

    if (x_perc == 0.0 && y_perc == 0.0) { return; }

    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    if (x_perc != 0.0) {
      GowNewScaleValue(x_scale, x_offset, GowOffsetValue);
      GowNewScaleValue(x_scale, x_range, GowRangeValue); }

    if (y_perc != 0.0) {
      GowNewScaleValue(y_scale, y_offset, GowOffsetValue);
      GowNewScaleValue(y_scale, y_range, GowRangeValue); }

    GowExpandScalesOfWidget(w, TRUE, TRUE);

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE);

    gow->gow_widget.cursor_type	= CrossHairCursor;
    XDefineCursor(XtDisplay(w), XtWindow(w),
		  gow->destination_widget.cursor); }
}

/************************************************************
 *
 * Public Functions.
 *
 ************************************************************/

/*	Function Name:	GowDelayWidgetOutput
 *	Description:	supends/resumes drawing output
 *	Arguments:	w - the gow widget.
 *			delay_output - new value
 *	Returns:	old value of delay_output.
 */

extern void GowDelayWidgetOutput(w, delay_output)
     Widget w;
     int delay_output;
{
  Arg arg;

  XtSetArg(arg, XtNdelayOutput,
	   (delay_output ? GowDelayBlock : GowDelayUnblock));
  XtSetValues(w, &arg, 1);
}

/*      Function Name: ExpandFullScaleOfWidget
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow widget.
 *			x/y_scale - the gow x/y-scale
 *			x/y_full - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

static void
ExpandFullScaleOfWidget(w, x_scale, x_full, y_scale, y_full)
Widget w;
GowScale x_scale;
int x_full;
GowScale y_scale;
int y_full;
{
  GowWidget gow = (GowWidget)w;
  GowObject *entry;

  if (x_full || y_full) {
    CursorUndisplay(w);

    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    if (x_full) { GowResetNewScaleOffsetRange(x_scale); }
    if (y_full) { GowResetNewScaleOffsetRange(y_scale); }

    ForAllObjects(gow, entry) {
      if (GowIsManaged ((Widget)*entry)) {
	GowExpandFullScaleOfObject((Widget)*entry, 
				   x_scale, x_full, y_scale, y_full); } }
    if (x_full) { GowNewScale(x_scale); }
    if (y_full) { GowNewScale(y_scale); }

    x_full= (x_full && !gow->gow_widget.local_full_x &&
	     gow->gow_widget.inheritable_scale == 'x');
    y_full= (y_full && !gow->gow_widget.local_full_y &&
	     gow->gow_widget.inheritable_scale == 'y');

    gow->gow_widget.local_full_x= TRUE;
    gow->gow_widget.local_full_y= TRUE;

    if (!x_full && y_full) { GowInheritFullScale(x_scale, y_scale); }
    if (!y_full && x_full) { GowInheritFullScale(y_scale, x_scale); }

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE); }

}

/*      Function Name: GowRedisplayWidget
 *      Description: forces redisplay of the widget.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

void
GowRedisplayWidget(w)
Widget w;
{
  Redisplay(w, (XEvent *)NULL, (Region)NULL);
}

/*      Function Name: GowUpdateWidget
 *      Description: Displays added contents of the widget.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

void
GowUpdateWidget(w)
Widget w;
{
  Update(w, (XEvent *)NULL, (Region)NULL);
}

static void NotifyWidgetScaleChange(w, scale)
     Widget w;
     GowScale scale;
{
  GowWidget gow= (GowWidget)w;

  if (!XtIsRealized(w)) { return; }

  if (gow->gow_widget.x_scale == scale ||
      gow->gow_widget.y_scale == scale) {
    XtCallCallbacks(w, XtNscaleCallback, NULL);
    Redisplay(w, (XEvent *)NULL, (Region)NULL); }
  return;
}

static void InheritFull(w, const_scale, full_scale)
     Widget w;
     GowScale const_scale, full_scale;
{
  GowWidget gow= (GowWidget)w;
  GowScale x_scale= gow->gow_widget.x_scale;
  GowScale y_scale= gow->gow_widget.y_scale;

  if (!XtIsRealized(w)) { return; }

  gow->gow_widget.local_full_x= TRUE;
  gow->gow_widget.local_full_y= TRUE;

  if (const_scale == x_scale && full_scale != y_scale) {
    ExpandFullScaleOfWidget(w, x_scale, FALSE, y_scale, TRUE); }
  else if (const_scale == y_scale && full_scale != x_scale) {
    ExpandFullScaleOfWidget(w, x_scale, TRUE, y_scale, FALSE); }
  return;
}

Boolean GowGetUserX(w, px, ux)
     Widget w;
     int px;
     double *ux;
{
  GowWidget gow = (GowWidget)w;
  GowScale x_scale= gow->gow_widget.x_scale;

  if (XtIsSubclass(w, gowWidgetClass)) {
    GowUserFunc user;
    GowScaleConversion conv;

    if (GowInitializeWidgetScales(w, TRUE, FALSE) &&
	(user= GowGetScaleUserFunction(x_scale)) &&
	(conv= user(x_scale, px, ux)) != GowToSmall && conv != GowToLarge) {
      return TRUE; } }
  return FALSE;
}

Boolean GowGetUserY(w, py, uy)
     Widget w;
     int py;
     double *uy;
{
  GowWidget gow = (GowWidget)w;
  GowScale y_scale= gow->gow_widget.y_scale;

  if (XtIsSubclass(w, gowWidgetClass)) {
    GowUserFunc user;
    GowScaleConversion conv;

    if (GowInitializeWidgetScales(w, FALSE, TRUE) &&
	(user= GowGetScaleUserFunction(y_scale)) &&
	(conv= user(y_scale, py, uy)) != GowToSmall && conv != GowToLarge) {
      return TRUE; } }
  return FALSE;
}

/************************************************************
 *
 * Private Functions.
 *
 ************************************************************/

/*      Function Name: CursorDisplay
 *      Description: event handler to draw a crosshair cursor.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

/* ARGSUSED */
static void CursorDisplay(w, x, y)
     Widget w;
     int x;
     int y;
{
  GowWidget gow= (GowWidget)w;

  if (gow->gow_widget.cursor_show) {
    CursorUndisplay(w);
    if (gow->gow_widget.cursor_type == ExpandRectangleCursor) {
      if (x < gow->gow_widget.cursor_rectangle_x) {
	gow->gow_widget.cursor_rectangle.x= x;
	gow->gow_widget.cursor_rectangle.width
	  = (gow->gow_widget.cursor_rectangle_x - x); }
      else {
	gow->gow_widget.cursor_rectangle.x= gow->gow_widget.cursor_rectangle_x;
	gow->gow_widget.cursor_rectangle.width
	  = (x - gow->gow_widget.cursor_rectangle_x); }
      if (y < gow->gow_widget.cursor_rectangle_y) {
	gow->gow_widget.cursor_rectangle.y= y;
	gow->gow_widget.cursor_rectangle.height
	  = (gow->gow_widget.cursor_rectangle_y - y); }
      else {
	gow->gow_widget.cursor_rectangle.y= gow->gow_widget.cursor_rectangle_y;
	gow->gow_widget.cursor_rectangle.height
	  = (y - gow->gow_widget.cursor_rectangle_y); }
      gow->gow_widget.cursor_drawn = TRUE;
      XDrawRectangles(gow->gow_widget.cursor_display,
		      gow->gow_widget.cursor_window,
		      gow->gow_widget.cursor_gc,
		      &gow->gow_widget.cursor_rectangle, 1); }
    else {
      switch (gow->gow_widget.cursor_type) {
      case CrossHairCursor:
	{ GowUserFunc user;
	  GowPixelFunc px;
	  GowScaleConversion conv;
	  GowScale x_scale= gow->gow_widget.x_scale;
	  double u_x, u_o= gow->gow_widget.cursor_grid_offset;

	  if (gow->gow_widget.cursor_grid &&
	      x_scale &&
	      GowInitializeWidgetScales(w, TRUE, FALSE) &&
	      (user= GowGetScaleUserFunction(x_scale)) &&
	      (px= GowGetScalePixelFunction(x_scale)) &&
	      (conv= user(x_scale, x, &u_x))
	      != GowToSmall && conv != GowToLarge) {
	    gow->gow_widget.cursor_grid[0].x1 = 0;
	    gow->gow_widget.cursor_grid[0].x2 = w->core.width;
	    gow->gow_widget.cursor_grid[0].y1 = y;
	    gow->gow_widget.cursor_grid[0].y2 = y;
	    gow->gow_widget.cursor_grid[1].x1 = x;
	    gow->gow_widget.cursor_grid[1].x2 = x;
	    gow->gow_widget.cursor_grid[1].y1 = 0;
	    gow->gow_widget.cursor_grid[1].y2 = w->core.height;
	    { int i, seg_num= 2, num= gow->gow_widget.cursor_grid_max;

	      for (i= -num + 1; i< num; i++) {
		int x_g;
		if (i != 0 &&
		    (conv= px(x_scale, &x_g, (double)(u_x + i * u_o)))
		    != GowToSmall && conv != GowToLarge) {
		  gow->gow_widget.cursor_grid[seg_num].x1 = x_g;
		  gow->gow_widget.cursor_grid[seg_num].x2 = x_g;
		  gow->gow_widget.cursor_grid[seg_num].y1 = 0;
		  gow->gow_widget.cursor_grid[seg_num].y2 = w->core.height;
		  seg_num++; } }
	      gow->gow_widget.cursor_grid_num= seg_num;
	      gow->gow_widget.cursor_drawn = TRUE;
	      XDrawSegments(gow->gow_widget.cursor_display,
			    gow->gow_widget.cursor_window,
			    gow->gow_widget.cursor_gc,
			    gow->gow_widget.cursor_grid, seg_num);
	    }
	    XFlush(gow->gow_widget.cursor_display);
	    return; }
	  else {
	    gow->gow_widget.cursor_segments[0].x1 = 0;
	    gow->gow_widget.cursor_segments[0].x2 = w->core.width;
	    gow->gow_widget.cursor_segments[0].y1 = y;
	    gow->gow_widget.cursor_segments[0].y2 = y;
	    gow->gow_widget.cursor_segments[1].x1 = x;
	    gow->gow_widget.cursor_segments[1].x2 = x;
	    gow->gow_widget.cursor_segments[1].y1 = 0;
	    gow->gow_widget.cursor_segments[1].y2 = w->core.height; } }
	break;
      case ExpandHorizontalCursor:
	gow->gow_widget.cursor_segments[1].x1 = x;
	gow->gow_widget.cursor_segments[1].x2 = x;
	gow->gow_widget.cursor_segments[1].y1 = 0;
	gow->gow_widget.cursor_segments[1].y2 = w->core.height;
	break;
      case ExpandVerticalCursor:
	gow->gow_widget.cursor_segments[1].x1 = 0;
	gow->gow_widget.cursor_segments[1].x2 = w->core.width;
	gow->gow_widget.cursor_segments[1].y1 = y;
	gow->gow_widget.cursor_segments[1].y2 = y;
	break;
      default:
	break; }
    
      gow->gow_widget.cursor_drawn = TRUE;
      XDrawSegments(gow->gow_widget.cursor_display,
		    gow->gow_widget.cursor_window,
		    gow->gow_widget.cursor_gc,
		    gow->gow_widget.cursor_segments, 2);

      if (gow->gow_widget.cursor_crosshair_type == DopplerCursor )
	{
	  GowUserFunc user;
	  GowPixelFunc px;
	  GowScaleConversion conv;
	  GowScale x_scale= gow->gow_widget.x_scale;
	  double user_x;
	  int doppler_channel;
	  int save_x = gow->gow_widget.cursor_segments[1].x1;
	  
	  if (x_scale &&
	      GowInitializeWidgetScales(w, TRUE, FALSE) &&
	      (user= GowGetScaleUserFunction(x_scale)) &&
	      (px= GowGetScalePixelFunction(x_scale))){
	    if ((conv= user(x_scale, save_x, &user_x)) != GowToSmall && conv != GowToLarge) {
	      
	      px (x_scale, &doppler_channel, user_x * ( 1 + gow->gow_widget.cursor_doppler_factor));
	      gow->gow_widget.cursor_segments[1].x1 = doppler_channel;
	      gow->gow_widget.cursor_segments[1].x2 = doppler_channel;
	      
	      XDrawSegments(gow->gow_widget.cursor_display,
			    gow->gow_widget.cursor_window,
			    gow->gow_widget.cursor_gc,
			    gow->gow_widget.cursor_segments, 2);
	      
	      px (x_scale, &doppler_channel, user_x * ( 1 - gow->gow_widget.cursor_doppler_factor));
	      gow->gow_widget.cursor_segments[1].x1 = doppler_channel;
	      gow->gow_widget.cursor_segments[1].x2 = doppler_channel;
	      
	      XDrawSegments(gow->gow_widget.cursor_display,
			    gow->gow_widget.cursor_window,
			    gow->gow_widget.cursor_gc,
			    gow->gow_widget.cursor_segments, 2);
	      gow->gow_widget.cursor_segments[1].x1 = save_x;
	      gow->gow_widget.cursor_segments[1].x2 = save_x;
	    }
	  }
	}
    } 
    XFlush(gow->gow_widget.cursor_display); }
  return;
}

static Boolean CursorGetXY(event, x, y)
     XEvent *event;
     int *x;
     int *y;
{
  switch (event->type) {
  case ButtonPress:
  case ButtonRelease:
    *x= event->xbutton.x;
    *y= event->xbutton.y;
    return TRUE;
  case KeyPress:
  case KeyRelease:
    *x= event->xkey.x;
    *y= event->xkey.y;
    return TRUE;
  default:
    return FALSE; }
}

/* ARGSUSED */ /* XtEventHandler */
static void CursorMotion(w, unused, event, unused_too)
     Widget w;
     XtPointer unused;
     XEvent *event;
     Boolean *unused_too;
{
  CursorUndisplay(w);
  switch (event->type) {
  case MotionNotify:
    CursorDisplay(w, event->xmotion.x, event->xmotion.y);
    break;
  case EnterNotify:
    CursorDisplay(w, event->xcrossing.x, event->xcrossing.y);
    break;
  default: {
    GowWidget gow= (GowWidget)w;
    gow->gow_widget.cursor_type	= CrossHairCursor; }
    break; }
}

/*      Function Name: CursorUndisplay
 *      Description: remove crosshair cursor from window.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

/* ARGSUSED */
static void CursorUndisplay(w)
     Widget w;
{
  GowWidget gow= (GowWidget)w;

  if (gow->gow_widget.cursor_drawn) {
    switch (gow->gow_widget.cursor_type) {
    case ExpandRectangleCursor:
      XDrawRectangles(gow->gow_widget.cursor_display,
		      gow->gow_widget.cursor_window,
		      gow->gow_widget.cursor_gc,
		      &gow->gow_widget.cursor_rectangle, 1);
      break;
    case CrossHairCursor:
      if (gow->gow_widget.cursor_grid_num) {
	XDrawSegments(gow->gow_widget.cursor_display,
		      gow->gow_widget.cursor_window,
		      gow->gow_widget.cursor_gc,
		      gow->gow_widget.cursor_grid,
		      gow->gow_widget.cursor_grid_num);
	gow->gow_widget.cursor_grid_num= 0;
	break; }
    default:
      XDrawSegments(gow->gow_widget.cursor_display,
		    gow->gow_widget.cursor_window,
		    gow->gow_widget.cursor_gc,
		    gow->gow_widget.cursor_segments, 2);
      
      if (gow->gow_widget.cursor_crosshair_type == DopplerCursor )
	{
	  GowUserFunc user;
	  GowPixelFunc px;
	  GowScaleConversion conv;
	  GowScale x_scale= gow->gow_widget.x_scale;
	  double user_x;
	  int doppler_channel;
	  int save_x = gow->gow_widget.cursor_segments[1].x1;
	  
	  if (x_scale &&
	      GowInitializeWidgetScales(w, TRUE, FALSE) &&
	      (user= GowGetScaleUserFunction(x_scale)) &&
	      (px= GowGetScalePixelFunction(x_scale))){
	    if ((conv= user(x_scale, save_x, &user_x)) != GowToSmall && conv != GowToLarge) {
	      
	      px (x_scale, &doppler_channel, user_x * (1 + gow->gow_widget.cursor_doppler_factor));
	      gow->gow_widget.cursor_segments[1].x1 = doppler_channel;
	      gow->gow_widget.cursor_segments[1].x2 = doppler_channel;
	      
	      XDrawSegments(gow->gow_widget.cursor_display,
			    gow->gow_widget.cursor_window,
			    gow->gow_widget.cursor_gc,
			    gow->gow_widget.cursor_segments, 2);
	      
	      px (x_scale, &doppler_channel, user_x * (1 - gow->gow_widget.cursor_doppler_factor));
	      gow->gow_widget.cursor_segments[1].x1 = doppler_channel;
	      gow->gow_widget.cursor_segments[1].x2 = doppler_channel;
	      
	      XDrawSegments(gow->gow_widget.cursor_display,
			    gow->gow_widget.cursor_window,
			    gow->gow_widget.cursor_gc,
			    gow->gow_widget.cursor_segments, 2);
	    }
	  }
	}
      break; }
    gow->gow_widget.cursor_drawn= FALSE; }
  return;
}

static Boolean CheckClip(clip_min, clip_max)
     float clip_min;
     float clip_max;
{
  return (clip_min + clip_max < 1.0 &&
	  0.0 <= clip_min && clip_min < 1.0 &&
	  0.0 <= clip_max && clip_max < 1.0);
}

/*      Function Name: CreateCursorGC
 *      Description: creates crosshair cursor GC.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

static void CreateCursorGC(w)
     Widget w;
{
  GowWidget gow= (GowWidget)w;
  XGCValues values;
  Screen *scr;
  XtGCMask  mask= (GCFunction | GCLineWidth  |
		   GCForeground | GCBackground | GCSubwindowMode);

  scr= XtScreenOfObject(w);

  if (XDefaultDepthOfScreen(scr) == 1) { /* static gray */
    values.foreground	= ~gow->core.background_pixel;
    values.background	= gow->core.background_pixel; }
  else {
    values.foreground	= gow->gow_widget.cursor_foreground;
    values.background	= gow->gow_widget.cursor_background; }
  values.line_width	= gow->gow_widget.cursor_linewidth;
  values.function	= GXxor;
  values.subwindow_mode	= IncludeInferiors; 

  gow->gow_widget.cursor_gc= XtGetGC(w, mask, &values);
}

/*      Function Name: GowExpandScalesOfWidget
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow widget.
 *			expand_x/y - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

void
GowExpandScalesOfWidget(w, expand_x, expand_y)
Widget w;
int expand_x;
int expand_y;
{
  GowWidget gow = (GowWidget)w;

  gow->gow_widget.needs_expand_x= expand_x;
  gow->gow_widget.needs_expand_y= expand_y;

  Update(w, (XEvent *)NULL, (Region)NULL);
}

/*      Function Name: GowExpandScalesOfWidgetOnUpdate
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow widget.
 *			expand_x/y - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

void
GowExpandScalesOfWidgetOnUpdate(w, expand_x, expand_y)
Widget w;
int expand_x;
int expand_y;
{
  GowWidget gow = (GowWidget)w;

  gow->gow_widget.needs_expand_x= expand_x;
  gow->gow_widget.needs_expand_y= expand_y;
}

/*      Function Name: GowInitializeWidgetScales
 *      Description: initializes x/y-scale as requested
 *      Arguments:	w - the gow widget.
 *			init_x/y request only one axis
 *      Returns: TRUE if succeded
 */

Boolean
GowInitializeWidgetScales(w, init_x, init_y)
Widget w;
int init_x;
int init_y;
{
  GowWidget gow = (GowWidget)w;
  GowDestination d= gow->destination_widget.destination;

  if (!XtIsRealized(w) || !XtIsManaged(w)) { return FALSE; }

  if (init_x &&
      !GowInitializeScale(gow->gow_widget.x_scale,
			  GowDXDirection(d),
			  GowDXOffset(d), GowDXRange(d),
			  (double)gow->gow_widget.x_clip_min,
			  (double)gow->gow_widget.x_clip_max)) {
    return FALSE; }
  if (init_y &&
      !GowInitializeScale(gow->gow_widget.y_scale,
			  GowDYDirection(d),
			  GowDYOffset(d), GowDYRange(d),
			  (double)gow->gow_widget.y_clip_min,
			  (double)gow->gow_widget.y_clip_max)) {
    return FALSE; }

  return TRUE;
}

/*      Function Name: Update
 *      Description: Displays added contents of the widget.
 *		     Redisplays if contents changed.
 *      Arguments: w - the gow widget.
 *                 event - unused always NULL.
 *                 region - unused always NULL. 
 *      Returns: none.
 */

/* ARGSUSED */
static void
Update(w, event, region)
Widget w;
XEvent * event;
Region region;
{
  GowWidget gow = (GowWidget)w;
  GowScale x_scale= gow->gow_widget.x_scale;
  GowScale y_scale= gow->gow_widget.y_scale;

  gow->gow_widget.needs_update= TRUE;

  if (!XtIsRealized(w) ||
      gow->gow_widget.recursive_redisplay ||
      gow->gow_widget.delaycount) {
    return; }

				/* scale check may force redisplay */
  gow->gow_widget.recursive_redisplay	= TRUE;
				/* suppress further SetValues-expose-events */
  gow->gow_widget.set_values_expose |= gow->gow_widget.needs_redisplay;

  gow->gow_widget.needs_expand_x= (gow->gow_widget.needs_expand_x ||
				   !GowValidScale(x_scale));
  gow->gow_widget.needs_expand_y= (gow->gow_widget.needs_expand_y ||
				   !GowValidScale(y_scale));

  if (gow->gow_widget.needs_expand_x || gow->gow_widget.needs_expand_y) {
    GowDelayScaleUpdate(x_scale, TRUE);
    GowDelayScaleUpdate(y_scale, TRUE);

    { Boolean x_full= FALSE;
      Boolean y_full= FALSE;

      if (gow->gow_widget.needs_expand_x && !GowNewScale(x_scale)) {
	x_full= TRUE; }

      if (gow->gow_widget.needs_expand_y && !GowNewScale(y_scale)) {
	y_full= TRUE; }

      if (x_full || y_full) {
	ExpandFullScaleOfWidget(w, x_scale, x_full, y_scale, y_full); } }

    if (gow->gow_widget.needs_expand_x) { GowResetNewScale(x_scale); }
    if (gow->gow_widget.needs_expand_y) { GowResetNewScale(y_scale); }

    GowDelayScaleUpdate(x_scale, FALSE);
    GowDelayScaleUpdate(y_scale, FALSE); }

  gow->gow_widget.recursive_redisplay	= FALSE;

  CursorUndisplay(w);

  if (gow->gow_widget.needs_erase ||
      (gow->gow_widget.needs_redisplay && region == NULL) ||
      (region && XRectInRegion(region, 0, 0,
			       (unsigned int)w->core.width,
			       (unsigned int)w->core.height) == RectangleIn)) {
    Erase(w); region= NULL; }

  /*
   * Check and Paint each of the entries
   */

  { if (GowInitializeWidgetScales(w, TRUE, TRUE)) {
      GowObject *entry;

      ForAllObjects(gow, entry) {
	if (!GowIsManaged ((Widget)*entry)) { continue; }

	if (region != NULL &&
	    ((GowObjectClass)(*entry)->object.widget_class)
	    ->gow_object_class.uses_rectangles) {
	  switch(XRectInRegion(region, (int)(*entry)->rectangle.x,
			       (int)(*entry)->rectangle.y,
			       (unsigned int)(*entry)->rectangle.width,
			       (unsigned int)(*entry)->rectangle.height)) {
	  case RectangleIn:
	  case RectanglePart:
	    break;
	  default:
	    continue; } }
	
	if ((*entry)->gow_object.needs_update || region) {
	  GowObjectClass class = (GowObjectClass)(*entry)->object.widget_class;

	  if (class->rect_class.expose != NULL) {
	    if ((*entry)->gow_object.needs_update) {
	      (class->rect_class.expose)((Widget)*entry,
					 (XEvent *)NULL, (Region)NULL); }
	    else {
	      (class->rect_class.expose)((Widget)*entry,
					 event, region); } } } } } }

  gow->gow_widget.needs_update		= FALSE;
  gow->gow_widget.needs_redisplay	= FALSE;
  gow->gow_widget.set_values_expose	= FALSE;
}

/*      Function Name: Erase
 *      Description: Notifies all associated GowObject's on erase.
 *      Arguments: w - the gow widget.
 *      Returns: none.
 */

static void EraseObjects(gow)
     GowWidget gow;
{
  GowObject *entry;

  ForAllObjects(gow, entry) {
    GowObjectClass class = (GowObjectClass)(*entry)->object.widget_class;

    if (class->gow_object_class.EraseObject != NULL) {
      (class->gow_object_class.EraseObject)((Widget)*entry); }
    (*entry)->gow_object.needs_update	= TRUE;
    (*entry)->gow_object.visible	= FALSE; }
}

static void
Erase(w)
Widget w;
{
  GowWidget gow = (GowWidget)w;

  if (!XtIsRealized(w)) { return; }
  else {
    CursorUndisplay(w);
    gow->gow_widget.cursor_type	= CrossHairCursor;
    GowWidgetResetScales(w, TRUE, TRUE);
    if (gow->gow_widget.delaycount) {
      gow->gow_widget.needs_erase= TRUE; return; } }

  XClearWindow(XtDisplay(w), XtWindow(w));

  /*
   * Notify each of the entries - including the not managed.
   */

  EraseObjects(gow);

  gow->gow_widget.needs_update	= TRUE;	/* everything must be redrawn ... */
  gow->gow_widget.needs_erase	= FALSE;
}

void GowWidgetPlot(w, fp, type, geo)
     Widget w;
     GowDestFile fp;
     int type;
     GowDestGeometry geo;
{
  GowWidget gow= (GowWidget)w;
  GowDestination d= gow->destination_widget.destination;

  if (XtIsRealized(w)) {
    if (GowSetupPlotDestination(d, type, fp, geo)) {
      EraseObjects(gow);
      if (GowInitializeWidgetScales(w, TRUE, TRUE)) {
	GowObject *entry;

	ForAllObjects(gow, entry) {
	  if (GowIsManaged ((Widget)*entry)) {
	    GowObjectClass class;
	    class= (GowObjectClass)(*entry)->object.widget_class;

	    if (class->rect_class.expose != NULL) {
	      (class->rect_class.expose)((Widget)*entry,
					 (XEvent *)NULL, (Region)NULL);
	    }
	  }
	}
      }
      GowFinishPlotDestination(d);
      Redisplay(w, (XEvent *)NULL, (Region)NULL);
    }
  }
}

void GowWidgetCursorDoppler(w, doppler_factor)
     Widget w;
     double doppler_factor;
{
  GowWidget gow= (GowWidget)w;
  gow->gow_widget.cursor_doppler_factor = doppler_factor;
  gow->gow_widget.cursor_crosshair_type = DopplerCursor;
}

void GowWidgetCursorCrossHair(w)
     Widget w;
{
  GowWidget gow= (GowWidget)w;
  gow->gow_widget.cursor_crosshair_type = CrossHairCursor;
}

void GowWidgetCursorGrid(w, num, offset)
     Widget w;
     int num;
     double offset;
{
  GowWidget gow= (GowWidget)w;

  CursorUndisplay(w);
  Free(gow->gow_widget.cursor_grid);

  if (num && offset != 0.0) {
    num= ABS(num);
    num++;
    if ((gow->gow_widget.cursor_grid
	 = (XSegment *)Malloc(2 * num * sizeof(XSegment)))) {
      gow->gow_widget.cursor_grid_offset= offset;
      gow->gow_widget.cursor_grid_num= 0;
      gow->gow_widget.cursor_grid_max= num; } }
}

/* ARGSUSED */
void GowWidgetCrossHair(w, u_x, conv_x, u_y, conv_y)
     Widget w;
     double u_x;
     int conv_x;
     double u_y;
     int conv_y;
{
  GowWidget gow= (GowWidget)w;

  if (gow->gow_widget.cursor_show) {
    CursorUndisplay(w);
    { GowPixelFunc px;
      int x, y;
      GowScaleConversion conv;
      GowScale x_scale= gow->gow_widget.x_scale;
      GowScale y_scale= gow->gow_widget.y_scale;

      if (conv_x && x_scale &&
	  GowInitializeWidgetScales(w, TRUE, FALSE) &&
	  (px= GowGetScalePixelFunction(x_scale)) &&
	  (conv= px(x_scale, &x, u_x)) != GowToSmall && conv != GowToLarge) {
	gow->gow_widget.cursor_segments[1].x1 = x;
	gow->gow_widget.cursor_segments[1].x2 = x;
	gow->gow_widget.cursor_segments[1].y1 = 0;
	gow->gow_widget.cursor_segments[1].y2 = w->core.height; }
      else { memset(gow->gow_widget.cursor_segments + 1,
		    0, sizeof(XSegment)); }

      if (conv_y && y_scale &&
	  GowInitializeWidgetScales(w, FALSE, TRUE) &&
	  (px= GowGetScalePixelFunction(y_scale)) &&
	  (conv= px(y_scale, &y, u_y)) != GowToSmall && conv != GowToLarge) {
	gow->gow_widget.cursor_segments[0].x1 = 0;
	gow->gow_widget.cursor_segments[0].x2 = w->core.width;
	gow->gow_widget.cursor_segments[0].y1 = y;
	gow->gow_widget.cursor_segments[0].y2 = y; }
      else { memset(gow->gow_widget.cursor_segments,
		    0, sizeof(XSegment)); } }

    gow->gow_widget.cursor_drawn = TRUE;
    XDrawSegments(gow->gow_widget.cursor_display,
		  gow->gow_widget.cursor_window,
		  gow->gow_widget.cursor_gc,
		  gow->gow_widget.cursor_segments, 2); }
  XFlush(gow->gow_widget.cursor_display);
  return;
}

void GowWidgetUndisplayCursor(w)
     Widget w;
{
  CursorUndisplay(w);
}

#endif				/* HAS_XT_GRAPHIC */
