/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/Gow/RCS/GowWidget.h,v 1.2 1994/05/26 19:49:50 vs Exp vs $ */

#ifndef _GowWidget_h
#define _GowWidget_h

#include <X11/Shell.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xfuncproto.h>

#include <Gow/GowConfig.h>
#include <Gow/GowDestWidget.h>
#include <Gow/GowScale.h>
#include <Gow/GowDest.h>

/****************************************************************
 *
 * GowWidget widget
 *
 ****************************************************************/

/* GowWidget Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 backgroundPixmap    BackgroundPixmap	Pixmap          None
 borderColor	     BorderColor	Pixel		XtDefaultForeground
 borderPixmap	     BorderPixmap	Pixmap		None
 borderWidth	     BorderWidth	Dimension	1
 bottomMargin        VerticalMargins    Dimension       VerticalSpace
 columnWidth         ColumnWidth        Dimension       Width of widest text
 cursor		     Cursor		Cursor		crosshair
 cursorForeground    CursorForeground	Pixel		XtDefaultForeground
 cursorBackground    CursorBackground	Pixel		XtDefaultBackground
 cursorLinestyle     CursorLinestyle	GowLineStyle	GowLineSolid
 cursorLinewidth     CursorLinewidth	Dimension	0
 cursorShow	     CursorShow		Boolean		True
 delayOutput	     DelayOutput	Boolean		False
 destroyCallback     Callback		Pointer		NULL
 delayCallback       Callback		Pointer		NULL
 scaleCallback       Callback		Pointer		NULL
 height		     Height		Dimension	0
 inheritableScale    InheritableScale	Int		'\0'
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 sensitive	     Sensitive		Boolean		True
 topMargin           VerticalMargins    Dimension       VerticalSpace
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
 xClipMin	     Clip		float		0.0
 xClipMax	     Clip		float		0.0
 xScale		     Scale		Pointer		NULL
 yClipMin	     Clip		float		0.0
 yClipMax	     Clip		float		0.0
 yScale		     Scale		Pointer		NULL

*/

typedef struct _GowWidgetClassRec	*GowWidgetClass;
typedef struct _GowWidgetRec		*GowWidget;

extern WidgetClass gowWidgetClass;

#ifndef XtNcursor
# define XtNcursor		"cursor"
#endif
#define XtNcursorForeground	"cursorForeground"
#define XtNcursorBackground	"cursorBackground"
#define XtNcursorLinestyle	"cursorLinestyle"
#define XtNcursorLinewidth	"cursorLinewidth"
#define XtNcursorShow		"cursorShow"
#define XtNdelayOutput		"delayOutput"
#define XtNdelayCallback	"delayCallback"
#define XtNscaleCallback	"scaleCallback"
#define XtNinheritableScale	"inheritableScale"
#define XtNxScale		"xScale"
#define XtNyScale		"yScale"
#define XtNxClipMin		"xClipMin"
#define XtNxClipMax		"xClipMax"
#define XtNyClipMin		"yClipMin"
#define XtNyClipMax		"yClipMax"

#define XtCCursorForeground	"CursorForeground"
#define XtCCursorBackground	"CursorBackground"
#define XtCCursorLinestyle	"CursorLinestyle"
#define XtCCursorLinewidth	"CursorLinewidth"
#define XtCCursorShow		"CursorShow"
#define XtCDelayOutput		"DelayOutput"
#define XtCInheritableScale	"InheritableScale"
#define XtCScale		"Scale"
#define XtCClip			"Clip"

typedef enum {
  GowDelayBlock,
  GowDelayUnblock,
  GowDelayReset,
  GowDelayUnchanged
} GowDelayOutput;

/************************************************************
 *
 * Public Functions.
 *
 ************************************************************/

_XFUNCPROTOBEGIN

/*	Function Name:	GowDelayWidgetOutput
 *	Description:	supends/resumes drawing output
 *	Arguments:	w - the gow widget.
 *			delay_output - new value
 *	Returns:	old value of delay_output.
 */

extern void GowDelayWidgetOutput(
#if NeedFunctionPrototypes
	Widget			/* w */,
	int			/* delay_output */
#endif
);

/*      Function Name: GowExpandScalesOfWidget
 *      Description: Expands x/y-scale as required by managed objects
 *      Arguments:	w - the gow widget.
 *			expand_x/y - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

extern void GowExpandScalesOfWidget(
#if NeedFunctionPrototypes
	Widget			/* w */,
	int			/* expand_x */,
	int			/* expand_y */
#endif
);

/*      Function Name:	GowExpandScalesOfWidgetOnUpdate
 *      Description: 	Expands x/y-scale as required by managed objects
 *			on next update
 *      Arguments:	w - the gow widget.
 *			expand_x/y - TRUE if x/y-scale is to be expanded
 *      Returns: none.
 */

extern void GowExpandScalesOfWidgetOnUpdate(
#if NeedFunctionPrototypes
	Widget			/* w */,
	int			/* expand_x */,
	int			/* expand_y */
#endif
);

/*      Function Name: GowInitializeWidgetScales
 *      Description: initializes x/y-scale as requested
 *      Arguments:	w - the gow widget.
 *			init_x/y requested axis
 *      Returns: TRUE if succeded
 */

extern Boolean GowInitializeWidgetScales(
#if NeedFunctionPrototypes
	Widget			/* w */,
	int			/* init_x */,
	int			/* init_y */
#endif
);

/*	Function Name:	GowRedisplayWidget
 *	Description:	forces redisplay of widget
 *	Arguments:	w - the gow widget.
 *	Returns:	none
 */

extern void GowRedisplayWidget(
#if NeedFunctionPrototypes
	Widget			/* w */
#endif
);

/*	Function Name:	GowUpdateWidget
 *	Description:	Displays added contents of the widget.
 *	Arguments:	w - the gow widget.
 *	Returns:	none
 */

extern void GowUpdateWidget(
#if NeedFunctionPrototypes
	Widget			/* w */
#endif
);

extern void GowDefineCursor(
#if NeedFunctionPrototypes
     Widget			/* w */
#endif
);

extern void GowNotifyWidgetScaleChange(
#if NeedFunctionPrototypes
     Widget			/* w */,
     GowScale			/* scale */
#endif
);

extern void GowWidgetCancelFull(
#if NeedFunctionPrototypes
     Widget			/* w */
#endif
);

extern void GowWidgetResetScales(
#if NeedFunctionPrototypes
     Widget			/* w */,
     int			/* x_reset */,
     int			/* y_reset */
#endif
);

extern void GowWidgetCenter(
#if NeedFunctionPrototypes
     Widget			/* w */,
     double			/* user_x */,
     double			/* user_y */,
     int			/* x_center */,
     int			/* y_center */
#endif
);

extern void GowWidgetFull(
#if NeedFunctionPrototypes
     Widget			/* w */,
     int			/* x_full */,
     int			/* y_full */
#endif
);

extern void GowWidgetLocalFull(
#if NeedFunctionPrototypes
     Widget			/* w */,
     int			/* x_full */,
     int			/* y_full */
#endif
);

extern void GowWidgetShift(
#if NeedFunctionPrototypes
     Widget			/* w */,
     double			/* x_perc */,
     double			/* y_perc */
#endif
);

extern void GowWidgetStretch(
#if NeedFunctionPrototypes
     Widget			/* w */,
     double			/* x_fac */,
     double			/* user_x */,
     double			/* y_fac */,
     double			/* user_y */
#endif
);

extern Boolean GowGetUserX(
#if NeedFunctionPrototypes
    Widget			/* w */,
    int				/* px */,
    double*			/* ux */
#endif
);

extern Boolean GowGetUserY(
#if NeedFunctionPrototypes
    Widget			/* w */,
    int				/* py */,
    double*			/* uy */
#endif
);

extern void GowWidgetPlot(
#if NeedFunctionPrototypes
     Widget			/* w */,
     GowDestFile		/* fp */,
     int			/* type */,
     GowDestGeometry		/* geo */
#endif
);

extern void GowWidgetCursorGrid(
#if NeedFunctionPrototypes
     Widget			/* w */,
     int			/* num */,
     double			/* offset */
#endif
);

extern void GowWidgetCursorDoppler(
#if NeedFunctionPrototypes
     Widget			/* w */,
     double			/* doppler_factor */
#endif
);

extern void GowWidgetCursorCrossHair(
#if NeedFunctionPrototypes
     Widget			/* w */
#endif
);

extern void GowWidgetCrossHair(
#if NeedFunctionPrototypes
     Widget			/* w */,
     double			/* u_x */,
     int			/* conv_x */,
     double			/* u_y */,
     int			/* conv_y */
#endif
);

extern void GowWidgetUndisplayCursor(
#if NeedFunctionPrototypes
     Widget			/* w */
#endif
);

_XFUNCPROTOEND

#endif /* _GowWidget_h */
