/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsAlloc.c,v 1.2 1994/05/26 19:49:59 vs Exp vs $ */
/* =========================================================================
 * some debugging tools by S.Krink
 * 18-Apr-91 (SK)
 * reorganised by J.Theuerkauf
 * 16-Jan-92 (JTH)
 * -------------------------------------------------------------------------
 */

#define _vsAlloc_c

#include "vsConfig.h"
#include "vsTypes.h"
#include "vsAlloc.h"
#include "vsAllocP.h"

#define PROTOCOL_FILE "VSALLOC.protocol"
#define MAX_ALC_NO	200000

#define AlcArg__	src, line
#define AlcArgDec__	char *src; int line;
#define ALC_WRITE(arg) \
  ((alc_fp != NULL || AlcOpen(PROTOCOL_FILE, "a", MAX_ALC_NO)) &&\
   fprintf(alc_fp, "#%-20.20s:%4d: ", src, line) != EOF &&\
   (fprintf arg) != EOF)
#define ALC_MAGIC "123456789012345678901234567890123456789012345678901234567890123"
#define ALC_MAGIC_NO 64
#define ALC_MAGIC_SIZE (sizeof(char) * ALC_MAGIC_NO)
#define ADD_MAGIC(at_adr) (void)strncpy((char *)(at_adr),ALC_MAGIC,ALC_MAGIC_NO)
#define ALC_ERROR() { if (AlcErrNo != ALCNOERR && \
			  (AlcErrNo == ALCWRITE || \
			   (alc_fp && \
			    fprintf(alc_fp, AlcError(AlcErrNo)) >= 0)) && \
			  alc_fp != NULL) { AlcClose(); } }

typedef struct {
  Alc_t *adr;
  Alc_size_t size;
  unsigned char freed;
} AlcEntry;

static AlcEntry *alc_table= NULL;
static int alc_no = 0;
static int max_alc_no = 0;
static FILE *alc_fp = NULL;

/* ------------------------------------------------------------------------- */
int AlcPCheck(AlcArg__)
     AlcArgDec__
{
  if (AlcErrNo || !alc_fp) { return FALSE; }
  else {
    int i;
    for ( i=0; i<alc_no; i++) {
      if (!alc_table[i].freed) {
	if (strncmp((char *)(alc_table[i].adr) + alc_table[i].size, ALC_MAGIC,
		    ALC_MAGIC_SIZE)) {
	  (void)ALC_WRITE((alc_fp,
			   "magic check failed at entry #%d (tail)\n", i));
	  AlcErrNo = ALCMAGLST;
	  fprintf(alc_fp, "found:'%s'", 
		  (char *)(alc_table[i].adr) + alc_table[i].size);
	  fprintf(alc_fp, "[%d] 0x%p (%d) %d\n",
		  i,(void *)alc_table[i].adr,(int)alc_table[i].size,
		  (int)alc_table[i].freed); }
	if (strncmp((char *)(alc_table[i].adr) - ALC_MAGIC_SIZE, ALC_MAGIC,
		    ALC_MAGIC_SIZE)) {
	  (void)ALC_WRITE((alc_fp,
			   "magic check failed at entry #%d (head)\n", i));
	  AlcErrNo = ALCMAG1ST;
	  fprintf(alc_fp, "found:'%s'", 
		  (char *)(alc_table[i].adr) - ALC_MAGIC_SIZE);
	  fprintf(alc_fp, "[%d] 0x%p (%d) %d\n",
		  i,(void *)alc_table[i].adr,(int)alc_table[i].size,
		  (int)alc_table[i].freed); } } }
    ALC_ERROR();
    return !AlcErrNo; }
}
/* ------------------------------------------------------------------------- */
int AlcPDumpMap(file, mode)
     char *file;
     char *mode;
{
  if (!AlcErrNo) {
    FILE *fp;
    extern char *getenv();
    char *gf;

    if (NULL != (gf= getenv("AlcDumpMap")) && *gf != '\0') { file = gf; }

    if (file == NULL || *file == '\0') { AlcErrNo = ALCNAME; }
    else if (NULL == (fp = fopen(file, mode))) { AlcErrNo = ALCFILE; }
    else { DumpMap(fp); fclose(fp); }

    if (AlcErrNo) { AlcClose(); } }
  return !AlcErrNo;
}
/* ------------------------------------------------------------------------- */
int AlcPDumpAllocated(file, mode)
     char *file;
     char *mode;
{
  FILE *fp;
  
  if (!AlcErrNo) {
    extern char *getenv();
    char *gf;
    if (NULL != (gf= getenv("AlcDumpAllocated")) && *gf != '\0') { file = gf; }

    if (file == NULL || *file == '\0') { AlcErrNo = ALCNAME; }
    else if (NULL == (fp = fopen(file, mode))) { AlcErrNo = ALCFILE; }
    else { DumpAllocated(fp); fclose(fp); }

    if (AlcErrNo) { AlcClose(); } }
  return !AlcErrNo;
}
/* ------------------------------------------------------------------------- */
char *AlcPError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* ALCNOERR:  */ "vsalloc: no error",
  /* ALCFREED:  */ "vsalloc: parameter is already free'd",
  /* ALCFIND:   */ "vsalloc: could not find pointer in alloc table",
  /* ALCTABLE:  */ "vsalloc: alloc table is full",
  /* ALCRQST:   */ "vsalloc: memory request failed",
  /* ALCMAG1ST: */ "vsalloc: lost 1st magic number",
  /* ALCMAGLST: */ "vsalloc: lost last magic number",
  /* ALCFILE:   */ "vsalloc: could not open file",
  /* ALCWRITE:  */ "vsalloc: could not write file",
  /* ALCNAME:   */ "vsalloc: illegal file name",
  /* ALCALLOC:  */ "vsalloc: could not create alloc table",
  /* default:   */ "vsalloc: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ------------------------------------------------------------------------- */
int AlcPOpen(AlcArg__, file, mode, size)
     AlcArgDec__
     char *file;
     char *mode;
     int size;
{
  if (!AlcErrNo) {
    extern char *getenv();
    char *gf;

    if (!(alc_table= (AlcEntry *)malloc(sizeof(AlcEntry) * size))) {
      AlcErrNo = ALCALLOC; return FALSE; }

    max_alc_no= size;

    if (NULL != (gf= getenv("AlcProtocol")) && *gf != '\0') { file = gf; }

    if (file == NULL || *file == '\0') { file = PROTOCOL_FILE; }
    if (alc_fp != NULL) { AlcClose(); }
    if (!AlcErrNo &&
	NULL == (alc_fp = fopen(file, mode))) {
      AlcErrNo = ALCFILE; }
    else { (void)ALC_WRITE((alc_fp, "protocol file opened\n")); } }
  else { AlcClose(); }
  return (alc_fp != NULL);
}
/* ------------------------------------------------------------------------- */
void AlcPClose (AlcArg__)
     AlcArgDec__
{
  if (alc_fp != NULL) {
    (void)ALC_WRITE((alc_fp, "closing protocol file\n"));
    DumpMap(alc_fp); 
    fclose(alc_fp); alc_fp = NULL;
    if (alc_table) { free(alc_table); max_alc_no= 0; } }
}
/* ------------------------------------------------------------------------- */
int AlcPProtocol(AlcArg__, str)
     AlcArgDec__
     char *str;
{
  return ALC_WRITE((alc_fp, str));
}
/* ------------------------------------------------------------------------- */
void AlcPFree(AlcArg__, ptr)
     AlcArgDec__
     Alc_t *ptr;
{
  int i;

  (void)ALC_WRITE((alc_fp, "free( 0x%p )\n", (void *)ptr ));
  if (AlcCheck() &&
      ptr != NULL) {
    AlcErrNo = ALCFIND;
    for (i=0;i<alc_no;i++) {
      if (alc_table[i].adr == ptr) {
	if (alc_table[i].freed) {
	  AlcErrNo = ALCFREED;
	  continue; }		/* try again... */
	else {
	  alc_table[i].freed = TRUE;
	  free(((char *)ptr) - ALC_MAGIC_SIZE);
	  AlcErrNo = ALCNOERR; } } }
    ALC_ERROR(); }
  return;
}
/* ------------------------------------------------------------------------- */
Alc_t *AlcPMalloc(AlcArg__, size)
     AlcArgDec__
     Alc_size_t size;
{
  Alc_t *ptr= NULL;

  if (AlcCheck()) {
    ptr = (Alc_t *)mallocP(size);
    (void)ALC_WRITE((alc_fp,"malloc( %d ) -> 0x%p\n",(int)size,(void *)ptr));
    ALC_ERROR(); }
  return ptr;
}
/* ------------------------------------------------------------------------- */
Alc_t *AlcPCalloc(AlcArg__, num, size)
     AlcArgDec__
     Alc_size_t num;
     Alc_size_t size;
{
  Alc_t *ptr = NULL;

  if (AlcCheck()) {
    Alc_size_t siz= num * size;
    ptr = (Alc_t *)mallocP(siz);
    (void)ALC_WRITE((alc_fp, "calloc( %d,%d ) -> 0x%p\n",
		     (int)num, (int)size, (void *)ptr));
    if (ptr != NULL) { memset((void *)ptr, 0, siz); } }
  return ptr;
}
/* ------------------------------------------------------------------------- */
Alc_t *AlcPRealloc(AlcArg__, ptr, size)
     AlcArgDec__
     Alc_t *ptr;
     Alc_size_t size;
{
  Alc_t *tmp = NULL;
  int i;

  if (AlcCheck()) {
    if (size == 0) { Free(ptr); }
    else if (ptr == NULL) { tmp = mallocP(size); }
    else {
      AlcErrNo = ALCFIND;
      for (i=0;i<alc_no;i++) {
	if (alc_table[i].adr == ptr) {
	  if (alc_table[i].freed) {
	    AlcErrNo = ALCFREED;
	    continue;}
	  else {
	    char *p = (char *)ptr - ALC_MAGIC_SIZE;
	    if (NULL == 
		(p = (Alc_t *)realloc(p, size + 2 * ALC_MAGIC_SIZE))) {
	      AlcErrNo = ALCRQST; }
	    else {
	      tmp = alc_table[i].adr = (Alc_t *)((char *)p + ALC_MAGIC_SIZE);
	      alc_table[i].size = size;
	      ADD_MAGIC(((char *)alc_table[i].adr) + size);
	      ADD_MAGIC(((char *)alc_table[i].adr) - ALC_MAGIC_SIZE);
	      AlcErrNo = ALCNOERR; }
	    break; } } } }
    (void)ALC_WRITE((alc_fp, "realloc( 0x%p, %d ) -> 0x%p\n",
		     (void *)ptr, (int)size, (void *)tmp));
    ALC_ERROR(); }
  return tmp;
}
/* ---------------------------------------------------------------------- */
Alc_t *Alc_Calloc(num, size)
     Alc_size_t num;
     Alc_size_t size;
{
  if (num == 0 || size == 0) { return NULL; }
  else {
    Alc_t *ptr;
    if ((ptr= (Alc_t *)malloc((Alc_size_t)(num * size)))) {
      memset(ptr, 0, num * size); }
    return ptr; }
}
/* ---------------------------------------------------------------------- */
Alc_t *Alc_Realloc(ptr, size)
     Alc_t *ptr;
     Alc_size_t size;
{
  if (size == 0) {
    if (ptr != NULL) { free(ptr); }
    return NULL; }
  else if (ptr == NULL) { return (Alc_t *)malloc(size); }
  else { return (Alc_t *)realloc(ptr, size); }
}
/* ========================================================================= */

/* AlcPStrdup() allocates memory for the given string copies it to the memory *
 * and returns the address. Coded for OS-9: N. Nicolay, 16-JUL-90,        *
 * IKP Cologne                                                            *
 */
char *AlcPStrdup(AlcArg__, s)
     AlcArgDec__
     char *s;
{
  char *buf= NULL;

  (void)ALC_WRITE((alc_fp, "strdup( 0x%p )\n", (void *)s));
  if (s) {
    Alc_size_t size;
    size = strlen (s) + 1;
    if ((buf = (char *)Malloc (size * sizeof(char))) != NULL) {
      strcpy (buf, s); } }
  return buf;
}
/* ---------------------------------------------------------------------- */
char *Alc_Strdup(s)
     char *s;
{
  int size;
  char *buf= NULL;
  if (s) {
    size = strlen (s) + 1;
    if ((buf = (char *)malloc (size * sizeof(char))) != NULL) {
      strcpy (buf, s); } }
  return buf;
}
/* ------------------------------------------------------------------------- */
/* AlcPStrndup() allocates memory for the given string copies it to the memory *
 * and returns the address. Coded for OS-9: N. Nicolay, 16-JUL-90,        *
 * IKP Cologne                                                            *
 */
char *AlcPStrndup(AlcArg__, s, maxlen)
     AlcArgDec__
     char *s;
     int maxlen;
{
  char *buf= NULL;

  (void)ALC_WRITE((alc_fp, "strndup( 0x%p, %d )\n", (void *)s, maxlen));
  if (s) {
    Alc_size_t size = strlen (s) + 1;
    if (size > maxlen + 1) { size= maxlen + 1; }
    if ((buf = (char *)Malloc(size * sizeof(char))) != NULL) {
      int i;
      for (i= 0; s[i] != '\0' && i < maxlen; i++) {
	buf[i]= s[i]; }
      buf[i]= '\0'; } }
  return buf;
}
/* ---------------------------------------------------------------------- */
char *Alc_Strndup(s, maxlen)
     char *s;
     int maxlen;
{
  int size, i;
  char *buf= NULL;
  if (s && maxlen >= 0) {
    size = strlen (s) + 1;
    if (size > maxlen + 1) { size= maxlen + 1; }
    if ((buf = (char *)malloc(size * sizeof(char))) != NULL) {
      for (i= 0; s[i] != '\0' && i < maxlen; i++) {
	buf[i]= s[i]; }
      buf[i]= '\0'; } }
  return buf;
}
/* ================ */
/* static functions */
/* ================ */

static Alc_t *mallocP(size)
     Alc_size_t size;
{
  Alc_t *ptr, *tmp = NULL;

  if (size != 0) {
    if (alc_no >= max_alc_no) {
      AlcErrNo = ALCTABLE; }
    else if (NULL == (ptr = (Alc_t *)malloc(size + 2 * ALC_MAGIC_SIZE))) {
      AlcErrNo = ALCRQST; }
    else {
      alc_table[alc_no].size = size;
      alc_table[alc_no].freed = FALSE;
      tmp = alc_table[alc_no].adr = (Alc_t *)((char *)ptr + ALC_MAGIC_SIZE);
      ADD_MAGIC(((char *)alc_table[alc_no].adr) + size);
      ADD_MAGIC(((char *)alc_table[alc_no].adr) - ALC_MAGIC_SIZE);
      alc_no++; } }
  return tmp;
}
/* ------------------------------------------------------------------------- */
static char *statVs[] = { "allocated", "freed", };

static void DumpMap(fp)
     FILE *fp;
{
  if (fp != NULL && fprintf(fp, "Map (%d entrys):\n", alc_no) != EOF) {
    int i;
    for (i=0;i<alc_no;i++) {
      if (fprintf(fp, "%10d:  0x%p | %10d | %s\n",
		  i,
		  (void *)alc_table[i].adr,
		  (int)alc_table[i].size,
		  statVs[(int)alc_table[i].freed]) == EOF) {
	if (!AlcErrNo) { AlcErrNo= ALCWRITE; }
	return; } }
    fprintf(fp, "=================\n"); }
  else if (!AlcErrNo) { AlcErrNo= ALCWRITE; }
}
/* ---------------------------------------------------------------------- */
static void DumpAllocated(fp)
     FILE *fp;
{
  if (fp != NULL && fprintf(fp, "Map (%d entrys):\n", alc_no) != EOF) {
    int i;
    for (i=0;i<alc_no;i++) {
      if (!alc_table[i].freed) {
	if (fprintf(fp, "%10d:  0x%p | %10d | %s\n",
		    i,
		    (void *)alc_table[i].adr,
		    (int)alc_table[i].size,
		    statVs[(int)alc_table[i].freed]) == EOF) {
	  if (!AlcErrNo) { AlcErrNo= ALCWRITE; }
	  return; } } }
    fprintf(fp, "=================\n"); }
  else if (!AlcErrNo) { AlcErrNo= ALCWRITE; }
}
/* =========================================================================
 *                     end of vsalloc.c
 * =========================================================================
 */
