/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsCal.c,v 1.3 1994/12/02 21:22:30 vs Exp vs $ */
/* -------------------------------------------------------------------------
 *                    vsCal.c
 * -------------------------------------------------------------------------
 * calibrations and related topics. (SK)
 */

#define _vsCal_c

#include "vsAlloc.h"
#include "vsTypes.h"
#include "vsCal.h"
#include "vsCalP.h"

#define NEWTON_PRECISION 0.01   /* exactness of calculation */
#define NEWTON_MAXITER	 10000  /* maximal number of iterations */
#define EPSILON 	 1.0e-10

#define COEF_DEF(c,t) (c->t.index >=0 && c->t.coef.num > 0)
#define CONS_DEF(c,t) (c->t.index >=0)
#define EFC_DEF(c,t) (c->t.index >=0 && c->t.efcp.num > 0)
#define STR_DEF(c,t) (c->t.index >=0 && *c->t.str != '\0')

#define COEF_EQ(c1,c2,t) (COEF_DEF(c1,t) == COEF_DEF(c2,t) &&\
			  c1->t.coef.num == c2->t.coef.num &&\
			  !memcmp(c1->t.coef.p, c2->t.coef.p,\
				  sizeof(double) * c1->t.coef.num))
#define CONS_EQ(c1,c2,t) (CONS_DEF(c1,t) == CONS_DEF(c2,t) &&\
			  c1->t.cons.c == c2->t.cons.c)
#define EFC_EQ(c1,c2,t) (EFC_DEF(c1,t) == EFC_DEF(c2,t) &&\
			 c1->t.efcp.num == c2->t.efcp.num &&\
			 !memcmp(c1->t.efcp.p, c2->t.efcp.p,\
				 sizeof(EfcPnt_t) * c1->t.efcp.num))
#define STR_EQ(c1,c2,t) (STR_DEF(c1,t) == STR_DEF(c2,t) &&\
			 !strcmp(c1->t.str, c2->t.str))

/* ---------------------------------------------------------------------- */
/* ============== */
/* error messages */
/* ============== */

char *CalError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* CALNOERR:  */ "vsCal: no error",
  /* CALPOINT:  */ "vsCal: illegal NULL-pointer",
  /* CALOPEN:   */ "vsCal: could not open calibration file",
  /* CALNOFIL:  */ "vsCal: illegal NULL-filepointer",
  /* CAL0READ:  */ "vsCal: no values read from file",
  /* CALMCOEF:  */ "vsCal: to many coefficients found",
  /* CALMCONS:  */ "vsCal: to many constants found",
  /* CALMEFC:   */ "vsCal: to many efficiency points found",
  /* CALMSTR:   */ "vsCal: to many strings found",
  /* CALSTRL:   */ "vsCal: internal string dimension exceeded",
  /* CALNODEF:  */ "vsCal: no legal calibration defined",
  /* CALWRITE:  */ "vsCal: could not write calibrationfile",
  /* CALRANGE:  */ "vsCal: position out of efc-calibration range",
  /* CALNEGEFC: */ "vsCal: negative value for efc-calibration",
  /* CALNOKEY:  */ "vsCal: keyword not found",
  /* CALEQXVAL: */ "vsCal: pairs with equal x-values for coefficient calculation",
  /* CALTYPE:   */ "vsCal: operation for calibration type not implemented",
  /* CALCLUST:  */ "vsCal: no clusterfile found",
  /* CALALLOC:  */ "vsCal: memory allocation failed",
  /* CALEFMT: */ "vsCal: wrong efc format (must be: x, y, dy)",
  /* CALUPAR: */ "vsCal: unexpected parameters found in file",
  /* default:   */ "vsCal: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
/* ============================ */
/* initialize/free calibrations */
/* ============================ */

#define INIT_CAL(c)  { (c).index = -1; *(c).file = '\0'; }

int CalInit(c)
     Cal_t c;
{
  if (NULL == c) { CalErrNo = CALPOINT; return FALSE; }
  else {
    INIT_CAL(c->position);
    INIT_CAL(c->width);
    INIT_CAL(c->tl);
    INIT_CAL(c->tr);
    INIT_CAL(c->sh);
    INIT_CAL(c->sw);
    INIT_CAL(c->vc);
    INIT_CAL(c->start); c->start.cons.c= 0.0;
    INIT_CAL(c->efc); c->efc.efcp.p= NULL; c->efc.efcp.num= 0;
    INIT_CAL(c->unit);
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int CalFree(c)
     Cal_t c;
{
  if (NULL == c) { CalErrNo = CALPOINT; return FALSE; }
  else {
    INIT_CAL(c->position);
    INIT_CAL(c->width);
    INIT_CAL(c->tl);
    INIT_CAL(c->tr);
    INIT_CAL(c->sh);
    INIT_CAL(c->sw);
    INIT_CAL(c->vc);
    INIT_CAL(c->start); c->start.cons.c= 0.0;
    INIT_CAL(c->efc);
    Free(c->efc.efcp.p); c->efc.efcp.num= 0; c->efc.efcp.siz= 0;
    INIT_CAL(c->unit);
    return TRUE; }
}

#undef INIT_CAL
/* ---------------------------------------------------------------------- */
/* =================== */
/* delete calibrations */
/* =================== */

int CalDelete(typ, c)
     int typ;
     Cal_t c;
{
  if (NULL == c) { CalErrNo = CALPOINT; return FALSE; }
  else {
    switch (typ) {
    case CAL_POS:   CalDelPos(c); break;
    case CAL_WIDTH: CalDelWidth(c); break;
    case CAL_TL:    CalDelTl(c); break;
    case CAL_TR:    CalDelTr(c); break;
    case CAL_SH:    CalDelSh(c); break;
    case CAL_SW:    CalDelSw(c); break;
    case CAL_VC:    CalDelVc(c); break;
    case CAL_START: CalDelStart(c); break;
    case CAL_EFC:   CalDelEfc(c); break;
    case CAL_UNIT:  CalDelUnit(c); break;
    case CAL_SET:
      CalDelPos(c);
      CalDelWidth(c);
      CalDelTl(c);
      CalDelTr(c);
      CalDelSh(c);
      CalDelSw(c);
      CalDelVc(c);
      CalDelStart(c);
      CalDelEfc(c);
      CalDelUnit(c);
      break;
    default:
      CalErrNo = CALTYPE; return FALSE; }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
#define DELETE_CAL(c,typ)  { (c)->typ.index = -1; *(c)->typ.file = '\0'; }
/* ---------------------------------------------------------------------- */
static void CalDelPos(c) Cal_t c; { DELETE_CAL(c, position); }
/* ---------------------------------------------------------------------- */
static void CalDelWidth(c) Cal_t c; { DELETE_CAL(c, width); }
/* ---------------------------------------------------------------------- */
static void CalDelTl(c) Cal_t c; { DELETE_CAL(c, tl); }
/* ---------------------------------------------------------------------- */
static void CalDelTr(c) Cal_t c; { DELETE_CAL(c, tr); }
/* ---------------------------------------------------------------------- */
static void CalDelSh(c) Cal_t c; { DELETE_CAL(c, sh); }
/* ---------------------------------------------------------------------- */
static void CalDelSw(c) Cal_t c; { DELETE_CAL(c, sw); }
/* ---------------------------------------------------------------------- */
static void CalDelVc(c) Cal_t c; { DELETE_CAL(c, vc); }
/* ---------------------------------------------------------------------- */
static void CalDelStart(c)
     Cal_t c;
{
  DELETE_CAL(c, start); c->start.cons.c = 0.0;
}
/* ---------------------------------------------------------------------- */
static void CalDelEfc(c)
     Cal_t c;
{
  DELETE_CAL(c, efc);
  c->efc.efcp.num= 0;
}
/* ---------------------------------------------------------------------- */
static void CalDelUnit(c) Cal_t c; { DELETE_CAL(c, unit); }

#undef DELETE_CAL
/* ---------------------------------------------------------------------- */
/* ================= */
/* read calibrations */
/* ================= */

static int cluster;

int CalRead(typ, c, file)
     int typ;
     Cal_t c;
     char *file;
{
  FILE *fp;
  int ret= FALSE;

  /*  printf("CalRead Typ: %d\n", typ);*/
  if (NULL == c) {
    CalErrNo = CALPOINT;
    return FALSE;
  }
  else if (file == NULL || *file == '\0') {
    CalErrNo = CALNOFIL;
  }
  else if (NULL == (fp= fopen(file, "r"))) {
    CalErrNo = CALOPEN;
  }
  else {
    cluster= FindKeyword(fp, "calibrations");
    if (!cluster && typ == CAL_SET) {
      CalErrNo = CALCLUST;
    }
    else {
      switch (typ) {
      case CAL_POS:
	if (cluster) {
	  ReadStrCal(fp, file, "unit", &c->unit);
	  ReadConsCal(fp, file, "startchannel", &c->start);
	}
	ret= ReadCoefCal(fp, file, "position", &c->position); break;
      case CAL_WIDTH:
	ret= ReadCoefCal(fp, file, "width", &c->width); break;
      case CAL_TL:
	ret= ReadCoefCal(fp, file, "tailleft", &c->tl); break;
      case CAL_TR:
	ret= ReadCoefCal(fp, file, "tailright", &c->tr); break;
      case CAL_SH:
	ret= ReadCoefCal(fp, file, "stepheight", &c->sh); break;
      case CAL_SW:
	ret= ReadCoefCal(fp, file, "stepwidth", &c->sw); break;
      case CAL_VC:
	ret= ReadConsCal(fp, file, "vc", &c->vc); break;
      case CAL_START:
	ret= ReadConsCal(fp, file, "startchannel", &c->start); break;
      case CAL_EFC:
	ret= ReadEfcCal(fp, file, "efficiency", &c->efc); break;
      case CAL_UNIT:
	ret= ReadStrCal(fp, file, "unit", &c->unit); break;
      case CAL_SET:
	if (!CalDelete(CAL_SET, c)) {
	  ret= FALSE;
	}
	else {
	  ret= (ReadCoefCal(fp, file, "position", &c->position) |
		ReadCoefCal(fp, file, "width", &c->width) |
		ReadCoefCal(fp, file, "tailleft", &c->tl) |
		ReadCoefCal(fp, file, "tailright", &c->tr) |
		ReadCoefCal(fp, file, "stepheight", &c->sh) |
		ReadCoefCal(fp, file, "stepwidth", &c->sw) |
		ReadConsCal(fp, file, "vc", &c->vc) |
		ReadConsCal(fp, file, "startchannel", &c->start) |
		ReadEfcCal(fp, file, "efficiency", &c->efc) |
		ReadStrCal(fp, file, "unit", &c->unit));
	}
	break;
      default:
	CalErrNo = CALTYPE;
	break;
      }
    } 
    fclose(fp);
  }
  return ret;
}
/* ---------------------------------------------------------------------- */
#define READ_CAL(read,cpy,kw,cp,fp,f,to,tmp,n_cal) \
  {							\
    if (cluster) { rewind(fp); }			\
    if (!cluster || FindKeyword(fp, kw)) {		\
      do {						\
	if (read(fp, tmp)) { n_cal++; }			\
      } while (cluster && FindKeyword(fp, kw)); }	\
    if (n_cal > 0 && cpy(to, tmp)) {			\
      CalErrNo = CALNOERR;				\
      strncpy((cp)->file, f, CAL_STR);			\
      (cp)->index = n_cal; }				\
  }

static int CpyCoef(to, from)
     Coef_t *to;
     Coef_t *from;
{
  if (NULL == from || NULL == to) { CalErrNo = CALPOINT; }
  else if (from->num > CAL_COE) { CalErrNo = CALMCOEF; }
  else {
    int i= from->num;
    while (i > 0 && from->p[i-1] == 0.0) { i--; } /* determination of grad */
    if (i > 0) {
      to->num = i;
      for ( i--; i >= 0; i--) { to->p[i]= from->p[i]; }
      return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
/*
 * Load calibration coefficient
 */
static int ReadCoef(fp, c)
     FILE *fp;
     Coef_t *c;
{
  double tmp;
  Coef_t coef;
  coef.num= 0;
  if (NULL == fp) { CalErrNo = CALNOFIL; }
  else if (NULL == c) { CalErrNo = CALPOINT; }
  else {
    while(1 == fscanf(fp, "%lf", &tmp)) {
      if (coef.num >= CAL_COE) { CalErrNo = CALMCOEF; return FALSE; }
      else {
	coef.p[coef.num]= tmp;
	coef.num++; } }
    if (coef.num <= 0) { CalErrNo = CAL0READ; }
    else if (CpyCoef(c, &coef)) { return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int ReadCoefCal(fp, file, keyword, cal)
     FILE      *fp;
     char      *file;
     char      *keyword;
     CoefCal_t *cal;
{
  int n_cal= 0;
  Coef_t coef;


  READ_CAL(ReadCoef, CpyCoef, keyword, cal, fp, file, 
	   &cal->coef, &coef, n_cal);
  return (n_cal > 0);
}
/* ---------------------------------------------------------------------- */
static int CpyCons(to, from)
     Cons_t *to;
     Cons_t *from;
{
  if (NULL == from || NULL == to) { CalErrNo = CALPOINT; return FALSE; }
  else { to->c = from->c; return TRUE; }
} 
static int ReadCons(fp, c)
     FILE *fp;
     Cons_t *c;
{
  if (NULL == fp) { CalErrNo = CALNOFIL; }
  else if (NULL == c) { CalErrNo = CALPOINT; }
  else {
    double tmp;
    Cons_t cons;
    int n_cons= 0;
    while(1 == fscanf(fp, "%lf", &tmp)) {
      if (n_cons >= 1) { CalErrNo = CALMCONS; return FALSE; }
      else { cons.c = tmp; n_cons++; } }
    if (n_cons <= 0) { CalErrNo = CAL0READ; }
    else if (CpyCons(c, &cons)) { return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int ReadConsCal(fp, file, keyword, cal)
     FILE      *fp;
     char      *file;
     char      *keyword;
     ConsCal_t *cal;
{
  int n_cal= 0;
  Cons_t cons;
  READ_CAL(ReadCons, CpyCons, keyword, cal, fp, file, 
	   &cal->cons, &cons, n_cal);
  return (n_cal > 0);
}
/* ---------------------------------------------------------------------- */
static int CpyEfc(to, from)
     Efc_t *to;
     Efc_t *from;
{
  /*  printf("CpyEfc\n");*/
  if (NULL == from || NULL == from->p || NULL == to) {
    CalErrNo = CALPOINT;
  }
  else if (from->num > MEF_C) {
    CalErrNo = CALMEFC;
  }
  else {
    EfcPnt_t *ep= from->p;
    { int no= from->num, i;
      for (i= 0; i< no; i++) {
	int j;
	for (j= i+1; j< no; j++) {
	  if (ep[i].en == ep[j].en) {
	    CalErrNo = CALEQXVAL; return FALSE; } } } }
    ep= NULL;
    if (from->num > to->siz) {
      if (NULL == (ep= (EfcPnt_t *)Malloc(from->num * sizeof(EfcPnt_t)))) {
	CalErrNo= CALALLOC; }
      else {
	/* Free(to->p); NO! to->p is a local variable in ReadEfcCal () !!!  os@ikp */
	to->p = ep; 
	to->siz = from->num; } }
    else { ep = to->p; }
    if (ep != NULL) {
      int i;
      for (i = 0; i < from->num; i++) {
	ep[i].en = from->p[i].en;
	ep[i].ef = from->p[i].ef;
	ep[i].df = from->p[i].df; }
      to->num = from->num;
      return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int ReadEfc(fp, c)
     FILE *fp;
     Efc_t *c;
{
  /*  printf("ReadCal\n");*/
  if (NULL == fp) {
    CalErrNo = CALNOFIL;
  }
  else if (NULL == c) {
    CalErrNo = CALPOINT;
  }
  else {
    char* lineTmp;
    double en_tmp, ef_tmp, df_tmp;
    EfcPnt_t ep[MEF_C];
    Efc_t efc;
    efc.num= 0;
    efc.p= ep;
    lineTmp = malloc (sizeof(char)*10240);
    while( NULL != fgets(lineTmp, 10240, fp)) {
      /* I hope this does not reject too much data from being read*/
      if (lineTmp && *lineTmp) {
	char* linePtr = NULL;
	/*	printf("%d\n", linePtr);*/
	linePtr = strtok (lineTmp, " ");
	sscanf(linePtr, "%lf", &en_tmp);
	if (NULL == (linePtr = strtok (NULL, " "))){
	  CalErrNo = CALEFMT;
	  return FALSE;
	}
	sscanf(linePtr, "%lf", &ef_tmp);
	if (NULL == (linePtr = strtok (NULL, " "))){
	  CalErrNo = CALEFMT;
	  return FALSE;
	}
	sscanf(linePtr, "%lf", &df_tmp);
	if (NULL != (linePtr = strtok (NULL, " "))){
	  CalErrNo = CALUPAR;
	  return FALSE;
	}

	if (efc.num >= MEF_C) {
	  CalErrNo = CALMEFC;
	  return FALSE;
	}
	else {
	  efc.p[efc.num].en = en_tmp;
	  efc.p[efc.num].ef = ef_tmp;
	  efc.p[efc.num].df = df_tmp;
	  efc.num++;
	}
      }
      else {
	CalErrNo = CAL0READ;
	return FALSE;
      }
    }
    if (efc.num <= 0) {
      CalErrNo = CAL0READ;
    }
    else if (CpyEfc(c, &efc)) {
      return TRUE;
    }
  }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int ReadEfcCal(fp, file, keyword, cal)
     FILE     *fp;
     char      *file;
     char     *keyword;
     EfcCal_t *cal;
{
  int n_cal= 0;
  EfcPnt_t ep[MEF_C];
  Efc_t efc;
  efc.num= 0;
  efc.p= ep;

  /*  printf("ReadEfcCal\n");*/
  READ_CAL(ReadEfc, CpyEfc, keyword, cal, fp, file, &cal->efcp, &efc, n_cal); 
  /* ok, here is what READ_CAL really does:
     if (cluster) { rewind(fp); }			
     if (!cluster || FindKeyword(fp, keyword)) {
     do {
     if (ReadEfc(fp, &efc)) { n_cal++; }
     } while (cluster && FindKeyword(fp, keyword)); }
     if (n_cal > 0 && CpyEfc(&cal->efcp, &efc)) {
     CalErrNo = CALNOERR;
     strncpy((cal)->file, file, CAL_STR);
     (cal)->index = n_cal; }
     */

  return n_cal;
}
/* ---------------------------------------------------------------------- */
static int CpyStr(to, from)
     char *to;
     char *from;
{
  int l;
  if (NULL == from || NULL == to) { CalErrNo = CALPOINT; }
  else if ((l= strlen(from)) >= CAL_STR) { CalErrNo = CALSTRL; }
  else {
    strcpy(to, from);
    if (l > 0 && to[l-1] == '\n') { to[l-1]= '\0'; }
    return TRUE; }
  return FALSE; 
}
/* ---------------------------------------------------------------------- */
static int ReadStr(fp, s)
     FILE *fp;
     char *s;
{
  char str[CAL_STR];
  if (NULL == fp) { CalErrNo = CALNOFIL; }
  else if (NULL == s) { CalErrNo = CALPOINT; }
  else {
    int n_lines= 0;
    int fpos= ftell(fp);
    char tmp[CAL_STR];
    while(NULL != fgets(tmp, CAL_STR, fp)) {
      if (!strncmp("$_", tmp, 2)) { fseek(fp, fpos, SEEK_SET); break; }
      else if (n_lines >= 1) { CalErrNo = CALMSTR; return FALSE; }
      else { CpyStr(str, tmp); n_lines++; } }
    if (n_lines <= 0) { CalErrNo = CAL0READ; }
    else if (CpyStr(s, str)) { return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int ReadStrCal(fp, file, keyword, cal)
     FILE     *fp;
     char      *file;
     char     *keyword;
     StrCal_t *cal;
{
  int n_cal= 0;
  char str[CAL_STR];

  READ_CAL(ReadStr, CpyStr, keyword, cal, fp, file, cal->str, str, n_cal);
  return n_cal;
}

#undef READ_CAL
/* ---------------------------------------------------------------------- */
static int FindKeyword(fp, keyword)
     FILE *fp;
     char *keyword;
{
  if (NULL == fp) { CalErrNo = CALPOINT; }
  else {
    char word[CAL_STR];
    char line[CAL_STR];
    while (NULL != fgets(line, CAL_STR, fp)) {
      if (sscanf(line, "$_%s", word) > 0 &&
	  !strcmp(word, keyword)) { return TRUE; } }
    CalErrNo = CALNOKEY;
    rewind(fp); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
/* ================== */
/* write calibrations */
/* ================== */

int CalWrite(typ, c, file)
     int typ;
     Cal_t c;
     char *file;
{
  FILE *fp;

  if (NULL == c) { CalErrNo = CALPOINT; }
  else if (file == NULL || *file == '\0') { CalErrNo = CALNOFIL; }
  else if (((typ == CAL_SET || typ == CAL_POS) && COEF_DEF(c, position)) ||
	   ((typ == CAL_SET || typ == CAL_WIDTH) && COEF_DEF(c, width)) ||
	   ((typ == CAL_SET || typ == CAL_TL) && COEF_DEF(c, tl)) ||
	   ((typ == CAL_SET || typ == CAL_TR) && COEF_DEF(c, tr)) ||
	   ((typ == CAL_SET || typ == CAL_SH) && COEF_DEF(c, sh)) ||
	   ((typ == CAL_SET || typ == CAL_SW) && COEF_DEF(c, sw)) ||
	   ((typ == CAL_SET || typ == CAL_VC) && CONS_DEF(c, vc)) ||
	   ((typ == CAL_SET || typ == CAL_START) && CONS_DEF(c, start)) ||
	   ((typ == CAL_SET || typ == CAL_EFC) && EFC_DEF(c, efc)) ||
	   ((typ == CAL_SET || typ == CAL_UNIT) && STR_DEF(c, unit))) {
				/* at least one calibration defined */
    if (NULL != (fp= fopen(file, "r")) && FindKeyword(fp, "calibrations")) {
      fclose(fp);
      if (NULL == (fp= fopen(file, "a"))) {
	CalErrNo = CALOPEN; } }
    else {
      if (NULL != fp) { fclose(fp); }
      if (NULL == (fp= fopen(file, "w"))) {
	CalErrNo = CALOPEN; }
      else if (fprintf(fp, "$_calibrations\n") == EOF) {
	fclose(fp); CalErrNo = CALWRITE; } }
    if (NULL != fp) {
      int ret= FALSE;
      switch (typ) {
      case CAL_POS:
	if (STR_DEF(c, unit)) { WriteStrCal(fp, file, "unit", &c->unit); }
	if (CONS_DEF(c, start)) {
	  WriteConsCal(fp, file, "startchannel", &c->start); }
	ret= WriteCoefCal(fp, file, "position", &c->position);
	break;
      case CAL_WIDTH: ret= WriteCoefCal(fp, file, "width", &c->width); break;
      case CAL_TL:    ret= WriteCoefCal(fp, file, "tailleft", &c->tl); break;
      case CAL_TR:    ret= WriteCoefCal(fp, file, "tailright", &c->tr); break;
      case CAL_SH:    ret= WriteCoefCal(fp, file, "stepheight", &c->sh); break;
      case CAL_SW:    ret= WriteCoefCal(fp, file, "stepwidth", &c->sw); break;
      case CAL_VC:    ret= WriteConsCal(fp, file, "vc", &c->vc); break;
      case CAL_START: ret= WriteConsCal(fp, file, "startchannel", &c->start);break;
      case CAL_EFC:   ret= WriteEfcCal(fp, file, "efficiency", &c->efc); break;
      case CAL_UNIT:  ret= WriteStrCal(fp, file, "unit", &c->unit); break;
      case CAL_SET:
	ret= ((!COEF_DEF(c, position) ||
	       WriteCoefCal(fp, file, "position", &c->position)) &&
	      (!COEF_DEF(c, width) ||
	       WriteCoefCal(fp, file, "width", &c->width)) &&
	      (!COEF_DEF(c, tl) ||
	       WriteCoefCal(fp, file, "tailleft", &c->tl)) &&
	      (!COEF_DEF(c, tr) ||
	       WriteCoefCal(fp, file, "tailright", &c->tr)) &&
	      (!COEF_DEF(c, sh) ||
	       WriteCoefCal(fp, file, "stepheight", &c->sh)) &&
	      (!COEF_DEF(c, sw) ||
	       WriteCoefCal(fp, file, "stepwidth", &c->sw)) &&
	      (!CONS_DEF(c, vc) ||
	       WriteConsCal(fp, file, "vc", &c->vc)) &&
	      (!CONS_DEF(c, start) ||
	       WriteConsCal(fp, file, "startchannel", &c->start)) &&
	      (!EFC_DEF(c, efc) ||
	       WriteEfcCal(fp, file, "efficiency", &c->efc)) &&
	      (!STR_DEF(c, unit) ||
	       WriteStrCal(fp, file, "unit", &c->unit)));
	break;
      default:
	CalErrNo = CALTYPE; ret= FALSE; break; }
      fclose(fp);
      return ret; } }
  else { CalErrNo = CALNODEF; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
#define WRITE_CAL(write,kw,cp,fp,f,src,st) \
  {							\
    if ((st = (fprintf(fp, "$_%s\n", kw) != EOF))) {	\
      if ((st = write(fp, src))) {			\
	strncpy((cp)->file, f, CAL_STR);		\
	(cp)->index = 1; } }				\
    else { CalErrNo = CALWRITE; }			\
  }

static int WriteCoef(fp, c)
     FILE *fp;
     Coef_t *c;
{
  if (NULL != fp && NULL != c) {
    int i;
    for (i= 0; i< c->num; i++) {
      if (fprintf(fp, "%20g\n", c->p[i]) == EOF) { return FALSE; } }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int WriteCoefCal(fp, file, keyword, cal)
     FILE      *fp;
     char      *file;
     char      *keyword;
     CoefCal_t *cal;
{
  int local_stat;
  WRITE_CAL(WriteCoef, keyword, cal, fp, file, &cal->coef, local_stat);
  return local_stat;
}
/* ---------------------------------------------------------------------- */
static int WriteCons(fp, c)
     FILE *fp;
     Cons_t *c;
{
  if (NULL != fp && NULL != c) {
    if (fprintf(fp, "%20g\n", c->c) == EOF) { return FALSE; }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int WriteConsCal(fp, file, keyword, cal)
     FILE      *fp;
     char      *file;
     char      *keyword;
     ConsCal_t *cal;
{
  int local_stat;
  WRITE_CAL(WriteCons, keyword, cal, fp, file, &cal->cons, local_stat);
  return local_stat;
}
/* ---------------------------------------------------------------------- */
static int WriteStr(fp, c)
     FILE *fp;
     char *c;
{
  if (NULL != fp && NULL != c) {
    if (fprintf(fp, "%s\n", c) == EOF) { return FALSE; }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int WriteStrCal(fp, file, keyword, cal)
     FILE     *fp;
     char     *file;
     char     *keyword;
     StrCal_t *cal;
{
  int local_stat;
  WRITE_CAL(WriteStr, keyword, cal, fp, file, cal->str, local_stat);
  return local_stat;
}
/* ---------------------------------------------------------------------- */
static int WriteEfc(fp, c)
     FILE *fp;
     Efc_t *c;
{
  if (NULL != fp && NULL != c) {
    int i;
    for (i= 0; i< c->num; i++) {
      if (fprintf(fp, "%20g %20g %15g\n",
		  c->p[i].en, c->p[i].ef, c->p[i].df) == EOF) {
	return FALSE; } }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int WriteEfcCal(fp, file, keyword, cal)
     FILE     *fp;
     char     *file;
     char     *keyword;
     EfcCal_t *cal;
{
  int local_stat;
  WRITE_CAL(WriteEfc, keyword, cal, fp, file, &cal->efcp, local_stat);
  return local_stat;
}

#undef WRITE_CAL
/* ---------------------------------------------------------------------- */
/* ================== */
/* copy calibrations */
/* ================== */

int CalCopy(typ, to, from)
     int typ;
     Cal_t to;
     Cal_t from;
{
  if (NULL == from || NULL == to) { CalErrNo = CALPOINT; }
  else if (((typ == CAL_SET || typ == CAL_POS) &&
	    COEF_DEF(from, position)) ||
	   ((typ == CAL_SET || typ == CAL_WIDTH) && COEF_DEF(from, width)) ||
	   ((typ == CAL_SET || typ == CAL_TL) && COEF_DEF(from, tl)) ||
	   ((typ == CAL_SET || typ == CAL_TR) && COEF_DEF(from, tr)) ||
	   ((typ == CAL_SET || typ == CAL_SH) && COEF_DEF(from, sh)) ||
	   ((typ == CAL_SET || typ == CAL_SW) && COEF_DEF(from, sw)) ||
	   ((typ == CAL_SET || typ == CAL_VC) && CONS_DEF(from, vc)) ||
	   ((typ == CAL_SET || typ == CAL_START) && CONS_DEF(from, start)) ||
	   ((typ == CAL_SET || typ == CAL_EFC) && EFC_DEF(from, efc)) ||
	   ((typ == CAL_SET || typ == CAL_UNIT) && STR_DEF(from, unit))) {
				/* at least one calibration defined */
    int ret= FALSE;
    switch (typ) {
    case CAL_POS:   ret= CopyCoefCal(&to->position, &from->position); break;
    case CAL_WIDTH: ret= CopyCoefCal(&to->width, &from->width); break;
    case CAL_TL:    ret= CopyCoefCal(&to->tl, &from->tl); break;
    case CAL_TR:    ret= CopyCoefCal(&to->tr, &from->tr); break;
    case CAL_SH:    ret= CopyCoefCal(&to->sh, &from->sh); break;
    case CAL_SW:    ret= CopyCoefCal(&to->sw, &from->sw); break;
    case CAL_VC:    ret= CopyConsCal(&to->vc, &from->vc); break;
    case CAL_START: ret= CopyConsCal(&to->start, &from->start); break;
    case CAL_EFC:   ret= CopyEfcCal(&to->efc, &from->efc); break;
    case CAL_UNIT:  ret= CopyStrCal(&to->unit, &from->unit); break;
    case CAL_SET:
      ret= ((!COEF_DEF(from, position) ||
	     CopyCoefCal(&to->position, &from->position)) &&
	    (!COEF_DEF(from, width) ||
	     CopyCoefCal(&to->width, &from->width)) &&
	    (!COEF_DEF(from, tl) ||
	     CopyCoefCal(&to->tl, &from->tl)) &&
	    (!COEF_DEF(from, tr) ||
	     CopyCoefCal(&to->tr, &from->tr)) &&
	    (!COEF_DEF(from, sh) ||
	     CopyCoefCal(&to->sh, &from->sh)) &&
	    (!COEF_DEF(from, sw) ||
	     CopyCoefCal(&to->sw, &from->sw)) &&
	    (!CONS_DEF(from, vc) ||
	     CopyConsCal(&to->vc, &from->vc)) &&
	    (!CONS_DEF(from, start) ||
	     CopyConsCal(&to->start, &from->start)) &&
	    (!EFC_DEF(from, efc) ||
	     CopyEfcCal(&to->efc, &from->efc)) &&
	    (!STR_DEF(from, unit) ||
	     CopyStrCal(&to->unit, &from->unit)));
      break;
    default:
      CalErrNo = CALTYPE; break; }
    return ret; }
  else { CalErrNo = CALNODEF; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
#define COPY_CAL(cpy,to,tp,from,fp,ret) \
  {							\
    if ((ret = cpy(tp, fp))) {				\
      (to)->index = (from)->index;			\
      strncpy((to)->file, (from)->file, CAL_STR); }	\
  }

static int CopyCoefCal(to, from)
     CoefCal_t *to;
     CoefCal_t *from;
{
  int ret;
  COPY_CAL(CpyCoef, to, &to->coef, from, &from->coef, ret);
  return ret;
}
/* ---------------------------------------------------------------------- */
static int CopyConsCal(to, from)
     ConsCal_t *to;
     ConsCal_t *from;
{
  int ret;
  COPY_CAL(CpyCons, to, &to->cons, from, &from->cons, ret);
  return ret;
}
/* ---------------------------------------------------------------------- */
static int CopyEfcCal(to, from)
     EfcCal_t *to;
     EfcCal_t *from;
{
  int ret;
  COPY_CAL(CpyEfc, to, &to->efcp, from, &from->efcp, ret);
  return ret;
}
/* ---------------------------------------------------------------------- */
static int CopyStrCal(to, from)
     StrCal_t *to;
     StrCal_t *from;
{
  int ret;
  COPY_CAL(CpyStr, to, to->str, from, from->str, ret);
  return ret;
}

#undef COPY_CAL
/* ---------------------------------------------------------------------- */
/* ================== */
/* exist calibrations */
/* ================== */

int CalExists(typ, c)
     int typ;
     Cal_t c;
{
  if (NULL == c) { CalErrNo = CALPOINT; return FALSE; }
  else {
    switch (typ) {
    case CAL_POS:   return COEF_DEF(c, position);
    case CAL_WIDTH: return COEF_DEF(c, width);
    case CAL_TL:    return COEF_DEF(c, tl);
    case CAL_TR:    return COEF_DEF(c, tr);
    case CAL_SH:    return COEF_DEF(c, sh);
    case CAL_SW:    return COEF_DEF(c, sw);
    case CAL_VC:    return CONS_DEF(c, vc);
    case CAL_START: return CONS_DEF(c, start);
    case CAL_EFC:   return EFC_DEF(c, efc);
    case CAL_UNIT:  return STR_DEF(c, unit);
    default:
      CalErrNo = CALTYPE; return FALSE; } }
}
/* ---------------------------------------------------------------------- */
/* ================== */
/* check calibrations */
/* ================== */

int CalCheck(typ, c)
     int typ;
     Cal_t c;
{
  if (NULL == c) { CalErrNo = CALPOINT; return FALSE; }
  else {
    switch (typ) {
    case CAL_POS:   return COEF_DEF(c, position);
    case CAL_WIDTH: return COEF_DEF(c, width);
    case CAL_TL:    return COEF_DEF(c, tl);
    case CAL_TR:    return COEF_DEF(c, tr);
    case CAL_SH:    return COEF_DEF(c, sh);
    case CAL_SW:    return COEF_DEF(c, sw);
    case CAL_VC:    return CONS_DEF(c, vc);
    case CAL_START: return CONS_DEF(c, start);
    case CAL_EFC:   return EFC_DEF(c, efc);
    case CAL_UNIT:  return STR_DEF(c, unit);
    case CAL_SET:   return (COEF_DEF(c, position) ||
			    COEF_DEF(c, width) ||
			    COEF_DEF(c, tl) ||
			    COEF_DEF(c, tr) ||
			    COEF_DEF(c, sh) ||
			    COEF_DEF(c, sw) ||
			    CONS_DEF(c, vc) ||
			    CONS_DEF(c, start) ||
			    EFC_DEF(c, efc) ||
			    STR_DEF(c, unit));
    default:
      CalErrNo = CALTYPE; return FALSE; } }
}
/* ---------------------------------------------------------------------- */
/* ==================== */
/* compare calibrations */
/* ==================== */

int CalEqual(typ, c1, c2)
     int typ;
     Cal_t c1;
     Cal_t c2;
{
  if (!c1 || !c2) { CalErrNo = CALPOINT; return FALSE; }
  else {
    switch (typ) {
    case CAL_POS:   return COEF_EQ(c1, c2, position);
    case CAL_WIDTH: return COEF_EQ(c1, c2, width);
    case CAL_TL:    return COEF_EQ(c1, c2, tl);
    case CAL_TR:    return COEF_EQ(c1, c2, tr);
    case CAL_SH:    return COEF_EQ(c1, c2, sh);
    case CAL_SW:    return COEF_EQ(c1, c2, sw);
    case CAL_VC:    return CONS_EQ(c1, c2, vc);
    case CAL_START: return CONS_EQ(c1, c2, start);
    case CAL_EFC:   return EFC_EQ(c1, c2, efc);
    case CAL_UNIT:  return STR_EQ(c1, c2, unit);
    case CAL_SET:   return (COEF_EQ(c1, c2, position) ||
			    COEF_EQ(c1, c2, width) ||
			    COEF_EQ(c1, c2, tl) ||
			    COEF_EQ(c1, c2, tr) ||
			    COEF_EQ(c1, c2, sh) ||
			    COEF_EQ(c1, c2, sw) ||
			    CONS_EQ(c1, c2, vc) ||
			    CONS_EQ(c1, c2, start) ||
			    EFC_EQ(c1, c2, efc) ||
			    STR_EQ(c1, c2, unit));
    default:
      CalErrNo = CALTYPE; return FALSE; } }
}
/* ---------------------------------------------------------------------- */
/* ===================== */
/* calibration functions */
/* ===================== */

#define HORNER(cs,x,ret) { register int i=(cs).num-1; register double *c=(cs).p; ret = 0.0; while (i >= 0) { ret = ret * (x) + c[i--]; } }

#define CAL_COEF(cp,typ,x,ret) \
  if ((cp) == NULL) {				\
    CalErrNo = CALPOINT; ret = channel; }	\
  else if ((cp)->typ.index < 0 ||		\
           (cp)->typ.coef.num <= 0) {		\
    ret = channel; }				\
  else {					\
    ret = 0.0;					\
    x += (cp)->start.cons.c;			\
    HORNER((cp)->typ.coef, x, ret); }

double CalP(cp, channel)
     Cal_t cp;
     double channel;
{
  register double tmp;
  CAL_COEF(cp, position, channel, tmp);
  return tmp;
}
/* ---------------------------------------------------------------------- */
/* ARGSUSED */
double CalUnity(cp, x)
     Cal_t cp;
     double x;
{
  return 1.0;
}
/* ---------------------------------------------------------------------- */
/* ARGSUSED */
double CalId(cp, x)
     Cal_t cp;
     double x;
{
  return x;
}
/* ---------------------------------------------------------------------- */
double CalS(cp, channel) Cal_t cp; double channel;
{
  register double tmp; CAL_COEF(cp, width, channel, tmp);
  return tmp * FWHM_TO_SIGMA;
}
/* ---------------------------------------------------------------------- */
double CalW(cp, channel) Cal_t cp; double channel;
{
  register double tmp; CAL_COEF(cp, width, channel, tmp);
  return tmp;
}
/* ---------------------------------------------------------------------- */
double CalTl(cp, channel) Cal_t cp; double channel;
{ register double tmp; CAL_COEF(cp, tl, channel, tmp); return tmp; }

double CalTr(cp, channel) Cal_t cp; double channel;
{ register double tmp; CAL_COEF(cp, tr, channel, tmp); return tmp; }

double CalSh(cp, channel) Cal_t cp; double channel;
{ register double tmp; CAL_COEF(cp, sh, channel, tmp); return tmp; }

double CalSw(cp, channel) Cal_t cp; double channel;
{ register double tmp; CAL_COEF(cp, sw, channel, tmp); return tmp; }

#undef CAL_COEF
#define CAL_COEF0(cp,typ,x,ret) \
  if ((cp) == NULL) {				\
    CalErrNo = CALPOINT; ret = channel; }	\
  else if ((cp)->typ.index < 0 ||		\
           (cp)->typ.coef.num <= 0) {		\
    ret = channel; }			 	\
  else {					\
    ret = 0.0;					\
    HORNER((cp)->typ.coef, x, ret); }

static double CalP0(cp, channel) Cal_t cp; double channel;
{ register double tmp; CAL_COEF0(cp, position, channel, tmp); return tmp; }

#undef CAL_COEF0
#undef HORNER
				/* ---------------------------------------- */
double CalC(cp, position)
     Cal_t cp;
     double position;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return position; }
  else if (cp->position.index < 0 || cp->position.coef.num < 2) {
    return position; }
  else {
    register int i, j;
    double *cfc= cp->position.coef.p;
    double x= 0.0;		/* iteration value for c_cal */
    double de;			/* deviation of e_cal(x,) from energy */
    double gradient;		/* gradient of polynom at x */
    int n_cfc1= cp->position.coef.num-1;

    if (cp->start.cons.c != 0.0 && cp->start.index >= 0) {
      x = (position - cfc[0]) / cfc[1];
      de = CalP0(cp, x) - position;
      for (j= 0; fabs(de) > NEWTON_PRECISION && j < NEWTON_MAXITER; j++) {
	gradient = 0.0;
	for (i = n_cfc1; i > 0; i--) {
	  gradient = gradient * x + (double) i * cfc[i]; }
	x -= de/gradient;
	de = CalP0(cp, x) - position; }
      return (x - cp->start.cons.c); }
    else {
      x = (position - cfc[0]) / cfc[1];
      de = CalP(cp, x) - position;
      for (j= 0; fabs(de) > NEWTON_PRECISION && j < NEWTON_MAXITER; j++) {
	gradient = 0.0;
	for (i = n_cfc1; i > 0; i--) {
	  gradient = gradient * x + (double) i * cfc[i]; }
	x -= de/gradient;
	de = CalP(cp, x) - position; }
      return x; } }
}
				/* ---------------------------------------- */
double CalVc(cp)
     Cal_t cp;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return 1.0; }
  else if (cp->vc.index < 0) { return 1.0; }
  else { return cp->vc.cons.c; }
}
				/* ---------------------------------------- */
double CalStart(cp)
     Cal_t cp;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return 0.0; }
  else if (cp->start.index < 0) { return 0.0; }
  else { return cp->start.cons.c; }
}
				/* ---------------------------------------- */
double CalEfc(cp, channel)
     Cal_t cp;
     double channel;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return 1.0; }
  else if (cp->efc.index < 0 || cp->efc.efcp.num == 0) {
    return 1.0; }
  else { return CalEfcP(cp, CalP(cp, channel)); }
}
				/* ---------------------------------------- */
double CalEfcP(cp, position)
     Cal_t cp;
     double position;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return 1.0; }
  else if (cp->efc.index < 0 || cp->efc.efcp.num <= 1) {
    return 1.0; }
  else {
    int low= 0;
    int high= cp->efc.efcp.num - 1;
    int middle= (low + high)/2;
    EfcPnt_t *pnt= cp->efc.efcp.p;
    double p1, p2, f1, f2, /* d1, d2, */ dp, t, fk/*, dk */;

    if (pnt[low].en > position || pnt[high].en < position) {
      CalErrNo = CALRANGE; return 1.0; }
    while (low+1 < high) {
      if (pnt[middle].en <= position) {
	low= middle; middle= (low + high)/2; }
      if (pnt[middle].en > position) {
	high= middle; middle= (low + high)/2; } }

    p1 = pnt[low].en; p2 = pnt[high].en;
    f1 = pnt[low].ef; f2 = pnt[high].ef;
    /*  d1 = pnt[low].df; d2 = pnt[high].df;*/
    dp = p2 - p1; 
  
    t = (position - p1) / dp;
    fk = f1 + t * (f2 - f1);
    if (fk < EPSILON) { CalErrNo = CALNEGEFC; return 1.0; }
    /*  dk = d1 + t * (d2 - d1); */
    return 1.0 / fk; }
}
				/* ---------------------------------------- */
char *CalUnit(cp)
     Cal_t cp;
{
  if (cp == NULL) { CalErrNo = CALPOINT; return NULL; }
  else if (cp->unit.index < 0) {
    return (CalCheck(CAL_POS, cp) ? "keV" : "Chn"); }
  else { return cp->unit.str; }
}
				/* ---------------------------------------- */
/* ARGSUSED */
char *CalChn(cp)
     Cal_t cp;
{
  return "Chn";
}
/* ============================= */
/* derived calibration functions */
/* ============================= */

double CalCWidth(cp, pos, w_pos)
     Cal_t cp;
     double pos;
     double w_pos;
{
  return (CalC(cp, pos + (w_pos * .5)) - CalC(cp, pos - (w_pos * .5)));
}
				/* ---------------------------------------- */
double CalCDwidth(cp, pos, w_pos, dw_pos)
     Cal_t cp;
     double pos;
     double w_pos;
     double dw_pos;
{
  if (w_pos == 0.0) { return 0.0; }
  return (CalCWidth(cp, pos, w_pos) * dw_pos / w_pos);
} 
				/* ---------------------------------------- */
double CalPWidth(cp, chn, w_chn)
     Cal_t cp;
     double chn;
     double w_chn;
{
  return (CalP(cp, chn + (w_chn * .5)) - CalP(cp, chn - (w_chn * .5)));
}
				/* ---------------------------------------- */
double CalPDwidth(cp, chn, w_chn, dw_chn)
     Cal_t cp;
     double chn;
     double w_chn;
     double dw_chn;
{
  if (w_chn == 0.0) { return 0.0; }
  else { return (CalPWidth(cp, chn, w_chn) * dw_chn / w_chn); }
}
/* ========================================================================= */
