/* $Header: /usr/files/home/Gamma/vs/tv/edit/tv/RCS/vsCal.h,v 1.1 1993/09/17 14:55:08 vs Exp vs $ */ 

#ifndef _vsCal_h
#define _vsCal_h

#include <tv/vsConfig.h>
#include <tv/vsTypes.h>

#define MEF_C 4096		/* max. number of stuetzstellen */

#define SIGMA_TO_FWHM	 2.35482004503094930000 /* == 2* SQRT(2* M_LN2) */
#define FWHM_TO_SIGMA	 0.42466090014400953000 /* == 1./SIGMA_TO_FWHM  */

#define CAL_SET		0
#define CAL_POS		1
#define CAL_WIDTH	2
#define CAL_TL		3
#define CAL_TR		4
#define CAL_SH		5
#define CAL_SW		6
#define CAL_VC		7
#define CAL_START	8
#define CAL_EFC		9
#define CAL_UNIT	10

#define CAL_STR	256		/* max. stringlength */
#define CAL_COE	  16		/* max. number of coefficients in calibration*/

typedef struct {		/* coefficient array */
  double p[CAL_COE];
  int num;
} Coef_t;

typedef struct {		/* polynom calibrations */
  int index;
  char file[CAL_STR];
  Coef_t coef;
} CoefCal_t;

typedef struct {		/* constant */
  double c;
} Cons_t;

typedef struct {		/* constant calibrations */
  int index;
  char file[CAL_STR];
  Cons_t cons;
} ConsCal_t;

typedef struct {		/* stuetzstelle */
  double en;			/* calibrated position */
  double ef;			/* efficiency */
  double df;			/* uncertainty of efficiency */
} EfcPnt_t;

typedef struct {		/* stuetzstellen array */
  int num;
  int siz;
  EfcPnt_t *p;
} Efc_t;

typedef struct {		/* efficiency calibration */
  int index;
  char file[CAL_STR];
  Efc_t efcp;
} EfcCal_t;

typedef struct {		/* unit of calibration */
  int index;
  char file[CAL_STR];
  char str[CAL_STR];
} StrCal_t;

typedef struct _CalRec {	/* calibrations */
  CoefCal_t position;
  CoefCal_t width;
  CoefCal_t tl;
  CoefCal_t tr;
  CoefCal_t sh;
  CoefCal_t sw;
  ConsCal_t vc;
  ConsCal_t start;
  EfcCal_t efc;
  StrCal_t unit;
} CalRec;

#define CALNOERR	 0 /* vsCal: no error */
#define CALPOINT	 1 /* vsCal: illegal NULL-pointer */
#define CALOPEN		 2 /* vsCal: could not open protocolfile */
#define CALNOFIL	 3 /* vsCal: illegal NULL-filepointer */
#define CAL0READ	 4 /* vsCal: no values read from file */
#define CALMCOEF	 5 /* vsCal: to many coefficients found */
#define CALMCONS	 6 /* vsCal: to many constants found */
#define CALMEFC		 7 /* vsCal: to many efficiency points found */
#define CALMSTR		 8 /* vsCal: to many strings found */
#define CALSTRL		 9 /* vsCal: internal string dimension exceeded */
#define CALNODEF	10 /* vsCal: no legal calibration defined */
#define CALWRITE	11 /* vsCal: could not write calibrationfile */
#define CALRANGE	12 /* vsCal: position out of efc-calibration range */
#define CALNEGEFC	13 /* vsCal: negative value for efc-calibration */
#define CALNOKEY	14 /* vsCal: keyword not found */
#define CALEQXVAL	15 /* vsCal: pairs with equal x-values for coef.calc */
#define CALTYPE		16 /* vsCal: operation for cal. type not implemented */
#define CALCLUST	17 /* vsCal: no clusterfile found */
#define CALALLOC	18 /* vsCal: memory allocation failed */
#define CALEFMT	19 /* vsCal: wrong efc format */
#define CALUPAR	20 /* vsCal: unexpected parameters found in file */

#ifdef _vsCal_c
int CalErrNo= CALNOERR;
#else
extern int CalErrNo;
#endif

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

#include <tv/vsCalF.h>

#undef P_

#endif				/* _vsCal_h */
/* ------------------------------------------------------------------------- */
