/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsCurFit.h,v 1.2 1994/05/26 19:50:06 vs Exp vs $ */
#ifndef _vsCurFit_h
#define _vsCurFit_h

#include <tv/vsConfig.h>
#include <tv/vsTypes.h>
#include <tv/vsPar.h>

#define CUR_MAXORDER	128	/* maximum of free parameters in fit */

				/* standard values for constants */

#define CUR_RATIO	(1.e-12)	/* minimal change in measure */
#define CUR_LAMBDA	(1.e-3)	/* lower limit for lambda */
#define CUR_MINLAMBDA	(1.e-20)/* lower limit for lambda */
#define CUR_MAXLAMBDA	(1.e30)	/* upper limit for lambda */
#define CUR_INCLAMBDA	10	/* while measure is getting worse,increment*/
				/* the control parameter times this faktor   */
#define CUR_DECLAMBDA	0.1	/* while its getting better decrement it.    */
#define CUR_ITERATIONS	50	/* maximum number of iterations */
#define CUR_ABSOLUTDX	(1.e-9)	/* parameter==0 change in numeric derivation */
#define CUR_RELATIVDX	(1.e-5)	/* parameter change in numeric derivation */

#define CUR_CHI		0
#define CUR_POI		1

typedef struct {		/* curfit datapoint struct */
  double x;			/* x-value */
  double y;			/* y-value */
  double o;			/* offset for calculation of weight */
  double w;			/* weight in chisqr */
} CurPnt_t;

#define CurPntX(p)		((p)->x)
#define CurPntY(p)		((p)->y)
#define CurPntO(p)		((p)->o)
#define CurPntW(p)		((p)->w)

typedef struct {		/* curfit datavector struct */
  VecCorePart	core;
  CurPnt_t 	*p;		/* vector */
} CurPntVecRec, *CurPntVec_t;

				/* access macros for CurDat_t */
#define CurPntVecX(cdp,i)	CurPntX(VecP(cdp)+i)
#define CurPntVecY(cdp,i)	CurPntY(VecP(cdp)+i)
#define CurPntVecO(cdp,i)	CurPntO(VecP(cdp)+i)
#define CurPntVecW(cdp,i)	CurPntW(VecP(cdp)+i)

typedef struct _CurRec {	/* curfit parameter struct */
  Par_t par;
  ParRec try;
  DVecRec covar;
  DVecRec grad;
  int freedom;
  double measure;
  double dMeasure;
  double measureRatio;
  double lambda;
  double minLambda;
  double maxLambda;
  double decLambda;
  double incLambda;
  int iterations;
  int maxIterations;
  CurPntVecRec data;
  int measureTyp;
  pfi_t functionSetup;
  pfi_t functionFinish;
  pfd_t function;
  pfi_t derivationSetup;
  pfi_t derivationFinish;
  pfi_t derivation;
  pfi_t protocol;
  pfi_t changeTry;
} CurRec;

				/* access macros for Cur_t */
#define CurPar(cp)		((cp)->par)
#define CurTry(cp)		(&(cp)->try)
#define CurFreeParNum(cp)	ParNumFree(CurPar(cp))
#define CurCovar(cp)		(&(cp)->covar)
#define CurGrad(cp)		(&(cp)->grad)
#define CurFreedom(cp)		((cp)->freedom)
#define CurMeasure(cp)		((cp)->measure)
#define CurDMeasure(cp)		((cp)->dMeasure)
#define CurMeasureRatio(cp)	((cp)->measureRatio)
#define CurLambda(cp)		((cp)->lambda)
#define CurMinLambda(cp)	((cp)->minLambda)
#define CurMaxLambda(cp)	((cp)->maxLambda)
#define CurDecLambda(cp)	((cp)->decLambda)
#define CurIncLambda(cp)	((cp)->incLambda)
#define CurIterations(cp)	((cp)->iterations)
#define CurMaxIterations(cp)	((cp)->maxIterations)
#define CurData(cp)		(&(cp)->data)
#define CurDataP(cp)		VecP(CurData(cp))
#define CurDataNum(cp)		VecNum(CurData(cp))
#define CurMeasureTyp(cp)	((cp)->measureTyp)
#define CurFunction(cp)		((cp)->function)
#define CurFunctionSetup(cp)	((cp)->functionSetup)
#define CurFunctionFinish(cp)	((cp)->functionFinish)
#define CurDerivation(cp)	((cp)->derivation)
#define CurDerivationSetup(cp)	((cp)->derivationSetup)
#define CurDerivationFinish(cp)	((cp)->derivationFinish)
#define CurFunction(cp)		((cp)->function)
#define CurProtocol(cp)		((cp)->protocol)
#define CurChangeTry(cp)	((cp)->changeTry)

#define CURNOERR	 0 /* vscurfit: no error */
#define CURPOINT	 1 /* vscurfit: illegal NULL-pointer */
#define CUR0FREE	 2 /* vscurfit: no free parameter */
#define CUR0DATA	 3 /* vscurfit: no data points to fit */
#define CURLDATA	 4 /* vscurfit: no degree of freedom */
#define CURMPAR		 5 /* vscurfit: too much degrees of freedom */
#define CURFCPNT	 6 /* vscurfit: illegal function NULL-pointer */
#define CURSING	 	 7 /* vscurfit: fit matrix has become singular */
#define CURSING2	 8 /* vscurfit: fit matrix has become singular-2*/
#define CURALLOC	 9 /* vscurfit: memory allocation failed */
#define CURINDEX	10 /* "vsCurFit: can not get free parameters" */

#ifdef _vsCurFit_c
int CurErrNo=CURNOERR;
#else
extern int CurErrNo;
#endif

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

#include <tv/vsCurFitF.h>

#undef P_

#endif				/* _vsCurFit_h */
/* -------------------------------------------------------------------------
 *                   end of vscurfit.h
 * -------------------------------------------------------------------------*/
