/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsCut.c,v 1.2 1994/05/26 19:50:07 vs Exp vs $ */
#define _vsCut_c

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include "vsTypes.h"
#include "vsAlloc.h"

#include "vsMark.h"
#include "vsPath.h"
#include "vsSpectra.h"
#include "vsTools.h"

#include "vsCut.h"
#include "vsCutP.h"

/* ------------------------------------------------------------------------- */
static char *gateTitle= "gate-marker";
static char *bgGateTitle= "bg-gate-marker";

#define CUTMAXBUFLEN		1024

/* ------------------------------------------------------------------------- */
char *CutError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* CUTNOERR:   */ "vsCut: no error",
  /* CUTPOINT:   */ "vsCut: illegal NULL-pointer",
  /* CUTALLOC:   */ "vsCut: can not allocate temporary arrays",
  /* CUTNOMRK:   */ "vsCut: not enough marker for normalization",
  /* CUTSPCBUF:  */ "vsCut: no spectrum buffer defined",
  /* CUTNOMTX:   */ "vsCut: no matrix file open",
  /* CUTNOGATE:  */ "vsCut: no gate marker defined",
  /* CUTCRESPC:  */ "vsCut: no gate spectra created",
  /* CUT0BG:     */ "vsCut: bg peak integral == 0.0",
  /* CUTWEIGHT:  */ "vsCut: illegal weight type for cut",
  /* CUTINSRT:   */ "vsCut: can not insert new CDIR-entry",
  /* CUTNOENTRY: */ "vsCut: no further CDIR-entry found",
  /* CUTPROJ:    */ "vsCut: can not create gate spectrum",
  /* CUTOPGATE:  */ "vsCut: can not create directory for gate file",
  /* CUTRDGATE:  */ "vsCut: can not read gate file",
  /* CUTWRGATE:  */ "vsCut: can not write gate file",
  /* CUTOPSPC:   */ "vsCut: can not create directory for spectra file",
  /* CUTRDSPC:   */ "vsCut: can not read cut spectra",
  /* CUTWRSPC:   */ "vsCut: can not write cut spectra",
  /* CUTILLFAC:  */ "vsCut: factors not correct defined",
  /* default:    */ "vsCut: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ------------------------------------------------------------------------- */
int CutCloseCDir(cdp)
     CDIR *cdp;
{
  if (cdp) {
    VecFree(CDirE(cdp));
    Free(CDirPath(cdp));
    Free(CDirGateExpr(cdp));
    Free(cdp); }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int CutCopy(t, f)
     Cut_t t;
     Cut_t f;
{
  if (!t || !f) { CutsetErrNo(CUTPOINT); return FALSE; }
  memcpy(CutB(t), CutB(f), sizeof(CutB_t));
  if (VecCopy(CutGate(t), CutGate(f)) &&
      VecCopy(CutBgGate(t), CutBgGate(f))) {
    CutFile(t)= Strdup(CutFile(f));
    return TRUE; }
  else { CutReset(t); return FALSE; }
}
/* ------------------------------------------------------------------------- */
static int CutCreateSpectrumUpdate(sfp, cp, gate, sp)
     SpcFile_t sfp;
     Cut_t cp;
     GVec_t gate;
     Spc_t sp;
{
  int i;

  CutTime(cp)= time(NULL);

  for (i= 0; i< VecNum(gate); i++) {
    SpcClear(sp);
    SpcBufNum(SpcSpc(sp))= SpcBufNum(SpcDSpc(sp))= SpcFileColumns(sfp);
    if (!SpcProject(sfp, GVecX1(gate,i), GVecX2(gate,i), sp)) {
      CutsetErrNo(CUTPROJ); return FALSE; }
    else if (!SpcMultiply(sp, GVecFac(gate,i)) ||
	     !SpcAdd(CutSpc(cp), sp, NULL, NULL)) {
      CutsetErrNo(CUTCRESPC); return FALSE; } }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
static int CheckFac(gt)
     GVec_t gt;
{
  int g;

  for (g= 0; g< VecNum(gt); g++) { if (GVecFac(gt,g) != 0.0) { return TRUE; } }
  return FALSE;
}
/* ------------------------------------------------------------------------- */
int CutCreateSpectrum(sfp, cp)
     SpcFile_t sfp;
     Cut_t cp;
{
  if (!sfp || !cp) { CutsetErrNo(CUTPOINT); return FALSE; }
  else if (!SpcFileMfp(sfp)) { CutsetErrNo(CUTNOMTX); return FALSE; }
  else if (!CutSpc(cp)) { CutsetErrNo(CUTSPCBUF); return FALSE; }
  else if (!VecNum(CutGate(cp))) { CutsetErrNo(CUTNOGATE); return FALSE; }
  else if (!CheckFac(CutGate(cp)) && !CheckFac(CutBgGate(cp))) {
    CutsetErrNo(CUTILLFAC); return FALSE; }
  else {
    /*SpcReal_t StackAlloc(SpcReal_t,spcBuf,2 * SpcMaxRes, 2 * SPC_MR); DUKE */
    SpcReal_t *spcBuf = (SpcReal_t*)malloc(2 * SpcMaxRes * sizeof (SpcReal_t));
    SpcRec spc;

    if (!spcBuf) { CutsetErrNo(CUTALLOC); return FALSE; }
    else {
      SpcSet(&spc, SPC_REAL, sizeof(SpcReal_t), (char *)spcBuf, SpcMaxRes);

      SpcReset(CutSpc(cp));
      {
	int ret = (CutCreateSpectrumUpdate(sfp, cp, CutGate(cp), &spc) &&
	  CutCreateSpectrumUpdate(sfp, cp, CutBgGate(cp), &spc));
      Free (spcBuf);		/*   CutCreateSpectrumUpdate() uses spcBuf !!! */
    return ret;
    }}}
}
/* ------------------------------------------------------------------------- */
int CutDeleteCDirEntry(cdp, idx)
     CDIR *cdp;
     int idx;
{
  if (cdp) {
    VecDelete(CDirE(cdp), idx, 1);
    return TRUE; }
  else { CutsetErrNo(CUTPOINT);  return FALSE; }
}
/* ------------------------------------------------------------------------- */
int CutCDirEntry(cdp, idx1, idx2, file)
     CDIR *cdp;
     int idx1;
     int idx2;
     char *file;
{
  if (!file || (!cdp && idx2 >= 0)) { CutsetErrNo(CUTPOINT); return FALSE; }
  else if (idx2 >= 0) { sprintf(file, "%s/%d_%d", CDirPath(cdp), idx1, idx2); }
  else if (cdp) { sprintf(file, "%s/%d_?", CDirPath(cdp), idx1); }
  else { sprintf(file, "%d_?", idx1); }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
static int CutGateFile(cdp, idx1, idx2, file)
     CDIR *cdp;
     int idx1;
     int idx2;
     char *file;
{
  CutCDirEntry(cdp, idx1, idx2, file);
  strcpy(file, PthSubstitute(file, CDirGateExpr(cdp)));
  return (*file != '\0');
}
/* ------------------------------------------------------------------------- */
int CutInit(cp)
     Cut_t cp;
{
  if (cp) {
    CutSpc(cp)= NULL;
    VecInit(CutGate(cp), sizeof(Gate_t));
    VecInit(CutBgGate(cp), sizeof(Gate_t));
    CutFile(cp)= NULL;
    return CutReset(cp); }
  else { CutsetErrNo(CUTPOINT); return FALSE; }
}
/* ------------------------------------------------------------------------- */
static int CutInsertCDirEntry(cdp, idx1, idx2, chn, tim)
     CDIR *cdp;
     int idx1;
     int idx2;
     Gate_t *chn;
     time_t tim;
{
  int i;
  for (i= 0; i< CDirNum(cdp); i++) {
    if (CDirGateX1(cdp,i) < GateX1(chn)) { break; }
    else if (CDirGateX1(cdp,i) == GateX1(chn) &&
	     CDirGateX2(cdp,i) < GateX2(chn)) { break; } }
  if (!VecInsert(CDirE(cdp), i, 1)) { CutsetErrNo(CUTINSRT); return -1; }
  CDirIdx1(cdp,i)= idx1;
  CDirIdx2(cdp,i)= idx2;
  CDirTime(cdp,i)= tim;
  memcpy(CDirGate(cdp,i), chn, sizeof(Pair_t));
  return i;
}
/* ------------------------------------------------------------------------- */
static int CutLegalCDirEntry(name)
     char *name;
{
  if ('0' <= *name && *name++ <= '9') {
    while ('0' <= *name && *name <= '9') { name++; }
    if (*name++ == '_' &&
	'0' <= *name && *name++ <= '9') {
      while ('0' <= *name && *name <= '9') { name++; }
      if (!*name) { return TRUE; } } }
  return FALSE;
}
/* ------------------------------------------------------------------------- */
CDIR *CutOpenCDir(path, gateExpr)
     char *path;
     char *gateExpr;
{
  CDIR *cdp;
  GVecRec gt;

  VecInit(&gt, sizeof(Gate_t));

  if ((cdp= (CDIR *)Malloc(sizeof(CDIR)))) {
    DIR *dp;

    VecInit(CDirE(cdp), sizeof(CDirEntry_t));

    if (!(CDirPath(cdp)= Strdup(path)) ||
	!(CDirGateExpr(cdp)= Strdup(gateExpr))) {
      CutCloseCDir(cdp); VecFree(&gt); return NULL; }

    if ((dp= opendir(path)) != NULL) {
      struct dirent *de;

      rewinddir(dp);
      while ((de= readdir(dp))) {
	int idx1;
	int idx2;
	char *name= de->d_name;
	if (CutLegalCDirEntry(name) &&
	    sscanf(name, "%d_%d", &idx1, &idx2) == 2) {
	  char file[MAXPATHLEN];
	  struct stat buf;

	  CutGateFile(cdp, idx1, idx2, file);
	  if (stat(file, &buf) == 0 && CutRead1stGate(file, &gt)) {
	    if (CutInsertCDirEntry(cdp, idx1, idx2,
				   VecP(&gt), buf.st_mtime) < 0) {
	      CutCloseCDir(cdp); VecFree(&gt); return NULL; } } } }
      closedir(dp); } }
  VecFree(&gt);
  return cdp;
}
/* ------------------------------------------------------------------------- */
int CutPosNextCDirEntry(cdp, idx, chn)
     CDIR *cdp;
     int idx;
     double chn;
{
  if (cdp == NULL) { CutsetErrNo(CUTPOINT); return -1; }
  else {
    int i;

    for (i= MAX(0, idx+1); i< CDirNum(cdp); i++) {
      if (CDirGateX1(cdp,i) <= chn && chn <= CDirGateX2(cdp,i)) { return i; } }
    CutsetErrNo(CUTNOENTRY); return -1; }
}
/* ------------------------------------------------------------------------- */
int CutPosPreviousCDirEntry(cdp, idx, chn)
     CDIR *cdp;
     int idx;
     double chn;
{
  if (cdp == NULL) {
    CutsetErrNo(CUTPOINT); return -1; }
  else {
    int i;

    for (i= MIN(CDirNum(cdp)-1, idx-1); i>= 0; i--) {
      if (CDirGateX1(cdp,i) <= chn && chn <= CDirGateX2(cdp,i)) {
	return i; } }
    CutsetErrNo(CUTNOENTRY); return -1; }
}
/* ------------------------------------------------------------------------- */
int CutNameCDirEntry(cdp, name)
     CDIR *cdp;
     char *name;
{
  if (cdp == NULL) {  CutsetErrNo(CUTPOINT); return -1; }
  else {
    int i;
    char entry[MAXPATHLEN];
    for (i= 0; i< CDirNum(cdp); i++) {
      sprintf(entry, "%d_%d", CDirIdx1(cdp,i), CDirIdx2(cdp,i));
      if (!strcmp(name, entry)) { return i; } }
    return -1; }
}
/* ------------------------------------------------------------------------- */
int CutReadGatesCDir(cdp, idx, cp)
     CDIR *cdp;
     int idx;
     Cut_t cp;
{
  /*  printf("CutReadGatesCDir\n");*/
  if (!cdp || !cp) {
    CutsetErrNo(CUTPOINT);
    return FALSE;
  }
  else {
    char file[MAXPATHLEN];

    if (!CutCDirEntry(cdp, CDirIdx1(cdp,idx), CDirIdx2(cdp,idx), file)) {
      return FALSE; }
    else {
      return CutReadGates(cp, file, CDirGateExpr(cdp));
    }
  }
}
/* ------------------------------------------------------------------------- */
static int CutPrintGate(fp, pp, idx)
     FILE *fp;
     Gate_t *pp;
     int idx;
{
  return (fprintf(fp, "%f %f %f\n", GateX1(pp + idx), GateX2(pp + idx),
		  GateFac(pp + idx)) != EOF);
}
/* ------------------------------------------------------------------------- */
int CutScanGate(buffer, pp, idx)
     char *buffer;
     Gate_t *pp;
     int idx;
{
  int num=sscanf(buffer, "%lf%lf%lf",
		 &GateX1(pp + idx), &GateX2(pp + idx), &GateFac(pp + idx));
  /*  printf("CutScanGate\n");*/
  if (num < 2) {
    return FALSE;
  }
  else if (num == 2) {
    GateFac(pp +idx)= 0.0;
  }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int CutReadGates(cp, entryName, gateExpr)
     Cut_t cp;
     char *entryName;
     char *gateExpr;
{
  /*  printf("CutReadGates\n");*/
  if (!entryName || !gateExpr || !cp) {
    CutsetErrNo(CUTPOINT);
    return FALSE;
  }
  else {
    FILE *fp;
    char *file= PthSubstitute(entryName, gateExpr);

    CutReset(cp);
    if ((fp= fopen(file, "r")) == NULL) {
      CutsetErrNo(CUTOPGATE);
      return FALSE;
    }
    else {
      int ret;

      ret= VecFRead(fp, gateTitle, CutGate(cp), CutScanGate);
      ret= ret && VecFRead(fp, bgGateTitle, CutBgGate(cp), CutScanGate);

      if ((ret= ret && (CutFile(cp)= Strdup(entryName)))) {
	struct stat buf;
	stat(file, &buf);
	CutTime(cp)= buf.st_mtime;
	CutExists(cp)= TRUE;
      }
      else {
	CutReset(cp);
	CutsetErrNo(CUTRDGATE);
      }
      fclose(fp);
      return ret;
    }
  }
}
/* ------------------------------------------------------------------------- */
static int CutRead1stGate(file, gat)
     char *file;
     GVec_t gat;
{
  FILE *fp;
  int status;

  /*  printf("CutRead1stGate\n");*/

  if ((fp= fopen(file, "r")) == NULL) {
    CutsetErrNo(CUTOPGATE);
    return FALSE;
  }
  status= VecFRead(fp, gateTitle, gat, CutScanGate);
  fclose(fp);

  if (!status) {
    CutsetErrNo(CUTRDGATE);
    return FALSE;
  }
  else if (!VecNum(gat)) {
    CutsetErrNo(CUTNOGATE);
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ------------------------------------------------------------------------- */
int CutReset(cp)
     Cut_t cp;
{
  if (cp) {
    CutWeight(cp)= CUT_WGAT;
    VecReset(CutGate(cp));
    VecReset(CutBgGate(cp));
    CutIdx2(cp)= -1;
    CutExists(cp)= FALSE;
    Free(CutFile(cp));
    return TRUE; }
  else { CutsetErrNo(CUTPOINT); return FALSE; }
}
/* ------------------------------------------------------------------------- */
static int CutResetSpectrum(cp)
     Cut_t cp;
{
  if (CutSpc(cp)) { SpcReset(CutSpc(cp)); }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int CutWriteGates(cdp, idx, cp)
     CDIR *cdp;
     int *idx;
     Cut_t cp;
{
  if (!cdp || !cp) { CutsetErrNo(CUTPOINT); return FALSE; }
  else if (!VecNum(CutGate(cp))) {
    CutsetErrNo(CUTNOGATE); return FALSE; }
  else {
    char file[MAXPATHLEN];
    FILE *fp;
    int idx1= CutIdx1(cp);
    int idx2= 0;

    { int i;
      int restart= FALSE;
      for (i= 0; i< CDirNum(cdp); i++) {
	if (CDirIdx1(cdp,i) == idx1) {
	  if (CDirIdx2(cdp,i) == idx2) {
	    idx2++; if (restart) { i= 0; restart= FALSE; } }
	  else if (CDirIdx2(cdp,i) > idx2) { restart= TRUE; } } } }

    if (!CutGateFile(cdp, idx1, idx2, file)) { return FALSE; }
    else if (!PthCreateRecursive(PthHead(file))) {
      CutsetErrNo(CUTOPGATE); return FALSE; }

    if ((fp= fopen(file, "w"))) {
      int ret;

      ret= VecFWrite(fp, gateTitle, CutGate(cp), CutPrintGate);
      ret= ret && VecFWrite(fp, bgGateTitle, CutBgGate(cp), CutPrintGate);
      ret= (fclose(fp) != EOF) && ret;
      if (ret) {
	struct stat buf; stat(file, &buf); CutTime(cp)= buf.st_mtime;
	*idx= CutInsertCDirEntry(cdp, idx1, idx2, VecP(CutGate(cp)),
				 CutTime(cp));
	CutIdx2(cp)= idx2;
	CutCDirEntry(cdp, idx1, idx2, file);
	Free(CutFile(cp));
	CutFile(cp)= Strdup(file);
	return (*idx >= 0); }
      else { CutsetErrNo(CUTWRGATE); return FALSE; } }
    else { CutsetErrNo(CUTOPGATE); return FALSE; } }
}
/* ------------------------------------------------------------------------- */
int CutReadSpectrum(cp, file, format)
     Cut_t cp;
     char *file;
     char *format;
{
  if (!cp || !file || !format) { CutsetErrNo(CUTPOINT); return FALSE; }
  else if (!CutSpc(cp)) {
    CutsetErrNo(CUTSPCBUF); return FALSE; }
  else if (!VecNum(CutGate(cp)) || !VecNum(CutGate(cp))) {
    CutsetErrNo(CUTNOGATE); return FALSE; }
  else if (!SpcRead(CutSpc(cp), file, format)) {
    CutResetSpectrum(cp);  CutsetErrNo(CUTRDSPC); return FALSE; }
  else { return TRUE; }
}
/* ------------------------------------------------------------------------- */
int CutWriteSpectrum(cp, file, format)
     Cut_t cp;
     char *file;
     char *format;
{
  if (!cp || !file) {
    CutsetErrNo(CUTPOINT);
    return FALSE;
  }
  else if (!VecNum(CutGate(cp))) {
    CutsetErrNo(CUTNOGATE);
    return FALSE;
  }
  else if (!CutSpc(cp)) {
    CutsetErrNo(CUTSPCBUF);
    return FALSE;
  }
  else if (!PthCreateRecursive(PthHead(file))) {
    CutsetErrNo(CUTOPSPC);
    return FALSE;
  }
  else if (!SpcWrite(CutSpc(cp), file, format)) {
    CutsetErrNo(CUTWRSPC);
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ------------------------------------------------------------------------- */
