/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsFit.c,v 1.5 1994/12/08 20:44:14 vs Exp vs $ */
#define _vsFit_c

#include "vsTypes.h"
#include "vsAlloc.h"

#include "vsPar.h"
#include "vsTools.h"

#include "vsFit.h"
#include "vsFitP.h"
/* ------------------------------------------------------------------------- */
static FitFctRec fitFctCA= {
  CAFunctionSetup,
  CAFunctionFinish,
  (FitOpt(*)())CAPeakFunctionSetup,
  (FitOpt(*)())CAPeakFunctionFinish,
  CADerivationSetup,
  CADerivationFinish,
  CADerivation,
  CAFunction,
  CANormVolume,
  CAStepFunction,
  CAPeakFunction,
  CAUnitPeakFunction,
  (FitOpt(*)())CAPeakFunctionSetup, /* unitPeakDerivationSetup */
  (FitOpt(*)())CAPeakFunctionFinish, /* unitPeakDerivationFinish */
  CAUnitPeakDerivation, };

FitFct_t FitFctCA= &fitFctCA;

static FitFctRec fitFctAE= {
  AEFunctionSetup,
  AEFunctionFinish,
  (FitOpt(*)())AEPeakFunctionSetup,
  (FitOpt(*)())AEPeakFunctionFinish,
  AEDerivationSetup,
  AEDerivationFinish,
  AEDerivation,
  AEFunction,
  AENormVolume,
  AEStepFunction,
  AEPeakFunction,
  AEUnitPeakFunction,
  (FitOpt(*)())AEPeakFunctionSetup, /* unitPeakDerivationSetup */
  (FitOpt(*)())AEPeakFunctionFinish, /* unitPeakDerivationFinish */
  AEUnitPeakDerivation, };

FitFct_t FitFctAE= &fitFctAE;

static double invff[]= {
  0.7978845608029,
  1.0,
  0.7978845608029,
  0.5,
  0.1329807601338,
  0.04166666666667,
  0.006649038006691,
  0.001388888888889,
  0.0001583104287307,
  2.480158730159e-05,
  2.198755954593e-06,
  2.755731922399e-07,
  1.99886904963e-08,
  2.087675698787e-09,
  1.281326313866e-10,
  1.147074559773e-11, };

/* ------------------------------------------------------------------------- */

char *FitError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* FITNOERR:  */ "vsFit: no error",
  /* FITPOINT:  */ "vsFit: illegal NULL-pointer",
  /* FITSIGM0:  */ "vsFit: illegal peak width == 0.0",
  /* FITPEAK:   */ "vsFit: maximal peak number exceeded",
  /* FITPARM:   */ "vsFit: maximal parameter number exceeded",
  /* FITILLST:  */ "vsFit: illegal sort type",
  /* FIT0AMPL:  */ "vsFit: no peak with non-zero amplitude",
  /* FITALLOC:  */ "vsFit: can not allocate temporary arrays",
  /* FITPKTYP:  */ "vsFit: illegal peak function typ",
  /* FITBGTYP:  */ "vsFit: illegal background function typ",
  /* FITSW0:    */ "vsFit: illegal step width == 0.0",
  /* default:   */ "vsFit: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ------------------------------------------------------------------------- */

/* ==================================================================== */
/* reset Fit_t, FitInt_t, FitStat_t bg, par(idx), peakVolumes and peaks */
/* ==================================================================== */

int FitReset(fp)
     Fit_t fp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    FitFitTyp(fp)	= FIT_FTUNDEF;

    FitRes(fp)		= 0;

    FitStatReset(FitStatBg(fp));
    FitStatReset(FitStatReg(fp));

    FitBgOffset(fp)	= 0;

    FitWCorr(fp)	= -1;
    FitPCorr(fp)	= -1;
    FitTlCorr(fp)	= -1;
    FitTrCorr(fp)	= -1;
    FitShCorr(fp)	= -1;
    FitSwCorr(fp)	= -1;

    FitPeakNum(fp)	= 0;
    FitPeakSortTyp(fp)	= FIT_STPOS;

    FitIntReset(FitIntegral(fp));
    FitIntReset(FitBgIntegral(fp));
    FitIntReset(FitSubIntegral(fp));

    ParReset(FitPar(fp));
    VecReset(FitRegMrk(fp));
    VecReset(FitBgMrk(fp));
    VecReset(FitBgCovar(fp));
    VecReset(FitPeakSortIdx(fp));
    VecReset(FitPeakSortRevIdx(fp));

    return (FitResetBg(fp) &&	/* reserves space */
	    FitDisablePars(fp, FIT_UTSH) &&
	    FitDisablePars(fp, FIT_UTTL) &&
	    FitDisablePars(fp, FIT_UTTR)); }
}
/* ------------------------------------------------------------------------- */
int FitStatReset(fp)
     FitStat_t *fp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    FitStatTime(fp)		= 0;
    FitStatMeasure(fp)		= 0.0;
    FitStatDMeasure(fp)		= 0.0;
    FitStatRms(fp)		= 0.0;
    FitStatMaxIterations(fp)	= 0;
    FitStatIterations(fp)	= 0;
    FitStatMeasureTyp(fp)	= FIT_MDY;
    FitStatFreedom(fp)		= 0;
    FitStatLambda(fp)		= 0.0;
    FitStatMinLambda(fp)	= 0.0;
    FitStatMaxLambda(fp)	= 0.0;
    FitStatStartLambda(fp)	= 0.0;
    return TRUE; }
}
/* ------------------------------------------------------------------------- */
int FitIntReset(fip)
     FitInt_t *fip;
{
  if (fip == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    FitIntI(fip)	= 0.0;
    FitIntDI(fip)	= 0.0;
    FitIntM1(fip)	= 0.0;
    FitIntDM1(fip)	= 0.0;
    FitIntM2(fip)	= 0.0;
    FitIntDM2(fip)	= 0.0;
    FitIntM3(fip)	= 0.0;
    FitIntDM3(fip)	= 0.0;
    return TRUE; }
}
/*---------------------------------------------------------------------------*/
int FitResetBg(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);

  FitTime(fp)	= 0;
  FitBgTime(fp)	= 0;
  { int i;			/* reset bg polynom */
    ParV_t pv;

    ParVInit(&pv);
    ParVUserStat(&pv)	= FIT_USUNDEF;
    ParVMode(&pv)	= PAR_MFREE;
    ParVStat(&pv)	= PAR_SUSED;
    ParVUserTyp(&pv)	= FIT_UTBGPOL;
    for (i=0; i< FIT_BGGRD; i++) { /* background */
      if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; } }
    for (i= FIT_BGGRD; i< FIT_1STPEAK; ) {
      ParVUserTyp(&pv)	= FIT_UTBGFAC;
      if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; }
      i++; 
      ParVUserTyp(&pv)	= FIT_UTBGEXP;
      if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; }
      i++; } }
  FitSetupBg(fp);
  FitResetPeakVolumes(fp);
  return TRUE;
}
/*---------------------------------------------------------------------------*/
int FitResetPeakVolumes(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int i;

  for (i=0; i< FitPeakNum(fp); i++) {
    ParV_t pv;
    int j= FitVolIdx(i);

    ParGet(pp, j, &pv);
    FitTime(fp)		= 0;
    ParVUserStat(&pv)	= FIT_USUNDEF;
    ParSet(pp, j, &pv, FALSE); }
  return TRUE;
}
/*---------------------------------------------------------------------------*/
static int FitDisablePars(fp, userTyp)
     Fit_t fp;
     int userTyp;
{
  int i;
  for (i=0; i< FitPeakNum(fp); i++) { FitDisablePar(fp, i, userTyp, FALSE); }
  return TRUE;
}
/*---------------------------------------------------------------------------*/
int FitDisablePar(fp, idx, userTyp, check_corr)
     Fit_t fp;
     int idx;
     int userTyp;
     int check_corr;
{
  Par_t pp= FitPar(fp);
  ParV_t pv;
  int j, k= -1;

  switch (userTyp) {
  case FIT_UTTL: j= FitTlIdx(idx); break;
  case FIT_UTTR: j= FitTrIdx(idx); break;
  case FIT_UTSH:
  case FIT_UTSW:
    j= FitShIdx(idx);
    k= FitSwIdx(idx);
    break;
  default: return TRUE; }

  ParGet(pp, j, &pv);
  if (ParVUserStat(&pv) != FIT_USNONE) {
    FitTime(fp) = 0;
    ParVUserStat(&pv) = FIT_USNONE;
    ParVStat(&pv) = PAR_SSUSP;
    ParVVal(&pv) = 0.0;
    ParSet(pp, j, &pv, check_corr); }

  if (k >= 0) {
    ParGet(pp, k, &pv);
    if (ParVUserStat(&pv) != FIT_USNONE) {
      FitTime(fp) = 0;
      ParVUserStat(&pv) = FIT_USNONE;
      ParVStat(&pv) = PAR_SSUSP;
      ParVVal(&pv) = 0.0;
      ParSet(pp, k, &pv, check_corr); } }
  return TRUE;
}
/*---------------------------------------------------------------------------*/
int FitEnablePar(fp, idx, userTyp, check_corr)
     Fit_t fp;
     int idx;
     int userTyp;
     int check_corr;
{
  Par_t pp= FitPar(fp);
  ParV_t pv;
  int j, k= -1;

  switch (userTyp) {
  case FIT_UTTL: j= FitTlIdx(idx); break;
  case FIT_UTTR: j= FitTrIdx(idx); break;
  case FIT_UTSH:
  case FIT_UTSW:
    j= FitShIdx(idx);
    k= FitSwIdx(idx);
    break;
  default: return TRUE; }

  ParGet(pp, j, &pv);
  if (ParVUserStat(&pv) == FIT_USNONE) {
    FitTime(fp) = 0;
    ParVUserStat(&pv) = FIT_USUNDEF;
    ParVStat(&pv) = PAR_SUSED;
    ParSet(pp, j, &pv, check_corr); }

  if (k >= 0) {
    ParGet(pp, k, &pv);
    if (ParVUserStat(&pv) == FIT_USNONE) {
      FitTime(fp) = 0;
      ParVUserStat(&pv) = FIT_USUNDEF;
      ParVStat(&pv) = PAR_SUSED;
      if (ParVUserTyp(&pv) == FIT_UTSW) { /* do not free step width */
	ParVVal(&pv) = 1.0;
	ParVUserStat(&pv) = FIT_USUSER;
	ParVMode(&pv)= PAR_MHOLD; }
      ParSet(pp, k, &pv, check_corr); } }
  return TRUE;
}
/*---------------------------------------------------------------------------*/
int FitResetPeaks(fp)
     Fit_t fp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    if (FitPeakNum(fp) > 0) {
      Par_t pp= FitPar(fp);

      { int i;
	for (i= FIT_1STPEAK; i < ParNum(pp); i++) {
	  ParStat(pp,i) = PAR_SDELD; } }
      ParCompress(pp);
      FitPeakNum(fp) = 0;
      FitTime(fp) = 0; }
    return TRUE; }
}
/* ------------------------------------------------------------------------- */

/* ================== */
/* suspend parameters */
/* ================== */

int FitSuspendBg(fp)
     Fit_t fp;
{
  if (FitBgTime(fp)) {
    Par_t pp= FitPar(fp);
    int i;
    for (i= 0; i< FitBgGrad(fp); i++) {
      if (ParMfree(pp,i)) { FitBgTime(fp)= 0; return TRUE; } }
    for (i= FIT_BGGRD; i< 2 * FitBgExpNum(fp); i++) {
      if (ParMfree(pp,i)) { FitBgTime(fp)= 0; return TRUE; } }
#if 0
    for (i= 0; i< FIT_1STPEAK; i++) {
      ParV_t pv;
      ParGet(pp, i, &pv); ParVStat(&pv) = PAR_SSUSP;
      ParSet(pp, i, &pv, FALSE); }
#endif
  }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int FitSuspendPeaks(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int i;

  for (i= FIT_1STPEAK; i< ParNum(pp); i++) {
      ParV_t pv;
      ParGet(pp, i, &pv); ParVStat(&pv) = PAR_SSUSP;
      ParSet(pp, i, &pv, FALSE); }
  return TRUE;
}
/* ------------------------------------------------------------------------- */

/* ============== */
/* use parameters */
/* ============== */

int FitUseBg(fp)
     Fit_t fp;
{
#if 0
  Par_t pp= FitPar(fp);
  int i;

  for (i= 0; i< FIT_1STPEAK; i++) {
    if (ParUserStat(pp,i) != FIT_USNONE) {
      ParV_t pv;
      ParGet(pp, i, &pv); ParVStat(&pv) = PAR_SUSED;
      ParSet(pp, i, &pv, FALSE); } }
#endif				/* 0 */
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int FitUsePeaks(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int i;

  for (i= FIT_1STPEAK; i< ParNum(pp); i++) {
    if (ParUserStat(pp,i) != FIT_USNONE) {
      ParV_t pv;
      ParGet(pp, i, &pv); ParVStat(&pv) = PAR_SUSED;
      ParSet(pp, i, &pv, FALSE); } }
  return TRUE;
}
/* ------------------------------------------------------------------------- */

/* ======================= */
/* init Fit_t and FitInt_t */
/* ======================= */

int FitInit(fp)
     Fit_t fp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    ParInit(FitPar(fp));
    FitOptInit(fp);
    VecInit(FitRegMrk(fp), sizeof(Pair_t));
    VecInit(FitBgMrk(fp), sizeof(Pair_t));
    VecInit(FitBgCovar(fp), sizeof(double));
    VecInit(FitPeakSortIdx(fp), sizeof(int));
    VecInit(FitPeakSortRevIdx(fp), sizeof(int));

    FitBgGrad(fp)	= 2;
    FitBgExpNum(fp)	= 0;
    FitBgTyp(fp)	= FIT_BGPOL;

    FitFct(fp)		= FitFctAE;
    FitPeakTyp(fp)	= FIT_PKAE;
    FitTlE(fp)		= 5;
    FitTrE(fp)		= 5;

    FitReset(fp);
    return TRUE; }
}
/* ------------------------------------------------------------------------- */

/* ========== */
/* free Fit_t */
/* ========== */

int FitFree(fp)
     Fit_t fp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    ParFree(FitPar(fp));
    VecFree(FitRegMrk(fp));
    VecFree(FitBgMrk(fp));
    VecFree(FitBgCovar(fp));
    VecFree(FitPeakSortIdx(fp));
    return TRUE; }
}
/* ------------------------------------------------------------------------- */

/* ======================= */
/* copy Fit_t and FitInt_t */
/* ======================= */

int FitCopy(t, f)
     Fit_t t;
     Fit_t f;
{
  if (t == NULL || f  == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    int result;
    memcpy((void *)FitB(t), (void *)FitB(f), sizeof(FitB_t));

    result= (FitStatCopy(FitStatBg(t), FitStatBg(f)) &&
	     FitStatCopy(FitStatReg(t), FitStatReg(f)) &&

	     FitIntCopy(FitIntegral(t), FitIntegral(f)) &&
	     FitIntCopy(FitBgIntegral(t), FitBgIntegral(f)) &&
	     FitIntCopy(FitSubIntegral(t), FitSubIntegral(f)) &&

	     ParCopy(FitPar(t), FitPar(f)) &&

	     VecCopy(FitRegMrk(t), FitRegMrk(f)) &&
	     VecCopy(FitBgMrk(t), FitBgMrk(f)) &&

	     VecCopy(FitBgCovar(t), FitBgCovar(f)) &&

	     VecCopy(FitPeakSortIdx(t), FitPeakSortIdx(f)) &&
	     VecCopy(FitPeakSortRevIdx(t), FitPeakSortRevIdx(f)));
    if (!result) { FitReset(t); }
    return result; }
}
/* ------------------------------------------------------------------------- */
int FitStatCopy(t, f)
     FitStat_t *t;
     FitStat_t *f;
{
  if (t == NULL || f  == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else { memcpy((void *)t, (void *)f, sizeof(FitStat_t)); return TRUE; }
}
/* ------------------------------------------------------------------------- */
int FitIntCopy(t, f)
     FitInt_t *t;
     FitInt_t *f;
{
  if (t == NULL || f  == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else { memcpy((void *)t, (void *)f, sizeof(FitInt_t)); return TRUE; }
}
/* ------------------------------------------------------------------------- */
static int FitOptInit(fp)
     Fit_t fp;
{
  FitOptPeak(fp)= NULL;
  VecInit(FitOptGradVec(fp), sizeof(double));
  return TRUE;
}
/*---------------------------------------------------------------------------*/

/* ============================ */
/* operations on fit parameters */
/* ============================ */

int FitDelPeak(fp, idx)
     Fit_t fp;
     int idx;
{
  Par_t pp= FitPar(fp);
  ParV_t pv[FIT_PEAKPAR];

  { int i;
    for (i= 0; i< FIT_PEAKPAR; i++) {
      if (!ParGet(pp, FitPosIdx(idx) + i, pv + i)) { return FALSE; } } }
  { int i;
    for (i= 0; i < FIT_PEAKPAR; i++) {
      ParStat(pp,FitPosIdx(idx) + i) = PAR_SDELD; } }

  if (ParCompress(pp)) {	/* eliminate deleted parameters */
    FitPeakNum(fp)--;		/* one peak deleted */
    FitTime(fp)= 0;
    return FitPeakSort(fp, FitPeakSortTyp(fp)); }
  else {			/* restore old values on error */
    int i;
    for (i= 0; i< FIT_PEAKPAR; i++) {
      ParSet(pp, FitPosIdx(idx) + i, pv + i, FALSE); }
    return FALSE; }
}
/*---------------------------------------------------------------------------*/
int FitSetupBg(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  ParV_t pv;
  int i;

  ParVInit(&pv);
  ParVUserStat(&pv)	= FIT_USNONE;
  ParVStat(&pv)		= PAR_SSUSP;
  ParVUserTyp(&pv)	= FIT_UTBGPOL;
  ParVMode(&pv)		= PAR_MFREE;

  for (i= FitBgGrad(fp); i< FIT_BGGRD; i++) {
    if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; } }
  for (i= FIT_BGGRD + 2 * FitBgExpNum(fp); i< FIT_1STPEAK; ) {
    ParVUserTyp(&pv)	= FIT_UTBGFAC;
    if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; }
    i++; 
    ParVUserTyp(&pv)	= FIT_UTBGEXP;
    if (!ParSet(pp, i, &pv, FALSE)) { return FALSE; }
    i++; }

  return TRUE;
}
/*---------------------------------------------------------------------------*/
int FitSetPeak(fp, x)
     Fit_t fp;
     double x;
{
  Par_t pp= FitPar(fp);
  ParV_t pv;

  FitTime(fp)= 0;
  ParVInit(&pv);
  ParVStat(&pv)		= PAR_SUSED;
  ParVMode(&pv)		= PAR_MFREE;

  ParVUserStat(&pv)	= FIT_USINIT;
  ParVVal(&pv)		= x;
  ParVUserVal(&pv)	= x;
  ParVUserTyp(&pv)	= FIT_UTPOS;
  if (ParSet(pp, -1, &pv, FALSE)) {	/* 1.: position */
    ParVUserStat(&pv)	= FIT_USUNDEF;
    ParVVal(&pv)	= 0.0;
    ParVUserVal(&pv)	= 0.0;
    ParVUserTyp(&pv)	= FIT_UTWIDTH;
    if (ParSet(pp, -1, &pv, FALSE)) {	/* 2.: width */
      ParVUserTyp(&pv)	= FIT_UTVOL;
      if (ParSet(pp, -1, &pv, FALSE)){ /* 3.: volume */
	ParVUserStat(&pv)	= FIT_USNONE;
	ParVStat(&pv)		= PAR_SSUSP;
	ParVUserTyp(&pv)	= FIT_UTTL;
	if (ParSet(pp, -1, &pv, FALSE)){ /* 4.: tailleft */
	  ParVUserTyp(&pv)	= FIT_UTTR;
	  if (ParSet(pp, -1, &pv, FALSE)){ /* 5.: tailright */
	    ParVUserTyp(&pv)	= FIT_UTSH;
	    if (ParSet(pp, -1, &pv, FALSE)){ /* 6.: step height */
	      ParVUserTyp(&pv)	= FIT_UTSW;
	      if (ParSet(pp, -1, &pv, FALSE)){ /* 7.: step width */
		int i= FitPeakNum(fp)++;

		if (i) {
		  if (i == 1) {	/* automatic correlation from 1st peak on */
		    if (FitWCorr(fp) < 0 &&
			ParVMfree(FitPTl(pp,0))) {
		      FitWCorr(fp)= FitWidthIdx(0); } /* width equal */
		    if (ParVSused(FitPTl(pp,0)) &&
			ParVMfree(FitPTl(pp,0))) {
		      FitTlCorr(fp)= FitTlIdx(0); } /* used tail equal */
		    if (ParVSused(FitPTr(pp,0)) &&
			ParVMfree(FitPTr(pp,0))) {
		      FitTrCorr(fp)= FitTrIdx(0); }
		    if (ParVSused(FitPSh(pp,0)) &&
			ParVMfree(FitPSh(pp,0))) { /* used step equal */
		      FitShCorr(fp)= FitShIdx(0); }
		    if (ParVSused(FitPSw(pp,0))) {
		      if (ParVMfree(FitPSw(pp,0))) {
			FitSwCorr(fp)= FitSwIdx(0); } } }

		  ParVCorr(&pv)[0]= 0.0;
		  ParVCorr(&pv)[1]= 1.0;
		  ParVMode(&pv)= PAR_MCORR;
		  ParVStat(&pv)= PAR_SUSED;
		  ParVUserStat(&pv)= FIT_USUSER;
		  if (FitWCorr(fp) >= 0) {
		    ParVUserTyp(&pv)= FIT_UTWIDTH;
		    ParVCorrIndex(&pv)= FitWCorr(fp);
		    ParSet(pp, FitWidthIdx(i), &pv, FALSE); }
		  if (FitTlCorr(fp) >= 0) {
		    ParVUserTyp(&pv)= FIT_UTTL;
		    ParVCorrIndex(&pv)= FitTlCorr(fp);
		    ParSet(pp, FitTlIdx(i), &pv, FALSE); }
		  if (FitTrCorr(fp) >= 0) {
		    ParVUserTyp(&pv)= FIT_UTTR;
		    ParVCorrIndex(&pv)= FitTrCorr(fp);
		    ParSet(pp, FitTrIdx(i), &pv, FALSE); }
		  if (FitShCorr(fp) >= 0) {
		    ParVUserTyp(&pv)= FIT_UTSH;
		    ParVCorrIndex(&pv)= FitShCorr(fp);
		    ParSet(pp, FitShIdx(i), &pv, FALSE); }
		  if (FitSwCorr(fp) >= 0) {
		    ParVUserTyp(&pv)= FIT_UTSW;
		    ParVCorrIndex(&pv)= FitSwCorr(fp);
		    ParSet(pp, FitSwIdx(i), &pv, FALSE); }
		  else if (ParVSused(FitPSw(pp,0)) &&
			   ParVMhold(FitPSw(pp,0))) {
		    ParVMode(&pv)= PAR_MHOLD; /* fix step width to 1.0 */
		    ParVVal(&pv)= 1.0;
		    ParSet(pp, FitSwIdx(i), &pv, FALSE); } }

		return FitPeakSort(fp, FitPeakSortTyp(fp)); }
				/* error while adding */
				/* delete already added parameters */
	      ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-6, &pv, FALSE); }
	    ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-5, &pv, FALSE); }
	  ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-4, &pv, FALSE); }
	ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-3, &pv, FALSE); }
      ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-2, &pv, FALSE); }
    ParVStat(&pv)= PAR_SDELD; ParSet(pp, ParNum(pp)-1, &pv, FALSE); }
  ParCompress(pp);
  return FALSE;
}
/*---------------------------------------------------------------------------*/
int FitPeakSort(fp, typ)
     Fit_t fp;
     int typ;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else if (!VecReset(FitPeakSortIdx(fp)) ||
	   !VecReset(FitPeakSortRevIdx(fp))) { return FALSE; }
  else if (FitPeakNum(fp) == 0) { return TRUE; }
  else if (!VecAlloc(FitPeakSortIdx(fp), FitPeakNum(fp)) ||
	   !VecAlloc(FitPeakSortRevIdx(fp), FitPeakNum(fp))) {
    return FALSE; }
  else {
    int *si= VecP(FitPeakSortIdx(fp));

    VecNum(FitPeakSortIdx(fp))= FitPeakNum(fp);
    VecNum(FitPeakSortRevIdx(fp))= FitPeakNum(fp);

    switch (typ) {
    case FIT_STPOS:
      { int i;
	for (i= 0; i< FitPeakNum(fp); i++) { si[i]= FitPosIdx(i); } }
      if (!ParSort(FitPar(fp), si, FitPeakNum(fp), PAR_VVAL)) {
	return FALSE; }
      else {
	int i;
	for (i= 0; i< FitPeakNum(fp); i++) { si[i]= FitPeakIdx(si[i]); } }
      break;
    case FIT_STHIST:
      { int i;
	for (i= 0; i< FitPeakNum(fp); i++) { si[i]= i; } }
      break;
    case FIT_STVOL:
      { int i;
	for (i= 0; i< FitPeakNum(fp); i++) { si[i]= FitVolIdx(i); } }
      if (!ParSort(FitPar(fp), si, FitPeakNum(fp), PAR_VVAL)) {
	return FALSE; }
      else {
	int i;
	for (i= 0; i< FitPeakNum(fp); i++) { si[i]= FitPeakIdx(si[i]); } }
      break;
    default: FitsetErrNo(FITILLST); return FALSE; }

    FitPeakSortTyp(fp)= typ;

    { int i;
      int *ri= VecP(FitPeakSortRevIdx(fp));
      for (i= 0; i< FitPeakNum(fp); i++) { ri[si[i]] = i; }
      return TRUE; } }
}
/* ---------------------------------------------------------------------- */
int FitMaxPeakIdx(fp, idx)
     Fit_t fp;
     int *idx;
{				/* find amplitude of largest peak */
  if (!fp) { FitsetErrNo(FITPOINT); return FALSE; }
  else {
    Par_t pp= FitPar(fp);
    int i;
    double maxAmpl= 0.0;
    int maxIdx= -1;
    for (i= 0; i< FitPeakNum(fp); i++) {
      int j= FitVolIdx(i);
      if (ParStat(pp,j) == PAR_SUSED &&
	  ParUserStat(pp,j) != FIT_USUNDEF &&
	  ParUserStat(pp,j) != FIT_USSTART) {
	double a= fabs(FitPVolVal(pp,i));
	if (a > maxAmpl) { maxAmpl= a; maxIdx= i; } } }

    if (maxIdx >= 0) { *idx= maxIdx; return TRUE; }
    else { FitsetErrNo(FIT0AMPL); return FALSE; } }
}
/* ---------------------------------------------------------------------- */
/* ============================================== */
/* fit function, component functions and gradient */
/* ============================================== */

int FitFunctionChange(fp, typ)
     Fit_t fp;
     int typ;
{
  if (FitPeakTyp(fp) != typ) {
    switch (typ) {
    case FIT_PKCA:
      if (FitTlE(fp) < 0 || 2 < FitTlE(fp)) { FitTlE(fp)= 0; }
      if (FitTrE(fp) < 0 || 2 < FitTrE(fp)) { FitTrE(fp)= 0; }
      FitFct(fp)= FitFctCA;
      break;
    case FIT_PKAE:
      if (FitTlE(fp) < 2 || 15 < FitTlE(fp)) { FitTlE(fp)= 5; }
      if (FitTrE(fp) < 2 || 15 < FitTrE(fp)) { FitTrE(fp)= 5; }
      FitFct(fp)= FitFctAE;
      break;
    default: FitsetErrNo(FITPKTYP); return FALSE; }

    FitPeakTyp(fp)= typ;
    FitTime(fp)= 0; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int FitBgFunctionSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  int i;
  pp = FitOptPp(fp) = (!pp) ? FitPar(fp) : pp;

  FitOptFitBg(fp) = FALSE;
  for (i= 0; i< FIT_1STPEAK; i++) {
    if (ParVSused(FitPBgCoef(pp,i))) {
      FitOptFitBg(fp) = TRUE; break; } }

  FitOptFitPeaks(fp) = FALSE; /* fit no peaks */
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int FitBgFunctionFinish(fp)
     Fit_t fp;
{
  return TRUE;
}
/* ------------------------------------------------------------------------- */
double FitBgFunction(fp, x)
     Fit_t fp;
     double x;
{
  Par_t pp= FitOptPp(fp);
  double result = 0.0;

  if (FitOptFitBg(fp)) {
    { int i=FitBgGrad(fp)-1;
      x -= FitBgOffset(fp);
      while (i >= 0) { result = result * x + FitPBgCoefVal(pp,i); i--; } }

    { int i, i_end= FIT_BGGRD + 2 * FitBgExpNum(fp);
      for (i= FIT_BGGRD; i< i_end; i += 2) {
	result += (FitPBgCoefVal(pp,i) * 
		   exp(-x * FitPBgCoefVal(pp,i+1))); } } }
  return result;
}
/*---------------------------------------------------------------------------*/
double FitDBg2Function(fp, x)
     Fit_t fp;
     double x;
{
  switch (FitBgTyp(fp)) {
  case FIT_BGPOL:	return FitDBg2Polynom(fp, x);
  default:		FitsetErrNo(FITBGTYP); return -1.0; }
}
/* ---------------------------------------------------------------------- */
static double FitDBg2Polynom(fp, x)	/* square of deviation of bg */
     Fit_t fp;
     double x;
{				/* takes only polynom in account !!! */
  double result= 0.0;
  if (FitBgTime(fp)) {
    double xPower[FIT_BGGRD];
    int i;
    int bgGrd= FitBgGrad(fp);
    double *c= VecP(FitBgCovar(fp));
    xPower[0]= 1.0;
    x -= FitBgOffset(fp);
    for (i= 1; i< bgGrd; i++) { xPower[i]= xPower[i-1] * x; }
    for (i=0;i<bgGrd;i++) {
      int j;
      for (j=0;j<=i;j++,c++) {
	double p= xPower[i] * xPower[j] * *c;
	result += p;
	if (i != j) { result += p; } } } } /* offdiagonal */

  return result;
}
/*---------------------------------------------------------------------------*/
double FitTotalBgFunction(fp, x)
     Fit_t fp;
     double x;
{
  double result= FitBgFunction(fp, x);
  if (FitOptFitPeaks(fp)) { return (result + FitStepFunction(fp, x)); }
  else { return result;}
}
/*---------------------------------------------------------------------------*/
int FitFunctionSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  return FitFctFunctionSetup(FitFct(fp))(fp, pp);
}
/*---------------------------------------------------------------------------*/
int FitFunctionFinish(fp)
     Fit_t fp;
{
  return FitFctFunctionFinish(FitFct(fp))(fp);
}
/*---------------------------------------------------------------------------*/
FitOpt FitPeakFunctionSetup(vp, fp, pp, peakIdx)
     FitOpt vp;
     Fit_t fp;
     Par_t pp;
     int peakIdx;
{
  if (!pp) { pp= FitPar(fp); }
  return FitFctPeakFunctionSetup(FitFct(fp))(vp, fp, pp, peakIdx);
}
/*---------------------------------------------------------------------------*/
FitOpt FitPeakFunctionFinish(op)
     FitOpt op;
{
  return (op ? FitOptPeakFunctionFinish(op)(op) : NULL);
}
/*---------------------------------------------------------------------------*/
int FitDerivationSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  return FitFctDerivationSetup(FitFct(fp))(fp, pp);
}
/*---------------------------------------------------------------------------*/
int FitDerivationFinish(fp)
     Fit_t fp;
{
  return FitFctDerivationFinish(FitFct(fp))(fp);
}
/*---------------------------------------------------------------------------*/
FitOpt FitUnitPeakDerivationSetup(vp, fp, pp, peakIdx)
     FitOpt vp;
     Fit_t fp;
     Par_t pp;
     int peakIdx;
{
  if (!pp) { pp= FitPar(fp); }
  return FitFctUnitPeakDerivationSetup(FitFct(fp))(vp, fp, pp, peakIdx);
}
/*---------------------------------------------------------------------------*/
FitOpt FitUnitPeakDerivationFinish(op)
     FitOpt op;
{
  return (op ? FitOptUnitPeakDerivationFinish(op)(op) : NULL);
}
/*---------------------------------------------------------------------------*/
int FitDerivation(fp, x, grad)
     Fit_t fp;
     double x;
     double *grad;
{
  return FitFctDerivation(FitFct(fp))(fp, x, grad);
}
/*---------------------------------------------------------------------------*/
int FitUnitPeakDerivation(op, x, x0, fct_val, grad_x0)
     FitOptCA op;
     double x;
     double x0;
     double *fct_val;
     double *grad_x0;
{
  return FitOptUnitPeakDerivation(op)(op, x, x0, fct_val, grad_x0);
}
/*---------------------------------------------------------------------------*/
double FitFunction(fp, x)
     Fit_t fp;
     double x;
{
  return FitFctFunction(FitFct(fp))(fp, x);
}
/*---------------------------------------------------------------------------*/
double FitNormVolume(op)
     FitOpt op;
{
  return FitOptNormVol(op)(op);
}
/*---------------------------------------------------------------------------*/
double FitStepFunction(fp, x)
     Fit_t fp;
     double x;
{
  return FitFctStepFunction(FitFct(fp))(fp, x);
}
/*---------------------------------------------------------------------------*/
double FitPeakFunction(op, x)
     FitOpt op;
     double x;
{
  return FitOptPeakFunction(op)(op, x);
}
/*---------------------------------------------------------------------------*/
double FitUnitPeakFunction(op, x, x0)
     FitOpt op;
     double x;
     double x0;
{
  return FitOptUnitPeakFunction(op)(op, x, x0);
}
/*---------------------------------------------------------------------------*/
static int BgDerivation(fp, grd, x)
     Fit_t fp;
     double *grd;
     double x;
{
  if (!FitOptFitBg(fp)) { return TRUE; }
  switch (FitBgTyp(fp)) {
  case FIT_BGPOL:
    { int i;
      for (i= 1, grd[0]= 1.0; i< FIT_BGGRD; i++) { grd[i]= grd[i-1] * x; } }
    { int i, p;
      Par_t pp= FitOptPp(fp);
      for (i= 0, p= FIT_BGGRD; i< FitBgExpNum(fp); i++, p += 2) {
	grd[p]= exp(-x * FitPBgCoefVal(pp,p+1));
	grd[p+1]= -x * FitPBgCoefVal(pp,p) * grd[p]; } }
    return TRUE;
  default:
    FitsetErrNo(FITBGTYP); return FALSE; }
}
/*---------------------------------------------------------------------------*/
				/* original leone fitfunction */
				/* tail: continuous exponential tail */
				/* step: arctan */

#define PEAK_VALUE(pp,x,p,result) \
{\
  double dx = x - ParVVal(pp + FIT_PEAKPOS); /* distance from actual peak */\
  double sig = ParVVal(pp + FIT_PEAKWIDTH); /* peak sigma */\
  double phi;			/* pure peak contribution */\
  if (ABS(dx) > FORGET_TAIL * ABS(sig)) { phi= 0.; }\
  else if (ParVSused(pp + FIT_PEAKTL) &&\
	   dx < (-FitOptCASl(p))) { /* tailleft */\
    phi= exp(FitOptCASl(p) * (dx + FitOptCASl(p) * 0.5) * FitOptCASs(p)); }\
  else if (ParVSused(pp + FIT_PEAKTR) && dx > FitOptCASr(p)) { /* tailright */\
    phi= exp(-(FitOptCASr(p) * (dx - FitOptCASr(p) * 0.5) * FitOptCASs(p))); }\
  else {	/* -FitOptCASl(p) <= dx <= FitOptCASr(p) : gaussian shape */\
    phi= exp(-dx * dx * 0.5 * FitOptCASs(p)); }\
  result += phi * (FitOptCAVolNorm(p) * ParVVal(pp + FIT_PEAKVOL));\
}
#if 0
{				/* for debugging */
  ParV_t *pp= par;
  FitOptCA p= (FitOptCA)pre;

  double dx = x - ParVVal(pp + FIT_PEAKPOS); /* distance from actual peak */
  double sig = ParVVal(pp + FIT_PEAKWIDTH); /* peak sigma */
  double phi;			/* pure peak contribution */
  if (ABS(dx) > FORGET_TAIL * ABS(sig)) { phi= 0.; }
  else if (ParVSused(pp + FIT_PEAKTL) &&
	   dx < (-FitOptCASl(p))) { /* tailleft */
    phi= exp(FitOptCASl(p) * (dx + FitOptCASl(p) * 0.5) * FitOptCASs(p)); }
  else if (ParVSused(pp + FIT_PEAKTR) && dx > FitOptCASr(p)) { /* tailright */
    phi= exp(-(FitOptCASr(p) * (dx - FitOptCASr(p) * 0.5) * FitOptCASs(p))); }
  else {	/* -FitOptCASl(p) <= dx <= FitOptCASr(p) : gaussian shape */
    phi= exp(-dx * dx * 0.5 * FitOptCASs(p)); }
  result += phi * (FitOptCAVolNorm(p) * ParVVal(pp + FIT_PEAKVOL));
}
#endif
/*---------------------------------------------------------------------------*/
static int CAFunctionSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  if (!FitBgFunctionSetup(fp, pp)) { return FALSE; }
  else if (!FitPeakNum(fp)) { return TRUE; }
  else {
    FitOptCA o;

    if ((o= (FitOptCA)Malloc(FitPeakNum(fp) * sizeof(FitOptCARec)))) {
      memset(o, 0, FitPeakNum(fp) * sizeof(FitOptCARec));

      pp = FitOptPp(fp) = (!pp) ? FitPar(fp) : pp;

      FitOptFitPeaks(fp) = FALSE; /* fit peaks only if there are any */
      { int i;
	FitOptCA p;

	for (i= 0, p= o; i< FitPeakNum(fp); i++, p++) {
	  if (CAPeakFunctionSetup(p, fp, pp, i)) {
	    FitOptFitPeaks(fp) = TRUE; }
	  else if (FitErrNo == FITSIGM0) { Free(o); return FALSE; } } }

      FitOptPeak(fp)= (void *)o;
      return TRUE; }
    else { FitsetErrNo(FITALLOC); return FALSE; } }
}
/*---------------------------------------------------------------------------*/
static int CAFunctionFinish(fp)
     Fit_t fp;
{
  Free(FitOptPeak(fp)); return TRUE;
}
/*---------------------------------------------------------------------------*/
static FitOptCA CAPeakFunctionSetup(vp, fp, pp, peakIdx)
     FitOptCA vp;
     Fit_t fp;
     Par_t pp;
     int peakIdx;
{
  int el= FitTlE(fp);
  int er= FitTrE(fp);
  FitOptCA op= vp;
  ParV_t *pv= VecP(ParPar(pp)) + FitPosIdx(peakIdx);

  if (!pv || !ParVSused(pv + FIT_PEAKWIDTH)) { op= NULL; }
  else {
    double sig = ParVVal(pv + FIT_PEAKWIDTH);

    if (sig == 0.0) { FitsetErrNo(FITSIGM0); return NULL; }
    else if (op ||
	     ((op= (FitOptCA)Malloc(sizeof(FitOptCARec))) &&
	      (FitOptCAPar(op)=
	       (ParV_t *)Malloc(FIT_PEAKPAR * sizeof(ParV_t))))) {
      if (vp) { FitOptCAPar(op)= pv; }
      else { memcpy(FitOptCAPar(op), pv, FIT_PEAKPAR * sizeof(ParV_t)); }

      FitOptCAFct(op)= FitFctCA;

      FitOptCASs(op) = 1.0 / (sig * sig);

      if (ParVSused(pv + FIT_PEAKTL)) {
	switch (el) {
	case 1:  FitOptCASlM(op) = sig; break;
	case 2:  FitOptCASlM(op) = sig * sig; break;
	case 0:
	default: FitOptCASlM(op) = 1.0; break; }
	FitOptCASl(op) = FitOptCASlM(op) * ParVVal(pv + FIT_PEAKTL); }

      if (ParVSused(pv + FIT_PEAKTR)) {
	switch (er) {
	case 1:  FitOptCASrM(op) = sig; break;
	case 2:  FitOptCASrM(op) = sig * sig; break;
	case 0:
	default: FitOptCASrM(op) = 1.0; break; }
	FitOptCASr(op) = FitOptCASrM(op) * ParVVal(pv + FIT_PEAKTR); }

      if (ParVSused(pv + FIT_PEAKSH)) {
	FitOptCAAtan(op) = (ParVVal(pv + FIT_PEAKSW) * M_SQRT1_2 / sig); }

      { double a= CANormVolume(op);
	FitOptCAVolNorm(op) = a ? (1.0 / a) : 1.0; } }
    else { Free(op); FitsetErrNo(FITALLOC); } }
  return op;
}
/* ------------------------------------------------------------------------- */
static FitOptCA CAPeakFunctionFinish(op)
     FitOptCA op;
{
  if (op) { Free(FitOptCAPar(op)); Free(op); }
  return NULL;
}
/*---------------------------------------------------------------------------*/
static int CADerivationSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  else if (!CAFunctionSetup(fp, pp)) { return FALSE; }
  else if (!VecAlloc(FitOptGradVec(fp), ParNum(FitPar(fp)))) {
    FitsetErrNo(FITALLOC); return FALSE; }
  else if (!FitOptFitPeaks(fp)) { return TRUE; }
  else {
    int i;
    FitOptCA p;

    for (i= 0, p= (FitOptCA)FitOptPeak(fp); i< FitPeakNum(fp);i++, p++) {
      ParV_t *par= FitOptCAPar(p);
      double sig = ParVVal(par + FIT_PEAKWIDTH);

      if (FitOptCASl(p) != 0.0) {
	double x = FitOptCASl(p) / sig;    
	double x2 = x * x;
	FitOptCAExpSl(p) =  FitOptCAVolNorm(p) * sig * exp(-x2 * 0.5) / x; }
      if (FitOptCASr(p) != 0.0) {
	double x = FitOptCASr(p) / sig;    
	double x2 = x * x;
	FitOptCAExpSr(p)= FitOptCAVolNorm(p) * sig * exp(-x2 *0.5) / x; } }
    return TRUE; }
}
/*---------------------------------------------------------------------------*/
static int CADerivationFinish(fp)
     Fit_t fp;
{
  VecFree(FitOptGradVec(fp));
  return CAFunctionFinish(fp);
}
/*---------------------------------------------------------------------------*/
static int CADerivation(fp, x, grad)
     Fit_t fp;			/* preinitialized values and parameters */
     double x;			/* location to calculate the derivations     */
     double *grad;		/* derivations of fitting function with      */
				/* respect to the fitting parameters         */
{
  Par_t pp= FitOptPp(fp);
  double *grd= FitOptGrad(fp);	/* gradient */

  BgDerivation(fp, grd, x);

  if (FitOptFitPeaks(fp)) {
    int i;
    FitOptCA p;
				/* loop over all peaks */
    for (i= 0, p= (FitOptCA)FitOptPeak(fp); i < FitPeakNum(fp); i++, p++) {
      ParV_t *par= FitOptCAPar(p);
      double sig = ParVVal(par + FIT_PEAKWIDTH); /* peak sigma */
      double are = ParVVal(par + FIT_PEAKVOL); /* peak volume */
      double sh= ParVVal(par + FIT_PEAKSH); /* step height */
      double sw= ParVVal(par + FIT_PEAKSW); /* step width */
      double dx = x - ParVVal(par + FIT_PEAKPOS); /* peak location */
      double f = FitOptCAVolNorm(p); /* 1 / (peak with unity amplitude) */
      double af = f * are;	/* f * fitted volume */
      double b= 0.;
      double rezi= 0.;
      double sig2 = FitOptCASs(p);
      double lef = FitOptCASl(p);	/* left tail border == sl */
      double rig = FitOptCASr(p);	/* right tail border == sr */
      double ko_xi;		/* koeff of phi, when derived to x[i] */
      double ko_tl;		/* koeff of phi, when derived to tl */
      double ko_tr;		/* koeff of phi, when derived to tr */
      double phiatan;		/*  == (phi + sh * atan (...)) */
      double phi;		/* peak function */
      double dphi;		/* derive phi to sigma */


      if (ParVSused(par + FIT_PEAKSH)) {
	double a = FitOptCAAtan(p) * dx;/* = dx * sw /(sigma(i) * sqrt(2))*/
	b = (M_PI * 0.5 + atan(a));
	rezi = sh / (a + 1/a);
	grd[FitShIdx(i)] = af * b;
	grd[FitSwIdx(i)] = af * rezi / sw; }

				/* get region dependend values */
      if (ABS(dx) > FORGET_TAIL * ABS(sig)) { /* far away from peak - forget */
	phi   = 0.;
	dphi  = 0.;
	ko_xi = 0.;
	ko_tl = 0.;
	ko_tr = 0.; }
      else if (ParVSused(par + FIT_PEAKTL) &&
	       dx < -lef) {	/* left tail region */
	double z;

	if (FORGET_EXP > -(z =  (dx + (lef * 0.5)) * lef * sig2)) {
	  phi   = exp (z);
	  dphi  = lef * (dx * (FitTlE(fp) - 2.) +
			 lef * (FitTlE(fp) - 1.)) * sig2 ; }
	else {
	  phi   = 0.;
	  dphi  = 0.; }

	ko_xi = - lef * sig2;
	ko_tl = (dx + lef) * sig2;
	ko_tr = 0.; }
      else if (ParVSused(par + FIT_PEAKTR) &&
	       dx > rig) {	/* right tail region */
	double z;
	if (FORGET_EXP >  (z =  (dx - (rig * 0.5)) * rig * sig2)) {
	  phi   = exp (-z);
	  dphi  = -rig * (dx * (FitTrE(fp) - 2.) -
			  rig * (FitTrE(fp) - 1.)) * sig2; }
	else {
	  phi   = 0.;
	  dphi  = 0.; }

	ko_xi = rig * sig2;
	ko_tl = 0.;
	ko_tr = -(dx - rig) * sig2; }
      else {			/* gaussian region */
	phi   = exp (-dx * dx * sig2 * 0.5);
	ko_xi = dx * sig2;
	ko_tl = 0.;
	ko_tr = 0.;
	dphi  = dx * dx * sig2; }
      phiatan = phi + b * sh;
      dphi *= phi;

      if (ParVSused(par + FIT_PEAKTL)) {
	grd[FitTlIdx(i)]= (af *
			   (ko_tl * phi + phiatan * FitOptCAExpSl(p) / lef) *
			   FitOptCASlM(p)); }
      if (ParVSused(par + FIT_PEAKTR)) {
	grd[FitTrIdx(i)]= (af *
			   (ko_tr * phi + phiatan * FitOptCAExpSr(p) / rig) *
			   FitOptCASrM(p)); }

      grd[FitPosIdx(i)]    = af * (ko_xi  * phi -
				   ((dx == 0.) ?
				    (sh * FitOptCAAtan(p)) : (rezi / dx)));
      grd[FitWidthIdx(i)]  = (af / sig *
			      (dphi - rezi - phiatan *
			       ((1. - FitTlE(fp)) * FitOptCAExpSl(p) +
				(1. - FitTrE(fp)) * FitOptCAExpSr(p) + 1.)));
      grd[FitVolIdx(i)]  = phiatan * f; } }
  { int i;
    for (i=0; i<ParNumFree(pp); i++)  { grad[i] = grd[ParIfree(pp,i)]; } }

  { int i;
    for (i=0; i<ParNumCorr(pp); i++) {
      int j= ParIcorr(pp,i);
      grad[ParIcofr(pp,i)] += grd[j] * ParCorr(pp,j)[1]; } }

  return TRUE;
}
/*---------------------------------------------------------------------------*/
static int CAUnitPeakDerivation(op, x, x0, fct_val, grad_x0)
     FitOptCA op;
     double x;
     double x0;
     double *fct_val;
     double *grad_x0;
{
  if (!op) { return FALSE; }
  else {
    ParV_t *par= FitOptCAPar(op);
    double sig = ParVVal(par + FIT_PEAKWIDTH); /* peak sigma */
    double dx = x - x0; /* peak location */
    double sig2 = FitOptCASs(op);
    double lef = FitOptCASl(op); /* left tail border == sl */
    double rig = FitOptCASr(op); /* right tail border == sr */

				/* get region dependend values */
    if (ABS(dx) > FORGET_TAIL * ABS(sig)) { /* far away from peak - forget */
      *fct_val= 0.;
      *grad_x0= 0.; }
    else if (ParVSused(par + FIT_PEAKTL) &&
	     dx < -lef) {	/* left tail region */
      double z;

      if (FORGET_EXP > -(z =  (dx + (lef * 0.5)) * lef * sig2)) {
	*fct_val= exp (z);
	*grad_x0= - *fct_val * lef * sig2; }
      else {
	*fct_val= 0.0;
	*grad_x0= 0.0; } }
    else if (ParVSused(par + FIT_PEAKTR) &&
	     dx > rig) {	/* right tail region */
      double z;
      if (FORGET_EXP >  (z =  (dx - (rig * 0.5)) * rig * sig2)) {
	*fct_val= exp(-z);
	*grad_x0= *fct_val * rig * sig2; }
      else {
	*fct_val= 0.0;
	*grad_x0= 0.0; } }
    else {			/* gaussian region */
      *fct_val= exp (-dx * dx * sig2 * 0.5);
      *grad_x0= *fct_val * dx * sig2; }

    return TRUE; }
}
/*---------------------------------------------------------------------------*/
static double CAFunction(fp, x)
				/* function to calculate the value of fitting*/
				/* function at location x in dependence of   */
				/* the parameters in fp                     */
     Fit_t fp;			/* preinitialized values and parameters */
     double x;			/* is the index of the correlated y value in */
				/* spectrum, could be used for x calibration */
{
  double result;		/* hold actual result */

  result= FitTotalBgFunction(fp, x); /* background */

  if (FitOptFitPeaks(fp)) {
    int i;		/* loop index */
    FitOptCA p;
    for (i= 0, p= (FitOptCA)FitOptPeak(fp); i < FitPeakNum(fp); i++, p++) {
      ParV_t *par= FitOptCAPar(p);
      PEAK_VALUE(par, x, p, result); } }
  return result;
}
/*---------------------------------------------------------------------------*/
static double CANormVolume(pre)
				/* volume of a 1-Peak (gauss) with exp-tail */
     FitOptCA pre;
{
  ParV_t *par= FitOptCAPar(pre);
  double sigma;			/* sigma of peak */
  double left;			/* >0, left start of exponential tail*/
  double right;			/* >0, right one ... */
  double term_l;		/* actual terms in sum */
  double term_r;
  double sum_l;			/* total sums (gauss-integral) */
  double sum_r;
  double denominator;		/* actual denominator in sum */
  double sl;			/* sigma / left */
  double sr;
  double sl2;			/* (sigma / left) ^ 2 */
  double sr2;
  register double result;	/* actual result to hold code readable */

  if (!ParVSused(par + FIT_PEAKWIDTH)) { return 0.; }

  sigma =  ParVVal(par + FIT_PEAKWIDTH);

  if (sigma <= 0)  return 0.; /* empty peak */
  
  left = FitOptCASl(pre);
  right = FitOptCASr(pre);

  if (right < TAIL_TOTAL) right = TAIL_TOTAL; /* no physical meaning !! */
  if (left < TAIL_TOTAL)   left = TAIL_TOTAL;

				/* initialize  */
  sum_l = term_l = sl  =  left / sigma;
  sum_r = term_r = sr  = right / sigma;
  sl2   = sl * sl;
  sr2   = sr * sr;
  denominator = 1.0;

  if ((!ParVSused(par + FIT_PEAKTL) || sl2 > FORGET_EXP) &&
      (!ParVSused(par + FIT_PEAKTR) || sr2 > FORGET_EXP)) {
    return (sigma * SQRT_2_PI); } /* pure gaussian peak shape */

  if (ParVSused(par + FIT_PEAKTL) && ParVSused(par + FIT_PEAKTR) &&
      sl2 <= FORGET_EXP && sr2 <= FORGET_EXP) {
				/* mixed shapes on both sides */
    do {			/* evaluate sums              */
      denominator += 2.;
      term_l  *= (sl2 / denominator);
      term_r  *= (sr2 / denominator);
      sum_l   += term_l;
      sum_r   += term_r;
    } while  ((term_l / sum_l > PRECISION) &&
	      (term_r / sum_r > PRECISION));

				/* let the other sum converge                */
    while (term_l / sum_l > PRECISION) {
      denominator += 2.;
      term_l  *= (sl2 / denominator);
      sum_l   += term_l; }
    while (term_r / sum_r > PRECISION) {
      denominator += 2.;
      term_r  *= (sr2 / denominator);
      sum_r   += term_r; }
				/* complete it                               */
    result  = exp(-sl2 * 0.5) * (sigma / left  + sum_l);
    result += exp(-sr2 * 0.5) * (sigma / right + sum_r);
    result *= sigma;
  }                   /*   if ((sl2 <= FORGET_EXP) &&  (sr2 <= FORGET_EXP))  */
  else if (!ParVSused(par + FIT_PEAKTR) || sr2 > FORGET_EXP) {
				/* pure gauss on right side                  */
				/* other side is mixed                       */
    do {
      denominator += 2.;
      term_l  *= (sl2 / denominator);
      sum_l   += term_l;
    } while (term_l / sum_l  > PRECISION);
				/* complete it                               */
    result  = exp(-sl2 * 0.5) * (sigma / left  + sum_l);
    result += SQRT_2_PI * 0.5;
    result *= sigma; }
  else {			/* if (!ParVSused(par + FIT_PEAKTL)          */
				/*     || sl2 > FORGET_EXP) {                */
				/* pure gauss on left side                   */
				/* other side is mixed                       */
    do {
      denominator += 2.;
      term_r  *= (sr2 / denominator);
      sum_r   += term_r;
    } while  (term_r / sum_r  > PRECISION);
				/* complete it */
    result  =  SQRT_2_PI / 2;
    result +=  exp(- sr2 / 2) * (sigma / right + sum_r);
    result *=  sigma; }
  return result;
}
/*---------------------------------------------------------------------------*/
static double CAStepFunction(fp, x)
				/* function to calculate the value of step   */
				/* function at location x in dependence of   */
				/* the parameters pp                         */
     Fit_t fp;			/* actual parameter set                      */
     double x;			/* is the index of the correlated y value in */
				/* spectrum, could be used for x calibration */
{
  double result= 0.0;		/* hold actual result */
  FitOptCA p;
  int i;			/* loop index */
				/* loop over all peaks */
  for (i= 0, p= (FitOptCA)FitOptPeak(fp); i < FitPeakNum(fp); i++, p++) {
    ParV_t *par= FitOptCAPar(p);

    if (ParVSused(par + FIT_PEAKPOS) && ParVSused(par + FIT_PEAKSH) &&
	ParVUserStat(par + FIT_PEAKPOS) != FIT_USUNDEF &&
	ParVUserStat(par + FIT_PEAKWIDTH) != FIT_USUNDEF &&
	ParVUserStat(par + FIT_PEAKVOL) != FIT_USUNDEF) {	  
      result += (ParVVal(par + FIT_PEAKSH) *
		 (atan ((x - ParVVal(par + FIT_PEAKPOS)) * FitOptCAAtan(p)) +
		  M_PI * 0.5) *
		 FitOptCAVolNorm(p) * ParVVal(par + FIT_PEAKVOL)); } }
  return result;
}
/*---------------------------------------------------------------------------*/
static double CAPeakFunction(pre, x)
     FitOptCA pre;		/* parameter (optimized) */
     double  x;			/* x value  */
{
  double result=0.0;		/* hold actual result */
				/* prepare for loop */
  if (pre) {
    ParV_t *par= FitOptPar(pre);
    PEAK_VALUE(par, x, pre, result); }
  return result;
}
/*---------------------------------------------------------------------------*/
static double CAUnitPeakFunction(op, x, x0)
     FitOptCA op;
     double x;
     double x0;
{
  ParV_t *par= FitOptCAPar(op);
  double dx = x - x0; /* distance from actual peak */
  double sig = ParVVal(par + FIT_PEAKWIDTH); /* peak sigma */
  double phi;			/* pure peak contribution */

  if (ABS(dx) > FORGET_TAIL * ABS(sig)) { phi= 0.; }
  else if (ParVSused(par + FIT_PEAKTL) &&
	   dx < (-FitOptCASl(op))) { /* tailleft */
    phi= exp(FitOptCASl(op) * (dx + FitOptCASl(op) * 0.5) * FitOptCASs(op)); }
  else if (ParVSused(par + FIT_PEAKTR) &&
	   dx > FitOptCASr(op)) { /* tailright */
    phi= exp(-(FitOptCASr(op) *
	       (dx - FitOptCASr(op) * 0.5) * FitOptCASs(op))); }
  else {	/* -FitOptCASl(op) <= dx <= FitOptCASr(op) : gaussian shape */
    phi= exp(-dx * dx * 0.5 * FitOptCASs(op)); }
  return phi;
}
/*---------------------------------------------------------------------------*/
#undef PEAK_VALUE
				/* new peak function */
				/* tail: additive (x-x0)**m */
				/* step: erf() */

#define PEAK_VALUE(x,op,result) \
{\
  ParV_t *par= FitOptPar(op);\
  double dx= x - ParVVal(par + FIT_PEAKPOS); /* distance from actual peak */\
  double dx2= dx * dx;\
  double ps, ms;\
  double amp= (ParVVal(par + FIT_PEAKVOL) * FitOptAEVolNorm(op) *\
	       exp(-dx2 * FitOptAESExp(op)));\
  if (dx < 0.0) {		/* left region */\
    ps= FitOptAEPl(op); ms= FitOptAEEl(op); }\
  else if (dx >= 0.0) {		/* right region */\
    ps= FitOptAEPr(op); ms= FitOptAEEr(op); }\
  if (ps != 0.0) {\
    result += (1.0 + pow(ABS(dx), ms) * ps) * amp; }\
  else {\
    result += amp; }\
}
/*---------------------------------------------------------------------------*/
static int AEFunctionSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  if (!FitBgFunctionSetup(fp, pp)) { return FALSE; }
  else if (!FitPeakNum(fp)) { return TRUE; }
  else {
    FitOptAE o;

    if ((o= (FitOptAE)Malloc(FitPeakNum(fp) * sizeof(FitOptAERec)))) {
      memset(o, 0, FitPeakNum(fp) * sizeof(FitOptAERec));

      pp = FitOptPp(fp) = (!pp) ? FitPar(fp) : pp;

      FitOptFitPeaks(fp) = FALSE; /* fit peaks only if there are any */
      { int i;
	FitOptAE p;

	for (i= 0, p= o; i< FitPeakNum(fp); i++, p++) {
	  if (AEPeakFunctionSetup(p, fp, pp, i)) {
	    FitOptFitPeaks(fp) = TRUE; }
	  else if (FitErrNo == FITSIGM0) { Free(o); return FALSE; } } }

      FitOptPeak(fp)= (void *)o;
      return TRUE; }
    else { FitsetErrNo(FITALLOC); return FALSE; } }
}
/*---------------------------------------------------------------------------*/
static int AEFunctionFinish(fp)
     Fit_t fp;
{
  Free(FitOptPeak(fp)); return TRUE;
}
/*---------------------------------------------------------------------------*/
static FitOptAE AEPeakFunctionSetup(vp, fp, pp, peakIdx)
     FitOptAE vp;
     Fit_t fp;
     Par_t pp;
     int peakIdx;
{
  int el= FitTlE(fp);
  int er= FitTrE(fp);
  FitOptAE op= vp;
  ParV_t *pv= VecP(ParPar(pp)) + FitPosIdx(peakIdx);

  if (!pv || !ParVSused(pv + FIT_PEAKWIDTH)) { op= NULL; }
  else {
    double sig = ParVVal(pv + FIT_PEAKWIDTH);

    if (sig == 0.0) { FitsetErrNo(FITSIGM0); return NULL; }
    else if (op ||
	     ((op= (FitOptAE)Malloc(sizeof(FitOptAERec))) &&
	      (FitOptAEPar(op)=
	       (ParV_t *)Malloc(FIT_PEAKPAR * sizeof(ParV_t))))) {
      double tl= 0.0, tr= 0.0;
      double vol_unit, amp;

      if (vp) { FitOptAEPar(op)= pv; }
      else { memcpy(FitOptAEPar(op), pv, FIT_PEAKPAR * sizeof(ParV_t)); }

      FitOptAEFct(op)= FitFctAE;

      FitOptAES1(op) = 1.0 / sig;
      FitOptAES2(op) = FitOptAES1(op) * FitOptAES1(op);
      FitOptAES3(op) = FitOptAES2(op) * FitOptAES1(op);
      FitOptAESExp(op) = 0.5 * FitOptAES2(op);

      tl= ParVSused(pv + FIT_PEAKTL) ? ABS(ParVVal(pv + FIT_PEAKTL)) : 0.0;
      tr= ParVSused(pv + FIT_PEAKTR) ? ABS(ParVVal(pv + FIT_PEAKTR)) : 0.0;

      vol_unit= sig * (SQRT_2_PI + tl + tr);
      FitOptAEVolNorm(op)= vol_unit ? (1.0 / vol_unit) : 0.0;
      amp= ParVVal(pv + FIT_PEAKVOL) * FitOptAEVolNorm(op);

      if (ParVSused(pv + FIT_PEAKTL)) {
	double sm= ((sig < 0.0) ?
		    -pow(-sig, -(double)el) : pow(sig, -(double)el));
	double fs= invff[el] * sm;
	FitOptAEPlT(op)= amp * fs;
	FitOptAEPl(op)= fs * tl;
	FitOptAEPlPos(op)= amp * FitOptAEPl(op) * el;
	FitOptAEPlSig(op)= FitOptAEPlPos(op) * FitOptAES1(op);
	FitOptAEEl(op) = el; }
      else { FitOptAEPl(op)= 0.0; }

      if (ParVSused(pv + FIT_PEAKTR)) {
	double sm= ((sig < 0.0) ?
		    -pow(-sig, -(double)er) : pow(sig, -(double)er));
	double fs= invff[er] * sm;
	FitOptAEPrT(op)= amp * fs;
	FitOptAEPr(op)= fs * tr;
	FitOptAEPrPos(op)= amp * FitOptAEPr(op) * er;
	FitOptAEPrSig(op)= FitOptAEPrPos(op) * FitOptAES1(op);
	FitOptAEEr(op) = er; }
      else { FitOptAEPr(op)= 0.0; }

      if (ParVSused(pv + FIT_PEAKSH)) {
	if (ParVVal(pv + FIT_PEAKSW) == 0.0) {
	  FitsetErrNo(FITSW0); return FALSE; }
	else {
	  double sw;
	  FitOptAESwInv(op) = 1.0 / ParVVal(pv + FIT_PEAKSW);
	  sw= FitOptAESwInv(op) * FitOptAES1(op);
	  FitOptAESwp(op)= ((1.0 / SQRT_2_PI) * sw *
			    ParVVal(pv + FIT_PEAKVOL) *
			    ParVVal(pv + FIT_PEAKSH));
	  FitOptAESw1(op)= M_SQRT1_2 * sw;
	  FitOptAESw2(op)= FitOptAESw1(op) * FitOptAESw1(op); } } }
    else { Free(op); FitsetErrNo(FITALLOC); } }
  return op;
}
/* ------------------------------------------------------------------------- */
static FitOptAE AEPeakFunctionFinish(op)
     FitOptAE op;
{
  if (op) { Free(FitOptAEPar(op)); Free(op); }
  return NULL;
}
/*---------------------------------------------------------------------------*/
static int AEDerivationSetup(fp, pp)
     Fit_t fp;
     Par_t pp;
{
  if (fp == NULL) { FitsetErrNo(FITPOINT); return FALSE; }
  if (!AEFunctionSetup(fp, pp)) { return FALSE; }
  else if (!VecAlloc(FitOptGradVec(fp), ParNum(FitPar(fp)))) {
    FitsetErrNo(FITALLOC); return FALSE; }
  else if (!FitOptFitPeaks(fp)) { return TRUE; }
  else { return TRUE; }
}
/*---------------------------------------------------------------------------*/
static int AEDerivationFinish(fp)
     Fit_t fp;
{
  VecFree(FitOptGradVec(fp));
  return AEFunctionFinish(fp);
}
/*---------------------------------------------------------------------------*/
static int AEDerivation(fp, x, grad)
     Fit_t fp;			/* preinitialized values and parameters */
     double x;			/* location to calculate the derivations     */
     double *grad;		/* derivations of fitting function with      */
				/* respect to the fitting parameters         */
{
  Par_t pp= FitOptPp(fp);
  double *grd= FitOptGrad(fp);	/* gradient */

  BgDerivation(fp, grd, x);

  if (FitOptFitPeaks(fp)) {
    int i;
    FitOptAE op;
				/* loop over all peaks */
    for (i= 0, op= (FitOptAE)FitOptPeak(fp); i < FitPeakNum(fp); i++, op++) {
      ParV_t *par= FitOptAEPar(op);
      double dx = x - ParVVal(par + FIT_PEAKPOS); /* peak location */
      double gauss, dxa, dx2, dxm, dxm1;
      int tsi, tails;
      double ps, psT, psPos, psSig, ss, ms;
      double phi, phinorm;	/* peak function */

      if (dx < 0.0) {		/* left region */
	ss= 1.0;
	tsi= FitTlIdx(i);
	if ((tails= ParVSused(par + FIT_PEAKTL))) {
	  ps= FitOptAEPl(op);
	  ms= FitOptAEEl(op);
	  psT= FitOptAEPlT(op);
	  psPos= FitOptAEPlPos(op);
	  psSig= FitOptAEPlSig(op); } }
      else if (dx >= 0.0) {	/* right region */
	ss= -1.0;
	tsi= FitTrIdx(i);
	if ((tails= ParVSused(par + FIT_PEAKTR))) {
	  ps= FitOptAEPr(op);
	  ms= FitOptAEEr(op);
	  psT= FitOptAEPrT(op);
	  psPos= FitOptAEPrPos(op);
	  psSig= FitOptAEPrSig(op); } }

      dxa= ABS(dx);
      dx2= dx * dx;

      gauss= exp(-dx2 * FitOptAESExp(op));
      phinorm= FitOptAEVolNorm(op) * gauss;

      phi= 1.0;

      if (tails) {
	dxm1= pow(dxa, ms - 1.0);
	dxm= dxm1 * dxa;
	phi += dxm * ps; }

      phi *= ParVVal(par + FIT_PEAKVOL) * phinorm;
	
      grd[FitPosIdx(i)]   = -ss * phi * dxa * FitOptAES2(op);
      grd[FitWidthIdx(i)] = phi * (-FitOptAES1(op) + dx2 * FitOptAES3(op));
      grd[FitVolIdx(i)]  = phinorm;
      grd[FitTlIdx(i)]= grd[FitTrIdx(i)]
	= -phi * FitOptAEVolNorm(op) * ParVVal(par + FIT_PEAKWIDTH);

      if (tails) {
	grd[FitPosIdx(i)] += ss * psPos * dxm1 * gauss;
	grd[FitWidthIdx(i)] -= psSig * dxm * gauss;
	grd[tsi] += psT * dxm * gauss; }

      if (ParVSused(par + FIT_PEAKSH)) {
	{ double ec= 0.5 * erfc(dx * FitOptAESw1(op));
	  grd[FitShIdx(i)] = ParVVal(par + FIT_PEAKVOL) * ec;
	  grd[FitVolIdx(i)] += ParVVal(par + FIT_PEAKSH) * ec; }
	{ double sx= ((ParVVal(par + FIT_PEAKSW) == 1.0) ? gauss :
		      exp(-dx2 * FitOptAESw2(op))) * FitOptAESwp(op);
	  grd[FitPosIdx(i)] += sx;
	  sx *= dx;
	  grd[FitSwIdx(i)] = sx * FitOptAESwInv(op);
	  grd[FitWidthIdx(i)] += sx * FitOptAES1(op); } } } }

  { int i;
    for (i=0; i<ParNumFree(pp); i++)  { grad[i] = grd[ParIfree(pp,i)]; } }

  { int i;
    for (i=0; i<ParNumCorr(pp); i++) {
      int j= ParIcorr(pp,i);
      grad[ParIcofr(pp,i)] += grd[j] * ParCorr(pp,j)[1]; } }

  return TRUE;
}
/*---------------------------------------------------------------------------*/
static int AEUnitPeakDerivation(op, x, x0, fct_val, grad_x0)
     FitOptAE op;
     double x;
     double x0;
     double *fct_val;
     double *grad_x0;
{
  if (!op) { return FALSE; }
  else {
    ParV_t *par= FitOptAEPar(op);
    double dx = x - x0;		/* peak location */
    double gauss, dxa, dx2, dxm, dxm1;
    int tails;
    double ps, psPos, ss, ms;
    double phi;			/* peak function */

    if (dx < 0.0) {		/* left region */
      ss= 1.0;
      if ((tails= ParVSused(par + FIT_PEAKTL))) {
	ps= FitOptAEPl(op);
	ms= FitOptAEEl(op);
	psPos= FitOptAEPlPos(op); } }
    else if (dx >= 0.0) {	/* right region */
      ss= -1.0;
      if ((tails= ParVSused(par + FIT_PEAKTR))) {
	ps= FitOptAEPr(op);
	ms= FitOptAEEr(op);
	psPos= FitOptAEPrPos(op); } }

    dxa= ABS(dx);
    dx2= dx * dx;

    gauss= exp(-dx2 * FitOptAESExp(op));

    phi= 1.0;

    if (tails) {
      dxm1= pow(dxa, ms - 1.0);
      dxm= dxm1 * dxa;
      phi = (1.0 + dxm * ps) * gauss; }
    else {
      phi= gauss; }

    *fct_val= phi;
	
    *grad_x0= -ss * phi * dxa * FitOptAES2(op);

    if (tails) {
      *grad_x0 += ss * psPos * dxm1 * gauss; }

    if (ParVSused(par + FIT_PEAKSH)) {
      double sx= ((ParVVal(par + FIT_PEAKSW) == 1.0) ? gauss :
		  exp(-dx2 * FitOptAESw2(op))) * FitOptAESwp(op);
      *grad_x0 += sx; }

    return TRUE; }
}
/*---------------------------------------------------------------------------*/
static double AEFunction(fp, x)
				/* function to calculate the value of fitting*/
				/* function at location x in dependence of   */
				/* the parameters in fp                     */
     Fit_t fp;			/* preinitialized values and parameters */
     double x;			/* is the index of the correlated y value in */
				/* spectrum, could be used for x calibration */
{
  double result;		/* hold actual result */

  result= FitTotalBgFunction(fp, x); /* background */

  if (FitOptFitPeaks(fp)) {
    int i;		/* loop index */
    FitOptAE op;
    for (i= 0, op= (FitOptAE)FitOptPeak(fp); i < FitPeakNum(fp); i++, op++) {
      PEAK_VALUE(x, op, result); } }
  return result;
}
/*---------------------------------------------------------------------------*/
static double AENormVolume(pre)
				/* volume of a 1-Peak (gauss) with exp-tail */
     FitOptAE pre;
{
  ParV_t *par= FitOptAEPar(pre);
  double tl= !ParVSsusp(par + FIT_PEAKTL) ? ParVVal(par + FIT_PEAKTL) : 0.0;
  double tr= !ParVSsusp(par + FIT_PEAKTR) ? ParVVal(par + FIT_PEAKTR) : 0.0;
  return ParVVal(par + FIT_PEAKWIDTH) * (SQRT_2_PI + tl + tr);
}
/*---------------------------------------------------------------------------*/
static double AEStepFunction(fp, x)
				/* function to calculate the value of step   */
				/* function at location x in dependence of   */
				/* the parameters pp                         */
     Fit_t fp;			/* actual parameter set                      */
     double x;			/* is the index of the correlated y value in */
				/* spectrum, could be used for x calibration */
{
  double result= 0.0;		/* hold actual result */
  FitOptAE op;
  int i;			/* loop index */
				/* loop over all peaks */
  for (i= 0, op= (FitOptAE)FitOptPeak(fp); i < FitPeakNum(fp); i++, op++) {
    ParV_t *par= FitOptAEPar(op);

    if (ParVSused(par + FIT_PEAKPOS) && ParVSused(par + FIT_PEAKSH) &&
	ParVUserStat(par + FIT_PEAKPOS) != FIT_USUNDEF &&
	ParVUserStat(par + FIT_PEAKWIDTH) != FIT_USUNDEF &&
	ParVUserStat(par + FIT_PEAKVOL) != FIT_USUNDEF) {	  
      double dx = x - ParVVal(par + FIT_PEAKPOS); /* peak location */
      result += (0.5 * ParVVal(par + FIT_PEAKVOL) *
		 ParVVal(par + FIT_PEAKSH) * erfc(dx * FitOptAESw1(op))); } }
  return result;
}
/*---------------------------------------------------------------------------*/
static double AEPeakFunction(op, x)
     FitOptAE op;		/* parameter (optimized) */
     double  x;			/* x value  */
{
  double result=0.0;		/* hold actual result */
				/* prepare for loop */
  if (op) {
    PEAK_VALUE(x, op, result); }
  return result;
}
/*---------------------------------------------------------------------------*/
static double AEUnitPeakFunction(op, x, x0)
     FitOptAE op;
     double x;
     double x0;
{
  if (op) {
    double dx= x - x0;		/* distance from actual peak */
    double dx2= dx * dx;
    double ps, ms;
    double gauss= exp(-dx2 * FitOptAESExp(op));

    if (dx < 0.0) {		/* left region */
      ps= FitOptAEPl(op); ms= FitOptAEEl(op); }
    else if (dx >= 0.0) {		/* right region */
      ps= FitOptAEPr(op); ms= FitOptAEEr(op); }

    if (ps != 0.0) {
      return ((1.0 + pow(ABS(dx), ms) * ps) * gauss) ; }
    else {
      return gauss; } }
  else {
    return 0.0; }
}
/*---------------------------------------------------------------------------*/
#undef PEAK_VALUE
