/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsFit.h,v 1.5 1994/12/08 20:44:15 vs Exp vs $ */
#ifndef _vsFit_h
#define _vsFit_h

#include <tv/vsConfig.h>
#include <tv/vsTypes.h>
#include <tv/vsPar.h>
				/* bgTyp */
#define FIT_BGPOL	0

				/* peakTyp */
#define FIT_PKCA	0	/* continuous tail, arctan-step */
#define FIT_PKAE	1	/* additive polynom tail, erf-step */

				/* fitTyp in Fit_t */
#define FIT_FTUNDEF	0
#define FIT_FTFIT	1
#define FIT_FTINT	2

				/* measure for extremization */
#define FIT_MDY		0	/* chisquare weighted by deviation*/
#define FIT_MY		1	/* chisquare weighted by value */
#define FIT_MPOI	2	/* poisson */

				/* sortTyp in Fit_t */
#define FIT_STPOS	0
#define FIT_STHIST	1
#define FIT_STVOL	2

				/* userStat in Par_t (parameter)*/
#define FIT_USUNDEF	0	/* not initialized */
#define FIT_USCALUNDEF	1	/* cal not initialized */
#define FIT_USCAL	2	/* calibrated value */
#define FIT_USINIT	3	/* initialized by program */
#define FIT_USUSER	4	/* initialized by user */
#define FIT_USFIT	5	/* set by fit */
#define FIT_USSTART	6	/* existant ? */
#define FIT_USNONE	7	/* not used */

				/* userTyp in Par_t (parameter)*/
#define FIT_UTBGPOL	0	/* background polynom coeff */
#define FIT_UTTL	1	/* tail left */
#define FIT_UTTR	2	/* tail right */
#define FIT_UTSH	3	/* step height */
#define FIT_UTSW	4	/* step width */
#define FIT_UTPOS	5	/* peak position */
#define FIT_UTWIDTH	6	/* peak width */
#define FIT_UTVOL	7	/* peak volume */
#define FIT_UTBGFAC	8	/* background exponent coeff */
#define FIT_UTBGEXP	9	/* background exponent */

				/* parameter inidices in Par_t */
#define FIT_BGGRD	5	/* max number background polynom coeffs */
#define FIT_BGMEXP	3	/* max number of exp. functions */
#define FIT_1STPEAK	(FIT_BGGRD + 2 * FIT_BGMEXP) /* index of first peak parameter */
#define FIT_PEAKPAR	7	/* number of parameter per peak */

				/* indices of peak parameter */
#define FIT_PEAKPOS	0
#define FIT_PEAKWIDTH	1
#define FIT_PEAKVOL	2
#define FIT_PEAKTL	3
#define FIT_PEAKTR	4
#define FIT_PEAKSH	5
#define FIT_PEAKSW	6
				/* numeric constants */
#ifndef M_LN2
#  define M_LN2          0.69314718055994530942	/* ln(2) */
#  define M_PI           3.14159265358979323846
#  define M_SQRT1_2      0.70710678118654752440	/* sqrt(1/2) */
#endif

#define SQRT_2_PI        2.50662827463100020000 /* sqrt(2*PI) */

#define TINY_DIAGONAL	(1E-20)	/*  */
#define TINY_WIDTH	(1E-5)	/* this is the smallest peakwidht of sense   */
#define TINY_STEPWIDTH	(1E-4)
#define TINY_STEPHEIGHT	(1E-10)
#define PRECISION	(1E-10)	/* precision to evaluate the sum in peak_vol*/
#define FORGET_EXP	50	/* value of exponent that does not matter    */
#define FORGET_TAIL	50	/* distance in units of sigma at wich you can
				   shurely ignore the peak contribution */
#define TAIL_TOTAL       0.1    /* nearest possible beginning of the tail    */
				/* to avoid divergenz of peak volume         */

				/* access macros for fit parameter */

				/* access on Par_t */
#define FitPosIdx(i)		(FIT_1STPEAK + FIT_PEAKPOS + FIT_PEAKPAR * i)
#define FitWidthIdx(i)		(FIT_1STPEAK + FIT_PEAKWIDTH + FIT_PEAKPAR * i)
#define FitVolIdx(i)		(FIT_1STPEAK + FIT_PEAKVOL + FIT_PEAKPAR * i)
#define FitTlIdx(i)		(FIT_1STPEAK + FIT_PEAKTL + FIT_PEAKPAR * i)
#define FitTrIdx(i)		(FIT_1STPEAK + FIT_PEAKTR + FIT_PEAKPAR * i)
#define FitShIdx(i)		(FIT_1STPEAK + FIT_PEAKSH + FIT_PEAKPAR * i)
#define FitSwIdx(i)		(FIT_1STPEAK + FIT_PEAKSW + FIT_PEAKPAR * i)

#define FitPeakIdx(i)		(int)((i - FIT_1STPEAK) / FIT_PEAKPAR)

#define FitPBgCoef(pp,i)	ParV(pp,i)

#define FitPBgCoefVal(pp,i)	ParVal(pp,i)

#define FitPPos(pp,i)		ParV(pp,FitPosIdx(i))
#define FitPWidth(pp,i)		ParV(pp,FitWidthIdx(i))
#define FitPVol(pp,i)		ParV(pp,FitVolIdx(i))
#define FitPTl(pp,i)		ParV(pp,FitTlIdx(i))
#define FitPTr(pp,i)		ParV(pp,FitTrIdx(i))
#define FitPSh(pp,i)		ParV(pp,FitShIdx(i))
#define FitPSw(pp,i)		ParV(pp,FitSwIdx(i))

#define FitPPosVal(pp,i)	ParVal(pp,FitPosIdx(i))
#define FitPWidthVal(pp,i)	ParVal(pp,FitWidthIdx(i))
#define FitPVolVal(pp,i)	ParVal(pp,FitVolIdx(i))
#define FitPTlVal(pp,i)		ParVal(pp,FitTlIdx(i))
#define FitPTrVal(pp,i)		ParVal(pp,FitTrIdx(i))
#define FitPShVal(pp,i)		ParVal(pp,FitShIdx(i))
#define FitPSwVal(pp,i)		ParVal(pp,FitSwIdx(i))

#define FitPSetPeakStat(pp,i,s) \
{\
  ParVStat(pp,FitPosIdx(i)) = s;\
  ParVStat(pp,FitWidthIdx(i)) = s;\
  ParVStat(pp,FitVolIdx(i)) = s;\
  ParVStat(pp,FitTlIdx(i)) = s;\
  ParVStat(pp,FitTrIdx(i)) = s;\
  ParVStat(pp,FitShIdx(i)) = s;\
  ParVStat(pp,FitSwIdx(i)) = s;\
}
				/* access on Fit_t */
#define FitBgCoef(fp,i)		ParV(FitPar(fp),i)

#define FitBgCoefVal(fp,i)	ParVal(FitPar(fp),i)

#define FitPos(fp,i)		FitPPos(FitPar(fp),i)
#define FitWidth(fp,i)		FitPWidth(FitPar(fp),i)
#define FitVol(fp,i)		FitPVol(FitPar(fp),i)
#define FitTl(fp,i)		FitPTl(FitPar(fp),i)
#define FitTr(fp,i)		FitPTr(FitPar(fp),i)
#define FitSh(fp,i)		FitPSh(FitPar(fp),i)
#define FitSw(fp,i)		FitPSw(FitPar(fp),i)

#define FitPosVal(fp,i)		FitPPosVal(FitPar(fp),i)
#define FitWidthVal(fp,i)	FitPWidthVal(FitPar(fp),i)
#define FitVolVal(fp,i)		FitPVolVal(FitPar(fp),i)
#define FitTlVal(fp,i)		FitPTlVal(FitPar(fp),i)
#define FitTrVal(fp,i)		FitPTrVal(FitPar(fp),i)
#define FitShVal(fp,i)		FitPShVal(FitPar(fp),i)
#define FitSwVal(fp,i)		FitPSwVal(FitPar(fp),i)


				/* access on sorted peaks */
#define FitPeakSIdx(fp,i)	(VecP(FitPeakSortIdx(fp))[i])
#define FitPosSIdx(fp,i)	FitPosIdx(FitPeakSIdx(fp,i))
#define FitWidthSIdx(fp,i)	FitWidthIdx(FitPeakSIdx(fp,i))
#define FitVolSIdx(fp,i)	FitVolIdx(FitPeakSIdx(fp,i))
#define FitTlSIdx(fp,i)		FitTlIdx(FitPeakSIdx(fp,i))
#define FitTrSIdx(fp,i)		FitTrIdx(FitPeakSIdx(fp,i))
#define FitShSIdx(fp,i)		FitShIdx(FitPeakSIdx(fp,i))
#define FitSwSIdx(fp,i)		FitSwIdx(FitPeakSIdx(fp,i))

#define FitPeakSRIdx(fp,i)	(VecP(FitPeakSortRevIdx(fp))[i])
#define FitPosSRIdx(fp,i)	FitPosIdx(FitPeakSRIdx(fp,i))
#define FitWidthSRIdx(fp,i)	FitWidthIdx(FitPeakSRIdx(fp,i))
#define FitVolSRIdx(fp,i)	FitVolIdx(FitPeakSRIdx(fp,i))
#define FitTlSRIdx(fp,i)	FitTlIdx(FitPeakSRIdx(fp,i))
#define FitTrSRIdx(fp,i)	FitTrIdx(FitPeakSRIdx(fp,i))
#define FitShSRIdx(fp,i)	FitShIdx(FitPeakSRIdx(fp,i))
#define FitSwSRIdx(fp,i)	FitSwIdx(FitPeakSRIdx(fp,i))

#define FitSortPos(fp,i)	ParV(FitPar(fp),FitPosSIdx(fp,i))
#define FitSortWidth(fp,i)	ParV(FitPar(fp),FitWidthSIdx(fp,i))
#define FitSortVol(fp,i)	ParV(FitPar(fp),FitVolSIdx(fp,i))
#define FitSortTl(fp,i)		ParV(FitPar(fp),FitTlSIdx(fp,i))
#define FitSortTr(fp,i)		ParV(FitPar(fp),FitTrSIdx(fp,i))
#define FitSortSh(fp,i)		ParV(FitPar(fp),FitShSIdx(fp,i))
#define FitSortSw(fp,i)		ParV(FitPar(fp),FitSwSIdx(fp,i))

#define FIT_PEAK	128
#define FIT_PARM	512

typedef struct {		/* integral struct */
  double i;			/* integral */
  double dI;			/* delta ... */
  double m1;			/* 1st moment, center of mass */
  double dM1;
  double m2;			/* 2nd moment, variance */
  double dM2;
  double m3;			/* 3rd moment, skewness */
  double dM3;
} FitInt_t;

#define FitIntI(fip)		((fip)->i)
#define FitIntDI(fip)		((fip)->dI)
#define FitIntM1(fip)		((fip)->m1)
#define FitIntDM1(fip)		((fip)->dM1)
#define FitIntM2(fip)		((fip)->m2)
#define FitIntDM2(fip)		((fip)->dM2)
#define FitIntM3(fip)		((fip)->m3)
#define FitIntDM3(fip)		((fip)->dM3)

typedef struct {		/* FitOpt_t basics */
  double offset;		/* offset for numeric stability */
  int fitPeaks;			/* fit any peaks */
  int fitBg;			/* fit bg */
} FitOptB_t;

typedef struct {
  int (*functionSetup)();
  int (*functionFinish)();
  FitOpt (*peakFunctionSetup)();
  FitOpt (*peakFunctionFinish)();
  int (*derivationSetup)();
  int (*derivationFinish)();
  int (*derivation)();
  double (*function)();
  double (*normVol)();
  double (*stepFunction)();
  double (*peakFunction)();
  double (*unitPeakFunction)();
  FitOpt (*unitPeakDerivationSetup)();
  FitOpt (*unitPeakDerivationFinish)();
  int (*unitPeakDerivation)();
} FitFctRec, *FitFct_t;

#define FitFctFunctionSetup(o)		((o)->functionSetup)
#define FitFctFunctionFinish(o)		((o)->functionFinish)
#define FitFctPeakFunctionSetup(o)	((o)->peakFunctionSetup)
#define FitFctPeakFunctionFinish(o)	((o)->peakFunctionFinish)
#define FitFctDerivationSetup(o)	((o)->derivationSetup)
#define FitFctDerivationFinish(o)	((o)->derivationFinish)
#define FitFctDerivation(o)		((o)->derivation)
#define FitFctFunction(o)		((o)->function)
#define FitFctNormVol(o)		((o)->normVol)
#define FitFctStepFunction(o)		((o)->stepFunction)
#define FitFctPeakFunction(o)		((o)->peakFunction)
#define FitFctUnitPeakFunction(o)	((o)->unitPeakFunction)
#define FitFctUnitPeakDerivationSetup(o) \
  ((o)->unitPeakDerivationSetup)
#define FitFctUnitPeakDerivationFinish(o) \
  ((o)->unitPeakDerivationFinish)
#define FitFctUnitPeakDerivation(o)	((o)->unitPeakDerivation)

typedef struct {
  FitFct_t fct;			/* fit function class rec */
  ParV_t *par;			/* peak parameter subvector */
} FitOptCoreRec, *FitOptCore;

typedef struct _FitOptRec {
  FitOptCoreRec core;
} FitOptRec;

#define FitOptFct(o)			((o)->core.fct)
#define FitOptFunctionSetup(o)		FitFctFunctionSetup(FitOptFct(o))
#define FitOptFunctionFinish(o)		FitFctFunctionFinish(FitOptFct(o))
#define FitOptPeakFunctionSetup(o)	FitFctPeakFunctionSetup(FitOptFct(o))
#define FitOptPeakFunctionFinish(o)	FitFctPeakFunctionFinish(FitOptFct(o))
#define FitOptDerivationSetup(o)	FitFctDerivationSetup(FitOptFct(o))
#define FitOptDerivationFinish(o)	FitFctDerivationFinish(FitOptFct(o))
#define FitOptDerivation(o)		FitFctDerivation(FitOptFct(o))
#define FitOptFunction(o)		FitFctFunction(FitOptFct(o))
#define FitOptNormVol(o)		FitFctNormVol(FitOptFct(o))
#define FitOptStepFunction(o)		FitFctStepFunction(FitOptFct(o))
#define FitOptPeakFunction(o)		FitFctPeakFunction(FitOptFct(o))
#define FitOptUnitPeakFunction(o)	FitFctUnitPeakFunction(FitOptFct(o))
#define FitOptUnitPeakDerivationSetup(o) \
  FitFctUnitPeakDerivationSetup(FitOptFct(o))
#define FitOptUnitPeakDerivationFinish(o) \
  FitFctUnitPeakDerivationFinish(FitOptFct(o))
#define FitOptUnitPeakDerivation(o)	FitFctUnitPeakDerivation(FitOptFct(o))
#define FitOptPar(o)			((o)->core.par)

typedef struct {
  double slM;			/* sigma weight of tailleft */
  double srM;			/* sigma weight of tailright */
  double ss;			/* 1.0 / (sigma * sigma) */
  double sl;			/* weighted left tail */
  double sr;			/* weighted right tail */
  double expSl;			/* left tail exp-factor */
  double expSr;			/* right tail exp-factor */
  double volNorm;		/* 1./(peak volume for unity-amplitude) */
  double atan;			/* atan-arg */
  ParV_t *par;			/* peak parameter subvector */
} FitOptCAPartRec;

typedef struct {
  FitOptCoreRec core;
  FitOptCAPartRec ca;
} FitOptCARec, *FitOptCA;

#define FitOptCAFct(p)		((p)->core.fct)
#define FitOptCAPar(p)		((p)->core.par)
#define FitOptCASlM(p)		((p)->ca.slM)
#define FitOptCASrM(p)		((p)->ca.srM)
#define FitOptCASs(p)		((p)->ca.ss)
#define FitOptCASl(p)		((p)->ca.sl)
#define FitOptCASr(p)		((p)->ca.sr)
#define FitOptCAExpSl(p)	((p)->ca.expSl)
#define FitOptCAExpSr(p)	((p)->ca.expSr)
#define FitOptCAVolNorm(p)	((p)->ca.volNorm)
#define FitOptCAAtan(p)		((p)->ca.atan)

typedef struct {
  int el;
  int er;
  double sExp;			/* 1.0 / (2.0 * sigma**2) */
  double s1;			/* 1.0 / sigma**1 */
  double s2;			/* 1.0 / sigma**2 */
  double s3;			/* 1.0 / sigma**3 */
				/* step components */
  double swInv;			/* 1.0 / sw */
  double sw1;
  double sw2;
  double swp;
				/* left tail components */
  double pl;			/* tail volume factor in peak function */
  double plT;			/* tail derivation factor*/
  double plPos;			/* pos derivation factor */
  double plSig;			/* sigma derivation factor */
  double pr;			/* same for right tail components*/
  double prT;
  double prPos;
  double prSig;
  double volNorm;		/* 1.0 / (peak vol for unity-amplitude) */
} FitOptAEPartRec;

typedef struct {
  FitOptCoreRec core;
  FitOptAEPartRec ae;
} FitOptAERec, *FitOptAE;

#define FitOptAEFct(p)		((p)->core.fct)
#define FitOptAEPar(p)		((p)->core.par)
#define FitOptAEEl(p)		((p)->ae.el)
#define FitOptAEEr(p)		((p)->ae.er)
#define FitOptAESExp(p)		((p)->ae.sExp)
#define FitOptAES1(p)		((p)->ae.s1)
#define FitOptAES2(p)		((p)->ae.s2)
#define FitOptAES3(p)		((p)->ae.s3)
#define FitOptAESwInv(p)	((p)->ae.swInv)
#define FitOptAESw1(p)		((p)->ae.sw1)
#define FitOptAESw2(p)		((p)->ae.sw2)
#define FitOptAESwp(p)		((p)->ae.swp)
#define FitOptAEPl(p)		((p)->ae.pl)
#define FitOptAEPlT(p)		((p)->ae.plT)
#define FitOptAEPlPos(p)	((p)->ae.plPos)
#define FitOptAEPlSig(p)	((p)->ae.plSig)
#define FitOptAEPr(p)		((p)->ae.pr)
#define FitOptAEPrT(p)		((p)->ae.prT)
#define FitOptAEPrPos(p)	((p)->ae.prPos)
#define FitOptAEPrSig(p)	((p)->ae.prSig)
#define FitOptAEVolNorm(p)	((p)->ae.volNorm)

typedef struct {		/* struct for PreFunction and PreDerivation */
  FitOptB_t b;
  Par_t pp;
  void *peak;
  DVecRec grad;			/* derivations */
} FitOpt_t;

#define FitOptB(fp)		(&FitParOpt(fp)->b)
#define FitOptOffset(fp)	(FitOptB(fp)->offset)
#define FitOptFitPeaks(fp)	(FitOptB(fp)->fitPeaks)
#define FitOptFitBg(fp)		(FitOptB(fp)->fitBg)

#define FitOptPp(fp)		(FitParOpt(fp)->pp)

#define FitOptGradVec(fp)	(&FitParOpt(fp)->grad)

#define FitOptPeak(fp)		(FitParOpt(fp)->peak)
#define FitOptGrad(fp)		VecP(FitOptGradVec(fp))

typedef struct {		/* FitS_t fit status */
  time_t time;			/* date of fit */

  double measure;		/* measure of fit */
  double dMeasure;		/* dmeasure of fit */
  double rms;			/* RMS src: paper winter (not used) */

  int maxIterations;		/* max. number of iterations */
  int iterations;		/* number of iterations */

  int measureTyp;		/* used measure for extremization */

  int freedom;			/* degrees of freedom in fit */

  double lambda;		/* lambda of curfit */
  double minLambda;		/* minimal lambda of curfit */
  double maxLambda;		/* maximal lambda of curfit */
  double startLambda;		/* start lambda for fit */
} FitStat_t;

#define FitStatTime(p)		((p)->time)
#define FitStatMeasure(p)	((p)->measure)
#define FitStatDMeasure(p)	((p)->dMeasure)
#define FitStatRms(p)		((p)->rms)
#define FitStatMaxIterations(p)	((p)->maxIterations)
#define FitStatIterations(p)	((p)->iterations)
#define FitStatMeasureTyp(p)	((p)->measureTyp)
#define FitStatFreedom(p)	((p)->freedom)
#define FitStatLambda(p)	((p)->lambda)
#define FitStatMinLambda(p)	((p)->minLambda)
#define FitStatMaxLambda(p)	((p)->maxLambda)
#define FitStatStartLambda(p)	((p)->startLambda)

typedef struct {		/* Fit_t basics */
  int fitTyp;			/* fit type */

  FitFct_t fct;			/* fit function class rec */

  int res;			/* resolution of fitted spectrum */

  int bgTyp;			/* background type */
  int bgGrad;			/* grad+1 of background polynom */
  int bgExpNum;			/* number of exponentials in background */
  double bgOffset;		/* background polynom relativ to offset */

  int wCorr;			/* widths correlated to index */
  int pCorr;			/* positions correlated to index */
  int tlCorr;			/* taillefts correlated to index */
  int trCorr;			/* tailrights correlated to index */
  int shCorr;			/* stepheights correlated to index */
  int swCorr;			/* stepwidths correlated to index */

  int peakTyp;			/* peak typ */

  int tlE;			/* ^tlE dependence of tailleft */
  int trE;			/* ^trE dependence of tailright */

  int peakNum;			/* number of peaks */
  int peakSortTyp;		/* sort criterion for peakSortIdx */
} FitB_t;

typedef struct _FitRec {	/* fit parameter struct */
  FitB_t b;

  FitStat_t bg;			/* fit status background */
  FitStat_t reg;		/* fit status region */

  FitInt_t integral;		/* total integral */
  FitInt_t bgIntegral;		/* background integral */
  FitInt_t subIntegral;		/* background subtracted integral */

  ParRec par;			/* parameter vector */

  FitOpt_t parOpt;		/* auxilary data for function evaluation */

  PVecRec regMrk;		/* fit region marker */
  PVecRec bgMrk;		/* background region marker */

  DVecRec bgCovar;		/* background covariance matrix */

  IVecRec peakSortIdx;		/* peak indices sorted on sortTyp */
  IVecRec peakSortRevIdx;	/* peak indices sorted on sortTyp */
} FitRec;

				/* access macros for Fit_t */
#define FitB(fp)		(&(fp)->b)

#define FitFct(fp)		(FitB(fp)->fct)
#define FitFitTyp(fp)		(FitB(fp)->fitTyp)

#define FitRes(fp)		(FitB(fp)->res)

#define FitBgTyp(fp)		(FitB(fp)->bgTyp)
#define FitBgGrad(fp)		(FitB(fp)->bgGrad)
#define FitBgExpNum(fp)		(FitB(fp)->bgExpNum)
#define FitBgOffset(fp)		(FitB(fp)->bgOffset)

#define FitWCorr(fp)		(FitB(fp)->wCorr)
#define FitPCorr(fp)		(FitB(fp)->pCorr)
#define FitTlCorr(fp)		(FitB(fp)->tlCorr)
#define FitTrCorr(fp)		(FitB(fp)->trCorr)
#define FitShCorr(fp)		(FitB(fp)->shCorr)
#define FitSwCorr(fp)		(FitB(fp)->swCorr)

#define FitPeakTyp(fp)		(FitB(fp)->peakTyp)

#define FitTlE(fp)		(FitB(fp)->trE)
#define FitTrE(fp)		(FitB(fp)->tlE)

#define FitPeakNum(fp)		(FitB(fp)->peakNum)
#define FitPeakSortTyp(fp)	(FitB(fp)->peakSortTyp)

#define FitStatBg(fp)		(&(fp)->bg)
#define FitStatReg(fp)		(&(fp)->reg)

#define FitIntegral(fp)		(&(fp)->integral)
#define FitBgIntegral(fp)	(&(fp)->bgIntegral)
#define FitSubIntegral(fp)	(&(fp)->subIntegral)

#define FitPar(fp)		(&(fp)->par)

#define FitParOpt(fp)		(&(fp)->parOpt)

#define FitRegMrk(fp)		(&(fp)->regMrk)
#define FitBgMrk(fp)		(&(fp)->bgMrk)

#define FitBgCovar(fp)		(&(fp)->bgCovar)

#define FitPeakSortIdx(fp)	(&(fp)->peakSortIdx)
#define FitPeakSortRevIdx(fp)	(&(fp)->peakSortRevIdx)

#define FitBgTime(fp)		FitStatTime(FitStatBg(fp))
#define FitTime(fp)		FitStatTime(FitStatReg(fp))

				/* error codes for FitErrNo */
#define FITNOERR	0 /* vsFit: no error */
#define FITPOINT	1 /* vsFit: illegal NULL-pointer */
#define FITSIGM0	2 /* vsFit: illegal peak width == 0.0 */
#define FITPEAK		3 /* vsFit: maximal peak number exceeded */
#define FITPARM		4 /* vsFit: maximal parameter number exceeded */
#define FITILLST	5 /* vsFit: illegal sort type */
#define FIT0AMPL	6 /* vsFit: no peak with non-zero amplitude */
#define FITALLOC	7 /* "vsFit: can not allocate temporary arrays" */
#define FITPKTYP	8 /* "vsFit: illegal peak function typ", */
#define FITBGTYP	9 /* "vsFit: illegal background function typ", */
#define FITSW0	       10 /* "vsFit: illegal step width == 0.0", */

#define FitsetErrNo(e)	{ FitErrNo= e; }

#ifdef _vsFit_c
int FitErrNo= FITNOERR;
#else
extern int FitErrNo;
extern FitFct_t FitFctCA;
extern FitFct_t FitFctAE;
#endif

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

#include <tv/vsFitF.h>

#undef P_

#endif				/* _vsFit_h */
