/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsFitFmt.c,v 1.4 1994/12/02 21:22:37 vs Exp vs $ */
#define _vsFitFmt_c

#define exp_10(x) (pow ((double)10,(double)(x)))

#include "vsTypes.h"

#include "vsAlloc.h"
#include "vsCal.h"
#include "vsFit.h"
#include "vsPath.h"
#include "vsSearch.h"
#include "vsUserFace.h"
/* ------------------------------------------------------------------------- */
typedef struct {
  double (*p)();
  double (*w)();
  double (*d)();
  double (*e)();
  char *(*u)();
  double (*P)();
  double (*W)();
  double (*D)();
  double (*E)();
  char *(*U)();
  Cal_t cal;
  char *file;
  int idx;
} FmtArgsRec, *FmtArgs;
/* ---------------------------------------------------------------------- */
#include "vsFitFmt.h"
#include "vsFitFmtP.h"
/* ------------------------------------------------------------------------- */
#define p_(cf,pp)	((cf)->p(cf->cal, pp))
#define w_(cf,pp,ww)	((cf)->w(cf->cal, pp, ww))
#define d_(cf,pp,ww,dd)	((cf)->d(cf->cal, pp, ww, dd))
#define e_(cf,pp)	((cf)->e(cf->cal, pp))
#define u_(cf)		((cf)->u(cf->cal))
#define P_(cf,pp)	((cf)->P(cf->cal, pp))
#define W_(cf,pp,ww)	((cf)->W(cf->cal, pp, ww))
#define D_(cf,pp,ww,dd)	((cf)->D(cf->cal, pp, ww, dd))
#define E_(cf,pp)	((cf)->E(cf->cal, pp))
#define U_(cf)		((cf)->U(cf->cal))
/* ====================================================================== */
/* Format i/o -- error handling */
/* ====================================================================== */
char *FmtError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* FMTNOERR	*/ "vsFitFmt: no error",
  /* FMTILLDBL	*/ "vsFitFmt: illegal double format specified",
  /* FMTILLINT	*/ "vsFitFmt: illegal integer format specified",
  /* FMTILLSTR	*/ "vsFitFmt: illegal string format specified",
  /* FMTILLCHR	*/ "vsFitFmt: illegal character found in format",
  /* FMTSETUP	*/ "vsFitFmt: can not setup fit function",
  /* FMTNOQUO	*/ "vsFitFmt: no quoted string found",
  /* FMTOVER	*/ "vsFitFmt: internal overflow during format scan",
  /* FMTOPEN	*/ "vsFitFmt: can not open format file",
  /* FMTPOINT:  */ "vsFitFmt: illegal NULL-pointer",
  /* FMTALLOC:  */ "vsFitFmt: can not allocate format string",
  /* default:   */ "vsFitFmt: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ------------------------------------------------------------------------- */
int FmtRead(fmt, file)
     char **fmt;
     char *file;
{
  FILE *fp;

  if ((fp= fopen(file, "r"))) {
    char item[IOBUFLEN], *f= item;
    int fl= IOBUFLEN;

    while (fgets(f, fl, fp)) {
      int l= strlen(f);
      f += l; fl -= l; }
    if (f > item && f[-1] == '\n') { f[-1]= '\0'; }
    fclose(fp);
    return FmtEnter(fmt, item, TRUE); }
  else { FmtsetErrNo(FMTOPEN); return FALSE; }
}
/* ------------------------------------------------------------------------- */
int FmtEnter(old, new, unquote)
     char **old;
     char *new;
     int unquote;
{
  char fmt[IOBUFLEN], *f= fmt;
  int fl= IOBUFLEN;

  *fmt= '\0';

  while (*new && fl > 1 && UFUnquoteChar(&new, &f, &fl, '\0')) { ; }
  if ((f= Strdup(fmt))) {
    if (old) { Free(*old); *old= f; }
    return TRUE; }
  else { FmtsetErrNo(FMTALLOC); return FALSE; }
}
/* ------------------------------------------------------------------------- */
int FmtWrite(fmt, file)
     char *fmt, *file;
{
  FILE *fp;
  if (!fmt) { FmtsetErrNo(FMTPOINT); return FALSE; }
  else if ((fp= fopen(file, "w"))) {
    char item[IOBUFLEN];
    char *i= item;
    int il= IOBUFLEN;

    *item= '\0';
    while (*fmt) {
      switch (*fmt) {
      case '\'':
	*i++= *fmt++; il--; *i= '\0';
	while (*fmt && il > 1 && (*fmt != '\'' || fmt[-1] == '\\') &&
	       UFQuoteChar(*fmt, &i, &il, '\'')) { fmt++; }
	if (il > 1) { *i++ = *fmt++; il--; *i= '\0'; }
	break;
      case '\n': fmt++;
	fputs(item, fp);
	i= item; il= IOBUFLEN; *item= '\0';
	break;
      default: *i++ = *fmt++; il--; *i= '\0'; } }

    fputs(item, fp);
    return (fclose(fp) != EOF); }
  else { FmtsetErrNo(FMTOPEN); return FALSE; }
}
/* ------------------------------------------------------------------------- */
char *FmtParseFit(pfct, fmt, fp, idx, file, cal,
		  data_calibrated, output_calibrated)
     FmtPrintFunc pfct;
     char **fmt;
     Fit_t fp;
     char *file;
     int idx;
     Cal_t cal;
     int data_calibrated, output_calibrated;
{
  /*  printf("FmtParseFit %d\n", data_calibrated);*/
  if (fmt && fp) {
    FmtArgsRec args;

    InitArgs(&args, file, idx, cal, data_calibrated, output_calibrated);

    switch (FitFitTyp(fp)) {
    case FIT_FTINT:
      return ParseIntegration(pfct, fmt, &args, fp);
    case FIT_FTFIT:
      return ParseFit(pfct, fmt, &args, fp);
    }
  }
  return NULL;
}
/* ------------------------------------------------------------------------- */
char *FmtParseFitPeak(pfct, fmt, fp, peakIdx, idx, file, cal,
		      data_calibrated, output_calibrated)
     FmtPrintFunc pfct;
     char *fmt;
     Fit_t fp;
     int peakIdx;
     char *file;
     int idx;
     Cal_t cal;
     int data_calibrated, output_calibrated;
{
  if (fmt && fp) {
    FmtArgsRec args;

    InitArgs(&args, file, idx, cal, data_calibrated, output_calibrated);

    return ParseFitPeak(pfct, fmt, &args, fp, peakIdx); }
  return NULL;
}
/* ------------------------------------------------------------------------- */
char *FmtParseSea(pfct, fmt, sp, idx, file, cal,
		  data_calibrated, output_calibrated)
     FmtPrintFunc pfct;
     char *fmt;
     SeaPeak sp;
     char *file;
     int idx;
     Cal_t cal;
     int data_calibrated, output_calibrated;
{
  if (fmt && sp) {
    FmtArgsRec args;

    InitArgs(&args, file, idx, cal, data_calibrated, output_calibrated);

    return ParseSeaPeak(pfct, fmt, &args, sp); }
  return NULL;
}
/* ====================================================================== */
/* Fit_t integration output */
/* ====================================================================== */
static char *ParseSeaPeak(pfct, fmt, args, sp)
     FmtPrintFunc pfct;
     char *fmt;
     FmtArgs args;
     SeaPeak sp;
{
  char *f= fmt;
  char *ret= NULL;
  double pos= SeaPos(sp);

  while (*f && !ret) {
    switch (*f) {
    case 'i': ret= PrintFormInt(pfct, &f, args->idx); break;
    case 'O': ret= PrintFormString(pfct, &f, "search"); break;
    case 'a':			/* amplitude */
      f++;
      ret= ParseValErr(pfct, &f, SeaAmpl(sp) * e_(args, pos),
		       SeaDAmpl(sp) * e_(args, pos), NULL);
      break;
    case 'A':			/* amplitude */
      f++;
      ret= ParseValErr(pfct, &f, SeaAmpl(sp) * E_(args, pos),
		       SeaDAmpl(sp) * E_(args, pos), NULL);
      break;
    case 'p':			/* position */
      f++;
      ret= PrintFormDouble(pfct, &f, p_(args, pos));
      break;
    case 'P':			/* position */
      f++;
      ret= PrintFormDouble(pfct, &f, P_(args, pos));
      break;
    case 'v':			/* volume */
      f++;
      ret= ParseValErr(pfct, &f, SeaVol(sp) * e_(args, pos),
		       SeaDVol(sp) * e_(args, pos), NULL);
      break;
    case 'V':			/* volume */
      f++;
      ret= ParseValErr(pfct, &f, SeaVol(sp) * E_(args, pos),
		       SeaDVol(sp) * E_(args, pos), NULL);
      break;
    case 'w':			/* width */
      f++;
      ret= PrintFormDouble(pfct, &f,
			   w_(args, pos, SeaWidth(sp)) * SIGMA_TO_FWHM);
      break;
    case 'W':			/* width */
      f++;
      ret= PrintFormDouble(pfct, &f,
			   W_(args, pos, SeaWidth(sp)) * SIGMA_TO_FWHM);
      break;
    case 'f': f++; ret= ParseFile(pfct, args, &f); break;
    case 'I': f++; ret= PrintFormInt(pfct, &f, args->idx); break;
    case 'M': f++; ret= PrintFormDouble(pfct, &f, SeaChi(sp)); break;
    case 'u': f++; ret= PrintFormString(pfct, &f, u_(args)); break;
    case 'U': f++; ret= PrintFormString(pfct, &f, U_(args)); break;
    case '%': f++; ret= PrintFormDouble(pfct, &f, SeaProb(sp)); break;
    case '\'': ret= PrintQuotedItem(pfct, &f); break;
    default: f++; break; } }	/* ignore */
  return ret;
}
/* ====================================================================== */
/* Fit_t integration output */
/* ====================================================================== */
static char *ParseIntegration(pfct, fmt, args, fp)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     Fit_t fp;
{
  char *f;
  char *ret= NULL;

  if ((f= fmt[FMT_INT])) {
    while (*f && !ret) {
      switch (*f) {
				/* private */
      case 'b': ret= ParseBgPol(pfct, fmt, fp); f++; break;
      case 'e': ret= ParseBgExp(pfct, fmt, fp); f++; break;
      case 'g': f++; ret= PrintFormInt(pfct, &f, FitBgGrad(fp)); break;
      case 'm':	f++; ret= ParseMark(pfct, fmt, args, &f, fp); break;
      case 'n': f++; ret= ParseMarkNum(pfct, &f, fp); break;
      case 'o': f++; ret= PrintFormDouble(pfct, &f, FitBgOffset(fp)); break;
      case 'i':	f++; ret= ParseIntPeaks(pfct, fmt, args, &f, fp); break;
				/* common */
      default: ret= ParseIntCommon(pfct, args, &f, fp); } } }
  return ret;
}
/* ---------------------------------------------------------------------- */
static char *ParseIntCommon(pfct, args, f, fp)
     FmtPrintFunc pfct;
     FmtArgs args;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'D': (*f)++; return ParseDate(pfct, f, fp);
  case 'f': (*f)++; return ParseFile(pfct, args, f);
  case 'G': (*f)++; return PrintFreedom(pfct, f, FitStatBg(fp));
  case 'I': (*f)++; return PrintFormInt(pfct, f, args->idx);
  case 'M': (*f)++; return ParseMeasure(pfct, f, fp);
  case 'N': (*f)++; return ParseIteration(pfct, f, fp);
  case 'u': (*f)++; return PrintFormString(pfct, f, u_(args));
  case 'U': (*f)++; return PrintFormString(pfct, f, U_(args));
  case '\'': return PrintQuotedItem(pfct, f);
  default: (*f)++; return NULL; } /* ignore */
}
/* ---------------------------------------------------------------------- */
static char *ParseIntPeaks(pfct, fmt, args, f, fp)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++;
    if (!VecNum(FitBgMrk(fp))) { return NULL; }
    return ParseIntPeak(pfct, fmt, args, fp, FitBgIntegral(fp), "background");
  case 's': (*f)++;
    if (!VecNum(FitBgMrk(fp))) { return NULL; }
    return ParseIntPeak(pfct, fmt, args, fp, FitSubIntegral(fp), "subtracted");
  case 't': (*f)++;
    return ParseIntPeak(pfct, fmt, args, fp, FitIntegral(fp), "total");
  default: FmtsetErrNo(FMTILLCHR); return *f; } /* ignore */
}
/* ---------------------------------------------------------------------- */
static char *ParseIntPeak(pfct, fmt, args, fp, ip, title)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     Fit_t fp;
     FitInt_t *ip;
     char *title;
{
  char *ret= NULL;

  char *f= fmt[FMT_PKI];
  double pos= FitIntM1(ip);

  while (*f && !ret) {
    switch (*f) {
    case 'i': ret= PrintFormString(pfct, &f, title); break;
    case 'O': ret= PrintFormString(pfct, &f, "integration"); break;
    case 'p':		/* position */
      f++;
      ret= ParseValErr(pfct, &f,
		       p_(args, pos), w_(args, pos, FitIntDM1(ip)), NULL);
      break;
    case 'P':		/* position */
      f++;
      ret= ParseValErr(pfct, &f,
		       P_(args, pos), W_(args, pos, FitIntDM1(ip)), NULL);
      break;
    case 's':		/* skewness */
      f++;
      ret= ParseValErr(pfct, &f, FitIntM3(ip), FitIntDM3(ip), NULL);
      break;
    case 'v':		/* volume */
      f++;
      ret= ParseValErr(pfct, &f, FitIntI(ip) * e_(args, pos),
		       FitIntDI(ip) * e_(args, pos), NULL);
      break;
    case 'V':		/* volume */
      f++;
      ret= ParseValErr(pfct, &f, FitIntI(ip) * E_(args, pos),
		       FitIntDI(ip) * E_(args, pos), NULL);
      break;
    case 'w':		/* width */
      f++;
      { double sigma= w_(args, FitIntM1(ip), FitIntM2(ip));
	double dSigma= d_(args, FitIntM1(ip), FitIntM2(ip), FitIntDM2(ip));

	sigma= (sigma > 0.0) ? sqrt(sigma) : -sqrt(-sigma);
	dSigma= (sigma != 0.0) ? dSigma / ABS(sigma) : 0;
	ret= ParseValErr(pfct, &f, sigma * SIGMA_TO_FWHM,
			 dSigma * SIGMA_TO_FWHM, NULL); }
      break;
    case 'W':		/* width */
      f++;
      { double sigma= W_(args, FitIntM1(ip), FitIntM2(ip));
	double dSigma= D_(args, FitIntM1(ip), FitIntM2(ip), FitIntDM2(ip));

	sigma= (sigma > 0.0) ? sqrt(sigma) : -sqrt(-sigma);
	dSigma= (sigma != 0.0) ? dSigma / ABS(sigma) : 0;
	ret= ParseValErr(pfct, &f, sigma * SIGMA_TO_FWHM,
			 dSigma * SIGMA_TO_FWHM, NULL); }
      break;
    default: ret= ParseIntCommon(pfct, args, &f, fp); break; } }
  return ret;
}
/* ====================================================================== */
/* Fit_t fit output */
/* ====================================================================== */
static char *ParseFit(pfct, fmt, args, fp)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     Fit_t fp;
{
  char *f;
  char *ret= NULL;

  /*  printf("ParseFit\n");*/
  if (fmt && fp && (f= fmt[FMT_FIT])) {
    while (*f && !ret) {
      switch (*f) {
				/* private */
      case 'b': ret= ParseBgPol(pfct, fmt, fp); f++; break;
      case 'e': ret= ParseBgExp(pfct, fmt, fp); f++; break;
      case 'g': f++; ret= PrintFormInt(pfct, &f, FitBgGrad(fp)); break;
      case 'm':	f++; ret= ParseMark(pfct, fmt, args, &f, fp); break;
      case 'n': f++; ret= ParseMarkNum(pfct, &f, fp); break;
      case 'o': f++; ret= PrintFormDouble(pfct, &f, FitBgOffset(fp)); break;
      case 'p':	f++; ret= ParseFitPeaks(pfct, fmt, args, fp); break;
				/* common */
      default: ret= ParseFitCommon(pfct, args, &f, fp);
      }
    }
  }
  return ret;
}
/* ---------------------------------------------------------------------- */
static char *ParseFitCommon(pfct, args, f, fp)
     FmtPrintFunc pfct;
     FmtArgs args;
     char **f;
     Fit_t fp;
{
  /*  printf("ParseFitCommon\n");*/
  switch (**f) {
  case 'D': (*f)++; return ParseDate(pfct, f, fp);
  case 'e': (*f)++; return ParseExponent(pfct, f, fp);
  case 'f': (*f)++; return ParseFile(pfct, args, f);
  case 'F': (*f)++; return PrintFunction(pfct, f, fp);
  case 'G': (*f)++; return ParseFreedom(pfct, f, fp);
  case 'I': (*f)++; return PrintFormInt(pfct, f, args->idx);
  case 'M': (*f)++; return ParseMeasure(pfct, f, fp);
  case 'N': (*f)++; return ParseIteration(pfct, f, fp);
  case 'u': (*f)++; return PrintFormString(pfct, f, u_(args));
  case 'U': (*f)++; return PrintFormString(pfct, f, U_(args));
  case '\'': return PrintQuotedItem(pfct, f);
  default: (*f)++;
    return NULL;
  } /* ignore */
}
/* ---------------------------------------------------------------------- */
static char *ParseDate(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++;
    return PrintDate(pfct, f, &FitBgTime(fp));
  case 'r': (*f)++;
    return PrintDate(pfct, f, &FitTime(fp));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseExponent(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'l': (*f)++; return PrintFormInt(pfct, f, FitTlE(fp));
  case 'r': (*f)++; return PrintFormInt(pfct, f, FitTrE(fp));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseMeasure(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++; return ParseMeasureStat(pfct, f, FitStatBg(fp));
  case 'r': (*f)++; return ParseMeasureStat(pfct, f, FitStatReg(fp));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseFreedom(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++; return PrintFreedom(pfct, f, FitStatBg(fp));
  case 'r': (*f)++; return PrintFreedom(pfct, f, FitStatReg(fp));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *PrintFreedom(pfct, f, fs)
     FmtPrintFunc pfct;
     char **f;
     FitStat_t *fs;
{
  if (FitStatTime(fs)) {
    return PrintFormInt2(pfct, f, FitStatFreedom(fs)); }
  else {
    char fmt[IOBUFLEN];
    { char *ret; if ((ret= GetIntFmt(f, fmt, IOBUFLEN))) { return ret; } }
    return PrintFormString(pfct, f, ""); }
}
/* ---------------------------------------------------------------------- */
static char *ParseIteration(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++; return ParseIterationStat(pfct, f, FitStatBg(fp));
  case 'r': (*f)++; return ParseIterationStat(pfct, f, FitStatReg(fp));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseFitPeaks(pfct, fmt, args, fp)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     Fit_t fp;
{
  char *ret= NULL;
  /*  printf("ParseFitPeaks\n");*/
  FitPeakSort(fp, FitPeakSortTyp(fp));
  if (fmt[FMT_PKF]) {
    int i;
    for (i= 0; i< FitPeakNum(fp) && !ret; i++) {
      ret= ParseFitPeak(pfct, fmt[FMT_PKF], args, fp, i); } }
  return ret;
}
/* ---------------------------------------------------------------------- */
static char *ParseFitPeak(pfct, fmt, args, fp, idx)
     FmtPrintFunc pfct;
     char *fmt;
     FmtArgs args;
     Fit_t fp;
     int idx;
{
  int sidx= FitPeakSIdx(fp,idx);
  char *ret= NULL;
  ParV_t *pi= NULL, *p= FitPos(fp,sidx);
  char *f= fmt;
  double pos= ParVVal(p + FIT_PEAKPOS);
  double nv;

  /*  printf("ParseFitPeak\n");*/
  while (*f && !ret) {
    switch (*f) {
    case 'a':			/* amplitude */
      f++; pi= p + FIT_PEAKVOL;
      { FitOpt op;
	if ((op= FitPeakFunctionSetup(NULL, fp, NULL, sidx))) {
	  nv= 1.0 / FitNormVolume(op);
	  FitPeakFunctionFinish(op); }
	else { nv= 0.0; } }
      ret= ParseParam
	(pfct, &f, fp, pi, ParVVal(pi) * nv * e_(args, pos),
	 ParVErr(pi) * nv * e_(args, pos));
      break;
    case 'A':
      f++; pi= p + FIT_PEAKVOL;
      { FitOpt op;
	if ((op= FitPeakFunctionSetup(NULL, fp, NULL, sidx))) {
	  nv= 1.0 / FitNormVolume(op);
	  FitPeakFunctionFinish(op); }
	else { nv= 0.0; } }
      ret= ParseParam
	(pfct, &f, fp, pi, ParVVal(pi) * nv * E_(args, pos),
	 ParVErr(pi) * nv * E_(args, pos));
      break;
    case 'i': f++; ret= PrintFormInt(pfct, &f, idx); break;
    case 'O': ret= PrintFormString(pfct, &f, "fit"); break;
    case 'p':			/* position */
      f++; pi= p + FIT_PEAKPOS;
      ret= ParseParam(pfct, &f, fp, pi,
		      p_(args, pos), w_(args, pos, ParVErr(pi)));
      break;
    case 'P':
      f++; pi= p + FIT_PEAKPOS;
      ret= ParseParam(pfct, &f, fp, pi,
		      P_(args, pos), W_(args, pos, ParVErr(pi)));
      break;
    case 's': f++;		/* step */
      switch (*f) {
      case 'h':		/* height */
	f++; pi= p + FIT_PEAKSH;
	ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi), ParVErr(pi));
	break;
      case 'w':		/* width */
	f++; pi= p + FIT_PEAKSW;
	ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi), ParVErr(pi));
	break; }
      break;
    case 't':			/* tail */
      f++;
      switch (*f) {
      case 'l':		/* left */
	f++; pi= p + FIT_PEAKTL;
	ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi), ParVErr(pi));
	break;
      case 'r':		/* right */
	f++; pi= p + FIT_PEAKTR;
	ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi), ParVErr(pi));
	break; }
      break;
    case 'v':			/* volume */
      f++; pi= p + FIT_PEAKVOL;
      ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi) * e_(args, pos),
		      ParVErr(pi) * e_(args, pos));
      break;
    case 'V':
      f++; pi= p + FIT_PEAKVOL;
      ret= ParseParam(pfct, &f, fp, pi, ParVVal(pi) * E_(args, pos),
		      ParVErr(pi) * E_(args, pos));
      break;
    case 'w':			/* width */
      f++; pi= p + FIT_PEAKWIDTH;
      ret= ParseParam(pfct, &f, fp, pi,
		      w_(args, pos, ParVVal(pi)) * SIGMA_TO_FWHM,
		      d_(args, pos,
			 ParVVal(pi), ParVErr(pi)) * SIGMA_TO_FWHM);
      break;
    case 'W':
      f++; pi= p + FIT_PEAKWIDTH;
      ret= ParseParam(pfct, &f, fp, pi,
		      W_(args, pos, ParVVal(pi)) * SIGMA_TO_FWHM,
		      D_(args, pos,
			 ParVVal(pi), ParVErr(pi)) * SIGMA_TO_FWHM);
      break;
    default: ret= ParseFitCommon(pfct, args, &f, fp); break; } }
  return ret;
}
/* ---------------------------------------------------------------------- */
static char *ParseMeasureStat(pfct, f, fs)
     FmtPrintFunc pfct;
     char **f;
     FitStat_t *fs;
{
  switch (**f) {
  case 't': (*f)++; return PrintMeasureType(pfct, f, FitStatMeasureTyp(fs));
  default: return ParseValErr(pfct, f, 
			      FitStatMeasure(fs), FitStatDMeasure(fs),
			      (FitStatTime(fs) ? NULL : "undefined")); }
}
/* ---------------------------------------------------------------------- */
static char *ParseIterationStat(pfct, f, fs)
     FmtPrintFunc pfct;
     char **f;
     FitStat_t *fs;
{
  switch (**f) {
  case 'm': (*f)++;
    return PrintFormInt(pfct, f, FitStatMaxIterations(fs));
  case 'n': (*f)++;
    return PrintFormInt(pfct, f, 
			(FitStatTime(fs) ? FitStatIterations(fs) : 0));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *PrintMeasureType(pfct, f, t)
     FmtPrintFunc pfct;
     char **f;
     int t;
{
  char *s;
  switch (t) {
  case FIT_MDY:		s= "chi(dy)**2"; break;
  case FIT_MY:		s= "chi(y)**2"; break;
  case FIT_MPOI:	s= "poisson(y)"; break;
  default:		s= "unknown"; break; }
  return PrintFormString(pfct, f, s);
}
/* ---------------------------------------------------------------------- */
static char *PrintFunction(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  char *s;
  switch (FitPeakTyp(fp)) {
  case FIT_PKCA:	s= "continuous tail, arctan-step";
  case FIT_PKAE:	s= "additive tail, erf-step";
  default:		s= "unknown"; break; }
  return PrintFormString(pfct, f, s);
}
/* ====================================================================== */
/* common: bg  / mark / file / parameter */
/* ====================================================================== */
static char *ParseParam(pfct, f, fp, p, val, err)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
     ParV_t *p;
     double val, err;
{
  switch (**f) {
  case 'c': (*f)++; return ParseCorr(pfct, f, p, fp);
  case 's': (*f)++; return PrintStatus(pfct, f, p);
  default: return ParseValErr(pfct, f, val, err,
			      (ParVSsusp(p) ? "none" : NULL)); }
}
/* ---------------------------------------------------------------------- */
static char *ParseValErr(pfct, f, val, err, val_string)
     FmtPrintFunc pfct;
     char **f;
     double val, err;
     char *val_string;
{
  switch (**f) {
  case 'd': (*f)++;
    return PrintFormVal(pfct, f, FALSE, val, err, val_string);
  case 'v': (*f)++;
    return PrintFormVal(pfct, f, TRUE, val, err, val_string);
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseCorr(pfct, f, p, fp)
     FmtPrintFunc pfct;
     char **f;
     ParV_t *p;
     Fit_t fp;
{
  if (ParVMode(p) != PAR_MCORR) {
    char fmt[IOBUFLEN];
    switch (**f) {
    case 'i': (*f)++;
      { char *ret; if ((ret= GetIntFmt(f, fmt, IOBUFLEN))) { return ret; } }
      break;
    case '0':
    case '1': (*f)++;
      { char *ret; if ((ret= GetDoubleFmt(f, fmt, IOBUFLEN))) { return ret; } }
      break;
    default: FmtsetErrNo(FMTILLCHR); return *f; }
    return PrintFormString(pfct, f, ""); }
  else {
    switch (**f) {
    case 'i': (*f)++;
      { int c= ParVCorrIndex(p);
	c= (c < FIT_1STPEAK) ? c : FitPeakSRIdx(fp,FitPeakIdx(c));
	return PrintFormInt2(pfct, f, c); }
    case '0': (*f)++; return PrintFormDouble2(pfct, f, ParVCorr(p)[0]);
    case '1': (*f)++; return PrintFormDouble2(pfct, f, ParVCorr(p)[1]);
    default: FmtsetErrNo(FMTILLCHR); return *f; } }
}
/* ---------------------------------------------------------------------- */
static char *PrintStatus(pfct, f, p)
     FmtPrintFunc pfct;
     char **f;
     ParV_t *p;
{
  char *s;

  if (ParVSsusp(p)) { s= ""; }
  else {
    switch (ParVUserStat(p)) {
    case FIT_USCAL:	s= "cal"; break;
    case FIT_USINIT:	s= "initialized"; break;
    case FIT_USUSER:
      switch (ParVMode(p)) {
      case PAR_MFREE:	s= "user"; break;
      case PAR_MHOLD:	s= "hold"; break;
      case PAR_MCORR:	s= "correlated"; break;
      default:		s= "undefined"; break; }
      break;
    case FIT_USFIT:
      switch (ParVMode(p)) {
      case PAR_MFREE:	s= ""; break;
      case PAR_MHOLD:	s= "hold"; break;
      case PAR_MCORR:	s= "correlated"; break;
      default:		s= "undefined"; break; }
      break;
    case FIT_USSTART:
    default:		s= "undefined"; } }
  return PrintFormString(pfct, f, s);
}
/* ---------------------------------------------------------------------- */
static char *ParseBgPol(pfct, fmt, fp)
     FmtPrintFunc pfct;
     char **fmt;
     Fit_t fp;
{
  char *ret= NULL;

  /*  printf("ParseBgPol\n");*/
  if (fmt[FMT_BGPOL]) {
    int i;

    for (i= 0; i< FitBgGrad(fp) && !ret; i++) {
      ParV_t *b= FitBgCoef(fp,i);
      char *f= fmt[FMT_BGPOL];
      while (*f && !ret) {
	switch (*f) {
	case 'b': f++;
	  ret= ParseParam(pfct, &f, fp, b, ParVVal(b), ParVErr(b)); break;
	case 'i': f++; ret= PrintFormInt(pfct, &f, i); break;
	case '\'': ret= PrintQuotedItem(pfct, &f); break;
	default: f++; break; } } } } /* ignore */
  return NULL;
}
/* ---------------------------------------------------------------------- */
static char *ParseBgExp(pfct, fmt, fp)
     FmtPrintFunc pfct;
     char **fmt;
     Fit_t fp;
{
  char *ret= NULL;

  /*  printf("ParseBgExp\n");*/
  if (fmt[FMT_BGEXP]) {
    int i;

    for (i= 0; i< FitBgExpNum(fp) && !ret; i++) {
      ParV_t *b;
      char *f= fmt[FMT_BGEXP];
      while (*f && !ret) {
	switch (*f) {
	case 'e': f++;
	  b= FitBgCoef(fp,FIT_BGGRD + 2 * i + 1);
	  ret= ParseParam(pfct, &f, fp, b, ParVVal(b), ParVErr(b)); break;
	case 'f': f++;
	  b= FitBgCoef(fp,FIT_BGGRD + 2 * i);
	  ret= ParseParam(pfct, &f, fp, b, ParVVal(b), ParVErr(b)); break;
	case 'i': f++; ret= PrintFormInt(pfct, &f, i); break;
	case '\'': ret= PrintQuotedItem(pfct, &f); break;
	default: f++; break; } } } } /* ignore */
  return NULL;
}
/* ---------------------------------------------------------------------- */
static char *ParseFile(pfct, args, f)
     FmtPrintFunc pfct;
     FmtArgs args;
     char **f;
{
  switch (**f) {
  case 's': (*f)++; return PrintFormString(pfct, f, PthShort(args->file));
  case 'l': (*f)++; return PrintFormString(pfct, f, PthExpand(args->file));
  case 't': (*f)++; return PrintFormString(pfct, f, PthTail(args->file));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ---------------------------------------------------------------------- */
static char *ParseMark(pfct, fmt, args, f, fp)
     FmtPrintFunc pfct;
     char **fmt;
     FmtArgs args;
     char **f;
     Fit_t fp;
{
  PVec_t mp;

  char *ret= NULL;

  switch (**f) {
  case 'b': (*f)++; mp= FitBgMrk(fp); break;
  case 'r': (*f)++; mp= FitRegMrk(fp); break;
  default: return *f; }

  if (fmt[FMT_MRK]) {
    int i;

    for (i= 0; i< VecNum(mp) && !ret; i++) {
      char *fm= fmt[FMT_MRK];
      while (*fm && !ret) {
	switch (*fm) {
	case 'l': fm++; ret= PrintFormDouble(pfct, &fm, p_(args,PVecX1(mp,i)));
	  break;
	case 'L': fm++; ret= PrintFormDouble(pfct, &fm, P_(args,PVecX1(mp,i)));
	  break;
	case 'r': fm++; ret= PrintFormDouble(pfct, &fm, p_(args,PVecX2(mp,i)));
	  break;
	case 'R': fm++; ret= PrintFormDouble(pfct, &fm, P_(args,PVecX2(mp,i)));
	  break;
	case '\'': ret= PrintQuotedItem(pfct, &fm); break;
	default: fm++; break; } } } }
  return ret;
}
/* ---------------------------------------------------------------------- */
static char *ParseMarkNum(pfct, f, fp)
     FmtPrintFunc pfct;
     char **f;
     Fit_t fp;
{
  switch (**f) {
  case 'b': (*f)++; return PrintFormInt(pfct, f, VecNum(FitBgMrk(fp)));
  case 'r': (*f)++; return PrintFormInt(pfct, f, VecNum(FitRegMrk(fp)));
  default: FmtsetErrNo(FMTILLCHR); return *f; }
}
/* ====================================================================== */
/* calibration setup */
/* ====================================================================== */
static double Id2(cal, p, w) Cal_t cal; double p, w; { return w; }
/* ---------------------------------------------------------------------- */
static double Id3(cal, p, w, dw) Cal_t cal; double p, w, dw; { return dw; }
/* ---------------------------------------------------------------------- */
static void InitArgs(args, file, idx, cal, data_calibrated, output_calibrated)
     FmtArgs args;
     char *file;
     int idx;
     Cal_t cal;
     int data_calibrated, output_calibrated;
{
  args->idx= idx;
  args->file= file;
  args->cal= cal;

  if (output_calibrated == data_calibrated) {
    if (!CalCheck(CAL_POS, cal)) {
      if (output_calibrated) { args->e= CalEfc; args->E= CalUnity; }
      else { args->e= CalUnity; args->E= CalEfc; }
      args->p= CalId; args->w= Id2; args->d= Id3; args->u= CalChn;
      args->P= CalId; args->W= Id2; args->D= Id3; args->U= CalChn; }
    else {
      args->p= CalId; args->w= Id2; args->d= Id3;
      if (output_calibrated) {
	args->e= CalEfcP; args->E= CalUnity;
	args->u= CalUnit; args->U= CalChn;
	args->P= CalC; args->W= CalCWidth; args->D= CalCDwidth; }
      else {
	args->e= CalUnity; args->E= CalEfc;
	args->u= CalChn; args->U= CalUnit;
	args->P= CalP; args->W= CalPWidth; args->D= CalPDwidth; } } }
  else if (output_calibrated) {
    args->e= CalEfc; args->E= CalUnity;
    args->p= CalP; args->w= CalPWidth; args->d= CalPDwidth; args->u= CalUnit;
    args->P= CalId; args->W= Id2; args->D= Id3; args->U= CalChn; }
  else {
    args->e= CalUnity; args->E= CalEfcP;
    args->p= CalC; args->w= CalCWidth; args->d= CalCDwidth; args->u= CalChn;
    args->P= CalId; args->W= Id2; args->D= Id3; args->U= CalUnit; }
  if (!CalCheck(CAL_EFC, cal)) { args->e= args->E= CalUnity; }
}
/* ====================================================================== */
/* general format routines */
/* ====================================================================== */
static char *PrintDate(pfct, f, t)
     FmtPrintFunc pfct;
     char **f;
     time_t *t;
{
  if (*t) {
    char ts[26];
    strcpy(ts, ctime(t)); ts[24]= '\0';
    return PrintFormString2(pfct, f, ts); }
  else {
    char fmt[IOBUFLEN];
    { char *ret; if ((ret= GetStringFmt(f, fmt, IOBUFLEN))) { return ret; } }
    return PrintFormString(pfct, f, ""); }
}
/* ---------------------------------------------------------------------- */
static void SprintVal(val, d_val, pval, fmtTyp, nfrac, fix_fmt, item)
     double val, d_val;
     double pval;
     ParamFmt fmtTyp;
     int nfrac;
     char *fix_fmt;
     char *item;
{
  int prec, mag;
  char fmt[IOBUFLEN];

  nfrac= ABS(nfrac);
  nfrac= MAX(1, nfrac);

  d_val= ABS(d_val);

  mag= (val == 0.0) ? 0 : floor(log10(ABS(val)));
  prec= nfrac - ((d_val == 0.0) ? 1 : (1 + (int)floor(log10(d_val))));

  switch (fmtTyp) {
  case FmtEng:
    mag -= mag % 3;
  case FmtSci:
    prec += mag;
    prec= MAX(0, prec);

    if (d_val == 0.0 && !mag) { strcpy(fmt, "%f"); }
    else if (d_val == 0.0) { sprintf(fmt, "%%fe%d", mag); }
    else if (!mag) { sprintf(fmt, "%%.%df", prec); }
    else { sprintf(fmt, "%%.%dfe%d", prec, mag); }

    sprintf(item, fmt, pval * exp_10((double)-mag));
    break;
  case FmtPar:
    if (d_val == 0.0) { sprintf(item, "%g(0)", val); }
    else {
      prec= MAX(0, prec);
      sprintf(fmt, "%%.%df(%%.0f)", prec);
      sprintf(item, fmt, val, d_val * exp_10((double)prec)); }
    break;
  case FmtNrm:
    if (d_val == 0.0) { sprintf(item, "%g", pval); }
    else {
      prec= MAX(0, prec);
      sprintf(fmt, "%%.%df", prec);
      sprintf(item, fmt, pval); }
    break;
  case FmtFix:
    if (*fix_fmt) { sprintf(item, fix_fmt, pval); break; }
  default: *item= '\0'; break; }
}
/* ---------------------------------------------------------------------- */
static int LegalDoubleFmt(fmt)
     char *fmt;
{
  return LegalFmt(fmt, "fge", FMTILLDBL);
}
/* ---------------------------------------------------------------------- */
static int LegalIntFmt(fmt)
     char *fmt;
{
  return LegalFmt(fmt, "doxu", FMTILLINT);
}
/* ---------------------------------------------------------------------- */
static int LegalStringFmt(fmt)
     char *fmt;
{
  return LegalFmt(fmt, "s", FMTILLSTR);
}
/* ---------------------------------------------------------------------- */
static int LegalFmt(fmt, typ, err)
     char *fmt, *typ;
     int err;
{
  char *fp= fmt;
  int count= 0;
  if (fmt) {
    while ((fp= strchr(fp, '%'))) {
      fp++;
      if (*fp == '%') { fp++; continue; }

      if (*fp == '-') { fp++; }
      while ('0' <= *fp && *fp < '9') { fp++; }
      if (*fp == '.') { fp++; }
      while ('0' <= *fp && *fp < '9') { fp++; }
      if (strchr(typ, *fp)) { count++; }
      else { count= 0; break; } } }
  if (count == 1) { return TRUE; }
  FmtsetErrNo(err);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int GetQuotedString(f, fmt, fl)
     char **f, *fmt;
     int fl;
{
  *fmt= '\0';
  while (**f && **f != '\'') { (*f)++; }
  if (**f) { (*f)++; }
  else { FmtsetErrNo(FMTNOQUO); return FALSE; }
  while (**f && fl > 1 && UFUnquoteChar(f, &fmt, &fl, '\'')) { ; }
  if (fl < 1) { FmtsetErrNo(FMTOVER); return FALSE; }
  *fmt= '\0';
  (*f)++;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static char *GetDoubleFmt(f, fmt, fl)
     char **f;
     char *fmt;
     int fl;
{
  char *ret= *f;
  if (!GetQuotedString(f, fmt, fl)) { return ret; }
  else { return (LegalDoubleFmt(fmt) ? NULL : ret); }
}
/* ---------------------------------------------------------------------- */
static char *GetIntFmt(f, fmt, fl)
     char **f;
     char *fmt;
     int fl;
{
  char *ret= *f;
  if (!GetQuotedString(f, fmt, fl)) { return ret; }
  else { return (LegalIntFmt(fmt) ? NULL : ret); }
}
/* ---------------------------------------------------------------------- */
static char *GetStringFmt(f, fmt, fl)
     char **f;
     char *fmt;
     int fl;
{
  char *ret= *f;
  if (!GetQuotedString(f, fmt, fl)) { return ret; }
  else { return (LegalStringFmt(fmt) ? NULL : ret); }
}
/* ---------------------------------------------------------------------- */
static char *PrintFormInt(pfct, f, v)
     FmtPrintFunc pfct;
     char **f;
     int v;
{
  char fmt[IOBUFLEN];
  char line[IOBUFLEN];
  char *ret= *f;

  if (GetIntFmt(f, fmt, IOBUFLEN)) { return ret; }
  sprintf(line, fmt, v);
  return (pfct("%s", line) ? NULL : ret);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormInt2(pfct, f, v)
     FmtPrintFunc pfct;
     char **f;
     int v;
{
  char fmt[IOBUFLEN];
  char line[IOBUFLEN];
  char *ret= *f;

  if (GetIntFmt(f, fmt, IOBUFLEN)) { return ret; }
  sprintf(line, fmt, v);
  return PrintFormString(pfct, f, line);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormDouble(pfct, f, v)
     FmtPrintFunc pfct;
     char **f;
     double v;
{
  char fmt[IOBUFLEN];
  char line[IOBUFLEN];
  char *ret= *f;

  if (GetDoubleFmt(f, fmt, IOBUFLEN)) { return ret; }
  sprintf(line, fmt, v);
  return (pfct("%s", line) ? NULL : ret);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormDouble2(pfct, f, v)
     FmtPrintFunc pfct;
     char **f;
     double v;
{
  char fmt[IOBUFLEN];
  char line[IOBUFLEN];
  char *ret= *f;

  if (GetDoubleFmt(f, fmt, IOBUFLEN)) { return ret; }
  sprintf(line, fmt, v);
  return PrintFormString(pfct, f, line);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormString(pfct, f, s)
     FmtPrintFunc pfct;
     char **f;
     char *s;
{
  char fmt[IOBUFLEN];
  char *ret= *f;
  if (GetStringFmt(f, fmt, IOBUFLEN)) { return ret; }
  return (pfct(fmt, s) ? NULL : ret);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormString2(pfct, f, s)
     FmtPrintFunc pfct;
     char **f;
     char *s;
{
  char fmt[IOBUFLEN];
  char line[IOBUFLEN];
  char *ret= *f;

  if (GetStringFmt(f, fmt, IOBUFLEN)) { return ret; }
  sprintf(line, fmt, s);
  return PrintFormString(pfct, f, line);
}
/* ---------------------------------------------------------------------- */
static char *PrintQuotedItem(pfct, f)
     FmtPrintFunc pfct;
     char **f;
{
  char item[IOBUFLEN];
  char *ret= *f;

  if (!GetQuotedString(f, item, IOBUFLEN)) { return ret; }
  return (pfct("%s", item) ? NULL : ret);
}
/* ---------------------------------------------------------------------- */
static char *PrintFormVal(pfct, f, print_val, val, err, val_string)
     FmtPrintFunc pfct;
     char **f;
     int print_val;
     double val, err;
     char *val_string;
{
  char fmt[IOBUFLEN];
  char item[IOBUFLEN];
  int fl;
  ParamFmt pf;
  char *ret;

  if ((ret= GetValFormat(f, fmt, &pf, &fl))) { return ret; }

  if (!val_string) {
    val_string= item;
    SprintVal(val, err, (print_val ? val : err), pf, fl, fmt, item); }
  return PrintFormString(pfct, f, val_string);
}
/* ---------------------------------------------------------------------- */
static char *GetValFormat(f, fmt, pf, fl)
     char **f;
     char *fmt;
     ParamFmt *pf;
     int *fl;
{
  char found= '\0';
  char *ret= *f;
  *fmt= '\0';
  while (**f && !found) {
    switch (**f) {
    case 'e':			/* engineer */
    case 'n':			/* normal */
    case 'p':			/* parenthesized */
    case 's':			/* scientific */
      found= **f; (*f)++;
      if (sscanf(*f, "%d", fl) != 1) { return ret; }
      { int dummy; while (sscanf(++(*f), "%d", &dummy)) { ; } }
      break;
    case 'f':			/* fixed */
      found= **f;
      if (GetDoubleFmt(f, fmt, IOBUFLEN)) { return ret; }
      break;
    default: (*f)++; } }	/* ignore */

  switch (found) {
  case 'e': *pf= FmtEng; break;
  case 'f': *pf= FmtFix; break;
  case 'n': *pf= FmtNrm; break;
  case 'p': *pf= FmtPar; break;
  case 's': *pf= FmtSci; break;
  default: return ret; }

  return NULL;
}
/* ---------------------------------------------------------------------- */
