/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsFitIo.c,v 1.5 1994/12/21 17:12:34 vs Exp vs $ */

#define _vsFitIo_c

#include <time.h>

#include "vsAlloc.h"
#include "vsTypes.h"

#include "vsFit.h"
#include "vsTools.h"

#include "vsFitIo.h"
#include "vsFitIoP.h"
/* ---------------------------------------------------------------------- */
char *FioError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* FIONOERR:  */ "vsFitIo: no error",
  /* FIOPOINT:  */ "vsFitIo: illegal NULL-pointer",
  /* FIOALLOC:  */ "vsFitIo: memory allocation failed",
  /* FIOOPEN:   */ "vsFitIo: could not open data file",
  /* FIOFILTYP: */ "vsFitIo: illegal data file type",
  /* FIOREAD:   */ "vsFitIo: could not read data file",
  /* FIOWRITE:  */ "vsFitIo: could not write data file",
  /* FIOTOCIDX: */ "vsFitIo: illegal toc index",
  /* FIORECTYP: */ "vsFitIo: illegal record type in data file",
  /* FIOCOMP:   */ "vsFitIo: could not compress data record",
  /* FIOUNCOMP: */ "vsFitIo: could not uncompress data record",
  /* FIOVERS:   */ "vsFitIo: could not handle compress version",
  /* FIOINTERN: */ "vsFitIo: severe internal error",
  /* FIOPEAINS: */ "vsFitIo: could not insert peak",
  /* FIORDONLY: */ "vsFitIo: data file is read only",
  /* FIONOPEAK: */ "vsFitIo: no peak found",
  /* FIOPKTYP:  */ "vsFitIo: illegal peak function typ",
  /* FIOBGSTO:  */ "vsFitIo: bg-dimensions exceed available arraysize",
  /* default:   */ "vsFitIo: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */

/* ================ */
/* initialize Fio_t */
/* ================ */

int FioInit(fp)
     Fio_t fp;
{
  if (fp == NULL) { FiosetErrNo(FIOPOINT); return FALSE; }
  else {
    FioRfp(fp)				= NULL;
    FioSea(fp)				= NULL;
    FioCurPeakTyp(FioCur(fp))		= FIO_PNONE;
    FioCurPeakPeaIdx(FioCur(fp))	= -1;
    FioCurPeakSeaPeak(FioCur(fp))	= NULL;

    return (VecInit(FioToc(fp), sizeof(FioToc_t)) &&
	    VecInit(FioBin(fp), sizeof(Bin_t)) &&
	    VecInit(FioBinMrk(fp), sizeof(Pair_t)) &&
	    VecInit(FioPea(fp), sizeof(FioPea_t)) &&
	    FioReset(fp)); }
}
/* ---------------------------------------------------------------------- */
int FioReset(fp)
     Fio_t fp;
{
  if (fp == NULL) {
    FiosetErrNo(FIOPOINT);
    return FALSE;
  }
  else if (!RecClose(FioRfp(fp))) {
    return FALSE;
  }
  else {
    FioRfp(fp)= NULL;

    SeaFreePeakList(FioSea(fp));
    FioSea(fp)= NULL;
    FioWriteTail(fp)= FALSE;
    if (FioFullPeaks(fp)) {
      int p;
      FioPeaVec_t pea=FioPea(fp);
      for (p= 0; p< VecNum(pea); p++) {
	FioPeaPPar(pea,p)= FitPeakFunctionFinish(FioPeaPPar(pea,p)); 
      }
      FioFullPeaks(fp)= FALSE;
    }

    return (VecReset(FioToc(fp)) &&
	    VecReset(FioBin(fp)) &&
	    VecReset(FioBinMrk(fp)) &&
	    VecReset(FioPea(fp)));
  }
}
/* ---------------------------------------------------------------------- */
int FioFree(fp)
     Fio_t fp;
{
  if (fp == NULL) {
    FiosetErrNo(FIOPOINT);
    return FALSE;
  }
  else if (!FioReset(fp)) {
    return FALSE;
  }
  else {
    return (VecFree(FioToc(fp)) &&
	    VecFree(FioBin(fp)) &&
	    VecFree(FioBinMrk(fp)) &&
	    VecFree(FioPea(fp)));
  }
}
/* ---------------------------------------------------------------------- */
/* ================ */
/* open/close Fio_t */
/* ================ */

int FioOpen(fp, file, mode)
     Fio_t fp;
     char *file;
     char *mode;
{
  /*  printf("FioOpen\n");*/
  if (!FioReset(fp)) {
    return FALSE;
  }
  else if (NULL == (FioRfp(fp) = RecOpen(file, mode))) {
    /*    printf("Opening file %s with mode %s failed\n", file, mode);*/
    FiosetErrNo(FIOOPEN);
    return FALSE;
  }
  else {
    if (RFilStat(FioRfp(fp)) == REC_SNEW) {
      RFilUserTyp(FioRfp(fp)) = FIO_RFILTYP;
      return TRUE;
    }
    else if (RFilUserTyp(FioRfp(fp)) != FIO_RFILTYP) {
      FiosetErrNo(FIOFILTYP);
      FioReset(fp);
      return FALSE;
    }
    else {
      int seatrue = FioTailReadSeaPeakList(fp, FIO_SEA);
      int vectrue = (FioTailReadVec(FioRfp(fp), FIO_TOC,
				    FioToc_r, (Vec_t)FioToc(fp)) &&
		     FioTailReadVec(FioRfp(fp), FIO_BIN,
				    Bin_r, (Vec_t)FioBin(fp)) &&
		     FioTailReadVec(FioRfp(fp), FIO_BINMRK,
				    Pair_r, (Vec_t)FioBinMrk(fp)) &&
		     FioTailReadVec(FioRfp(fp), FIO_PEA,
				    FioPea_r, (Vec_t)FioPea(fp)));
      if (seatrue || vectrue) {
	return TRUE;
      }
      else {
	FioReset(fp);
	return FALSE;
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
int FioClose(fp)
     Fio_t fp;
{
  if (fp == NULL) { FiosetErrNo(FIOPOINT); return FALSE; }
  else {
    if (!FioTailWrite(fp)) { return FALSE; }
    return FioReset(fp); }
}
/* ---------------------------------------------------------------------- */

/* ============================= */
/* read/write Fio*Vec_t of Fio_t */
/* ============================= */
  
static int FioTailWrite(fp)
     Fio_t fp;
{
  /*  printf("FioTailWrite\n");*/
  if (!FioWriteTail(fp)) {
    return TRUE;
  }
  else if (FioTailWriteSeaPeakList(fp, FIO_SEA) &&
	   FioTailWriteVec(FioRfp(fp), FIO_TOC,
			   FioToc_w, (Vec_t)FioToc(fp)) &&
	   FioTailWriteVec(FioRfp(fp), FIO_BIN,
			   Bin_w, (Vec_t)FioBin(fp)) &&
	   FioTailWriteVec(FioRfp(fp), FIO_BINMRK,
			   Pair_w, (Vec_t)FioBinMrk(fp)) &&
	   FioTailWriteVec(FioRfp(fp), FIO_PEA,
			   FioPea_w, (Vec_t)FioPea(fp)) &&
	   FioTailFinish(fp)) {
    return TRUE;
  }
  return FALSE;
}

static int FioTailFinish(fp)
     Fio_t fp;
{
  /*  printf("FioTailFinish\n");*/
  if (!RecBackupFinish(FioRfp(fp))) {
    FioWriteTail(fp)= TRUE;
    FiosetErrNo(FIOWRITE);
    return FALSE;
  }
  else {
    FioWriteTail(fp)= FALSE;
    return TRUE;
  }
}

static int FioTailReadVec(rfp, recTyp, read_fct, vec)
     RFILE *rfp;
     int recTyp;
     pfi_t read_fct;
     Vec_t vec;
{
  /*  printf("FioTailReadVec\n");*/
  if (RecTailReadStart(rfp, recTyp) &&
      Vec_r(rfp, read_fct, vec) &&
      RecTailReadFinish(rfp)) {
    return TRUE;
  }
  else {
    FiosetErrNo(FIOREAD);
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
static int FioTailWriteVec(rfp, recTyp, write_fct, vec)
     RFILE *rfp;
     int recTyp;
     pfi_t write_fct;
     Vec_t vec;
{
  if (RecTailWriteStart(rfp, recTyp) &&
      Vec_w(rfp, write_fct, vec) &&
      RecTailWriteFinish(rfp)) {
    return TRUE; }
  else {
    FiosetErrNo(FIOWRITE); return FALSE; }
}
/* ---------------------------------------------------------------------- */
/* ================ */
/* read/write Fit_t */
/* ================ */

int FioReadFit(fp, idx, ftp)
     Fio_t fp;
     int idx;
     Fit_t ftp;
{
  if (fp == NULL || ftp == NULL) {
    FiosetErrNo(FIOPOINT);
    return FALSE;
  }
  else {
    FioTocVec_t tp= FioToc(fp);
    if (idx < 0 || VecNum(tp) <= idx) {
      FiosetErrNo(FIOTOCIDX);
      return FALSE;
    }
    else if (!RecReadStart(FioRfp(fp), FioTocPPos(tp,idx))) {
      FiosetErrNo(FIOREAD);
      return FALSE;
    }
    else if (RecTyp(RFilF(FioRfp(fp))) != FIO_FIT) {
      FiosetErrNo(FIORECTYP);
      return FALSE;
    }
    else if (Fit_r(FioRfp(fp), ftp) &&
	     RecReadFinish(FioRfp(fp))) {
      return FitPeakSort(ftp, FitPeakSortTyp(ftp));
    }
    else {
      FiosetErrNo(FIOREAD);
      return FALSE;
    }
  }
}
/* ---------------------------------------------------------------------- */
int FioWriteFit(fp, idx, ftp, binMarks)
     Fio_t fp;
     int *idx;
     Fit_t ftp;
     Pair_t *binMarks;
{
  if (fp == NULL || ftp == NULL) {
    /*
     * No file or no fit defined
     */
    FiosetErrNo(FIOPOINT);
    return FALSE;
  }
  else {
    if (*idx < 0) {
      unsigned int pos;

      if (!(RecWriteStart(FioRfp(fp), FIO_FIT) &&
	    Fit_w(FioRfp(fp), ftp) &&
	    RecWriteFinish(FioRfp(fp), &pos))) {
	FiosetErrNo(FIOWRITE);
	return FALSE;
      }
      FioWriteTail(fp)= TRUE;
      if (!FioUpdateToc(fp, pos, ftp, idx)) {
	return FALSE;
      }
    }

    if (FitFitTyp(ftp) != FIT_FTFIT) {
      return TRUE;
    }
    else if (FioUpdateBin(fp, binMarks, ftp, *idx)) {
      if (!FioTailWrite(fp)) { /* write tail records */
	return FALSE;
      }
      else {
	return TRUE;
      }
    }
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
/* ====================== */
/* read/write SeaPeakList */
/* ====================== */

int FioReadSea(fp, sea)
     Fio_t fp;
     SeaPeakList *sea;
{
  if (!fp || !sea) { FiosetErrNo(FIOPOINT); return FALSE; }
  SeaFreePeakList(*sea); *sea= NULL;
  if (!FioSea(fp) || (*sea= SeaDupPeakList(FioSea(fp)))) { return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int FioWriteSea(fp, sea)
     Fio_t fp;
     SeaPeakList sea;
{
  /*  printf("FioWriteSea\n");*/
  if (!fp || !FioRfp(fp)) {
    FiosetErrNo(FIOPOINT);
    return FALSE;
  }
  else if (strchr(RFilMode(FioRfp(fp)), '+') ||
	   strchr(RFilMode(FioRfp(fp)), 'w')) { /* writable */
    SeaFreePeakList(FioSea(fp));
    FioSea(fp)= SeaDupPeakList(sea);
    if (FioWriteTail(fp)) { /* write tail records */
      return FioTailWrite(fp);
    }
    else if (FioTailWriteSeaPeakList(fp, FIO_SEA) &&
	     FioTailFinish(fp)) {
      return TRUE;
    }
    else {
      return FALSE;
    }
  }
  else {
    FiosetErrNo(FIORDONLY);
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */

/* =============== */
/* update FioToc_t */
/* =============== */

static int FioUpdateToc(fp, pos, ftp, idx)
     Fio_t fp;
     unsigned int pos;
     Fit_t ftp;
     int *idx;
{
  FioTocVec_t tp= FioToc(fp);
  *idx= VecNum(tp);

  if (!VecRealloc(tp, *idx+1)) { FiosetErrNo(FIOALLOC); return FALSE; }
  else {
    FioTocPPos(tp,*idx) = pos;
    FioTocPTyp(tp,*idx) = FitFitTyp(ftp);
    FioTocPTime(tp,*idx) = FitTime(ftp);
    VecNum(tp)++;
    return TRUE; }
}
/* ---------------------------------------------------------------------- */

/* ============ */
/* update Bin_t */
/* ============ */

static int FioUpdateBin(fp, binMarks, ftp, tocIdx)
     Fio_t fp;
     Pair_t *binMarks;
     Fit_t ftp;
     int tocIdx;
{
  BinVec_t bp= FioBin(fp);
  PVec_t mp= FioBinMrk(fp);
  int b1= NINT(PairX1(binMarks));
  int b2= NINT(PairX2(binMarks));

  if (!FioPeakDel(fp, binMarks)) { return FALSE; }
				/* binMarks, FioBinMrk in channels */
				/* => compare as integers */
  { int i= VecNum(bp)-1;
    while (i >= 0) {
      int m1= NINT(PVecX1(mp,i));
      int m2= NINT(PVecX2(mp,i));

      if (m2 < b1) {		     			       /*   [b]     */
	return FioBinInsert(fp, i+1, binMarks, ftp, tocIdx); } /*[i]   [i+1]*/
      else if (b1 <= m2 && m2 <= b2) {
	if (m1 < b1) {		/*   [ b ] */
	  m2 = b1-1;		/* [ i ]   */
	  BinPTime(bp,i)= time(NULL);
	  return FioBinInsert(fp, i+1, binMarks, ftp, tocIdx); }
	else {						/* [ b ] */
	  if (!FioBinSub(fp,i)) { return FALSE; }	/* [i]   */
	  i--; } }
      else if (m1 <= b2 && b2 < m2) {
	if (m1 >= b1) {		/* [b]   */
	  m1= b2+1;		/* [ i ] */
	  BinPTime(bp,i)= time(NULL);
	  i--; }
	else {			/*  [b]  */
	  return (FioBinSplit(fp, i, binMarks) &&
		  FioBinInsert(fp, i+1, binMarks, ftp, tocIdx)); } }
      else if (b2 < m1) {	/* [b]     */
	i--; }			/*     [i] */
      else {
	FiosetErrNo(FIOINTERN); return FALSE; } } }
  return FioBinInsert(fp, 0, binMarks, ftp, tocIdx);
}
/* ---------------------------------------------------------------------- */
static int FioBinInsert(fp, idx, binMarks, ftp, tocIdx)
     Fio_t fp;
     int idx;
     Pair_t *binMarks;
     Fit_t ftp;
     int tocIdx;
{
  if (FioPeakInsert(fp, idx, binMarks, ftp)) {
    if (!FioBinAdd(fp, idx, binMarks, tocIdx)) {
      FioPeakDel(fp, binMarks); return FALSE; }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int FioBinSub(fp, idx)
     Fio_t fp;
     int idx;
{
  BinVec_t bp= FioBin(fp);
  PVec_t mp= FioBinMrk(fp);
  FioPeakDecIdx(fp, PVecX2(mp,idx));
  VecDelete(bp, idx, 1);
  VecDelete(mp, idx, 1);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FioBinAdd(fp, idx, binMarks, tocIdx)
     Fio_t fp;
     int idx;
     Pair_t *binMarks;
     int tocIdx;
{
  BinVec_t bp= FioBin(fp);
  PVec_t mp= FioBinMrk(fp);
  if (!VecInsert(bp, idx, 1)) {
    FiosetErrNo(FIOALLOC); return FALSE; }
  else if (!VecInsert(mp, idx, 1)) {
    VecDelete(bp, idx, 1); FiosetErrNo(FIOALLOC); return FALSE; }

  FioPeakIncIdx(fp, PairX2(binMarks));

  PVecX1(mp,idx)= PairX1(binMarks);
  PVecX2(mp,idx)= PairX2(binMarks);
  BinPTocIdx(bp,idx)= tocIdx;
  BinPTime(bp,idx)= time(NULL);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FioBinSplit(fp, idx, binMarks)
     Fio_t fp;
     int idx;
     Pair_t *binMarks;
{
  BinVec_t bp= FioBin(fp);
  PVec_t mp= FioBinMrk(fp);
  if (!VecInsert(bp, idx+1, 1)) { FiosetErrNo(FIOALLOC); return FALSE; }
  else if (!VecInsert(mp, idx, 1)) {
    VecDelete(bp, idx, 1); FiosetErrNo(FIOALLOC); return FALSE; }
  FioPeakIncIdx(fp, PairX2(binMarks));
  PVecX2(mp,idx+1)= PVecX2(mp,idx);
  PVecX1(mp,idx+1)= PairX2(binMarks);
  BinPTocIdx(bp,idx+1)= BinPTocIdx(bp,idx);
  BinPTime(bp,idx+1)= time(NULL);
  PVecX2(mp,idx)= PairX1(binMarks);
  BinPTime(bp,idx)= time(NULL);
  return TRUE;
}
/* ---------------------------------------------------------------------- */

/* =============== */
/* update FioPea_t */
/* =============== */

static int FioPeakDecIdx(fp, limit)
     Fio_t fp;
     double limit;
{
  FioPeaVec_t pp= FioPea(fp);
  double ge= limit + 0.5; /* upper limit of bin */
  int i;

  for (i= VecNum(pp)-1; i >=0 && ge < FioPeaPChn(pp,i); i--) {
    FioPeaPBinIdx(pp,i)--; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FioPeakIncIdx(fp, limit)
     Fio_t fp;
     double limit;
{
  FioPeaVec_t pp= FioPea(fp);
  double ge= limit + 0.5; /* upper limit of bin */
  int i;

  for (i= VecNum(pp)-1; i >=0 && ge < FioPeaPChn(pp,i); i--) {
    FioPeaPBinIdx(pp,i)++; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FioPeakDel(fp, binMarks)
     Fio_t fp;
     Pair_t *binMarks;
{
  FioPeaVec_t pp= FioPea(fp);
  int first, last= -1;
  int num= VecNum(pp);

  { int i;
    double ge= (double)PairX2(binMarks)+0.5; /* upper limit of bin */
    for (i= num-1; i >= 0; i--) {
      if (FioPeaPChn(pp,i) <= ge) { last= i; break; } } }

  if (last < 0) { return TRUE; }

  { int i;
    double lt= (double)PairX1(binMarks)-0.5; /* lower limit of bin */
    first= last+1;
    for (i= last; i >= 0; i--) {
      if (FioPeaPChn(pp,i) < lt) { first= i+1; break; } } }

  if (first <= last) {
    if (FioFullPeaks(fp)) {
      int i;
      for (i= first; i<=last; i++) { Free(FioPeaPPar(pp,i)); } }

    VecDelete(pp, first, last-first+1); }

  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FioPeakInsert(fp, idx, binMarks, ftp)
     Fio_t fp;
     int idx;
     Pair_t *binMarks;
     Fit_t ftp;
{
  double lt= PairX1(binMarks)-0.5;
  double ge= PairX2(binMarks)+0.5;
  int peakNum= 0;
  int peakIdx[FIT_PEAK];
  Par_t pp= FitPar(ftp);
  int num= VecNum(FioPea(fp));
  FioPeaVec_t pea= FioPea(fp);

  { int i;
    for (i= 0; i< FitPeakNum(ftp); i++) {
      int j= FitPosIdx(i);
      double p= ParVal(pp,j);
      if (lt < p && p <= ge) {
	peakIdx[peakNum]= j;
	peakNum++; } } }

  if (peakNum == 0) { return TRUE; }
  else if (!ParSort(FitPar(ftp), peakIdx, peakNum, PAR_VVAL)) {
    FiosetErrNo(FIOPEAINS); return FALSE; }

  { int i;
    for (i= num-1; i >= 0; i--) {
      if (FioPeaPChn(pea,i) <= ge) {
	if (lt < FioPeaPChn(pea,i)) { FiosetErrNo(FIOINTERN); return FALSE; }
	break; } }
    i++;

    if (!VecInsert(pea, i, peakNum)) {
      FiosetErrNo(FIOALLOC); return FALSE; }

    { int j;
      for (j= 0; j< peakNum; j++, i++) {
	FioPeaPPar(pea,i)= NULL;
	FioPeaPChn(pea,i)= ParVal(pp,peakIdx[j]);
	FioPeaPPeakIdx(pea,i)= FitPeakIdx(peakIdx[j]);
	FioPeaPBinIdx(pea,i)= idx;
	if (FioFullPeaks(fp)) {
	  if (!(FioPeaPPar(pea,i)=
		FitPeakFunctionSetup(NULL, ftp, pp, FioPeaPPeakIdx(pea,i)))) {
	    FiosetErrNo(FIOALLOC); } } } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int FioPeakRead(fp, ftp)
     Fio_t fp;
     Fit_t ftp;
{
  FioPeaVec_t pea= FioPea(fp);
  BinVec_t bin= FioBin(fp);
  int peaIdx;
  int binIdx= -1;
  int ret_val= TRUE;

  FioFullPeaks(fp)= TRUE;
  for (peaIdx= 0; peaIdx< VecNum(pea); peaIdx++) {
    if (FioPeaPPar(pea,peaIdx)) { continue; }
    else if (binIdx != FioPeaPBinIdx(pea,peaIdx)) {
      int tocIdx;
      int b= FioPeaPBinIdx(pea,peaIdx);
      tocIdx= BinPTocIdx(bin,b);
      if (!FioReadFit(fp, tocIdx, ftp)) { ret_val= FALSE; continue; }
      binIdx= b; }
    { int p= FioPeaPPeakIdx(pea,peaIdx);
      if (!(FioPeaPPar(pea,peaIdx)=
	    FitPeakFunctionSetup(NULL, ftp, FitPar(ftp), p))) {
	FiosetErrNo(FIOALLOC); ret_val= FALSE; } } }
  return ret_val;
}
/* ---------------------------------------------------------------------- */
FioCurPeak_t FioNextPeak(fio, x1, x2, cur)
     Fio_t fio;
     double x1;
     double x2;
     FioCurPeak_t cur;
{
  if (fio) {
    FioPeaVec_t pp= FioPea(fio);
    SeaPeakList sp= FioSea(fio);
    PVec_t mp= FioBinMrk(fio);
    int p;
    SeaPeak s;

    if (!cur) {
      cur= FioCur(fio);
      FioCurPeakTyp(cur)	= FIO_PNONE;
      FioCurPeakPeaIdx(cur)	= -1;
      FioCurPeakSeaPeak(cur)	= NULL; }

    p= FioCurPeakPeaIdx(cur);
    s= FioCurPeakSeaPeak(cur);

    p= (p < 0) ? 0 : p+1;
    s= (s ? SeaNext(s) : (sp ? SeaListFirst(sp) : NULL));

    while (p< VecNum(pp) && FioPeaPChn(pp,p)< x1) { p++; }
    while (s && SeaPos(s)< x1) { s= SeaNext(s); }

    if (p < VecNum(pp)) {
      if (p > 0) {		/* skip searched peaks in last bin */
	int binIdx= FioPeaPBinIdx(pp,p-1);
	double lt= PVecX1(mp,binIdx) - 0.5;
	double ge= PVecX2(mp,binIdx) + 0.5;
	while (s && lt< SeaPos(s) && SeaPos(s) <= ge) { s= SeaNext(s); } }
      { int binIdx= FioPeaPBinIdx(pp,p); /* skip searched peaks in bin */
	double lt= PVecX1(mp,binIdx) - 0.5; /* skip searched peaks in bin */
	double ge= PVecX2(mp,binIdx) + 0.5;
	while (s && lt< SeaPos(s) && SeaPos(s) <= ge) { s= SeaNext(s); } }
      if (s) {
	if (FioPeaPChn(pp,p)<= SeaPos(s) &&
	    (FioPeaPChn(pp,p) <= x2 || x2 < x1)) {
	  FioCurPeakPeaIdx(cur)	= p;
	  FioCurPeakSeaPeak(cur)= SeaPrev(s);
	  FioCurPeakTyp(cur)	= FIO_PBIN;
	  return cur; }
	else if (SeaPos(s) <= x2 || x2 < x1) {
	  FioCurPeakPeaIdx(cur)	= p-1;
	  FioCurPeakSeaPeak(cur)= s;
	  FioCurPeakTyp(cur)	= FIO_PSEA;
	  return cur; } }
      else if (FioPeaPChn(pp,p) <= x2 || x2 < x1) {
	FioCurPeakPeaIdx(cur)	= p;
	FioCurPeakSeaPeak(cur)	= s;
	FioCurPeakTyp(cur)	= FIO_PBIN;
	return cur; } }
    else if (s && (SeaPos(s) <= x2 || x2 < x1)) {
      FioCurPeakPeaIdx(cur)	= p;
      FioCurPeakSeaPeak(cur)	= s;
      FioCurPeakTyp(cur)	= FIO_PSEA;
      return cur; }
    FioCurPeakPeaIdx(cur)	= p;
    FioCurPeakSeaPeak(cur)	= s;
    FioCurPeakTyp(cur)		= FIO_PNONE;

    FiosetErrNo(FIONOPEAK); return NULL; }

  else { FiosetErrNo(FIOPOINT); return NULL; }
}
/* ---------------------------------------------------------------------- */
FioCurPeak_t FioPrevPeak(fio, x1, x2, cur)
     Fio_t fio;
     double x1;
     double x2;
     FioCurPeak_t cur;
{
  if (fio) {
    FioPeaVec_t pp= FioPea(fio);
    SeaPeakList sp= FioSea(fio);
    PVec_t mp= FioBinMrk(fio);
    int p;
    SeaPeak s;

    if (!cur) {
      cur= FioCur(fio);
      FioCurPeakTyp(cur)	= FIO_PNONE;
      FioCurPeakPeaIdx(cur)	= -1;
      FioCurPeakSeaPeak(cur)	= NULL; }

    p= FioCurPeakPeaIdx(cur);
    s= FioCurPeakSeaPeak(cur);

    p= (p < 0) ? VecNum(pp)-1 : p-1;
    s= (s ? SeaNext(s) : (sp ? SeaListLast(sp) : NULL));

    while (p>= 0 && FioPeaPChn(pp,p)> x2) { p--; }
    while (s && SeaPos(s)> x2) { s= SeaPrev(s); }

    if (p >= 0) {
      if (p < VecNum(pp)-1) {		/* skip searched peaks in last bin */
	int binIdx= FioPeaPBinIdx(pp,p+1);
	double lt= PVecX1(mp,binIdx) - 0.5;
	double ge= PVecX2(mp,binIdx) + 0.5;
	while (s && lt< SeaPos(s) && SeaPos(s) <= ge) { s= SeaPrev(s); } }
      { int binIdx= FioPeaPBinIdx(pp,p); /* skip searched peaks in bin */
	double lt= PVecX1(mp,binIdx) - 0.5; /* skip searched peaks in bin */
	double ge= PVecX2(mp,binIdx) + 0.5;
	while (s && lt< SeaPos(s) && SeaPos(s) <= ge) { s= SeaPrev(s); } }
      if (s) {
	if (FioPeaPChn(pp,p)> SeaPos(s) &&
	    (FioPeaPChn(pp,p) >= x1 || x2 < x1)) {
	  FioCurPeakPeaIdx(cur)	= p;
	  FioCurPeakSeaPeak(cur)= SeaNext(s);
	  FioCurPeakTyp(cur)	= FIO_PBIN;
	  return cur; }
	else if (SeaPos(s) <= x2 || x2 < x1) {
	  FioCurPeakPeaIdx(cur)	= p+1;
	  FioCurPeakSeaPeak(cur)= s;
	  FioCurPeakTyp(cur)	= FIO_PSEA;
	  return cur; } }
      else if (FioPeaPChn(pp,p) <= x2 || x2 < x1) {
	FioCurPeakPeaIdx(cur)	= p;
	FioCurPeakSeaPeak(cur)	= s;
	FioCurPeakTyp(cur)	= FIO_PBIN;
	return cur; } }
    else if (s && (SeaPos(s) <= x2 || x2 < x1)) {
      FioCurPeakPeaIdx(cur)	= p;
      FioCurPeakSeaPeak(cur)	= s;
      FioCurPeakTyp(cur)	= FIO_PSEA;
      return cur; }
    FioCurPeakPeaIdx(cur)	= p;
    FioCurPeakSeaPeak(cur)	= s;
    FioCurPeakTyp(cur)		= FIO_PNONE;

    FiosetErrNo(FIONOPEAK); return NULL; }

  else { FiosetErrNo(FIOPOINT); return NULL; }
}
/* ---------------------------------------------------------------------- */

/* ================= */
/* read/write Fit_t */
/* ================= */

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/* !!! only 4 byte float/integer !!! */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */

#define WRITE_FLT(rp,f,r) { float F=f; r = r && RecWriteFloat4(rp,F); }
#define READ_FLT(rp,f,r) { float F; r = r && RecReadFloat4(rp,&F); f= F; }
#define WRITE_INT(rp,i,r) { int I= i; r = r && RecWriteInt4(rp,I); }
#define READ_INT(rp,i,r) { int I; r = r && RecReadInt4(rp, &I); i= I; }

static int Fit_w(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int r= TRUE;
  r = r && FitB_w(rfp, fp);
  /*  printf("Fit_W\n");*/
  if (r) {
    int i;			/* translate positions by offset */
    for (i= 0; i< FitPeakNum(fp); i++) {
      FitPPosVal(pp,i) -= FitBgOffset(fp); }

    r = r && FioWriteFitPar(rfp, fp);
    
    for (i= 0; i< FitPeakNum(fp); i++) { /* and back ! */
      FitPPosVal(pp,i) += FitBgOffset(fp); } }

  r = r && Vec_w(rfp, Pair_w, (Vec_t)FitRegMrk(fp));
  r = r && Vec_w(rfp, Pair_w, (Vec_t)FitBgMrk(fp));
      r = r && (FitStat_w(rfp, FitStatBg(fp)) &&
		FitStat_w(rfp, FitStatReg(fp)));
      r = r && (FitFitTyp(fp) != FIT_FTINT ||
		(FitInt_w(rfp, FitIntegral(fp)) &&
		 FitInt_w(rfp, FitBgIntegral(fp)) &&
		 FitInt_w(rfp, FitSubIntegral(fp))));
      r = r && Vec_w(rfp, doubleP_w, (Vec_t)FitBgCovar(fp));
  return r;
}
/* ---------------------------------------------------------------------- */
static int Fit_r(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  int r= TRUE;
  r = (FitReset(fp) &&
       FitB_r(rfp, fp) &&
       FioReadFitPar(rfp, fp) &&
       Vec_r(rfp, Pair_r, (Vec_t)FitRegMrk(fp)) &&
       Vec_r(rfp, Pair_r, (Vec_t)FitBgMrk(fp)) &&
	 FitStat_r(rfp, FitStatBg(fp)) &&
	 FitStat_r(rfp, FitStatReg(fp)) &&
	 (FitFitTyp(fp) != FIT_FTINT ||
	  (FitInt_r(rfp, FitIntegral(fp)) &&
	   FitInt_r(rfp, FitBgIntegral(fp)) &&
	   FitInt_r(rfp, FitSubIntegral(fp)))) &&
       Vec_r(rfp, doubleP_r, (Vec_t)FitBgCovar(fp)));
  { int i;			/* translate positions by offset */
    Par_t pp= FitPar(fp);
    for (i= 0; i< FitPeakNum(fp); i++) {
      FitPPosVal(pp,i) += FitBgOffset(fp); } }
  return r;
}
/* ---------------------------------------------------------------------- */

/* ================== */
/* read/write FitB_t */
/* ================== */

static int FitB_w(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  int r= TRUE;

  WRITE_INT(rfp, FitFitTyp(fp), r);

  WRITE_INT(rfp, FitRes(fp), r);

  WRITE_INT(rfp, FitBgTyp(fp), r);
  WRITE_INT(rfp, FitBgGrad(fp), r);
  WRITE_INT(rfp, FitBgExpNum(fp), r);
  WRITE_FLT(rfp, FitBgOffset(fp), r);

  WRITE_INT(rfp, FitWCorr(fp), r);
  WRITE_INT(rfp, FitPCorr(fp), r);
  WRITE_INT(rfp, FitTlCorr(fp), r);
  WRITE_INT(rfp, FitTrCorr(fp), r);
  WRITE_INT(rfp, FitShCorr(fp), r);
  WRITE_INT(rfp, FitSwCorr(fp), r);

  WRITE_INT(rfp, FitPeakTyp(fp), r);

  WRITE_INT(rfp, FitTlE(fp), r);
  WRITE_INT(rfp, FitTrE(fp), r);

  WRITE_INT(rfp, FitPeakNum(fp), r);
  WRITE_INT(rfp, FitPeakSortTyp(fp), r);

  return r;
}
/* ---------------------------------------------------------------------- */
static int FitB_r(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  int r= TRUE;

  READ_INT(rfp, FitFitTyp(fp), r);

  READ_INT(rfp, FitRes(fp), r);

  READ_INT(rfp, FitBgTyp(fp), r);
  READ_INT(rfp, FitBgGrad(fp), r);
  READ_INT(rfp, FitBgExpNum(fp), r);
  READ_FLT(rfp, FitBgOffset(fp), r);

  READ_INT(rfp, FitWCorr(fp), r);
  READ_INT(rfp, FitPCorr(fp), r);
  READ_INT(rfp, FitTlCorr(fp), r);
  READ_INT(rfp, FitTrCorr(fp), r);
  READ_INT(rfp, FitShCorr(fp), r);
  READ_INT(rfp, FitSwCorr(fp), r);

  READ_INT(rfp, FitPeakTyp(fp), r);
  if (r) {
    switch (FitPeakTyp(fp)) {
    case FIT_PKCA: FitFct(fp)= FitFctCA; break;
    case FIT_PKAE: FitFct(fp)= FitFctAE; break;
    default: FitsetErrNo(FIOPKTYP); return FALSE; } }

  READ_INT(rfp, FitTlE(fp), r);
  READ_INT(rfp, FitTrE(fp), r);

  READ_INT(rfp, FitPeakNum(fp), r);
  READ_INT(rfp, FitPeakSortTyp(fp), r);

  return r;
}
/* ---------------------------------------------------------------------- */

/* ==================== */
/* read/write FitStat_t */
/* ==================== */

static int FitStat_w(rfp, fp)
     RFILE *rfp;
     FitStat_t *fp;
{
  int r= TRUE;

  WRITE_INT(rfp, FitStatTime(fp), r);
  if (FitStatTime(fp)) {
    WRITE_FLT(rfp, FitStatMeasure(fp), r);
    WRITE_FLT(rfp, FitStatDMeasure(fp), r);
    WRITE_FLT(rfp, FitStatRms(fp), r);

    WRITE_INT(rfp, FitStatMaxIterations(fp), r);
    WRITE_INT(rfp, FitStatIterations(fp), r);

    WRITE_INT(rfp, FitStatMeasureTyp(fp), r);

    WRITE_INT(rfp, FitStatFreedom(fp), r);

    WRITE_FLT(rfp, FitStatLambda(fp), r);
    WRITE_FLT(rfp, FitStatMinLambda(fp), r);
    WRITE_FLT(rfp, FitStatMaxLambda(fp), r);
    WRITE_FLT(rfp, FitStatStartLambda(fp), r); }
  return r;
}
/* ---------------------------------------------------------------------- */
static int FitStat_r(rfp, fp)
     RFILE *rfp;
     FitStat_t *fp;
{
  int r= TRUE;

  FitStatReset(fp);
  READ_INT(rfp, FitStatTime(fp), r);
  if (FitStatTime(fp)) {
    READ_FLT(rfp, FitStatMeasure(fp), r);
    READ_FLT(rfp, FitStatDMeasure(fp), r);
    READ_FLT(rfp, FitStatRms(fp), r);

    READ_INT(rfp, FitStatMaxIterations(fp), r);
    READ_INT(rfp, FitStatIterations(fp), r);

    READ_INT(rfp, FitStatMeasureTyp(fp), r);

    READ_INT(rfp, FitStatFreedom(fp), r);

    READ_FLT(rfp, FitStatLambda(fp), r);
    READ_FLT(rfp, FitStatMinLambda(fp), r);
    READ_FLT(rfp, FitStatMaxLambda(fp), r);
    READ_FLT(rfp, FitStatStartLambda(fp), r); }
  return r;
}
/* ---------------------------------------------------------------------- */

/* ==================== */
/* read/write FitInt_t */
/* ==================== */

static int FitInt_w(rfp, fip)
     RFILE *rfp;
     FitInt_t *fip;
{
  int r= TRUE;

  WRITE_FLT(rfp, FitIntI(fip), r);
  WRITE_FLT(rfp, FitIntDI(fip), r);
  WRITE_FLT(rfp, FitIntM1(fip), r);
  WRITE_FLT(rfp, FitIntDM1(fip), r);
  WRITE_FLT(rfp, FitIntM2(fip), r);
  WRITE_FLT(rfp, FitIntDM2(fip), r);
  WRITE_FLT(rfp, FitIntM3(fip), r);
  WRITE_FLT(rfp, FitIntDM3(fip), r);
  
  return r;
}
/* ---------------------------------------------------------------------- */
static int FitInt_r(rfp, fip)
     RFILE *rfp;
     FitInt_t *fip;
{
  int r= TRUE;

  READ_FLT(rfp, FitIntI(fip), r);
  READ_FLT(rfp, FitIntDI(fip), r);
  READ_FLT(rfp, FitIntM1(fip), r);
  READ_FLT(rfp, FitIntDM1(fip), r);
  READ_FLT(rfp, FitIntM2(fip), r);
  READ_FLT(rfp, FitIntDM2(fip), r);
  READ_FLT(rfp, FitIntM3(fip), r);
  READ_FLT(rfp, FitIntDM3(fip), r);
  
  return r;
}
/* ---------------------------------------------------------------------- */

/* ================ */
/* read/write Vec_t */
/* ================ */

static int Vec_w(rfp, write_fct, vp)
     RFILE *rfp;
     pfi_t write_fct;
     Vec_t vp;
{
  int r= TRUE;
  int i;
  char *p= (char *)VecP(vp);

  WRITE_INT(rfp, VecNum(vp), r);
  for (i= 0; r && i< VecNum(vp); i++, p += VecTypSiz(vp)) {
    r= write_fct(rfp, p);
  }
  if (!r) {
    FiosetErrNo(FIOCOMP);
  }
  return r;
}
/* ---------------------------------------------------------------------- */
static int Vec_r(rfp, read_fct, vp)
     RFILE *rfp;
     pfi_t read_fct;
     Vec_t vp;
{
  int r= TRUE;
  int i, num;
  char *p;

  READ_INT(rfp, num, r);
  r = r && VecAlloc(vp, num);
  VecNum(vp)= num;
  for (i= 0, p= (char *)VecP(vp);
       r && i< num;
       i++, p += VecTypSiz(vp)) {
    r= read_fct(rfp, p); }
  if (!r) { FiosetErrNo(FIOUNCOMP); }
  return r;
}
/* ---------------------------------------------------------------------- */
static int doubleP_w(rfp, p)
     RFILE *rfp;
     double *p;
{
  int r= TRUE;
  WRITE_FLT(rfp, *p, r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int doubleP_r(rfp, p)
     RFILE *rfp;
     double *p;
{
  int r= TRUE;
  READ_FLT(rfp, *p, r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int ParV_w(rfp, pvp)
     RFILE *rfp;
     ParV_t *pvp;
{
  int r= TRUE;
  WRITE_FLT(rfp, ParVVal(pvp), r);
  WRITE_FLT(rfp, ParVErr(pvp), r);
  WRITE_FLT(rfp, ParVUserVal(pvp), r);
  WRITE_INT(rfp, ParVMode(pvp), r);
  if (ParVMcorr(pvp)) {
    int j;
    for (j= 0; r && j< PAR_COR; j++) {
      WRITE_FLT(rfp, ParVCorr(pvp)[j], r); } }
  WRITE_INT(rfp, ParVStat(pvp), r);
  WRITE_INT(rfp, ParVUserStat(pvp), r);
  WRITE_INT(rfp, ParVUserTyp(pvp), r);
  if (ParVMcorr(pvp)) {
    WRITE_INT(rfp, ParVCorrIndex(pvp), r); }
  return r;
}
/* ---------------------------------------------------------------------- */
static int ParV_r(rfp, pvp)
     RFILE *rfp;
     ParV_t *pvp;
{
  int r= TRUE;
  READ_FLT(rfp, ParVVal(pvp), r);
  READ_FLT(rfp, ParVErr(pvp), r);
  READ_FLT(rfp, ParVUserVal(pvp), r);
  READ_INT(rfp, ParVMode(pvp), r);
  if (ParVMcorr(pvp)) {
    int j;
    for (j= 0; r && j< PAR_COR; j++) {
      READ_FLT(rfp, ParVCorr(pvp)[j], r); } }
  READ_INT(rfp, ParVStat(pvp), r);
  READ_INT(rfp, ParVUserStat(pvp), r);
  READ_INT(rfp, ParVUserTyp(pvp), r);
  if (ParVMcorr(pvp)) {
    READ_INT(rfp, ParVCorrIndex(pvp), r); }
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioWriteFitPar(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  int i;
  int r= TRUE;
  Par_t pp= FitPar(fp);

  WRITE_INT(rfp, ParNum(pp), r);
  for (i= 0; r && i< FitBgGrad(fp); i++) {
    r = r && ParV_w(rfp, ParV(pp,i)); }
  for (i= FIT_BGGRD; i< FIT_BGGRD + 2 * FitBgExpNum(fp); i++) {
    r = r && ParV_w(rfp, ParV(pp,i)); }
  for (i= FIT_1STPEAK; i< ParNum(pp); i++) {
    r = r && ParV_w(rfp, ParV(pp,i)); }
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioReadFitPar(rfp, fp)
     RFILE *rfp;
     Fit_t fp;
{
  int i;
  int r;
  Par_t pp= FitPar(fp);
  ParV_t pv;
  int parnum;

  r= FitSetupBg(fp);
  READ_INT(rfp, parnum, r);
  if (FIT_BGGRD < FitBgGrad(fp) || FIT_BGMEXP < FitBgExpNum(fp)) {
    FiosetErrNo(FIOBGSTO); return FALSE; }
  for (i= 0; r && i< FitBgGrad(fp); i++) {
    r = r && ParV_r(rfp, &pv);
    r = r && ParSet(pp, i, &pv, FALSE); }
  for (i= FIT_BGGRD; i< FIT_BGGRD + 2 * FitBgExpNum(fp); i++) {
    r = r && ParV_r(rfp, &pv);
    r = r && ParSet(pp, i, &pv, FALSE); }
  for (i= FIT_1STPEAK; i< parnum; i++) {
    r = r && ParV_r(rfp, &pv);
    r = r && ParSet(pp, i, &pv, FALSE); }
  ParNum(pp)= parnum;
  return r;
}
/* ---------------------------------------------------------------------- */
static int Pair_w(rfp, p)
     RFILE *rfp;
     Pair_t *p;
{
  int r= TRUE;
  WRITE_FLT(rfp, PairX1(p), r);
  WRITE_FLT(rfp, PairX2(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int Pair_r(rfp, p)
     RFILE *rfp;
     Pair_t *p;
{
  int r= TRUE;
  READ_FLT(rfp, PairX1(p), r);
  READ_FLT(rfp, PairX2(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioTailWriteSeaPeakList(fp, recTyp)
     Fio_t fp;
     int recTyp;
{
  RFILE *rfp= FioRfp(fp);
  SeaPeakList pl= FioSea(fp);
  int r= TRUE;

  /*  printf("FioTailWriteSeaPeakList\n");*/
  if (!RecTailWriteStart(rfp, recTyp)) {
    FiosetErrNo(FIOWRITE);
    return FALSE;
  }
  if (!pl) {
    WRITE_INT(rfp, 0, r);
  }
  else {
    SeaPeak p;

    WRITE_INT(rfp, SeaListNum(pl), r);
    ForAllSeaPeaks(pl, p) {
      if (!(r= SeaPeak_w(rfp, p))) {
	FiosetErrNo(FIOWRITE);
	return FALSE;
      }
    }
  }

  if (RecTailWriteFinish(rfp)) {
    return TRUE;
  }

  FiosetErrNo(FIOWRITE);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int FioTailReadSeaPeakList(fp, recTyp)
     Fio_t fp;
     int recTyp;
{
  RFILE *rfp= FioRfp(fp);

  /*  printf("FioTailReadSeaPeakList\n");*/
  SeaFreePeakList(FioSea(fp));
  FioSea(fp)= NULL;

  if (!RecTailReadStart(rfp, recTyp)) {
    FiosetErrNo(FIOREAD);
    return FALSE;
  }
  else {
    int num;
    int r= TRUE;
    SeaPeakList pl= NULL;

    READ_INT(rfp, num, r);

    if (!r) {
      FiosetErrNo(FIOREAD);
      return FALSE;
    }

    if (!num) {
      return TRUE;
    }
    else if (!(pl= SeaCreatePeakList())) {
      FiosetErrNo(FIOREAD);
      return FALSE;
    }
    else {
      int i;
      for (i= 0; r && i< num; i++) {
	SeaPeakRec p;

	r= (SeaPeak_r(rfp, &p) &&
	    SeaEnterPeak(pl, 
			 SeaPos(&p),
			 SeaVol(&p), SeaDVol(&p),
			 SeaProb(&p),
			 SeaAmpl(&p), SeaDAmpl(&p),
			 SeaBg(&p), SeaDBg(&p),
			 SeaChi(&p),
			 SeaWidth(&p)) != NULL);
      }
      if (!r) {
	SeaFreePeakList(pl);
	FiosetErrNo(FIOREAD);
	return FALSE;
      }
      else {
	FioSea(fp)= pl;
	RecTailReadFinish(rfp);
	return TRUE;
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
static int SeaPeak_w(rfp, p)
     RFILE *rfp;
     SeaPeak p;
{
  int r= TRUE;
  WRITE_FLT(rfp, SeaPos(p), r);
  WRITE_FLT(rfp, SeaProb(p), r);
  WRITE_FLT(rfp, SeaAmpl(p), r);
  WRITE_FLT(rfp, SeaDAmpl(p), r);
  WRITE_FLT(rfp, SeaBg(p), r);
  WRITE_FLT(rfp, SeaDBg(p), r);
  WRITE_FLT(rfp, SeaVol(p), r);
  WRITE_FLT(rfp, SeaDVol(p), r);
  WRITE_FLT(rfp, SeaChi(p), r);
  WRITE_FLT(rfp, SeaWidth(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int SeaPeak_r(rfp, p)
     RFILE *rfp;
     SeaPeak p;
{
  int r= TRUE;
  /*  printf("SeaPeak_r\n");*/
  READ_FLT(rfp, SeaPos(p), r);
  READ_FLT(rfp, SeaProb(p), r);
  READ_FLT(rfp, SeaAmpl(p), r);
  READ_FLT(rfp, SeaDAmpl(p), r);
  READ_FLT(rfp, SeaBg(p), r);
  READ_FLT(rfp, SeaDBg(p), r);
  READ_FLT(rfp, SeaVol(p), r);
  READ_FLT(rfp, SeaDVol(p), r);
  READ_FLT(rfp, SeaChi(p), r);
  READ_FLT(rfp, SeaWidth(p), r);
  SeaNext(p)= NULL;
  SeaPrev(p)= NULL;
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioToc_w(rfp, p)
     RFILE *rfp;
     FioToc_t *p;
{
  int r= TRUE;
  WRITE_INT(rfp, FioTocTyp(p), r);
  WRITE_INT(rfp, FioTocPos(p), r);
  WRITE_INT(rfp, FioTocTime(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioToc_r(rfp, p)
     RFILE *rfp;
     FioToc_t *p;
{
  int r= TRUE;
  READ_INT(rfp, FioTocTyp(p), r);
  READ_INT(rfp, FioTocPos(p), r);
  READ_INT(rfp, FioTocTime(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int Bin_w(rfp, p)
     RFILE *rfp;
     Bin_t *p;
{
  int r= TRUE;
  WRITE_INT(rfp, BinTocIdx(p), r);
  WRITE_INT(rfp, BinTime(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int Bin_r(rfp, p)
     RFILE *rfp;
     Bin_t *p;
{
  int r= TRUE;
  READ_INT(rfp, BinTocIdx(p), r);
  READ_INT(rfp, BinTime(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioPea_w(rfp, p)
     RFILE *rfp;
     FioPea_t *p;
{
  int r= TRUE;
  WRITE_FLT(rfp, FioPeaChn(p), r);
  WRITE_INT(rfp, FioPeaBinIdx(p), r);
  WRITE_INT(rfp, FioPeaPeakIdx(p), r);
  return r;
}
/* ---------------------------------------------------------------------- */
static int FioPea_r(rfp, p)
     RFILE *rfp;
     FioPea_t *p;
{
  int r= TRUE;
  READ_FLT(rfp, FioPeaChn(p), r);
  READ_INT(rfp, FioPeaBinIdx(p), r);
  READ_INT(rfp, FioPeaPeakIdx(p), r);
  FioPeaPar(p)= NULL;
  return r;
}
/* ---------------------------------------------------------------------- */
