/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsFitSetup.c,v 1.5 1994/12/21 17:12:36 vs Exp vs $ */
#define _vsFitSetup_c

#include "vsTypes.h"

#include "vsCal.h"
#include "vsFit.h"
#include "vsSpectra.h"

#include "vsFitSetup.h"
#include "vsFitSetupP.h"
/* ------------------------------------------------------------------------- */
char *FSError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* FSNOERR   */ "vsFitSetup: no error",
  /* FSPOINT   */ "vsFitSetup: illegal NULL-pointer",
  /* FSNOMARK  */ "vsFitSetup: no marker defined",
  /* FSSORTPK  */ "vsFitSetup: can not sort peak positions",
  /* FSNOPEAK  */ "vsFitSetup: no peaks found",
  /* FSPKSUM0  */ "vsFitSetup: sum of peaks == 0.0 -- no step",
  /* FSNOBG    */ "vsFitSetup: can not evaluate background",
  /* FS0AMPL   */ "vsFitSetup: no peak with non-zero amplitude",
  /* FSNOWIDTH */ "vsFitSetup: can not determine width",
  /* FSNOSPC   */ "vsFitSetup: no spectrum defined -- only position init",
  /* FSPKTYP:  */ "vsFitSetup: illegal peak function typ",
  /* default:  */ "vsFitSetup: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
/* ===================== */
/* initialize background */
/* ===================== */

/* ARGSUSED */
int FSBgSetup(fit, mrk, spc, cal)
     Fit_t fit;
     PVec_t mrk;
     Spc_t spc;
     Cal_t cal;
{
  if (fit == NULL || mrk == NULL) { /* DUKE */
    FSsetErrNo(FSPOINT); return FALSE; }
  return FSBgSetupPar(fit, mrk, spc); /* DUKE: function handles spc==NULL */
}
/* ---------------------------------------------------------------------- */
static int FSBgSetupUserStat(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int i;
  for (i= 0; i< FitBgGrad(fp); i++) {
    if (ParVUserStat(FitPBgCoef(pp,i)) == FIT_USINIT) {
      ParVUserStat(FitPBgCoef(pp,i))= FIT_USUNDEF; } }
  for (i= FIT_BGGRD; i< FIT_BGGRD + 2 * FitBgExpNum(fp); i++) {
    if (ParVUserStat(FitPBgCoef(pp,i)) == FIT_USINIT) {
      ParVUserStat(FitPBgCoef(pp,i))= FIT_USUNDEF; } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FSBgSetupPar(fp, mp, sp)
     Fit_t fp;
     PVec_t mp;
     Spc_t sp;
{
  Par_t pp= FitPar(fp);
  int i;
  int ini= TRUE;

  FSBgSetupUserStat(fp);
  for (i= 0; i< FitBgGrad(fp); i++) {
    ParV_t *p= FitPBgCoef(pp,i);

    if (ParVUserStat(p) == FIT_USUNDEF) {
      ParVMode(p) = PAR_MFREE;
      ParVUserStat(p) = FIT_USINIT;
      ParVVal(p) = 0.0;
      ParVErr(p) = 0.0; }
    else { ini= FALSE; } }
  for (i= 0; i< FitBgExpNum(fp); i++) {
    ParV_t *p= FitPBgCoef(pp,FIT_BGGRD + 2 * i);
    ini= FALSE;			/* no linear initialization  */
    if (ParVUserStat(p) == FIT_USUNDEF) {
      ParVMode(p) = PAR_MFREE;
      ParVUserStat(p) = FIT_USINIT;
      ParVVal(p) = 1.0;
      ParVErr(p) = 0.0; }

    p= FitPBgCoef(pp,FIT_BGGRD + 2 * i + 1);
    if (ParVUserStat(p) == FIT_USUNDEF) {
      ParVMode(p) = PAR_MFREE;
      ParVUserStat(p) = FIT_USINIT;
      ParVVal(p) = 0.0;
      ParVErr(p) = 0.0; } }
  if (!FSOffsetSet(fp, mp)) { return TRUE; }
  else {
    if (ini) {
      double xl;
      double xg;
      double yl;
      double yg;
      SpcReal_t *sb;

      if (sp){
	xl = PVecX1(mp,0);
	xg = PVecX2(mp,VecNum(mp)-1);
	sb= SpcBufReal(SpcSpc(sp));
	yl = sb[MAX(0, NINT(xl))];
	yg = sb[MIN(SpcNum(sp)-1, NINT(xg))];}
      else{
	xl = 0;
	xg = 0;
	yl = 0;
	yg = 0;}

      yl = sb[MAX(0, NINT(xl))];
      yg = sb[MIN(SpcNum(sp)-1, NINT(xg))];
      xl -= FitBgOffset(fp);
      xg -= FitBgOffset(fp);

      if (xl == xg || FitBgGrad(fp) == 1) { /* one channel or constant bg */
	FitPBgCoefVal(pp,0) = (yl + yg) * 0.5; }
      else {			/* linear start bg */
	FitPBgCoefVal(pp,1) = (yg - yl) / (xg - xl);
	FitPBgCoefVal(pp,0) = yl - FitPBgCoefVal(pp,1) * xl; } }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
/* ======================================= */
/* initialize offset for numeric stability */
/* ======================================= */

static int FSOffsetSet(fp, mp)
     Fit_t fp;
     PVec_t mp;
{
  Par_t pp= FitPar(fp);
  int i;

  for (i=1; i< FitBgGrad(fp); i++) { /* constant bg is offset independent */
    if (ParVal(pp,i) != 0.0 &&	/* bg at least not constant */
	(ParMhold(pp,i) ||	/* fixed */
	 ParUserStat(pp,i) == FIT_USUSER || /* user defined */
	 ParUserStat(pp,i) == FIT_USFIT)) { /* fitted */
      return TRUE; } }

  for (i=FIT_BGGRD; i< FIT_BGGRD + FitBgExpNum(fp) * 2; i++) { 
    if ((ParMhold(pp,i) ||	/* fixed */
	 ParUserStat(pp,i) == FIT_USUSER || /* user defined */
	 ParUserStat(pp,i) == FIT_USFIT)) { /* fitted */
      return TRUE; } }
    
  if (VecNum(mp) != 0) {	/* legal marker */
    FitBgOffset(fp)= NPVecX1(mp,0); return TRUE; }
  else {			/* illegal marker */
    FitBgOffset(fp)= 0.0; FSsetErrNo(FSNOMARK); return FALSE; }
}
/* ---------------------------------------------------------------------- */
/* ======================= */
/* setup region parameters */
/* ======================= */

/* ARGSUSED */
int FSRegSetup(fit, mrk, spc, cal)
     Fit_t fit;
     PVec_t mrk;
     Spc_t spc;
     Cal_t cal;
{
  if (FitFitTyp(fit) != FIT_FTFIT) {
    FitFitTyp(fit)= FIT_FTFIT;
    FitTime(fit)= 0; }

/*  if (!VecNum(FitRegMrk(fit))) { FSsetErrNo(FSNOMARK); return FALSE; } */

  if (!FSRegSetupPar(fit, spc, cal)) { return FALSE; }
  else if (!FitPeakSort(fit, FitPeakSortTyp(fit))) {
    FSsetErrNo(FSSORTPK); return FALSE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FSPeakSetupUserStat(fp)
     Fit_t fp;
{
  Par_t pp= FitPar(fp);
  int i;
  for (i= FIT_1STPEAK; i< ParNum(pp); i++) {
    if (ParUserTyp(pp,i) == FIT_UTVOL &&
	(ParUserStat(pp,i) == FIT_USINIT ||
	 ParUserStat(pp,i) == FIT_USSTART)) {
      ParUserStat(pp,i)= FIT_USUNDEF; }
    else if (ParUserStat(pp,i) == FIT_USINIT) {
      ParUserStat(pp,i)= FIT_USUNDEF; } }

  return TRUE;
}
/* ---------------------------------------------------------------------- */
/* ============================ */
/* initialize region parameters */
/* ============================ */

static int FSRegSetupPar(fit, spc, cal)
     Fit_t fit;
     Spc_t spc;
     Cal_t cal;
{
  FSPeakSetupUserStat(fit);
  return (FSBgSetup(fit, FitRegMrk(fit), spc, NULL) && /* if not yet fitted */
	  FSInitPeaks(fit, spc, cal) && /* initialize peaks */
	  FSInitTails(fit, cal) && /* initialize tails (width needed) */
	  FSInitStep(fit, spc, cal)); /* initialize step (amplitudes needed) */
}
/* ---------------------------------------------------------------------- */
/* ========================== */
/* initialize peak parameters */
/* ========================== */

static int FSInitPeaks(fit, spc, cal)
     Fit_t fit;
     Spc_t spc;
     Cal_t cal;
{
  if (FitPeakNum(fit) == 0) { return TRUE; }
  else if (FSInitPos(fit)) {		/* initialize start values of positions */
    if (spc == NULL) { FSsetErrNo(FSNOSPC); return FALSE; }
    return (FSInitAmpl(fit, spc) && /* amplitudes are needed for volumes */
	    FSInitWidth(fit, spc, cal) && /* uses width of highest peak */
	    FSInitVol(fit)); }	/* gauss peaks of given width and amplitude */
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int FSInitPos(fit)
     Fit_t fit;
{				/* user given start value */
  Par_t pp= FitPar(fit);
  int i;			/* init positions */
  for (i= 0; i< FitPeakNum(fit); i++) {
    int j= FitPosIdx(i);
    if (ParSused(pp,j) &&
	ParUserStat(pp,j) == FIT_USUNDEF) {
      ParV_t pv;
      ParGet(pp, j, &pv);
      ParVVal(&pv) = ParVUserVal(&pv); /* set value and status */
      ParVMode(&pv) = PAR_MFREE;
      ParVUserStat(&pv) = FIT_USINIT;
      ParSet(pp, j, &pv, FALSE); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FSInitAmpl(fit, spc)
     Fit_t fit;
     Spc_t spc;
{				/* ampl= spectrum(pos) - bg */
  Par_t pp= FitPar(fit);
  int i;
  SpcReal_t *b= SpcBufReal(SpcSpc(spc));

  if (!FitBgFunctionSetup(fit, NULL)) { FSsetErrNo(FSNOBG); return FALSE; }

  for (i= 0; i< FitPeakNum(fit); i++) {
    int j= FitVolIdx(i);
    if (ParSused(pp,j)) {
      if (ParUserStat(pp,j) == FIT_USUNDEF) {
	ParV_t pv;

	ParGet(pp, j, &pv);	/* start is value of amplitude */
	ParVUserVal(&pv) = (b[NINT(ParVal(pp,FitPosIdx(i)))] -
			    FitBgFunction(fit, ParVal(pp,FitPosIdx(i))));
	ParVMode(&pv) = PAR_MFREE; /* set stat */
	ParVUserStat(&pv)= FIT_USSTART;
	ParSet(pp, j, &pv, FALSE); } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FSLargestUserAmpl(fit, idx)
     Fit_t fit;
     int *idx;
{
  Par_t pp= FitPar(fit);
  int maxPeakIdx= -1;
  double maxAmpl= 0.0;
  int i;
  for (i= 0; i< FitPeakNum(fit); i++) {
    int j= FitVolIdx(i);
    if (ParUserStat(pp,j) != FIT_USUNDEF &&
	fabs(ParUserVal(pp,j)) > maxAmpl) {
      maxPeakIdx= i;		/* largest peak */
      maxAmpl= fabs(ParUserVal(pp,j)); } }
  if (i < 0) { return FALSE; }
  else { *idx= maxPeakIdx; return TRUE; }
}
/* ---------------------------------------------------------------------- */
static int FSInitWidth(fit, spc, cal)
     Fit_t fit;
     Spc_t spc;
     Cal_t cal;
{				/* width= fwhm(largest peak) */
  Par_t pp= FitPar(fit);
  double width= -1.0;

  if (FitWCorr(fit) >= 0) {	/* use parent width of correlations */
    int i;
    int wc= FitWCorr(fit);
    for (i= 0; i< FitPeakNum(fit); i++) {
      int j= FitWidthIdx(i);
      if (j != wc) {
	if (ParSused(pp,j)) {
	  if (ParUserStat(pp,j) == FIT_USUNDEF) {
	    ParV_t pv;
	    ParGet(pp, j, &pv);
	    ParVMode(&pv) = PAR_MCORR;
	    ParVCorrIndex(&pv) = wc;
	    ParVUserStat(&pv) = FIT_USUSER;
	    ParSet(pp, j, &pv, FALSE); } } }
      else if (ParUserStat(pp,j) != FIT_USUNDEF &&
	       ParUserStat(pp,j) != FIT_USNONE) {
	width= FitPWidthVal(pp,i); } } }

  if (width < 0.0) {		/* use mean value of defined widths */
    int i;
    int widthNum= 0;
    double widthSum= 0.0;
    for (i= 0; i< FitPeakNum(fit); i++) {
      int j= FitWidthIdx(i);
      if (ParSused(pp,j) &&	/* used */
	  !ParMcorr(pp,j) &&	/* free or hold */
	  ParUserStat(pp,j) != FIT_USUNDEF &&
	  ParUserStat(pp,j) != FIT_USNONE) {
	widthNum++;
	widthSum += FitPWidthVal(pp,i); } }
    if (widthNum > 0 && widthSum > 0.0) {
      width= widthSum / widthNum; } }

  if (width < 0.0) {		/* use width of largest peak */
    int peakIdx= -1;

    if (!VecNum(FitRegMrk(fit))) { return TRUE; }

    if (!FSLargestUserAmpl(fit, &peakIdx)) {
      FSsetErrNo(FSNOPEAK); return FALSE; }

    { int maxPeakPos= FitPPosVal(pp,peakIdx);
      double maxAmpl= ABS(ParVUserVal(FitPVol(pp,peakIdx)));
      int rle= MAX(0, NPVecX1(FitRegMrk(fit),0));
      int rge= MIN(SpcNum(spc), NPVecX2(FitRegMrk(fit),0));
      double halfAmpl;
      int w;
      SpcReal_t *b= SpcBufReal(SpcSpc(spc));

      rle= MIN(rle, rge);

      if (!FitBgFunctionSetup(fit, NULL)) { FSsetErrNo(FSNOBG); return FALSE; }

      { int i;			/* find maximum of largest peak */
	for (i= maxPeakPos; i >= rle; i--) {
	  double ampl;
	  ampl = fabs(b[i] - FitBgFunction(fit, (double)i));
	  if (ampl > maxAmpl) { maxAmpl= ampl; maxPeakPos= i; }
	  else { break; } } }
      { int i;
	for (i= maxPeakPos; i <= rge; i++) {
	  double ampl;
	  ampl = fabs(b[i] - FitBgFunction(fit, (double)i));
	  if (ampl > maxAmpl) { maxAmpl= ampl; maxPeakPos= i; }
	  else { break; } } }

      halfAmpl= maxAmpl * 0.5;	/* find fwhm of largest peak */
      for (w=1; ;w++) {
	int jg, jl;
	if ((jg=maxPeakPos + w) < rge &&
	    fabs(b[jg] - FitBgFunction(fit, (double)jg)) < halfAmpl &&
	    fabs(b[jg+1] -
		 FitBgFunction(fit, (double)(jg+1))) < halfAmpl){ break; }
	else if ((jl=maxPeakPos - w) > rle &&
		 fabs(b[jl] - FitBgFunction(fit, (double)jl)) < halfAmpl &&
		 fabs(b[jl+1] -
		      FitBgFunction(fit, (double)(jl-1))) < halfAmpl){
	  break; }
	else if (jg >= rge && jl <= rle) { break; } }

      width = w * FWHM_TO_SIGMA; } }

  if (width < 0.0) { FSsetErrNo(FSNOWIDTH); return FALSE; }

  { int i;			/* set all peak width to designated value */
    for (i= 0; i< FitPeakNum(fit); i++) {
      int j= FitWidthIdx(i);
      if (ParSused(pp,j)) {
	if (ParUserStat(pp,j) == FIT_USUNDEF ||
	    ParUserStat(pp,j) == FIT_USCALUNDEF) {
	  ParV_t pv;
	  ParGet(pp, j, &pv);
	  if (ParVUserStat(&pv) == FIT_USCALUNDEF) { /* calibrated width */
	    ParVUserStat(&pv)= FIT_USCAL;
	    ParVVal(&pv)= CalS(cal, FitPPosVal(pp,i)); }
	  else {
	    ParVVal(&pv) = width; /* set value and stat */
	    ParVMode(&pv) = PAR_MFREE;
	    ParVUserStat(&pv)= FIT_USINIT; }
	  ParSet(pp, j, &pv, FALSE); } } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int FSInitVol(fit)
     Fit_t fit;
{				/* volume = amplitude * unity-gauss-volume */
  Par_t pp= FitPar(fit);
  int i;

  for (i= 0; i< FitPeakNum(fit); i++) {
    int j= FitVolIdx(i);
    if (ParSused(pp,j)) {
      if (ParUserStat(pp,j) == FIT_USSTART &&
	  ParVUserStat(FitPWidth(pp,i)) != FIT_USUNDEF) {
	ParV_t pv;

	ParGet(pp, j, &pv);
	ParVVal(&pv) = ParVUserVal(&pv) * FitPWidthVal(pp,i) * SQRT_2_PI;
	ParVMode(&pv) = PAR_MFREE; /* set stat */
	ParVUserStat(&pv) = FIT_USINIT;
	ParSet(pp, j, &pv, FALSE); } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
/* ========================== */
/* initialize tail parameters */
/* ========================== */

static int FSInitTails(fit, cal)
     Fit_t fit;
     Cal_t cal;
{
  Par_t pp= FitPar(fit);
  int i;

  for (i= 0; i< FitPeakNum(fit); i++) {
    ParV_t pv;
    int k;

    for (k=0; k< 2; k++) {
      int j;
      double (*fct)();
      switch (k) {
      case 0: j= FitTlIdx(i); fct= CalTl; break;
      case 1: j= FitTrIdx(i); fct= CalTr; break;
      default: return FALSE; }

      if (ParUserStat(pp,j) == FIT_USUNDEF || /* tail left */
	  ParUserStat(pp,j) == FIT_USINIT ||
	  ParUserStat(pp,j) == FIT_USCALUNDEF) {
	ParGet(pp, j, &pv);
	if (ParVUserStat(&pv) == FIT_USCALUNDEF) { /* calibrated tail */
	  ParVUserStat(&pv)= FIT_USCAL;
	  ParVVal(&pv)= fct(cal, FitPPosVal(pp,i)); }
	else {
	  switch (FitPeakTyp(fit)) {
	  case FIT_PKCA: ParVVal(&pv) = 1.0; break;
	  case FIT_PKAE: ParVVal(&pv) = 0.0; break;
	  default: FSsetErrNo(FSPKTYP); return FALSE; }
	  ParVMode(&pv) = PAR_MFREE; /* set status */
	  ParVUserStat(&pv) = FIT_USINIT; }
	ParSet(pp, j, &pv, FALSE); } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
/* ========================== */
/* initialize step parameters */
/* ========================== */

static int FSInitStep(fit, spc, cal)
     Fit_t fit;
     Spc_t spc;
     Cal_t cal;
{
  Par_t pp= FitPar(fit);
  double amplSum= 0.0;
  int fit_step= FALSE;

  { int i;
    for (i= 0; i< FitPeakNum(fit); i++) {
      int j= FitVolIdx(i);

      if (ParSused(pp,FitShIdx(i))) { fit_step= TRUE; }

      if (ParSused(pp,j)) {
	if (ParUserStat(pp,j) != FIT_USUNDEF) {
	  amplSum += ParUserVal(pp,j); } } } }

  if (!fit_step) { return TRUE; }
  else if (!VecNum(FitRegMrk(fit))) { return TRUE; }
  else if (amplSum == 0.0) { FSsetErrNo(FSPKSUM0); return FALSE; }
  else if (!FitBgFunctionSetup(fit, NULL)) {
    FSsetErrNo(FSNOBG); return FALSE; }
  else {
    int i;

    for (i= 0; i< FitPeakNum(fit); i++) {
      ParV_t pv;
      int j= FitShIdx(i);

      if (ParUserStat(pp,j) == FIT_USUNDEF || /* step height */
	  ParUserStat(pp,j) == FIT_USINIT ||
	  ParUserStat(pp,j) == FIT_USCALUNDEF) {
	ParGet(pp, j, &pv);
	if (ParVUserStat(&pv) == FIT_USCALUNDEF) { /* calibrated sh */
	  ParVUserStat(&pv)= FIT_USCAL;
	  ParVVal(&pv)= CalSh(cal, FitPPosVal(pp,i)); }
	else {
	  PVec_t mp= FitRegMrk(fit);
	  double xl = PVecX1(mp,0);
	  double xg = PVecX2(mp,VecNum(mp)-1);
	  double yl;
	  double yg;
	  SpcReal_t *sb= SpcBufReal(SpcSpc(spc));
	  yl = sb[MAX(0, NINT(xl))];
	  yg = sb[MIN(SpcNum(spc)-1, NINT(xg))];

	  ParVVal(&pv) = ((yg - yl) * 0.25 / amplSum);
	  ParVMode(&pv) = PAR_MFREE;	/* set status */
	  ParVUserStat(&pv) = FIT_USINIT; }
	ParSet(pp, j, &pv, FALSE); }

      j= FitSwIdx(i);
      if (ParUserStat(pp,j) == FIT_USUNDEF || /* step width */
	  ParUserStat(pp,j) == FIT_USINIT ||
	  ParUserStat(pp,j) == FIT_USCALUNDEF) {
	ParGet(pp, j, &pv);
	if (ParVUserStat(&pv) == FIT_USCALUNDEF) { /* calibrated width */
	  ParVUserStat(&pv)= FIT_USCAL;
	  ParVVal(&pv)= CalSw(cal, FitPPosVal(pp,i)); }
	else {
	  ParVVal(&pv) = 1.0;
	  ParVMode(&pv) = PAR_MFREE; /* set status */
	  ParVUserStat(&pv) = FIT_USINIT; }
	ParSet(pp, j, &pv, FALSE); } } }

  return TRUE;
}
/* ---------------------------------------------------------------------- */
