/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsKey.c,v 1.3 1994/09/20 14:47:55 vs Exp vs $ */
#define _vsKey_c

				/* ======================== */
				/* vs library include files */
				/* ======================== */
#include "vsTypes.h"
#include "vsAlloc.h"
#include "vsPath.h"
#include "vsUserFace.h"

#include "vsKey.h"
#include "vsKeyP.h"
				/* ==================== */
				/* system include files */
				/* ==================== */
#include <string.h>
#include <stdio.h>

/* ---------------------------------------------------------------------- */
char *KTError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* KTNOERR  */ "vsKey: no error",
  /* KTPOINT  */ "vsKey: illegal NULL-pointer",
  /* KTOPEN   */ "vsKey: can not open keytable file",
  /* KTREAD   */ "vsKey: can not read keytable",
  /* KTWRITE  */ "vsKey: can not write keytable",
  /* KTALLOC  */ "vsKey: can not allocate buffer",
  /* KTSCAN   */ "vsKey: scan error on keytable file",
  /* KTBUFEND */ "vsKey: buffer overflow",
  /* KTFIND   */ "vsKey: can not find keysequence",
  /* default  */ "vsKey: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
Keytable_t KTCreateKeytable(title, path)
     char *title;
     char *path;
{
  Keytable_t kt;

  if ((kt= (Keytable_t)Malloc(sizeof(KeytableRec)))) {
    KeytableTitle(kt)	= Strdup(title);
    KeytableFile(kt)	= Strdup(path ? path : "");
    KeytableFirst(kt)	= NULL;
    KeytableNext(kt)	= NULL; }
  return kt;
}
/* ---------------------------------------------------------------------- */
Keytable_t KTDestroyKeytable(keytable)
     Keytable_t keytable;
{
  if (keytable) {
    Key_t key= KeytableFirst(keytable);

    while (key) {
      Key_t next= KeyNext(key);

      Free(KeySequence(key));
      Free(key);
      key= next; }
    Free(KeytableTitle(keytable));
    Free(KeytableFile(keytable));
    Free(keytable); }
  return NULL;
}
/* ---------------------------------------------------------------------- */
int KTGetKeyAlias(keytable, sequence, alias)
     Keytable_t keytable;
     char *sequence;
     char **alias;
{
  if (keytable && sequence && alias) {
    Key_t key= KeytableNext(keytable);

    if (!key || strcmp(KeySequence(key), sequence) >= 0) {
      key= KeytableFirst(keytable); }

    while (key) {
      char *k= KeySequence(key);
      char *s= sequence;

      while (*s && *k && *k == *s) { s++; k++; }

      if (*s > *k) { key= KeyNext(key); }
      else if (*s == *k) {
	*alias= KeyAlias(key);
	KeytableNext(keytable)= key;
	return KTKEYFOUND; }
      else if (!*s && *k) { *alias= NULL; return KTINCOMPL; }
      else { break; } } }
  *alias= NULL;
  return KTILLKEY;
}
/* ---------------------------------------------------------------------- */
int KTReadTitle(title, file)
     char *title;
     char *file;
{
  FILE *fp;

  if (!title || !file) { KTsetErrNo(KTPOINT); return FALSE; }
  else if (*title) { return TRUE; }
  else if (!(fp= fopen(file, "r"))) {
    KTsetErrNo(KTOPEN); return FALSE; }
  else {
    if (fscanf(fp, "#keytable %s", title) <= 0) {
      strcpy(title, PthTail(file)); }
    fclose(fp);
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
Keytable_t KTReadKeytable(old, title, file, merge)
     Keytable_t old;
     char *title;
     char *file;
     int merge;
{
  if (!file) { KTsetErrNo(KTPOINT); return NULL; }
  else {
    char *path= PthExpand(file);
    FILE *fp;

    if (NULL == (fp= fopen(path, "r"))) {
      KTsetErrNo(KTOPEN); return NULL; }
    else {
      char line[IOBUFLEN];
      char token[IOBUFLEN];
      char *s;
      Keytable_t kt;

      if (!(kt= (old && merge) ? old : KTCreateKeytable(title, path))) {
	fclose(fp); return NULL; }

      while (fgets(line, IOBUFLEN, fp)) {
	if (*line == '"') {
	  char *t= token;
	  int tl= IOBUFLEN;

	  s= line + 1;
	  while (*s && tl > 1 && UFUnquoteChar(&s, &t, &tl, '"')) { ; }

	  *t= '\0';

	  if (*s != '"') {
	    fclose(fp); KTDestroyKeytable(kt);
	    KTsetErrNo(KTSCAN); return FALSE; }

	  s++;

	  while (*s == ' ' || *s == '\t') { s++; } /* skip whitespace */

	  if (!KTInsertKeySequence(kt, token, s)) {
	    fclose(fp); KTDestroyKeytable(kt); return NULL; } } }

      fclose(fp);
      if (old && !merge) { KTDestroyKeytable(old); }
      return kt; } }
}
/* ---------------------------------------------------------------------- */
int KTInsertKeySequence(keytable, sequence, alias)
     Keytable_t keytable;
     char *sequence;
     char *alias;
{
  Key_t new;

  if (!keytable || !sequence || !alias) {
    KTsetErrNo(KTPOINT); return FALSE; }
  else if (!(new= (Key_t)Malloc(sizeof(KeyRec)))) {
    KTsetErrNo(KTALLOC); return FALSE; }
  else if (!(KeySequence(new)= (char *)Malloc((strlen(sequence)
					       +strlen(alias)+2)))) {
    Free(new);
    KTsetErrNo(KTALLOC); return FALSE; }
  else {
    Key_t key= KeytableNext(keytable); /* start loop at last requested key */
    Key_t *next_of_prev= &KeytableNext(keytable);

    strcpy(KeySequence(new), sequence);
    KeyAlias(new)= KeySequence(new)+strlen(sequence)+1;
    strcpy(KeyAlias(new), alias);
    KeyNext(new)= NULL;

    if (!key || strcmp(KeySequence(key), sequence) >= 0) {
      key= KeytableFirst(keytable);
      next_of_prev= &KeytableFirst(keytable); }

    while (key) {
      char *k= KeySequence(key);
      char *s= sequence;

      while (*s && *k && *k == *s) { s++; k++; }

      if (*s > *k) { next_of_prev= &KeyNext(key); key= KeyNext(key); }
      else if (*s == *k) {
	if (key == KeytableNext(keytable)) {
	  KeytableNext(keytable)= KeyNext(key); }
	KeyNext(new)= KeyNext(key); Free(KeySequence(key)); Free(key);
	break; }
      else { KeyNext(new)= key; break; } }

    *next_of_prev= new;
    KeytableNext(keytable)= new;
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int KTDeleteKeySequence(keytable, sequence)
     Keytable_t keytable;
     char *sequence;
{
  if (!keytable || !sequence) {
    KTsetErrNo(KTPOINT); return FALSE; }
  else {
    Key_t key= KeytableNext(keytable); /* start loop at last requested key */
    Key_t *next_of_prev= &KeytableNext(keytable);

    if (!key || strcmp(KeySequence(key), sequence) >= 0) {
      key= KeytableFirst(keytable);
      next_of_prev= &KeytableFirst(keytable); }

    while (key) {
      char *k= KeySequence(key);
      char *s= sequence;

      while (*s && *k && *k == *s) { s++; k++; }

      if (*s > *k) { next_of_prev= &KeyNext(key); key= KeyNext(key); }
      else if (*s == *k) {
	if (key == KeytableNext(keytable)) {
	  KeytableNext(keytable)= KeyNext(key); }
	*next_of_prev= KeyNext(key);
	Free(KeySequence(key)); Free(key);
	return TRUE; } }
    KTsetErrNo(KTFIND); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int KTWriteKeytable(keytable, file)
     Keytable_t keytable;
     char *file;
{
  if (!keytable || !file) { KTsetErrNo(KTPOINT); return FALSE; }
  else {
    char *path= PthExpand(file);
    FILE *fp;

    if (NULL == (fp= fopen(file, "w"))) { KTsetErrNo(KTOPEN); return FALSE; }
    else {
      if (fprintf(fp, "#keytable %s\n", KeytableTitle(keytable)) < 0) {
	KTsetErrNo(KTWRITE); fclose(fp); return FALSE; }
      else {
	Key_t k= KeytableFirst(keytable);
	char line[IOBUFLEN];

	while (k) {
	  if (!KTSprintKey(line, IOBUFLEN, KeySequence(k), KeyAlias(k)) ||
	      fprintf(fp, "%s\n", line) < 0) {
	    fclose(fp); return FALSE; }
	  k= KeyNext(k); }

	fclose(fp);
	Free(KeytableFile(keytable));
	KeytableFile(keytable)= Strdup(path);
	return TRUE; } } }
}
/* ---------------------------------------------------------------------- */
char *KTSprintKey(buffer, buflen, sequence, alias)
     char *buffer;
     int buflen;
     char *sequence;
     char *alias;
{
  char *b= buffer;
  int bl= buflen;

  if (strlen(sequence) + strlen(alias) + 5 >= buflen) {
    KTsetErrNo(KTBUFEND); return NULL; }
  else {
    *b++ = '"'; bl--;
    if (!UFItemSPrint(sequence, b, bl, '"')) {
       KTsetErrNo(KTBUFEND); return NULL; }
    else {
      bl -= strlen(b);
      b += strlen(b);
      if (bl > 0) { *b++ = '"'; bl--; }

      if (strlen(alias) + 2 >= bl) { KTsetErrNo(KTBUFEND); return NULL; }
      else {
	sprintf(b, "\t%s", alias);
	if (b[strlen(b)-1] == '\n') { b[strlen(b)-1]= '\0'; }
	return buffer; } } }
}
/* ---------------------------------------------------------------------- */
