/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsPar.c,v 1.2 1994/05/26 19:50:23 vs Exp vs $ */

#define _vsPar_c

#include <stdlib.h>

#include "vsTypes.h"

#include "vsAlloc.h"

#include "vsTools.h"

#include "vsPar.h"
#include "vsParP.h"

/* ---------------------------------------------------------------------- */
char *ParError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* PARNOERR:  */ "vsPar: no error",
  /* PARPOINT:  */ "vsPar: illegal NULL-pointer",
  /* PARALLOC:  */ "vsPar: memory allocation failed",
  /* PARILLIDX: */ "vsPar: illegal index",
  /* PARILLMOD: */ "vsPar: illegal parameter mode",
  /* PARILLSTT: */ "vsPar: illegal parameter status",
  /* PARCORCYC: */ "vsPar: cyclic correlation not accepted",
  /* PARCORZER: */ "vsPar: no correlation defined",
  /* PARCORIDX: */ "vsPar: illegal parameter index",
  /* PARCORDEL: */ "vsPar: correlation to deleted parameter",
  /* PARCORSUS: */ "vsPar: correlation to suspended parameter",
  /* PARCOREX:  */ "vsPar: parameter in correlation in use",
  /* PARSRTLEN: */ "vsPar: illegal number of sort elements",
  /* PARMEMBER: */ "vsPar: illegal member of ParV_t specified",
  /* default:   */ "vsPar: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
int ParInit(pp)
     Par_t pp;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    VecInit(ParPar(pp), sizeof(ParV_t));
    VecInit(ParFr(pp), sizeof(int));
    VecInit(ParHld(pp), sizeof(int));
    VecInit(ParCor(pp), sizeof(int));
    VecInit(ParCorFr(pp), sizeof(int));
    return ParReset(pp); }
}
/* ---------------------------------------------------------------------- */
int ParVInit(pvp)
     ParV_t *pvp;
{
  if (pvp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    ParVVal(pvp)	= 0.0;
    ParVErr(pvp)	= 0.0;
    ParVUserVal(pvp)	= 0.0;
    ParVCorr(pvp)[0]	= 0.0;
    ParVCorr(pvp)[1]	= 0.0;
    ParVUserStat(pvp)	= 0;
    ParVUserTyp(pvp)	= 0;
    ParVStat(pvp)	= PAR_SUNDEF;
    ParVMode(pvp)	= PAR_MUNDEF;
    ParVCorrIndex(pvp)	= -1;
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int ParReset(pp)
     Par_t pp;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    VecReset(ParPar(pp));
    VecReset(ParFr(pp));
    VecReset(ParHld(pp));
    VecReset(ParCor(pp));
    VecReset(ParCorFr(pp));
    ParMaxUsed(pp) = 0;
    ParModeCounted(pp)= FALSE;
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int ParFree(pp)
     Par_t pp;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    VecFree(ParPar(pp));
    VecFree(ParFr(pp));
    VecFree(ParHld(pp));
    VecFree(ParCor(pp));
    VecFree(ParCorFr(pp));
    return ParReset(pp); }
}
/* ---------------------------------------------------------------------- */
static int ParRealloc(pp, siz)
     Par_t pp;
     int siz;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    int oldNum= VecNum(ParPar(pp));
    if (VecRealloc(ParPar(pp), siz) &&
	VecRealloc(ParFr(pp), siz) &&
	VecRealloc(ParHld(pp), siz) &&
	VecRealloc(ParCor(pp), siz) &&
	VecRealloc(ParCorFr(pp), siz)) {
      int i;
      for (i= oldNum; i< siz; i++) { ParVInit(ParV(pp,i)); }
      return TRUE; }
    else { ParErrNo = PARALLOC; return FALSE; } }
}
/* ---------------------------------------------------------------------- */
int ParCopy(t, f)
     Par_t t;
     Par_t f;
{
  if (t == NULL || f == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else if (VecCopy(ParPar(t), ParPar(f)) &&
	   VecCopy(ParFr(t), ParFr(f)) &&
	   VecCopy(ParHld(t), ParHld(f)) &&
	   VecCopy(ParCor(t), ParCor(f)) &&
	   VecCopy(ParCorFr(t), ParCorFr(f))) {
    ParMaxUsed(t) = ParMaxUsed(f);
    ParModeCounted(t) = ParModeCounted(f);
    return TRUE; }
  else { ParErrNo= PARALLOC; ParReset(t); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int ParGet(pp, idx, pv)
     Par_t pp;
     int idx;
     ParV_t *pv;
{
  if (pp == NULL || pv == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    if (idx < 0 || idx >= ParNum(pp)) { ParErrNo= PARILLIDX; return FALSE; }
    else {
      memcpy((void *)pv, (void *)ParV(pp,idx), sizeof(ParV_t));
      return TRUE; } }
}
/* ---------------------------------------------------------------------- */
int ParSet(pp, idx, pvp, check_corr)
     Par_t pp;
     int idx;
     ParV_t *pvp;
     int check_corr;
{
  if (pp == NULL || pvp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    if (idx < 0) { idx= ParNum(pp); }

    if (ILLEGAL_STAT(ParVStat(pvp))) {
      ParErrNo = PARILLSTT; return FALSE; }
    else if (ILLEGAL_MODE(ParVMode(pvp))) {
      ParErrNo = PARILLMOD; return FALSE; }
    else if (check_corr && ParVMcorr(pvp) && !ParCorrelate(pp, idx, pvp)) {
      return FALSE; }
    else {
      if (idx >= ParNum(pp)) {
	if (!ParRealloc(pp, idx+1)) { return FALSE; }
	else { ParModeCounted(pp)= FALSE; ParNum(pp)= idx+1; } }
      else if (ParStat(pp,idx) != ParVStat(pvp) ||
	       ParMode(pp,idx) != ParVMode(pvp)) {
	ParModeCounted(pp) = FALSE; }
      memcpy((void *)ParV(pp,idx), (void *)pvp, sizeof(ParV_t));
      if (check_corr) { return ParCorrelateAll(pp); }
      else { return ParUpdateCorr(pp); } } }
}
/* ---------------------------------------------------------------------- */
int ParCorrelateAll(pp)
     Par_t pp;
{
  int i;
  for (i= 0; i< ParNum(pp); i++) {
    if (ParSused(pp,i) && ParMcorr(pp,i)) {
      if (!ParCorrelate(pp, i, ParV(pp,i))) { return FALSE; } } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int ParCompress(pp)
     Par_t pp;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else {
    int i;

    for (i= 0; i< ParNum(pp); i++) {
      if (!ParSdeld(pp,i) && ParMcorr(pp,i)) {
	int j= ParCorrIndex(pp,i); /* get parent of correlation */
	if (ParSdeld(pp,j)) {	/* don't delete correlation parents */
	  ParErrNo = PARCOREX; return FALSE; } } }

    for (i= ParNum(pp)-1; i >= 0; i--) {
      if (ParSdeld(pp,i)) {
	VecDelete(ParPar(pp), i, 1);
	ParModeCounted(pp) = FALSE; } }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
static int ParCorrelate(pp, idx, pv)
     Par_t pp;
     int idx;
     ParV_t *pv;
{
  double v0= ParVCorr(pv)[0];
  double v1= ParVCorr(pv)[1];
  int i= ParVCorrIndex(pv);
  int count= -1;

  while (ParSused(pp,i)) {
    if (ParMcorr(pp,i)) {
      v0 += ParCorr(pp,i)[0] * v1;
      v1 *= ParCorr(pp,i)[1];
      i= ParCorrIndex(pp,i); }
    else { break; }
    if (i == idx || count > ParNum(pp)) {
      ParErrNo = PARCORCYC; return FALSE; }
    else { count++; } }
  if (ParSdeld(pp,idx)) { ParErrNo = PARCORDEL; return FALSE; }
  else if (ParSsusp(pp,idx)) { ParErrNo = PARCORSUS; return FALSE; }
  else if (ParVCorr(pv)[1] == 0.0) { ParErrNo = PARCORZER; return FALSE; }
  else {
    ParVCorr(pv)[0]= v0;
    ParVCorr(pv)[1]= v1;
    ParVCorrIndex(pv)= i;
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
static ParV_t *ParSort_pp;

static int ParSortVal(i1, i2)
     int *i1;
     int *i2;
{
  if (ParVVal(ParSort_pp + *i1) > ParVVal(ParSort_pp + *i2)) { return 1; }
  else if (ParVVal(ParSort_pp + *i1) == ParVVal(ParSort_pp + *i2)) {
    return 0; }
  else { return -1; }
}
/* ---------------------------------------------------------------------- */
static int ParSortErr(i1, i2)
     int *i1;
     int *i2;
{
  if (ParVErr(ParSort_pp + *i1) > ParVErr(ParSort_pp + *i2)) { return 1; }
  else if (ParVErr(ParSort_pp + *i1) == ParVErr(ParSort_pp + *i2)) {
    return 0; }
  else { return -1; }
}
/* ---------------------------------------------------------------------- */
static int ParSortStart(i1, i2)
     int *i1;
     int *i2;
{
  if (ParVUserVal(ParSort_pp + *i1) > ParVUserVal(ParSort_pp + *i2)) {
    return 1; }
  else if (ParVUserVal(ParSort_pp + *i1) == ParVUserVal(ParSort_pp + *i2)) {
    return 0; }
  else { return -1; }
}
/* ---------------------------------------------------------------------- */
int ParSort(pp, idx, numidx, member)
     Par_t pp;
     int *idx;
     int numidx;
     int member;
{
  if (pp == NULL || idx == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else if (numidx < 0) { ParErrNo = PARSRTLEN; return FALSE; }
  else if (numidx <= 1) { return TRUE; }
  else {
    ParSort_pp = ParP(pp);
    switch (member) {
    case PAR_VVAL:
      qsort((void *)idx, (size_t)numidx, sizeof(int), (pfi_t)ParSortVal);
      break;
    case PAR_VERR:
      qsort((void *)idx, (size_t)numidx, sizeof(int), (pfi_t)ParSortErr);
      break;
    case PAR_VSTART:
      qsort((void *)idx, (size_t)numidx, sizeof(int), (pfi_t)ParSortStart);
      break;
    default: ParErrNo = PARMEMBER; return FALSE; }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int ParIndex(pp)
     Par_t pp;
{
  if (pp == NULL) { ParErrNo = PARPOINT; return FALSE; }
  else if (!ParModeCounted(pp)) {
    int i;
    if (!ParRealloc(pp,ParNum(pp))) { return FALSE; }
    ParNumFree(pp)= 0;
    ParNumHold(pp)= 0;
    ParNumCorr(pp)= 0;
    for (i= 0; i< ParNum(pp); i++) {
      if (ParSused(pp,i)) {
	ParMaxUsed(pp) = i;
	switch (ParMode(pp,i)) {
	case PAR_MFREE:
	  ParIfree(pp, ParNumFree(pp)) = i; ParNumFree(pp)++; break;
	case PAR_MCORR:		/* correlated to a free parameter */
	  { int j= ParCorrIndex(pp,i);
	    if (ParSdeld(pp,j)) { ParErrNo = PARCORDEL; return FALSE; }
	    else if (ParSsusp(pp,j)) { ParErrNo = PARCORSUS; return FALSE; }
	    else if (ParMfree(pp, j)) {
	      ParIcorr(pp, ParNumCorr(pp)) = i; ParNumCorr(pp)++; break; } }
	case PAR_MHOLD:		/* held or correlated to a held param */
	  ParIhold(pp, ParNumHold(pp)) = i; ParNumHold(pp)++; break;
	default: ParErrNo = PARILLMOD; return FALSE; } } }
    ParNumCoFr(pp)= ParNumCorr(pp);
    for (i= 0; i< ParNumCorr(pp); i++) {
      int j;
      int ci= ParCorrIndex(pp,ParIcorr(pp,i));
      for (j= 0; j< ParNumFree(pp); j++) {
	if (ci == ParIfree(pp,j)) { ParIcofr(pp,i) = j; break; } } }
    ParModeCounted(pp) = TRUE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int ParUpdate(pp)
     Par_t pp;
{
  if (!ParIndex(pp)) { return FALSE; }
  else { return ParUpdateCorr(pp); }
}
/* ---------------------------------------------------------------------- */
static int ParUpdateCorr(pp)
     Par_t pp;
{
  int i;
  for (i= 0; i< ParNum(pp); i++) {
    if (ParSused(pp,i) && ParMcorr(pp,i)) { ParCorrel(pp,i); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
