/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsPath.c,v 1.3 1994/12/02 21:22:49 vs Exp vs $ */
#define _vsPath_c

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#include <unistd.h> /* getcwd() */

#include "vsTypes.h"
#include "vsAlloc.h"
#include "vsPath.h"
#include "vsPathP.h"
/* ---------------------------------------------------------------------- */
static char pthPath[MAXPATHLEN]; /* return string for all functions */
static char *pthHome= NULL;	/* canonical home directory */
/* ---------------------------------------------------------------------- */
char *PthError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* PTHNOERR: */ "vsPath: no error",
  /* PTHPOINT: */ "vsPath: illegal NULL-pointer",
  /* PTHCREAT: */ "vsPath: could not create path",
  /* PTHILLSP: */ "vsPath: illegal path specification",
  /* PTHILLCP: */ "vsPath: illegal path component specification",
  /* PTHFOUND: */ "vsPath: could not find search string",
  /* PTHLEN:   */ "vsPath: maximal path length exceeded",
  /* default:  */ "vsPath: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
static int PthArgCheck(path)
     char *path;
{
  if (path == NULL) {
    PthErrNo= PTHPOINT;
    return FALSE;
  }
  else if (strlen(path) >= MAXPATHLEN) {
    PthErrNo= PTHLEN;
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
char *PthCreate(path)
     char *path;
{
  if (PthArgCheck(path)) {
    strcpy(pthPath, PthExpand(path));
    if (*pthPath) {
      DIR *dp;
      if (NULL == (dp= opendir(pthPath))) { /* Does directory exist? */
	if (mkdir(pthPath, /* No, create it */
		   (mode_t)(S_IRUSR | S_IWUSR | S_IXUSR | /* mode 755 */
		  S_IRGRP | S_IXGRP |
		  S_IROTH | S_IXOTH)) < 0) {
	  PthErrNo= PTHCREAT;
	  *pthPath= '\0';
	}
      }
      else {
	closedir(dp);
      }
    }
  }
  else {
    *pthPath= '\0';
  }

  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthCreateRecursive(path)
     char *path;
{
  if (PthArgCheck(path)) {
    char pthTmp[MAXPATHLEN];
    strcpy(pthTmp, path);
    if (!*PthCreate(pthTmp)) { /* Create directory path */
      if (PthCreate(PthHead(pthTmp))) {
	return PthCreate(pthTmp);
      }
    }
  }
  else {
    *pthPath= '\0';
  }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthCompress(path)
     char *path;
{
  if (PthArgCheck(path)) {
    return PthPCompress(PthExpand(path)); }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
static char *PthPCompress(path)
     char *path;
{
  int pl= 0;
  if (PthArgCheck(path)) {
    char pthTmp[MAXPATHLEN];
    char *p= pthTmp;
    char *pp= pthPath;

    strcpy(pthTmp, path);
    while (*p) {
      if (!strncmp(p, "//", 2)) { p++; }
      else if (!strncmp(p, "/./", 3)) { p += 2; }
      else if (!strncmp(p,"/..", 3) && (p[3] == '\0' || p[3] == '/')) {
	while (pl > 0) {
	  if (pp[pl-1] == '/') { /* path part to go up */
	    if (p[3] == '/' || pl > 1) { pl--; } /* don't remove root '/' */
	    break; }
	  pl--; }
	p += 3; }
      else { pp[pl++] = *p++; } } } /* copy */
  pthPath[pl]= '\0';
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthShort(path)
     char *path;
{
  if (PthArgCheck(path)) {
    char pthTmp[MAXPATHLEN];

#if defined(HAS_GETWD)
    char ref[MAXPATHLEN];
#else
    /* getcwd requires 2 more bytes! */
    char ref[MAXPATHLEN+2];
#endif

    int pl= 0;
    int rl;
    char *hp;
#if 0
    strcpy(pthTmp, PthCompress(path));
#else				/* 0 */
    strcpy(pthTmp, PthPCompress(path));	/* should not expand unneccessarily */
#endif				/* 0 */

#ifndef AVOID_GETWD
#if defined(HAS_GETWD)
    if (getwd(ref) != NULL) {
#else
    if (getcwd(ref,MAXPATHLEN+2) != NULL) {
#endif
#else
      {
	strcpy (ref,rememberCwd);
/*	printf ("avoided getwd(cashed %s)\n", rememberCwd); */
#endif

     strcpy(ref, PthCompress(ref));
      rl= strlen(ref);
      if (rl > 0 && ref[rl-1] == '/') { rl--; }
      if (!strncmp(ref, pthTmp, (size_t)rl) &&
	  (pthTmp[rl] == '/' || pthTmp[rl] == '\0') &&
	  strcmp("/", ref)) {
	strcpy(pthPath, "./"); pl= 2; } }

    if (!pl && (hp= PthHome())) {
      strcpy(ref, PthCompress(hp));
      rl= strlen(ref);
      if (rl > 0 && ref[rl-1] == '/') { rl--; }
      if (!strncmp(ref, pthTmp, (size_t)rl) &&
	  (pthTmp[rl] == '/' || pthTmp[rl] == '\0') &&
	  strcmp("/", ref)) {
	strcpy(pthPath, "~/"); pl= 2; }
      else { rl= 0; }}
    
    strcpy(pthPath + pl, pthTmp + rl); /* concatenate */
    strcpy(pthPath, PthPCompress(pthPath)); }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthExpand(path)
     char *path;
{				/* since PthCompress calls this routine */
				/* it must not call PthCompress */
  if (PthArgCheck(path)) {
    extern char *getenv();
    char pthTmp[MAXPATHLEN];
    /*    char envVar[MAXPATHLEN]; BEFORE DUKE*/
    char *envVar;		          /* AFTER DUKE */
    *pthTmp= '\0';

    if (!strncmp(path, "~/", 2) || !strcmp(path, "~")) {
      char *hp;
      if ((hp= PthHome())) { strcpy(pthTmp, hp); }
      strcat(pthTmp, "/");	/* root '/' if no home found */
      strcat(pthTmp, path+1); }


    else if (*path == '~' && path[1] != '/') { /* home of someone else */
      int i;
      struct passwd *pw;
      for (i=1; path[i] != '\0' && path[i] != '/'; i++) {
	pthTmp[i-1] = path[i]; }
      pthTmp[i-1] = '\0';
      pw= getpwnam(pthTmp);
      if (pw != NULL) { strcpy(pthTmp, pw->pw_dir); strcat(pthTmp, path+i); }
      else { strcpy(pthTmp, path); }}

    else {
      if (*path != '/') {	/* no absolute path */

#ifndef AVOID_GETWD
#if defined(HAS_GETWD)
	char ref[MAXPATHLEN];
	if (getwd(ref) != NULL) {
#else
	  /* + 2:  */
	char ref[MAXPATHLEN + 2];
	if (getcwd(ref, MAXPATHLEN+2) != NULL) {
#endif
#else
	  {
	    char ref[MAXPATHLEN];

	    strcpy (ref,rememberCwd);
/*	    printf ("avoided getwd(cashed %s)\n", rememberCwd); */
#endif

	  strcpy(pthTmp, ref); strcat(pthTmp, "/"); }}
	strcat(pthTmp, path); }
      
      { int i= 0;			/* expand environment variables */
      int tl= strlen(pthTmp);
      int pl= 0;
      while (i < tl) { /* at least expecting 3 further char $(c) */
	pthPath[pl++]= pthTmp[i];
	if (pthTmp[i] == '$' && pthTmp[i+1] == '(') {
	  int j;
	  for (j= 0; j + i + 2 < tl; j++) {
	    if (pthTmp[j + i + 2] == ')') {
	      char *ep;
	      /*	      envVar[j] = '\0'; BEFORE DUKE */
	      envVar = &pthTmp[i+2]; /* AFTER DUKE */
	      envVar[j] = '\0';	/* AFTER DUKE */
	      if (NULL != (ep= getenv(envVar))) {
		strcpy(pthPath + pl-1, ep); /* pl-1: supersede copied '$' */
		pl= strlen(pthPath);
		i += j + 3;
		break; }
	      else{		/* env variable not found: go on */
		envVar[j] =')';
		i++; }
	    }}}
	else { i++; }}
      pthPath[pl] = '\0'; }
    strcpy(pthPath, PthPCompress(pthPath)); }
  else { *pthPath = '\0'; }	      
  
  return pthPath;
}
/* ---------------------------------------------------------------------- */
#define P_COPY		0
#define P_SUBST		1
#define strcatchr(s,c)	{ int sl= strlen(s); s[sl++]= c; s[sl]= '\0'; }

char *PthSubstitute(path, expression)
     char *path;
     char *expression;
{
  if (PthArgCheck(path)) {
    char pthTmp[MAXPATHLEN];	/* temporary path */
    char p[MAXPATHLEN];		/* further temporary path */
    char s[MAXPATHLEN];		/* search */
    char r[MAXPATHLEN];		/* replace */
    char *e= expression;
    int i= 0;


    strcpy(pthTmp, PthCompress(path));
    *pthPath= '\0';

    while (e[i]) {
      int action= P_SUBST;
      int sl= 0;
      int rl= 0;

      *s= '\0';
      *r= '\0';
      strcpy(p, pthTmp);
      if (e[i] == '\\') { i++; }
      while (e[i] && e[i] != '\\') {
	if (e[i] == '!') { action = P_COPY; }
	s[sl++]= e[i++]; }
      s[sl]= '\0';		/* search or copy string */

      if (e[i] == '\\') { i++; }

      switch (action) {
      case P_COPY:
	if (!PthPCopy(pthTmp, p, s)) { *pthPath= '\0'; return pthPath; }
	break;
      case P_SUBST:
	while (e[i] && e[i] != '\\') { r[rl++]= e[i++]; } /* replace string*/
	r[rl]= '\0';
	if (!PthPSubst(pthTmp, p, s, r)) { *pthPath= '\0'; return pthPath; }
	break; } }

    strcpy(pthPath, PthCompress(pthTmp)); }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
static int PthPCopy(t, f, e)
     char *t;
     char *f;
     char *e;
{
  *t= '\0';
  while (*e) {
    if (*e == '!') {
      e++;
      switch (*e) {
      case 'h': strcat(t, PthHead(f)); break; /* head */
      case 't': strcat(t, PthTail(f)); break; /* tail */
      case 'r': strcat(t, PthRoot(f)); break; /* root */
      case 'e': strcat(t, PthExtension(f)); break; /* extension */
      case 'f': strcat(t, f); break; /* file */
      default:
	PthErrNo= PTHILLCP; *t= '\0'; return FALSE; } }
    else { strcatchr(t, *e); }
    e++; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int PthPSubst(t, f, s, r)
     char *t;
     char *f;
     char *s;
     char *r;
{
  int i;
  for (i= strlen(f)-1; i>= 0; i--) {
    if (!strcmp(f, s)) {
      strncpy(t, f, (size_t)i);
      strcpy(t + i, r);
      strcat(t, f + i + strlen(s));
      return TRUE; } }
  PthErrNo= PTHFOUND; *t= '\0';
  return FALSE;
}
/* ---------------------------------------------------------------------- */
/*
 * PthHome:
 *
 */

char *PthHome()
{
  char *hp;
  extern char *getenv();

  if (!pthHome && (hp= getenv("HOME"))) {
#if defined(HAS_GETWD)
    char cwd[MAXPATHLEN];
    if (getwd(cwd)) {
#else
    char cwd[MAXPATHLEN + 2];
    if (getcwd(cwd,MAXPATHLEN+2)) {
#endif
      extern int chdir();

      if (!chdir(hp)) {
#if defined(HAS_GETWD)
	char home[MAXPATHLEN];
	if (getwd(home)) { pthHome= Strdup(home); } }
#else
	char home[MAXPATHLEN + 2];
	if (getcwd(home,MAXPATHLEN+2)) { pthHome= Strdup(home); } }
#endif
      chdir(cwd); } }

  return pthHome;
}
/* ---------------------------------------------------------------------- */
char *PthHead(path)
     char *path;
{
  if (PthArgCheck(path)) {
    int i;
    for (i= strlen(path)-1; i >= 0; i--) {
      if (path[i] == '/') {
	strncpy(pthPath, path, (size_t)i);
	pthPath[i] = '\0';
	break;
      }
    }
    strcpy(pthPath, PthCompress(pthPath));
  }
  else {
    *pthPath= '\0';
  }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthTail(path)
     char *path;
{
  if (PthArgCheck(path)) {
    int i;
    int pl1= strlen(path)-1;
    for (i= pl1; i >= 0; i--) {
      if (path[i] == '/') {
	strncpy(pthPath, path + i + 1, (size_t)(pl1 - i));
	pthPath[pl1 - i] = '\0'; break; } }
    if (i == -1) { strcpy(pthPath, path); } }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthRoot(path)
     char *path;
{
  if (PthArgCheck(path)) {
    int i;
    for (i= strlen(path)-1; i >= 0; i--) {
      if (path[i] == '.') {
	strncpy(pthPath, path, (size_t)i);
	pthPath[i] = '\0'; break; }
      else if (path[i] == '/') { strcpy(pthPath, path); break; } }
    strcpy(pthPath, PthCompress(pthPath)); }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthExtension(path)
     char *path;
{
  if (PthArgCheck(path)) {
    int i;
    int pl1= strlen(path)-1;
    for (i= pl1; i >= 0; i--) {
      if (path[i] == '.') {
	strncpy(pthPath, path + i + 1, (size_t)(pl1 - i));
	pthPath[pl1 - i] = '\0'; break; }
      else if (path[i] == '/') { *pthPath = '\0'; break; } } }
  else { *pthPath= '\0'; }
  return pthPath;
}
/* ---------------------------------------------------------------------- */
char *PthFind(path, searchPath)
     char *path;
     char *searchPath;
{
  if (PthArgCheck(path) && PthArgCheck(searchPath)) {
    char pthTmp[MAXPATHLEN];
    char *s= searchPath;
    struct stat buf;

    if (strchr(path, '/')) {
      strcpy(pthPath, PthCompress(path));
      if (stat(pthPath, &buf) >= 0) { return pthPath; } }
    else {
      do {
	int tl= 0;
	while (*s && *s != ':') { pthTmp[tl++] = *s++; }
	if (*pthTmp) { pthTmp[tl++] = '/'; }
	strcpy(pthTmp + tl, path);
	strcpy(pthPath, PthCompress(pthTmp));
	if (stat(pthPath, &buf) >= 0) { return pthPath; }
	if (*s) { s++; }
      } while (*s); } }
  *pthPath= '\0';
  return pthPath;
}
/* ---------------------------------------------------------------------- */
/* must be called after every chdir !!! */
void PthChangedCwd (char* path)
{

#ifdef AVOID_GETWD
  printf ("chdir %s -> ", rememberCwd);
  strcpy (rememberCwd, path);
  printf ("%s\n",rememberCwd);
#endif
}
