/* $Header: /usr/files/home/Gamma/vs/tv/edit/tv/RCS/vsRecal.c,v 1.1 1993/09/17 14:55:08 vs Exp vs $ */
#define _vsRecal_c

#include "vsTypes.h"

#include "vsSpectra.h"
#include "vsTools.h"
#include "vsAlloc.h"
#include "vsRecal.h"
#include "vsRecalP.h"
/* ====================================================================== */

/* the value of a channel c will become split */
/* to the intervall [Rec(c-0.5), Rec(c+0.5)] */
/* where the relevant channels range from */
/* NINT(Rec(c-0.5)) to NINT(Rec(c+0.5)) and */
/* the overlap with the limiting channels is */
/* obeyed */

#define SPLIT(tLeft,tRight,num,t,f,stat) \
    { double tDelta = tRight - tLeft;\
      if (tDelta <= 0.0) {\
	RclsetErrNo(RCLGRAD);\
	{ int j; for (j = (tLeft < 0) ? 0 : NINT(tLeft); j < num; j++) {\
	  t[j] = 0.0; } }\
	stat= FALSE; }\
      else {\
	int l= NINT(tLeft);\
	int r= NINT(tRight);\
	if (r == l) {		/* target is one channel */\
	  if (0 <= l && l < num) { t[l] += f; } }\
	else {			/* split content of f */\
	  double fFraction= f / tDelta;\
	  if (0 <= l && l < num) { /* start */\
	    t[l] += fFraction * (l + 0.5 - tLeft); }\
	  l++;\
	  if (l < 0) { l= 0; }\
	  if (r > num) { r= num; }\
	  while (l < r) {	/* loop */\
	    t[l] +=  fFraction; l++; }\
	  if (0 <= r && l < num) { /* finish */\
	    t[l] += fFraction * (tRight - l + 0.5); } } } }
/* ------------------------------------------------------------------------- */
#define DELTA_OFFSET		1.0 /* regime to vary fractional offset */
#define INC_OFFSET		0.1 /* increment of fractional offset */
/* ------------------------------------------------------------------------- */
char *RclError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* RCLNOERR  */ "vsRecal: no error",
  /* RCLPOINT  */ "vsRecal: illegal NULL-pointer",
  /* RCLGRAD   */ "vsRecal: zero gradient - spectrum zeroed",
  /* RCLNOMARK */ "vsRecal: no (valid) recalibration marker specified",
  /* RCLMAXRES */ "vsRecal: maximum resolution exceeded",
  /* RCLALLOC  */ "vsRecal: can not extend recalibration data",
  /* RCLEQPAIR */ "vsRecal: equal recalibration marker",
  /* RCLTYP    */ "vsRecal: illegal recalibration typ",
  /* RCLNORES  */ "vsRecal: undefined resolution",
  /* RCLNONORM */ "vsRecal: undefined normalization factor",
  /* default   */ "vsRecal: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ------------------------------------------------------------------------- */
int RclFree(rp)
     Rcl_t rp;
{
  if (!rp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    VecFree(RclPair(rp));
    VecFree(RclPoly(rp));
    RclTyp(rp)= RCL_TUNDEF;
    return TRUE; }
}
/* ------------------------------------------------------------------------- */
int RclInit(rp)
     Rcl_t rp;
{
  if (!rp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    VecInit(RclPair(rp), sizeof(RclPair_t));
    VecInit(RclPoly(rp), sizeof(double));
    return RclReset(rp); }
}
/* ------------------------------------------------------------------------- */
int RclPairCreate(ref, rcl, refNormFac, rclNormFac, num, mp, rp, typ)
     SpcReal_t *ref;
     SpcReal_t *rcl;
     double refNormFac;
     double rclNormFac;
     int num;
     PVec_t mp;
     RclPairVec_t rp;
     int typ;
{
  if (!ref || !rcl || !mp || !rp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else if (!VecNum(mp)) { RclsetErrNo(RCLNOMARK); return FALSE; }
  else if (num <= 0) { RclsetErrNo(RCLNORES); return FALSE; }
  else if (refNormFac == 0.0 || rclNormFac == 0.0) {
    RclsetErrNo(RCLNONORM); return FALSE; }
  else if (num > SPC_MR) { RclsetErrNo(RCLMAXRES); return FALSE; }
  else if (!VecAlloc(rp, VecNum(mp))) { RclsetErrNo(RCLALLOC); return FALSE; }
  else {
    /*
       SpcReal_t StackAlloc(SpcReal_t, rclTmp, num, SPC_MR);
       SpcReal_t StackAlloc(SpcReal_t, rclNrm, num, SPC_MR);
       SpcReal_t StackAlloc(SpcReal_t, refNrm, num, SPC_MR);
       DUKE */

    SpcReal_t *rclTmp = (SpcReal_t*)malloc (num * sizeof (SpcReal_t));
    SpcReal_t *rclNrm = (SpcReal_t*)malloc (num * sizeof (SpcReal_t));
    SpcReal_t *refNrm = (SpcReal_t*)malloc (num * sizeof (SpcReal_t));

    int m;
    int r;
  
    if (!rclTmp || !rclNrm || !refNrm) { RclsetErrNo(RCLALLOC); return FALSE; }

    VecNum(rp)= 0;

    { int i;			/* normalize spectra for recalibration */
      for (i= 0; i< num; i++) {
	rclNrm[i]= rclNormFac * rcl[i];
	refNrm[i]= refNormFac * ref[i]; } }
  
    for (m= 0, r= 0; m< VecNum(mp); m++) {
      int left= MIN(PVecX1(mp,m), PVecX2(mp,m));
      int right= MAX(PVecX1(mp,m), PVecX2(mp,m))+1;
      double bestChnChi;
      double bestChnOffset;
      double bestChi;
      double bestOffset;
      double bestRevChi;
      double bestRevOffset;
      double bestCoM;
      double dBestCoM;
      double chiCoM;
    
      if (left < 0 || num <= right) {
	continue; } /* recalibrate only in windows within spectrum*/

    
      RclPairCreateShiftChn(refNrm, rclNrm, num, left, right,
			    &bestChnOffset, &bestChnChi);
    
      bestOffset= bestChnOffset;
      bestChi= bestChnChi;
      bestRevOffset= -bestChnOffset;
      bestRevChi= bestChnChi;
      bestCoM= bestChnOffset;

      VecNum(rp)++;
      RclPairVecX1(rp,r)= (left + right - 1) / 2;
      RclPairVecXChn(rp,r)= RclPairVecX1(rp,r) + bestChnOffset; 
      RclPairVecChiChn(rp,r)= bestChnChi;

      if (typ != RCL_TSQUARE) {
	RclPairCreateCoM(refNrm, rclNrm, rclTmp, num, left, right,
			 &bestCoM, &dBestCoM, &chiCoM);
	RclPairVecDCoM(rp,r)= dBestCoM;
	RclPairVecXCoM(rp,r)= RclPairVecX1(rp,r) + bestCoM;
	RclPairVecChiCoM(rp,r)= chiCoM; }

      if (typ != RCL_TCOM) {
	RclPairCreateShiftFraction(refNrm, rclNrm, rclTmp, num, left, right,
				   &bestOffset, &bestChi);
	RclPairVecXRcl(rp,r)= RclPairVecX1(rp,r) + bestOffset;
	RclPairVecDRcl(rp,r)= RclPairVecXChn(rp,r) - RclPairVecXRcl(rp,r);
	RclPairVecChiRcl(rp,r)= bestChi; 
	RclPairCreateShiftFraction(rclNrm, refNrm, rclTmp, num,
				   left + (int)bestRevOffset,
				   right + (int)bestRevOffset,
				   &bestRevOffset, &bestRevChi);
	RclPairVecXRev(rp,r)= RclPairVecX1(rp,r) - bestRevOffset;
	RclPairVecDRev(rp,r)= RclPairVecXChn(rp,r) - RclPairVecXRev(rp,r);
	RclPairVecChiRev(rp,r)= bestRevChi; }

      if (typ == RCL_TCOM) {
	RclPairVecX2(rp,r)= RclPairVecX1(rp,r) + bestCoM; }
      else if (bestChi < bestRevChi) {
	RclPairVecX2(rp,r)= RclPairVecX1(rp,r) + bestOffset; }
      else {
	RclPairVecX2(rp,r)= RclPairVecX1(rp,r) - bestRevOffset; }
      r++; } 
    Free (rclTmp);
    Free (rclNrm);
    Free (refNrm);
 }

  if (VecNum(rp) > 0) { 
    return TRUE; }
  else { 
    RclsetErrNo(RCLNOMARK);
    return FALSE; }
}
/* ------------------------------------------------------------------------- */
static int RclPairCreateShiftChn(ref, rcl, num, left, right,
				 bestOffset, bestChi)
     SpcReal_t *ref;
     SpcReal_t *rcl;
     int num;
     int left;
     int right;
     double *bestOffset;
     double *bestChi;

{
  int stat= FALSE;
  int i;
  int width = (right - left - 1) / 2; /* width of shift gate */
  int startOffset
    = -MIN(width, left - DELTA_OFFSET - (int)(DELTA_OFFSET + 0.5));
				/* shift only within spectrum (>=D'O)*/
  int endOffset
    = MIN(width, num - right - (int)(DELTA_OFFSET + 0.5));
				/* (<num-D'O) next step +/-D'O */

  for (i = startOffset; i < endOffset; i++) {
    int j;
    double chi = 0.0;
    for (j = left; j < right; j++) {
      double delta= ref[j] - rcl[j+i]; /* offset = -i */
      chi += delta * delta; }
    if (i == startOffset || chi < *bestChi ||
	(chi == *bestChi && ABS(i) < ABS(*bestOffset))) {
      *bestOffset = i;		/* DUKE -i -> i */
      *bestChi = chi;
      stat= TRUE; } }
  return stat;
}
/* ------------------------------------------------------------------------- */
static int RclPairCreateShiftFraction(ref, rcl, rclTmp, num, left, right,
				      bestOffset, bestChi)
     SpcReal_t *ref;
     SpcReal_t *rcl;
     SpcReal_t *rclTmp;
     int num;
     int left;
     int right;
     double *bestOffset;
     double *bestChi;
{
  int stat= FALSE;
  double pol[2];
  double b= *bestOffset;
  double x;
  int localStart= MIN(left, left - b - (int)(DELTA_OFFSET + 0.5));
				/* >= 0 */
  int localNum= right - left + ABS(b) + 1 + (int)(DELTA_OFFSET + 0.5);
				/* locStart+locNum<=SpcNum */
  pol[1] = 1.0;
      
  for (x = b - DELTA_OFFSET; x < b + DELTA_OFFSET; x += INC_OFFSET) {
    pol[0] = x;
    RclPolynomRecal(rclTmp + localStart, rcl + localStart, localNum, pol, 2);
    { int j;
      double chi = 0.0;
      for (j = left; j < right; j++) {
	double delta = ref[j] - rclTmp[j];
	chi += delta*delta; }
      if (chi < *bestChi || (chi == *bestChi && ABS(x) < ABS(*bestOffset))) {
	*bestOffset = x;
	*bestChi= chi;
	stat= TRUE; } } }
  return stat;
}
/* ------------------------------------------------------------------------- */
static int RclPairCreateCoM(ref, rcl, rclTmp, num, left, right,
			    offset, dOffset, chi)
     SpcReal_t *ref;
     SpcReal_t *rcl;
     SpcReal_t *rclTmp;
     int num;
     int left;
     int right;
     double *offset;
     double *dOffset;
     double *chi;
{
  int i;
  double x= 0.0;
  double ref0= 0.0;
  double ref1= 0.0;
  double rcl0= 0.0;
  double rcl1= 0.0;

  rcl= rcl - NINT(*offset);
    
  for (i = left; i < right; i++, x += 1.0) {
    ref0 += ref[i]; ref1 += x * ref[i];
    rcl0 += rcl[i]; rcl1 += x * rcl[i]; }
  if (ref0 == 0.0 && rcl0 == 0.0) { *dOffset= num; return FALSE; }
  ref1 /= ref0; rcl1 /= rcl0;
  *dOffset= ref1 - rcl1;
  *offset += *dOffset;

  { double pol[2];
    int localStart= MIN(left, left - *offset - (int)(DELTA_OFFSET + 0.5));
				/* >= 0 */
    int localNum= right - left + ABS(*offset) + 1 + (int)(DELTA_OFFSET + 0.5);
				/* locStart+locNum<=SpcNum */
    pol[1] = 1.0;
    pol[0] = *offset;
    RclPolynomRecal(rclTmp + localStart, rcl + localStart, localNum, pol, 2); }

  { int j;
    double c = 0.0;
    for (j = left; j < right; j++) {
      double delta = ref[j] - rclTmp[j];
      c += delta*delta; }
    *chi= c; }
  return TRUE;
}
/* ------------------------------------------------------------------------- */
int RclPairRecal(t, f, num, pp)
     SpcReal_t *t;
     SpcReal_t *f;
     int num;
     RclPairVec_t pp;
{
  if (!t || !f || !pp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    { int i; for (i = 0; i < num; i++) { t[i]= 0.0; } }

    { int p= 0;
      double m;

      if (VecNum(pp) == 1) {	/* simple translation */
	m= 1.0;
	p= 1; }
      else if (RclPairVecX1(pp,0) == RclPairVecX1(pp,1)) {
	RclsetErrNo(RCLEQPAIR); return FALSE; }
      else {			/* linear extrapolation before first point */
	m= ((RclPairVecX2(pp,1) - RclPairVecX2(pp,0)) /
	    (RclPairVecX1(pp,1) - RclPairVecX1(pp,0)));
	p= 2; }
      
      { int i;
	int pNum= VecNum(pp)-1;
	double fOffset= RclPairVecX1(pp,p-1);
	double tOffset= RclPairVecX2(pp,p-1);
	double tRight = tOffset + m * ( -0.5 - fOffset);
	int stat= TRUE;
	int nextPoint= NINT(RclPairVecX1(pp,p-1));

	for (i = 0; i < num && stat; i++) {
	  double tLeft= tRight;
	  if (i > nextPoint && /* change polynom */
	      p < pNum) {	/* linear extrapolation beyond last point */
	    double m_old= m;

	    p++;
	    if (RclPairVecX1(pp,p-1) == RclPairVecX1(pp,p)) {
	      RclsetErrNo(RCLEQPAIR); return FALSE; }
	    m= ((RclPairVecX2(pp,p) - RclPairVecX2(pp,p-1))
		/ (RclPairVecX1(pp,p) - RclPairVecX1(pp,p-1)));
	    fOffset= RclPairVecX1(pp,p-1);
	    tOffset= RclPairVecX2(pp,p-1);
	    nextPoint= NINT(RclPairVecX1(pp,p));
	    tRight= tOffset + m * ((double)i + 0.5 - fOffset);
	    if (tRight <= tLeft) { /* mean distribution of bounding channel */
	      tLeft= tOffset + (m_old + m) * 0.5 * ((double)i - 0.5 - fOffset);
	      tRight= tLeft + (m_old + m) * 0.5; }
	    SPLIT(tLeft, tRight, num, t, f[i], stat);
	    tRight= tOffset + m * ((double)i + 0.5 - fOffset); }
	  else {
	    tRight= tOffset + m * ((double)i + 0.5 - fOffset);
	    SPLIT(tLeft, tRight, num, t, f[i], stat); } }
	return stat; } } }
}
/* ------------------------------------------------------------------------- */
int RclPairVecPrint(fp, pp, idx)
     FILE *fp;
     RclPair_t *pp;
     int idx;
{
  if (!fp || !pp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    RclPair_t *p= pp + idx;

    return (fprintf(fp, "%f %f", RclPairX1(p), RclPairX2(p)) != EOF &&
	    fprintf(fp, " #chn: %.0f %.0f",
		    RclPairXChn(p), RclPairChiChn(p)) != EOF &&
	    fprintf(fp, " #rcl: %.1f %.1f %.0f",
		    RclPairXRcl(p), RclPairDRcl(p), RclPairChiRcl(p)) != EOF &&
	    fprintf(fp, " #rev: %.1f %.1f %.0f",
		    RclPairXRev(p), RclPairDRev(p), RclPairChiRev(p)) != EOF &&
	    fprintf(fp, " #CoM: %.3f %.3f %.0f",
		    RclPairXCoM(p), RclPairDCoM(p), RclPairChiCoM(p)) != EOF &&
	    fprintf(fp, "\n")); }
}
/* ------------------------------------------------------------------------- */
int RclPairVecPrintShort(fp, pp, idx)
     FILE *fp;
     RclPair_t *pp;
     int idx;
{
  if (!fp || !pp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    RclPair_t *p= pp + idx;

    return (fprintf(fp, "%f %f\n", RclPairX1(p), RclPairX2(p)) != EOF); }
}
/* ------------------------------------------------------------------------- */
int RclPairVecScan(buffer, pp, idx)
     char *buffer;
     RclPair_t *pp;
     int idx;
{
  if (!buffer || !pp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    RclPair_t *p= pp + idx;
    memset(p, 0, sizeof(RclPair_t));
    return (sscanf(buffer,
      "%lf%lf #chn: %lf%lf #rcl: %lf%lf%lf #rev: %lf%lf%lf #CoM: %lf%lf%lf\n",
		   &RclPairX1(p), &RclPairX2(p),
		   &RclPairXChn(p), &RclPairChiChn(p),
		   &RclPairXRcl(p), &RclPairDRcl(p), &RclPairChiRcl(p),
		   &RclPairXRev(p), &RclPairDRev(p), &RclPairChiRev(p),
		   &RclPairXCoM(p), &RclPairDCoM(p), &RclPairChiCoM(p))
	    >= 2); }
}
/* ------------------------------------------------------------------------- */
int RclPolynomRecal (t, f, num, pol, polNum)
     SpcReal_t *t;
     SpcReal_t *f;
     int num;
     double *pol;
     int polNum;
{
  if (!t || !f || !pol) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    { int i; for (i = 0; i < num; i++) { t[i] = 0.0; } }

    { int i;
      int stat= TRUE;
      double tRight = Horner(pol, polNum, -0.5);

      for (i = 0; i < num && stat; i++) {
	double tLeft = tRight;
	tRight = Horner(pol, polNum, (double)(i + 0.5));

	SPLIT(tLeft, tRight, num, t, f[i], stat); }
      return stat; } }
}
/* ------------------------------------------------------------------------- */
int RclRead(typ, rp, file)
     int typ;
     Rcl_t rp;
     char *file;
{
  if (!rp || !file) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    switch (typ) {
    case RCL_TPOLYNOM:
      if (VecRead(file, "", RclPoly(rp), TlsScanDVec)) { break; }
      else { return FALSE; }
    case RCL_TPAIR:
    case RCL_TCOM:
    case RCL_TSQUARE:
      if (VecRead(file, "", RclPair(rp), RclPairVecScan)) { break; }
      else { return FALSE; }
    default:
      RclsetErrNo(RCLTYP); return FALSE; }
    RclTyp(rp)= typ;
    return TRUE; }
}
/* ------------------------------------------------------------------------- */
int RclRecalibrate(sp, rp)
     Spc_t sp;
     Rcl_t rp;
{
  if (!sp || !rp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else if (SpcNum(sp) <= 0) { RclsetErrNo(RCLNORES); return FALSE; }
  else {
    int stat;
/*    SpcReal_t StackAlloc(SpcReal_t, rclTmp, SpcNum(sp), SPC_MR);*/
    SpcReal_t *rclTmp = (SpcReal_t*)malloc (SpcNum (sp) * sizeof (SpcReal_t));

    if (!rclTmp) { RclsetErrNo(RCLALLOC); return FALSE; }

    switch (RclTyp(rp)) {
    case RCL_TPOLYNOM:
      { SpcReal_t *s= SpcBufReal(SpcSpc(sp));
	stat= RclPolynomRecal(rclTmp, s, SpcNum(sp),
			      RclPolyP(rp), RclPolyNum(rp));
      { int i; for (i = 0; i < SpcNum(sp); i++) { s[i] = rclTmp[i]; } } }

      { SpcReal_t *s= SpcBufReal(SpcDSpc(sp));
	stat= RclPolynomRecal(rclTmp, s, SpcNum(sp),
			      RclPolyP(rp), RclPolyNum(rp)) && stat;
	{ int i; for (i = 0; i < SpcNum(sp); i++) { s[i] = rclTmp[i]; } } }
      break;
    case RCL_TPAIR:
    case RCL_TCOM:
    case RCL_TSQUARE:
      { SpcReal_t *s= SpcBufReal(SpcSpc(sp));
	stat= RclPairRecal(rclTmp, s, SpcNum(sp), RclPair(rp));
	{ int i; for (i = 0; i < SpcNum(sp); i++) { s[i] = rclTmp[i]; } } }

      { SpcReal_t *s= SpcBufReal(SpcDSpc(sp));
	stat= RclPairRecal(rclTmp, s, SpcNum(sp), RclPair(rp)) && stat;
	{ int i; for (i = 0; i < SpcNum(sp); i++) { s[i] = rclTmp[i]; } } }
      break;
    default:
      RclsetErrNo(RCLTYP); stat= FALSE; }
    Free (rclTmp);
    return stat; }
}
/* ------------------------------------------------------------------------- */
int RclReset(rp)
     Rcl_t rp;
{
  if (!rp) { RclsetErrNo(RCLPOINT); return FALSE; }
  else {
    VecReset(RclPair(rp));
    VecReset(RclPoly(rp));
    RclTyp(rp)= RCL_TUNDEF;
    return TRUE; }
}
/* ------------------------------------------------------------------------- */
