
/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsRecord.c,v 1.4 1994/12/21 17:12:49 vs Exp vs $ */

#define _vsRecord_c

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "vsTypes.h"
#include "vsAlloc.h"
#include "vsEndian.h"

#include "vsPath.h"

#include "vsRecord.h"
#include "vsRecordP.h"

/* ---------------------------------------------------------------------- */

/* ================ */
/* static constants */
/* ================ */

#define REC_COPSIZ	32768


/* ---------------------------------------------------------------------- */
char *RecError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* RECNOERR  */ "vsRecord: no error",
  /* RECPOINT  */ "vsRecord: illegal NULL-pointer",
  /* RECALLOC  */ "vsRecord: memory allocation failed",
  /* RECOPEN   */ "vsRecord: could not open file",
  /* RECCLOSE  */ "vsRecord: could not close file",
  /* RECNOFIL  */ "vsRecord: no file specified",
  /* RECBOPEN  */ "vsRecord: could not open backup file",
  /* RECBLINK  */ "vsRecord: could not create temporary backup file",
  /* RECBREMOV */ "vsRecord: could not remove backup file",
  /* RECILLTYP */ "vsRecord: illegal record type (<=0) specified",
  /* RECTLUPDT */ "vsRecord: specified record type not in toc; update !",
  /* RECTLFND  */ "vsRecord: specified record type not in toc",
  /* RECREAD   */ "vsRecord: read error",
  /* RECILLPOS */ "vsRecord: illegal file position specified",
  /* RECSEEK   */ "vsRecord: could not position file",
  /* RECMAGIC  */ "vsRecord: magic number not matched",
  /* RECVERS   */ "vsRecord: incompatible vsRecord-versions used",
  /* RECTYPNUM */ "vsRecord: illegal size type specified",
  /* RECWRITE  */ "vsRecord: write error",
  /* RECLOCKED */ "vsRecord: backup exists; file locked; recover!",
  /* RECFOPEN  */ "vsRecord: file not closed before reset",
  /* RECMOD    */ "vsRecord: sizeof(array) % sizeof(type) != 0",
  /* RECSIZEOF */ "vsRecord: sizeof() had unexpected result",
  /* RECPUTEOR */ "vsRecord: put found eor of subrecord",
  /* RECGETEOR */ "vsRecord: get found eor of subrecord",
  /* RECSTAT   */ "vsRecord: could not stat() file",
  /* RECMODIF  */ "vsRecord: file modified; - reopen!",
  /* RECBFILE  */ "vsRecord: can not create backup file name",
  /* default:  */ "vsRecord: undefined error code", };

  if (err_no < sizeof(errors)) {
    return errors[err_no];
  }
  else {
    return errors[sizeof(errors)-1];
  }
}
/* ---------------------------------------------------------------------- */

/* ================ */
/* open/close RFILE */
/* ================ */

RFILE *RecOpen(file, mode)
     char *file;
     char *mode;
{
  RFILE *rfp;

  if ((sizeof(char) != 1) || (sizeof(short) != 2) ||
      (sizeof(int) != 4) || (sizeof(float) != 4)) {
    RecErrNo= RECSIZEOF;
    return NULL;
  }

  /*  printf("RecOpen\n");*/

  if (file == NULL || mode == NULL) {
    RecErrNo= RECPOINT;
    return NULL;
  }
  else if (NULL == (rfp= (RFILE *)Malloc(sizeof(RFILE)))) {
    RecErrNo= RECALLOC;
    return NULL;
  }
  else if (!RecFInit(rfp)) {
    Free(rfp);
    return NULL;
  }
  else if (!(RFilFile(rfp)= Strdup(file))) {
    RecErrNo= RECALLOC;
    Free(rfp);
    return NULL;
  }
  else {
    extern int errno;
    struct stat buf;

    RFilStat(rfp) = ((stat(file, &buf) < 0 ||
		      buf.st_size == 0 ||
		      strchr(mode, 'w') != NULL) ?
		     REC_SNEW : REC_SREAD);

    
    if ((RFilStat(rfp) == REC_SNEW) && (strcmp(mode, "r+") == 0)) {
      strcpy(RFilMode(rfp), "w+");
    }
    else {
      strcpy(RFilMode(rfp), mode);
    }

    errno= 0;
    RFilModifTime(rfp)= buf.st_mtime;

    /*    printf("stat: %d mode: %s\n", RFilStat(rfp), RFilMode(rfp));*/

    if (RecLocked(rfp) && RecErrNo == RECLOCKED) {
      Free(RFilFile(rfp));
      Free(rfp);
      return NULL;
    }
    else if (NULL == (RFilFfp(rfp)= FOPEN(file, RFilMode(rfp)))) {
      RecErrNo= RECOPEN;
      Free(RFilFile(rfp));
      Free(rfp);
      return NULL;
    }
    else if (RFilStat(rfp) != REC_SNEW && !RecReadToc(RFilF(rfp))) {
      FCLOSE(RFilFfp(rfp));
      Free(RFilFile(rfp));
      Free(rfp);
      return NULL;
    }
    else if (stat(file, &buf) < 0) {
      RecErrNo= RECSTAT;
      FCLOSE(RFilFfp(rfp));
      Free(RFilFile(rfp));
      Free(rfp);
      return NULL;
    }
    else {
      RFilDevice(rfp)= (int)buf.st_dev;
      RFilINode(rfp)= (int)buf.st_ino;
      RFilModifTime(rfp)= buf.st_mtime;
      return rfp;
    }
  }
}
/* ---------------------------------------------------------------------- */
int RecClose(rfp)
     RFILE *rfp;
{
  /*  printf("RecClose\n");*/
  if (rfp != NULL) {
    if (!RecBackupFinish(rfp)) {
      return FALSE;
    }
    else if (strchr(RFilMode(rfp), '+') && fflush(RFilFfp(rfp)) == EOF) {
      RFilFfp(rfp)= NULL;
      RecErrNo= RECCLOSE;
      return FALSE;
    }
    else {
      FCLOSE(RFilFfp(rfp));
      if (RFilStat(rfp) == REC_SNEW) { /* unlink new file */
	extern int unlink();
	unlink(RFilFile(rfp));
      }

      RFilFfp(rfp)= NULL;
      RFilStat(rfp)= REC_SREAD;

      RecFilFree(RFilF(rfp));
      RecFilFree(RFilB(rfp));
      Free(RFilFile(rfp));
      Free(rfp);
    }
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecCloseForce(rfp)		/* close anyway */
     RFILE *rfp;
{
  /*  printf("RecCloseForce\n");*/
  if (RecClose(rfp)) {
    return TRUE;
  }
  else {
    if (RFilFfp(rfp) != NULL) {
      FCLOSE(RFilFfp(rfp));
    }
    RecFilFree(RFilF(rfp));
    RecFilFree(RFilB(rfp));
    Free(RFilFile(rfp));
    Free(rfp);
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
int RecBackupFinish(rfp)
     RFILE *rfp;
{
  /*  printf("RecBackupFinish\n");*/
  if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else if (!RFilBackup(rfp)) {
    return TRUE;
  }
  else {
    char bfile[MAXPATHLEN];
    if (!RecBackupFile(bfile, RFilFile(rfp), "!h/#!t#")) {
      return FALSE;
    }
    if (NULL == (RFilBfp(rfp)= FOPEN(bfile, "r"))) {
      RecErrNo= RECBOPEN;
      return FALSE;
    }
    else {
      struct stat buf;
      int result= TRUE;
      extern int unlink();

      if (!RecReadToc(RFilB(rfp))) {
	result= FALSE;
      }
      else if (!RecTailCopy(RFilF(rfp), RFilB(rfp))) {
	result= FALSE;
      }
      FCLOSE(RFilBfp(rfp));
      RFilBfp(rfp)= NULL;

      if (!result || !RecWriteToc(RFilF(rfp))) {
	return FALSE;
      }
      else if (unlink(bfile) < 0) {
	RecErrNo= RECBREMOV;
	return FALSE;
      }
      else {
	RFilBackup(rfp)= FALSE;
	if (stat(RFilFile(rfp), &buf) < 0) {
	  RecErrNo= RECSTAT;
	  return FALSE;
	}
	else {
	  RFilModifTime(rfp)= buf.st_mtime;
	  return TRUE;
	}
      }
    }
  }
}
/* ---------------------------------------------------------------------- */

/* ======================= */
/* read records from RFILE */
/* ======================= */

static int ReadStart(rtp, pos)
     RFil_t *rtp;
     unsigned int pos;
{
  /*  printf ("ReadStart : %d\n",pos);*/
  if (RFtFp(rtp) == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else if (pos == 0) {
    RecErrNo= RECILLPOS;
    return FALSE;
  }
  else if (fseek(RFtFp(rtp), (long)pos, SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  else if (!ReadInt4(rtp, &RecMagic(rtp)) ||
	   !ReadInt4(rtp, &RecTyp(rtp)) ||
	   !ReadInt4(rtp, &RecVersion(rtp)) ||
	   !ReadInt4(rtp, &RecNum(rtp))) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  else if (RecMagic(rtp) != REC_RMAGIC) {
    RecErrNo= RECMAGIC;
    return FALSE;
  }
  else if (RecVersion(rtp) > REC_RVERSION) {
    RecErrNo= RECVERS;
    return FALSE;
  }
  else {
    RecPos(rtp)= pos;
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
static int ReadFinish(rtp)
     RFil_t *rtp;
{
  /*  printf("ReadFinish\n");*/
  if (fseek(RFtFp(rtp), (long)(RecPos(rtp) + RecNum(rtp)), SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecReadStart(rfp, pos)
     RFILE *rfp;
     unsigned int pos;
{
  /*  printf("RecReadStart\n");*/
  if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else {
    return ReadStart(RFilF(rfp), pos);
  }
}
/* ---------------------------------------------------------------------- */
int RecReadFinish(rfp)
     RFILE *rfp;
{
  /*  printf("RecReadFinish\n");*/
  return ReadFinish(RFilF(rfp));
}
/* ---------------------------------------------------------------------- */
int RecTailReadStart(rfp, typ)
     RFILE *rfp;
     int typ;
{
  int idx;
  /*  printf("RecTailReadStart\n");*/
  if (typ <= 0) {
    RecErrNo= RECILLTYP;
    return FALSE;
  }	/* 0 reserved */
  else if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else if (RecLocked(rfp)) {
    return FALSE;
  }
  else if ((idx= RecTocIdx(RFilF(rfp), typ)) < 0) {
    RecErrNo= RFilBackup(rfp) ? RECTLUPDT : RECTLFND;
    return FALSE;
  }
  else if (RFtTocPos(RFilF(rfp),idx) == 0) {
    return RecReset(RFilF(rfp));
  }
  else {
    return RecReadStart(rfp, RFtTocPos(RFilF(rfp),idx));
  }
}
/* ---------------------------------------------------------------------- */
/* RecTailReadFinish() is a cpp-macro */
/* ---------------------------------------------------------------------- */
static int RecTocIdx(rtp, typ)
     RFil_t *rtp;
     int typ;
{
  int i;
  /*  printf("RecTocIdx\n");*/
  for (i= 0; i< RFtTailRecNum(rtp); i++) {
    if (RFtTocTyp(rtp,i) == typ) {
      return i;
    }
  }
  return -1;
}
/* ---------------------------------------------------------------------- */

/* ====================== */
/* write records to RFILE */
/* ====================== */

static int WriteStart(rtp, typ, startPos)
     RFil_t *rtp;
     int typ;
     unsigned int startPos;
{
  /*  printf("WriteStart\n");*/
  RecMagic(rtp)= REC_RMAGIC;
  RecVersion(rtp)= REC_RVERSION;
  RecTyp(rtp)= typ;
  RecNum(rtp)= 4 * 4;
  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
  /* this offset must correspond to WriteFinish() */
  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
  /*  printf("Startposition: %8o\n", RecPos(rtp)+RecNum(rtp));*/
  RecPos(rtp)= startPos;
  if (fseek(RFtFp(rtp), (long)(RecPos(rtp)+RecNum(rtp)), SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int WriteFinish(rtp)
     RFil_t *rtp;
{
  int num= RecNum(rtp);
  /*  printf("WriteFinish\n");*/
  if (fseek(RFtFp(rtp), (long)(RecPos(rtp)), SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  else if (!WriteInt4(rtp, RecMagic(rtp)) ||
	   !WriteInt4(rtp, RecTyp(rtp)) || 
	   !WriteInt4(rtp, RecVersion(rtp)) ||
	   !WriteInt4(rtp, num)) {
    RecNum(rtp)= num;
    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    /* 4 4-byte integers, the preset-value in WriteStart() must correspond */
    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    RecErrNo= RECWRITE;
    return FALSE;
  }
  else {
    RecNum(rtp)= num;
    if (fseek(RFtFp(rtp),
		 (long)(RecPos(rtp)+RecNum(rtp)), SEEK_SET) != 0) {
      RecErrNo= RECSEEK;
      return FALSE;
    }
    else {
      return TRUE;
    }
  }
}
/* ---------------------------------------------------------------------- */
int RecWriteStart(rfp, typ)
     RFILE *rfp;
     int typ;
{
  /*  printf("RecWriteStart\n");*/
  if (typ <= 0) {
    RecErrNo= RECILLTYP;
    return FALSE;
  }	/* 0 reserved */
  else if (rfp == NULL) {
    /*
     * No file specified
     */
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else if (!RecBackup(rfp) ||
	   /* RFilStat is set in function RecOpen */
	   (RFilStat(rfp) == REC_SNEW && !RecWriteHead(RFilF(rfp))) ||
	   !RecWriteMode(rfp) ||
	   !RecTailReset(RFilF(rfp), RFilTailStart(rfp))) {
    return FALSE;
  }
  else {
    return WriteStart(RFilF(rfp), typ, RFilTailStart(rfp));
  }
}
/* ---------------------------------------------------------------------- */
int RecWriteFinish(rfp, posStart)
     RFILE *rfp;
     unsigned int *posStart;
{
  /*  printf("RecWriteFinish\n");*/
  if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else if (posStart == NULL) {
    RecErrNo= RECILLPOS;
    return FALSE;
  }
  else {
    RFil_t *rtp= RFilF(rfp);

    if (WriteFinish(rtp)) {
      RFilRecNum(rfp)++;
      *posStart= RecPos(rtp);
      RecTailReset(rtp, (unsigned)(RecPos(rtp)+RecNum(rtp)));
      return TRUE;
    }
    else {
      return FALSE;
    }
  }
}
/* ---------------------------------------------------------------------- */
int RecTailWriteStart(rfp, typ)
     RFILE *rfp;
     int typ;
{
  /*  printf("RecTailWriteStart\n");*/
  if (typ <= 0) {
    RecErrNo= RECILLTYP;
    return FALSE;
  }	/* 0 reserved */
  else if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else { return (RecBackup(rfp) &&
		 (RFilStat(rfp) != REC_SNEW || RecWriteHead(RFilF(rfp))) &&
		 RecWriteMode(rfp) &&
		 WriteStart(RFilF(rfp), typ, RFtTailToc(RFilF(rfp))));
  }
}
/* ---------------------------------------------------------------------- */
int RecTailWriteFinish(rfp)
     RFILE *rfp;
{
  /*  printf("RecTailWriteFinish\n");*/
  if (rfp == NULL) {
    RecErrNo= RECNOFIL;
    return FALSE;
  }
  else {
    RFil_t *rtp= RFilF(rfp);

    if (WriteFinish(rtp) &&
	RecUpdateToc(rtp, (unsigned)RecPos(rtp), RecTyp(rtp))) {
      RFtTailToc(rtp)= RecPos(rtp) + RecNum(rtp);
      return TRUE;
    }
    else {
      return FALSE;
    }
  }
}
/* ---------------------------------------------------------------------- */
static int RecUpdateToc(rtp, pos, typ)
     RFil_t *rtp;
     unsigned int pos;
     int typ;
{
  int idx;

  /*  printf("RecUpdateToc\n");*/
  if ((idx= RecTocIdx(rtp, typ)) < 0) {
    idx= RFtTailRecNum(rtp);
    if (!RecTocRealloc(rtp, idx+1)) {
      return FALSE;
    }
    RFtTailRecNum(rtp)= idx + 1;
  }
  RFtTocTyp(rtp,idx)= typ;
  RFtTocPos(rtp,idx)= pos;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int RecWriteMode(rfp)
     RFILE *rfp;
{
  RFil_t *rtp= RFilF(rfp);

  /*  printf("RecWriteMode\n");*/
  if (RFtStat(rtp) == REC_SREAD &&
      RFtFp(rtp) != NULL &&
      RFtTailEnd(rtp) != 0) {
    if (fseek(RFtFp(rtp), (long)RFtTailEnd(rtp), SEEK_SET) != 0) {
      RecErrNo= RECSEEK;
      return FALSE;
    }
    else if (!WriteInt4(rtp, 0)) { /* eliminate end-of-tail-mark */
      return FALSE;
    }
  }
  RFtTailEnd(rtp) = 0;
  RFtStat(rtp) = REC_SWRITE;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int RecLocked(rfp)
     RFILE *rfp;
{
  char bfile[MAXPATHLEN];
  struct stat buf;

  /*  printf("RecLocked\n");*/
  if (RFilBackup(rfp)) {
    return FALSE;
  }
  else if (!RecBackupFile(bfile, RFilFile(rfp), "!h/#!t#")) {
    return FALSE;
  }
  else if (stat(RFilFile(rfp), &buf) >= 0 &&
      buf.st_mtime > RFilModifTime(rfp)) {
    RecErrNo= RECMODIF;
    return TRUE;
  }
  else if (stat(bfile, &buf) < 0) {
    extern int errno;
    errno= 0;
    return FALSE;
  }
  else {
    RecErrNo= RECLOCKED;
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */

/* =================== */
/* write/read head/toc */
/* =================== */

static int RecWriteHead(rtp)
     RFil_t *rtp;
{
  /*  printf("RecWriteHead\n");*/
  rewind(RFtFp(rtp));
  if (fseek(RFtFp(rtp), 0L, SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  else if (!(WriteInt4(rtp, RFtMagic(rtp)) &&
	     WriteInt4(rtp, RFtUserTyp(rtp)) &&
	     WriteInt4(rtp, RFtVersion(rtp)) &&
	     WriteInt4(rtp, RFtRecNum(rtp)) &&
	     WriteInt4(rtp, RFtTailRecNum(rtp)) &&
	     WriteUInt4(rtp, RFtRecStart(rtp)) &&
	     WriteUInt4(rtp, RFtTailStart(rtp)) &&
	     WriteUInt4(rtp, RFtTailToc(rtp)) &&
	     WriteUInt4(rtp, RFtTailEnd(rtp)))) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  else { 
    unsigned int pos= (unsigned int)ftell(RFtFp(rtp));
    /*    printf("position: %d\n", pos);*/
    if (RFtRecStart(rtp) == 0) {
      RFtRecStart(rtp) = RFtTailStart(rtp) = RFtTailToc(rtp) = pos;
    }
    if (RFtStat(rtp) == REC_SNEW) {
      RFtStat(rtp) = REC_SWRITE;
    }
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
static int RecReadHead(rtp)
     RFil_t *rtp;
{
  /*  printf("RecReadHead\n");*/
  if (fseek(RFtFp(rtp), (long)0, SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  else if (!(ReadInt4(rtp, &RFtMagic(rtp)) &&
	     ReadInt4(rtp, &RFtUserTyp(rtp)) &&
	     ReadInt4(rtp, &RFtVersion(rtp)) &&
	     ReadInt4(rtp, &RFtRecNum(rtp)) &&
	     ReadInt4(rtp, &RFtTailRecNum(rtp)) &&
	     ReadUInt4(rtp, &RFtRecStart(rtp)) &&
	     ReadUInt4(rtp, &RFtTailStart(rtp)) &&
	     ReadUInt4(rtp, &RFtTailToc(rtp)) &&
	     ReadUInt4(rtp, &RFtTailEnd(rtp)))) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  else if (RFtMagic(rtp) != REC_HMAGIC) {
    RecErrNo= RECMAGIC;
    return FALSE;
  }
  else if (RFtVersion(rtp) != REC_HVERSION) {
    RecErrNo= RECVERS;
    return FALSE;
  }
  else {
    if (RFtRecStart(rtp) == 0) {
      RFtStat(rtp)= REC_SNEW;
    }
    return TRUE;
  }
}
/* ----------------------------------------------------------------------
 *
 */
static int RecWriteToc(rtp)
     RFil_t *rtp;
{
  /*  printf("RecWriteToc1\n");*/
  if (fseek(RFtFp(rtp), (long)RFtTailToc(rtp), SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
    return FALSE;
  }
  /* Write magic number for beginning of table of contents */
  else if (!WriteInt4(rtp, REC_BMAGIC)) {
    return FALSE;
  }
  else {
    int i;
    for (i= 0; i< RFtTailRecNum(rtp); i++) {
      if (!WriteInt4(rtp, RFtTocTyp(rtp,i)) ||
	  !WriteUInt4(rtp, RFtTocPos(rtp,i))) {
	return FALSE;
      }
    }
    RFtTailEnd(rtp)= (unsigned int)ftell(RFtFp(rtp));
    if (RecWriteHead(rtp)) {
      /*      printf("Executed RecWriteHead in function RecWriteToc\n");*/
      if (fseek(RFtFp(rtp), (long)RFtTailEnd(rtp), SEEK_SET) != 0) {
	RecErrNo= RECSEEK;
      }
      /* Write magic number for end of table of contents */
      else if (WriteInt4(rtp, REC_EMAGIC)) {
	if (fflush(RFtFp(rtp)) == EOF) {
	  RecErrNo= RECWRITE;
	}
	else {
	  return TRUE;
	}
      }
    }
    RFtTailEnd(rtp)= 0;
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
static int RecReadToc(rtp)
     RFil_t *rtp;
{
  /*  printf("RecReadToc\n");*/
  if (!RecReadHead(rtp)) {
    return FALSE;
  }
  else if (fseek(RFtFp(rtp), (long)RFtTailToc(rtp), SEEK_SET) != 0) {
    RecErrNo= RECSEEK;
  }
  else {
    unsigned int bot;
    unsigned int eot;

    if (ReadInt4(rtp, &bot)) {
      if (bot != REC_BMAGIC) {
	/*
	 * Was geht hier schief???
	 * Irgendwas mit der Typkonversion!!
	 */
	RecErrNo= RECMAGIC;
      }
      else if (RFtTailRecNum(rtp) == 0) {
	return RecTailReset(rtp, 0);
      }
      else if (RecTocRealloc(rtp, RFtTailRecNum(rtp))) {
	int i;
	for (i= 0; i< RFtTailRecNum(rtp); i++) {
	  if (!ReadInt4(rtp, &RFtTocTyp(rtp,i)) ||
	      !ReadUInt4(rtp, &RFtTocPos(rtp,i))) {
	    RFtTailEnd(rtp)= 0;
	    return FALSE;
	  }
	}
	if (ReadInt4(rtp, &eot)) {
	  if (eot != REC_EMAGIC) {
	    RecErrNo= RECMAGIC;
	  }
	  else {
	    RFtStat(rtp) = REC_SREAD;
	    return TRUE;
	  }
	}
      }
    }
  }
  RFtTailEnd(rtp)= 0;
  return FALSE; 
}
/* ---------------------------------------------------------------------- */

/* ============== */
/* backup/recover */
/* ============== */

int RecRecover(file)
     char *file;
{
  RFILE *rfp;

  /*  printf("RecRecover\n");*/
  if (file == NULL) {
    RecErrNo= RECPOINT;
    return FALSE;
  }
  else if (NULL == (rfp= (RFILE *)Malloc(sizeof(RFILE)))) {
     RecErrNo= RECALLOC;
     return FALSE;
  }
  else if (!RecFInit(rfp)) {
    Free(RFilFile(rfp));
    Free(rfp);
    return FALSE;
  }
  else if (!(RFilFile(rfp)= Strdup(file))) {
    RecErrNo= RECALLOC;
    Free(rfp);
    return FALSE;
  }
  else {
    strcpy(RFilMode(rfp), "a+");

    if (NULL == (RFilFfp(rfp)= FOPEN(file, "a+"))) {
      RecErrNo= RECOPEN;
      Free(RFilFile(rfp));
      Free(rfp);
      return FALSE;
    }
    else if (RecReadToc(RFilF(rfp))) {
      return RecClose(rfp);
    }
    else if (RecTailReset(RFilF(rfp), RFilTailStart(rfp))) {
      RFilBackup(rfp)= TRUE;
      return RecClose(rfp);
    }
    else {
      return FALSE;
    }
  }
}
/* ---------------------------------------------------------------------- */
static int RecBackup(rfp)
     RFILE *rfp;
{
  /*  printf("RecBackup\n");*/
  /*
   * Does a backup exist?
   */
  if (RFilBackup(rfp)) {
    return TRUE;
  }
  /*  else if (RecLocked(rfp)) {
   *    return FALSE;
  }*/
  else {
    /*
     * Create backup!
     */
    return RecBackupIntern(rfp);
  }
}
/* ----------------------------------------------------------------------
 * Generate pathname from filename f with format expr
 * Put pathname in variable t
 */
static int RecBackupFile(t, f, expr)
     char *t;
     char *f;
     char *expr;
{
  char *r;
  /*  printf("RecBackupFile\n");*/
  if ((r= PthSubstitute(f, expr)) && *r) {
    strcpy(t, r);
    return TRUE;
  }
  else {
    RecErrNo= RECBFILE;
    return FALSE;
  }
}
/* ----------------------------------------------------------------------
 * Function to create record backup file
 */
static int RecBackupIntern(rfp)
     RFILE *rfp;
{
  /* Create filename */
  char bfile[MAXPATHLEN];
  /*  printf("RecBackupIntern\n");*/
  if (!RecBackupFile(bfile, RFilFile(rfp), "!h/#!t#")) {
    return FALSE;
  }
  /* Reset all necessary variables */
  else if (!RecFilReset(RFilB(rfp))) {
    return FALSE;
  }
  /* Open backup file */
  else if (NULL == (RFilBfp(rfp)= FOPEN(bfile, "w"))) {
    RecErrNo= RECBOPEN;
    return FALSE;
  }
  else {
    int result;
    /*
     * Write some data to backupfile
     * Write header
     * Copy taildata to backupfile
     * Write table of contents
     */
    result = (RecWriteHead(RFilB(rfp)) &&
	      RecTailCopy(RFilB(rfp), RFilF(rfp)) &&
	      RecWriteToc(RFilB(rfp)));
    FCLOSE(RFilBfp(rfp));
    RFilBfp(rfp)= NULL;
    RFilBackup(rfp)= TRUE;
    return result;
  }
}
/* ----------------------------------------------------------------------
 * Copy record tail from file f to file t
 */
static int RecTailCopy(t, f)
     RFil_t *t;
     RFil_t *f;
{
  int i;
  /*  printf("RecTailCopy\n");*/
  for (i= 0; i< RFtTailRecNum(f); i++) {
    int typ= RFtTocTyp(f,i);
    if (RecTocIdx(t, typ) < 0) { /* do not supersede records */
      unsigned int fPos= RFtTocPos(f,i);
      unsigned int tPosStart= RFtTailToc(t);
      unsigned int tPosEnd;

      if (fPos == 0) {		/* do not copy deleted records */
	if (!RecUpdateToc(t, 0, typ)) {
	  return FALSE;
	}
	continue;
      }
      else if (!ReadStart(f, fPos)) {
	return FALSE;
      }
      else if (!WriteStart(t, typ, tPosStart)) {
	return FALSE;
      }
      else {
	int j;
	char buf[REC_COPSIZ];

	for (j= RecNum(f); j > 0; j -= REC_COPSIZ) {
	  unsigned int siz = MIN(j,REC_COPSIZ);
	  if (!ReadChars(f, buf, siz) || !WriteChars(t, buf, siz)) {
	    return FALSE;
	  }
	}
	ReadFinish(f);
	WriteFinish(t);
	tPosEnd= (unsigned int)ftell(RFtFp(t));
	if (!RecUpdateToc(t, tPosStart, typ)) {
	  return FALSE;
	}
	RFtTailToc(t)= RecPos(t) + RecNum(t);
      }
    }
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */

/* ================================= */
/* initialize RFILE and sub-struct's */
/* ================================= */

static int RecFInit(rfp)
     RFILE *rfp;
{
  /*  printf("RecFInit\n");*/
  RFilFile(rfp)= NULL;
  *RFilMode(rfp) = '\0';
  RFilStat(rfp) = REC_SREAD;
  RFilBackup(rfp) = FALSE;
  return (RecFilInit(RFilF(rfp)) && RecFilInit(RFilB(rfp)));
}
/* ---------------------------------------------------------------------- */
static int RecFilInit(rtp)
     RFil_t *rtp;
{
  /*  printf("RecFilInit\n");*/
  RFtToc(rtp)= NULL;
  RFtSiz(rtp)= 0;
  RFtFp(rtp)= NULL;
  return RecFilReset(rtp);
}
/* ----------------------------------------------------------------------
 * Reset variables for record file
 */
static int RecFilReset(rtp)
     RFil_t *rtp;
{
  /*  printf("RecFilReset\n");*/
  if (RFtFp(rtp) != NULL) {
    RecErrNo= RECFOPEN;
    return FALSE;
  }
  else {
    RFtMagic(rtp)= REC_HMAGIC;
    RFtUserTyp(rtp) = 0;
    RFtVersion(rtp) = REC_HVERSION;
    RFtRecNum(rtp)= 0;
    RFtRecStart(rtp)= 0;
    return RecTailReset(rtp, 0);
  }
}
/* ----------------------------------------------------------------------
 * Reset variables for record tail
 */
static int RecTailReset(rtp, pos)
     RFil_t *rtp;
     unsigned int pos;
{
  /*  printf("RecTailReset\n");*/
  RFtTailRecNum(rtp)= 0;
  RFtTailToc(rtp)= RFtTailStart(rtp)= pos;
  RFtTailEnd(rtp)= 0;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int RecTocRealloc(rtp, num)
     RFil_t *rtp;
     int num;
{
  /*  printf("RecTocRealloc\n");*/
  if (RFtSiz(rtp) < num) {
    RFilToc_t *t;
    if (NULL ==
	(t= (RFilToc_t *)Realloc(RFtToc(rtp), num * sizeof(RFilToc_t)))) {
      RecErrNo= RECALLOC;
      return FALSE;
    }
    else { RFtToc(rtp)= t; RFtSiz(rtp)= num; } }

  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int RecFilFree(rtp)
     RFil_t *rtp;
{
  Free(RFtToc(rtp));
  RFtSiz(rtp)= 0;
  return TRUE;
}
/* ---------------------------------------------------------------------- */

/* ================= */
/* endian conversion */
/* ================= */

/* ========== */
/* char-array */
/* ========== */

static int WriteChars(rtp, c, num)
     RFil_t *rtp;
     char *c;
     unsigned int num;
{
  /*  printf ("wc: %8lo %s %d\n", ftell(RFtFp(rtp)), c, num);*/
  if (FWRITE(c, sizeof(char), (size_t)num, RFtFp(rtp)) != num) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wcnp: %8lo\n", ftell(RFtFp(rtp)));*/
  RecNum(rtp) += num * sizeof(char);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadChars(rtp, c, num)
     RFil_t *rtp;
     char *c;
     unsigned int num;
{
  if (fread(c, sizeof(char), (size_t)num, RFtFp(rtp)) != num) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteChars(rfp, c, num)
     RFILE *rfp;
     char *c;
     unsigned int num;
{
  return WriteChars(RFilF(rfp), c, num);
}
/* ---------------------------------------------------------------------- */
int RecReadChars(rfp, c, num)
     RFILE *rfp;
     char *c;
     unsigned int num;
{
  return ReadChars(RFilF(rfp), c, num);
}
/* ---------------------------------------------------------------------- */
int RecWriteString(rfp, s)
     RFILE *rfp;
     char *s;
{
  unsigned int num=strlen(s);

  return (WriteUInt4(RFilF(rfp), num) &&
	  WriteChars(RFilF(rfp), s, num+1));
}
/* ---------------------------------------------------------------------- */
int RecReadString(rfp, s)
     RFILE *rfp;
     char *s;
{
  unsigned int num;

  return (ReadUInt4(RFilF(rfp), &num) &&
	  ReadChars(RFilF(rfp), s, num+1));
}
/* ---------------------------------------------------------------------- */
/* ============== */
/* 2 byte integer */
/* ============== */

static int WriteInt2(rtp, i)
     RFil_t *rtp;
     short i;
{
  short i2= i;
  i2= GETLE2(i2);
  /*  printf ("wi2: %8lo %8x %8x\n",  ftell(RFtFp(rtp)), i, i2);*/
  if (FWRITE(&i2, sizeof(short), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wi2np: %8lo\n", ftell(RFtFp(rtp)));*/
  RecNum(rtp) += sizeof(short);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadInt2(rtp, i)
     RFil_t *rtp;
     short *i;
{
  int i2;
  if (fread(&i2, sizeof(short), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  *i= GETLE2(i2);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteInt2(rfp, i)
     RFILE *rfp;
     short i;
{
  return WriteInt2(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
int RecReadInt2(rfp, i)
     RFILE *rfp;
     short *i;
{
  return ReadInt2(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
/* ======================= */
/* 2 byte unsigned integer */
/* ======================= */

static int WriteUInt2(rtp, i)
     RFil_t *rtp;
     unsigned short i;
{
  unsigned short i2= i;
  i2= GETLE2(i2);
  /*  printf ("wui2:%8lo %8x %8x\n", ftell (RFtFp(rtp)), i, i2);*/
  if (FWRITE(&i2, sizeof(unsigned short), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wui2np:%8lo\n", ftell (RFtFp(rtp)));*/
  RecNum(rtp) += sizeof(unsigned short);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadUInt2(rtp, i)
     RFil_t *rtp;
     unsigned short *i;
{
  unsigned short i2;
  if (fread(&i2, sizeof(unsigned short), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  *i= GETLE2(i2);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteUInt2(rfp, i)
     RFILE *rfp;
     unsigned short i;
{
  return WriteUInt2(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
int RecReadUInt2(rfp, i)
     RFILE *rfp;
     unsigned short *i;
{
  return ReadUInt2(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
/* ============== */
/* 4 byte integer */
/* ============== */

static int WriteInt4(rtp, i)
     RFil_t *rtp;
     int i;
{
  int i4= i;
  i4= GETLE4(i4);
  /*  printf ("wi4: %8lo %8x %8x\n", ftell(RFtFp(rtp)), i, i4);*/
  if (FWRITE(&i4, sizeof(int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wi4np: %8lo\n", ftell(RFtFp(rtp)));*/
  RecNum(rtp) += sizeof(int);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadInt4(rtp, i)
     RFil_t *rtp;
     int *i;
{
  int i4;
  /*  long int ttt = ftell (RFtFp(rtp));*/
  if (fread(&i4, sizeof(int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  *i= GETLE4(i4);
  /*  printf ("ri4: %8lo %8x %8x\n", ttt, *i, i4);*/
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteInt4(rfp, i)
     RFILE *rfp;
     int i;
{
  return WriteInt4(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
int RecReadInt4(rfp, i)
     RFILE *rfp;
     int *i;
{
  return ReadInt4(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
/* ======================= */
/* 4 byte unsigned integer */
/* ======================= */

static int WriteUInt4(rtp, i)
     RFil_t *rtp;
     unsigned int i;
{
  unsigned int i4= i;
  i4= GETLE4(i4);
  /*  printf ("wui4:%8lo %8x %8x\n", ftell(RFtFp(rtp)), i, i4);*/
  if (FWRITE(&i4, sizeof(unsigned int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wui4np:%8lo\n", ftell(RFtFp(rtp)));*/
  RecNum(rtp) += sizeof(unsigned int);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadUInt4(rtp, i)
     RFil_t *rtp;
     unsigned int *i;
{
  unsigned int i4;
  /*  long int ttt = ftell (RFtFp(rtp));*/
  if (fread(&i4, sizeof(unsigned int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  *i= GETLE4(i4);
  /*  printf ("rui4:%8lo %8x %8x\n", ttt, *i, i4);*/
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteUInt4(rfp, i)
     RFILE *rfp;
     unsigned int i;
{
  return WriteUInt4(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
int RecReadUInt4(rfp, i)
     RFILE *rfp;
     unsigned int *i;
{
  return ReadUInt4(RFilF(rfp), i);
}
/* ---------------------------------------------------------------------- */
/* ============ */
/* 4 byte float */
/* ============ */

static int WriteFloat4(rtp, f)
     RFil_t *rtp;
     float f;
{
  float fi= f;
  int f4= *((int *)&fi);
  f4= GETLE4(f4);
  /*  printf ("wf4: %8lo %8.2f %8x\n", ftell(RFtFp(rtp)), fi, f4);*/
  if (FWRITE(&f4, sizeof(int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECWRITE;
    return FALSE;
  }
  /*  printf ("wf4np: %8lo\n", ftell(RFtFp(rtp)));*/
  RecNum(rtp) += sizeof(int);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int ReadFloat4(rtp, f)
     RFil_t *rtp;
     float *f;
{
  int f4;
  /*  long int ttt = ftell (RFtFp(rtp));*/
  if (fread(&f4, sizeof(int), 1, RFtFp(rtp)) != 1) {
    RecErrNo= RECREAD;
    return FALSE;
  }
  f4= GETLE4(f4);
  *f= *((float *)&f4);
  /*  printf ("rf4: %8lo %8.2f %8x\n", ttt, *f, f4);*/
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int RecWriteFloat4(rfp, f)
     RFILE *rfp;
     float f;
{
  return WriteFloat4(RFilF(rfp), f);
}
/* ---------------------------------------------------------------------- */
int RecReadFloat4(rfp, f)
     RFILE *rfp;
     float *f;
{
  return ReadFloat4(RFilF(rfp), f);
}
/* ---------------------------------------------------------------------- */
/* ================== */
/* Rec_t manipulation */
/* ================== */

int RecReset(rtp)
     RFil_t *rtp;
{
  if (rtp == NULL) {
    RecErrNo= RECPOINT;
    return FALSE;
  }
  else {
    RecMagic(rtp) = REC_RMAGIC;
    RecTyp(rtp) = -1;
    RecVersion(rtp) = REC_RVERSION;
    RecNum(rtp) = 0;
    RecPos(rtp)= 0;
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
