/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsRecord.h,v 1.2 1994/12/08 20:44:32 vs Exp vs $ */
#ifndef _vsRecord_h
#define _vsRecord_h

#include <tv/vsConfig.h>
#include <tv/vsTypes.h>

				/* Test-debug fuer fit-write-file */

#ifdef DEBUGFIT
#define FWRITE dfwrite
#define FCLOSE dfclose
#define FOPEN dfopen
size_t dfwrite(const void *ptr, size_t size, size_t nitems,
          FILE *stream);
FILE *dfopen(const char *filename, const char *type);
int dfclose(FILE *stream);
#else
#define FWRITE fwrite
#define FOPEN fopen
#define FCLOSE fclose
#endif

				/* implemented conversion types */
#define REC_CHR		0	/* (unsigned) char */
#define REC_SHT		1	/* (unsigned) short */
#define REC_INT		2	/* (unsigned) int */
#define REC_FLT		3	/* float */

/* REC_TYPNUM(vstypes.h) has to be equal to the number of implemented types */

				/* types choosen by length */
#define REC_1BYTE	REC_CHR
#define REC_2BYTE	REC_SHT
#define REC_4BYTE	REC_INT

				/* magics, types and versions */
				/* RFILE header */
#define REC_HMAGIC	11211957 /*0x00ab14b5 fitz*/
#define REC_HVERSION	1
				/* record */
#define REC_RMAGIC	21111957 /*0x01422495 fitz*/
#define REC_RVERSION	1
				/* toc */
#define REC_BMAGIC	19571121 /*0x012aa1b1 fitz ;  beginning of toc */
#define REC_EMAGIC	11195721 /*0x00aad549 fitz ; end of toc */


				/* RFILE.typ */
#define REC_FTYPE	1

				/* RFILE.version */
#define REC_FVERSION	1

				/* RFILE.stat */
#define REC_SNEW	1
#define REC_SREAD	2
#define REC_SWRITE	3

#define REC_TYPNUM	4

typedef struct {		/* RFILE header */
  int magic;			/* magic number of RFilHead_t */
  int userTyp;			/* user defined file type */
  int version;			/* file version */
  int recNum;			/* number of normal records */
  int tailRecNum;		/* number of tail records */
  unsigned int recStart;	/* file position of 1st normal record */
  unsigned int tailStart;	/* file position of 1st tail record */
  unsigned int tailToc;		/* file position of tail toc */
  unsigned int tailEnd;		/* file position of end-of-tail */
} RFilHead_t;

typedef struct {		/* toc info on tail record */
  int typ;			/* type -- only one tail record per type */
  unsigned int pos;		/* file position of tail record in question */
} RFilToc_t;

typedef struct {		/* RFILE tail (RFilToc_t vector */
  int siz;			/* allocated size */
  int num;			/* number of entries */
  RFilToc_t *toc;		/* memory location of toc */
} RFilTail_t;

typedef struct {		/* fit record header */
  int magic;			/* magic number of RecHead_t */
  int typ;			/* type */
  int version;			/* version */
  int num;			/* length in char */
  int pos;			/* position in file */
} RecRec;

typedef struct {		/* file info */
  FILE *fp;			/* file pointer */
  int stat;			/* file status */
  RFilHead_t hd;		/* header of file */
  RFilTail_t tl;		/* tail info of file */
  RecRec rec;			/* info of actual record */
} RFil_t;

#define RFtFp(rtp)		((rtp)->fp)
#define RFtStat(rtp)		((rtp)->stat)
#define RFtTl(rtp)		(&(rtp)->tl)
#define RFtHd(rtp)		(&(rtp)->hd)
#define RFtRec(rtp)		(&(rtp)->rec)
				/* RFilHead_t of RFil_t */
#define RFtMagic(rtp)		(RFtHd(rtp)->magic)
#define RFtUserTyp(rtp)		(RFtHd(rtp)->userTyp)
#define RFtVersion(rtp)		(RFtHd(rtp)->version)
#define RFtRecNum(rtp)		(RFtHd(rtp)->recNum)
#define RFtTailRecNum(rtp)	(RFtHd(rtp)->tailRecNum)
#define RFtRecStart(rtp)	(RFtHd(rtp)->recStart)
#define RFtTailStart(rtp)	(RFtHd(rtp)->tailStart)
#define RFtTailToc(rtp)		(RFtHd(rtp)->tailToc)
#define RFtTailEnd(rtp)		(RFtHd(rtp)->tailEnd)
				/* RFilTail_t of RFil_t */
#define RFtToc(rtp)		(RFtTl(rtp)->toc)
#define RFtSiz(rtp)		(RFtTl(rtp)->siz)
				/* RFilToc_t of RFilTail_t of RFil_t */
#define RFtTocTyp(rtp,i)	(RFtToc(rtp)[i].typ)
#define RFtTocPos(rtp,i)	(RFtToc(rtp)[i].pos)

#define RecMagic(rtp)		(RFtRec(rtp)->magic)
#define RecTyp(rtp)		(RFtRec(rtp)->typ)
#define RecVersion(rtp)		(RFtRec(rtp)->version)
#define RecNum(rtp)		(RFtRec(rtp)->num)
#define RecPos(rtp)		(RFtRec(rtp)->pos)

typedef struct _RFILERec {	/* RFILE struct; */
  char *file;			/* file name */
  char mode[8];			/* file mode */
  time_t modifTime;		/* last modification time at open */
  int backup;			/* backup file exists */
  int device;			/* device of file */
  int iNode;			/* inode number of file */
  RFil_t f;			/* info of file */
  RFil_t b;			/* info of backup file for head/tail */
} RFILERec;

#define RFilFile(rfp)		((rfp)->file)
#define RFilMode(rfp)		((rfp)->mode)
#define RFilModifTime(rfp)	((rfp)->modifTime)
#define RFilBackup(rfp)		((rfp)->backup)
#define RFilDevice(rfp)		((rfp)->device)
#define RFilINode(rfp)		((rfp)->iNode)
#define RFilF(rfp)		(&(rfp)->f)
#define RFilB(rfp)		(&(rfp)->b)
    
#define RFilFfp(rfp)		(RFtFp(RFilF(rfp)))
#define RFilRecNum(rfp)		(RFtRecNum(RFilF(rfp)))
#define RFilTailStart(rfp)	(RFtTailStart(RFilF(rfp)))
#define RFilStat(rfp)		(RFtStat(RFilF(rfp)))
#define RFilUserTyp(rfp)	(RFtUserTyp(RFilF(rfp)))

#define RFilBfp(rfp)		(RFtFp(RFilB(rfp)))

				/* error codes of RecErrNo */
#define RECNOERR	 0 /* vsRecord: no error */
#define RECPOINT	 1 /* vsRecord: illegal NULL-pointer */
#define RECALLOC	 2 /* vsRecord: memory allocation failed */
#define RECOPEN		 3 /* vsRecord: could not open file */
#define RECCLOSE	 4 /* vsRecord: could not close file */
#define RECNOFIL	 5 /* vsRecord: no file specified */
#define RECBOPEN	 6 /* vsRecord: could not open backup file */
#define RECBLINK	 7 /* vsRecord: could not create temporary backup file */
#define RECBREMOV	 8 /* vsRecord: could not remove backup file */
#define RECILLTYP	 9 /* vsRecord: illegal record type (<=0) specified */
#define RECTLUPDT	10 /* vsRecord: spec. record type not in toc; update ! */
#define RECTLFND	11 /* vsRecord: specified record type not in toc */
#define RECREAD		12 /* vsRecord: read error */
#define RECILLPOS	13 /* vsRecord: illegal file position specified */
#define RECSEEK		14 /* vsRecord: could not position file */
#define RECMAGIC	15 /* vsRecord: magic number not matched */
#define RECVERS		16 /* vsRecord: incompatible vsRecord-versions used */
#define RECTYPNUM	17 /* vsRecord: illegal size type specified */
#define RECWRITE	18 /* vsRecord: write error */
#define RECLOCKED	19 /* vsRecord: backup exists; file locked; recover! */
#define RECFOPEN	20 /* vsRecord: file not closed before reset */
#define RECMOD		21 /* vsRecord: sizeof(array) % sizeof(type) != 0 */
#define RECSIZEOF	22 /* vsRecord: sizeof() had unexpected result */
#define RECPUTEOR	23 /* vsRecord: put found end of subrecord */
#define RECGETEOR	24 /* vsRecord: get found end of subrecord */
#define RECSTAT		25 /* vsRecord: could not stat() file */
#define RECMODIF	26 /* "vsRecord: file modified; - reopen!" */
#define RECBFILE	27 /* "vsRecord: can not create backup file name" */

#ifdef _vsRecord_c
int RecErrNo=RECNOERR;
#else
extern int RecErrNo;
#endif

#define RecTailReadFinish(rfp)	RecReadFinish(rfp)

#if !defined(NOPROTO) && (defined(__STDC__) || defined(VSPROTO) || defined(__cplusplus))
# define P_(s) s
#else
# define P_(s) ()
#endif

#include <tv/vsRecordF.h>

#undef P_

#endif				/* _vsRecord_h */
