/* 
 * vsSearch.h
 * includefile for vsSearch.c
 *
 * peaksearch library Sassi 
 *
 * Martin Luig,   IKP Cologne
 *
 * started 17. December 1991
 *
 * copyright: sassi library version 1.0 (c) (1991,1992), M. Luig
 *            Institut of Nuclear Physics, University of Cologne.
 *            The library and the header files may be used free of charge,
 *            provided that the following conditions are met:
 *            1. There is no warranty of any kind for all parts of the library.
 *            2. If the use of the library is connected to any form of
 *               publication, the authors must be cited correctly.
 *            3. The authors prohibit, that other person give away or sell copies
 *               of this software. Copies are given only by the authors under the
 *               following postal address:  Universitaet zu Koeln,
 *               Zuelpicherstr. 77, D-5000 Koeln 41, Fed. Rep. Germany.
 */

#include <tv/vsConfig.h>
#include <tv/vsTypes.h>

#define SEANOERR	0 /* "vsSearch: no error", */
#define SEAPOINT	1 /* "vsSearch: illegal NULL-pointer", */
#define SEAALLOC	2 /* "vsSearch: can not allocate buffer", */
#define SEAEXIST	3 /* "vsSearch: peak already exists", */
#define SEAUSER		4 /* "vsSearch: error occured in user function", */
#define SEASPCTYP	5 /* "vsSearch: unknown spectrum type", */
#define SEANOPEAK	6 /* "vsSearch: no peak found", */

#define SeasetErrNo(e)		{ SeaErrNo= e; }

#ifndef _vsSearch_h
#define _vsSearch_h
#ifdef _vsSearch_c
int SeaErrNo= SEANOERR;
#else
extern int SeaErrNo;
#endif

#ifndef T_CHAR
#  define  T_CHAR     0          /* character          */
#  define  T_SHORT    1          /* short integer      */
#  define  T_INT      2          /* integer            */
#  define  T_LONG     3          /* long integer       */
#  define  T_FLOAT    4          /* floating point     */
#  define  T_DOUBLE   5          /* double word        */
#  define  T_UCHAR    6          /* unsigned character */
#  define  T_USHORT   7          /* unsigned short     */
#  define  T_UINT     8          /* unsigned integer   */
#  define  T_ULONG    9          /* unsigned long      */
#endif

#ifndef TRUE
#	define TRUE 1
#	define FALSE 0
#endif

typedef struct _SeaPeakRec {
  float pos;
  float prob;
  float amplitude, amplitude_err;
  float background, background_err;
  float vol, vol_err; 
  float chi;
  float width;
  struct _SeaPeakRec *next;
  struct _SeaPeakRec *prev;
} SeaPeakRec;

typedef struct {
  int		number_peaks;
  SeaPeak	start_list;
  SeaPeak	end_list;
} SeaPeakListRec, *SeaPeakList;

#ifdef __STDC__ 
#define P_(s) s
#else
#define P_(s) ()
#endif

extern char *SeaError P_((int err_no));
extern SeaPeakList SeaPeakSearch P_((void *spec, void *err_spec,
				     int spec_type, int res,
				     int left, int right, double def_prob,
				     double (*w_cal)(), Cal_t cal,
				     int (*peak_func)()));
extern SeaPeakList SeaCreatePeakList P_((void));
extern SeaPeakList SeaDupPeakList P_((SeaPeakList src));
extern SeaPeak SeaEnterPeak P_((SeaPeakList peak_list, double channel,
				double Vol, double dVol, double Prob,
				double Amp, double dAmp, double BG, double dBG,
				double Chi, double width));
extern SeaPeak SeaInsertPeak P_((SeaPeakList peak_list, void *spec,
				 void *err_spec, int spec_type, int res,
				 double channel, double min_prob,
				 double (*w_cal)(), Cal_t cal));
extern SeaPeak SeaClosestPeak P_((SeaPeakList pl, double chan));
extern void SeaFreePeakList P_((SeaPeakList pl));
extern void SeaFreeClosestPeak P_((SeaPeakList pl, double chan));
extern double SeaBgFunction P_((SeaPeak sp, double x));
extern double SeaFunction P_((SeaPeak sp, double x));
extern double SeaPeakFunction P_((SeaPeak sp, double x));

#undef P_
#define ForAllSeaPeaks(pl, p) for(p= SeaListFirst(pl);p != NULL;p= SeaNext(p))

#define SeaPos(_p)		((_p)->pos)
#define SeaProb(_p)		((_p)->prob)
#define SeaAmpl(_p)		((_p)->amplitude)
#define SeaDAmpl(_p)		((_p)->amplitude_err)
#define SeaBg(_p)		((_p)->background)
#define SeaDBg(_p)		((_p)->background_err)
#define SeaVol(_p)		((_p)->vol)
#define SeaDVol(_p)		((_p)->vol_err)
#define SeaChi(_p)		((_p)->chi)
#define SeaWidth(_p)		((_p)->width)
#define SeaNext(_p)		((_p)->next)
#define SeaPrev(_p)		((_p)->prev)

#define SeaListFirst(_pl)	((_pl)->start_list)
#define SeaListLast(_pl)	((_pl)->end_list)
#define SeaListNum(_pl)		((_pl)->number_peaks)

#endif				/* _vsSearch_h */



