/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsSpectra.c,v 1.3 1994/12/02 21:23:20 vs Exp vs $ */
#define _vsSpectra_c

#include "vsConfig.h"

#include <mfile.h>

#ifdef HAS_SHM
#	include <sys/types.h>
#	include <sys/stat.h>
#	include <fcntl.h>
#	include <limits.h>

#ifdef __STDC__
# include <stdarg.h>
#else
# include <varargs.h>
#endif

#	include <sysdep/shm_alloc.h>
#endif				/* HAS_SHM */

#include "vsTypes.h"
#include "vsAlloc.h"
#include "vsCal.h"

#include "vsTools.h"

#include "vsSpectra.h"
#include "vsSpectraP.h"
/* ---------------------------------------------------------------------- */
char *SpcError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* SPCNOERR:  */ "vsSpectra: no error",
  /* SPCPOINT:  */ "vsSpectra: illegal NULL-pointer",
  /* SPCOPEN:   */ "vsSpectra: could not open spectrafile",
  /* SPCGTINF:  */ "vsSpectra: could not get mfile info",
  /* SPCINVAL:  */ "vsSpectra: unknown or invalid filetype",
  /* SPCSTFMT:  */ "vsSpectra: could not set mfile format",
  /* SPCSTINF:  */ "vsSpectra: could not set mfile info",
  /* SPCMRES:   */ "vsSpectra: maximal resolution exceeded",
  /* SPCWRITE:  */ "vsSpectra: could not write spectrum",
  /* SPCLEV:    */ "vsSpectra: illegal level specified",
  /* SPC0READ:  */ "vsSpectra: nothing read from file",
  /* SPCTYP:    */ "vsSpectra: spectrum type not implemented",
  /* SPCILRES:  */ "vsSpectra: illegal resolution specified",
  /* SPCFIXED:  */ "vsSpectra: fixed buffer can not be freed",
  /* SPCDRES:   */ "vsSpectra: different resolutions",
  /* SPCDTYP:   */ "vsSpectra: different buffer types",
  /* SPCWRTLV:  */ "vsSpectra: can not write specified level",
  /* SPCALLOC:  */ "vsSpectra: can not allocate buffer",
  /* SPCLIN:    */ "vsSpectra: illegal line specified",
  /* SPCTYPSIZ: */ "vsSpectra: internal error: incompatible size of SpcReal_t",
  /* SPCNOFIL:  */ "vsSpectra: internal error: no MFILE open",
  /* SPCSHMALC: */ "vsSpectra: shared memory allocation error",
  /* default:   */ "vsSpectra: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */

/* =============== */
/* i/o of SpcBuf_t */
/* =============== */

SpcFile_t SpcOpen(file, format, mode)
     char *file;
     char *format;
     char *mode;
{
  SpcFile_t sfp;
  char *mf;

/*   printf("SpcOpen %s %s\n", file, format); */

  /* mopen understands mode and format concatenated*/
  if (0 == strcmp(mode,"w")) {
    mf = (char*) malloc(strlen(format)+strlen(mode)+2);
    sprintf(mf, "%s,%s", mode, format);
  } else {
    mf = mode;
  }
/*   printf("1\n"); */
  if (!file || !mode) {
    SpcsetErrNo(SPCPOINT);
    return NULL;
  } else {
    sfp = (SpcFile_t)Malloc(sizeof(SpcFileRec));
/*     printf("2\n"); */
    if (!sfp) {
      SpcsetErrNo(SPCALLOC);
      return NULL;
    } else {
      SpcFileMfp(sfp)= mopen(file, mf);
/*       printf("3 %d\n", SpcFileMfp(sfp)); */
      if (SpcFileMfp(sfp)) {
/*  	printf("4\n"); */
	if (format != NULL && *format != '\0' &&
	    msetfmt(SpcFileMfp(sfp), format)< 0) {
	  SpcsetErrNo(SPCSTFMT);
	  mclose(SpcFileMfp(sfp));
	  Free(mf);
	  Free(sfp);
	  return NULL;
	} else {
/*  	  printf("5\n"); */
	  if (mgetinfo(SpcFileMfp(sfp), SpcFileInfo(sfp))< 0 ||
	      !mgetfmt(SpcFileMfp(sfp), SpcFileFormat(sfp))) {
	    SpcsetErrNo(SPCGTINF);
	    mclose(SpcFileMfp(sfp));
	    Free(mf);
	    Free(sfp);
	    return NULL;
	  } else {
	    strcpy(SpcFileMode(sfp), mode);
	    return sfp;
	  }
	}
      } else {
/*  	printf("6\n"); */
	SpcsetErrNo(SPCOPEN);
	if (0 == strcmp(mode,"w")) {
	  Free(mf);
	}
	Free(sfp);
	return NULL;
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
int SpcClose(sfp)
     SpcFile_t sfp;
{
  int status;

  /*  printf("SpcClose\n");*/
  if (!sfp) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else if (!SpcFileMfp(sfp)) {
    return TRUE;
  }
  else {
    status= (mclose(SpcFileMfp(sfp)) >= 0);
    Free(sfp);
    return status;
  }
}
/* ---------------------------------------------------------------------- */
static int SpcBufRead(sbp, sfp, level, line)
     SpcBuf_t *sbp;
     SpcFile_t sfp;
     int level;
     int line;
{
  int res;

  if (!sfp || !sbp) { SpcsetErrNo(SPCPOINT); return -1; }
  else if (!SpcFileMfp(sfp)) { SpcsetErrNo(SPCNOFIL); return -1; }
  else if (level< 0 || level> SpcFileLevels(sfp)) {
    SpcsetErrNo(SPCLEV); return -1; }
  else if (line< 0 || line> SpcFileLines(sfp)) {
    SpcsetErrNo(SPCLIN); return -1; }
  else if (SpcBufSiz(sbp) <
	   (res= SpcFileColumns(sfp)) * VecTypSiz(SpcBufB(sbp))) {
      SpcsetErrNo(SPCMRES); return -1; }
  else {
    MFILE *mfp= SpcFileMfp(sfp);
    int r;
    switch (SpcBufTyp(sbp)) {
    case SPC_FLT:
      r = mgetflt(mfp, SpcBufFlt(sbp), level, line, 0, res);
      break;
    case SPC_REAL:
      if (sizeof(double) != sizeof(SpcReal_t)) {
	SpcsetErrNo(SPCTYPSIZ); return -1; }
    case SPC_DBL:
      r = mgetdbl(mfp, SpcBufDbl(sbp), level, line, 0, res);
      break;
    case SPC_INT:
      r = mgetint(mfp, SpcBufInt(sbp), level, line, 0, res);
      break;
    default: SpcsetErrNo(SPCTYP); return -1; }

    if (r <= 0) { SpcsetErrNo(SPC0READ); return -1; }
    else { SpcBufNum(sbp) = r; return r; } }
}
/* ---------------------------------------------------------------------- */
static int SpcBufWrite(sbp, sfp, level, line)
     SpcBuf_t *sbp;
     SpcFile_t sfp;
     int line;
     int level;
{
  /*  printf("SpcBufWrite\n");*/
  if (sfp == NULL || sbp == NULL) {
    SpcsetErrNo(SPCPOINT);
    return -1;
  }
  else if (!SpcFileMfp(sfp)) {
    SpcsetErrNo(SPCNOFIL);
    return -1;
  }
  else if (SpcFileFiletype(sfp) == MAT_INVALID) {
    SpcsetErrNo(SPCINVAL);
    return -1;
  }
  else if (level< 0 || level>= SpcFileLevels(sfp)) {
    SpcsetErrNo(SPCLEV);
    return -1;
  }
  else if (line< 0 || line >= SpcFileLines(sfp)) {
    SpcsetErrNo(SPCLIN);
    return FALSE;
  }
  else {
    MFILE *mfp= SpcFileMfp(sfp);
    int res= SpcBufNum(sbp);
    int r;

    switch (SpcBufTyp(sbp)) {
    case SPC_FLT:
      /*      printf("SPC_FLT\n");*/
      r = mputflt(mfp, SpcBufFlt(sbp), level, line, 0, res);
      break;
    case SPC_REAL:
      /*      printf("SPC_REAL\n");*/
      if (sizeof(double) != sizeof(SpcReal_t)) {
	SpcsetErrNo(SPCTYPSIZ);
	return -1;
      }
    case SPC_DBL:
      /*      printf("SPC_DBL\n");*/
      r = mputdbl(mfp, SpcBufDbl(sbp), level, line, 0, res);
      break;
    case SPC_INT:
      /*      printf("SPC_INT\n");*/
      r = mputint(mfp, SpcBufInt(sbp), level, line, 0, res);
      break;
    default: SpcsetErrNo(SPCTYP);
      return -1;
    }

    if (r != res) {
      SpcsetErrNo(SPCWRITE);
      return -1;
    }
    else {
      return r;
    }
  }
}
/* ---------------------------------------------------------------------- */

/* =================== */
/* initialize SpcBuf_t */
/* =================== */

static int SpcBufAlloc(sbp, siz)
     SpcBuf_t *sbp;
     int siz;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (siz <= 0) { SpcsetErrNo(SPCILRES); return FALSE; }
  else if (siz > SpcMaxRes) { SpcsetErrNo(SPCMRES); return FALSE; }
  else { return VecAlloc(SpcBufB(sbp), siz); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufRealloc(sbp, siz)
     SpcBuf_t *sbp;
     int siz;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (siz <= 0) { SpcsetErrNo(SPCILRES); return FALSE; }
  else if (siz > SpcMaxRes) { SpcsetErrNo(SPCMRES); return FALSE; }
  else { return VecRealloc(SpcBufB(sbp), siz); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufFree(sbp)
     SpcBuf_t *sbp;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else { return VecFree(SpcBufB(sbp)); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufInit(sbp, typ)
     SpcBuf_t *sbp;
     int typ;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    int typSiz;
    switch (typ) {
    case SPC_INT: typSiz= sizeof(int); break;
    case SPC_FLT: typSiz= sizeof(float); break;
    case SPC_REAL:typSiz= sizeof(SpcReal_t); break;
    case SPC_DBL: typSiz= sizeof(double); break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }

    SpcBufTyp(sbp) = typ;
#ifdef HAS_SHM
    SpcBufShm(sbp) = 0;
#endif				/* HAS_SHM */
    return (VecInit(SpcBufB(sbp), typSiz) &&
	    SpcBufReset(sbp)); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufReset(sbp)
     SpcBuf_t *sbp;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else { return VecReset(SpcBufB(sbp)); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufClear(sbp)
     SpcBuf_t *sbp;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else { return VecClear(SpcBufB(sbp)); }
}
/* ---------------------------------------------------------------------- */
static int SpcBufSet(sbp, buf, typ, siz, num)
     SpcBuf_t *sbp;
     void *buf;
     int typ;
     int siz;
     int num;
{
  if (sbp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    int typSiz;
    switch (typ) {
    case SPC_INT: typSiz= sizeof(int); break;
    case SPC_FLT: typSiz= sizeof(float); break;
    case SPC_REAL:typSiz= sizeof(SpcReal_t); break;
    case SPC_DBL: typSiz= sizeof(double); break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    SpcBufTyp(sbp) = typ;
    return VecSet(SpcBufB(sbp), buf, typSiz, siz, num); }
}
/* ---------------------------------------------------------------------- */

/* ======================================================= */
/* operations on SpcBuf_t (copy, add, sub, equal, integer, and) */
/* ======================================================= */

#define LOOP(a,b,r,ret)  { int i; for (i=0;i<r;i++) {OP(a[i],b[i],ret)} }
#define LOOPCAL(a,b,r,ret)  { int i; for (i=0;i<r;i++) {OP(a[i],b[(int)(0.5+CalC (c_f, (CalP (c_t, (double)i))))],ret)} }

#define OP_BUF(a,sbp,res,ret) \
  {\
    switch (SpcBufTyp(sbp)) {\
    case SPC_INT: {int *b=SpcBufInt(sbp);LOOP(a,b,res,ret);}break;\
    case SPC_FLT: {float *b=SpcBufFlt(sbp);LOOP(a,b,res,ret);}break;\
    case SPC_REAL:{SpcReal_t *b=SpcBufReal(sbp);LOOP(a,b,res,ret);}break;\
    case SPC_DBL: {double *b=SpcBufDbl(sbp);LOOP(a,b,res,ret);}break;\
    default: SpcsetErrNo(SPCTYP); ret = FALSE; }\
  }

static int SpcBufCopy(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    int res= SpcBufNum(f);
    int ret= TRUE;
    switch (SpcBufTyp(t)) {
    case SPC_INT: {
#define OP(a,b,r) {(a)=NINT(b);}
      int *a=SpcBufInt(t); OP_BUF(a,f,res,ret); } break;
#undef OP
    case SPC_FLT: {
#define OP(a,b,r) {(a)=(b);}
      float *a=SpcBufFlt(t); OP_BUF(a,f,res,ret); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t); OP_BUF(a,f,res,ret); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t); OP_BUF(a,f,res,ret); } break;
#undef OP
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    if (ret) { SpcBufNum(t)= res; }
    return ret; }
}
/* ---------------------------------------------------------------------- */
static int SpcBufAbsCopy(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    int res= SpcBufNum(f);
    int ret= TRUE;
    switch (SpcBufTyp(t)) {
#define OP(a,b,r) {(a)=NINT(ABS(b));}
    case SPC_INT: {
      int *a=SpcBufInt(t); OP_BUF(a,f,res,ret); } break;
#undef OP
#define OP(a,b,r) {(a)=ABS(b);}
    case SPC_FLT: {
      float *a=SpcBufFlt(t); OP_BUF(a,f,res,ret); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t); OP_BUF(a,f,res,ret); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t); OP_BUF(a,f,res,ret); } break;
#undef OP
    default: SpcsetErrNo(SPCTYP); ret= FALSE; }
    if (ret) { SpcBufNum(t)= res; }
    return ret; }
}
/* ---------------------------------------------------------------------- */
static int SpcBufAdd(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)+=(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufAddCalib(t, f, c_t, c_f)
     SpcBuf_t *t;
     SpcBuf_t *f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)+=(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOPCAL(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOPCAL(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufAbsAdd(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)+=ABS(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufAbsAddCalib(t, f, c_t, c_f)
     SpcBuf_t *t;
     SpcBuf_t *f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)+=ABS(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOPCAL(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOPCAL(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufSub(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)-=(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}

/* ---------------------------------------------------------------------- */
static int SpcBufSubCalib(t, f, c_t, c_f)
     SpcBuf_t *t;
     SpcBuf_t *f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)-=(b);}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: { int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOPCAL(a,b,res,()); }break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOPCAL(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOPCAL(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}

/* ---------------------------------------------------------------------- */
static int SpcBufMin(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)=MIN((a),(b));}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
static int SpcBufMinCalib(t, f, c_t, c_f)
     SpcBuf_t *t;
     SpcBuf_t *f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)=MIN((a),(b));}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOPCAL(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOPCAL(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufMax(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)=MAX((a),(b));}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufMaxCalib(t, f, c_t, c_f)
     SpcBuf_t *t;
     SpcBuf_t *f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) {
    SpcsetErrNo(SPCDRES); return FALSE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) {
    SpcsetErrNo(SPCDTYP); return FALSE; }
  else {
#define OP(a,b,r) {(a)=MAX((a),(b));}
    int res= SpcBufNum(f);
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOPCAL(a,b,res,()); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOPCAL(a,b,res,()); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOPCAL(a,b,res,()); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufAbsEqual(t, f)
     SpcBuf_t *t;
     SpcBuf_t *f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return TRUE; }
  else if (SpcBufTyp(t) != SpcBufTyp(f)) { return FALSE; }
  else if (SpcBufNum(t) != SpcBufNum(f)) { return FALSE; }
  else {
#define OP(a,b,r) {if (ABS(a)!=ABS(b)) {r=FALSE; break;}}
    int res= SpcBufNum(f);
    int ret= TRUE;
    switch (SpcBufTyp(f)) {
    case SPC_INT: {
      int *a=SpcBufInt(t), *b=SpcBufInt(f); LOOP(a,b,res,ret); } break;
    case SPC_FLT: {
      float *a=SpcBufFlt(t), *b=SpcBufFlt(f); LOOP(a,b,res,ret); } break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(t), *b=SpcBufReal(f); LOOP(a,b,res,ret); } break;
    case SPC_DBL: {
      double *a=SpcBufDbl(t), *b=SpcBufDbl(f); LOOP(a,b,res,ret); } break;
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return ret; }
#undef OP
}
/* ---------------------------------------------------------------------- */
static int SpcBufInteger(sbp)
     SpcBuf_t *sbp;
{
  if (sbp==NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
#define OP(a,b,r) {if ((double)((int)(a))!=(double)(a)) {r=FALSE;break;}}
    int res= SpcBufNum(sbp);
    int ret= TRUE;
    switch (SpcBufTyp(sbp)) {	/* start of equal switch */
    case SPC_INT: return TRUE;
    case SPC_FLT: {
      float *a=SpcBufFlt(sbp); LOOP(a,(),res,ret);}break;
    case SPC_REAL: {
      SpcReal_t *a=SpcBufReal(sbp); LOOP(a,(),res,ret);}break;
    case SPC_DBL: {
      double *a=SpcBufDbl(sbp); LOOP(a,(),res,ret);}break;
    default: SpcsetErrNo(SPCTYP); ret= FALSE; } /* end of equal switch */
  return ret; }
#undef OP
}
#undef LOOP

/* ---------------------------------------------------------------------- */
/* ================ */
/* initialize Spc_t */
/* ================ */

int SpcAlloc(sp, num)
     Spc_t sp;
     int num;
{
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
#ifdef HAS_SHM
    Vec_t v= (Vec_t)SpcBufB(SpcSpc(sp));
    int siz= num * VecTypSiz(v);
    /*    printf("try shm\n");*/
    if (SpcSharedMemory && SpcBufShm(SpcSpc(sp))) {
      char file[128];

      if (VecSiz(v) < siz) {
	void *p;
	if (VecP(v)) { shm_free((void *)VecP(v)); VecInit(v, VecTypSiz(v)); }

	sprintf(file, "shm.tv_buffer%d", SpcBufShm(SpcSpc(sp))-1);
	if ((p= (void *)shm_alloc((unsigned int)(2 * siz),
				  file, O_CREAT | O_RDWR,
				  00444))) { /* r--r--r-- */
	  VecSet(v, p, VecTypSiz(v), siz, 0);
	  VecSet(SpcBufB(SpcDSpc(sp)), ((char *)p + siz),
		 VecTypSiz(v), siz, 0); }
	else { SpcsetErrNo(SPCSHMALC); SpcReset(sp); return FALSE; } }
      else {
	VecSet(SpcBufB(SpcDSpc(sp)), ((char *)VecP(v) + siz),
	       VecTypSiz(v), siz, 0); }
      SpcReset(sp); return TRUE; }
#endif				/* HAS_SHM */
    return ((SpcBufAlloc(SpcSpc(sp), num) &&
	     SpcBufAlloc(SpcDSpc(sp), num)) &
	    SpcReset(sp)); }
}
/* ---------------------------------------------------------------------- */
int SpcRealloc(sp, num)
     Spc_t sp;
     int num;
{
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (num <= 0) { SpcsetErrNo(SPCILRES); return FALSE; }
  else {
#ifdef HAS_SHM
    Vec_t v= (Vec_t)SpcBufB(SpcSpc(sp));
    int siz= num * VecTypSiz(v);
    if (SpcSharedMemory && SpcBufShm(SpcSpc(sp))) {
      char file[128];

      if (VecSiz(v) < siz) {
	void *p;
	if (VecP(v)) { SpcsetErrNo(SPCSHMALC); return FALSE; }

	sprintf(file, "shm.tv_buffer%d", SpcBufShm(SpcSpc(sp))-1);
	if ((p= (void *)shm_alloc((unsigned int)(2 * siz),
				  file, O_CREAT | O_RDWR,
				  00444))) { /* r--r--r-- */
	  VecSet(v, p, VecTypSiz(v), siz, 0);
	  VecSet(SpcBufB(SpcDSpc(sp)), ((char *)p + siz),
		 VecTypSiz(v), siz, 0); }
	else { SpcsetErrNo(SPCSHMALC); return FALSE; } }
      else {
	VecSet(SpcBufB(SpcDSpc(sp)), ((char *)VecP(v) + siz),
	       VecTypSiz(v), siz, 0); }
      return TRUE; }
#endif				/* HAS_SHM */
    return (SpcBufRealloc(SpcSpc(sp), num) &&
	    SpcBufRealloc(SpcDSpc(sp), num)); }
}
/* ---------------------------------------------------------------------- */
int SpcFree(sp)
     Spc_t sp;
{
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
#ifdef HAS_SHM
  else if (SpcSharedMemory && SpcBufShm(SpcSpc(sp))) {
    Vec_t v= (Vec_t)SpcBufB(SpcSpc(sp));
    if (VecP(v)) { shm_free(VecP(v)); }
    VecInit(v, VecTypSiz(v));
    v= (Vec_t)SpcBufB(SpcDSpc(sp));
    VecInit(v, VecTypSiz(v));
    return TRUE; }
#endif				/* HAS_SHM */
  else {
    return ((SpcBufFree(SpcSpc(sp)) &&
	     SpcBufFree(SpcDSpc(sp))) &
	    SpcReset(sp)); }
}
/* ---------------------------------------------------------------------- */
int SpcInit(sp, typ)
     Spc_t sp;
     int typ;
{
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    SpcBufInit(SpcSpc(sp), typ);
    SpcBufInit(SpcDSpc(sp), typ);
    SpcFile(sp)= NULL;
    SpcTyp(sp)= typ;
    SpcReset(sp);
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int SpcShmInit(sp, typ)
     Spc_t sp;
     int typ;
{
#ifdef HAS_SHM
  static int bufNum= 0;
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (SpcSharedMemory) {
    SpcBufInit(SpcSpc(sp), typ);
    SpcBufInit(SpcDSpc(sp), typ);
    SpcFile(sp)= NULL;
    bufNum++;
    SpcBufShm(SpcSpc(sp))= bufNum;
    SpcBufShm(SpcDSpc(sp))= -bufNum;
    SpcTyp(sp)= typ;
    SpcReset(sp);
    return TRUE; }
#endif				/* HAS_SHM */
  return SpcInit(sp, typ);
}
/* ---------------------------------------------------------------------- */
int SpcReset(sp)
     Spc_t sp;
{
  if (sp == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    Free(SpcFile(sp));
    *SpcFormat(sp)= '\0';
    SpcBufReset(SpcSpc(sp));
    SpcBufReset(SpcDSpc(sp));
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int SpcClear(sp)
     Spc_t sp;
{
  if (sp) { return (SpcBufClear(SpcSpc(sp)) && SpcBufClear(SpcDSpc(sp))); }
  else { SpcsetErrNo(SPCPOINT); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int SpcCopy(t, f)
     Spc_t t;
     Spc_t f;
{
  if (t == NULL || f == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    if (SpcAlloc(t, SpcNum(f)) &&
	SpcBufCopy(SpcSpc(t),SpcSpc(f)) &&
	SpcBufAbsCopy(SpcDSpc(t),SpcDSpc(f))) {
      Free(SpcFile(t));
      if (SpcFile(f) && !(SpcFile(t)= Strdup(SpcFile(f)))) {
	SpcReset(t); SpcsetErrNo(SPCALLOC); return FALSE; }
      strcpy(SpcFormat(t),SpcFormat(f));
      return TRUE; }
    else {
      SpcReset(t);
      return FALSE; } }
}
/* ---------------------------------------------------------------------- */
int SpcAdd(t, f, c_t, c_f)
     Spc_t t;
     Spc_t f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (c_t != NULL && c_f != NULL){
    return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	    (SpcBufAddCalib(SpcSpc(t),SpcSpc(f), c_t, c_f) &&
	     SpcBufAbsAddCalib(SpcDSpc(t),SpcDSpc(f), c_t, c_f)));}

  return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	  (SpcBufAdd(SpcSpc(t),SpcSpc(f)) &&
	   SpcBufAbsAdd(SpcDSpc(t),SpcDSpc(f))));
}     
/* ---------------------------------------------------------------------- */
int SpcIntegrate(sp, x1, x2, sum, dSum)
     Spc_t sp;
     double x1;
     double x2;
     double *sum;
     double *dSum;
{
  if (!sp || !sum) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    double I= 0.0;
    double dI= 0.0;
    SpcReal_t *s= SpcBufReal(SpcSpc(sp));
    SpcReal_t *dS= SpcBufReal(SpcDSpc(sp));
    int j;
    int jStart= MAX(0, NINT(x1));
    int jEnd= MIN(SpcNum(sp), NINT(x2)+1);
    for (j= jStart; j< jEnd; j++) { I += s[j]; dI += dS[j]; }

    *sum= I; if (dSum) { *dSum= dI; }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int SpcIntegratePVec(sp, pp, sum, dSum)
     Spc_t sp;
     PVec_t pp;
     double *sum;
     double *dSum;
{
  if (!sp || !pp || !sum) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else {
    double I= 0.0;
    double dI= 0.0;
    SpcReal_t *s= SpcBufReal(SpcSpc(sp));
    SpcReal_t *dS= SpcBufReal(SpcDSpc(sp));
    int i;

    for (i= 0; i< VecNum(pp); i++) {
      int j;
      int jStart= MAX(0, NPVecX1(pp,i));
      int jEnd= MIN(SpcNum(sp), NPVecX2(pp,i)+1);
      for (j= jStart; j< jEnd; j++) { I += s[j]; dI += dS[j]; } }

    *sum= I; if (dSum) { *dSum= dI; }
    return TRUE; }
}     
/* ---------------------------------------------------------------------- */
int SpcProject(sfp, x1, x2, sp)
     SpcFile_t sfp;
     double x1;
     double x2;
     Spc_t sp;
{
  if (!sfp || !sp) { SpcsetErrNo(SPCPOINT); return FALSE; }
  if (SpcFileColumns(sfp) != SpcNum(sp)) {SpcsetErrNo(SPCDRES); return FALSE;}
  else {
    int l= NINT(x1);
    int r= NINT(x2)+1;
    int i;

    l= MAX(0, l);
    r= MIN(r, SpcFileLines(sfp));

    for (i= l; i< r; i++) {
      if (!SpcReadAddLine(sp, sfp, i)) { return FALSE; } }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int SpcSub(t, f, c_t, c_f)
     Spc_t t;
     Spc_t f;
     Cal_t c_t;
     Cal_t c_f;
{

  if (c_t != NULL && c_f != NULL){
    return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	    (SpcBufSubCalib(SpcSpc(t),SpcSpc(f), c_t, c_f) &&
	     SpcBufAbsAddCalib(SpcDSpc(t),SpcDSpc(f), c_t, c_f)));}

  return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	  (SpcBufSub(SpcSpc(t),SpcSpc(f)) &&
	   SpcBufAbsAdd(SpcDSpc(t),SpcDSpc(f))));
}     

/* ---------------------------------------------------------------------- */
int SpcMin(t, f, c_t, c_f)
     Spc_t t;
     Spc_t f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (c_t != NULL && c_f != NULL){
    return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	    (SpcBufMinCalib(SpcSpc(t),SpcSpc(f), c_t, c_f) &&
	     SpcBufAbsCopy(SpcDSpc(t),SpcSpc(t))));}
  
  return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	  (SpcBufMin(SpcSpc(t),SpcSpc(f)) &&
	   SpcBufAbsCopy(SpcDSpc(t),SpcSpc(t))));
}     
/* ---------------------------------------------------------------------- */
int SpcMax(t, f, c_t, c_f)
     Spc_t t;
     Spc_t f;
     Cal_t c_t;
     Cal_t c_f;
{
  if (c_t != NULL && c_f != NULL){
    return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	    (SpcBufMaxCalib(SpcSpc(t),SpcSpc(f), c_t, c_f) &&
	     SpcBufAbsCopy(SpcDSpc(t),SpcSpc(t))));}

  return ((SpcNum(t) == 0) ? SpcCopy(t,f) :
	  (SpcBufMax(SpcSpc(t),SpcSpc(f)) &&
	   SpcBufAbsCopy(SpcDSpc(t),SpcSpc(t))));
}     
/* ---------------------------------------------------------------------- */
int SpcMultiply(sp, fac)
     Spc_t sp;
     double fac;
{
#define MUL(s,dS,r,f,f2)  { int i; for (i=0;i<r;i++) { s[i] *= f; dS[i]= f2 * dS[i]; } }

  if (sp==NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (fac == 1.0) { return TRUE; }
  else if (fac == 0.0) { return SpcClear(sp); }
  else {
    int res= SpcNum(sp);
    double fac2= fac * fac;
    
    switch (SpcTyp(sp)) {	/* start of equal switch */
    case SPC_INT: {
      int *s= SpcBufInt(SpcSpc(sp)), *dS= SpcBufInt(SpcDSpc(sp));
      MUL(s,dS,res,fac,fac2); break; }
    case SPC_FLT: {
      float *s= SpcBufFlt(SpcSpc(sp)), *dS= SpcBufFlt(SpcDSpc(sp));
      MUL(s,dS,res,fac,fac2); break; }
    case SPC_REAL: {
      SpcReal_t *s= SpcBufReal(SpcSpc(sp)), *dS= SpcBufReal(SpcDSpc(sp));
      MUL(s,dS,res,fac,fac2); break; }
    case SPC_DBL: {
      double *s= SpcBufDbl(SpcSpc(sp)), *dS= SpcBufDbl(SpcDSpc(sp));
      MUL(s,dS,res,fac,fac2); break; }
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef MUL
}
/* ---------------------------------------------------------------------- */
int SpcOffset(sp, offset)
     Spc_t sp;
     double offset;
{
#define ADD(s,r,o)  { int i; for (i=0;i<r;i++) { s[i] += o; } }

  if (sp==NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (offset == 0.0) { return TRUE; }
  else {
    int res= SpcNum(sp);
    
    switch (SpcTyp(sp)) {	/* start of equal switch */
    case SPC_INT: {
      int *s= SpcBufInt(SpcSpc(sp));
      ADD(s,res,offset); break; }
    case SPC_FLT: {
      float *s= SpcBufFlt(SpcSpc(sp));
      ADD(s,res,offset); break; }
    case SPC_REAL: {
      SpcReal_t *s= SpcBufReal(SpcSpc(sp));
      ADD(s,res,offset); break; }
    case SPC_DBL: {
      double *s= SpcBufDbl(SpcSpc(sp));
      ADD(s,res,offset); break; }
    default: SpcsetErrNo(SPCTYP); return FALSE; }
    return TRUE; }
#undef ADD
}
/* ---------------------------------------------------------------------- */
int SpcRead(sp, file, format)
     Spc_t sp;
     char *file;
     char *format;
{
/*   printf("SpcRead\n"); */
  SpcReset(sp);
#ifdef DUKE
  return SpcReadOperate(sp, file, format, NULL, NULL);
#else
  return SpcReadOperate(sp, file, format, SpcBufCopy, SpcBufAbsCopy);
#endif
}
/* ---------------------------------------------------------------------- */
int SpcReadLine(sp, sfp, line)
     Spc_t sp;
     SpcFile_t sfp;
     int line;
{
  return (SpcReset(sp) &&
#ifdef DUKE
	  SpcReadOperateLine(sp, sfp, line, NULL, NULL));
#else
  SpcReadOperateLine(sp, sfp, line, SpcBufCopy, SpcBufAbsCopy));
#endif
}
/* ---------------------------------------------------------------------- */
int SpcReadAdd(sp, file, format)
     Spc_t sp;
     char *file;
     char *format;
{
  return SpcReadOperate(sp, file, format, SpcBufAdd, SpcBufAbsAdd);
}
/* ---------------------------------------------------------------------- */
int SpcReadAddLine(sp, sfp, line)
     Spc_t sp;
     SpcFile_t sfp;
     int line;
{
  return SpcReadOperateLine(sp, sfp, line, SpcBufAdd, SpcBufAbsAdd);
}
/* ---------------------------------------------------------------------- */
static int SpcReadOperate(sp, file, format, fct, dFct)
     Spc_t sp;
     char *file;
     char *format;
     int (*fct)();
     int (*dFct)();
{
  SpcFile_t sf;
  int cpy;

/*   printf("SpcReadOperate\n"); */

  if (sp == NULL || file == NULL) { SpcsetErrNo(SPCPOINT); return FALSE; }

  cpy= (SpcNum(sp) == 0);

  if ((sf= SpcOpen(file, format, "r"))) {
/*     printf("SpcOpen succ\n"); */
    if (SpcReadOperateLine(sp, sf, 0, fct, dFct)) {
      strcpy(format, SpcFileFormat(sf));
      if (cpy &&
	  !(SpcFile(sp)= Strdup(file))) {
	SpcReset(sp);
	SpcsetErrNo(SPCALLOC);
	return FALSE;
      }
    } else {
      SpcReset(sp);
    }
    SpcClose(sf);
  } else {
    return FALSE;
  }
  if (SpcNum(sp) > 0) {
    return TRUE;
  } else {
    SpcReset(sp);
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
static int SpcReadOperateLine(sp, sfp, line, fct, dFct)
     Spc_t sp;
     SpcFile_t sfp;
     int line;
     int (*fct)();
     int (*dFct)();
{
/*   printf("SpcReadOperateLine\n"); */
  if (!sp || !sfp) { SpcsetErrNo(SPCPOINT); return FALSE; }
  else if (!SpcFileMfp(sfp)) { SpcsetErrNo(SPCNOFIL); return FALSE; }
  else {
#ifdef DUKE
    SpcReal_t *tmp = (double*)malloc( SpcMaxRes * sizeof(double));
#else
    SpcReal_t StackAlloc(double, tmp, SpcMaxRes, SPC_MR);
#endif
    SpcBuf_t sb;
    int cpy;
    int res;

    if (!tmp) { SpcsetErrNo(SPCALLOC); return FALSE; }

    SpcBufInit(&sb, SpcTyp(sp));
    SpcBufSet(&sb, tmp, SpcTyp(sp), (int)sizeof(SpcReal_t) * SpcMaxRes, 0);

    cpy= (SpcNum(sp) == 0);

    if (SpcFileFiletype(sfp) == MAT_UNKNOWN ||
	SpcFileFiletype(sfp) == MAT_INVALID ||
	SpcFileColumns(sfp) <= 0 ||
	SpcFileLevels(sfp) <= 0 ||
	SpcFileLines(sfp) <= 0) {
      SpcsetErrNo(SPCINVAL);
#ifdef DUKE
      Free (tmp);
#endif
      return FALSE; }
    else if (SpcMaxRes < (res= SpcFileColumns(sfp))) {
      SpcsetErrNo(SPCMRES); 
#ifdef DUKE
      Free (tmp);
#endif
      return FALSE; }
    else if (cpy && !SpcAlloc(sp, res)) { 
#ifdef DUKE
      Free (tmp); 
#endif
      return FALSE; }
    else if ((SpcBufRead(&sb, sfp, MAT_DATA, line) < 0) || /* not read */
	     (cpy && !SpcBufCopy(SpcSpc(sp), &sb)) ||	/* not copied */
	     (fct == NULL ? 0 :!(*fct)(SpcSpc(sp), &sb))) {/* not sub'ed/added */
      SpcReset(sp); }
    else if ((MAT_DATAERR < SpcFileLevels(sfp) && /* dSpc exists */
	      (SpcBufRead(&sb, sfp, 
			  MAT_DATAERR, line) != SpcNum(sp) && /* not read and*/
	       SpcBufRead(&sb, sfp,
			  MAT_DATA, line) != SpcNum(sp))) || /* spc not read */
	     (cpy && !SpcBufAbsCopy(SpcDSpc(sp), &sb)) || /* not copied */
	     (dFct == NULL ? 0 :!(*dFct)(SpcDSpc(sp), &sb))) { /* not sub'ed/added */
      SpcReset(sp); }
    else { SpcsetErrNo(SPCNOERR); }
    if (SpcNum(sp) > 0) {
      if (cpy) { strcpy(SpcFormat(sp), SpcFileFormat(sfp)); }
#ifdef DUKE
      Free (tmp); 
#endif
      return TRUE; }
    else { SpcReset(sp); 
#ifdef DUKE
	   Free (tmp);
#endif
	   return FALSE; } 
#ifdef DUKE
    Free (tmp);
#endif
  }
}
/* ---------------------------------------------------------------------- */
int SpcReadSub(sp, file, format)
     Spc_t sp;
     char *file;
     char *format;
{
  return SpcReadOperate(sp, file, format, SpcBufSub, SpcBufAbsAdd);
}
/* ---------------------------------------------------------------------- */
int SpcReadSubLine(sp, sfp, line)
     Spc_t sp;
     SpcFile_t sfp;
     int line;
{
  return SpcReadOperateLine(sp, sfp, line, SpcBufSub, SpcBufAbsAdd);
}
/* ---------------------------------------------------------------------- */
int SpcSet(sp, typ, typSiz, buffer, spcLen)
     Spc_t sp;
     int typ;
     size_t typSiz;
     char *buffer;
     int spcLen;
{
  if (sp && buffer) {
    return (SpcInit(sp, typ) &&
	    SpcBufSet(SpcSpc(sp), buffer,
		      typ, (int)(typSiz * spcLen), 0) &&
	    SpcBufSet(SpcDSpc(sp), buffer + typSiz * spcLen,
		      typ, (int)(typSiz * spcLen), 0)); }
  else { SpcsetErrNo(SPCPOINT); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int SpcSetInfo(sfp, mip)
     SpcFile_t sfp;
     minfo *mip;
{
  /*  printf("SpcSetInfo\n");*/
  if (!sfp || !mip) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else if (!SpcFileMfp(sfp)) {
    SpcsetErrNo(SPCNOFIL);
    return FALSE;
  }
  else if (!msetinfo(SpcFileMfp(sfp), mip) < 0) {
    SpcsetErrNo(SPCSTINF);
    return FALSE;
  }
  else if (mgetinfo(SpcFileMfp(sfp), SpcFileInfo(sfp))< 0 ||
	   !mgetfmt(SpcFileMfp(sfp), SpcFileFormat(sfp))) {
    SpcsetErrNo(SPCGTINF);
    return FALSE;
  }
  else {
    /*    printf("SpcFileFormat(sfp): %s\n", SpcFileFormat(sfp));*/
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
int SpcWrite(sp, file, format)
     Spc_t sp;
     char *file;
     char *format;
{
/*   printf("SpcWrite file %s with format %s\n", file, format); */
  if (sp == NULL || file == NULL) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else if (SpcBufNum(SpcSpc(sp)) == 0) {
    SpcsetErrNo(SPCILRES);
    return FALSE;
  }
  else if (SpcBufNum(SpcDSpc(sp)) != SpcBufNum(SpcSpc(sp))) {
    SpcsetErrNo(SPCDRES);
    return FALSE;
  }
  else {
    SpcFile_t sf;

    if (!(sf= SpcOpen(file, format, "w"))) {
      return FALSE;
    }
    else {
      int ret;
      minfo bestInfo;

      if ((ret= (SpcGetInfo(sf, &bestInfo) &&
		 SpcGetBestInfo(sp, SpcFileInfo(sf), &bestInfo) &&
		 SpcSetInfo(sf, &bestInfo) &&
		 SpcWriteLine(sp, sf, 0)))) {
	strcpy(format, SpcFileFormat(sf));
      }
      ret= SpcClose(sf) && ret;
      return ret;
    }
  }
}
/* ---------------------------------------------------------------------- */
int SpcGetBestInfo(sp, actInfo, bestInfo)
     Spc_t sp;
     minfo *actInfo;
     minfo *bestInfo;
{
  /*  printf("SpcGetBestInfo %d %d\n", actInfo->filetype, bestInfo->filetype);*/
  if (!sp || !actInfo || !bestInfo) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else {
    int spcEqual;
    int res= SpcBufNum(SpcSpc(sp));

    spcEqual= SpcBufAbsEqual(SpcSpc(sp), SpcDSpc(sp));

    if (actInfo->filetype == MAT_UNKNOWN &&
	(bestInfo->filetype == MAT_UNKNOWN ||
	 bestInfo->filetype == MAT_STD_INT)) {
      if (SpcBufInteger(SpcSpc(sp)) && 
	  (spcEqual || SpcBufInteger(SpcDSpc(sp)))) {
	bestInfo->filetype = MAT_STD_INT;
      }
      else {
	bestInfo->filetype = MAT_STD_FLT;
      }
    }

    res = ((actInfo->columns <= 0) ? res : MIN(actInfo->columns, res));
    bestInfo->levels= spcEqual ? 1 : actInfo->levels;
    bestInfo->lines= (actInfo->lines <= 0) ? 1 : actInfo->lines;
    bestInfo->columns = res;
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
int SpcGetInfo(sfp, mip)
     SpcFile_t sfp;
     minfo *mip;
{
  /*  printf("SpcGetInfo\n");*/
  if (!sfp || !mip) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else if (!SpcFileMfp(sfp)) {
    SpcsetErrNo(SPCNOFIL);
    return FALSE;
  }
  else if (mgetinfo(SpcFileMfp(sfp), mip) < 0) {
    SpcsetErrNo(SPCGTINF);
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
int SpcWriteLine(sp, sfp, line)
     Spc_t sp;
     SpcFile_t sfp;
     int line;
{
  /*  printf("SpcWriteLine\n");*/
  if (sp == NULL || sfp == NULL) {
    SpcsetErrNo(SPCPOINT);
    return FALSE;
  }
  else if (!SpcFileMfp(sfp)) {
    SpcsetErrNo(SPCNOFIL);
    return FALSE;
  }
  else if (SpcBufNum(SpcSpc(sp)) == 0) {
    SpcsetErrNo(SPCILRES);
    return FALSE;
  }
  else if (SpcBufNum(SpcDSpc(sp)) != SpcBufNum(SpcSpc(sp))) {
    SpcsetErrNo(SPCDRES);
    return FALSE;
  }
  else if (SpcBufWrite(SpcSpc(sp), sfp, MAT_DATA, line) < 0) {
    return FALSE;
  }
  else if (SpcFileLevels(sfp) > MAT_DATAERR &&
	   SpcBufWrite(SpcDSpc(sp), sfp, MAT_DATAERR, line) < 0) {
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
