/* $Header: /usr/files/home/Gamma/vs/tv/edit/clients/tv/RCS/VdKey.c,v 1.2 1994/05/26 19:51:16 vs Exp vs $ */

#define _VdKey_c
#include <limits.h>		/* CHAR_MAX */

#include <tv/vsTypes.h>
#include <tv/vsAlloc.h>
#include <tv/vsAscii.h>
#include <tv/vsKey.h>
#include <tv/vsPath.h>

typedef struct _VdKeyRec {
  Keytable_t keytable;
  struct _VdKeyRec *next;
} VdKeyRec, *VdKey;

#define ForAllKeytables(vkt)	for (vkt= root; vkt; vkt= vkt->next)
#include "VsUserFace.h"
#include "VsError.h"
#include "VdKey.h"
#include "VdKeyP.h"

/* ---------------------------------------------------------------------- */
static VdKey current= NULL;
static VdKey root= NULL;
/* ---------------------------------------------------------------------- */
int KT_Activate()
{
  char title[IOBUFLEN];

  if (UF_ItemRead(title, IOBUFLEN, "keytable title") && *title) {
    VdKey vkt;

    if ((vkt= GetKeytable(title))) { UseKeytable(vkt); return TRUE; }
    else { ErrPrintR("fatal: no keytable '%s' found\n", title); } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_Ascii()
{
  int c_min, c_max;
  if (UF_NoArgs()) { c_min= 0; c_max= CHAR_MAX; }
  else if (UF_IntRead(FALSE, &c_min, "char-value") && UF_NoArgs()) {
    c_max= c_min; }
  else if (UF_IntRead(FALSE, &c_max, "char-value")) { ; }
  else { return FALSE; }

  if (c_min > c_max) { int tmp= c_min; c_min= c_max; c_max= tmp; }
  c_min= MAX(0, c_min);
  c_max= MIN(CHAR_MAX, c_max);
  c_min= MIN(c_min, c_max);

  { int i;

    for (i= c_min; i< c_max + 1; i++) {
      UF_PrintM("%3d %4x  %03o  %-3s  %-3s\n",
		i, i, i, ascii_code[i], ascii_ctrl[i]); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int KT_Create()
{
  char title[IOBUFLEN];

  if (UF_ItemRead(title, IOBUFLEN, "title of keytable to create") &&
      *title) {
    VdKey vkt;
    if ((vkt= GetKeytable(title))) {
      UF_PrintE("warning: keytable '%s' already exists\n", title);
      return TRUE; }
    else {
      Keytable_t kt;
      if ((kt= KTCreateKeytable(title, ""))) {
	return InsertKeytable(kt); }
      else {
	ErrPrintR("fatal: can not create keytable '%s'\n", title); } } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_Delete()
{
  char title[IOBUFLEN];

  if (UF_ItemRead(title, IOBUFLEN, "title of keytable to delete") &&
      *title) {
    VdKey destroy;

    if ((destroy= GetKeytable(title))) {
      VdKey vkt;

      ForAllKeytables(vkt) {
	if (vkt->next == destroy) {
	  vkt->next= vkt->next->next;
	  KTDestroyKeytable(destroy->keytable);
	  Free(destroy);
	  UF_PrintM("keytable '%s' deleted\n", title);
	  if (current == destroy) { UseKeytable(NULL); }
	  return TRUE; } } }
    UF_PrintE("warning: keytable '%s' not found\n", title); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_ReadReplace()
{
  char title[IOBUFLEN];
  char file[IOBUFLEN];

  if (UF_ItemRead(file, IOBUFLEN, "keytable input file") && *file) {
    if (UF_NoArgs()) { *title= '\0'; }
    else if (!UF_ItemRead(title, IOBUFLEN, "title keytable to read")) {
      return FALSE; }
    return KT_ReadFile(title, file, FALSE, FALSE); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_ReadMerge()
{
  char title[IOBUFLEN];
  char file[IOBUFLEN];

  if (UF_ItemRead(file, IOBUFLEN, "keytable input file") && *file) {
    if (UF_NoArgs()) { *title= '\0'; }
    else if (!UF_ItemRead(title, IOBUFLEN, "title keytable to read")) {
      return FALSE; }
    return KT_ReadFile(title, file, TRUE, FALSE); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_ReadFile(title, file, merge, try)
     char *title;
     char *file;
     int merge;
     int try;
{
  char *path;
  char t[IOBUFLEN];
  if (!title) { title= t; *t= '\0'; }

  if ((path= UF_Find(file))) {
    if (!KTReadTitle(title, path)) { ; }
    else {
      VdKey vkt;
      Keytable_t kt;
      kt= ((vkt= GetKeytable(title)) && merge) ? vkt->keytable : NULL;
      if ((kt= KTReadKeytable(kt, title, path, merge))) {
	if (vkt) {
	  if (merge) {
	    UF_PrintM("keytable '%s'(%s) merged\n",
		      KeytableTitle(kt), PthShort(KeytableFile(kt)));
	    return TRUE; }
	  else { return ReplaceKeytable(vkt, kt); } }
	else { return InsertKeytable(kt); } } }
    ErrPrintR("fatal: can not read keytable '%s'(%s)\n",
	      (title ? title : ""), PthShort(path)); }
  else if (!try) {
    ErrPrintR("fatal: can not find keytable file '%s'\n", file); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_Show()
{
  if (!root) {
    UF_PrintE("warning: no keytables defined\n"); return FALSE; }
  else {
    VdKey vkt;

    if (current) {
      UF_PrintM("title: %-20s file: %-30s current\n",
		KeytableTitle(current->keytable),
		PthShort(KeytableFile(current->keytable))); }

    ForAllKeytables(vkt) {
      if (vkt != current) {
	UF_PrintM("title: %-20s file: %-30s\n",
		  KeytableTitle(vkt->keytable),
		  PthShort(KeytableFile(vkt->keytable))); } }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int KT_Write()
{
  char title[IOBUFLEN];
  char file[IOBUFLEN];

  if (UF_ItemRead(file, IOBUFLEN, "keytable output file") && *file &&
      UF_ItemRead(title, IOBUFLEN, "keytable title") && *title) {
      
    char *path;

    if ((path= PthExpand(file))) {
      VdKey vkt;

      if ((vkt= GetKeytable(title))) {
	if (KTWriteKeytable(vkt->keytable, path)) {
	  UF_PrintM("keytable '%s' written to '%s'\n",
		    title, PthShort(path));
	  return TRUE; }
	else {
	  ErrPrint("error: can not write keytable '%s' to '%s'\n",
		   title, PthShort(path));
	  return FALSE; } }
      else { ErrPrint("error: can not find keytable '%s'\n", title); } }
    else { ErrPrint("error: can not expand '%s'\n", file); } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_KeyAlias()
{
  char name[IOBUFLEN];
  char alias[IOBUFLEN];

  if (current) {
    if (UF_NoArgs()) { return ShowKey(current->keytable, NULL); }
    else if (UF_ItemRead(name, IOBUFLEN, "keysequence to alias")) {
      if (UF_NoArgs()) { return ShowKey(current->keytable, name); }
      else if (UF_ItemRead(alias, IOBUFLEN, "alias")) {
	if (alias[strlen(alias)-1] != '\n') {
	  strncat(alias, "\n", IOBUFLEN); }
	if (!KTInsertKeySequence(current->keytable, name, alias)) {
	  ErrPrintR("fatal: can not insert keysequence '%s'\n", name);
	  return FALSE; }
	else { return TRUE; } } } }
  else { ErrPrintR("fatal: use of keytable disabled\n"); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_KeyUnalias()
{
  char name[IOBUFLEN];

  if (current) {
    if (UF_ItemRead(name, IOBUFLEN, "keysequence to unalias")) {
      if (!KTDeleteKeySequence(current->keytable, name)) {
	UF_PrintE("warning: can not unalias keysequence '%s'\n", name);
	return FALSE; }
      else { return TRUE; } } }
  else { UF_PrintE("error: use of keytable disabled\n"); }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int KT_List()
{
  if (current) { ShowKey(current->keytable, NULL); return TRUE; }
  else { UF_PrintE("error: use of keytable disabled\n"); return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int ShowKey(kt, name)
     Keytable_t kt;
     char *name;
{
  Key_t k;
  char line[IOBUFLEN];
  int printed= FALSE;

  if (!name && KeytableFirst(kt)) {
    UF_PrintM("keytable '%s'(%s):\n", KeytableTitle(kt), KeytableFile(kt)); }

  for (k= KeytableFirst(kt); k; k= KeyNext(k)) {
    if (!name || !strcmp(KeySequence(k), name)) {
      if (!KTSprintKey(line, IOBUFLEN, KeySequence(k), KeyAlias(k))) {
	ErrPrintR("fatal internal: can not print key alias of '%s'\n",
		  KeySequence(k)); }
      else { UF_PrintM("%s\n", line); }
      printed= TRUE; } }

  if (!printed) {
    if (name) {
      UF_PrintE("warning: no alias defined for '%s' in keytable '%s'\n",
		name, KeytableTitle(kt)); }
    else {
      UF_PrintE("warning: keytable '%s' empty\n", KeytableTitle(kt)); }
    return FALSE; }
  else { return TRUE; }
}
/* ---------------------------------------------------------------------- */
static VdKey GetKeytable(title)
     char *title;
{
  VdKey vkt;

  ForAllKeytables(vkt) {
    if (!strcmp(title, KeytableTitle(vkt->keytable))) { return vkt; } }
  return NULL;
}
/* ---------------------------------------------------------------------- */
static int InsertKeytable(kt)
     Keytable_t kt;
{
  VdKey new;

  if ((new= (VdKey)Malloc(sizeof(VdKeyRec)))) {
    new->next= NULL;
    new->keytable= kt;
    if (root) {
      VdKey vkt;

      ForAllKeytables(vkt) {
	if (!vkt->next) { vkt->next= new; break; } } }
    else { root= new; }
    UF_PrintM("keytable '%s'(%s) created\n",
	      KeytableTitle(kt), PthShort(KeytableFile(kt)));
    UseKeytable(new);
    return TRUE; }
  else {
    ErrPrintR("fatal: can not create keytable list entry for '%s'\n",
	      KeytableTitle(kt));
    KTDestroyKeytable(kt);
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int ReplaceKeytable(vkt, kt)
     VdKey vkt;
     Keytable_t kt;
{
  KTDestroyKeytable(vkt->keytable);
  vkt->keytable= kt;
  UF_PrintM("keytable '%s' replaced\n", KeytableTitle(kt));
  UseKeytable(vkt);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static void UseKeytable(vkt)
     VdKey vkt;
{
  if (vkt) {
    UF_PrintM("using keytable '%s'\n", KeytableTitle(vkt->keytable)); }
  else { UF_PrintM("use of keytable disabled\n"); }
  current= vkt;
  { extern void XUFSetKeytable(); XUFSetKeytable(vkt->keytable); }
}
/* ---------------------------------------------------------------------- */
