/* $Header: /usr/files/home/Gamma/vs/tv/edit/clients/tv/RCS/VsCal.c,v 1.2 1994/05/26 19:51:46 vs Exp vs $ */ 

/* ============================ */
/* vs calls to the vsCal-module */
/* ============================ */

#define _VsCal_c
				/* vs library */
#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsPath.h>
				/* vs private */
#include "VsActiv.h"
#include "VsError.h"
#include "VsIndex.h"
#include "VsPath.h"
#include "VsUserFace.h"

#include "VdNotify.h"

#include "VsCal.h"
#include "VsCalP.h"
/* ---------------------------------------------------------------------- */
static void Cal_Changed(idx, sid)
     int idx;
     SerialId_t sid;
{
  Vcal_t *vp= VcalsP(&vCal) + idx;
  if (VcalsTyp(&vCal) == CAL_POS || VcalsTyp(&vCal) == CAL_SET) {
    DN_ChangedCal(idx); }
  if (sid) { SerialIdCp(VcalSerialId(vp), sid); }
  else { SerialIdGet(VcalSerialId(vp)); }
}
/* ---------------------------------------------------------------------- */

/* ========================== */
/* vs initialize calibrations */
/* ========================== */

int Cal_Init(num)
     int num;
{
  if (NULL == (VcalsP(&vCal)= (Vcal_t *)Malloc(num * sizeof(Vcal_t)))) {
    ErrPrintR("fatal: can not initialize calibrations\n");
    return FALSE; }
  else {
    int i;
    VcalsNum(&vCal)= num;
    for (i= 0; i< num; i++) { Cal_VcalInit(VcalsP(&vCal)+i); }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
static int Cal_VcalInit(cp)
     Vcal_t *cp;
{
  SerialIdInit(VcalSerialId(cp));
  CalInit(VcalCal(cp));
  return TRUE;
}
/* ---------------------------------------------------------------------- */
/* ==================== */
/* vs reset calibration */
/* ==================== */

int Cal_Reset(i)
     int i;
{
  if (i < 0 || i >= VcalsNum(&vCal)) { return FALSE; }
  else {
    CalDelete(CAL_SET, VcalsCal(&vCal,i));
    Cal_Changed(i, 0);
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
/* =================== */
/* vs calibration read */
/* =================== */

static int Cal_Read(typ, name)
     int typ;
     char *name;
{
  /*  printf("Cal_Read\n");*/
  if (PTH_TUNDEF ==
      Pth_TokenRead(PTH_TUNDEF, PTH_CAL, "%s input file | wc [%s]",
		    name, PTH_CAL)) {
    UF_Reset();
    return FALSE;
  }
  else if (UF_NoArgs()) {
    Idx_SetAll();		/* default if no further input */
  }
  else if (!Idx_Read(1, "read %s for index(s)", name)) {
    return FALSE;
  }

  VcalsTyp(&vCal)= typ;
  VcalsName(&vCal)= name;

  return Pth_LoopIndices(Cal_ReadExec, Cal_ReadFinish);
}
/* ---------------------------------------------------------------------- */
static int Cal_ReadExec(idx, num)
     int idx;
     int num;
{
  /*  printf("Cal_ReadExec\n");*/
  switch(VpthTyp(&vPth)) {
  case PTH_TPATTERN:		/* each file must be read */
    /*    printf("Typ PTH_TPATTERN: %d\n", VpthTyp(&vPth));*/
    if (CalRead(VcalsTyp(&vCal), VcalsCal(&vCal,idx), VpthPath(&vPth))) {
      Cal_Changed(idx, 0);
      UF_PrintM("%s read from '%s' for #%d\n",
		VcalsName(&vCal), PthShort(VpthPath(&vPth)), idx);
      return IDX_EDONE;
    }
    else {			/* read error not fatal */
      Cal_Changed(idx, 0);
      ErrPrint("error: can not read %s from '%s' for #%d\n",
	       VcalsName(&vCal), PthShort(VpthPath(&vPth)), idx);
      return IDX_ESKIPPED;
    } 
  case PTH_TPATH:
    /*    printf("Typ PTH_TPATH: %d\n", VpthTyp(&vPth));*/
    if (!num) {			/* first calibration must be read */
      if (CalRead(VcalsTyp(&vCal), VcalsCal(&vCal,idx), VpthPath(&vPth))) {
	Cal_Changed(idx, 0);
	UF_PrintM("%s read from '%s' for #%d\n",
		  VcalsName(&vCal), PthShort(VpthPath(&vPth)), idx);
	return IDX_EDONE;
      }
      else {
	Cal_Changed(idx, 0);
	ErrPrint("error: can not read %s from '%s'\n",
		 VcalsName(&vCal), PthShort(VpthPath(&vPth)));
	return IDX_EERROR;
      }
    }
    else {
      return Cal_CopyExec(VidxIdx(&vIdx,0), idx, num-1);
    }
  default:
    ErrPrintR("fatal internal: illegal path type\n");
    return IDX_EERROR; }
}
/* ---------------------------------------------------------------------- */
static int Cal_ReadFinish(num)
     int num;
{
  /*  printf("Cal_ReadFinish\n");*/
  if (num) {
    return TRUE;
  }
  else {
    UF_PrintE("warning: no %s read for '%s'\n",
	      VcalsName(&vCal), VpthToken(&vPth));
    return FALSE;
  }
}
/* ---------------------------------------------------------------------- */
int Cal_ReadP() { return Cal_Read(CAL_POS, "position calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadW() { return Cal_Read(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadTl() { return Cal_Read(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadTr() { return Cal_Read(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadVc() { return Cal_Read(CAL_VC, "v/c calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadEfc() { return Cal_Read(CAL_EFC, "efficiency calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ReadSet() { return Cal_Read(CAL_SET, "calibration set"); }
/* ---------------------------------------------------------------------- */
/* ==================== */
/* vs calibration write */
/* ==================== */

static int Cal_Write(typ, name)
     int typ;
     char *name;
{
  if (PTH_TUNDEF ==		/* read vPth.token */
      Pth_TokenRead(PTH_TUNDEF, PTH_CAL,
		    "%s output file | wc [%s]", name, PTH_CAL)) {
    UF_Reset();
    return FALSE; }
  else if ((UF_NoArgs()) &&
	   Idx_SetActSpc()) { ; } /* default if no further input */
  else if (!Idx_Read(1, "write %s for index(s)", name)) {
    return FALSE; }

  VcalsTyp(&vCal)= typ;
  VcalsName(&vCal)= name;

  return Pth_LoopIndices(Cal_WriteExec, Cal_WriteFinish);
}
/* ---------------------------------------------------------------------- */
static int Cal_WriteExec(idx, num)
     int idx;
     int num;
{
  if (CalWrite(VcalsTyp(&vCal), VcalsCal(&vCal,idx), VpthPath(&vPth))) {
    UF_PrintM("%s written to '%s' for #%d\n",
	      VcalsName(&vCal), PthShort(VpthPath(&vPth)), idx);
    return IDX_EDONE; }
  else {
    ErrPrint("error: can not write %s to '%s' for #%d\n",
	     VcalsName(&vCal), PthShort(VpthPath(&vPth)), idx);
    return IDX_ESKIPPED; }
}
/* ---------------------------------------------------------------------- */
static int Cal_WriteFinish(num)
     int num;
{
  if (num == 0) {
    UF_PrintE("warning: no %s written for '%s'\n",
	      VcalsName(&vCal), VpthToken(&vPth));
    UF_Reset();
    return FALSE; }
  else { return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Cal_WriteP()		/* writes startchannel and unit too */
{
  return Cal_Write(CAL_POS, "position calibration");
}
/* ---------------------------------------------------------------------- */
int Cal_WriteW() { return Cal_Write(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_WriteTl() { return Cal_Write(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_WriteTr() { return Cal_Write(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_WriteVc() { return Cal_Write(CAL_VC, "vc calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_WriteEfc() { return Cal_Write(CAL_EFC, "efficiency calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_WriteSet() { return Cal_Write(CAL_SET, "calibration set"); }
/* ---------------------------------------------------------------------- */
/* ==================== */
/* vs calibration enter */
/* ==================== */

static int Cal_Enter(typ, name)
     int typ;
     char *name;
{
  CalRec c;
  double x[2], y[2];
  int stat= FALSE;

  CalInit(&c);

  switch (typ) {
  case CAL_POS:
    stat= (Cal_EnterPairs("channel", x, "position", y, 2) &&
	   Cal_CalcLinCoef(&c.position, x, y));
    break;
  case CAL_WIDTH:
    stat= (Cal_EnterPairs("channel", x, "width", y, 2) &&
	   Cal_CalcLinCoef(&c.width, x, y));
    break;
  case CAL_TL:
    stat= (Cal_EnterPairs("channel", x, "tailleft", y, 2) &&
	   Cal_CalcLinCoef(&c.tl, x, y));
    break;
  case CAL_TR:
    stat= (Cal_EnterPairs("channel", x, "tailright", y, 2) &&
	   Cal_CalcLinCoef(&c.tr, x, y));
    break;
  case CAL_VC:
    { double d;
      double p;
      if ((stat= UF_RealRead(FALSE, &d, &p, "v/c-factor"))) {
	if (stat && d <= 0.0) {
	  ErrPrintR("fatal: v/c-factor value %f is illegal\n", d);
	  stat= FALSE; } }
      c.vc.index= 0; c.vc.cons.c= d; }
    break;
  case CAL_START:
    { int i;
      stat= UF_IntRead(FALSE, &i, "startchannel");
      c.start.index= 0; c.start.cons.c= i; }
    break;
  case CAL_UNIT:
    { char s[CAL_STR];
      stat= UF_ItemRead(s, CAL_STR, "unit");
      c.unit.index= 0;
      if (stat) { strcpy(c.unit.str, s); } } }
  if (!stat) {
    UF_Reset();
    return FALSE; }
  else {
    int idx;

    if (UF_NoArgs()) {
      Idx_SetAll(); }		/* default if no further input */
    else if (!Idx_Read(1, "enter %s for index(s)", name)) {
      return FALSE; }

    VcalsTyp(&vCal)= typ;
    VcalsName(&vCal)= name;

    idx= VidxIdx(&vIdx,0);

    if (CalCopy(typ, VcalsCal(&vCal,idx), &c)) {
      int ret_val;
      DN_Delay(TRUE);
      Cal_Changed(idx, 0);
      UF_PrintM("%s entered for #%d\n", name, idx);
      VidxSingle(&vIdx)= idx;
      ret_val= Idx_Loop2(Cal_CopyExec, Cal_CopyFinish);
      DN_Delay(FALSE);
      return ret_val; }
    else {
      Cal_Changed(idx, 0);
      ErrPrintR("fatal: can not copy %s to #%d\n", name, idx);
      return FALSE; } }
}
/* ---------------------------------------------------------------------- */
static int Cal_EnterPairs(xname, x, yname, y, num)
     char *xname;
     double *x;
     char *yname;
     double *y;
     int num;
{
  int i;
  double p;

  for (i= 0; i< num; i++) {
    if (!UF_RealRead(FALSE, (x + i), &p, "%s[%d]", xname, i) ||
	!UF_RealRead(FALSE, (y + i), &p, "%s[%d]", yname, i)) {
      UF_Reset();
      return FALSE; } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cal_CalcLinCoef(cp, x, y)
     CoefCal_t *cp;
     double *x;
     double *y;
{
  if (x[0] == x[1]) {
    ErrPrintR("fatal: pairs with equal x-values for coef. calculation");
    return FALSE; }
  cp->coef.p[1]= (y[1] - y[0]) / (x[1] - x[0]);
  cp->coef.p[0]= y[0] - cp->coef.p[1] * x[0];
  cp->coef.num = 2;
  cp->index= 0;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cal_EnterP() { return Cal_Enter(CAL_POS, "position calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_EnterW() { return Cal_Enter(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_EnterTl() { return Cal_Enter(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_EnterTr() { return Cal_Enter(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_EnterVc() { return Cal_Enter(CAL_VC, "v/c calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_EnterStart()
{
  return Cal_Enter(CAL_START, "calibration startchannel");
}
/* ---------------------------------------------------------------------- */
int Cal_EnterUnit() { return Cal_Enter(CAL_UNIT, "calibration unit"); }
/* ---------------------------------------------------------------------- */
/* =================== */
/* vs calibration list */
/* =================== */

static int Cal_List(typ, name)
     int typ;
     char *name;
{
  if (!Idx_Read(1, "list %s for index(s)", name)) {
    return FALSE; }
  else {
    int i;
    int calNum= 0;
    for (i= 0; i< VidxNum(&vIdx); i++) {
      int idx= VidxIdx(&vIdx,i);
      if (CalCheck(typ, VcalsCal(&vCal,idx))) {
	calNum++;
	Cal_Print(typ, idx); } }
    if (!calNum) {
      UF_PrintM("info: no %s found to list\n", name);
      return FALSE; }
    else { return TRUE; } }
}
/* ---------------------------------------------------------------------- */
#define PRINT_COEF(cp,memb) {\
  int i;\
  Coef_t *c= &cp->memb.coef;\
    UF_PrintP("[%d]", (int)cp->start.cons.c);\
  for (i= 0; i< c->num; i++) {\
    UF_PrintP("%10g ", c->p[i]); }\
  UF_PrintP("\n"); }

static int Cal_Print(typ, idx)
     int typ;
     int idx;
{
  Cal_t cp= VcalsCal(&vCal,idx);

  UF_PrintP("calibration(s) for #%d\n", idx);
  if (typ == CAL_POS || (typ == CAL_SET && CalCheck(CAL_POS, cp))) {
    UF_PrintP("position [%s]: ", cp->unit.str);
    PRINT_COEF(cp, position); }
  if (typ == CAL_WIDTH || (typ == CAL_SET && CalCheck(CAL_WIDTH, cp))) {
    UF_PrintP("width: ");
    PRINT_COEF(cp, width); }
  if (typ == CAL_TL || (typ == CAL_SET && CalCheck(CAL_TL, cp))) {
    UF_PrintP("tailleft: ");
    PRINT_COEF(cp, tl); }
  if (typ == CAL_TR || (typ == CAL_SET && CalCheck(CAL_TR, cp))) {
    UF_PrintP("tailright: ");
    PRINT_COEF(cp, tr); }
#if 0
  if (typ == CAL_VC || (typ == CAL_SET && CalCheck(CAL_VC, cp))) {
    UF_PrintP("v/c-factor: %10g\n", cp->start.cons.c); }
#endif				/* 0 */
  if (typ == CAL_EFC || (typ == CAL_SET && CalCheck(CAL_EFC, cp))) {
    UF_PrintP("efficiency calibration with %d points\n",
	      cp->efc.efcp.num); }
  return TRUE;
}

#undef PRINT_COEF
/* ---------------------------------------------------------------------- */
int Cal_ListP() { return Cal_List(CAL_POS, "position calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListW() { return Cal_List(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListTl() { return Cal_List(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListTr() { return Cal_List(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListVc() { return Cal_List(CAL_VC, "v/c calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListEfc() { return Cal_List(CAL_EFC, "efficiency calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListStart() { return Cal_List(CAL_START, "startchannel calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_ListSet() { return Cal_List(CAL_SET, "calibration set"); }
/* ---------------------------------------------------------------------- */
int Cal_ListUnit() { return Cal_List(CAL_UNIT, "unit calibration"); }
/* ---------------------------------------------------------------------- */
/* ===================== */
/* vs calibration delete */
/* ===================== */

static int Cal_Delete(typ, name)
     int typ;
     char *name;
{
  if (!Idx_Read(1, "delete %s for index(s)", name)) {
    return FALSE; }
  else {
    VcalsTyp(&vCal)= typ;
    VcalsName(&vCal)= name;
    return Idx_Loop(Cal_DeleteExec, Cal_DeleteFinish); }
}
/* ---------------------------------------------------------------------- */
static int Cal_DeleteExec(idx, num)
     int idx;
     int num;
{
  if (CalCheck(VcalsTyp(&vCal), VcalsCal(&vCal,idx)) &&
      CalDelete(VcalsTyp(&vCal), VcalsCal(&vCal,idx))) {
    Cal_Changed(idx, 0);
    UF_PrintM("%s deleted for #%d\n", VcalsName(&vCal), idx);
    return IDX_EDONE; }
  return IDX_ESKIPPED;
}
/* ---------------------------------------------------------------------- */
static int Cal_DeleteFinish(num)
     int num;
{
  if (num) { return TRUE; } 
  else {
    UF_PrintE("warning: no %s found to delete\n", VcalsName(&vCal));
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Cal_DeleteP() { return Cal_Delete(CAL_POS, "position calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteW() { return Cal_Delete(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteTl() { return Cal_Delete(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteTr() { return Cal_Delete(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteVc() { return Cal_Delete(CAL_VC, "v/c calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteEfc() { return Cal_Delete(CAL_EFC, "efficiency calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteStart()
{
  return Cal_Delete(CAL_START, "calibration startchannel");
}
/* ---------------------------------------------------------------------- */
int Cal_DeleteSet() { return Cal_Delete(CAL_SET, "calibration set"); }
/* ---------------------------------------------------------------------- */
int Cal_DeleteUnit() { return Cal_Delete(CAL_UNIT, "calibration unit"); }
/* ---------------------------------------------------------------------- */
/* =================== */
/* vs calibration copy */
/* =================== */

static int Cal_Copy(typ, name)
     int typ;
     char *name;
{
  if (!Idx_ReadOne("%s copy source index", name) ||
      !Idx_Read(1, "%s copy destination index(s) for #%d",
		name, VidxSingle(&vIdx))) {
    return FALSE; }
  else {
    VcalsTyp(&vCal)= typ;
    VcalsName(&vCal)= name;
    return Idx_Loop2(Cal_CopyExec, Cal_CopyFinish); }
}
/* ---------------------------------------------------------------------- */
static int Cal_CopyExec(idx0, idx, num)
     int idx0;
     int idx;
     int num;
{
  if (!CalCheck(VcalsTyp(&vCal), VcalsCal(&vCal,idx0))) {
    UF_PrintE("error: %s #%d not defined\n", VcalsName(&vCal), idx0);
    return IDX_EERROR; }
  else if (CalCopy(VcalsTyp(&vCal),
		   VcalsCal(&vCal,idx), VcalsCal(&vCal,idx0))) {
    UF_PrintM("%s #%d copied to #%d\n", VcalsName(&vCal), idx0, idx);
    Cal_Changed(idx, VcalsSerialId(&vCal,idx0));
    return IDX_EDONE; }
  else {
    Cal_Changed(idx, 0);
    ErrPrintR("fatal: can not copy %s #%d to #%d\n",
	      VcalsName(&vCal), idx0, idx);
    return IDX_EERROR; }
}
/* ---------------------------------------------------------------------- */
static int Cal_CopyFinish(num)
     int num;
{
  if (num) { return TRUE; }
  else {
    UF_PrintE("warning: no %s copied\n", VcalsName(&vCal)); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Cal_CopyP() { return Cal_Copy(CAL_POS, "position calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyW() { return Cal_Copy(CAL_WIDTH, "width calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyTl() { return Cal_Copy(CAL_TL, "tailleft calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyTr() { return Cal_Copy(CAL_TR, "tailright calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyVc() { return Cal_Copy(CAL_VC, "v/c calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyEfc() { return Cal_Copy(CAL_EFC, "efficiency calibration"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyStart() { return Cal_Copy(CAL_START, "calibration startchannel"); }
/* ---------------------------------------------------------------------- */
int Cal_CopySet() { return Cal_Copy(CAL_SET, "calibration set"); }
/* ---------------------------------------------------------------------- */
int Cal_CopyUnit() { return Cal_Copy(CAL_UNIT, "calibration unit"); }
/* ---------------------------------------------------------------------- */
