/* $Header: /usr/files/home/Gamma/vs/tv/edit/clients/tv/RCS/VsCut.c,v 1.4 1994/12/02 21:24:04 vs Exp vs $ */

# define _VsCut_c
				/* library include files*/
#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCurFit.h>
#include <tv/vsMark.h>
#include <tv/vsPath.h>
#include <tv/vsTools.h>

#include <errno.h>
#include <sys/stat.h>

				/* private include files */
#include "VsActiv.h"
#include "VsCal.h"
#include "VsError.h"
#include "VsFit.h"
#include "VsFitIo.h"
#include "VsIndex.h"
#include "VsMark.h"
#include "VsPath.h"
#include "VsSpectra.h"
#include "VsUserFace.h"

#include "VdNotify.h"

#include "VsCut.h"
#include "VsCutP.h"
/* ---------------------------------------------------------------------- */

#define CUT_FIRST		1
#define CUT_LAST		2
#define CUT_NEXT		3
#define CUT_PREVIOUS		4

#define WEIGHT(w)	((w==CUT_WGAT) ? "gatewidth" : ((w==CUT_WFIT) ? "fit" : "integral"))

/* ---------------------------------------------------------------------- */
static void Cut_Changed(idx)
     int idx;
{
  DN_ChangedCut(idx);
}
/* ---------------------------------------------------------------------- */
/* ===================== */
/* initialize cut module */
/* ===================== */

int Cut_Init(num)
     int num;
{
  if ((VcutsC(&vCut)= (Vcut_t *)Malloc(num * sizeof(Vcut_t))) &&
      (VcutsM(&vCut)= (Vmtx_t *)Malloc(num * sizeof(Vmtx_t))) &&
      (VcutsP(&vCut)= (VcutPath_t *)Malloc(num * sizeof(VcutPath_t)))) {
    { int i; for (i= 0; i< num; i++) {
      Cut_InitVcut(VcutsC(&vCut)+i);
      Cut_InitVmtx(VcutsM(&vCut)+i);
      Cut_InitVcutPath(VcutsP(&vCut)+i); } }

    *VcutsMtxFmt(&vCut)= '\0';
    *VcutsCutFmt(&vCut)= '\0';
    VcutsWeight(&vCut)= CUT_WGAT;
    VcutsIgnore(&vCut)= 0.001;
    VcutsScale(&vCut)= ACT_SDEFAULT;
    VecInit(VcutsGate(&vCut), sizeof(Pair_t));
    VecInit(VcutsBgGate(&vCut), sizeof(Pair_t));


    VcutsNum(&vCut)= num;

    Cut_InitBuffer(num);

    return TRUE; }
  else {
    ErrPrintR("fatal: can not initialize cut-module\n");
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_InitBuffer(num)
     int num;
{
				/* projections */

  { int i; for (i= 0; i< num; i++) { /* last buffer reserved for projection */
    VmtxProIdx(VcutsM(&vCut)+i)= num-1; } }

				/* cuts */

  { /* buffer:		S0  S1  S2  ... S(n-1)  */
    /* proj.:				p       */
    /* cuts:		C00 C10 C20 ... C(n-1)0 */
    int c;
    for (c= 0; c< num; c++) {
      Vcut_t *vp= VcutsC(&vCut)+c;
      Cut_t cp= VcutCut(vp);
      VcutSpcIdx(vp)= c; CutSpc(cp)= VspcsSpc(&vSpc,c); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_InitVcut(cp)
     Vcut_t *cp;
{
  CutInit(VcutCut(cp));
  SerialIdInit(VcutSpcSerialId(cp));
  VcutSpcExists(cp)= FALSE;
  SerialIdInit(VcutGatSerialId(cp));
  SerialIdInit(VcutBgGatSerialId(cp));

  VcutMtxIdx(cp)= 0;
  VcutMtx(cp)= VcutsM(&vCut);
  VcutCutPathIdx(cp)= 0;
  VcutCutPath(cp)= VcutsP(&vCut);
    
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ResetVcut(cutIdx)
     int cutIdx;
{
  int ret_val;
  DN_Delay(TRUE);
  ret_val= (Cut_ResetSpectrum(cutIdx) && Cut_ResetGates(cutIdx));
  DN_Delay(FALSE);
  return ret_val;
}
/* ---------------------------------------------------------------------- */
static int Cut_ResetGates(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);
  int changed= FALSE;

  if (VecNum(CutGate(cp))) {
    VecReset(CutGate(cp));
    SerialIdGet(VcutGatSerialId(vcp));
    changed= TRUE; }
  if (VecNum(CutBgGate(cp))) {
    VecReset(CutBgGate(cp));
    SerialIdGet(VcutGatSerialId(vcp));
    changed= TRUE; }
  if (changed) { Cut_Changed(cutIdx); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ResetSpectrum(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);
  int spcIdx= VcutSpcIdx(vcp);

  if (CutSpc(cp) && !Spc_Reset(spcIdx)) {
    ErrPrintR("fatal: can not delete cut spectrum #%d of cut #%d\n",
	      spcIdx, cutIdx);
    return FALSE; }
  VcutSpcExists(vcp)= FALSE;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_InitVcutPath(pp)
     VcutPath_t *pp;
{
  VcutPathCdp(pp)= NULL;
  *VcutPathCutFmt(pp)= '\0';
  VcutPathFitExpr(pp)= NULL;
  VcutPathSpcExpr(pp)= NULL;
  SerialIdInit(VcutPathSerialId(pp));
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_InitVmtx(mp)
     Vmtx_t *mp;
{
  VmtxMtxExists(mp)= FALSE;
  SerialIdInit(VmtxMtxSerialId(mp));
  VmtxProIdx(mp)= 0;
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_CreateCut()
{
  int cutIdx= VactCutIdx(vAct);

  return (Cut_ResetSpectrum(cutIdx) &&
	  Cut_CreateHeadP(cutIdx) &&
	  Cut_CreateSpectrumP(cutIdx));
}
/* ---------------------------------------------------------------------- */
int Cut_CreateHead()
{
  int cutIdx= VactCutIdx(vAct);

  if (!Cut_ResetVcut(cutIdx)) { return FALSE; }
  else { return Cut_CreateHeadP(cutIdx); }
}
/* ---------------------------------------------------------------------- */
static int Cut_CreateHeadP(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);

  DN_Delay(TRUE);
  if (!Cut_ResetVcut(cutIdx)) { DN_Delay(FALSE); return FALSE; }
  else if (!Cut_MrkGet(cutIdx)) {
    ErrPrintR("fatal: can not get marker for cut #%d\n", cutIdx); }
  else if (Cut_CreateFac(cutIdx)) {
    char entry[MAXPATHLEN];
    CDIR *cdp= VcutPathCdp(VcutsP(&vCut)+ VcutCutPathIdx(vcp));
    double center= (GVecX1(CutGate(cp),0) + GVecX2(CutGate(cp),0)) * 0.5;

    CutIdx1(cp)= NINT(CalP(VcalsCal(&vCal,VcutProIdx(vcp)), center));
    CutIdx2(cp)= -1;
    CutCDirEntry(cdp, CutIdx1(cp), CutIdx2(cp), entry);
    if ((CutFile(cp)= Strdup(entry))) {
      CutExists(cp)= TRUE;
      DN_Delay(FALSE);
      return TRUE; } }
  Cut_ResetVcut(cutIdx);
  DN_Delay(FALSE);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int Cut_CreateFac(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);

  switch (VcutsWeight(&vCut)) {
  case CUT_WFIT:
    if (VecNum(CutGate(cp)) != 1) {
      UF_PrintE("warning: fit weight only for single gate possible\n"); }
    else if (!VspcsExists(&vSpc,VcutProIdx(vcp))) {
      UF_PrintE("warning: no projection available for fit weight\n"); }
    else if (!VecNum(CutBgGate(cp))) {
      UF_PrintE("warning: no background gate(s) available for fit weight\n"); }
    else  { return Cut_CreateFacFit(cutIdx); }
  case CUT_WGAT: return Cut_CreateFacGate(cutIdx);
  default:
    ErrPrintR("fatal internal: illegal weight type for cut\n"); return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_CreateFacFit(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);
  int proIdx= VcutProIdx(vcp);
  Vfio_t *vp;

  if (!(vp= Fio_Reopen(proIdx, "r"))) {
    ErrPrintR("fatal: no fit data available\n");
    return FALSE; }
  else {
    Fio_t fio= VfioFio(vp);
    GVec_t gt= CutGate(cp);
    GVec_t bgGt= CutBgGate(cp);
    double x1= GVecX1(gt,0);
    double x2= GVecX2(gt,0);
    double *matrix= NULL;
    double *bgInt= NULL;
    double *tmpInt= NULL;
    int *cont= NULL;
    int lastIdx= VecNum(bgGt);
    int matDim= (lastIdx + 1);
    double gatVol= 0.0;
    if (!(matrix= (double *)Calloc(matDim * matDim, sizeof(double))) ||
	!(bgInt= (double *)Calloc(matDim, sizeof(double))) ||
	!(tmpInt= (double *)Calloc(matDim, sizeof(double))) ||
	!(cont= (int *)Calloc(matDim, sizeof(int)))) {
      Free(matrix); Free(bgInt); Free(tmpInt);
      ErrPrintR("fatal: can not create temporary arrays for fit weight\n");
      return FALSE; }

    x1= MIN(x1, GVecX1(bgGt,0)); x2= MAX(x2, GVecX2(bgGt,VecNum(bgGt)-1));

    Cut_IntegrateBg(cp, proIdx, bgInt);

    { FioCurPeak_t cur= NULL;
      int gtIdx= 0;

      while ((cur= FioNextPeak(fio, x1, x2, cur))) {
	double vol;
	switch (FioCurPeakTyp(cur)) {
	case FIO_PBIN:
	  if (!Cut_IntegratePeak(cp, proIdx, FioCurPeakPeaIdx(cur),
				 tmpInt, &gtIdx, &vol)) {
	    Free(matrix); Free(bgInt); Free(tmpInt); Free(cont);
	    return FALSE; }
	  break;
	case FIO_PSEA:
	  if (Cut_IntegrateSearch(cp, proIdx, FioCurPeakSeaPeak(cur),
				  tmpInt, &gtIdx, &vol)) {
	    Free(matrix); Free(bgInt); Free(tmpInt); Free(cont);
	    return FALSE; }
	  break; }
				/* subtract all peak contributions */
	{ int i; for (i= 0; i< matDim; i++) {
	  bgInt[i] -= tmpInt[i]; } }

	if (gtIdx >= 0 && 
	    (!cont[gtIdx] || matrix[gtIdx * matDim + gtIdx] < tmpInt[gtIdx])) {
	  if (gtIdx == lastIdx) { gatVol= vol; }
	  cont[gtIdx]= TRUE;
	  memcpy(matrix + gtIdx * matDim, tmpInt,
		 matDim * sizeof(double)); } } }

    if (!cont[lastIdx]) {
      cont[lastIdx]= TRUE;
      matrix[lastIdx * matDim + lastIdx]= 1.0;
      gatVol= 1.0; }

    { int bgIdx, dim= 1;

      for (bgIdx= 0; bgIdx< lastIdx && cont[bgIdx]; bgIdx++) { ; }
      { int i; for (i= 0; i< matDim; i++) { if (cont[i]) { dim++; } } }

      if (bgIdx >= lastIdx) {
	ErrPrintR("fatal: no background gate defined\n");
	Free(matrix); Free(bgInt); Free(tmpInt); Free(cont);
	return FALSE; }
      else {
	memcpy(matrix + bgIdx * matDim, bgInt,
	       matDim * sizeof(double));

	{ int i;
	  double *m= matrix, *mc= matrix;
	  for (i= 0; i< matDim; i++) {
	    if (cont[i] || i == bgIdx) {
	      double *b;
	      int j;
	      for (j= 0; j< matDim; j++) {
		if (j == bgIdx) { b= mc; *mc++ = *m++; }
		else if (cont[j] ) { *mc++= *m++; }
		else { *b += *m++; } } }
	    else { m += matDim; } } } }

      memset(tmpInt, 0, dim * sizeof(double));
      tmpInt[lastIdx]= gatVol; /* full peak intensity */

      if (!CurGaussJordan(matrix, dim, dim, tmpInt, 1, 1)) {
	ErrPrintR("fatal: can not determine weights\n");
	Free(matrix); Free(bgInt); Free(tmpInt); Free(cont);
	return FALSE; }
      else {
	int i;
	Gate_t *bgGat= VecP(CutBgGate(cp));
	double *t= tmpInt;
	for (i= 0; i< lastIdx; i++) {
	  if (cont[i] || i == bgIdx) { GateFac(bgGat++) = *t++; }
	  else { GateFac(bgGat++) = tmpInt[bgIdx]; } }
	GVecFac(CutGate(cp),0)= tmpInt[lastIdx];
	UF_PrintM("fit weighted cut #%d created\n", cutIdx);
	Free(matrix);
	Free(bgInt);
	Free(tmpInt);
	Free(cont);
	return TRUE; } } }
}
/* ---------------------------------------------------------------------- */
static int Cut_IntegratePeak(cp, proIdx, peaIdx, tmpInt, idx, vol)
     Cut_t cp;
     int proIdx;
     int peaIdx;
     double *tmpInt;
     int *idx;
     double *vol;
{
  Vfio_t *vfio= VfiosF(&vFio) + proIdx;
  FioPeaVec_t pea= FioPea(VfioFio(vfio));
  int binIdx= FioPeaPBinIdx(pea,peaIdx);
  int peakIdx= FioPeaPPeakIdx(pea,peaIdx);
  Vfit_t *vfit= VfitsTmpFit(&vFit);
  Fit_t fit= VfitFit(vfit);
  FitOpt op;

  if (!Fio_ReadBin(proIdx, binIdx, vfit)) {
    ErrPrintR("fatal: can not read bin %d for #%d from '%s'\n",
	      binIdx, proIdx, PthShort(VfioFile(vfio)));
    return FALSE; }
  else if (!(op= FitPeakFunctionSetup(NULL, fit, FitPar(fit), peakIdx))) {
    ErrPrintR("fatal: can not evaluate peak function\n"); return FALSE; }
  else {
    GVec_t bgGt= CutBgGate(cp);
    double pos= FitPosVal(fit,peakIdx);
    int lastIdx= VecNum(bgGt);

    *idx= -1;

    { int g;
      for (g= 0; g< lastIdx; g++) {
	double sum= 0.0;
	int chn;
	int last= NGVecX2(bgGt,g);
	if (GVecX1(bgGt,g) <= pos && pos <= GVecX2(bgGt,g)) { *idx= g; }
	for (chn= NGVecX1(bgGt,g); chn<= last; chn++) {
	  sum += FitPeakFunction(op, (double)chn); }
	tmpInt[g]= sum; } }
    { double sum= 0.0;
      int chn;
      GVec_t gt= CutGate(cp);
      int last= NGVecX2(gt,0);
      if (GVecX1(gt,0) <= pos && pos <= GVecX2(gt,0)) { *idx= lastIdx; }
      for (chn= NGVecX1(gt,0); chn<= last; chn++) {
	sum += FitPeakFunction(op, (double)chn); }
      tmpInt[lastIdx]= sum; }
    *vol= FitVolVal(fit,peakIdx);
    FitPeakFunctionFinish(op);
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_IntegrateSearch(cp, proIdx, seaPeak, tmpInt, idx, vol)
     Cut_t cp;
     int proIdx;
     SeaPeak seaPeak;
     double *tmpInt;
     int *idx;
     double *vol;
{
  GVec_t bgGt= CutBgGate(cp);
  double pos= SeaPos(seaPeak);
  int lastIdx= VecNum(bgGt);

  *idx= -1;

  { int g;
    for (g= 0; g< lastIdx; g++) {
      double sum= 0.0;
      int chn;
      int last= NGVecX2(bgGt,g);
      if (GVecX1(bgGt,g) <= pos && pos <= GVecX2(bgGt,g)) { *idx= g; }
      for (chn= NGVecX1(bgGt,g); chn<= last; chn++) {
	sum += SeaPeakFunction(seaPeak, (double)chn); }
      tmpInt[g]= sum; } }
  { double sum= 0.0;
    int chn;
    GVec_t gt= CutGate(cp);
    int last= NGVecX2(gt,0);
    if (GVecX1(gt,0) <= pos && pos <= GVecX2(gt,0)) { *idx= lastIdx; }
    for (chn= NGVecX1(gt,0); chn<= last; chn++) {
      sum += SeaPeakFunction(seaPeak, (double)chn); }
    tmpInt[lastIdx]= sum; }
  *vol= SeaVol(seaPeak);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static void Cut_IntegrateBg(cp, proIdx, bgInt)
     Cut_t cp;
     int proIdx;
     double *bgInt;
{
  Spc_t sp= VspcsSpc(&vSpc,proIdx);
  GVec_t gt= CutGate(cp);
  GVec_t bgGt= CutBgGate(cp);
  int i;

  SpcIntegrate(sp, GVecX1(gt,0), GVecX2(gt,0), bgInt + VecNum(bgGt), NULL);
  for (i= 0; i< VecNum(bgGt); i++) {
    SpcIntegrate(sp, GVecX1(bgGt,i), GVecX2(bgGt,i), bgInt + i, NULL); }
}
/* ---------------------------------------------------------------------- */
static int Cut_CreateFacGate(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);
  { int i;
    GVec_t gat= CutGate(cp);
    for (i= 0; i< VecNum(gat); i++) { GVecFac(gat,i)= 1.0; } }
  if (Cut_CpGatToMrk(VcutsGate(&vCut), CutGate(cp)) &&
      Cut_CpGatToMrk(VcutsBgGate(&vCut), CutBgGate(cp))) {
    int i;
    GVec_t bgGat= CutBgGate(cp);
    double tmp= -(MrkRange(VcutsGate(&vCut)) / MrkRange(VcutsBgGate(&vCut)));
    for (i= 0; i< VecNum(bgGat); i++) { GVecFac(bgGat,i)= tmp; } }
  UF_PrintM("gate width weighted cut #%d created\n", cutIdx);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_CreateSpectrum()
{
  int cutIdx= VactCutIdx(vAct);
  return Cut_CreateSpectrumP(cutIdx);
}
/* ---------------------------------------------------------------------- */
static int Cut_CreateSpectrumP(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
  Cut_t cp= VcutCut(vcp);
  Vmtx_t *mp= VcutMtx(vcp);

  DN_Delay(TRUE);
  if (!CutExists(cp)) {
    ErrPrintR("fatal: cut head not defined for cut #%d\n", cutIdx); }
  else if (!VmtxMtxExists(mp)) {
    ErrPrintR("fatal: matrix file #%d for cut #%d not opened\n",
	      VcutMtxIdx(vcp), cutIdx); }
  else if (!Cut_ResetSpectrum(cutIdx)) { return FALSE; }
  else if (!CutCreateSpectrum(VmtxMtx(mp), cp)) {
    ErrPrintR("fatal: can not create spectrum for cut #%d\n", cutIdx); }
  else {
    VcutPath_t *pp= VcutCutPath(vcp);
    Spc_t sp= CutSpc(cp);
    if (sp && SpcNum(sp)) {
      int idx= VcutSpcIdx(vcp);
      Spc_InitCut(idx,
		  PthSubstitute(CutFile(cp), VcutPathSpcExpr(pp)),
		  VcutPathFitExpr(pp),
		  SPC_OMEMORY);
      SerialIdCp(VcutSpcSerialId(vcp), VspcsSerialId(&vSpc,idx));
      VcutSpcExists(vcp)= TRUE;
      UF_PrintM("spectrum of cut #%d copied to buffer #%d\n",
		cutIdx, VcutSpcIdx(vcp)); }
    else { VcutSpcExists(vcp)= FALSE; }

    { int proIdx= VmtxProIdx(mp);
      if (VspcsExists(&vSpc,proIdx) &&
	  SpcFileLines(VmtxMtx(mp)) != SpcNum(VspcsSpc(&vSpc,proIdx))) {
	UF_PrintE
	  ("warning: resolutions of matrix (%d) and projection (%d) differ\n",
	   SpcFileLines(VmtxMtx(mp)), SpcNum(VspcsSpc(&vSpc,proIdx))); } }
    DN_Delay(FALSE);
    return TRUE; }
  DN_Delay(FALSE);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_CutPathClose()
{
  return (Idx_ReadOne("cut directory index") &&
	  Cut_CutPathCloseP(VidxSingle(&vIdx)));
}
/* ---------------------------------------------------------------------- */
static int Cut_CutPathCloseP(pathIdx)
     int pathIdx;
{
  VcutPath_t *pp= VcutsP(&vCut) + pathIdx;

  if (VcutPathCdp(pp)) {
    UF_PrintM("closing cut directory #%d '%s'\n",
	      pathIdx, PthShort(CDirPath(VcutPathCdp(pp))));

    CutCloseCDir(VcutPathCdp(pp));
    VcutPathCdp(pp)= NULL;
    *VcutPathCutFmt(pp)= '\0';
    Free(VcutPathFitExpr(pp));
    Free(VcutPathSpcExpr(pp));
    SerialIdGet(VcutPathSerialId(pp));
    DN_ChangedCutPath(pathIdx); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_CutPathOpen()
{
  strcpy(VpthFormat(&vPth), VcutsCutFmt(&vCut));
  return (Idx_ReadOne("cut directory index") &&
	  Pth_TokenFormatRead(PTH_TPATH, NULL,
			      "path of cut directory #%d",
			      VidxSingle(&vIdx)) &&
	  Cut_CutPathOpenP(VidxSingle(&vIdx),
			   VpthPath(&vPth), VpthFormat(&vPth)));
}
/* ---------------------------------------------------------------------- */
static int Cut_CutPathOpenP(pathIdx, path, cutFmt)
     int pathIdx;
     char *path;
     char *cutFmt;
{
  if (Cut_CutPathCloseP(pathIdx)) {
    VcutPath_t *pp= VcutsP(&vCut) + pathIdx;

    if ((VcutPathCdp(pp)= CutOpenCDir(path,
				     PatGet(VpthPattern(&vPth), PTH_CUTGAT)))) {
      VcutPathCutIdx(pp)= -1;
      SerialIdGet(VcutPathSerialId(pp));
      strcpy(VcutsCutFmt(&vCut), cutFmt);
      strcpy(VcutPathCutFmt(pp), cutFmt);
      Free(VcutPathFitExpr(pp));
      VcutPathFitExpr(pp)= Strdup(PatGet(VpthPattern(&vPth), PTH_FIT));
      Free(VcutPathSpcExpr(pp));
      VcutPathSpcExpr(pp)= Strdup(PatGet(VpthPattern(&vPth), PTH_CUTSPC));
      UF_PrintM("opening cut directory #%d '%s'\n",
		pathIdx, PthShort(CDirPath(VcutPathCdp(pp))));
      DN_ChangedCutPath(pathIdx);
      return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_CutFormat()
{
  return Pth_FormatRead("cut spectrum", VcutsCutFmt(&vCut));
}
/* ---------------------------------------------------------------------- */
int Cut_MtxClose()
{
  return (Idx_ReadOne("matrix index") &&
	  Cut_MtxCloseP(VidxSingle(&vIdx)));
}
/* ---------------------------------------------------------------------- */
static int Cut_MtxCloseP(mtxIdx)
     int mtxIdx;
{
  Vmtx_t *mp= VcutsM(&vCut) + mtxIdx;

  if (VmtxMtxExists(mp)) {
    UF_PrintM("closing matrix #%d '%s'\n",
	      mtxIdx, PthShort(SpcFileName(VmtxMtx(mp))));
    SpcClose(VmtxMtx(mp));
    SerialIdGet(VmtxMtxSerialId(mp));
    VmtxMtxExists(mp)= FALSE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_MtxFormat()
{
  return Pth_FormatRead("matrix", VcutsMtxFmt(&vCut));
}
/* ---------------------------------------------------------------------- */
int Cut_MtxOpen()
{
  strcpy(VpthFormat(&vPth), VcutsMtxFmt(&vCut));
  return (Idx_ReadOne("matrix index") &&
	  Pth_TokenFormatRead(PTH_TPATH, NULL,
			      "path of matrix #%d",
			      VidxSingle(&vIdx)) &&
	  Cut_MtxOpenP(VidxSingle(&vIdx),
		       VpthPath(&vPth), VpthFormat(&vPth)));
}
/* ---------------------------------------------------------------------- */
static int Cut_MtxOpenP(mtxIdx, path, mtxFmt)
     int mtxIdx;
     char *path;
     char *mtxFmt;
{
  Vmtx_t *mp= VcutsM(&vCut) + mtxIdx;
  if (VmtxMtxExists(mp) && !strcmp(path, SpcFileName(VmtxMtx(mp)))) {
    return TRUE; }
  else if (Cut_MtxCloseP(mtxIdx)) {
    if ((VmtxMtx(mp)= SpcOpen(path, mtxFmt, "r"))) {
      SerialIdGet(VmtxMtxSerialId(mp));
      VmtxMtxExists(mp)= TRUE;
      UF_PrintM("opening matrix #%d %s'%s\n",
		mtxIdx, PthShort(SpcFileName(VmtxMtx(mp))), SpcFileFormat(VmtxMtx(mp)));
      return TRUE; } }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static char *Cut_Substitute(path, pattern)
     char *path;
     char *pattern;
{
  char *expr;
  if (!(expr= PatGet(VpthPattern(&vPth), pattern)) || !*expr) {
    UF_PrintE("error: can not find expression for pattern '%s'\n", pattern); }
  else {
    char *file;
    if (!(file= PthSubstitute(path, expr)) && !*file) {
      UF_PrintE("error: can not decode expression '%s' for '%s'\n",
		expr, path); }
    else { return file; } }
  return NULL;
}
/* ---------------------------------------------------------------------- */
int Cut_EnvOpen()
{
  char path[MAXPATHLEN];
  int cutIdx= VactCutIdx(vAct);
  if (UF_ItemRead(path, MAXPATHLEN,
		  "root directory for environment of cut #%d", cutIdx)) {
    Vcut_t *cp= VcutsC(&vCut) + cutIdx;
    int mtxIdx= VcutMtxIdx(cp);
    char file[MAXPATHLEN];
    if (strcpy(file, Cut_Substitute(path,PTH_MTX))) {
      Cut_MtxOpenP(mtxIdx, file, VcutsMtxFmt(&vCut)); }
    if (strcpy(file, Cut_Substitute(path,PTH_PROSPC))) {
      strcpy(VpthFormat(&vPth), VspcsInFmt(&vSpc));
      Pth_TokenSet(PTH_TPATH, file);
      Spc_ReadExec(VmtxProIdx(VcutsM(&vCut) + mtxIdx), 0); }
    if (strcpy(file, Cut_Substitute(path,PTH_CUTDIR))) {
      Cut_CutPathOpenP(VcutCutPathIdx(cp), file, VcutsCutFmt(&vCut)); }
  }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_CutPathIdx()
{
  int cutIdx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut) + cutIdx;
  if (Idx_ReadOne("cut directory for active cut #%d [%d]",
		  cutIdx, VcutCutPathIdx(cp))) {
    if (VcutCutPathIdx(cp) != VidxSingle(&vIdx)) {
      VcutCutPathIdx(cp)= VidxSingle(&vIdx);
      VcutCutPath(cp)= VcutsP(&vCut) + VidxSingle(&vIdx);
      DN_ChangedCutPathIdx(cutIdx); }
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_MtxIdx()
{
  int cutIdx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut) + cutIdx;
  if (Idx_ReadOne("matrix for active cut #%d [%d]",
		  cutIdx, VcutMtxIdx(cp))) {
    VcutMtxIdx(cp)= VidxSingle(&vIdx);
    VcutMtx(cp)= VcutsM(&vCut) + VidxSingle(&vIdx);
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_ProIdx()
{
  int cutIdx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut) + cutIdx;
  int mtxIdx= VcutMtxIdx(cp);
  Vmtx_t *mp= VcutsM(&vCut) + mtxIdx;
  if (Idx_ReadOne("projection for matrix #%d [%d]",
		  mtxIdx, VmtxProIdx(mp))) {
    VmtxProIdx(mp)= VidxSingle(&vIdx); return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_SpcIdx()
{
  int cutIdx= VactCutIdx(vAct);
/*  int cutIdx= VidxSingle(&vIdx); DUKE: wrong ? */
  Vcut_t *vcp= VcutsC(&vCut)+cutIdx;
  if (!Idx_ReadOne("spectrum for active cut #%d [%d]",
		   cutIdx, VcutSpcIdx(vcp))) {
    return FALSE; }
  else {
    Cut_t cp= VcutCut(vcp);
    VcutSpcIdx(vcp)= VidxSingle(&vIdx);
    CutSpc(cp)= VspcsSpc(&vSpc,VidxSingle(&vIdx));
    return TRUE; }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
static int Cut_EntryIdx(cDirIdx, cutPathIdx, incTyp)
     int *cDirIdx;
     int cutPathIdx;
     int incTyp;
{
  if (!Cut_CheckCutPath(cutPathIdx)) { return FALSE; }
  else {
    VcutPath_t *pp= VcutsP(&vCut)+cutPathIdx;
    CDIR *cdp= VcutPathCdp(pp);

    if (!CDirNum(cdp)) {
      ErrPrintR("fatal: cut directory #%d has no entries\n", cutPathIdx);
      return FALSE; }
    else {
      switch (incTyp) {
      case CUT_FIRST: *cDirIdx= 0; return TRUE;
      case CUT_LAST: *cDirIdx= CDirNum(cdp)-1; return TRUE;
      case CUT_NEXT:
	*cDirIdx= VcutPathCutIdx(pp)+1;
	if (VcutPathCutIdx(pp) < 0) { *cDirIdx= 0; }
	break;
      case CUT_PREVIOUS:
	*cDirIdx= VcutPathCutIdx(pp)-1;
	if (VcutPathCutIdx(pp) >= CDirNum(cdp)) { *cDirIdx= CDirNum(cdp)-1; }
	break; }

      if (0 <= *cDirIdx && *cDirIdx < CDirNum(cdp)) { return TRUE; }
      else {
	ErrPrintR("fatal: no further entry found in cut directory #%d '%s'\n",
		  cutPathIdx, PthShort(CDirPath(cdp)));
	return FALSE; } } }
}
/* ---------------------------------------------------------------------- */
static int Cut_EntryName(cp, cDirIdx)
     Vcut_t *cp;
     int *cDirIdx;
{
  char name[MAXPATHLEN];
  CDIR *cdp= VcutPathCdp(VcutsP(&vCut)+ VcutCutPathIdx(cp));

  if (UF_ItemRead(name, MAXPATHLEN,
		  "cut entry name in %s", PthShort(CDirPath(cdp)))) {
    *cDirIdx= CutNameCDirEntry(cdp, name);
    return (*cDirIdx >= 0); }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_EntryPos(cp, cDirIdx)
     Vcut_t *cp;
     int *cDirIdx;
{
  double pos;
  CDIR *cdp= VcutPathCdp(VcutCutPath(cp));
  Cal_t cal= VcalsCal(&vCal,VcutProIdx(cp));

  if (Mrk_EnterSingleV(VcutsScale(&vCut), cal, &pos, "cut position")) {
    int first;
    double chn= (VactCalibrated(VactScale(vAct)) ? CalC(cal, pos) : pos);

    first= CutPosNextCDirEntry(cdp, -1, chn);
    if (first < 0) {
      ErrPrintR
	("fatal: no cut entry found at %.1f channels in cut directory #%d\n",
	 chn, VcutCutPathIdx(cp));
      return FALSE; }
    else if (CutPosNextCDirEntry(cdp, first, chn) < 0) {
      *cDirIdx= first; return TRUE; }
    else {
      Cut_ListP(VcutCutPathIdx(cp), cal, chn, chn);
      return Cut_EntryName(cp, cDirIdx); } }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Cut_ListRange()
{
  Vcut_t *cp= VcutsC(&vCut)+VactCutIdx(vAct);
  Cal_t cal= VcalsCal(&vCal,VcutProIdx(cp));
  
  /*
   *  if (Mrk_EnterRange(VcutsScale(&vCut), ACT_SUNCALIB, VcutProIdx(cp))) {
   * PVec_t rp= VmrkMrk(VmrksRange(&vMrk));
   *    Cut_ListP(VcutCutPathIdx(cp), cal, PVecX1(rp,0), PVecX1(rp,0));
   */
  double pos;
  if (Mrk_EnterSingleV(VcutsScale(&vCut), cal, &pos, "list-point")) {
    Cut_ListP(VcutCutPathIdx(cp), cal, pos-VmrksDfltRange(&vMrk), pos+VmrksDfltRange(&vMrk));
    return TRUE;
  }
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_ListAll()
{
  Vcut_t *cp= VcutsC(&vCut)+VactCutIdx(vAct);
  Cal_t cal= VcalsCal(&vCal,VcutProIdx(cp));
  Cut_ListP(VcutCutPathIdx(cp), cal, 0.0, -1.0);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ListP(cpIdx, cal, chn1, chn2)
     int cpIdx;
     Cal_t cal;
     double chn1;
     double chn2;
{
  CDIR *cdp= VcutPathCdp(VcutsP(&vCut) + cpIdx);
  if (!cdp) { ErrPrintR("fatal: cut directory #%d not defined\n", cpIdx); }
  else if (CDirNum(cdp)) {
    UF_PrintM("cut entries of '%s':\n", PthShort(CDirPath(cdp)));
    { int i; for (i= 0; i< CDirNum(cdp); i++) {
      if (chn1 > chn2 ||
	  (chn1 <= CDirGateX2(cdp,i) && CDirGateX1(cdp,i) <= chn2)) {
	UF_PrintM("%d_%d  [%f,%f] [%.1f,%.1f] %24s\n",
		  CDirIdx1(cdp,i), CDirIdx2(cdp,i),
		  CDirGateX1(cdp,i), CDirGateX2(cdp,i),
		  CalP(cal, CDirGateX1(cdp,i)), CalP(cal, CDirGateX2(cdp,i)),
		  ctime(&CDirTime(cdp,i))); } } } }
  else {
    UF_PrintE("warning: no cut entries in '%s'\n", PthShort(CDirPath(cdp))); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_CpMrkToGat(gt, mrk)
     GVec_t gt;
     PVec_t mrk;
{
  if (!VecAlloc(gt, VecNum(mrk))) {
    ErrPrintR("fatal: can not copy cut marker\n"); return FALSE; }
  else {
    int g;
    VecNum(gt)= VecNum(mrk);
    for (g= 0; g< VecNum(mrk); g++) {
      GVecX1(gt,g)= PVecX1(mrk,g);
      GVecX2(gt,g)= PVecX2(mrk,g);
      GVecFac(gt,g)= 0.0; }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Cut_CpGatToMrk(mrk, gt)
     PVec_t mrk;
     GVec_t gt;
{
  if (!VecAlloc(mrk, VecNum(gt))) {
    ErrPrintR("fatal: can not copy cut marker\n"); return FALSE; }
  else {
    int g;
    VecNum(mrk)= VecNum(gt);
    for (g= 0; g< VecNum(gt); g++) {
      PVecX1(mrk,g)= GVecX1(gt,g);
      PVecX2(mrk,g)= GVecX2(gt,g); }
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_MrkGet(idx)
     int idx;
{
  Vcut_t *vcp= VcutsC(&vCut) + idx;
  Cut_t cp= VcutCut(vcp);
  int proIdx= VcutProIdx(vcp);
  int gatNum= VecNum(VmrkMrk(VmrksGate(&vMrk)));

  if (!gatNum) { ErrPrintR("fatal: no gate marker defined\n"); return FALSE; }
  else if (CutWeight(cp) == CUT_WFIT && gatNum != 1) {
    ErrPrintR("fatal: exact one pair gate markers required for fit weight\n");
    return FALSE; }
  else if (Mrk_Get(VcutsGate(&vCut), VmrksGate(&vMrk), proIdx,
		   ACT_SUNCALIB) &&
	   Mrk_Get(VcutsBgGate(&vCut), VmrksGateBg(&vMrk), proIdx,
		   ACT_SUNCALIB) &&
	   Cut_MrkAndNot(VcutsBgGate(&vCut), VmrkName(VmrksGateBg(&vMrk)),
			 VcutsGate(&vCut), VmrkName(VmrksGate(&vMrk))) &&
	   Cut_CpMrkToGat(CutGate(cp), VcutsGate(&vCut)) &&
	   Cut_CpMrkToGat(CutBgGate(cp), VcutsBgGate(&vCut))) {

    SerialIdCp(VcutGatSerialId(vcp),
	       VmrkMrkSerialId(VmrksGate(&vMrk)));
    SerialIdCp(VcutBgGatSerialId(vcp),
	       VmrkMrkSerialId(VmrksGateBg(&vMrk)));
    Cut_Changed(idx);

    return TRUE; }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_MrkPut(vcp)
     Vcut_t *vcp;
{
  Cut_t cp= VcutCut(vcp);
  int proIdx= VcutProIdx(vcp);

  if (Cut_CpGatToMrk(VcutsGate(&vCut), CutGate(cp)) &&
      Cut_CpGatToMrk(VcutsBgGate(&vCut), CutBgGate(cp)) &&
      Mrk_Put(VmrksGate(&vMrk), VcutsGate(&vCut), proIdx, ACT_SUNCALIB) &&
      Mrk_Put(VmrksGateBg(&vMrk), VcutsBgGate(&vCut), proIdx, ACT_SUNCALIB)) {
    SerialIdCp(VmrkMrkSerialId(VmrksGate(&vMrk)),
	       VcutGatSerialId(vcp));
    SerialIdCp(VmrkMrkSerialId(VmrksGateBg(&vMrk)),
	       VcutBgGatSerialId(vcp));
    return TRUE; }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
static int Cut_MrkAndNot(aMrk, aName, bMrk, bName)
     PVec_t aMrk;
     char *aName;
     PVec_t bMrk;
     char *bName;
{
  double aRange;
  double bRange;

  aRange= MrkRange(aMrk); bRange= MrkRange(bMrk);

  MrkAndNot(aMrk, bMrk);

  if (MrkRange(aMrk) != aRange) {
    aRange= MrkRange(aMrk);
    UF_PrintE("warning: eliminating overlaps with %s in %s\n",
	      bName, aName); }
  MrkCompress(aMrk); MrkCompress(bMrk);
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_CheckCutPath(cpIdx)
     int cpIdx;
{
  if (!VcutPathCdp(VcutsP(&vCut) + cpIdx)) {
    ErrPrintR("fatal: cut directory #%d not defined\n", cpIdx);
    return FALSE; }
  else { return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCut(cDirIdx, cutIdx)
     int cDirIdx;
     int cutIdx;
{
  /*  printf("Cut_ReadCut\n");*/
  DN_Delay(TRUE);
  if (Cut_ReadHead(cDirIdx, cutIdx)) {
    Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
    Cut_t cp= VcutCut(vcp);
    VcutPath_t *pp= VcutCutPath(vcp);
    char format[SPC_FMTLEN];
    char *file= PthSubstitute(CutFile(cp), VcutPathSpcExpr(pp));

    strcpy(format, VcutPathCutFmt(pp));

    if (!CutReadSpectrum(cp, file, format)) {
      ErrPrintR("fatal: can not read cut spectrum %s'%s\n",
		PthShort(file), format);
      Cut_ResetVcut(cutIdx);
    }
    else {
      Spc_t sp= CutSpc(cp);
      if (sp && SpcNum(sp)) {
	int spcIdx= VcutSpcIdx(vcp);
	Spc_InitCut(spcIdx,
		    PthSubstitute(CutFile(cp), VcutPathSpcExpr(pp)),
		    VcutPathFitExpr(pp),
		    SPC_ODISK);
	SerialIdCp(VcutSpcSerialId(vcp), VspcsSerialId(&vSpc,spcIdx));
	VcutSpcExists(vcp)= TRUE;
	UF_PrintM("spectrum of cut #%d read to buffer #%d\n",
		  cutIdx, spcIdx);
      }
      else {
	VcutSpcExists(vcp)= FALSE;
      }
      DN_Delay(FALSE);
      return TRUE;
    }
  }
  DN_Delay(FALSE);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutName()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryName(cp, &cDirIdx) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutPos()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryPos(cp, &cDirIdx) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutCursor()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryLabel(cp, &cDirIdx) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutFirst()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_FIRST) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutLast()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_LAST) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutNext()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_NEXT) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadCutPrevious()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_PREVIOUS) &&
	  Cut_ReadCut(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
static int Cut_ReadHead(cDirIdx, idx)
     int cDirIdx;
     int idx;
{
  Vcut_t *cp= VcutsC(&vCut) + idx;
  VcutPath_t *pp= VcutCutPath(cp);

  /*  printf("Cut_ReadHead\n");*/
  DN_Delay(TRUE);
  if (!Cut_ResetVcut(idx)) {
    ;
  }
  else if (!CutReadGatesCDir(VcutPathCdp(pp), cDirIdx, VcutCut(cp))) {
    ErrPrintR("fatal: can not read gates");
    Cut_ResetVcut(idx);
  }
  else {
    UF_PrintM("gate of cut #%d read\n", idx);
    SerialIdGet(VcutGatSerialId(cp));
    SerialIdGet(VcutBgGatSerialId(cp));
    Cut_MrkPut(cp);
    VcutPathCutIdx(pp)= cDirIdx;
    Cut_Changed(idx);
    DN_Delay(FALSE);
    return TRUE;
  }
  DN_Delay(FALSE);
  return FALSE;
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadName()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryName(cp, &cDirIdx) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadPos()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryPos(cp, &cDirIdx) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadCursor()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryLabel(cp, &cDirIdx) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
static int Cut_EntryLabel(cp, idx)
     Vcut_t *cp;
     int *idx;
{
#ifndef NO_VD_GRAPHIC
  extern int DCL_GetIndex();
  if (DCL_GetIndex(idx)) { return TRUE; }
  ErrPrintR
    ("fatal: no cut entry identified by cursor in cut directory #%d\n",
     VcutCutPathIdx(cp));
  return FALSE;
#else
  ErrPrintR("fatal: cursor not available\n");
  return FALSE;
#endif
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadFirst()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_FIRST) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadLast()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_LAST) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadNext()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_NEXT) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
int Cut_ReadHeadPrevious()
{
  int idx= VactCutIdx(vAct);
  Vcut_t *cp= VcutsC(&vCut)+idx;
  int cDirIdx;

  return (Cut_EntryIdx(&cDirIdx, VcutCutPathIdx(cp), CUT_PREVIOUS) &&
	  Cut_ReadHead(cDirIdx, idx));
}
/* ---------------------------------------------------------------------- */
static int Cut_Rename(vcp)
     Vcut_t *vcp;
{
  char entry[MAXPATHLEN];
  VcutPath_t *pp= VcutCutPath(vcp);
  CDIR *cdp= VcutPathCdp(pp);
  int idx= VcutSpcIdx(vcp);

  if (VcutSpcExists(vcp) &&
      !SerialIdCmp(VcutSpcSerialId(vcp), VspcsSerialId(&vSpc,idx))) {
    if (!Fio_Close(idx)) { return FALSE; }
    else {
      Cut_t cp= VcutCut(vcp);
      CutCDirEntry(cdp, CutIdx1(cp), CutIdx2(cp), entry);
      Free(VspcsFile(&vSpc,idx));
      VspcsFile(&vSpc,idx)=
	Strdup(PthSubstitute(entry, VcutPathSpcExpr(pp)));
      DN_ChangedSpectrum(idx);
      Fio_FileInit(idx, VcutPathFitExpr(pp)); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_Use()
{
  return (Idx_ReadOne("use cut marker of cut index") &&
	  Cut_MrkPut(VcutsC(&vCut) + VidxSingle(&vIdx)));
}
/* ---------------------------------------------------------------------- */
static int Cut_Rm(cp, cDirIdx, expr)
     Vcut_t *cp;
     int cDirIdx;
     char *expr;
{
  CDIR *cdp= VcutPathCdp(VcutCutPath(cp));
  char entry[MAXPATHLEN];
  char *file;

  CutCDirEntry(cdp, CDirIdx1(cdp,cDirIdx), CDirIdx2(cdp,cDirIdx), entry);

  if (!expr || !*expr) {
    UF_PrintE("fatal: no pattern defined for generation of filename\n");
    return FALSE; }
  else if (!(file= PthSubstitute(entry, expr)) || !*file) {
    ErrPrintR("fatal: can not decode expression '%s' for entry '%s'\n",
	      expr, entry);
    return FALSE; }

  { struct stat buf;
    if (stat(file, &buf) < 0) {extern int errno; errno= 0; return TRUE; }
    else if (buf.st_mode & S_IFDIR) {
      if (rmdir(file) < 0) {
	ErrPrint("error: can not remove directory '%s'\n", PthShort(file));
	return FALSE; }
      else { return TRUE; } }
    else if (unlink(file) < 0) {
      ErrPrint("error: can not unlink file '%s'\n", PthShort(file));
      return FALSE; }
    else { return TRUE; } }
}
/* ---------------------------------------------------------------------- */
int Cut_RmCut()
{
  Vcut_t *cp= VcutsC(&vCut)+VactCutIdx(vAct);
  VcutPath_t *pp= VcutCutPath(cp);
  int cDirIdx;

  if (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
      Cut_EntryName(cp, &cDirIdx) &&
      Cut_Rm(cp, cDirIdx, VcutPathSpcExpr(pp)) &&
      Cut_Rm(cp, cDirIdx, PatGet(VpthPattern(&vPth), PTH_FIT)) &&
      Cut_Rm(cp, cDirIdx, PatGet(VpthPattern(&vPth), PTH_CUTGAT)) &&
      Cut_Rm(cp, cDirIdx, "!f") &&
      CutDeleteCDirEntry(VcutPathCdp(pp), cDirIdx)) {
    DN_ChangedCutPath(VcutCutPathIdx(cp)); return TRUE; }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Cut_RmFit()
{
  Vcut_t *cp= VcutsC(&vCut)+VactCutIdx(vAct);
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryName(cp, &cDirIdx) &&
	  Cut_Rm(cp, cDirIdx, PatGet(VpthPattern(&vPth), PTH_FIT)));
}
/* ---------------------------------------------------------------------- */
int Cut_RmSpectrum()
{
  Vcut_t *cp= VcutsC(&vCut)+VactCutIdx(vAct);
  VcutPath_t *pp= VcutCutPath(cp);
  int cDirIdx;

  return (Cut_CheckCutPath(VcutCutPathIdx(cp)) &&
	  Cut_EntryName(cp, &cDirIdx) &&
	  Cut_Rm(cp, cDirIdx, VcutPathSpcExpr(pp)));
}
/* ---------------------------------------------------------------------- */
int Cut_Scale()
{
  int scale= VcutsScale(&vCut);
  if (Act_Scale(&scale, "cut i/o-scale")) { VcutsScale(&vCut)= scale; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_ShowCut()
{
  UF_PrintM("active cut: %d  cut weight: %s\n",
	    VactCutIdx(vAct), WEIGHT(VcutsWeight(&vCut)));

  { int i; for (i= 0; i< VcutsNum(&vCut); i++) { Cut_ShowCutP(i); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ShowCutP(cutIdx)
     int cutIdx;
{
  Vcut_t *vcp= VcutsC(&vCut)+cutIdx;
  Cut_t cp= VcutCut(vcp);
  UF_PrintM("#%d:   m: %d (p: %d)  d: %d  s: %d",
	    cutIdx, VcutMtxIdx(vcp),
	    VmtxProIdx(VcutsM(&vCut) + VcutMtxIdx(vcp)),
	    VcutCutPathIdx(vcp), VcutSpcIdx(vcp));
  if (CutExists(cp)) {
    char entry[MAXPATHLEN];
    CutCDirEntry(VcutPathCdp(VcutCutPath(vcp)),
		 CutIdx1(cp), CutIdx2(cp), entry);
    UF_PrintM("  w: %s  %s\n", WEIGHT(CutWeight(cp)), PthShort(entry)); }
  else { UF_PrintM("\n"); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_ShowCutPath()
{
  UF_PrintM("cut format: %s\n", VcutsCutFmt(&vCut));

  { int i; for (i= 0; i< VcutsNum(&vCut); i++) { Cut_ShowCutPathP(i); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ShowCutPathP(cutPathIdx)
     int cutPathIdx;
{
  VcutPath_t *pp= VcutsP(&vCut)+cutPathIdx;
  CDIR *cdp= VcutPathCdp(pp);

  if (!cdp) { UF_PrintM("#%d: not defined\n", cutPathIdx); }
  else {
    UF_PrintM("#%d: %s  no: %d  idx: %d  fmt: %s  fit: %s  spc: %s\n",
	      cutPathIdx, PthShort(CDirPath(cdp)),
	      CDirNum(cdp), VcutPathCutIdx(pp),
	      VcutPathCutFmt(pp),
	      VcutPathFitExpr(pp), VcutPathSpcExpr(pp)); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_ShowMtx()
{
  UF_PrintM("matrix format: %s\n", VcutsMtxFmt(&vCut));

  { int i; for (i= 0; i< VcutsNum(&vCut); i++) { Cut_ShowMtxP(i); } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Cut_ShowMtxP(mtxIdx)
     int mtxIdx;
{
  Vmtx_t *mp= VcutsM(&vCut)+mtxIdx;
  SpcFile_t sf= VmtxMtx(mp);

  if (!VmtxMtxExists(mp)) { UF_PrintM("#%d: not defined\n", mtxIdx); }
  else {
    UF_PrintM("#%d: %s'%s\n",
	      mtxIdx, PthShort(SpcFileName(sf)), SpcFileFormat(sf)); }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Cut_WeightGat() { VcutsWeight(&vCut)= CUT_WGAT; return TRUE; }
/* ---------------------------------------------------------------------- */
int Cut_WeightFit() { VcutsWeight(&vCut)= CUT_WFIT; return TRUE; }
/* ---------------------------------------------------------------------- */
int Cut_WriteHead()
{
  return Cut_WriteHeadP(VactCutIdx(vAct));
}
/* ---------------------------------------------------------------------- */
static int Cut_WriteHeadP(cutIdx)
     int cutIdx;
{
  Vcut_t *cp= VcutsC(&vCut) + cutIdx;
  CDIR *cdp= VcutPathCdp(VcutCutPath(cp));
  int cDirIdx;

  if (!Cut_CheckCutPath(VcutCutPathIdx(cp))) { return FALSE; }
  else if (!CutWriteGates(cdp, &cDirIdx, VcutCut(cp))) {
    ErrPrintR("fatal: can not write gate file for cut #%d\n", cutIdx);
    return FALSE; }
  else {
    Cut_Changed(cutIdx);

    if (!Cut_Rename(cp)) { return FALSE; }
    else {
      char entry[MAXPATHLEN];

      CutCDirEntry(cdp, CDirIdx1(cdp,cDirIdx), CDirIdx2(cdp,cDirIdx), entry);
      UF_PrintM("cut #%d written to entry '%s'\n", cutIdx, PthShort(entry));
      return TRUE; } }
}
/* ---------------------------------------------------------------------- */
int Cut_WriteCut()
{
  int cutIdx= VactCutIdx(vAct);

  if (!Cut_WriteHeadP(cutIdx)) {
    return FALSE;
  }
  else {
    Vcut_t *vcp= VcutsC(&vCut) + cutIdx;
    
    if (SerialIdCmp(VcutSpcSerialId(vcp),
		    VspcsSerialId(&vSpc,VcutSpcIdx(vcp)))) {
      UF_PrintE("error: spectrum of cut #%d not available\n",
		cutIdx);
      return FALSE;
    }
    else {
      Cut_t cp= VcutCut(vcp);

      { char entry[MAXPATHLEN];
	char format[SPC_FMTLEN];
	char file[MAXPATHLEN];
	VcutPath_t *pp= VcutCutPath(vcp);
	CDIR *cdp= VcutPathCdp(pp);
	CutCDirEntry(cdp, CutIdx1(cp), CutIdx2(cp), entry);
	strcpy(file, PthSubstitute(entry, VcutPathSpcExpr(pp)));
	strcpy(format, VcutPathCutFmt(pp));
	if (!CutWriteSpectrum(cp, file, format)) {
	  ErrPrintR("fatal: can not write spectrum of cut #%d to %s'%s\n",
		    cutIdx, PthShort(file), format);
	  return FALSE;
	}
	else {
	  return TRUE;
	}
      }
    }
  }
}
/* ---------------------------------------------------------------------- */
