/* $Header: /usr/files/home/Gamma/vs/tv/edit/prg/RCS/VsRecal.c,v 1.1 1993/09/17 14:53:34 vs Exp vs $ */ 

/* ============================== */
/* vs calls to the vsRecal-module */
/* ============================== */

#define _VsRecal_c
				/* library include files */
#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsPath.h>
#include <tv/vsTools.h>

				/* private include files */
#include "VsActiv.h"
#include "VsIndex.h"
#include "VsMark.h"
#include "VsNorm.h"
#include "VsPath.h"
#include "VsSpectra.h"
#include "VsUserFace.h"

#include "VsRecal.h"
#include "VsRecalP.h"

#include "VsError.h"
/* ---------------------------------------------------------------------- */
int Rcl_Append()
{
  if (PTH_TUNDEF ==		/* read vPth.token */
      Pth_TokenRead(PTH_TUNDEF, PTH_RCL,
		    "pair recalibration output file | wc [%s]", PTH_RCL)) {
    UF_Reset();
    return FALSE; }
  else if (!Idx_Read(1, "write pair recalibration of index(s)")) {
    return FALSE; }

  return Pth_LoopIndices(Rcl_AppendExec, Rcl_WriteFinish);
}
/* ---------------------------------------------------------------------- */
static int Rcl_AppendExec(idx, num)
     int idx;
     int num;
{
  return Rcl_WriteIntern(idx, VspcsFile(&vSpc,idx), "a", "appended");
}
/* ---------------------------------------------------------------------- */
int Rcl_Create()
{
  if (!Idx_ReadOne("recalibration reference index") ||
      !Idx_Read(1, "index(s) for recalibration to #%d", VidxSingle(&vIdx))) {
    return FALSE; }
  else {
    return Idx_Loop2(Rcl_CreateExec, Rcl_CreateFinish); }
}
/* ---------------------------------------------------------------------- */
static int Rcl_CreateExec(idx0, idx, num) /* reverse order !!! */
     int idx0;
     int idx;
     int num;
{
  Rcl_t rp= VrclsRcl(&vRcl,idx);
  int typ= (VrclsTyp(&vRcl) != RCL_TPOLYNOM) ? VrclsTyp(&vRcl) : RCL_TSQUARE;
  if (!VspcsExists(&vSpc,idx0)) {
    ErrPrintR("fatal: spectrum #%d not defined\n");
    return IDX_EERROR; }
  else if (!VspcsExists(&vSpc,idx)) { return IDX_ESKIPPED; }
  else if (SpcNum(VspcsSpc(&vSpc,idx0)) != SpcNum(VspcsSpc(&vSpc,idx))) {
    UF_PrintE("warning: spectrum #%d not recalibrated; resolutions differ\n",
	      idx);
    return IDX_ESKIPPED; }
  else if (!Mrk_Get(VrclsMrk(&vRcl), VmrksRcl(&vMrk), idx, ACT_SUNCALIB)) {
    return IDX_EERROR; }
  else if (!Nrm_Update()) { return IDX_EERROR; }
  else if (RclPairCreate(SpcBufReal(SpcSpc(VspcsSpc(&vSpc,idx))),
			 SpcBufReal(SpcSpc(VspcsSpc(&vSpc,idx0))),
			 VnrmsFac(&vNrm,idx), VnrmsFac(&vNrm,idx0),
			 SpcNum(VspcsSpc(&vSpc,idx0)),
			 VrclsMrk(&vRcl), RclPair(rp),typ)) {
    RclTyp(rp)= typ;
    UF_PrintM("pair recalibration relative to #%d created for #%d\n",
	      idx0, idx);

    if (!Rcl_Recalibrate(idx)) {
      ErrPrint("error: incomplete recalibration of spectrum #%d\n", idx); }
    return IDX_EDONE; }
  else {
    ErrPrintR("fatal: can not create pair recalibration #%d\n", idx);
    return IDX_EERROR; }
}
/* ---------------------------------------------------------------------- */
static int Rcl_CreateFinish(num)
     int num;
{
  if (num) { return TRUE; }
  else { UF_PrintE("warning: no pair recalibration created\n"); return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Rcl_Delete()
{
  if (!Idx_Read(1, "delete recalibration of index(s)")) {
    return FALSE; }
  else {
    return Idx_Loop(Rcl_DeleteExec, Rcl_DeleteFinish); }
}
/* ---------------------------------------------------------------------- */
static int Rcl_DeleteExec(idx, num)
     int idx;
     int num;
{
  if (Rcl_Reset(idx)) {
    UF_PrintM("recalibration for #%d deleted", idx); }
  return IDX_EDONE;
}
/* ---------------------------------------------------------------------- */
static int Rcl_DeleteFinish(num)
     int num;
{
  if (num) { return TRUE; } 
  else {
    UF_PrintE("warning: no recalibration found to delete\n");
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Rcl_Init(num)
     int num;
{
  if (NULL == (VrclsP(&vRcl)= (Vrcl_t *)Malloc(num * sizeof(Vrcl_t)))) {
    ErrPrintR("fatal: can not initialize recalibrations\n");
    return FALSE; }
  else {
    int i;
    VrclsNum(&vRcl)= num;
    VrclsTyp(&vRcl)= RCL_TSQUARE;
    for (i= 0; i< num; i++) {
      *VrclsSpcFile(&vRcl,i)= '\0';
      RclInit(VrclsRcl(&vRcl,i)); }
    return VecInit(VrclsMrk(&vRcl), sizeof(Pair_t)); }
}
/* ---------------------------------------------------------------------- */
int Rcl_Read()
{
  char *name= (VrclsTyp(&vRcl) == RCL_TPOLYNOM) ? "polynomial" : "pair";
  if (PTH_TUNDEF ==
      Pth_TokenRead(PTH_TUNDEF, PTH_RCL, "%s recalibration file | wc [%s]",
		    name, PTH_RCL)) {
    UF_Reset();
    return FALSE; }
  else if (!Idx_Read(1, "read %s recalibration for index(s)", name)) {
    return FALSE; }

  VrclsName(&vRcl)= name;

  return Pth_LoopIndices(Rcl_ReadExec, Rcl_ReadFinish);
}
/* ---------------------------------------------------------------------- */
static int Rcl_ReadExec(idx, num)
     int idx;
     int num;
{
  if (!VspcsExists(&vSpc,idx) ||
      (VpthTyp(&vPth) == PTH_TPATH && num)) {
    return IDX_ESKIPPED; }

  if (RclRead(VrclsTyp(&vRcl), VrclsRcl(&vRcl,idx), VpthPath(&vPth))) {
    UF_PrintM("%s recalibration read from '%s' for #%d\n",
	      VrclsName(&vRcl), PthShort(VpthPath(&vPth)), idx);
    if (!Rcl_Recalibrate(idx)) {
      ErrPrint("warning: incomplete recalibration of spectrum #%d\n", idx); }
    return IDX_EDONE; }
  else {			/* read error not fatal */
    ErrPrint("warning: can not read %s recalibration from '%s' for #%d\n",
	     VrclsName(&vRcl), PthShort(VpthPath(&vPth)), idx); }
  return IDX_ESKIPPED;
}
/* ---------------------------------------------------------------------- */
static int Rcl_ReadFinish(num)
     int num;
{
  if (num == 0) {
    UF_PrintE("warning: no %s recalibration read for '%s'\n",
	     VrclsName(&vRcl), VpthToken(&vPth));
    UF_Reset();
    return FALSE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int Rcl_Recalibrate(idx)
     int idx;
{
  SerialIdGet(VspcsSerialId(&vSpc,idx));
  VspcsChange(&vSpc,idx) |= SPC_CRECAL;
  strcpy(VrclsSpcFile(&vRcl,idx), VspcsFile(&vSpc,idx));
  return RclRecalibrate(VspcsSpc(&vSpc,idx), VrclsRcl(&vRcl,idx));
}
/* ---------------------------------------------------------------------- */
static int Rcl_Reset(i)
     int i;
{
  if (i < 0 || i >= VrclsNum(&vRcl)) { return FALSE; }
  else {
    *VrclsSpcFile(&vRcl,i)= '\0';
    RclReset(VrclsRcl(&vRcl,i));
    return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Rcl_List()
{
  if (!Idx_Read(1, "list recalibration for index(s)")) {
    return FALSE; }

  { char *t;
    switch (VrclsTyp(&vRcl)) {
    case RCL_TPOLYNOM:	t= "polynom"; break;
    case RCL_TPAIR:	t= "pair"; break;
    case RCL_TSQUARE:	t= "squared-distance"; break;
    case RCL_TCOM:	t= "center-of-mass"; break; }

    UF_PrintM("default recalibration type is %s\n", t); }

  { int i;
    int rclNum= 0;
    
    for (i= 0; i< VidxNum(&vIdx); i++) {
      int idx= VidxIdx(&vIdx,i);
      if (VspcsExists(&vSpc,idx) &&
	  (VspcsChange(&vSpc,idx) & SPC_CRECAL) &&
	  !strcmp(VspcsFile(&vSpc,idx), VrclsSpcFile(&vRcl,idx))) {
	switch (RclTyp(VrclsRcl(&vRcl,idx))) {
	case RCL_TPOLYNOM:
	  rclNum++;
	  { int j;
	    DVec_t dp= RclPoly(VrclsRcl(&vRcl,idx));

	    UF_PrintM("#%d '%s' polynom recalibration:\n",
		      idx, VspcsFile(&vSpc,idx));
	    for (j= 0; j< VecNum(dp); j++) {
	      UF_PrintM(" %f", VecP(dp)[j]); }
	    UF_PrintM("\n"); }
	  break;
	case RCL_TPAIR:
	case RCL_TSQUARE:
	case RCL_TCOM:
	  rclNum++;
	  UF_PrintM("#%d '%s' pair recalibration:\n",
		    idx, VspcsFile(&vSpc,idx));
	  { int j;
	    RclPairVec_t pp= RclPair(VrclsRcl(&vRcl,idx));
	    for (j= 0; j< VecNum(pp); j++) {
	      UF_PrintM(" (%f, %f)", RclPairVecX1(pp,j), RclPairVecX2(pp,j)); }
	    UF_PrintM("\n"); }
	  break; } } }
    if (!rclNum) {
      UF_PrintM("info: no recalibration found to show\n");
      return FALSE; } }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
int Rcl_TypSetPair() { VrclsTyp(&vRcl)= RCL_TPAIR; return TRUE; }
/* ------------------------------------------------------------------------- */
int Rcl_TypSetSquare() { VrclsTyp(&vRcl)= RCL_TSQUARE; return TRUE; }
/* ------------------------------------------------------------------------- */
int Rcl_TypSetCoM() { VrclsTyp(&vRcl)= RCL_TCOM; return TRUE; }
/* ------------------------------------------------------------------------- */
int Rcl_TypSetPolynom() { VrclsTyp(&vRcl)= RCL_TPOLYNOM; return TRUE; }
/* ------------------------------------------------------------------------- */
int Rcl_Write()
{
  if (PTH_TUNDEF ==		/* read vPth.token */
      Pth_TokenRead(PTH_TUNDEF, PTH_RCL,
		    "pair recalibration output file | wc [%s]", PTH_RCL)) {
    UF_Reset();
    return FALSE; }
  else if (!Idx_Read(1, "write pair recalibration for index(s)")) {
    return FALSE; }

  return Pth_LoopIndices(Rcl_WriteExec, Rcl_WriteFinish);
}
/* ---------------------------------------------------------------------- */
static int Rcl_WriteIntern(idx, title, mode, action)
     int idx;
     char *title;
     char *mode;
     char *action;
{
  Rcl_t rp= VrclsRcl(&vRcl,idx);
  if (!VspcsExists(&vSpc,idx) ||
      !(VspcsChange(&vSpc,idx) & SPC_CRECAL)  ||
      RclTyp(rp) == RCL_TPOLYNOM ||
      strcmp(VspcsFile(&vSpc,idx), VrclsSpcFile(&vRcl,idx))) {
    return IDX_ESKIPPED; }
  else if (VecWrite(VpthPath(&vPth), mode, title, RclPair(rp),
		    ((RclTyp(rp) == RCL_TPAIR) ?
		     RclPairVecPrint : RclPairVecPrintShort))) {
    UF_PrintM("pair recalibration %s to '%s' for #%d\n",
	      action, PthShort(VpthPath(&vPth)), idx);
    return IDX_EDONE; }
  else {
    ErrPrint("warning: pair recalibration not %s to '%s' for #%d\n",
	     action, PthShort(VpthPath(&vPth)), idx);
    return IDX_ESKIPPED; }
}
/* ---------------------------------------------------------------------- */
static int Rcl_WriteExec(idx, num)
     int idx;
     int num;
{
  return Rcl_WriteIntern(idx, "", "w", "written");
}
/* ---------------------------------------------------------------------- */
static int Rcl_WriteFinish(num)
     int num;
{
  if (num == 0) {
    UF_PrintE("warning: no pair recalibration written for '%s'\n",
	      VpthToken(&vPth));
    UF_Reset();
    return FALSE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
