/* $Header: /usr/files/home/Gamma/vs/tv/edit/clients/tv/RCS/VsSearch.c,v 1.2 1994/05/26 19:52:14 vs Exp vs $ */ 

/* =============================== */
/* vs calls to the vsSearch-module */
/* =============================== */

#define _VsSearch_c
				/* library include files*/
#include <tv/vsAlloc.h>
#include <tv/vsTypes.h>
#include <tv/vsCurFit.h>
#include <tv/vsMark.h>
#include <tv/vsPath.h>
#include <tv/vsTools.h>

				/* private include files */
#include "VsActiv.h"
#include "VsCal.h"
#include "VsError.h"
#include "VsFit.h"
#include "VsFitIo.h"
#include "VsIndex.h"
#include "VsMark.h"
#include "VsSpectra.h"
#include "VsUserFace.h"

#include "VdNotify.h"

#include "VsSearch.h"
#include "VsSearchP.h"
/* ---------------------------------------------------------------------- */
int Sea_Init(num)
     int num;
{
  if ((VseasS(&vSea)= (Vsea_t *)Malloc(num * sizeof(Vsea_t)))) {
    int i;
    for (i= 0; i< num; i++) {
      SerialIdInit(VseasSpcSerialId(&vSea,i));
      VseasSea(&vSea,i)= NULL; }

    VseasNum(&vSea)= num;
    VseasProb(&vSea)= 0.9999;

    VecInit(VseasMrk(&vSea), sizeof(Pair_t));

    return TRUE; }
  else {
    ErrPrintR("fatal: can not initialize peak-search-module\n");
    return FALSE; }
}
/* ---------------------------------------------------------------------- */
int Sea_Search()
{
  if (UF_NoArgs()) {
    if (!Idx_SetActSpc()) { return FALSE; } }
  else if (!Idx_Read(1, "search peaks for index(s)")) {
    return FALSE; }

  return Idx_Loop(Sea_SearchExec, Sea_SearchFinish);
}
/* ---------------------------------------------------------------------- */
static int Sea_SearchExec(idx, num)
     int idx;
     int num;
{
  if (!VspcsExists(&vSpc,idx)) { return IDX_ESKIPPED; }
  if (!CalCheck(CAL_WIDTH, VcalsCal(&vCal, idx))) {
    UF_PrintE("error: no width calibration available for spectrum #%d '%s'\n",
	      idx, PthShort(VspcsFile(&vSpc,idx)));
    return IDX_ESKIPPED; }
  else {
    Spc_t spc= VspcsSpc(&vSpc,idx);
    int left, right;

    SeaFreePeakList(VseasSea(&vSea,idx));

    if (Mrk_Get(VseasMrk(&vSea), VmrksFitSea(&vMrk), idx, ACT_SUNCALIB) &&
	Mrk_Consistent(VseasMrk(&vSea), SpcNum(spc)) &&
	VecNum(VseasMrk(&vSea))) {
      left= PVecX1(VseasMrk(&vSea),0);
      right= PVecX2(VseasMrk(&vSea),0); }
    else { left= 0; right= SpcNum(spc) - 1; }

    if ((VseasSea(&vSea,idx)= SeaPeakSearch((void*)SpcBufReal(SpcSpc(spc)),
					   (void*)SpcBufReal(SpcDSpc(spc)),
					   T_DOUBLE, SpcNum(spc),
					   left, right, VseasProb(&vSea),
					   CalW, VcalsCal(&vCal,idx), NULL))) {
      UF_PrintM("%d peaks found in spectrum #%d '%s'\n",
		SeaListNum(VseasSea(&vSea,idx)), idx,
		PthShort(VspcsFile(&vSpc,idx)));
      SerialIdCp(VseasSpcSerialId(&vSea,idx), VspcsSerialId(&vSpc,idx));
      return IDX_EDONE; }
    else {
      SerialIdGet(VseasSpcSerialId(&vSea,idx));
      ErrPrintR("fatal: error during peak search in spectrum #%d '%s'\n",
		idx, PthShort(VspcsFile(&vSpc,idx)));
      return IDX_EERROR; } }
}
/* ---------------------------------------------------------------------- */
static int Sea_SearchFinish(num)
     int num;
{
  if (!num) {
    ErrPrintR("fatal: no spectrum searched for peaks\n"); return FALSE; }
  else { return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Sea_Write()
{
  if (UF_NoArgs()) {
    if (!Idx_SetActSpc()) {
      return FALSE;
    }
  }
  else if (!Idx_Read(1, "write search list for index(s)")) {
    return FALSE;
  }

  return Idx_Loop(Sea_WriteExec, Sea_WriteFinish);
}
/* ---------------------------------------------------------------------- */
static int Sea_WriteExec(idx, num)
     int idx;
     int num;
{
  /*  printf("Sea_WriteExec\n");*/
  if (!VspcsExists(&vSpc,idx) ||
      SerialIdCmp(VseasSpcSerialId(&vSea,idx),
		  VspcsSerialId(&vSpc,idx))) {
    return IDX_ESKIPPED;
  }
  else {
    Vfio_t *vp;
    if (!(vp= Fio_Reopen(idx, "r+")) ||
	!FioWriteSea(VfioFio(vp), VseasSea(&vSea,idx))) {
      ErrPrint("warning: can not write search list of #%d to '%s'\n",
	       idx, PthShort(VfiosFFile(&vFio,idx)));
      return IDX_ESKIPPED;
    }
    else {
      UF_PrintM("search list of #%d written to '%s'\n",
		idx, PthShort(VfioFile(vp)));
      DN_ChangedPeak(idx);
      return IDX_EDONE;
    }
  }
}
/* ---------------------------------------------------------------------- */
static int Sea_WriteFinish(num)
     int num;
{
  if (!num) {
    ErrPrintR("fatal: no search list written\n");
    return FALSE;
  }
  else {
    return TRUE;
  }
}
/* ---------------------------------------------------------------------- */
int Sea_Read()
{
  if (UF_NoArgs()) {
    if (!Idx_SetActSpc()) { return FALSE; } }
  else if (!Idx_Read(1, "read search list for index(s)")) {
    return FALSE; }

  return Idx_Loop(Sea_ReadExec, Sea_ReadFinish);
}
/* ---------------------------------------------------------------------- */
int Sea_ReadExec(idx, num)
     int idx;
     int num;
{
  /*  printf("Sea_ReadExec\n");*/
  if (VspcsExists(&vSpc,idx)) {
    Vfio_t *vp;
    if (!(vp= Fio_Reopen(idx, "r")) ||
	!FioReadSea(VfioFio(vp), &VseasSea(&vSea,idx))) {
      ErrPrint("warning: can not read search list for #%d\n", idx);
      SerialIdGet(VseasSpcSerialId(&vSea,idx));
      return IDX_ESKIPPED;
    }
    else if (VseasSea(&vSea,idx)) {
      SerialIdCp(VseasSpcSerialId(&vSea,idx), VspcsSerialId(&vSpc,idx));
      UF_PrintM("search list of %d peaks for #%d read\n",
		SeaListNum(VseasSea(&vSea,idx)), idx);
      return IDX_EDONE;
    }
  }
  return IDX_ESKIPPED;
}
/* ---------------------------------------------------------------------- */
static int Sea_ReadFinish(num)
     int num;
{
  if (!num) {
    ErrPrintR("fatal: no search list read\n");
    return FALSE; }
  else { return TRUE; }
}
/* ---------------------------------------------------------------------- */
int Sea_Prob()
{
  double value;
  double precision;

  if (UF_NoArgs()) {
    UF_PrintM("acceptance limit for peak search probability: %.2f %%\n",
	      VseasProb(&vSea) * 100.0);
    return TRUE; }
  else if (UF_RealRead(FALSE, &value, &precision, "probability")) {
    value= MIN(100.0, value) * 0.01;
    value= ABS(value);
    VseasProb(&vSea)= value;
    return TRUE; }
  else { return FALSE; }
}
/* ---------------------------------------------------------------------- */
